#include <zls.h>

//======================================================================
//
// User problem parameter initialisation/default values routine
// Called before the user's problem config file is read to set
// all problem parameters to sensible defaults in case the init.conf
// is faultly or even missing
//
//  The fields of the problem structure are defined by the user in
//  zls_problem.h and can be any variables desired by the user's routines.
//  The problem data is stored by reference in gLevels[0] and shared by all 
//  other levels, and can be accessed by any zone at need (see zlsp_UserZoneUpdate
//  and zlsp_InitAtmosphere for examples.
//
//======================================================================

int zlsp_InitProblem    ( ProblemPtr   prob){
  int err = noErr;
  
  if (prob != NULL) {

    
    prob->gD0   =  1.0;
    prob->gP0   =  2.5;
    prob->gv0   = -0.5;
    
    prob->gD1   =  2.0;
    prob->gP1   =  2.5;
    prob->gv1   =  0.5;
    
    prob->gDeltaY  =  0.25; // slab half-height, ie +/- DeltaY
    prob->gNoise   =  0.01;
    prob->gUseFreq    =  1;
    prob->gFreq    =  0.01;
    
  }
  return (err);
}

//======================================================================
// standard allocation and destructor for problem strucutres, called
// by the ZLS system at the right time, and should not need to be
// modified in any way by the user.  Here to keep all problem code together
//
// The only exception is if levels are destroyed during a run and the user
// needs to deallocate any dynamic memory set up in zlsp_InitProblem
// or in zlsp_SetUpProblemFromConfigFile.  If this is the case then 
// the user needs to deallocate the memory in zlsp_DisposeProblem
// before  free(prob) is called.
//
// Currently the level list is fixed for the entire run and this is not
// necessary, but it will be in AMR.  Will be called by zls_DisposeLevel
// which doesn't exist yet. :)
//======================================================================

int zlsp_NewProblem(ProblemPtr * prob){
  int err = noErr;
  
  *prob = (ProblemPtr)malloc(sizeof(Problem));
  
  if (*prob == NULL) {
    err = memErr;
#if (ZLS_DEBUG >0)
    zlsio_ErrPrintf( " ERROR: zlsp_NewProblem allocation failed %d \n",err);
#endif
  }
  
  return (err);
}

int zlsp_DisposeProblem(ProblemPtr prob){
  int err = noErr;
  
  if ( prob != NULL ) {
    free(prob);
  }
  
  return (err);
}


int zlsp_InitZoneData( ZonePtr z);
int zlsp_InitZoneData( ZonePtr z){

  int err = noErr;
    
  if (z != NULL){
    
    WHD zoneW = z->zoneW;    // get zone physical variables
    
    Integer level = z->zoneLevel;
    
    ProblemPtr prob = gLevels[level].levelProblem;  // get the user's problem parameters

#ifdef SYM2D
    Integer    i,  j;   
    Integer nti  = z->zoneGrid.nt[kXDim];
    Integer ntj  = z->zoneGrid.nt[kYDim];
#endif SYM2D
#ifdef SYM3D
    Integer    i,  j, k;   
    Integer nti  = z->zoneGrid.nt[kXDim];
    Integer ntj  = z->zoneGrid.nt[kYDim];
    Integer ntk  = z->zoneGrid.nt[kZDim];
#endif SYM3D
    
    Integer seed = -123*z->zoneID;
    zls_ran1_(&seed);
//======================================================================
// Atmosphere parameters
//======================================================================
    
    Real d0  = prob->gD0;
    Real P0  = prob->gP0;
    Real v0  = prob->gv0;
    
    Real n0  = prob->gNoise;
    Integer useFreq = prob->gUseFreq;
    Real f0 = prob->gFreq;
    
    Real dv  = 0.0;
    Real dvx  = 0.0;
    Real dvy  = 0.0;
#ifdef SYM3D
    Real dvz  = 0.0;
#endif SYM3D
    
    ACoordPtr xx  = z->zoneGrid.x;
    ACoordPtr yx  = z->zoneGrid.y;
#ifdef SYM3D
    ACoordPtr zx  = z->zoneGrid.z;
#endif SYM3D
    Real xc = 0.0;
    Real yc = 0.0;
#ifdef SYM3D
    Real zc = 0.0;
#endif SYM3D
 
//======================================================================
//
// Put on grid first to fill everywhere, replaced later by
// slab in some or all locations
//
//======================================================================
#ifdef SYM2D
    
    if ( useFreq == 1 ) {
      for(j = 0; j < ntj; j++){
       yc      = yx[j].c;
       for(i = 0; i< nti; i++){
          xc      = xx[i].c;
          
          zoneW.d [j][i]      =  d0;
          zoneW.P [j][i]      =  P0;
          
          dvx  = n0*sin(f0*twopi*xc);
          dvy  = n0*cos(f0*twopi*xc);
          zoneW.vx[j][i]      =  v0+dvx;
          zoneW.vy[j][i]      =  dvy;
          
        } //i
      } //j
    } else { // noise
      for(j = 0; j < ntj; j++){
       yc      = yx[j].c;
       for(i = 0; i< nti; i++){
          xc      = xx[i].c;
          
          zoneW.d [j][i]      =  d0;
          zoneW.P [j][i]      =  P0;
          
          dv  = n0*(zls_ran1_(&seed)-0.5);
          zoneW.vx[j][i]      =  v0+dv;
          dv  = n0*(zls_ran1_(&seed)-0.5);
          zoneW.vy[j][i]      =  dv;
          
        } //i
      } //j
    }
    
#endif SYM2D
#ifdef SYM3D
    
    if ( useFreq == 1 ) {
      for(k = 0; k < ntk; k++){
       zc      = zx[k].c;
      for(j = 0; j < ntj; j++){
       yc      = yx[j].c;
       for(i = 0; i< nti; i++){
          xc      = xx[i].c;
          
          zoneW.d [k][j][i]      =  d0;
          zoneW.P [k][j][i]      =  P0;
          
          dvx  = n0*sin(f0*twopi*xc);
          dvy  = n0*cos(f0*twopi*xc);
          dvz  = n0*cos(f0*twopi*yc);
          zoneW.vx[k][j][i]      =  v0+dvx;
          zoneW.vy[k][j][i]      =  dvy;
          zoneW.vz[k][j][i]      =  dvz;
          
        } //i
      } //j
      } //k
  } else { // noise
      for(k = 0; k < ntk; k++){
      for(j = 0; j < ntj; j++){
       for(i = 0; i< nti; i++){
          
          zoneW.d [k][j][i]      =  d0;
          zoneW.P [k][j][i]      =  P0;
          
          dv  = n0*(zls_ran1_(&seed)-0.5);
          zoneW.vx[k][j][i]      =  v0+dv;
          dv  = n0*(zls_ran1_(&seed)-0.5);
          zoneW.vy[k][j][i]      =  dv;
          dv  = n0*(zls_ran1_(&seed)-0.5);
          zoneW.vz[k][j][i]      =  dv;
          
        } //i
      } //j
      } //k
    }
    
#endif SYM3D
//======================================================================
// Slab parameters
//======================================================================
    
    Real d1  = prob->gD1;
    Real P1  = prob->gP1;
    Real v1  = prob->gv1;
    Real dY  = prob->gDeltaY;
    
#ifdef SYM2D
    if ( useFreq == 1 ) {
      for(j = 0; j < ntj; j++){
        yc      = yx[j].c;
         if ( zls_abs(yc) < dY ) {
          for(i = 0; i< nti; i++){
            xc      = xx[i].c;
          
          zoneW.d [j][i]      =  d1;
          zoneW.P [j][i]      =  P1;
          
          dvx  = n0*sin(f0*twopi*xc);
          dvy  = n0*cos(f0*twopi*xc);
          zoneW.vx[j][i]      =  v1+dvx;
          zoneW.vy[j][i]      =  dvy;
          
          } //i
        }
      } //j
    } else { // noise
      for(j = 0; j < ntj; j++){
        yc      = yx[j].c;
         if ( zls_abs(yc) < dY ) {
          for(i = 0; i< nti; i++){
          
          zoneW.d [j][i]      =  d1;
          zoneW.P [j][i]      =  P1;
          
          dv  = n0*(zls_ran1_(&seed)-0.5);
          zoneW.vx[j][i]      =  v1+dv;
          dv  = n0*(zls_ran1_(&seed)-0.5);
          zoneW.vy[j][i]      =  dv;
          
          } //i
        }
      } //j
    }
#endif SYM2D
        
#ifdef SYM3D
    if ( useFreq == 1 ) {
      for(k = 0; k < ntk; k++){
       zc      = zx[k].c;
      for(j = 0; j < ntj; j++){
        yc      = yx[j].c;
         if ( zls_abs(yc) < dY ) {
          for(i = 0; i< nti; i++){
            xc      = xx[i].c;
          
          zoneW.d [k][j][i]      =  d1;
          zoneW.P [k][j][i]      =  P1;
          
          dvx  = n0*sin(f0*twopi*xc);
          dvy  = n0*cos(f0*twopi*xc);
          dvz  = n0*cos(f0*twopi*yc);
          zoneW.vx[k][j][i]      =  v1+dvx;
          zoneW.vy[k][j][i]      =  dvy;
          zoneW.vz[k][j][i]      =  dvz;
          
          } //i
        }
      } //j
    } //k
    
    } else { // noise
      for(k = 0; k < ntk; k++){
       zc      = zx[k].c;
      for(j = 0; j < ntj; j++){
        yc      = yx[j].c;
         if ( zls_abs(yc) < dY ) {
          for(i = 0; i< nti; i++){
          
          zoneW.d [k][j][i]      =  d1;
          zoneW.P [k][j][i]      =  P1;
          
          dv  = n0*(zls_ran1_(&seed)-0.5);
          zoneW.vx[k][j][i]      =  v1+dv;
          dv  = n0*(zls_ran1_(&seed)-0.5);
          zoneW.vy[k][j][i]      =  dv;
          dv  = n0*(zls_ran1_(&seed)-0.5);
          zoneW.vz[k][j][i]      =  dv;
          
          } //i
        }
      } //j
      } // k
    }
#endif SYM3D
  } // z != NULL
  
  return(err);
}


//======================================================================
//  Standard level initialsation routine, loops through all zones of
//   a level. Called by ZLS after all parameters, coordinates, physics
//   etc have been set up.  Sets the gravity field arrays before the
//   calling zlsp_InitZoneData.
//======================================================================

int zlsp_InitLevelZones(Integer level){
  
  
  int err = noErr;
  Integer zone;
  ZonePtr  z;
  
#if (ZLS_VERBOSE >0)
#ifdef MPI
    if ( gMPIrank == 0){
#endif MPI
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " Level %d User Zones Initialisation: \n", level);
      zlsio_Printf( "%s\n",  DIVIDERLINE );
#ifdef MPI
    }
#endif MPI
#endif
  
  for ( zone = 0; zone <gLevels[level].nZones; zone++){
    z =  gLevels[level].levelZones[zone];
	  if (z != NULL){
#ifdef MPI	
		  if (z->zoneRankID == gMPIrank) {
#endif MPI
			  if ( z->zoneType == kAlgorithmZoneType ){ // only process appropriate zones, ie skip image zones
				  
#if (ZLS_DEBUG>0)
				  zlsio_ErrPrintf( " Initialising Zone: %2.2d:%2.2d \n", level, zone );
#endif
				  
				  err = zlsp_InitZoneData( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsp_InitZoneData Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
          
				  err = zls_SetZoneConstantBCValues( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zls_SetZoneConstantBCValues Failed: %d,  %d \n",zone, err );
					  return(err);
				  }

#ifdef MEANSPLITSWEEPS
#ifdef SYM2D
          err =  zlsa_Synch2DData(z);
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsa_Synch2DSymmetricSweeps Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
#endif SYM2D
#endif   MEANSPLITSWEEPS   
          
         
			  } // z->zoneType == kAlgorithmZoneType 
#ifdef MPI			
		  }
#endif MPI
	  } else { err = zonErr; }
  } // zone
  
  
  return(err);
  
}

//======================================================================
// Optional routine to allow overriding the hard wired zlsp_InitProblem
// parameters from a config file, specified in zls.conf - usually init.conf
// Called by ZLS at the right time(tm)
//======================================================================

int   zlsp_SetUpProblemFromConfigFile(char * confFile){
  
//======================================================================
// Open the config file confFile, and parse for any user definded
// key,value pairs to set problem parameters (not grid initialisation)
// Calls zlsp_InitProblem, and before zlsp_InitLevelZones
//======================================================================
  
  int err = noErr;
  struct ccl_t  config;
  const char    *str1;
  
  Integer level;
  
  ProblemPtr  prob = NULL;
  err = zlsp_NewProblem(&prob);
  
  if ( prob != NULL ) {
    
    config.comment_char = '#';
    config.sep_char     = '=';
    config.str_char     = '"';
    
    if (ccl_parse(&config, confFile)==noErr) {
      
#if (ZLS_VERBOSE > 0) 
      zlsio_Printf( "\n%s\n", DIVIDERLINE );
#endif
      zlsio_Printf( " FYRIS Hydro Problem Initialised from %s \n", confFile);
#if (ZLS_VERBOSE > 0) 
      zlsio_Printf(  "%s\n", DIVIDERLINE );
#endif
      
// Default Values
      
      err = zlsp_InitProblem(prob);
      
//======================================================================
// The user is free to define any key strings and interpret the 
// returned values in any way.  Some problems are simple (see the SOD setup)
// this one is more complex as it sets up a multi-component model with many
// different types of parameters.
//======================================================================

//======================================================================
// Atmosphere parameters
//======================================================================
      {
//
// WIERD C NOTE:
// brace sets up a local scope for temp variables to help reduce brief 
// named variables colliding with any used later in the routine. ie Th 
// could be used here and later in the cluster parameter setup without
// interfering with eachother
//
        
        Real P0 = prob->gP0;
        if (( str1 = ccl_get( &config, "Atmos_Pressure"))){
          P0 = atof( str1);
#if (ZLS_DEBUG>0)
          zlsio_Printf( "   %s = %s\n", "Atmos_Pressure", str1);
#endif
        }
        prob->gP0   = P0;
        
        Real D0 = prob->gD0;
        if (( str1 = ccl_get( &config, "Atmos_Density"))){
          D0 = atof( str1);
#if (ZLS_DEBUG>0)
          zlsio_Printf( "   %s = %s\n", "Atmos_Density", str1);
#endif
        }
        prob->gD0   = D0;

        
        Real v0 = prob->gv0;
        if (( str1 = ccl_get( &config, "Atmos_Velocity"))){
          v0 = atof( str1);
#if (ZLS_DEBUG>0)
          zlsio_Printf( "   %s = %s\n", "Atmos_Velocity", str1);
#endif
        }
        prob->gv0   = v0;

        
      }
      
      zlsio_Printf("%s\n",  DIVIDERLINE );
      zlsio_Printf( " Atmosphere Parameters:\n");
      zlsio_Printf("   d        %g \n", prob->gD0);    
      zlsio_Printf("   P        %g \n", prob->gP0); 
      zlsio_Printf("   vx       %g \n", prob->gv0); 
      zlsio_Printf("%s\n",  DIVIDERLINE );
      
      
//======================================================================
// slab parameters
//======================================================================
      {
        
        Real P1 = prob->gP1;
        if (( str1 = ccl_get( &config, "Slab_Pressure"))){
          P1 = atof( str1);
#if (ZLS_DEBUG>0)
          zlsio_Printf( "   %s = %s\n", "Slab_Pressure", str1);
#endif
        }
        prob->gP1   = P1;
        
        Real D1 = prob->gD1;
        if (( str1 = ccl_get( &config, "Slab_Density"))){
          D1 = atof( str1);
#if (ZLS_DEBUG>0)
          zlsio_Printf( "   %s = %s\n", "Slab_Density", str1);
#endif
        }
        prob->gD1   = D1;

        
        Real v1 = prob->gv1;
        if (( str1 = ccl_get( &config, "Slab_Velocity"))){
          v1 = atof( str1);
#if (ZLS_DEBUG>0)
          zlsio_Printf( "   %s = %s\n", "Slab_Velocity", str1);
#endif
        }
        prob->gv1   = v1;

        Real t1 = prob->gDeltaY;
        if (( str1 = ccl_get( &config, "Slab_Thickness"))){
          t1 = atof( str1);
#if (ZLS_DEBUG>0)
          zlsio_Printf( "   %s = %s\n", "Slab_Thickness", str1);
#endif
        }
        prob->gDeltaY   = t1;

        
      }
        Integer useFreq = prob->gUseFreq;
        if (( str1 = ccl_get( &config, "Use_Frequency"))){

          useFreq = atoi( str1);
            
          if (strcmp(str1, "no")==0) {
            useFreq = 0;
          }
          if (strcmp(str1, "No")==0) {
            useFreq = 0;
          }
          if (strcmp(str1, "N")==0) {
            useFreq = 0;
          }
          if (strcmp(str1, "n")==0) {
            useFreq = 0;
          }
          if (strcmp(str1, "none")==0) {
            useFreq = 0;
          }
          if (strcmp(str1, "None")==0) {
            useFreq = 0;
          }
          
          if (strcmp(str1, "yes")==0) {
            useFreq = 1;
          }
          if (strcmp(str1, "Yes")==0) {
            useFreq = 1;
          }
          if (strcmp(str1, "Y")==0) {
            useFreq = 1;
          }
          if (strcmp(str1, "y")==0) {
            useFreq = 1;
          }
          
          useFreq = (useFreq != 0)? 1: 0;

#if (ZLS_DEBUG>0)
          zlsio_Printf( "   %s = %s\n", "Use_Frequency", str1);
#endif
        }
        prob->gUseFreq   = useFreq;
        Real f0 = prob->gFreq;
        if (( str1 = ccl_get( &config, "Frequency"))){
          f0 = atof( str1);
#if (ZLS_DEBUG>0)
          zlsio_Printf( "   %s = %s\n", "Frequency", str1);
#endif
        }
        prob->gFreq   = f0;

        Real n0 = prob->gNoise;
        if (( str1 = ccl_get( &config, "Noise"))){
          n0 = atof( str1);
#if (ZLS_DEBUG>0)
          zlsio_Printf( "   %s = %s\n", "Noise", str1);
#endif
        }
        prob->gNoise   = n0;


      zlsio_Printf( " Slab Parameters:\n");
      zlsio_Printf("   d        %g \n", prob->gD1);    
      zlsio_Printf("   P        %g \n", prob->gP1); 
      zlsio_Printf("   vx       %g \n", prob->gv1); 
      zlsio_Printf("   h        %g \n", prob->gDeltaY); 
      
      if ( useFreq == 1 ) {
        zlsio_Printf("%s\n",  DIVIDERLINE );
        zlsio_Printf("   Frequency    %g \n", prob->gFreq);    
        zlsio_Printf("   Amplitude    %g \n", prob->gNoise);    
        zlsio_Printf("%s\n",  DIVIDERLINE );
      } else {
        zlsio_Printf("%s\n",  DIVIDERLINE );
        zlsio_Printf("   Noise    %g \n", prob->gNoise);    
        zlsio_Printf("%s\n",  DIVIDERLINE );
      }
      
//======================================================================
// This code sets the problem into the level structures, mulit-zone
// multi-level or simple single structures 
//======================================================================
      
      
      gLevels[0].levelProblem = prob; // register the problem with all levels from 0 down.      
      if (gNLevels>1){
        for (level = 1; level < gNLevels; level++){
          gLevels[level].levelProblem = gLevels[0].levelProblem;
        }
      }
      
//======================================================================
// Now the level contents can be set - standard calls here dont' change
//======================================================================
      
      for (level = 0; level < gNLevels; level++){
        
#ifdef GRAVITY
        err =  zlsg_InitLevelGravityField( level ); // precompute the potential of to the gravity field over the level
        if (err != noErr ) {
          zlsio_ErrPrintf( " ERROR: zlsg_InitLevelGravityField: Level Field Failed: %d,  %d \n",level, err );
          return(err);
        }
#endif GRAVITY
#ifdef NEQIONISATION
        err =  zlsi_InitLevelIonisationState( level ); 
        if (err != noErr ) {
          zlsio_ErrPrintf( " ERROR: zlsi_InitLevelIonisationState: Level Failed: %d,  %d \n",level, err );
          return(err);
        }
#endif NEQIONISATION
        
        err = zlsp_InitLevelZones( level ); // last: user's levelProblem, initialised top down
        
      }
      
      if (gNLevels>1){
        for (level = (gNLevels-1); level > 0; level--){
          err = zlsbc_FillSuperLevelFromLevel(level);
        }
      }
      
#if (ZLS_VERBOSE>0)
      zlsio_Printf("%s\n",  DIVIDERLINE );
#else
      zlsio_LogPrintf("%s\n",  DIVIDERLINE );
#endif
      
      ccl_release(&config);
      
    }// config  
    
  } else {
    err = memErr;
    zlsio_ErrPrintf( " ERROR: Unable to allocate level 0 ProblemPtr \n");
    return (err);
  }
  
  return (err);
  
}

//======================================================================
// Routine for 'User; boundary cells set in grid.conf not used here
// but often used for jet inlet models
//======================================================================
#ifdef ZLS_USER_BOUNDARIES
int  zlsp_CalcUserZoneBCs ( ZonePtr z, int direction, int bndryKind ){
  return (noErr);
}
#endif ZLS_USER_BOUNDARIES


//======================================================================
// Routine for modifying anywhere in the grid in any way as part of
// a custom problem model. Not restricted to boundary cells, here used for 
// wind bubble/mass loss bubble energy injection and mass injection
// called after all other boundary conditions have been computed.
// Use sparingly, if not careful this routine can impact heavily on 
// the performance of the simulation.
//======================================================================
// 
// Normally deactivated (ie not even checked for calling) unless the macro
// ZLS_USER_UPDATE_ROUTINE is defined in zls_defs.  If not defined then
// this routine is not even callable - and this is the default setting.
// Also normally provided as an empty routine.
//

#ifdef ZLS_USER_UPDATE_ROUTINE

int  zlsp_UserZoneUpdate ( ZonePtr z  ){
  
  Integer err = noErr;
  ProblemPtr prob = NULL;
  PhysicsPtr phys = NULL;
  
  // A different type of BC, not the grid edges as all but setting cell values
  // on the grid.  Uses zoneW and is based on the zone data init routine rather than
  // the grid edge code of the basic boundary value routines.
  
  if ( z != NULL ){
    
    Integer level = z->zoneLevel;
    prob = gLevels[level].levelProblem;  // get the user's problem parameters
    phys = gLevels[level].levelPhysics;  // get the global eos parameters
    
    if ( level == (gNLevels-1)) {
      
#ifdef COOLING
    PhysicsPtr phys = gLevels[0].levelPhysics;  // get the global Physics parameters
    CoolingPtr cp   = NULL;  // get the coolng parameters
    cp   = &(phys->coolParameters);  // get the coolng parameters
#endif
     
	   Real t      = zlst_GlobalTime;

     WHD zoneW    = z->zoneW;    

        
    } // level
    
  } else { err = memErr; } // z
  
  return (err);
}
#endif ZLS_USER_UPDATE_ROUTINE

int zlsp_KHZoneOutput( zFile f, ZonePtr z){ 
  int err = noErr;
  
  IOSettingsPtr io = gLevels[ z->zoneLevel].levelIO;
  
  if (( f!= NULL) &&(io != NULL) && (z != NULL)){
    
    WHD         zoneW =  z->zoneW;
    
    Real t      = zlst_GlobalTime;

    
    Real d      = 0.0;
    Real vy     = 0.0;
    Real sumKE      = 0.0;
    Real KEy      = 0.0;
    
#ifdef SYM2D
    Integer i, j;
    Integer is   = z->zoneGrid.ns[kXDim];
    Integer ie   = z->zoneGrid.ne[kXDim];
    Integer js   = z->zoneGrid.ns[kYDim];
    Integer je   = z->zoneGrid.ne[kYDim];
    //     
    for(j = js; j< je; j++){
      for(i = is; i< ie; i++){
        vy  = zoneW.vy [j][i];
        d   = zoneW.d [j][i];
        KEy = 0.5*d*vy*vy;
        sumKE += KEy;
        
      } //i
    } //j
#endif SYM2D
    
#ifdef SYM3D
    Integer i, j, k;
    Integer is   = z->zoneGrid.ns[kXDim];
    Integer ie   = z->zoneGrid.ne[kXDim];
    Integer js   = z->zoneGrid.ns[kYDim];
    Integer je   = z->zoneGrid.ne[kYDim];
    Integer ks   = z->zoneGrid.ns[kZDim];
    Integer ke   = z->zoneGrid.ne[kZDim];
    //     
    for(k = ks; k< ke; k++){
    for(j = js; j< je; j++){
      for(i = is; i< ie; i++){
        vy  = zoneW.vy [k][j][i];
        d   = zoneW.d  [k][j][i];
        KEy = 0.5*d*vy*vy;
        sumKE += KEy;
        
      } //i
    } //j
    } //k
#endif SYM3D
    
#ifdef MPI  
  Real sum;    
  MPI_Allreduce( &sumKE, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  sumKE = sum/gLevels[ z->zoneLevel].levelNCells;
  if ( gMPIrank == 0) {
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    zlsio_ZFPrintf( f, " Y-Kinetic Energy     = \t %#12.6g \t at \t %#12.6g \n",sumKE, t);
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    zlsio_UserLogPrintf("%#12.6g \t %#12.6g \n",t, sumKE);
  }
#else
    sumKE = sumKE/z->zoneNCells;
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    zlsio_ZFPrintf( f, " Y-Kinetic Energy     = \t %#12.6g \t at \t %#12.6g \t %d\n",sumKE, t, z->zoneID);
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    zlsio_UserLogPrintf("%#12.6g \t %#12.6g \n",t, sumKE);
#endif
    
  }
  
  return(err);
}

