#include <zls.h>

//======================================================================
//
// User problem parameter initialisation/default values routine
// Called before the user's problem config file is read to set
// all problem parameters to sensible defaults in case the init.conf
// is faultly or even missing
//
//  The fields of the problem structure are defined by the user in
//  zls_problem.h and can be any variables desired by the user's routines.
//  The problem data is stored by reference in gLevels[0] and shared by all 
//  other levels, and can be accessed by any zone at need (see zlsp_UserZoneUpdate
//  and zlsp_InitAtmosphere for examples.
//
//======================================================================

int zlsp_InitProblem    ( ProblemPtr   prob){
  int err = noErr;
  
  if (prob != NULL) {
  
    prob->gX  = 0.0;
    prob->gY  = 0.0;
    
    prob->gVX = 0.0;
    prob->gVY = 0.0;
   
  }
  
  return (err);

}

//======================================================================
// standard allocation and destructor for problem strucutres, called
// by the ZLS system at the right time, and should not need to be
// modified in any way by the user.  Here to keep all problem code together
//
// The only exception is if levels are destroyed during a run and the user
// needs to deallocate any dynamic memory set up in zlsp_InitProblem
// or in zlsp_SetUpProblemFromConfigFile.  If this is the case then 
// the user needs to deallocate the memory in zlsp_DisposeProblem
// before  free(prob) is called.
//
// Currently the level list is fixed for the entire run and this is not
// necessary, but it will be in AMR.  Will be called by zls_DisposeLevel
// which doesn't exist yet. :)
//======================================================================

int zlsp_NewProblem(ProblemPtr * prob){
  int err = noErr;
  
  *prob = (ProblemPtr)malloc(sizeof(Problem));
  
  if (*prob == NULL) {
    err = memErr;
#if (ZLS_DEBUG >0)
    zlsio_ErrPrintf( " ERROR: zlsp_NewProblem allocation failed %d \n",err);
#endif
  }
  
  return (err);
}

int zlsp_DisposeProblem(ProblemPtr prob){
  int err = noErr;
  
  if ( prob != NULL ) {
    free(prob);
  }
  
  return (err);
}

int zlsp_InitZone2DVortex( ZonePtr z );
int zlsp_InitZone2DVortex( ZonePtr z ){
  
//======================================================================
//  Initialise a zone system variables (ie W), usually self contained, 
//  here the initialisation is a bit messy so subroutines are called from
//  here.  When called all the variables for the zone are initialised and 
//  ready to go, grid coordinates, problem vars etc.
//======================================================================
  
  int err = noErr;
  
  if (z != NULL){
    
    Integer level = z->zoneLevel;
    
    if ( level == (gNLevels-1)) {
      
      WHD zoneW = z->zoneW; 
      ProblemPtr prob = gLevels[level].levelProblem;  // get the user's problem parameters
      
      ACoordPtr xx  = z->zoneGrid.x;
      ACoordPtr yx  = z->zoneGrid.y;
      
      Integer nti   = z->zoneGrid.nt[kXDim];
      Integer ntj   = z->zoneGrid.nt[kYDim];
      Integer i, j;
      
      Real fi,  fj; // axis radial components
      
      
      Real r; 
      Real vx      = 0.0;
      Real vy      = 0.0;
      
	    Real vPhi0, vPhi1, vPhi2;
      Real    P0,    P1,    P2;
      
      Real    d0 = 1.0;
      
      Real    vx0 = prob->gVX;
      Real    vy0 = prob->gVY;
      
      Real    x0 = prob->gX;
      Real    y0 = prob->gY;
      
      
      Real x      = 0.0;
      Real y      = 0.0;
      Real dx     = 0.0;
      Real dy     = 0.0;
      
      Real radius = 0.0;
      Real f      = 1.0;
      
      Real t0, t1;
      
      Real c0 = 9.0 - 4.0*zls_Log(0.2);
      Real c1 = 3.0 + 4.0*zls_Log(2.0);
      
      for(j = 0; j < ntj; j++){
        y      = yx[j].c-y0;
        dy     = yx[j].dx;
        for(i = 0; i< nti; i++){
          x      = xx[i].c-x0;
          dx     = xx[i].dx;
          
          zoneW.d[j][i] = d0;
          
          radius = zls_CvtCoordsToProjRadius2D( &fi, &fj, x, y);
          
          r      = radius;
          
          t0    = 5.0*r;
          
          vPhi0 = t0;
          vPhi1 = 2.0-t0;
          vPhi2 = 0.0;
          
          t0 = 5.0*r;
          t1 = t0*t0*0.5;
          
          P0 = 5.0+t1;
          P1 = c0 + t1 -20.0*r + 4.0*zls_Log(r);
          P2 = c1;
          
          if ( radius < 0.20) { 
            zoneW.P[j][i] = P0;
            vx = -fj*vPhi0;
            vy =  fi*vPhi0;
            zoneW.vx[j][i] = vx + vx0;
            zoneW.vy[j][i] = vy + vy0;
          }  else if ( radius < 0.40) {
            zoneW.P[j][i] = P1;
            vx = -fj*vPhi1;
            vy =  fi*vPhi1;
            zoneW.vx[j][i] = vx + vx0;
            zoneW.vy[j][i] = vy + vy0;
          }  else {
            zoneW.P[j][i] = P2;
            zoneW.vx[j][i] = vx0;
            zoneW.vy[j][i] = vy0;
          }
                    
        } //i
      } //j
      
      
    } // end wind on lowest level
    
  } else { err = memErr; } //z
  
  return(err);
}

//======================================================================
//  Standard zone initialsation routine
//======================================================================

int zlsp_InitZoneData( ZonePtr z ){
  
//======================================================================
//  Initialise a zone system variables (ie W), usually self contained, 
//  here the initialisation is a bit messy so subroutines are called from
//  here.  When called all the variables for the zone are initialised and 
//  ready to go, grid coordinates, problem vars etc.
//======================================================================
  
  int err = noErr;
  
  if (z != NULL){
    
    Integer level = z->zoneLevel;
    
    if ( level > 1) {
      err = zlsbc_FillZoneFromSuperZone( z, kInterpBCsFromSuper );
    } else {
      err = zlsp_InitZone2DVortex(z);   
    }
        
  } else { err = memErr; }//z
  
  return(err);
}

//======================================================================
//  Standard level initialsation routine, loops through all zones of
//   a level. Called by ZLS after all parameters, coordinates, physics
//   etc have been set up.  Sets the gravity field arrays before the
//   calling zlsp_InitZoneData.
//======================================================================

int zlsp_InitLevelZones(Integer level){
  
  
  int err = noErr;
  Integer zone;
  ZonePtr  z;
  
#if (ZLS_VERBOSE >0)
#ifdef MPI
    if ( gMPIrank == 0){
#endif MPI
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " Level %d User Zones Initialisation: \n", level);
      zlsio_Printf( "%s\n",  DIVIDERLINE );
#ifdef MPI
    }
#endif MPI
#endif
  
  for ( zone = 0; zone <gLevels[level].nZones; zone++){
    z =  gLevels[level].levelZones[zone];
	  if (z != NULL){
#ifdef MPI	
		  if (z->zoneRankID == gMPIrank) {
#endif MPI
			  if ( z->zoneType == kAlgorithmZoneType ){ // only process appropriate zones, ie skip image zones
				  
#if (ZLS_DEBUG>0)
				  zlsio_ErrPrintf( " Initialising Zone: %2.2d:%2.2d \n", level, zone );
#endif
				  
				  err = zlsp_InitZoneData( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsp_InitZoneData Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
          
				  err = zls_SetZoneConstantBCValues( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zls_SetZoneConstantBCValues Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
#ifdef MEANSPLITSWEEPS
#ifdef SYM2D
          err =  zlsa_Synch2DData(z);
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsa_Synch2DSymmetricSweeps Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
#endif SYM2D
#endif   MEANSPLITSWEEPS   
          
          
			  } // z->zoneType == kAlgorithmZoneType 
#ifdef MPI			
		  }
#endif MPI
	  } else { err = zonErr; }
  } // zone
  
  return(err);
  
}


//======================================================================
// Optional routine to allow overriding the hard wired zlsp_InitProblem
// parameters from a config file, specified in zls.conf - usually init.conf
// Called by ZLS at the right time(tm)
//======================================================================

int   zlsp_SetUpProblemFromConfigFile(char * confFile){
  
//======================================================================
// Open the config file confFile, and parse for any user definded
// key,value pairs to set problem parameters (not grid initialisation)
// Calls zlsp_InitProblem, and before zlsp_InitLevelZones
//======================================================================
  
  int err = noErr;
  struct ccl_t  config;
  const char    *str1;
  
  Integer level;
  
  ProblemPtr  prob = NULL;
  err = zlsp_NewProblem(&prob);
  
  if ( prob != NULL ) {
    
    config.comment_char = '#';
    config.sep_char     = '=';
    config.str_char     = '"';
    
    if (ccl_parse(&config, confFile)==noErr) {
      
#if (ZLS_VERBOSE > 0) 
      zlsio_Printf( "\n%s\n", DIVIDERLINE );
#endif
      zlsio_Printf( " FYRIS Hydro Problem Initialised from %s \n", confFile);
#if (ZLS_VERBOSE > 0) 
      zlsio_Printf(  "%s\n", DIVIDERLINE );
#endif
      
// Default Values
      
      err = zlsp_InitProblem(prob);
      
//======================================================================
// The user is free to define any key strings and interpret the 
// returned values in any way.  Some problems are simple (see the SOD setup)
// this one is more complex as it sets up a multi-component model with many
// different types of parameters.
//======================================================================
          
      
      Real x = prob->gX;
      if (( str1 = ccl_get( &config, "X0"))){
        x = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "X0", str1);
#endif
      }
      prob->gX = x;
      
      Real y = prob->gY;
      if (( str1 = ccl_get( &config, "Y0"))){
        y = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Y0", str1);
#endif
      }
      prob->gY = y;
            
      Real vx = prob->gVX;
      if (( str1 = ccl_get( &config, "VX0"))){
        vx = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "VX0", str1);
#endif
      }
      prob->gVX = vx;
      
      Real vy = prob->gVY;
      if (( str1 = ccl_get( &config, "VY0"))){
        vy = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "VY0", str1);
#endif
      }
      prob->gVY = vy;
            
      
#if (ZLS_VERBOSE>0)
      
      zlsio_Printf("%s\n",  DIVIDERLINE );
      zlsio_Printf(" Gresho Vortex Parameters:\n");
      zlsio_Printf("    (x0, y0)   :  (%g, %g) \n", prob->gX, prob->gY);    
      zlsio_Printf("    (vx0, vy0) :  (%g, %g)  \n", prob->gVX, prob->gVY);       
      zlsio_Printf("%s\n",  DIVIDERLINE );
      
#else
// log even if not verbose
      
      zlsio_LogPrintf("%s\n",  DIVIDERLINE );
      zlsio_LogPrintf(" Gresho Vortex Parameters:\n");
      zlsio_LogPrintf("    (x0, y0)   :  (%g, %g) \n", prob->gX, prob->gY);    
      zlsio_LogPrintf("    (vx0, vy0) :  (%g, %g)  \n", prob->gVX, prob->gVY);       
      zlsio_LogPrintf("%s\n",  DIVIDERLINE );

#endif 
      
      
//======================================================================
// This code sets the problem into the level structures, mulit-zone
// multi-level or simple single structures 
//======================================================================
      
      
      gLevels[0].levelProblem = prob; // register the problem with all levels from 0 down.      
      if (gNLevels>1){
        for (level = 1; level < gNLevels; level++){
          gLevels[level].levelProblem = gLevels[0].levelProblem;
        }
      }
      
//======================================================================
// Now the level contents can be set - standard calls here dont' change
//======================================================================
      
      for (level = 0; level < gNLevels; level++){
        
#ifdef GRAVITY
        err =  zlsg_InitLevelGravityField( level ); // precompute the potential of to the gravity field over the level
        if (err != noErr ) {
          zlsio_ErrPrintf( " ERROR: zlsg_InitLevelGravityField: Level Field Failed: %d,  %d \n",level, err );
          return(err);
        }
#endif GRAVITY
#ifdef NEQIONISATION
        err =  zlsi_InitLevelIonisationState( level ); 
        if (err != noErr ) {
          zlsio_ErrPrintf( " ERROR: zlsi_InitLevelIonisationState: Level Failed: %d,  %d \n",level, err );
          return(err);
        }
#endif NEQIONISATION
        
        err = zlsp_InitLevelZones( level ); // last: user's levelProblem, initialised top down
        
      }
      
      if (gNLevels>1){
        for (level = (gNLevels-1); level > 0; level--){
          err = zlsbc_FillSuperLevelFromLevel(level);
        }
      }
      
#if (ZLS_VERBOSE>0)
      zlsio_Printf("%s\n",  DIVIDERLINE );
#else
      zlsio_LogPrintf("%s\n",  DIVIDERLINE );
#endif
      
      ccl_release(&config);
      
    }// config  
    
  } else {
    err = memErr;
    zlsio_ErrPrintf( " ERROR: Unable to allocate level 0 ProblemPtr \n");
    return (err);
  }
  
  return (err);
  
}

//======================================================================
// Routine for 'User; boundary cells set in grid.conf not used here
// but often used for jet inlet models
//======================================================================
#ifdef ZLS_USER_BOUNDARIES
int  zlsp_CalcUserZoneBCs ( ZonePtr z, int direction, int bndryKind ){
  return (noErr);
}
#endif ZLS_USER_BOUNDARIES


//======================================================================
// Routine for modifying anywhere in the grid in any way as part of
// a custom problem model. Not restricted to boundary cells, here used for 
// wind bubble/mass loss bubble energy injection and mass injection
// called after all other boundary conditions have been computed.
// Use sparingly, if not careful this routine can impact heavily on 
// the performance of the simulation.
//======================================================================
// 
// Normally deactivated (ie not even checked for calling) unless the macro
// ZLS_USER_UPDATE_ROUTINE is defined in zls_defs.  If not defined then
// this routine is not even callable - and this is the default setting.
// Also normally provided as an empty routine.
//

#ifdef ZLS_USER_UPDATE_ROUTINE

int  zlsp_UserZoneUpdate ( ZonePtr z  ){
  
  Integer err = noErr;
  ProblemPtr prob = NULL;
  PhysicsPtr phys = NULL;
  
  // A different type of BC, not the grid edges as all but setting cell values
  // on the grid.  Uses zoneW and is based on the zone data init routine rather than
  // the grid edge code of the basic boundary value routines.
  
  if ( z != NULL ){
    
    Integer level = z->zoneLevel;
    prob = gLevels[level].levelProblem;  // get the user's problem parameters
    phys = gLevels[level].levelPhysics;  // get the global eos parameters
    
  } else { err = memErr; } // z
  
  return (err);
}
#endif ZLS_USER_UPDATE_ROUTINE


// Custom user TEXT io routine.

int zlsp_2DVortexZoneOutput( zFile f, ZonePtr z){ 
  int err = noErr;
  
  IOSettingsPtr io = gLevels[ z->zoneLevel].levelIO;
  
  if (( f!= NULL) &&(io != NULL) && (z != NULL)){
    
    WHD               zoneW;
    err     = zlsp_GetZoneWHD(&zoneW, z);
    ProblemPtr prob = gLevels[0].levelProblem;  // get the user's problem parameters
   
    zlsio_ZFPrintf( f, " Zone ID = %d\n", z->zoneID);
    Real t      = zlst_GlobalTime;
    zlsio_ZFPrintf( f, " Time = %d\n", t);

#if (ZLS_VERBOSE > 1)
    zls_ZPrintZoneCoords( f,  z);
#endif (ZLS_VERBOSE > 1)
    
    ACoordPtr xx  = z->zoneGrid.x;
    ACoordPtr yx  = z->zoneGrid.y;
    
    Integer i, j;
    
    Real x      = 0.0;
    Real y      = 0.0;
    Real dx     = 0.0;
    Real dy     = 0.0;
    
    Real vxp = prob->gVX;
    Real vyp = prob->gVY;
    Real xp = prob->gX;
    Real yp = prob->gY;
    
    Real x0     = xp+(t*vxp);
    Real y0     = yp+(t*vyp);
    
    Real d      = 0.0;
    Real P      = 0.0;
    Real vx     = 0.0;
    Real vy     = 0.0;
    Real vm     = 0.0;
    Real KE     = 1.0;

    Real rv      = 0.0;
    Real vx0     = 0.0;
    Real vy0     = 0.0;
    Real vx1     = 0.0;
    Real vy1     = 0.0;
    Real vx2     = 0.0;
    Real vy2     = 0.0;
    Real vx3     = 0.0;
    Real vy3     = 0.0;
    
    Real vorticity    = 0.0;
    Real dvydx  = 0.0;
    Real dvxdy  = 0.0;
    Real idx    = 0.0;
    Real idy    = 0.0;

    Real vorticity0    = 0.0;
    Real dvydx0  = 0.0;
    Real dvxdy0  = 0.0;
    
    Real gVtx    = 0.0;
    Real gVorticity0    = 0.0;
    Real gVorticity1    = 0.0;
    Real gVorticity2    = 0.0;
    Real sumVtx    = 0.0;
    Real sumN      = 0.0;
    
    Real sumKE     = 0.0;
    Real sumNKE    = 0.0;
    
    Real sumD      = 0.0;
    Real sumND     = 0.0;
    
    Real sumP      = 0.0;
    Real sumNP     = 0.0;
    
    Real gD  = 1.0;
    Real gKE = 1.0;
    Real gP  = 1.0;

    Real t0, t1;
    
    Real c0 = 9.0 - 4.0*zls_Log(0.2);
    Real c1 = 3.0 + 4.0*zls_Log(2.0);
 
    Real vPhi0, vPhi1, vPhi2;
    Real P0, P1, P2;
    
    Real radius = 0.0;
    Real fi, fj, fr;
    
    Integer is   = z->zoneGrid.ns[kXDim];
    Integer ie   = z->zoneGrid.ne[kXDim];
    Integer js   = z->zoneGrid.ns[kYDim];
    Integer je   = z->zoneGrid.ne[kYDim];
    
     zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
     zlsio_ZFPrintf( f, " x \t y \t r \t d \t Gd \t P \t GP \t vm \t vx \t vy \t dvydx \t dvxdy \t vtx \t Gvtx \t KE \t GKE \n");
     zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
     
     for(j = js; j< je; j++){
      y      = yx[j].c - y0;
       for(i = is; i< ie; i++){
         x      = xx[i].c - x0;
         
         if (( x > -0.5) && ( x <= 0.5) &&
             ( y > -0.5 ) && ( y <= 0.5)){
            // keep to box around vortex

         radius = zls_CvtCoordsToProjRadius2D(&fi, &fj, x, y);
         
         rv = zls_CvtCoordsToProjRadius2D(&fi, &fj, xx[i+1].c - x0, y);
         t0    = 5.0*rv;
         vPhi0 = t0;
         vPhi1 = 2.0-t0;
         vPhi2 = 0.0;
         if ( rv < 0.2) {
           vx0  = -fj*vPhi0 + vxp;
           vy0  =  fi*vPhi0 + vyp;
         }  else if ( rv < 0.4) {
           vx0  = -fj*vPhi1 + vxp;
           vy0  =  fi*vPhi1 + vyp;
         }  else {
           vx0  = vxp;
           vy0  = vyp;
         }
         
         rv = zls_CvtCoordsToProjRadius2D(&fi, &fj, xx[i-1].c - x0, y);
         t0    = 5.0*rv;
         vPhi0 = t0;
         vPhi1 = 2.0-t0;
         vPhi2 = 0.0;
         if ( rv < 0.2) {
           vx1  = -fj*vPhi0 + vxp;
           vy1  =  fi*vPhi0 + vyp;
         }  else if ( rv < 0.4) {
           vx1  = -fj*vPhi1 + vxp;
           vy1  =  fi*vPhi1 + vyp;
         }  else {
           vx1  = vxp;
           vy1  = vyp;
         }

         rv = zls_CvtCoordsToProjRadius2D(&fi, &fj, x, yx[j+1].c - y0);
         t0    = 5.0*rv;
         vPhi0 = t0;
         vPhi1 = 2.0-t0;
         vPhi2 = 0.0;
         if ( rv < 0.2) {
           vx2  = -fj*vPhi0 + vxp;
           vy2  =  fi*vPhi0 + vyp;
         }  else if ( rv < 0.4) {
           vx2  = -fj*vPhi1 + vxp;
           vy2  =  fi*vPhi1 + vyp;
         }  else {
           vx2  = vxp;
           vy2  = vyp;
         }
         
         rv = zls_CvtCoordsToProjRadius2D(&fi, &fj, x, yx[j-1].c - y0);
         t0    = 5.0*rv;
         vPhi0 = t0;
         vPhi1 = 2.0-t0;
         vPhi2 = 0.0;
         if ( rv < 0.2) {
           vx3  = -fj*vPhi0 + vxp;
           vy3  =  fi*vPhi0 + vyp;
         }  else if ( rv < 0.4) {
           vx3  = -fj*vPhi1 + vxp;
           vy3  =  fi*vPhi1 + vyp;
         }  else {
           vx3  = vxp;
           vy3  = vyp;
         }
         
         d = zoneW.d [j][i];
         P = zoneW.P [j][i];
         
         vx = zoneW.vx [j][i];
         vy = zoneW.vy [j][i];
         vm = zls_sqrt(vx*vx+vy*vy);
         
         KE = 0.5*d*(vx*vx+vy*vy);
         
         dx = xx[i].dx;
         dy = yx[j].dx;
         
         idx = xx[i].idx;
         idy = yx[j].idx;
         
         dvydx0 = 0.5*( vy0 - vy1)*idx;
         dvxdy0 = 0.5*( vx2 - vx3)*idy;
         vorticity0 = (dvydx0 - dvxdy0);

         dvydx = 0.5*( zoneW.vy [j][i+1] - zoneW.vy [j][i-1])*idx;
         dvxdy = 0.5*( zoneW.vx [j+1][i] - zoneW.vx [j-1][i])*idy;
         vorticity = (dvydx - dvxdy);
         
         gVtx = 10.0;
         
         gVorticity0 = 10.0;
         gVorticity1 = (2.0/radius - 10.0);
         gVorticity2 = 0.0;
         
         t0    = 5.0*radius;
         vPhi0 = t0;
         vPhi1 = 2.0-t0;
         vPhi2 = 0.0;
         
         t0 = 5.0*radius;
         t1 = t0*t0*0.5;
         P0 = 5.0+t1;
         P1 = c0 + t1 -20.0*radius + 4.0*zls_Log(radius);
         P2 = c1;
         
         gP = 5.0;
         
         if ( radius < 0.2) {
           gVtx = gVorticity0;
           gKE  = 0.5*vPhi0*vPhi0 + 0.5*(vxp*vxp+vyp*vyp);
           gP   = P0;
         }  else if ( radius < 0.4) {
           gVtx = gVorticity1;
           gKE  = 0.5*vPhi1*vPhi1 + 0.5*(vxp*vxp+vyp*vyp);
           gP   = P1;
         }  else {
           gVtx = gVorticity2;
           gKE  = 0.5*vPhi2*vPhi2 + 0.5*(vxp*vxp+vyp*vyp);
           gP   = P2;
         }
                  
         zlsio_ZFPrintf( f, " %#12.6g \t %#12.6g \t %#12.6g \t",  x,  y,  radius);
         zlsio_ZFPrintf( f, " %#12.6g \t %#12.6g \t %#12.6g \t %#12.6g \t",  d, gD,  P, gP);
         zlsio_ZFPrintf( f, " %#12.6g \t %#12.6g \t %#12.6g \t", vm,  vx, vy);
         zlsio_ZFPrintf( f, " %#12.6g \t %#12.6g \t %#12.6g \t %#12.6g \t", dvydx,     dvxdy,     vorticity,  vorticity0);
         zlsio_ZFPrintf( f, " %#12.6g \t %#12.6g  \n", KE, gKE);
         
         sumVtx += zls_abs((vorticity-vorticity0));
         sumN   += zls_abs(vorticity0);
         
         sumKE  += zls_abs(KE);
         sumNKE += zls_abs(gKE);
         
         sumD   += zls_abs((d-gD));
         sumND  += zls_abs(gD);
         
         sumP   += zls_abs((P-gP));
         sumNP  += zls_abs(gP);
         } // keep to box around vortex
       } //i
    } //j
    
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    zlsio_ZFPrintf( f, "mean L1 Vtx     = \t %#12.6g\%\n",100.0*sumVtx/sumN);
    zlsio_ZFPrintf( f, "Rel. Total KE   = \t %#12.6g\%\n",100.0*zls_abs(((sumKE-sumNKE)/sumNKE)));
    zlsio_ZFPrintf( f, "mean L1 Den     = \t %#12.6g\%\n",100.0*sumD/sumND);
    zlsio_ZFPrintf( f, "mean L1 Pre     = \t %#12.6g\%\n",100.0*sumP/sumNP);
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    zlsp_ReleaseZoneWHD(zoneW);
    
  }
  
  return(err);
}
