#include <zls.h>

//======================================================================
//
// User problem parameter initialisation/default values routine
// Called before the user's problem config file is read to set
// all problem parameters to sensible defaults in case the init.conf
// is faultly or even missing
//
//  The fields of the problem structure are defined by the user in
//  zls_problem.h and can be any variables desired by the user's routines.
//  The problem data is stored by reference in gLevels[0] and shared by all 
//  other levels, and can be accessed by any zone at need (see zlsp_UserZoneUpdate
//  and zlsp_InitAtmosphere for examples.
//
//======================================================================

int zlsp_InitProblem    ( ProblemPtr   prob){
  int err = noErr;
  
  if (prob != NULL) {
    
    prob->gA0       = 0.01;
    prob->gF0       = 3.0;

    prob->gHeight   = 0.5;
    prob->gDeltaD   = 2.0;
    prob->gAttwood  = (2.0-1.0)/(2.0+1.0);
    
    prob->gD0   = 1.0;
    prob->gP0   = 1.0;
    prob->gC2   = 1.0;
    
    prob->gAntiAlias     = kSmooth; 
    prob->gSmooth        = 2.0 ; 

  }
  return (err);
}

//======================================================================
// standard allocation and destructor for problem structures, called
// by the ZLS system at the right time, and should not need to be
// modified in any way by the user.  Here to keep all problem code together
//
// The only exception is if levels are destroyed during a run and the user
// needs to deallocate any dynamic memory set up in zlsp_InitProblem
// or in zlsp_SetUpProblemFromConfigFile.  If this is the case then 
// the user needs to deallocate the memory in zlsp_DisposeProblem
// before  free(prob) is called.
//
// Currently the level list is fixed for the entire run and this is not
// necessary, but it will be in AMR.  Will be called by zls_DisposeLevel
// which doesn't exist yet. :)
//======================================================================

int zlsp_NewProblem(ProblemPtr * prob){
  int err = noErr;
  
  *prob = (ProblemPtr)malloc(sizeof(Problem));
  
  if (*prob == NULL) {
    err = memErr;
#if (ZLS_DEBUG >0)
    zlsio_ErrPrintf( " ERROR: zlsp_NewProblem allocation failed %d \n",err);
#endif
  }
  
  return (err);
}

int zlsp_DisposeProblem(ProblemPtr prob){
  int err = noErr;
  
  if ( prob != NULL ) {
    free(prob);
  }
  
  return (err);
}

//======================================================================


//======================================================================
//
// A static pointer to hold the fluctuations between zone initialisation
// to reduce memory and file IO repetition.
// allocated by the level loop outside the InitZoneData routine
//
//======================================================================

static   QPtr qStructure = NULL;

//======================================================================
// Routine to set up a hydrostatic atmosphere in a gravity field or 
// a uniform/cloudy medium if no gravity field is present
//======================================================================

int zlsp_InitAtmosphere( ZonePtr z);
int zlsp_InitAtmosphere( ZonePtr z){

//======================================================================
// Routine to set up a hydrostatic atmosphere in a gravity field or 
// a uniform/cloudy medium if no gravity field is present
//======================================================================
  
  int err = noErr;
    
  if (z != NULL){
    
#ifdef GRAVITY
    Field zoneF ;
    zls_GetZoneField(&zoneF, z); // nothing is allocated here, so no dispose is needed later
#endif GRAVITY
    
    WHD zoneW = z->zoneW;    // get zone physical variables
    
    Integer level = z->zoneLevel;
    
    ProblemPtr prob = gLevels[level].levelProblem;  // get the user's problem parameters
    PhysicsPtr phys = gLevels[0].levelPhysics;  // get the global Physics parameters
    
#ifdef SYM2D 
    Integer    i,  j;   
    Integer nti  = z->zoneGrid.nt[kXDim];
    Integer ntj  = z->zoneGrid.nt[kYDim];
#endif SYM2D 

    Real h = prob->gHeight;

    Real  fi   =  prob->gF0;
    Real  ampl =  prob->gA0;
    
    Real C2  = prob->gC2;
    Real C22 = C2/prob->gDeltaD;
    
    Real PScale1 = 1.0/C2 ; 
    Real PScale2 = 1.0/C22; 
    
#ifdef GRAVITY
    PotentialPtr gp = NULL; 
    gp       = &(phys->gravPotential);  // get the potential 
    Real gH  =  zlsg_phi(h, gp);
    Real phi1, phi2;
#endif GRAVITY
    
    Real  rho1   =  prob->gD0;
    Real  rho2   =  prob->gD0*prob->gDeltaD;
    
    Real d1,d2; 
    
    Integer nSmooth = 1024;
    Real          s =  prob->gSmooth;
    
#ifdef SYM2D
    
    
    for(j = 0; j < ntj; j++){
      
      Real y =  z->zoneGrid.y[j].c;
      Real dy =  z->zoneGrid.y[j].dx;
      
      for(i = 0; i< nti; i++){
      
        Real x =  z->zoneGrid.x[i].c;
        Real dx =  z->zoneGrid.x[i].dx;
        
#ifdef GRAVITY
        phi1             = -PScale1*(zoneF.p[j][i].a-gH);
        phi2             = -PScale2*(zoneF.p[j][i].a-gH);
        d1               = rho1*exp(phi1);
        d2               = rho2*exp(phi2);
#else
        d1               = rho1;
        d2               = rho2;
#endif !GRAVITY
        
        if ( prob->gAntiAlias != kSmooth){
        
           Real c = ampl*cos(twopi*fi*x)+h;
           if ( y < c ){
            zoneW.d [j][i]  = d1;
            zoneW.P [j][i]  = d1*C2;
           } else {
            zoneW.d [j][i]  = d2;
            zoneW.P [j][i]  = d2*C22;
           }
          
        } else {
          
          if ( (z->zoneGrid.y[j].l) >  (h+ampl+ampl) ){
            
            zoneW.d [j][i]  = d2;
            zoneW.P [j][i]  = d2*C22;
            
          } else if ( (z->zoneGrid.y[j].r) < (h-ampl-ampl) ) {
            
            zoneW.d [j][i]  = d1;
            zoneW.P [j][i]  = d1*C2;
            
          } else {
            
            Real f = zls_SmoothCos2D(ampl, fi, 0.0, h, x, y, dx, dy, nSmooth, s ); //fraction below curve
            
            zoneW.d [j][i]  =  d1*f + (1.0-f)*d2;
            zoneW.P [j][i]  =  d1*C2*f + (1.0-f)*d2*C22;
            
          }
          
        }
        
        zoneW.vx[j][i]      = 0.0;
        zoneW.vy[j][i]      = 0.0;
        
      } //i
    } //j
    
    
#endif SYM2D 
    
    
  } // z != NULL
  
  return(err);
}


//======================================================================
//  Standard zone initialsation routine
//======================================================================

int zlsp_InitZoneData( ZonePtr z ){
  
//======================================================================
//  Initialise a zone system variables (ie W), usually self contained, 
//  here the initialisation is a bit messy so subroutines are called from
//  here.  When called all the variables for the zone are initialised and 
//  ready to go, grid coordinates, problem vars etc.
//======================================================================
  
  int err = noErr;
  
  if (z != NULL){
    
    Integer level = z->zoneLevel;
    
    if ( level == 0) {
      err = zlsp_InitAtmosphere(z);
    } else {
      err = zlsbc_FillZoneFromSuperZone( z, kInterpBCsFromSuper);
    }
	    
    
  } else { err = memErr; }//z
  
  
  return(err);
}


//======================================================================
//  Standard level initialsation routine, loops through all zones of
//   a level. Called by ZLS after all parameters, coordinates, physics
//   etc have been set up.  Sets the gravity field arrays before the
//   calling zlsp_InitZoneData.
//======================================================================

int zlsp_InitLevelZones(Integer level){
  
  
  int err = noErr;
  Integer zone;
  ZonePtr  z;
  
#if (ZLS_VERBOSE >0)
#ifdef MPI
    if ( gMPIrank == 0){
#endif MPI
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " Level %d User Zones Initialisation: \n", level);
      zlsio_Printf( "%s\n",  DIVIDERLINE );
#ifdef MPI
    }
#endif MPI
#endif

  qStructure = NULL;
  
  for ( zone = 0; zone <gLevels[level].nZones; zone++){
    z =  gLevels[level].levelZones[zone];
	  if (z != NULL){
#ifdef MPI	
		  if (z->zoneRankID == gMPIrank) {
#endif MPI
			  if ( z->zoneType == kAlgorithmZoneType ){ // only process appropriate zones, ie skip image zones
				  
#if (ZLS_DEBUG>0)
				  zlsio_ErrPrintf( " Initialising Zone: %2.2d:%2.2d \n", level, zone );
#endif
				  
				  err = zlsp_InitZoneData( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsp_InitZoneData Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
          
				  err = zls_SetZoneConstantBCValues( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zls_SetZoneConstantBCValues Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
#ifdef MEANSPLITSWEEPS
#ifdef SYM2D
          err =  zlsa_Synch2DData(z);
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsa_Synch2DSymmetricSweeps Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
#endif SYM2D
#endif   MEANSPLITSWEEPS   
          
          
			  } // z->zoneType == kAlgorithmZoneType 
#ifdef MPI			
		  }
#endif MPI
	  } else { err = zonErr; }
  } // zone
  
  zls_DisposeQuantity(qStructure);
  
  
  return(err);
  
}

//======================================================================
// Optional routine to allow overriding the hard wired zlsp_InitProblem
// parameters from a config file, specified in zls.conf - usually init.conf
// Called by ZLS at the right time(tm)
//======================================================================

int   zlsp_SetUpProblemFromConfigFile(char * confFile){
  
//======================================================================
// Open the config file confFile, and parse for any user definded
// key,value pairs to set problem parameters (not grid initialisation)
// Calls zlsp_InitProblem, and before zlsp_InitLevelZones
//======================================================================
  
  int err = noErr;
  struct ccl_t  config;
  const char    *str1;
  
  Integer level;
  PhysicsPtr phys = gLevels[0].levelPhysics;  // get the global Physics parameters
  
#ifdef VARIABLE_EOS  
  CoolingPtr cp   = NULL; 
#ifdef COOLING
  cp   = &(phys->coolParameters);  // get the coolng parameters
#endif COOLING
#endif VARIABLE_EOS  

  PotentialPtr gp;
#ifdef GRAVITY
   gp   = &(phys->gravPotential);  // get the gravity parameters
#endif GRAVITY

//  Real        mu  = k_mu;
  
  ProblemPtr  prob = NULL;
  err = zlsp_NewProblem(&prob);
  
  if ( prob != NULL ) {
    
    config.comment_char = '#';
    config.sep_char     = '=';
    config.str_char     = '"';
    
    if (ccl_parse(&config, confFile)==noErr) {
      
#if (ZLS_VERBOSE > 0) 
      zlsio_Printf( "\n%s\n", DIVIDERLINE );
#endif
      zlsio_Printf( " FYRIS Hydro Problem Initialised from %s \n", confFile);
#if (ZLS_VERBOSE > 0) 
      zlsio_Printf(  "%s\n", DIVIDERLINE );
#endif
      
// Default Values
      
      err = zlsp_InitProblem(prob);
      
//======================================================================
// The user is free to define any key strings and interpret the 
// returned values in any way.  Some problems are simple (see the SOD setup)
// this one is more complex as it sets up a multi-component model with many
// different types of parameters.
//======================================================================
      
//======================================================================
// atmosphere parameters
//======================================================================
      {
//
// WIERD C NOTE:
// brace sets up a local scope for temp variables to help reduce brief 
// named variables colliding with any used later in the routine. ie Th 
// could be used here and later in the cluster parameter setup without
// interfering with eachother
//
        
        Real P = prob->gP0;
        if (( str1 = ccl_get( &config, "P0"))){
          P = atof( str1);
#if (ZLS_DEBUG>0)
          zlsio_Printf( "   %s = %s\n", "P0", str1);
#endif
        }
        prob->gP0 = P;
        
        Real d = prob->gD0;
        if (( str1 = ccl_get( &config, "D0"))){
          d = atof( str1);
#if (ZLS_DEBUG>0)
          zlsio_Printf( "   %s = %s\n", "D0", str1);
#endif
        }
        prob->gD0 = d;
        prob->gC2 = P/d;
        
      }
      
      Real del = prob->gDeltaD;
      if (( str1 = ccl_get( &config, "Delta"))){
        del = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Delta", str1);
#endif
      }
      prob->gDeltaD   = del;
      prob->gAttwood  =  ( del-1.0)/(del+1.0);
      
      Real h = prob->gHeight;
      if (( str1 = ccl_get( &config, "Z0"))){
        h = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Z0", str1);
#endif
      }
      prob->gHeight = h;
            
      Real ampl = prob->gA0;
      if (( str1 = ccl_get( &config, "A0"))){
          ampl = atof( str1);
#if (ZLS_DEBUG>0)
          zlsio_Printf( "   %s = %s\n", "A0", str1);
#endif
      }
      prob->gA0 = ampl;
      
      Real f = prob->gF0;
      if (( str1 = ccl_get( &config, "F0"))){
          f = atof( str1);
#if (ZLS_DEBUG>0)
          zlsio_Printf( "   %s = %s\n", "F0", str1);
#endif
      }
      prob->gF0 = f;
      
      
      Integer aa = prob->gAntiAlias;
      if (( str1 = ccl_get( &config, "Anti-Alias"))){
        aa = atoi( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Anti-Alias", str1);
#endif
      }
      if ( aa == 0) {
        prob->gAntiAlias = kDontSmooth;
      } else {
        prob->gAntiAlias = kSmooth;
      
      Real s = prob->gSmooth;
      if (( str1 = ccl_get( &config, "Smooth_Cells"))){
        s = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Smooth_Cells", str1);
#endif
      }
      prob->gSmooth = s;
      }
      
      
      zlsio_Printf("%s\n",  DIVIDERLINE );
      zlsio_Printf( " RT Parameters:\n");
      zlsio_Printf("    H       %#6.2g \n", prob->gHeight);    
      zlsio_Printf("    DelD    %#6.2g \n", prob->gDeltaD);    
      zlsio_Printf("    Attwood %#6.2g \n", prob->gAttwood);    
#ifdef GRAVITY
      zlsio_Printf("%s\n",  DIVIDERLINE );
      Real gPot    = zlsg_phi(prob->gHeight, gp);
      Real dPotdz  = zlsg_dphi(prob->gHeight, gp);
      zlsio_Printf("    Phi     %#12.6g \n", gPot);    
      zlsio_Printf("    dPhi/z  %#12.6g \n", dPotdz);    
#else
      zlsio_Printf("    Phi     %#12.6g \n", 0.0);    
      zlsio_Printf("    dPhi/z  %#12.6g \n", 0.0);    
#endif !GRAVITY
      zlsio_Printf("%s\n",  DIVIDERLINE );
      if ( prob->gAntiAlias == kSmooth){
      zlsio_Printf("   Anti-Alias   :  ON \n");    
      zlsio_Printf("   Smooth Cells :  %g \n", prob->gSmooth);    
      } else{
      zlsio_Printf("   Anti-Alias   :  OFF \n");    
      }
      zlsio_Printf("%s\n",  DIVIDERLINE );
      
      zlsio_Printf("     A0 %#6.2g \n", prob->gA0);    
      zlsio_Printf("     F0 %#6.2g \n", prob->gF0);    
      
      zlsio_Printf("%s\n",  DIVIDERLINE );
      zlsio_Printf(" Atmosphere Parameters:\n");
      zlsio_Printf("   d    %#6.2g (g/cm^3) \n", prob->gD0*f_d0);    
      zlsio_Printf("   P    %#6.2g (dyne/cm^2) \n", prob->gP0*f_P0); 
      zlsio_Printf("   Cs   %#6.2g (isothermal) \n", zls_sqrt(prob->gC2)*f_v0); 
      zlsio_Printf("   C2   %#6.2g (P/rho) \n", prob->gC2); 
      zlsio_Printf("%s\n",  DIVIDERLINE );
      
      
//======================================================================
// This code sets the problem into the level structures, mulit-zone
// multi-level or simple single structures 
//======================================================================
      
      
      gLevels[0].levelProblem = prob; // register the problem with all levels from 0 down.      
      if (gNLevels>1){
        for (level = 1; level < gNLevels; level++){
          gLevels[level].levelProblem = gLevels[0].levelProblem;
        }
      }
      
//======================================================================
// Now the level contents can be set - standard calls here dont' change
//======================================================================
      
      for (level = 0; level < gNLevels; level++){
        
#ifdef GRAVITY
        err =  zlsg_InitLevelGravityField( level ); // precompute the potential of to the gravity field over the level
        if (err != noErr ) {
          zlsio_ErrPrintf( " ERROR: zlsg_InitLevelGravityField: Level Field Failed: %d,  %d \n",level, err );
          return(err);
        }
#endif GRAVITY
#ifdef NEQIONISATION
        err =  zlsi_InitLevelIonisationState( level ); 
        if (err != noErr ) {
          zlsio_ErrPrintf( " ERROR: zlsi_InitLevelIonisationState: Level Failed: %d,  %d \n",level, err );
          return(err);
        }
#endif NEQIONISATION
        
        err = zlsp_InitLevelZones( level ); // last: user's levelProblem, initialised top down
        
      }
      
      if (gNLevels>1){
        for (level = (gNLevels-1); level > 0; level--){
          err = zlsbc_FillSuperLevelFromLevel(level);
        }
      }
      
#if (ZLS_VERBOSE>0)
      zlsio_Printf("%s\n",  DIVIDERLINE );
#else
      zlsio_LogPrintf("%s\n",  DIVIDERLINE );
#endif
      
      ccl_release(&config);
      
    }// config  
    
  } else {
    err = memErr;
    zlsio_ErrPrintf( " ERROR: Unable to allocate level 0 ProblemPtr \n");
    return (err);
  }
  
  return (err);
  
}

//======================================================================
// Routine for 'User; boundary cells set in grid.conf not used here
// but often used for jet inlet models
//======================================================================
#ifdef ZLS_USER_BOUNDARIES
int  zlsp_CalcUserZoneBCs ( ZonePtr z, int direction, int bndryKind ){
  
  int err = noErr;
  ProblemPtr prob = NULL;
  PhysicsPtr phys = NULL;
  
  
  
#ifdef GRAVITY
  Real phi;
#endif GRAVITY
  
  if (z != NULL){
#ifdef GRAVITY
    Field zoneF ;
    zls_GetZoneField(&zoneF, z); // nothing is allocated here, so no dispose is needed later
#endif GRAVITY
    WHD zoneW = z->zoneW;    // get zone physical variables    
    Integer level = z->zoneLevel;
    prob = gLevels[level].levelProblem;  // get the user's problem parameters
    phys = gLevels[0].levelPhysics;  // get the global Physics parameters
    
  } else { err = memErr;} // z != NULL
  
  return (err);
}
#endif ZLS_USER_BOUNDARIES


//======================================================================
// Routine for modifying anywhere in the grid in any way as part of
// a custom problem model. Not restricted to boundary cells, here used for 
// wind bubble/mass loss bubble energy injection and mass injection
// called after all other boundary conditions have been computed.
// Use sparingly, if not careful this routine can impact heavily on 
// the performance of the simulation.
//======================================================================
// 
// Normally deactivated (ie not even checked for calling) unless the macro
// ZLS_USER_UPDATE_ROUTINE is defined in zls_defs.  If not defined then
// this routine is not even callable - and this is the default setting.
// Also normally provided as an empty routine.
//

#ifdef ZLS_USER_UPDATE_ROUTINE

int  zlsp_UserZoneUpdate ( ZonePtr z  ){
  
  Integer err = noErr;
  ProblemPtr prob = NULL;
  PhysicsPtr phys = NULL;
  
  // A different type of BC, not the grid edges as all but setting cell values
  // on the grid.  Uses zoneW and is based on the zone data init routine rather than
  // the grid edge code of the basic boundary value routines.
  
  if ( z != NULL ){
    
    Integer level = z->zoneLevel;
    prob = gLevels[level].levelProblem;  // get the user's problem parameters
    phys = gLevels[level].levelPhysics;  // get the global eos parameters
    
    if ( level == (gNLevels-1)) {
      
        
    } // level
    
  } else { err = memErr; } // z
  
  return (err);
}
#endif ZLS_USER_UPDATE_ROUTINE
