#include <zls.h>

//======================================================================
//
// User problem parameter initialisation/default values routine
// Called before the user's problem config file is read to set
// all problem parameters to sensible defaults in case the init.conf
// is faultly or even missing
//
//  The fields of the problem structure are defined by the user in
//  zls_problem.h and can be any variables desired by the user's routines.
//  The problem data is stored by reference in gLevels[0] and shared by all 
//  other levels, and can be accessed by any zone at need (see zlsp_UserZoneUpdate
//  and zlsp_InitAtmosphere for examples.
//
//======================================================================

int zlsp_InitProblem    ( ProblemPtr   prob){
  int err = noErr;
  
  if (prob != NULL) {
   
    prob->gD0          = 1.0; 
    prob->gP0          = 1.0e-6; 
    prob->gvMag        = 1.0; 
    
  }
  
  return (err);

}

//======================================================================
// standard allocation and destructor for problem strucutres, called
// by the ZLS system at the right time, and should not need to be
// modified in any way by the user.  Here to keep all problem code together
//
// The only exception is if levels are destroyed during a run and the user
// needs to deallocate any dynamic memory set up in zlsp_InitProblem
// or in zlsp_SetUpProblemFromConfigFile.  If this is the case then 
// the user needs to deallocate the memory in zlsp_DisposeProblem
// before  free(prob) is called.
//
// Currently the level list is fixed for the entire run and this is not
// necessary, but it will be in AMR.  Will be called by zls_DisposeLevel
// which doesn't exist yet. :)
//======================================================================

int zlsp_NewProblem(ProblemPtr * prob){
  int err = noErr;
  
  *prob = (ProblemPtr)malloc(sizeof(Problem));
  
  if (*prob == NULL) {
    err = memErr;
#if (ZLS_DEBUG >0)
    zlsio_ErrPrintf( " ERROR: zlsp_NewProblem allocation failed %d \n",err);
#endif
  }
  
  return (err);
}

int zlsp_DisposeProblem(ProblemPtr prob){
  int err = noErr;
  
  if ( prob != NULL ) {
    free(prob);
  }
  
  return (err);
}

int zlsp_InitZoneValues( ZonePtr z );
int zlsp_InitZoneValues( ZonePtr z ){
  
//======================================================================
//  Initialise a zone system variables (ie W), usually self contained, 
//  here the initialisation is a bit messy so subroutines are called from
//  here.  When called all the variables for the zone are initialised and 
//  ready to go, grid coordinates, problem vars etc.
//======================================================================
  
  int err = noErr;
  
  if (z != NULL){
    
    Integer level = z->zoneLevel;
    
    if ( level == (gNLevels-1)) {
      
      WHD zoneW = z->zoneW;    
      ProblemPtr prob = gLevels[level].levelProblem;  // get the user's problem parameters
      
      
#ifdef SYM1D
      
      Integer nti   = z->zoneGrid.nt[kXDim];
      Integer i;
      for(i = 0; i< nti; i++){
        zoneW.d [i] =  prob->gD0;
        zoneW.P [i] =  prob->gP0;
        zoneW.vx[i] = -prob->gvMag;
      } //i
      
#endif SYM1D
      
#ifdef SYM2D
      
      ACoordPtr xx  = z->zoneGrid.x;
      ACoordPtr yx  = z->zoneGrid.y;
      
      Integer nti   = z->zoneGrid.nt[kXDim];
      Integer ntj   = z->zoneGrid.nt[kYDim];
      Integer i, j;
      
      Real x      = 0.0;
      Real y      = 0.0;
      Real radius = 0.0;
      Real fi, fj;
      
      for(j = 0; j < ntj; j++){
        y      = yx[j].c;
        for(i = 0; i< nti; i++){
          x      = xx[i].c;
          
          zoneW.d [j][i] = prob->gD0;
          zoneW.P [j][i] = prob->gP0;
          
          zls_CvtCoordsToProjRadius2D(&fi, &fj, x, y);
          
          zoneW.vx[j][i] = -fi*prob->gvMag;
          zoneW.vy[j][i] = -fj*prob->gvMag;
          
        } //i
      } //j
#endif SYM2D
      
#ifdef SYM3D
      
      
      ACoordPtr xx  = z->zoneGrid.x;
      ACoordPtr yx  = z->zoneGrid.y;
      ACoordPtr zx  = z->zoneGrid.z;
      
      Integer nti   = z->zoneGrid.nt[kXDim];
      Integer ntj   = z->zoneGrid.nt[kYDim];
      Integer ntk   = z->zoneGrid.nt[kZDim];
      Integer i, j, k;
      
      Real x      = 0.0;
      Real y      = 0.0;
      Real z      = 0.0;
      Real radius = 0.0;
      Real fi, fj, fk;
      
      for(k = 0; k < ntk; k++){
        z      = zx[k].c;
        for(j = 0; j < ntj; j++){
          y      = yx[j].c;
          for(i = 0; i< nti; i++){
            x      = xx[i].c;
            
            zoneW.d [k][j][i] = prob->gD0;
            zoneW.P [k][j][i] = prob->gP0;
            
            zls_CvtCoordsToProjRadius3D(&fi, &fj, &fk, x, y, z);
            
            zoneW.vx [k][j][i] = -fi*prob->gvMag;
            zoneW.vy [k][j][i] = -fj*prob->gvMag;
            zoneW.vz [k][j][i] = -fk*prob->gvMag;
            
            
          } //i
        } //j
      } //k
#endif SYM2D
      
      
    } 
    
  } else { err = memErr; } //z
  
  return(err);
}

//======================================================================
//  Standard zone initialsation routine
//======================================================================

int zlsp_InitZoneData( ZonePtr z ){
  
//======================================================================
//  Initialise a zone system variables (ie W), usually self contained, 
//  here the initialisation is a bit messy so subroutines are called from
//  here.  When called all the variables for the zone are initialised and 
//  ready to go, grid coordinates, problem vars etc.
//======================================================================
  
  int err = noErr;
  
  if (z != NULL){
    
    Integer level = z->zoneLevel;
    
    if ( level > 1) {
      err = zlsbc_FillZoneFromSuperZone( z, kInterpBCsFromSuper );
    } else {
      err = zlsp_InitZoneValues(z);   
    }
        
  } else { err = memErr; }//z
  
  return(err);
}

//======================================================================
//  Standard level initialsation routine, loops through all zones of
//   a level. Called by ZLS after all parameters, coordinates, physics
//   etc have been set up.  Sets the gravity field arrays before the
//   calling zlsp_InitZoneData.
//======================================================================

int zlsp_InitLevelZones(Integer level){
  
  
  int err = noErr;
  Integer zone;
  ZonePtr  z;
  
#if (ZLS_VERBOSE >0)
#ifdef MPI
    if ( gMPIrank == 0){
#endif MPI
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " Level %d User Zones Initialisation: \n", level);
      zlsio_Printf( "%s\n",  DIVIDERLINE );
#ifdef MPI
    }
#endif MPI
#endif
  
  for ( zone = 0; zone <gLevels[level].nZones; zone++){
    z =  gLevels[level].levelZones[zone];
	  if (z != NULL){
#ifdef MPI	
		  if (z->zoneRankID == gMPIrank) {
#endif MPI
			  if ( z->zoneType == kAlgorithmZoneType ){ // only process appropriate zones, ie skip image zones
				  
#if (ZLS_DEBUG>0)
				  zlsio_ErrPrintf( " Initialising Zone: %2.2d:%2.2d \n", level, zone );
#endif
				  
				  err = zlsp_InitZoneData( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsp_InitZoneData Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
          
				  err = zls_SetZoneConstantBCValues( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zls_SetZoneConstantBCValues Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
#ifdef MEANSPLITSWEEPS
#ifdef SYM2D
          err =  zlsa_Synch2DData(z);
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsa_Synch2DSymmetricSweeps Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
#endif SYM2D
#endif   MEANSPLITSWEEPS   
          
          
			  } // z->zoneType == kAlgorithmZoneType 
#ifdef MPI			
		  }
#endif MPI
	  } else { err = zonErr; }
  } // zone
  
  return(err);
  
}


//======================================================================
// Optional routine to allow overriding the hard wired zlsp_InitProblem
// parameters from a config file, specified in zls.conf - usually init.conf
// Called by ZLS at the right time(tm)
//======================================================================

int   zlsp_SetUpProblemFromConfigFile(char * confFile){
  
//======================================================================
// Open the config file confFile, and parse for any user definded
// key,value pairs to set problem parameters (not grid initialisation)
// Calls zlsp_InitProblem, and before zlsp_InitLevelZones
//======================================================================
  
  int err = noErr;
  struct ccl_t  config;
  const char    *str1;
  
  Integer level;
  
  ProblemPtr  prob = NULL;
  err = zlsp_NewProblem(&prob);
  
  if ( prob != NULL ) {
    
    config.comment_char = '#';
    config.sep_char     = '=';
    config.str_char     = '"';
    
    if (ccl_parse(&config, confFile)==noErr) {
      
#if (ZLS_VERBOSE > 0) 
      zlsio_Printf( "\n%s\n", DIVIDERLINE );
#endif
      zlsio_Printf( " FYRIS Hydro Problem Initialised from %s \n", confFile);
#if (ZLS_VERBOSE > 0) 
      zlsio_Printf(  "%s\n", DIVIDERLINE );
#endif
      
// Default Values
      
      err = zlsp_InitProblem(prob);
      
//======================================================================
// The user is free to define any key strings and interpret the 
// returned values in any way.  Some problems are simple (see the SOD setup)
// this one is more complex as it sets up a multi-component model with many
// different types of parameters.
//======================================================================
      
      Real g = kZLS_eos_gamma;
      if (( str1 = ccl_get( &config, "Gamma"))){
        g = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Gamma", str1);
#endif
      }
      if ( g != kZLS_eos_gamma){
         PhysicsPtr phys = gLevels[0].levelPhysics;
         zlsp_InitGlobalEOS_Gamma( phys, g);
      }
      
      Real d = prob->gD0;
      if (( str1 = ccl_get( &config, "Density"))){
        d = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Density", str1);
#endif
      }
      prob->gD0 = d;
      
      Real P = prob->gP0;
      if (( str1 = ccl_get( &config, "Pressure"))){
        P = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Pressure", str1);
#endif
      }
      prob->gP0 = P;
      
      Real v = prob->gvMag;
      if (( str1 = ccl_get( &config, "Velocity"))){
        v = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Velocity", str1);
#endif
      }
      prob->gvMag = v;
      
#if (ZLS_VERBOSE>0)
      
      zlsio_Printf("%s\n",  DIVIDERLINE );
      zlsio_Printf(" 2D Noh Parameters:\n");
      zlsio_Printf("    Gamma      :  %g \n", kZLS_eos_gamma);    
      zlsio_Printf("    Density    :  %g \n", prob->gD0);    
      zlsio_Printf("    Pressure   :  %g \n", prob->gP0);    
      zlsio_Printf("    Velocity   :  %g \n", prob->gvMag);    
      zlsio_Printf("%s\n",  DIVIDERLINE );
      
#else
// log even if not verbose

      zlsio_LogPrintf("%s\n",  DIVIDERLINE );
      zlsio_LogPrintf(" 2D Noh Parameters:\n");
      zlsio_LogPrintf("    Gamma      :  %g \n", kZLS_eos_gamma);    
      zlsio_LogPrintf("    Density    :  %g \n", prob->gD0);    
      zlsio_LogPrintf("    Pressure   :  %g \n", prob->gP0);    
      zlsio_LogPrintf("    Velocity   :  %g \n", prob->gvMag);    
      zlsio_LogPrintf("%s\n",  DIVIDERLINE );
      
#endif 
      
      
//======================================================================
// This code sets the problem into the level structures, mulit-zone
// multi-level or simple single structures 
//======================================================================
      
      
      gLevels[0].levelProblem = prob; // register the problem with all levels from 0 down.      
      if (gNLevels>1){
        for (level = 1; level < gNLevels; level++){
          gLevels[level].levelProblem = gLevels[0].levelProblem;
        }
      }
      
//======================================================================
// Now the level contents can be set - standard calls here dont' change
//======================================================================
      
      for (level = 0; level < gNLevels; level++){
        
#ifdef GRAVITY
        err =  zlsg_InitLevelGravityField( level ); // precompute the potential of to the gravity field over the level
        if (err != noErr ) {
          zlsio_ErrPrintf( " ERROR: zlsg_InitLevelGravityField: Level Field Failed: %d,  %d \n",level, err );
          return(err);
        }
#endif GRAVITY
#ifdef NEQIONISATION
        err =  zlsi_InitLevelIonisationState( level ); 
        if (err != noErr ) {
          zlsio_ErrPrintf( " ERROR: zlsi_InitLevelIonisationState: Level Failed: %d,  %d \n",level, err );
          return(err);
        }
#endif NEQIONISATION
        
        err = zlsp_InitLevelZones( level ); // last: user's levelProblem, initialised top down
        
      }
      
      if (gNLevels>1){
        for (level = (gNLevels-1); level > 0; level--){
          err = zlsbc_FillSuperLevelFromLevel(level);
        }
      }
      
#if (ZLS_VERBOSE>0)
      zlsio_Printf("%s\n",  DIVIDERLINE );
#else
      zlsio_LogPrintf("%s\n",  DIVIDERLINE );
#endif
      
      ccl_release(&config);
      
    }// config  
    
  } else {
    err = memErr;
    zlsio_ErrPrintf( " ERROR: Unable to allocate level 0 ProblemPtr \n");
    return (err);
  }
  
  return (err);
  
}

//======================================================================
// Routine for 'User; boundary cells set in grid.conf not used here
// but often used for jet inlet models
//======================================================================
#ifdef ZLS_USER_BOUNDARIES
int  zlsp_CalcUserZoneBCs ( ZonePtr z, int direction, int bndryKind ){
  
  int err = noErr;
  if ( z != NULL ) {  
    
    Integer level = z->zoneLevel;
    
    WHD zoneW = z->zoneW;    
    ProblemPtr prob = gLevels[level].levelProblem;  // get the user's problem parameters
    
#ifdef SYM1D
    Integer   id;   
    QPtr       q;

    ACoordPtr xx  = z->zoneGrid.x;
    
    Integer is    = z->zoneGrid.ns[kXDim];
    Integer ie    = z->zoneGrid.ne[kXDim];
    
    Integer nti   = z->zoneGrid.nt[kXDim];
    
    Integer i;

    Integer  nbc  = z->zoneBCells;
   
    
    Real P0     = prob->gP0;
    Real d0     = prob->gD0;
    Real vMag   = prob->gvMag;

    Real x      = 0.0;
    Real radius = 0.0;
    Real fi;

	   Real t      = zlst_GlobalTime;
     Real dt     = zlst_GlobalTimeStep;
    
 		for (id = 0; id <  z->nQ; id++){
      
			if (z->zoneData[id]!=NULL){
				
				q = z->zoneData[id];
				
				if (q!=NULL){
                        
          if (direction == kDirLeft)  {
            for(i = 0; i < nbc; i++){
              x = xx[is-i-1].c;
                
                radius = zls_CvtCoordsToProjRadius1D(&fi, x);
                
                Real vx = -fi*vMag;

#ifdef CARTESIAN          
                Real d  = d0;
#endif CARTESIAN          
#ifdef CIRCULAR          
                Real d  = d0*(1.0+ (t/radius));
#endif CIRCULAR          
#ifdef SPHERICAL          
                Real d  = d0*(1.0+ (t/radius))*(1.0+ (t/radius));
#endif SPHERICAL          
               
                switch (id) {
                  case kWDens:
                    q[is-i-1] = d;
                    break;
                  case kWPres:
                    q[is-i-1] = P0;
                    break;
                  case kWVx:
                    q[is-i-1] = vx;
                    break;
                  default:
                    ;
                } // id switch
            } //i
          } // direction switch
          
          if (direction == kDirRight)  {
            for(i = 0; i < nbc; i++){
              x = xx[ie+i].c;
                radius = zls_CvtCoordsToProjRadius1D(&fi, x);
                
                Real vx = -fi*vMag;

#ifdef CARTESIAN          
                Real d  = d0;
#endif CARTESIAN          
#ifdef CIRCULAR          
                Real d  = d0*(1.0+ (t/radius));
#endif CIRCULAR          
#ifdef SPHERICAL          
                Real d  = d0*(1.0+ (t/radius))*(1.0+ (t/radius));
#endif SPHERICAL          
                
                switch (id) {
                  case kWDens:
                    q[ie+i] = d;
                    break;
                  case kWPres:
                    q[ie+i] = P0;
                    break;
                  case kWVx:
                    q[ie+i] = vx;
                    break;
                  default:
                    ;
                } // id switch
            } //i
          } // direction switch
          
          
                                                  
				} else { err = memErr; } // q
			} // z->zoneData[id]!=NULL
		} // var id loop

#endif SYM1D

#ifdef SYM2D
    Integer   id;   
    QArr       q;

    ACoordPtr xx  = z->zoneGrid.x;
    ACoordPtr yx  = z->zoneGrid.y;
    
    Integer is    = z->zoneGrid.ns[kXDim];
    Integer ie    = z->zoneGrid.ne[kXDim];
    
    Integer js    = z->zoneGrid.ns[kYDim];
    Integer je    = z->zoneGrid.ne[kYDim];
    
    Integer nti   = z->zoneGrid.nt[kXDim];
    Integer ntj   = z->zoneGrid.nt[kYDim];
    
    Integer i, j;

    Integer  nbc  = z->zoneBCells;
   
    
    Real P0     = prob->gP0;
    Real d0     = prob->gD0;
    Real vMag   = prob->gvMag;

    Real x      = 0.0;
    Real y      = 0.0;
    Real radius = 0.0;
    Real fi, fj;

	   Real t      = zlst_GlobalTime;
     Real dt     = zlst_GlobalTimeStep;
    
 		for (id = 0; id <  z->nQ; id++){
      
			if (z->zoneData[id]!=NULL){
				
#ifdef ZONEDATA_ARE_PLAINBLOCKS
				zls_SetQuantity2DArr(&q, z->zoneData[id], ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
				
#ifdef ZONEDATA_ARE_MULTIDARRAYS
				q = z->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
				
				if (q!=NULL){
                        
          if (direction == kDirLeft)  {
            for(i = 0; i < nbc; i++){
              x = xx[is-i-1].c;
              for(j = 0; j < ntj; j++){ 
                y = yx[j].c;
                
                radius = zls_CvtCoordsToProjRadius2D(&fi, &fj, x, y);
                
                Real vx = -fi*vMag;
                Real vy = -fj*vMag;
                
#ifdef CARTESIAN          
                Real d  = d0*(1.0+ (t/radius));
#endif CARTESIAN          
#ifdef CIRCULAR          
                Real d  = d0*(1.0+ (t/radius));
#endif CIRCULAR          
#ifdef SPHERICAL          
                Real d  = d0*(1.0+ (t/radius))*(1.0+ (t/radius));
#endif SPHERICAL          
                
                switch (id) {
                  case kWDens:
                    q[j][is-i-1] = d;
                    break;
                  case kWPres:
                    q[j][is-i-1] = P0;
                    break;
                  case kWVx:
                    q[j][is-i-1] = vx;
                    break;
                  case kWVy:
                    q[j][is-i-1] = vy;
                    break;
                  default:
                    ;
                } // id switch
              } //j
            } //i
          } // direction switch
          
          if (direction == kDirRight)  {
            for(i = 0; i < nbc; i++){
              x = xx[ie+i].c;
              for(j = 0; j < ntj; j++){ 
                y = yx[j].c;
                
                radius = zls_CvtCoordsToProjRadius2D(&fi, &fj, x, y);
                
                Real vx = -fi*vMag;
                Real vy = -fj*vMag;
                

#ifdef CARTESIAN          
                Real d  = d0*(1.0+ (t/radius));
#endif CARTESIAN          
#ifdef CIRCULAR          
                Real d  = d0*(1.0+ (t/radius));
#endif CIRCULAR          
#ifdef SPHERICAL          
                Real d  = d0*(1.0+ (t/radius))*(1.0+ (t/radius));
#endif SPHERICAL          
                
                switch (id) {
                  case kWDens:
                    q[j][ie+i] = d;
                    break;
                  case kWPres:
                    q[j][ie+i] = P0;
                    break;
                  case kWVx:
                    q[j][ie+i] = vx;
                    break;
                  case kWVy:
                    q[j][ie+i] = vy;
                    break;
                  default:
                    ;
                } // id switch
              } //j
            } //i
          } // direction switch
          
          if (direction == kDirDown)  {
            for(i = 0; i < nti; i++){
              x = xx[i].c;
              for(j = 0; j < nbc; j++){ 
                y = yx[js-j-1].c;
                
                radius = zls_CvtCoordsToProjRadius2D(&fi, &fj, x, y);
                
                Real vx = -fi*vMag;
                Real vy = -fj*vMag;
                

#ifdef CARTESIAN          
                Real d  = d0*(1.0+ (t/radius));
#endif CARTESIAN          
#ifdef CIRCULAR          
                Real d  = d0*(1.0+ (t/radius));
#endif CIRCULAR          
#ifdef SPHERICAL          
                Real d  = d0*(1.0+ (t/radius))*(1.0+ (t/radius));
#endif SPHERICAL          
                
                switch (id) {
                  case kWDens:
                    q[js-j-1][i] = d;
                    break;
                  case kWPres:
                    q[js-j-1][i] = P0;
                    break;
                  case kWVx:
                    q[js-j-1][i] = vx;
                    break;
                  case kWVy:
                    q[js-j-1][i] = vy;
                    break;
                  default:
                    ;
                } // id switch
              } //j
            } //i
          } // direction switch
          
          
          if (direction == kDirUp)  {
            for(i = 0; i < nti; i++){
              x = xx[i].c;
              for(j = 0; j < nbc; j++){ 
                y = yx[je+j].c;
                
                radius = zls_CvtCoordsToProjRadius2D(&fi, &fj, x, y);
                
                Real vx = -fi*vMag;
                Real vy = -fj*vMag;
                
#ifdef CARTESIAN          
                Real d  = d0*(1.0+ (t/radius));
#endif CARTESIAN          
#ifdef CIRCULAR          
                Real d  = d0*(1.0+ (t/radius));
#endif CIRCULAR          
#ifdef SPHERICAL          
                Real d  = d0*(1.0+ (t/radius))*(1.0+ (t/radius));
#endif SPHERICAL          
                
                switch (id) {
                  case kWDens:
                    q[je+j][i] = d;
                    break;
                  case kWPres:
                    q[je+j][i] = P0;
                    break;
                  case kWVx:
                    q[je+j][i] = vx;
                    break;
                  case kWVy:
                    q[je+j][i] = vy;
                    break;
                  default:
                    ;
                } // id switch
              } //j
            } //i
          } // direction switch
                              
#ifdef ZONEDATA_ARE_PLAINBLOCKS
					zls_ReleaseQuantity2DArr(q);
#endif ZONEDATA_ARE_PLAINBLOCKS
          
				} else { err = memErr; } // q
			} // z->zoneData[id]!=NULL
		} // var id loop

#endif SYM2D
    

#ifdef SYM3D
    Integer   id;   
    QArr       q;

    ACoordPtr xx  = z->zoneGrid.x;
    ACoordPtr yx  = z->zoneGrid.y;
    ACoordPtr zx  = z->zoneGrid.z;
    
    Integer is    = z->zoneGrid.ns[kXDim];
    Integer ie    = z->zoneGrid.ne[kXDim];
    
    Integer js    = z->zoneGrid.ns[kYDim];
    Integer je    = z->zoneGrid.ne[kYDim];
    
    Integer ks    = z->zoneGrid.ns[kZDim];
    Integer ke    = z->zoneGrid.ne[kZDim];
    
    Integer nti   = z->zoneGrid.nt[kXDim];
    Integer ntj   = z->zoneGrid.nt[kYDim];
    Integer ntk   = z->zoneGrid.nt[kZDim];
    
    Integer i, j, k;

    Integer  nbc  = z->zoneBCells;
   
    
    Real P0     = prob->gP0;
    Real d0     = prob->gD0;
    Real vMag   = prob->gvMag;

    Real x      = 0.0;
    Real y      = 0.0;
    Real zc      = 0.0;
    Real radius = 0.0;
    Real fi, fj, fk;

	   Real t      = zlst_GlobalTime;
     Real dt     = zlst_GlobalTimeStep;
    
 		for (id = 0; id <  z->nQ; id++){
      
			if (z->zoneData[id]!=NULL){
				
#ifdef ZONEDATA_ARE_PLAINBLOCKS
				zls_SetQuantity3DArr(&q, z->zoneData[id], ntk, ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
				
#ifdef ZONEDATA_ARE_MULTIDARRAYS
				q = z->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
				
				if (q!=NULL){
          
          if (direction == kDirLeft)  {
            for(k = 0; k < ntk; k++){ 
              zc = zx[k].c;
              for(i = 0; i < nbc; i++){
                x = xx[is-i-1].c;
                for(j = 0; j < ntj; j++){ 
                  y = yx[j].c;
                  
                  radius = zls_CvtCoordsToProjRadius3D(&fi, &fj, &fk, x, y, zc);
                  
                  
                  Real vx = -fi*vMag;
                  Real vy = -fj*vMag;
                  Real vz = -fk*vMag;
                  
                  Real d  = d0*(1.0+ (t/radius))*(1.0+ (t/radius));
                  
                  switch (id) {
                    case kWDens:
                      q[k][j][is-i-1] = d;
                      break;
                    case kWPres:
                      q[k][j][is-i-1] = P0;
                      break;
                    case kWVx:
                      q[k][j][is-i-1] = vx;
                      break;
                    case kWVy:
                      q[k][j][is-i-1] = vy;
                      break;
                    case kWVz:
                      q[k][j][is-i-1] = vz;
                      break;
                    default:
                      ;
                  } // id switch
                } //j
              } //i
            } //k
          } // direction switch
          
          if (direction == kDirRight)  {
            for(k = 0; k < ntk; k++){ 
              zc = zx[k].c;
              for(i = 0; i < nbc; i++){
                x = xx[ie+i].c;
                for(j = 0; j < ntj; j++){ 
                  y = yx[j].c;
                  
                  radius = zls_CvtCoordsToProjRadius3D(&fi, &fj, &fk, x, y, zc);
                  
                  
                  Real vx = -fi*vMag;
                  Real vy = -fj*vMag;
                  Real vz = -fk*vMag;
                  
                  Real d  = d0*(1.0+ (t/radius))*(1.0+ (t/radius));
                  
                  switch (id) {
                    case kWDens:
                      q[k][j][ie+i] = d;
                      break;
                    case kWPres:
                      q[k][j][ie+i] = P0;
                      break;
                    case kWVx:
                      q[k][j][ie+i] = vx;
                      break;
                    case kWVy:
                      q[k][j][ie+i] = vy;
                      break;
                    case kWVz:
                      q[k][j][ie+i] = vz;
                      break;
                    default:
                      ;
                  } // id switch
                } //j
              } //i
            } //k
          } // direction switch
          
          if (direction == kDirDown)  {
            for(k = 0; k < ntk; k++){ 
              zc = zx[k].c;
              for(i = 0; i < nti; i++){
                x = xx[i].c;
                for(j = 0; j < nbc; j++){ 
                  y = yx[js-j-1].c;
                  
                  radius = zls_CvtCoordsToProjRadius3D(&fi, &fj, &fk, x, y, zc);
                  
                  Real vx = -fi*vMag;
                  Real vy = -fj*vMag;
                  Real vz = -fk*vMag;
                  
                  Real d  = d0*(1.0+ (t/radius))*(1.0+ (t/radius));
                  
                  switch (id) {
                    case kWDens:
                      q[k][js-j-1][i] = d;
                      break;
                    case kWPres:
                      q[k][js-j-1][i] = P0;
                      break;
                    case kWVx:
                      q[k][js-j-1][i] = vx;
                      break;
                    case kWVy:
                      q[k][js-j-1][i] = vy;
                      break;
                    case kWVz:
                      q[k][js-j-1][i] = vz;
                      break;
                    default:
                      ;
                  } // id switch
                } //j
              } //i
            }// k
          } // direction switch
          
          
          if (direction == kDirUp)  {
            for(k = 0; k < ntk; k++){ 
              zc = zx[k].c;
              for(i = 0; i < nti; i++){
                x = xx[i].c;
                for(j = 0; j < nbc; j++){ 
                  y = yx[je+j].c;
                  
                  radius = zls_CvtCoordsToProjRadius3D(&fi, &fj, &fk, x, y, zc);
                  
                  Real vx = -fi*vMag;
                  Real vy = -fj*vMag;
                  Real vz = -fk*vMag;
                  
                  Real d  = d0*(1.0+ (t/radius))*(1.0+ (t/radius));
                  
                  switch (id) {
                    case kWDens:
                      q[k][je+j][i] = d;
                      break;
                    case kWPres:
                      q[k][je+j][i] = P0;
                      break;
                    case kWVx:
                      q[k][je+j][i] = vx;
                      break;
                    case kWVy:
                      q[k][je+j][i] = vy;
                      break;
                    case kWVz:
                      q[k][je+j][i] = vz;
                      break;
                    default:
                      ;
                  } // id switch
                } //j
              } //i
            } //k
          } // direction switch
          
          if (direction == kDirYon)  {
            for(i = 0; i < nti; i++){
              x = xx[i].c;
              for(j = 0; j < ntj; j++){
                y = yx[j].c;
                for(k = 0; k < nbc; k++){ 
                  zc = zx[ks-k-1].c;
                  
                  radius = zls_CvtCoordsToProjRadius3D(&fi, &fj, &fk, x, y, zc);
                  
                  Real vx = -fi*vMag;
                  Real vy = -fj*vMag;
                  Real vz = -fk*vMag;
                  
                  Real d  = d0*(1.0+ (t/radius))*(1.0+ (t/radius));
                  
                  switch (id) {
                    case kWDens:
                      q[ks-k-1][j][i] = d;
                      break;
                    case kWPres:
                      q[ks-k-1][j][i] = P0;
                      break;
                    case kWVx:
                      q[ks-k-1][j][i] = vx;
                      break;
                    case kWVy:
                      q[ks-k-1][j][i] = vy;
                      break;
                    case kWVz:
                      q[ks-k-1][j][i] = vz;
                      break;
                    default:
                      ;
                  } // id switch
                } // k
              } //j
            } //i
          } // direction switch
          
          
          if (direction == kDirHither)  {
            for(i = 0; i < nti; i++){
              x = xx[i].c;
              for(j = 0; j < ntj; j++){
                y = yx[j].c;
                for(k = 0; k < nbc; k++){ 
                  zc = zx[ke+k].c;
                  
                  radius = zls_CvtCoordsToProjRadius3D(&fi, &fj, &fk, x, y, zc);
                  
                  Real vx = -fi*vMag;
                  Real vy = -fj*vMag;
                  Real vz = -fk*vMag;
                  
                  Real d  = d0*(1.0+ (t/radius))*(1.0+ (t/radius));
                  
                  switch (id) {
                    case kWDens:
                      q[ke+k][j][i] = d;
                      break;
                    case kWPres:
                      q[ke+k][j][i] = P0;
                      break;
                    case kWVx:
                      q[ke+k][j][i] = vx;
                      break;
                    case kWVy:
                      q[ke+k][j][i] = vy;
                      break;
                    case kWVz:
                      q[ke+k][j][i] = vz;
                      break;
                    default:
                      ;
                  } // id switch
                } //k
              } //j
            } //i
          } // direction switch
          
#ifdef ZONEDATA_ARE_PLAINBLOCKS
					zls_ReleaseQuantity3DArr(q);
#endif ZONEDATA_ARE_PLAINBLOCKS
          
				} else { err = memErr; } // q
			} // z->zoneData[id]!=NULL
		} // var id loop

#endif SYM3D
    
    
  } else { err = memErr; } //z
    
    return(err);
  }
#endif ZLS_USER_BOUNDARIES


//======================================================================
// Routine for modifying anywhere in the grid in any way as part of
// a custom problem model. Not restricted to boundary cells, here used for 
// wind bubble/mass loss bubble energy injection and mass injection
// called after all other boundary conditions have been computed.
// Use sparingly, if not careful this routine can impact heavily on 
// the performance of the simulation.
//======================================================================
// 
// Normally deactivated (ie not even checked for calling) unless the macro
// ZLS_USER_UPDATE_ROUTINE is defined in zls_defs.  If not defined then
// this routine is not even callable - and this is the default setting.
// Also normally provided as an empty routine.
//

#ifdef ZLS_USER_UPDATE_ROUTINE

int  zlsp_UserZoneUpdate ( ZonePtr z  ){
    
  Integer err = noErr;
  
  if ( z != NULL ){
    
    WHD zoneW = z->zoneW;    
        
    
  } else { err = memErr; } // z
  
  return (err);
}
#endif ZLS_USER_UPDATE_ROUTINE


#ifdef SYM1D
int zlsp_1DNohZoneOutput( zFile f, ZonePtr z){ 
  int err = noErr;
  
  IOSettingsPtr io = gLevels[ z->zoneLevel].levelIO;
  
  if (( f!= NULL) &&(io != NULL) && (z != NULL)){
    
    WHD         zoneW =  z->zoneW;
    
    Real t      = zlst_GlobalTime;
    
#if (ZLS_VERBOSE > 1)
    zls_ZPrintZoneCoords( f,  z);
#endif (ZLS_VERBOSE > 1)
    
    ACoordPtr xx  = z->zoneGrid.x;
    
    Integer i, j;
    
    Real x      = 0.0;
    
    Real d      = 0.0;
    Real P      = 0.0;
    Real vx     = 0.0;
    Real vm     = 0.0;
    
    Real sumD      = 0.0;
    Real sumND     = 0.0;
    
    Real sumP      = 0.0;
    Real sumNP     = 0.0;
    
    Real gD  = 1.0;
    Real gP  = 1.0;
    
    Real t3 = t/3.0;

#ifdef CARTESIAN          
    Real d3 = 4.0;
    Real P3 = 4.0/3.0;
#endif CARTESIAN          
#ifdef CIRCULAR          
    Real d3 = 16.0;
    Real P3 = 16.0/3.0;
#endif CIRCULAR          
#ifdef SPHERICAL          
    Real d3 = 64.0;
    Real P3 = 64.0/3.0;
#endif SPHERICAL          
    
    Real radius = 0.0;
    Real fi;
    
    Integer is   = z->zoneGrid.ns[kXDim];
    Integer ie   = z->zoneGrid.ne[kXDim];
    
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    zlsio_ZFPrintf( f, " r \t x \t y \t d \t Nd \t P \t NP \t vm \t vx\n");
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    //     
      for(i = is; i< ie; i++){
        x      = xx[i].c;
        
        radius = zls_CvtCoordsToProjRadius1D(&fi, x);
        
        d = zoneW.d [i];
        P = zoneW.P [i];
        
        if ( radius < t3 ){
          gD = d3;
          gP = P3;
        }  else {
          gP = 1.0e-6;
#ifdef CARTESIAN          
          gD = (1.0); 
#endif CARTESIAN          
#ifdef CIRCULAR          
          gD = (1.0+(t/radius));
#endif CIRCULAR          
#ifdef SPHERICAL          
          gD = (1.0+(t/radius))*(1.0+(t/radius));
#endif SPHERICAL          
        }
        
        vx = zoneW.vx [i];
        vm = zls_abs(vx);
        
        if ( radius < 1.0){
          zlsio_ZFPrintf( f, " %#12.6g \t %#12.6g \t",  radius, x );
          zlsio_ZFPrintf( f, " %#12.6g \t %#12.6g \t %#12.6g \t %#12.6g \t",  d, gD,  P, gP );
          zlsio_ZFPrintf( f, " %#12.6g \t %#12.6g \n", vm,  vx);
        }
        
        sumD   += zls_abs((d-gD));
        sumND  += zls_abs(gD);
        
        sumP   += zls_abs((P-gP));
        sumNP  += zls_abs(gP);
        
      } //i
          
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    zlsio_ZFPrintf( f, "mean L1 Den     = \t %#12.6g percent\n",100.0*sumD/sumND);
    zlsio_ZFPrintf( f, "mean L1 Pre     = \t %#12.6g percent\n",100.0*sumP/sumNP);
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    
  }
  
  return(err);
}
#endif SYM1D

#ifdef SYM2D
int zlsp_2DNohZoneOutput( zFile f, ZonePtr z){ 
  int err = noErr;
  
  IOSettingsPtr io = gLevels[ z->zoneLevel].levelIO;
  
  if (( f!= NULL) &&(io != NULL) && (z != NULL)){
    
    WHD         zoneW =  z->zoneW;
    
    Real t      = zlst_GlobalTime;
    
#if (ZLS_VERBOSE > 1)
    zls_ZPrintZoneCoords( f,  z);
#endif (ZLS_VERBOSE > 1)
    
    ACoordPtr xx  = z->zoneGrid.x;
    ACoordPtr yx  = z->zoneGrid.y;
    
    Integer i, j;
    
    Real x      = 0.0;
    Real y      = 0.0;
    
    Real d      = 0.0;
    Real P      = 0.0;
    Real vx     = 0.0;
    Real vy     = 0.0;
    Real vm     = 0.0;
    
    Real sumD      = 0.0;
    Real sumND     = 0.0;
    
    Real sumP      = 0.0;
    Real sumNP     = 0.0;
    
    Real gD  = 1.0;
    Real gP  = 1.0;
    
    Real t3 = t/3.0;
#ifdef CARTESIAN           
    Real d3 = 16.0;// circular 2D cartesian test
    Real P3 = 16.0/3.0;
#endif CARTESIAN          
#ifdef CIRCULAR          
    Real d3 = 16.0;
    Real P3 = 16.0/3.0;
#endif CIRCULAR          
#ifdef SPHERICAL          
    Real d3 = 64.0;
    Real P3 = 64.0/3.0;
#endif SPHERICAL          
    
    Real radius = 0.0;
    Real fi, fj;
    
    Integer is   = z->zoneGrid.ns[kXDim];
    Integer ie   = z->zoneGrid.ne[kXDim];
    Integer js   = z->zoneGrid.ns[kYDim];
    Integer je   = z->zoneGrid.ne[kYDim];
    
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    zlsio_ZFPrintf( f, " r \t x \t y \t d \t Nd \t P \t NP \t vm \t vx \t vy\n");
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    //     
    for(j = js; j< je; j++){
      y      = yx[j].c;
      for(i = is; i< ie; i++){
        x      = xx[i].c;
        
        radius = zls_CvtCoordsToProjRadius2D(&fi, &fj, x, y);
        
        d = zoneW.d [j][i];
        P = zoneW.P [j][i];
        
        if ( radius < t3 ){
          gD = d3;
          gP = P3;
        }  else {
          gP = 1.0e-6;
#ifdef CARTESIAN          
          gD = (1.0+(t/radius));  // circular 2D cartesian test
#endif CARTESIAN          
#ifdef CIRCULAR          
          gD = (1.0+(t/radius));
#endif CIRCULAR          
#ifdef SPHERICAL          
          gD = (1.0+(t/radius))*(1.0+(t/radius));
#endif SPHERICAL          
       }
        
        vx = zoneW.vx [j][i];
        vy = zoneW.vy [j][i];
        vm = zls_sqrt(vx*vx+vy*vy);
        
        if ( radius < 1.0){
          zlsio_ZFPrintf( f, " %#12.6g \t %#12.6g \t %#12.6g \t",  radius, x,  y );
          zlsio_ZFPrintf( f, " %#12.6g \t %#12.6g \t %#12.6g \t %#12.6g \t",  d, gD,  P, gP );
          zlsio_ZFPrintf( f, " %#12.6g \t %#12.6g \t %#12.6g \n", vm,  vx, vy);
        }
        
        sumD   += zls_abs((d-gD));
        sumND  += zls_abs(gD);
        
        sumP   += zls_abs((P-gP));
        sumNP  += zls_abs(gP);
        
      } //i
    } //j
    
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    zlsio_ZFPrintf( f, "mean L1 Den     = \t %#12.6g percent\n",100.0*sumD/sumND);
    zlsio_ZFPrintf( f, "mean L1 Pre     = \t %#12.6g percent\n",100.0*sumP/sumNP);
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    
  }
  
  return(err);
}
#endif SYM2D


#ifdef SYM3D
int zlsp_3DNohZoneOutput( zFile f, ZonePtr z){ 
  int err = noErr;
  
  IOSettingsPtr io = gLevels[ z->zoneLevel].levelIO;
  
  if (( f!= NULL) &&(io != NULL) && (z != NULL)){
    
    WHD         zoneW =  z->zoneW;
    
    Real t      = zlst_GlobalTime;
    
#if (ZLS_VERBOSE > 1)
    zls_ZPrintZoneCoords( f,  z);
#endif (ZLS_VERBOSE > 1)
    
    ACoordPtr xx  = z->zoneGrid.x;
    ACoordPtr yx  = z->zoneGrid.y;
    ACoordPtr zx  = z->zoneGrid.z;
    
    Integer i, j, k;
    
    Real xc      = 0.0;
    Real yc      = 0.0;
    Real zc      = 0.0;
    
    Real d      = 0.0;
    Real P      = 0.0;
    Real vx     = 0.0;
    Real vy     = 0.0;
    Real vz     = 0.0;
    Real vm     = 0.0;
    
    Real sumD      = 0.0;
    Real sumND     = 0.0;
    
    Real sumP      = 0.0;
    Real sumNP     = 0.0;
    
    Real gD  = 1.0;
    Real gP  = 1.0;
    
    Real t3 = t/3.0;
    Real d3 = 64.0;
    Real P3 = 64.0/3.0;
    
    Real radius = 0.0;
    Real fi, fj, fk;
    
    Integer is   = z->zoneGrid.ns[kXDim];
    Integer ie   = z->zoneGrid.ne[kXDim];
    Integer js   = z->zoneGrid.ns[kYDim];
    Integer je   = z->zoneGrid.ne[kYDim];
    Integer ks   = z->zoneGrid.ns[kZDim];
    Integer ke   = z->zoneGrid.ne[kZDim];
    
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    zlsio_ZFPrintf( f, " r \t x \t d \t Nd \t P \t NP \t vm \t vx \t vy\n");
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    //     
    for(k = ks; k< ke; k++){
      zc      = zx[k].c;
    for(j = js; j< je; j++){
      yc      = yx[j].c;
      for(i = is; i< ie; i++){
        xc      = xx[i].c;
        
        radius = zls_CvtCoordsToProjRadius3D(&fi, &fj, &fk, xc, yc, zc);
        
        d = zoneW.d [k][j][i];
        P = zoneW.P [k][j][i];
        
        if ( radius < t3 ){
          gD = d3;
          gP = P3;
        }  else {
          gP = 1.0e-6;
          gD = (1.0+(t/radius))*(1.0+(t/radius));
       }
        
        vx = zoneW.vx [k][j][i];
        vy = zoneW.vy [k][j][i];
        vz = zoneW.vz [k][j][i];
        vm = zls_sqrt(vx*vx+vy*vy+vz*vz);
        
        if ( (j == js) && ( k == ks) ){
          zlsio_ZFPrintf( f, " %#12.6g \t %#12.6g \t %#12.6g \t %#12.6g \t",  radius, xc,  yc, zc );
          zlsio_ZFPrintf( f, " %#12.6g \t %#12.6g \t %#12.6g \t %#12.6g \t",  d, gD,  P, gP );
          zlsio_ZFPrintf( f, " %#12.6g \t %#12.6g \t %#12.6g \t %#12.6g \n", vm,  vx, vy, vz);
        }
        
        sumD   += zls_abs((d-gD));
        sumND  += zls_abs(gD);
        
        sumP   += zls_abs((P-gP));
        sumNP  += zls_abs(gP);
        
      } //i
    } //j
    } //k
    
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    zlsio_ZFPrintf( f, "mean L1 Den     = \t %#12.6g percent\n",100.0*sumD/sumND);
    zlsio_ZFPrintf( f, "mean L1 Pre     = \t %#12.6g percent\n",100.0*sumP/sumNP);
    zlsio_ZFPrintf( f, "%s\n",  DIVIDERLINE );
    
  }
  
  return(err);
}
#endif SYM3D
