#include <zls.h>


//======================================================================
//
// User problem parameter initialisation/default values routine
// Called before the user's problem config file is read to set
// all problem parameters to sensible defaults in case the init.conf
// is faultly or even missing
//
//  The fields of the problem structure are defined by the user in
//  zls_problem.h and can be any variables desired by the user's routines.
//  The problem data is stored by reference in gLevels[0] and shared by all 
//  other levels, and can be accessed by any zone at need (see zlsp_UserZoneUpdate
//  and zlsp_InitAtmosphere for examples.
//
//======================================================================

int zlsp_InitProblem    ( ProblemPtr   prob){
  int err = noErr;
  
// Std SOD Shock problem	
	
  if (prob != NULL) {

    prob->gX0 =    0.0; 
    prob->gX1 = 9999.0; 
    
    prob->gP0  =   1.0;
    prob->gd0  =   1.0; 
    prob->gv0  =   0.0; 
    
    prob->gP1  =   0.1;
    prob->gd1  = 0.125; 
    prob->gv1  =   0.0; 
    
    prob->gP2  =   0.1;
    prob->gd2  = 0.125; 
    prob->gv2  =   0.0;
    
  }
	
  return (err);
}

//======================================================================
// standard allocation and destructor for problem strucutres, called
// by the ZLS system at the right time, and should not need to be
// modified in any way by the user.  Here to keep all problem code together
//
// The only exception is if levels are destroyed during a run and the user
// needs to deallocate any dynamic memory set up in zlsp_InitProblem
// or in zlsp_SetUpProblemFromConfigFile.  If this is the case then 
// the user needs to deallocate the memory in zlsp_DisposeProblem
// before  free(prob) is called.
//
// Currently the level list is fixed for the entire run and this is not
// necessary, but it will be in AMR.  Will be called by zls_DisposeLevel
// which doesn't exist yet. :)
//======================================================================


int zlsp_NewProblem(ProblemPtr * prob){
  int err = noErr;
	
  *prob = (ProblemPtr)malloc(sizeof(Problem));
	
  if (*prob == NULL) {
    err = memErr;
#if (ZLS_DEBUG >0)
    fprintf(stderr, " ERROR: zlsp_NewProblem allocation failed %d \n",err);
#endif
  }
	
  return (err);
}


int zlsp_DisposeProblem(ProblemPtr prob){
  int err = noErr;
	
  if ( prob != NULL ) {
    free(prob);
  }
	
  return (err);
}




int zlsp_InitZoneData( ZonePtr z){
  
//======================================================================
//  Initialise a zone system variables (ie W), usually self contained, 
//  here the initialisation is a bit messy so subroutines are called from
//  here.  When called all the variables for the zone are initialised and 
//  ready to go, grid coordinates, problem vars etc.
//======================================================================

  int err = noErr;
	
  WHD       zoneW;
  ACoordPtr  xx;
  ProblemPtr prob = NULL;
  
	
  if (z != NULL){
		
    err     = zlsp_GetZoneWHD(&zoneW, z);
		
		
    if (err == noErr) {
      
      prob = gLevels[0].levelProblem;  // get the user's problem parameters
      
      Real x0  = prob->gX0;
      Real x1  = prob->gX1;
      
      Real d0  = prob->gd0;
      Real P0  = prob->gP0;
      Real v0  = prob->gv0;
      
      Real d1  = prob->gd1;
      Real P1  = prob->gP1;
      Real v1  = prob->gv1;
      
      Real d2  = prob->gd2;
      Real P2  = prob->gP2;
      Real v2  = prob->gv2;
      
      
#ifdef SYM1D
      Integer    i ;
      xx = z->zoneGrid.x;
      
      Integer nti   = z->zoneGrid.nt[kXDim];
      
        for(i = 0; i< nti; i++){
          
          if (xx[i].l < x0){
            zoneW.d [i]      =  d0;
            zoneW.P [i]      =  P0;
            zoneW.vx[i]      =  v0;
            
          } else if (xx[i].l < x1){
            zoneW.d [i]      =  d1;
            zoneW.P [i]      =  P1;
            zoneW.vx[i]      =  v1;
            
          } else {
            zoneW.d [i]      =  d2;
            zoneW.P [i]      =  P2;
            zoneW.vx[i]      =  v2;
            
          }
          
        }//i
        
      
        // set wind density after outburst to low value for BC  updates
#endif SYM1D
      
#ifdef SYM2D
       Integer    i, j ;
       xx = z->zoneGrid.x;
      
      Integer nti   = z->zoneGrid.nt[kXDim];
      Integer ntj   = z->zoneGrid.nt[kYDim];
      
    for(j = 0; j < ntj; j++){
	   for(i = 0; i< nti; i++){
        
      if (xx[i].l < x0){
        zoneW.d [j][i]      =  d0;
        zoneW.P [j][i]      =  P0;
        zoneW.vx[j][i]      =  v0;
        zoneW.vy[j][i]      =  0.0;

      } else if (xx[i].l < x1) {
        zoneW.d [j][i]      =  d1;
        zoneW.P [j][i]      =  P1;
        zoneW.vx[j][i]      =  v1;
        zoneW.vy[j][i]      =  0.0;
        
      }  else  {
         zoneW.d [j][i]      =  d2;
         zoneW.P [j][i]      =  P2;
         zoneW.vx[j][i]      =  v2;
         zoneW.vy[j][i]      =  0.0;
         
       }
       
	}//i
       
      } //j
        
        // set wind density after outburst to low value for BC  updates
#endif SYM2D
            
#ifdef SYM3D
        Integer    i, j, k ;
     xx = z->zoneGrid.x;
      
      Integer nti   = z->zoneGrid.nt[kXDim];
      Integer ntj   = z->zoneGrid.nt[kYDim];
      Integer ntk   = z->zoneGrid.nt[kZDim];
      
      for(k = 0; k < ntk; k++){
      for(j = 0; j < ntj; j++){
	for(i = 0; i< nti; i++){
        
    if (xx[i].l < x0){
	    zoneW.d [k][j][i]      =  d0;
	    zoneW.P [k][j][i]      =  P0;
	    zoneW.vx[k][j][i]      =  v0;
	    zoneW.vy[k][j][i]      =  0.0;
	    zoneW.vz[k][j][i]      =  0.0;
      
    } else if (xx[i].l < x1){
        zoneW.d [k][j][i]      =  d1;
        zoneW.P [k][j][i]      =  P1;
        zoneW.vx[k][j][i]      =  v1;
        zoneW.vy[k][j][i]      =  0.0;
        zoneW.vz[k][j][i]      =  0.0;
        
      } else {
        zoneW.d [k][j][i]      =  d2;
        zoneW.P [k][j][i]      =  P2;
        zoneW.vx[k][j][i]      =  v2;
        zoneW.vy[k][j][i]      =  0.0;
        zoneW.vz[k][j][i]      =  0.0;
          
          }
          
	}//i
       
      } //j
      } //k
        
        // set wind density after outburst to low value for BC  updates
#endif SYM3D
      zlsp_ReleaseZoneWHD(zoneW);
        
    } // zoneW
		
  } //z
	
	
  return(err);
}



//======================================================================
//  Standard level initialsation routine, loops through all zones of
//   a level. Called by ZLS after all parameters, coordinates, physics
//   etc have been set up.  Sets the gravity field arrays before the
//   calling zlsp_InitZoneData.
//======================================================================

int zlsp_InitLevelZones(Integer level){
  
  
  int err = noErr;
  Integer zone;
  ZonePtr  z;
  
#if (ZLS_VERBOSE >0)
#ifdef MPI
    if ( gMPIrank == 0){
#endif MPI
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " Level %d User Zones Initialisation: \n", level);
      zlsio_Printf( "%s\n",  DIVIDERLINE );
#ifdef MPI
    }
#endif MPI
#endif
  
  for ( zone = 0; zone <gLevels[level].nZones; zone++){
    z =  gLevels[level].levelZones[zone];
	  if (z != NULL){
#ifdef MPI	
		  if (z->zoneRankID == gMPIrank) {
#endif MPI
			  if ( z->zoneType == kAlgorithmZoneType ){ // only process appropriate zones, ie skip image zones
				  
#if (ZLS_DEBUG>0)
				  zlsio_ErrPrintf( " Initialising Zone: %2.2d:%2.2d \n", level, zone );
#endif
				  
				  err = zlsp_InitZoneData( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsp_InitZoneData Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
          
				  err = zls_SetZoneConstantBCValues( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zls_SetZoneConstantBCValues Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
#ifdef MEANSPLITSWEEPS
#ifdef SYM2D
          err =  zlsa_Synch2DData(z);
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsa_Synch2DSymmetricSweeps Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
#endif SYM2D
#endif   MEANSPLITSWEEPS   
          
          
			  } // z->zoneType == kAlgorithmZoneType 
#ifdef MPI			
		  }
#endif MPI
	  } else { err = zonErr; }
  } // zone
  
  return(err);
  
}

//======================================================================
// Optional routine to allow overriding the hard wired zlsp_InitProblem
// parameters from a config file, specified in zls.conf - usually init.conf
// Called by ZLS at the right time(tm)
//======================================================================

int   zlsp_SetUpProblemFromConfigFile(char * confFile){
  
//======================================================================
// Open the config file confFile, and parse for any user definded
// key,value pairs to set problem parameters (not grid initialisation)
// Calls zlsp_InitProblem, and before zlsp_InitLevelZones
//======================================================================
  
	
  int err = noErr;
  struct ccl_t	config;
  const char		*str1;

  Integer level;
	
  ProblemPtr  prob = NULL;
  err = zlsp_NewProblem(&prob);
  
  if ( prob != NULL ) {
    
    config.comment_char = '#';
    config.sep_char     = '=';
    config.str_char     = '"';
    
    if (ccl_parse(&config, confFile)==noErr) {
      
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " FYRIS Hydro Problem Initialising from %s \n", confFile);
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      
      // Default Values: std SOD tube
      err = zlsp_InitProblem(prob);
      
      if ((str1= ccl_get( &config,"SOD_X0"))){
        prob->gX0 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_X1"))){
        prob->gX1 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_P0"))){
        prob->gP0 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_D0"))){
        prob->gd0 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_V0"))){
        prob->gv0 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_P1"))){
        prob->gP1 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_D1"))){
        prob->gd1 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_V1"))){
        prob->gv1 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_P2"))){
        prob->gP2 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_D2"))){
        prob->gd2 = atof( str1);
      }
      
      if ((str1= ccl_get( &config,"SOD_V2"))){
        prob->gv2 = atof( str1);
      }
      
      gLevels[0].levelProblem = prob; // register the problem with all levels from 0 down.      
      if (gNLevels>1){
        for (level = 1; level < gNLevels; level++){
          gLevels[level].levelProblem = gLevels[0].levelProblem;
        }
      }
      
      for (level = 0; level < gNLevels; level++){
        err = zlsp_InitLevelZones( level ); // uses levelProblem
      }
      
      zlsio_Printf( "%s\n\n",  DIVIDERLINE );
      
      ccl_release(&config);
      
    }// config	
    
  } else {
    err = memErr;
    fprintf( stderr, " ERROR: Unable to allocate level 0 ProblemPtr \n");
    return (err);
  }
  
  return (err);
  
}

#ifdef ZLS_USER_BC_ROUTINE
//
int  zlsp_CalcUserZoneBCs ( ZonePtr z, int direction, int bndryKind ){
  Integer err = noErr;
	return (err);
}
#endif ZLS_USER_BC_ROUTINE


#ifdef ZLS_USER_UPDATE_ROUTINE
int zlsp_UserZoneUpdate( ZonePtr z){
  //
  // This routine is called just before sweeps are done, and after edge boundaries are set up.
  // The user has an opportunity to include any customised zone values, eg a stellar wind update or
  // coordinate evolution for a star.  Basicly for all boundary conditions that don't fit in the
  // grid boundary direction updates scheme.
  //  Many simulations will have nothing to do here so it is best to 
  // comment out as all the code unless a zone update is actually needed.
  // Define the global macro in zls_defs.h ZLS_USER_UPDATE_ROUTINE to enable this
  // routine.
  //
  int err = noErr;
  
  ProblemPtr prob = NULL;
  PhysicsPtr phys = NULL;
  
  if (z != NULL){
    
    WHD zoneW = z->zoneW;    
    prob = gLevels[0].levelProblem;  // get the user's problem parameters
    phys = gLevels[0].levelPhysics;  // get the global eos parameters
        
  } else { err = memErr; }//z
  
  
  return(err);
}
#endif ZLS_USER_UPDATE_ROUTINE

