#include <zls.h>

//======================================================================
//
// User problem parameter initialisation/default values routine
// Called before the user's problem config file is read to set
// all problem parameters to sensible defaults in case the init.conf
// is faultly or even missing
//
//  The fields of the problem structure are defined by the user in
//  zls_problem.h and can be any variables desired by the user's routines.
//  The problem data is stored by reference in gLevels[0] and shared by all 
//  other levels, and can be accessed by any zone at need (see zlsp_UserZoneUpdate
//  and zlsp_InitAtmosphere for examples.
//
//======================================================================

int zlsp_InitProblem    ( ProblemPtr   prob){
  int err = noErr;
  
  if (prob != NULL) {

    
#ifdef SYM1D
    prob->gVISM[kXDim]  = 0.0*f_cmkm/f_v0; // 30 km/s
#endif SYM1D
#ifdef SYM2D
    prob->gVISM[kXDim]  = 0.0*f_cmkm/f_v0; // 30 km/s
    prob->gVISM[kYDim]  = 0.0*f_cmkm/f_v0; 
#endif SYM2D
#ifdef SYM3D
    prob->gVISM[kXDim]  = 0.0*f_cmkm/f_v0; // 30 km/s
    prob->gVISM[kYDim]  = 0.0*f_cmkm/f_v0; 
    prob->gVISM[kZDim]  = 0.0*f_cmkm/f_v0; 
#endif SYM3D

    prob->gNstars             = 1; 
    prob->gRWindR0[0]         = 0.01;
    
#ifdef SYM1D
    prob->gXStar[0]    = 0.0;
#endif SYM1D
#ifdef SYM2D
    prob->gXStar[0]    = 0.0;
    prob->gYStar[0]    = 0.0;
#endif SYM2D
#ifdef SYM3D
    prob->gXStar[0]    = 0.0;
    prob->gYStar[0]    = 0.0;
    prob->gZStar[0]    = 0.0;
#endif SYM3D
    
    prob->gVWind[0]    =  1.0;
    prob->gDWind[0]    = 0.01;
    prob->gPWind[0]    = 0.01;
    
    prob->gD0   = 1.0;
    prob->gP0   = 1.0;
    
  }
  return (err);
}

//======================================================================
// standard allocation and destructor for problem strucutres, called
// by the ZLS system at the right time, and should not need to be
// modified in any way by the user.  Here to keep all problem code together
//
// The only exception is if levels are destroyed during a run and the user
// needs to deallocate any dynamic memory set up in zlsp_InitProblem
// or in zlsp_SetUpProblemFromConfigFile.  If this is the case then 
// the user needs to deallocate the memory in zlsp_DisposeProblem
// before  free(prob) is called.
//
// Currently the level list is fixed for the entire run and this is not
// necessary, but it will be in AMR.  Will be called by zls_DisposeLevel
// which doesn't exist yet. :)
//======================================================================

int zlsp_NewProblem(ProblemPtr * prob){
  int err = noErr;
  
  *prob = (ProblemPtr)malloc(sizeof(Problem));
  
  if (*prob == NULL) {
    err = memErr;
#if (ZLS_DEBUG >0)
    zlsio_ErrPrintf( " ERROR: zlsp_NewProblem allocation failed %d \n",err);
#endif
  }
  
  return (err);
}

int zlsp_DisposeProblem(ProblemPtr prob){
  int err = noErr;
  
  if ( prob != NULL ) {
    free(prob);
  }
  
  return (err);
}


//======================================================================
// Routine to set up a hydrostatic atmosphere in a gravity field or 
// a uniform/cloudy medium if no gravity field is present
//======================================================================

int zlsp_InitAtmosphere( ZonePtr z);
int zlsp_InitAtmosphere( ZonePtr z){

//======================================================================
// Routine to set up a hydrostatic atmosphere in a gravity field or 
// a uniform/cloudy medium if no gravity field is present
//======================================================================
  
  int err = noErr;
  
#ifdef GRAVITY
  Real phi;
#endif GRAVITY
  
  if (z != NULL){
    
#ifdef GRAVITY
    Field zoneF ;
    zls_GetZoneField(&zoneF, z); // nothing is allocated here, so no dispose is needed later
#endif GRAVITY
    
    WHD zoneW = z->zoneW;    // get zone physical variables
    
    Integer level = z->zoneLevel;
    
    ProblemPtr prob = gLevels[level].levelProblem;  // get the user's problem parameters
#ifdef COOLING
    PhysicsPtr phys = gLevels[0].levelPhysics;  // get the global Physics parameters
    CoolingPtr cp   = NULL;  // get the coolng parameters
    cp   = &(phys->coolParameters);  // get the coolng parameters
#endif

#ifdef SYM2D 
    Integer    i,  j;   
    Integer nti  = z->zoneGrid.nt[kXDim];
    Integer ntj  = z->zoneGrid.nt[kYDim];
    
#endif SYM2D 
    
#ifdef SYM3D 
    Integer    ic, jc, kc;   
    Integer    i, j, k;   
    Integer nti  = z->zoneGrid.nt[kXDim];
    Integer ntj  = z->zoneGrid.nt[kYDim];
    Integer ntk  = z->zoneGrid.nt[kZDim];
#endif SYM3D 
    
//    Integer seed = -1234;

    Real v_ismx = prob->gVISM[kXDim];
#if ( NDIMS > 1)
    Real v_ismy = prob->gVISM[kYDim];
#endif
#if ( NDIMS > 2)
    Real v_ismz = prob->gVISM[kZDim];
#endif

//======================================================================
// Atmosphere parameters
//======================================================================
    
    Real d1  = prob->gD0;
    Real P1  = prob->gP0;

//======================================================================
//
// Put on grid first to fill everywhere, replaced later by
// cool atmosphere in some or all locations
//
//======================================================================
    
#ifdef SYM2D 
    for(j = 0; j < ntj; j++){
      for(i = 0; i< nti; i++){
        
        zoneW.d [j][i]      = d1;
        zoneW.P [j][i]      = P1;
        
        zoneW.vx[j][i]      =  v_ismx;
        zoneW.vy[j][i]      =  v_ismy;
        
#if (NTRACERS > 0)    
        zoneW.tr[0][j][i]   =  0.0;
#endif
#if (NTRACERS > 1)    
        zoneW.tr[1][j][i]   =  0.0;
#endif
#if (NIONS == 1)    
        zoneW.ion[0][j][i]   = zlsp_XHI_T( T1, cp)*d1*f_nHtot;
#endif
        
      } //i
    } //j
#endif SYM2D 
#ifdef SYM3D 
    for(k = 0; k < ntk; k++){
      for(j = 0; j < ntj; j++){
        for(i = 0; i< nti; i++){
          
          zoneW.d [k][j][i]      = d1;
          zoneW.P [k][j][i]      = P1;
          
          zoneW.vx[k][j][i]      =  v_ismx;
          zoneW.vy[k][j][i]      =  v_ismy;
          zoneW.vz[k][j][i]      =  v_ismz;
          
#if (NTRACERS > 0)    
          zoneW.tr[0][k][j][i]   =  0.0;
#endif
#if (NTRACERS > 1)    
          zoneW.tr[1][k][j][i]   =  0.0;
#endif
#if (NIONS == 1)    
         zoneW.ion[0][k][j][i]   = zlsp_XHI_T( T1, cp)*d1*f_nHtot;
#endif
          
        } //i
      } //j
    } //k
#endif SYM3D 
        
  } // z != NULL
  
  return(err);
}

int zlsp_InitZoneWind( ZonePtr z );
int zlsp_InitZoneWind( ZonePtr z ){
  
//======================================================================
//  Initialise a zone system variables (ie W), usually self contained, 
//  here the initialisation is a bit messy so subroutines are called from
//  here.  When called all the variables for the zone are initialised and 
//  ready to go, grid coordinates, problem vars etc.
//======================================================================
  
  int err = noErr;
  
  if (z != NULL){
    
    Integer level = z->zoneLevel;
    
//======================================================================
// stellar wind
//
// clear some mass out of the initial region to reduce overloading the 
// young bubbles. Then call UserZone update incase a wind bubble is being
// used and it needs to write the initial wind onto the grid.
// writing the wind completely replaces the grid values, the clearing is
// only needed for the direct Edot and Mdot injection models.
//======================================================================
    
    if ( level == (gNLevels-1)) {
      
      WHD zoneW = z->zoneW;    
      ProblemPtr prob = gLevels[level].levelProblem;  // get the user's problem parameters
      
#ifdef SYM2D
      ACoordPtr xx  = z->zoneGrid.x;
      ACoordPtr yx  = z->zoneGrid.y;
      
      Integer nti   = z->zoneGrid.nt[kXDim];
      Integer ntj   = z->zoneGrid.nt[kYDim];
      Integer i, j;
      
      Real fi,  fj; // axis radial components
#endif SYM2D
      
#ifdef SYM3D
      ACoordPtr xx  = z->zoneGrid.x;
      ACoordPtr yx  = z->zoneGrid.y;
      ACoordPtr zx  = z->zoneGrid.z;
      
      Integer nti   = z->zoneGrid.nt[kXDim];
      Integer ntj   = z->zoneGrid.nt[kYDim];
      Integer ntk   = z->zoneGrid.nt[kZDim];
      Integer i, j, k;
      
      Real fi,  fj,  fk; // axis radial components
#endif SYM3D
      
        
        Counter starID;
        
        for ( starID = 0; starID<prob->gNstars; starID++){
          
          Real rWind  = prob->gRWindR0[starID] ;

#ifdef SYM2D
          
          Real x0       = prob->gXStar[starID];
          Real y0       = prob->gYStar[starID];
          
          Real x        = 0.0;
          Real y        = 0.0;
          Real dx        = 0.0;
          Real dy        = 0.0;
          Real radius   = 0.0;
          Real smoothFrac     = 1.0;
//======================================================================
// reduce the mass in the bubble inlet region to 0.1% of it's original 
//======================================================================
          
          for(j = 0; j < ntj; j++){
            y      = yx[j].c - y0;
            dy     = yx[j].dx;
            for(i = 0; i< nti; i++){
              x      = xx[i].c - x0;
              dx     = xx[i].dx;
              
              radius = zls_CvtCoordsToProjRadius2D( &fi, &fj, x, y);
              if ( radius < 2*rWind ){
                smoothFrac = 1.0-zls_SmoothCircle2D( rWind*1.1, 0.0, 0.0,
                                                     x,  y, dx, dy, 128, 3.0);
                smoothFrac *= (radius/rWind);
                smoothFrac  = smoothFrac > 0.001? smoothFrac: 0.001;
                smoothFrac  = smoothFrac > 1.0? 1.0: smoothFrac;
                zoneW.d[j][i] *= smoothFrac;
#if (NTRACERS > 1)    
//======================================================================
//   This tracer follows the fate of the cool HI ism , not needed 
//    if no hot  atmosphere is on the grid.
//======================================================================
                zoneW.tr[1][j][i]   =  zoneW.d[j][i];
#endif
              }
            } //i
          } //j
#endif SYM2D
          
#ifdef SYM3D
          
          Real x0       = prob->gXStar[starID];
          Real y0       = prob->gYStar[starID];
          Real z0       = prob->gZStar[starID];
          
          Real x        = 0.0;
          Real y        = 0.0;
          Real z        = 0.0;
          Real dx       = 0.0;
          Real dy       = 0.0;
          Real dz       = 0.0;
          Real radius   = 0.0;
          Real smoothFrac     = 1.0;
          
//======================================================================
// reduce the mass in the bubble inlet region to 0.1% of it's original 
//======================================================================
          
          for(k = 0; k < ntk; k++){
            z      = zx[k].c - z0;
            dz     = zx[k].dx;
            for(j = 0; j < ntj; j++){
              y      = yx[j].c - y0;
              dy     = yx[j].dx;
              for(i = 0; i< nti; i++){
                x      = xx[i].c - x0;
                dx     = xx[i].dx;
                radius = zls_CvtCoordsToProjRadius3D( &fi, &fj, &fk, x, y, z);
                if ( radius < 2*rWind ){
                  smoothFrac = 1.0-zls_SmoothSphere3D( rWind*1.1, 0.0, 0.0,0.0,
                                                       x,  y,  z, dx, dy, dz, 32, 3.0);
                  smoothFrac *= (radius/rWind);
                  smoothFrac  = smoothFrac > 0.001? smoothFrac: 0.001;
                  smoothFrac  = smoothFrac > 1.0? 1.0: smoothFrac;
                  
                  zoneW.d[k][j][i] *= smoothFrac;
#if (NTRACERS > 1)    
//======================================================================
//   This tracer follows the fate of the cool HI ism , not needed 
//    if no hot  atmosphere is onthe grid.
//======================================================================
                  zoneW.tr[1][k][j][i]   =  zoneW.d[k][j][i];
#endif

                }
              } //i
            } //j
          } //k
#endif SYM3D
          
          
        } // starID
        
      
      err = zlsp_UserZoneUpdate ( z ); // call wind boundary condition
      
    } // end wind on lowest level
    
  } else { err = memErr; } //z
  
  return(err);
}

//======================================================================
//  Standard zone initialsation routine
//======================================================================

int zlsp_InitZoneData( ZonePtr z ){
  
//======================================================================
//  Initialise a zone system variables (ie W), usually self contained, 
//  here the initialisation is a bit messy so subroutines are called from
//  here.  When called all the variables for the zone are initialised and 
//  ready to go, grid coordinates, problem vars etc.
//======================================================================
  
  int err = noErr;
  
  if (z != NULL){
    
    Integer level = z->zoneLevel;
    
    if ( level > 0) {
      err = zlsbc_FillZoneFromSuperZone( z, kInterpBCsFromSuper );
    } else {
      err = zlsp_InitAtmosphere(z);   
    }
        
  } else { err = memErr; }//z
  
  return(err);
}

int zlsp_InitLevelVelocityField(Integer level);
int zlsp_InitLevelVelocityField(Integer level){
  
  
  int err = noErr;
  Integer zone;
  ZonePtr  z;
  if ( level > 0) {
    err = zlsbc_ImposeAllLevelBCs(level, 0);
  }
   //======================================================================
   // intial turbulence
   //
   //======================================================================
  
  for ( zone = 0; zone <gLevels[level].nZones; zone++){
    z =  gLevels[level].levelZones[zone];
	  if (z != NULL){
#ifdef MPI	
		  if (z->zoneRankID == gMPIrank) {
#endif MPI
			  if ( z->zoneType == kAlgorithmZoneType ){ // only process appropriate zones, ie skip image zones
				  
#if (ZLS_DEBUG>0)
				  zlsio_ErrPrintf( " Initialising Zone: %2.2d:%2.2d \n", level, zone );
#endif

#ifdef DRIVEN_ISM      
				  err = zlsp_InitVelocityField( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelVelocityField: zlsp_InitVelocityField Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
#endif DRIVEN_ISM      
          
          err = zlsp_InitZoneWind( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelVelocityField: zlsp_InitZoneWind Failed: %d,  %d \n",zone, err );
					  return(err);
				  }

			  } // z->zoneType == kAlgorithmZoneType 
#ifdef MPI			
		  }
#endif MPI
	  } else { err = zonErr; }
  } // zone
  
  
  
  return(err);
  
}

//======================================================================
//  Standard level initialsation routine, loops through all zones of
//   a level. Called by ZLS after all parameters, coordinates, physics
//   etc have been set up.  Sets the gravity field arrays before the
//   calling zlsp_InitZoneData.
//======================================================================

int zlsp_InitLevelZones(Integer level){
  
  
  int err = noErr;
  Integer zone;
  ZonePtr  z;
  
#if (ZLS_VERBOSE >0)
#ifdef MPI
    if ( gMPIrank == 0){
#endif MPI
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " Level %d User Zones Initialisation: \n", level);
      zlsio_Printf( "%s\n",  DIVIDERLINE );
#ifdef MPI
    }
#endif MPI
#endif
  
  for ( zone = 0; zone <gLevels[level].nZones; zone++){
    z =  gLevels[level].levelZones[zone];
	  if (z != NULL){
#ifdef MPI	
		  if (z->zoneRankID == gMPIrank) {
#endif MPI
			  if ( z->zoneType == kAlgorithmZoneType ){ // only process appropriate zones, ie skip image zones
				  
#if (ZLS_DEBUG>0)
				  zlsio_ErrPrintf( " Initialising Zone: %2.2d:%2.2d \n", level, zone );
#endif
				  
				  err = zlsp_InitZoneData( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsp_InitZoneData Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
          
				  err = zls_SetZoneConstantBCValues( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zls_SetZoneConstantBCValues Failed: %d,  %d \n",zone, err );
					  return(err);
				  }

#ifdef MEANSPLITSWEEPS
#ifdef SYM2D
          err =  zlsa_Synch2DData(z);
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsa_Synch2DSymmetricSweeps Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
#endif SYM2D
#endif   MEANSPLITSWEEPS   
          
         
			  } // z->zoneType == kAlgorithmZoneType 
#ifdef MPI			
		  }
#endif MPI
	  } else { err = zonErr; }
  } // zone
  

  err =  zlsp_InitLevelVelocityField( level );
  
  return(err);
  
}

//======================================================================
// Optional routine to allow overriding the hard wired zlsp_InitProblem
// parameters from a config file, specified in zls.conf - usually init.conf
// Called by ZLS at the right time(tm)
//======================================================================

int   zlsp_SetUpProblemFromConfigFile(char * confFile){
  
//======================================================================
// Open the config file confFile, and parse for any user definded
// key,value pairs to set problem parameters (not grid initialisation)
// Calls zlsp_InitProblem, and before zlsp_InitLevelZones
//======================================================================
  
  int err = noErr;
  struct ccl_t  config;
  const char    *str1;
  char keyStr[512];
  
  Integer level;
  
  ProblemPtr  prob = NULL;
  err = zlsp_NewProblem(&prob);
  
  if ( prob != NULL ) {
    
    config.comment_char = '#';
    config.sep_char     = '=';
    config.str_char     = '"';
    
    if (ccl_parse(&config, confFile)==noErr) {
      
#if (ZLS_VERBOSE > 0) 
      zlsio_Printf( "\n%s\n", DIVIDERLINE );
#endif
      zlsio_Printf( " FYRIS Hydro Problem Initialised from %s \n", confFile);
#if (ZLS_VERBOSE > 0) 
      zlsio_Printf(  "%s\n", DIVIDERLINE );
#endif
      
// Default Values
      
      err = zlsp_InitProblem(prob);
      
//======================================================================
// The user is free to define any key strings and interpret the 
// returned values in any way.  Some problems are simple (see the SOD setup)
// this one is more complex as it sets up a multi-component model with many
// different types of parameters.
//======================================================================
            
            Real v_0 = prob->gVISM[kXDim];
            if (( str1 = ccl_get( &config,"V_ISMX"))){
              v_0 = atof( str1)*f_cmkm/f_v0;
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", "V_ISMX", str1);
#endif
            }
            prob->gVISM[kXDim] =  v_0; 
#if (NDIMS > 1)

            v_0 = prob->gVISM[kYDim];
            if (( str1 = ccl_get( &config,"V_ISMY"))){
              v_0 = atof( str1)*f_cmkm/f_v0;
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", "V_ISMY", str1);
#endif
            }
            prob->gVISM[kYDim] =  v_0; 
            
#endif
#if (NDIMS > 2)

            v_0 = prob->gVISM[kZDim];
            if (( str1 = ccl_get( &config,"V_ISMZ"))){
              v_0 = atof( str1)*f_cmkm/f_v0;
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", "V_ISMZ", str1);
#endif
            }
            prob->gVISM[kZDim] =  v_0; 
            
#endif

//======================================================================
// Atmosphere parameters
//======================================================================
      {
//
// WIERD C NOTE:
// brace sets up a local scope for temp variables to help reduce brief 
// named variables colliding with any used later in the routine. ie Th 
// could be used here and later in the cluster parameter setup without
// interfering with eachother
//
        
        Real P0 = prob->gP0;
        if (( str1 = ccl_get( &config, "Atmos_Pressure"))){
          P0 = atof( str1);
#if (ZLS_DEBUG>0)
          zlsio_Printf( "   %s = %s\n", "Atmos_Pressure", str1);
#endif
        }
        prob->gP0   = P0;
        
        Real D0 = prob->gD0;
        if (( str1 = ccl_get( &config, "Atmos_Density"))){
          D0 = atof( str1);
#if (ZLS_DEBUG>0)
          zlsio_Printf( "   %s = %s\n", "Atmos_Density", str1);
#endif
        }
        prob->gD0   = D0;

        
      }
      
      zlsio_Printf("%s\n",  DIVIDERLINE );
      zlsio_Printf( " Atmosphere Parameters:\n");
      zlsio_Printf("   d        %g \n", prob->gD0);    
      zlsio_Printf("   P        %g \n", prob->gP0); 
      zlsio_Printf("%s\n",  DIVIDERLINE );
      
        
        Integer n = prob->gNstars;
        if (( str1 = ccl_get( &config, "Nstars"))){
          n = atoi( str1);
#if (ZLS_DEBUG>0)
          zlsio_Printf( "   %s = %s\n", "Nstars", str1);
#endif
        }
        prob->gNstars  = n;
        
        
        zlsio_Printf( " Stellar Wind Parameters:\n");
        zlsio_Printf("   N_Stars   %d \n", prob->gNstars );   
        
        Integer i;
        for ( i = 0; i < n; i++) { // set up all the stars
          
          zlsio_Printf("   Star:\n" );   
          
          sprintf(keyStr, "Star_R0_%2.2d",i); // cm
          
          Real r_0 = prob->gRWindR0[i];
          if (( str1 = ccl_get( &config, keyStr))){
            r_0 = atof( str1);
#if (ZLS_DEBUG>0)
            zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
          }
          prob->gRWindR0[i] =  r_0/f_x0;
            
            
            sprintf(keyStr, "Star_X_%2.2d",i);
            
            Real x_0 = prob->gXStar[i];
            if (( str1 = ccl_get( &config, keyStr))){
              x_0 = atof( str1);
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            prob->gXStar[i] =  x_0;
#if (NDIMS>1)         
            
            sprintf(keyStr, "Star_Y_%2.2d",i); // pc
            
            Real y_0 = prob->gYStar[i];
            if (( str1 = ccl_get( &config, keyStr))){
              y_0 = atof( str1);
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            prob->gYStar[i] =  y_0;
#endif (NDIMS>1)          
            
#if (NDIMS>2)         
            sprintf(keyStr, "Star_Z_%2.2d",i); // pc
            
            Real z_0 = prob->gZStar[i];
            if (( str1 = ccl_get( &config, keyStr))){
              z_0 = atof( str1);
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            prob->gZStar[i] =  z_0;
#endif (NDIMS>2)          
            
            
            sprintf(keyStr, "Star_Dw_%2.2d",i); //             
            Real d_0 = prob->gDWind[i];
            if (( str1 = ccl_get( &config, keyStr))){
              d_0 = atof( str1);
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            prob->gDWind[i] =  d_0; // normalised
            
            
            sprintf(keyStr, "Star_Vw_%2.2d",i); // km/s
            
            Real v_0 = prob->gVWind[i];
            if (( str1 = ccl_get( &config, keyStr))){
              v_0 = atof( str1);
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            prob->gVWind[i] =  v_0; // normalised from km/s
            
            sprintf(keyStr, "Star_Pw_%2.2d",i); //             
            Real P_0 = prob->gPWind[i];
            if (( str1 = ccl_get( &config, keyStr))){
              P_0 = atof( str1);
#if (ZLS_DEBUG>0)
              zlsio_Printf( "   %s = %s\n", keyStr, str1);
#endif
            }
            prob->gPWind[i] =  P_0; // normalised

            zlsio_Printf("     %2.2d:  R0_Wind  %#8.3g \n", i,  prob->gRWindR0[i] ); 
            zlsio_Printf("     %2.2d:  V_Wind   %#8.3g \n", i,  prob->gVWind[i] ); 
            zlsio_Printf("     %2.2d:  d_Wind   %#8.3g \n"  , i, prob->gDWind[i] );    
            zlsio_Printf("     %2.2d:  P_Wind   %#8.3g \n", i, prob->gPWind[i] ); 

            zlsio_Printf("%s\n",  DIVIDERLINE );
            
            
        } // i
        
      
      
//======================================================================
// This code sets the problem into the level structures, mulit-zone
// multi-level or simple single structures 
//======================================================================
      
      
      gLevels[0].levelProblem = prob; // register the problem with all levels from 0 down.      
      if (gNLevels>1){
        for (level = 1; level < gNLevels; level++){
          gLevels[level].levelProblem = gLevels[0].levelProblem;
        }
      }
      
//======================================================================
// Now the level contents can be set - standard calls here dont' change
//======================================================================
      
      for (level = 0; level < gNLevels; level++){
        
#ifdef GRAVITY
        err =  zlsg_InitLevelGravityField( level ); // precompute the potential of to the gravity field over the level
        if (err != noErr ) {
          zlsio_ErrPrintf( " ERROR: zlsg_InitLevelGravityField: Level Field Failed: %d,  %d \n",level, err );
          return(err);
        }
#endif GRAVITY
#ifdef NEQIONISATION
        err =  zlsi_InitLevelIonisationState( level ); 
        if (err != noErr ) {
          zlsio_ErrPrintf( " ERROR: zlsi_InitLevelIonisationState: Level Failed: %d,  %d \n",level, err );
          return(err);
        }
#endif NEQIONISATION
        
        err = zlsp_InitLevelZones( level ); // last: user's levelProblem, initialised top down
        
      }
      
      if (gNLevels>1){
        for (level = (gNLevels-1); level > 0; level--){
          err = zlsbc_FillSuperLevelFromLevel(level);
        }
      }
      
#if (ZLS_VERBOSE>0)
      zlsio_Printf("%s\n",  DIVIDERLINE );
#else
      zlsio_LogPrintf("%s\n",  DIVIDERLINE );
#endif
      
      ccl_release(&config);
      
    }// config  
    
  } else {
    err = memErr;
    zlsio_ErrPrintf( " ERROR: Unable to allocate level 0 ProblemPtr \n");
    return (err);
  }
  
  return (err);
  
}

//======================================================================
// Routine for 'User; boundary cells set in grid.conf not used here
// but often used for jet inlet models
//======================================================================
#ifdef ZLS_USER_BOUNDARIES
int  zlsp_CalcUserZoneBCs ( ZonePtr z, int direction, int bndryKind ){
  return (noErr);
}
#endif ZLS_USER_BOUNDARIES


//======================================================================
// Routine for modifying anywhere in the grid in any way as part of
// a custom problem model. Not restricted to boundary cells, here used for 
// wind bubble/mass loss bubble energy injection and mass injection
// called after all other boundary conditions have been computed.
// Use sparingly, if not careful this routine can impact heavily on 
// the performance of the simulation.
//======================================================================
// 
// Normally deactivated (ie not even checked for calling) unless the macro
// ZLS_USER_UPDATE_ROUTINE is defined in zls_defs.  If not defined then
// this routine is not even callable - and this is the default setting.
// Also normally provided as an empty routine.
//

#ifdef ZLS_USER_UPDATE_ROUTINE

int  zlsp_UserZoneUpdate ( ZonePtr z  ){
  
  Integer err = noErr;
  ProblemPtr prob = NULL;
  PhysicsPtr phys = NULL;
  
  // A different type of BC, not the grid edges as all but setting cell values
  // on the grid.  Uses zoneW and is based on the zone data init routine rather than
  // the grid edge code of the basic boundary value routines.
  
  if ( z != NULL ){
    
    Integer level = z->zoneLevel;
    prob = gLevels[level].levelProblem;  // get the user's problem parameters
    phys = gLevels[level].levelPhysics;  // get the global eos parameters
    
    if ( level == (gNLevels-1)) {
      
#ifdef COOLING
    PhysicsPtr phys = gLevels[0].levelPhysics;  // get the global Physics parameters
    CoolingPtr cp   = NULL;  // get the coolng parameters
    cp   = &(phys->coolParameters);  // get the coolng parameters
#endif
//======================================================================
//  Inject a wind or just hot gas into multiple centres on the grid
//  to represent stars or clusters of stars.  Parameters above.
//  RSS-2006
//======================================================================
     
	   Real t      = zlst_GlobalTime;

     WHD zoneW    = z->zoneW;    

     Integer seed = zlst_GlobalCycles;
     zls_ran0_(&seed);
     
     Integer    i;
#ifdef SYM2D
     Integer    j;
     
     ACoordPtr xx  = z->zoneGrid.x;
     ACoordPtr yx  = z->zoneGrid.y;
     
     Real x        = 0.0;
     Real y        = 0.0;
     
     Integer nti   = z->zoneGrid.nt[kXDim];
     Integer ntj   = z->zoneGrid.nt[kYDim];
     

#endif SYM2D      
     	   
       if (t>=0.0){
         
// Real dt    = gLevels[level].levelClock.dt;
// Real  t    = gLevels[level].levelClock.t;
                
        Integer  starID = 0;
        
        for ( starID = 0; starID<prob->gNstars; starID++){
          
          Real d0     = prob->gDWind  [starID];
          Real P0     = prob->gPWind  [starID];
          Real v0     = prob->gVWind  [starID];
          Real r0     = prob->gRWindR0[starID];
                    
          Real radius;
          
#ifdef SYM2D
          
          Real fi,  fj; // axis radial components
          
//======================================================================
// add some artifical jitter to the centre of the injection region to 
// help smooth out grid artifacts when the region is small
// keep small < +/- CFL number , ie 0.5
//======================================================================
//          Real dr  = 0.25*zls_abs(zls_ran1_(&seed)+zls_ran1_(&seed)-1.0);
//          Real dp  = twopi*zls_ran1_(&seed);
//          Real dx  = dr*xx[0].dx*cos(dp);
//          Real dy  = dr*yx[0].dx*sin(dp);
         
          Real dx  = 0.0;
          Real dy  = 0.0;
          
          Real x0  = prob->gXStar[starID]+dx;
          Real y0  = prob->gYStar[starID]+dy;
          

            
            for(j = 0; j < ntj; j++){
              y      = yx[j].c - y0;
              for(i = 0; i< nti; i++){
                x      = xx[i].c - x0;
                
                radius = zls_CvtCoordsToProjRadius2D( &fi, &fj, x, y);
                
                if ( radius < (r0+kZLS_Small) ){
                  zoneW.d [j][i]      =  d0;
                  zoneW.P [j][i]      =  P0;
                 // zoneW.vx[j][i]      =  fi*v0*radius/r0;
                 // zoneW.vy[j][i]      =  fj*v0*radius/r0;
                  zoneW.vx[j][i]      =  fi*v0;
                  zoneW.vy[j][i]      =  fj*v0;
                  
#if (NTRACERS > 0)    
                  zoneW.tr[0][j][i]   = zoneW.d[j][i];
#endif
                } 
                
              } //j
            }//i
#endif SYM2D
          
          
#ifdef SYM3D
          Integer    i, j, k;
          
          ACoordPtr xx  = z->zoneGrid.x;
          ACoordPtr yx  = z->zoneGrid.y;
          ACoordPtr zx  = z->zoneGrid.z;
          
          Integer nti   = z->zoneGrid.nt[kXDim];
          Integer ntj   = z->zoneGrid.nt[kYDim];
          Integer ntk   = z->zoneGrid.nt[kZDim];
          
          
          Real fi,  fj, fk; // axis radial components
          
//======================================================================
// add some artifical jitter to the centre of the injection region to 
// help smooth out grid artifacts when the region is small
//======================================================================
          
//          seed = zlst_GlobalCycles;
          
//          Real dr  = 0.2*zls_abs(zls_ran1_(&seed)+zls_ran1_(&seed)-1.0);
//          Real dp  = twopi*zls_ran1_(&seed);
//          Real df  = pi*(zls_ran1_(&seed)-0.5);
//          Real dx  = dr*xx[0].dx*cos(dp);
//          Real dy  = dr*yx[0].dx*sin(dp);
//          Real dz  = dr*zx[0].dx*sin(df);
//         
          Real dx  = 0.0;
          Real dy  = 0.0;
          Real dz  = 0.0;
          
          Real x0  = prob->gXStar[starID]+dx;
          Real y0  = prob->gYStar[starID]+dy;
          Real z0  = prob->gZStar[starID]+dz;
          
          Real x        = 0.0;
          Real y        = 0.0;
          Real z        = 0.0;
            
            for(k = 0; k < ntk; k++){
              z      = zx[k].c - z0;
              for(j = 0; j < ntj; j++){
                y      = yx[j].c - y0;
                for(i = 0; i< nti; i++){
                  x      = xx[i].c - x0;
                  radius = zls_CvtCoordsToProjRadius3D( &fi, &fj, &fk, x, y, z);
                  
                  if ( radius < r0+kZLS_Small ){
                    zoneW.d [k][j][i]      =  d0;
                    zoneW.P [k][j][i]      =  P0;
                 //   zoneW.vx[k][j][i]      =  fi*v0*radius/r0;
                 //   zoneW.vy[k][j][i]      =  fj*v0*radius/r0;
                 //   zoneW.vz[k][j][i]      =  fk*v0*radius/r0;
                    
                    zoneW.vx[k][j][i]      =  fi*v0;
                    zoneW.vy[k][j][i]      =  fj*v0;
                    zoneW.vz[k][j][i]      =  fk*v0;
                    
#if (NTRACERS > 0)    
                    zoneW.tr[0][k][j][i]   = d0;
#endif
                  }                  
                } //i
              } //j
            }//k
            
          
#endif SYM3D
          
        } // star loop
  
    } // time
        
    } // level
    
  } else { err = memErr; } // z
  
  return (err);
}
#endif ZLS_USER_UPDATE_ROUTINE

  int zlsp_TextZoneBubbleAnalysis( zFile f, ZonePtr z){ 
  int err = noErr;
#ifdef BUBBLES
  //
  // Write out a row from a zone.  
  // in 1D k, and j are ignored
  // in 2D the jth row is written
  // in 3D the kth,jth row is written
  //
  //
  WHD               zoneW;
  Field             zoneF;
  
  PhysicsPtr phys  = NULL;
  Real       gamma = kZLS_eos_gamma;  // global plasma adiabtic index
  Real       gamm  = kZLS_eos_gamma_1;  // global plasma adiabtic index-1.0
  Real       mu    = k_mu;  // global plasma mu

  CoolingPtr cp    = NULL;
  
#ifdef COOLING
  
  Real      fH[3] = {1.0, 0.0, 0.0};  // atom fractions of all H atoms, H2, HI, HII (= 0.0 here)
  
#ifdef VARIABLE_EOS
#ifdef NEQIONISATION
  Real      nd_HI; // HI atom number density
  Real     nd_tot; // total H nuclei density
#endif NEQIONISATION
#endif VARIABLE_EOS
  Real        C  = 0.0;  // cooling rate
  Real      Ctot = 0.0;  // atomic/ion cooling rate
  Real       CAt = 0.0;  // atomic/ion cooling rate
  Real       CH2 = 0.0;  // H2 cooling rate
  Real       CCO = 0.0;  // CO cooling rate
  Real      Cnet = 0.0;  // net cooling - heating rate
#endif COOLING
 
  Real            d0;
  Real            P0;
  Real           Pd0; // P0/d0
  Real             e; // internal energy
  Real             T;  // Temperature
#if (NTRACERS > 0)    
  Real            tr0;
#endif
    
#ifdef SYM1D
  ACoordPtr          xx;
#endif SYM1D
#ifdef SYM2D
  ACoordPtr          xx, yx;
#endif SYM2D
#ifdef SYM3D
  ACoordPtr          xx, yx, zx;
#endif SYM3D
  Integer level =  z->zoneLevel;
  
  IOSettingsPtr io = gLevels[level].levelIO;
  
  ProblemPtr prob = gLevels[level].levelProblem;  // get the user's problem parameters
  
  phys  = gLevels[level].levelPhysics;
  if (phys != NULL) {
      gamma  = phys->gamma;  // global plasma adiabtic index
      gamm   = phys->gamm;  // global plasma adiabtic index -1.0
      mu     = phys->mu;     // global plasma mu
#ifdef COOLING
      cp    = &(phys->coolParameters);
#endif COOLING  
  }
  
  if ((io != NULL) && (z != NULL)){
    
    err     = zlsp_GetZoneWHD(&zoneW, z);
    err     = zls_GetZoneField(&zoneF, z);
    
    if (err == noErr) {
      
      
      Integer i;
      Integer is  =  z->zoneGrid.ns[kXDim];
      Integer ie  =  z->zoneGrid.ne[kXDim];
#ifdef SYM2D
      Integer j;
      Integer js  =  z->zoneGrid.ns[kYDim];
      Integer je  =  z->zoneGrid.ne[kYDim];
#endif SYM2D
#ifdef SYM3D
      Integer j;
      Integer js  =  z->zoneGrid.ns[kYDim];
      Integer je  =  z->zoneGrid.ne[kYDim];
      Integer k;
      Integer ks  =  z->zoneGrid.ns[kZDim];
      Integer ke  =  z->zoneGrid.ne[kZDim];
#endif SYM3D
      
#ifdef SYM1D
      xx = z->zoneGrid.x;
#endif SYM1D
#ifdef SYM2D
      xx = z->zoneGrid.x;
      yx = z->zoneGrid.y;
#endif SYM2D
#ifdef SYM3D
      xx = z->zoneGrid.x;
      yx = z->zoneGrid.y;
      zx = z->zoneGrid.z;
#endif SYM3D
      
      if (f) {
        
        zlsio_ZFPrintf( f, " Zone ID = %d\n", z->zoneID);
 	      Real t      = zlst_GlobalSeconds;
       
        Counter bubbleCells   =   0;
        Real bubbleRadius     = 0.0;
        
#if (NTRACERS > 0)    
        Counter bubbleWCells  =   0;
        Real bubbleWVolume    = 0.0;
        Real bubbleWRadius    = 0.0;
#endif
        
        Real bubbleDRadius    = 0.0;
        Real bubbleD2Radius   = 0.0;
        Real bubblePRadius    = 0.0;
        Real bubbleERadius    = 0.0;
        Real bubbleKERadius   = 0.0;
        Real bubbleMomRadius  = 0.0;
        
        Real bubbleWint       = 0.0;
        Real bubbleDint       = 0.0;
        Real bubbleD2int      = 0.0;
        Real bubblePint       = 0.0;
        Real bubbleEint       = 0.0;
        Real bubbleKEint      = 0.0;
        Real bubbleMomint     = 0.0;
        
        Real kE, M;
        
#ifdef SYM2D
        Real bubbleArea       = 0.0;
        Real cellArea = xx[0].dx*yx[0].dx; // cartesian only for now
        Real fi,  fj;      // axis radial components
        Real x0   = prob->gXStar[0];
        Real y0   = prob->gYStar[0];
        
        Real x        = 0.0;
        Real y        = 0.0;
        
          for(j = js; j< je; j++){
            y      = yx[j].c - y0;
            for(i = is; i< ie; i++){
              x      = xx[i].c - x0;
              
              
              d0   = zoneW.d[j][i];
              P0   = zoneW.P[j][i];
#if (NTRACERS > 0)    
              tr0  = zoneW.tr[0][j][i];
#endif
              
              Pd0  = P0/d0;
              
#ifdef NEQIONISATION
#if ( NIONS == 1)
              nd_HI = zoneW.ion[0][j][i]; // HI atom number density
              nd_tot = zlsp_fH_ndHd( fH, nd_HI, d0);
#endif
#if ( NIONS == 2)
              nd_HI = zoneW.ion[0][j][i]; // HI atom number density
              nd_HI = zoneW.ion[1][j][i]; // HII ion number density
              nd_tot = zlsp_fH_ndHd( fH, nd_HI, nd_HII, d0);
#endif
              gamma  = zlsp_Gamma_fH(fH);
              T = zlsp_T_Pd(Pd0, fH, cp); 
#else
              gamma  = zlsp_Gamma_Pd(Pd0, cp);
              T      = zlsp_T_Pd(Pd0, cp); // works also for variable Mu and gamma, cp may be null if not needed
#ifdef H_IONISATION
              fH[1] = zlsp_XHI_T ( T, cp );  
              fH[2] = zlsp_XHII_T( T, cp );
              fH[0] = 1.0-(fH[1]+fH[2]);
#endif H_IONISATION
#endif !NEQIONISATION
              gamm = gamma -1.0;
              
              e    = P0/gamm;
              kE   = d0*zlsp_KE_L_WHD( zoneW, j, i); // 0.5 m v^2
              M    = d0*zlsp_Mom_L_WHD( zoneW, j, i); // |m v|
              
              
#ifdef COOLING
              C    = 0.0;
              if ( cp->coolType != kPhysFileCooling) {
#ifdef NEQIONISATION
                C  = d0*d0*zlsp_Lambda(T,fH,cp);
#else
                C  = d0*d0*zlsp_Lambda(T,cp);
#endif !NEQIONISATION
                
              } else {
                
#ifdef NEQIONISATION
                C  = zlsp_LambdaComponents(&Ctot, &CAt, &CH2, &CCO, T, fH, cp);
#else
                C  = zlsp_LambdaComponents(&Ctot, &CAt, &CH2, &CCO, T, cp);
#endif !NEQIONISATION
                Ctot *= d0*d0;
                CAt  *= d0*d0;
                CH2  *= d0*d0;
                CCO  *= d0*d0;
                C    *= d0*d0;
              }
#endif COOLING
              
              if (kE > 1.0e-6) {
                
                bubbleRadius = zls_CvtCoordsToProjRadius2D( &fi, &fj, x, y);
                
                bubbleCells      += 1;
#if (NTRACERS > 0)    
                bubbleWCells     += (tr0>1.0e-6)? 1:0;
                bubbleWint       += tr0;
                bubbleWRadius    += tr0*bubbleRadius;
#endif
                
                bubbleDint       += d0;
                bubbleD2int      += d0*d0;
                bubblePint       += P0;
                
                bubbleDRadius    += d0*bubbleRadius;
                bubbleD2Radius   += d0*d0*bubbleRadius;
                bubblePRadius    += P0*bubbleRadius;
                
                bubbleEint       += e;
                bubbleKEint      += kE;
                bubbleMomint     += M;
                
                bubbleERadius    +=  e*bubbleRadius;
                bubbleKERadius   += kE*bubbleRadius;
                bubbleMomRadius  += M*bubbleRadius;
                
              } 
              
            } // i
          } // j
        
        bubbleArea  = bubbleCells*cellArea;
        
#ifdef MPI  
  Real sum;    
  MPI_Allreduce( &bubbleArea, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleArea = sum;

  MPI_Allreduce( &bubbleDint, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleDint = sum;
  
  MPI_Allreduce( &bubbleD2int, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleD2int = sum;
  
  MPI_Allreduce( &bubblePint, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubblePint = sum;
  
  MPI_Allreduce( &bubbleEint, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleEint = sum;
  
  MPI_Allreduce( &bubbleKEint, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleKEint = sum;
  
  MPI_Allreduce( &bubbleMomint, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleMomint = sum;
  
  
  MPI_Allreduce( &bubbleDRadius, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleDRadius = sum;
  
  MPI_Allreduce( &bubbleD2Radius, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleD2Radius = sum;
  
  MPI_Allreduce( &bubblePRadius, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubblePRadius = sum;
  
  MPI_Allreduce( &bubbleERadius, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleERadius = sum;
  
  MPI_Allreduce( &bubbleKERadius, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleKERadius = sum;
  
  MPI_Allreduce( &bubbleMomRadius, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleMomRadius = sum;
#endif MPI

  Real effectiveRadius  = zls_Pow(bubbleArea/pi, 0.5);
  
#if (NTRACERS > 0)    
  bubbleWArea = bubbleWCells*cellArea;
#ifdef MPI  
  MPI_Allreduce( &bubbleWArea, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleWArea = sum;

  MPI_Allreduce( &bubbleWint, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleWint = sum;  
  
  MPI_Allreduce( &bubbleWRadius, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleWRadius = sum;
#endif MPI
  Real effectiveWRadius = zls_Pow(bubbleArea/pi, 0.5);
  bubbleWRadius    /= bubbleWint*effectiveRadius;
#endif
        
        bubbleDRadius    /= bubbleDint*effectiveRadius;
        bubbleD2Radius   /= bubbleD2int*effectiveRadius;
        bubblePRadius    /= bubblePint*effectiveRadius;
        bubbleERadius    /= bubbleEint*effectiveRadius;
        bubbleKERadius   /= bubbleKEint*effectiveRadius;
        bubbleMomRadius  /= bubbleMomint*effectiveRadius;
        
#ifdef MPI
  if ( gMPIrank == 0) {
#endif MPI
#if (NTRACERS > 0)    

        zlsio_ZFPrintf(f,"Level \t t \t Area \t R_eff \t RWeff \t ratio \t R_W \t R_D \t R_D2 \t R_P \t R_E \t R_KE \t R_mv \n");
        zlsio_ZFPrintf(f,"L%d \t  %12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \n", 
        level, t, bubbleArea, effectiveRadius, effectiveWRadius, effectiveWRadius/effectiveRadius,
        bubbleWRadius,bubbleDRadius, bubbleD2Radius, bubblePRadius,
        bubbleERadius, bubbleKERadius, bubbleMomRadius );
        zlsio_LogPrintf("Level \t t \t Volume \t R_eff \t RWeff \t ratio \t R_W \t R_D \t R_D2 \t R_P \t R_E \t R_KE \t R_mv \n");
        zlsio_LogPrintf("L%d \t %12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \n", 
        level, t, bubbleArea, effectiveRadius, effectiveWRadius, effectiveWRadius/effectiveRadius,
        bubbleWRadius,bubbleDRadius, bubbleD2Radius, bubblePRadius,
        bubbleERadius, bubbleKERadius, bubbleMomRadius );
#else
        zlsio_ZFPrintf(f,"Level \t t \t Area \t R_eff \t RWeff \t R_D \t R_D2 \t R_P \t R_E \t R_KE \t R_mv \n");
        zlsio_ZFPrintf(f,"L%d \t %12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \n", 
        level, t, bubbleArea, effectiveRadius,
        bubbleDRadius, bubbleD2Radius, bubblePRadius,
        bubbleERadius, bubbleKERadius, bubbleMomRadius );
        zlsio_LogPrintf("Level \t t \t Volume \t R_eff \t RWeff \t R_D \t R_D2 \t R_P \t R_E \t R_KE \t R_mv \n");
        zlsio_LogPrintf("L%d \t %12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \n", 
        level, t, bubbleArea, effectiveRadius,
        bubbleDRadius, bubbleD2Radius, bubblePRadius,
        bubbleERadius, bubbleKERadius, bubbleMomRadius );
#endif
#ifdef MPI
  }
#endif MPI
        
#endif SYM2D
         
#ifdef SYM3D
        Real bubbleVolume     = 0.0;
        Real cellVolume = xx[0].dx*yx[0].dx*zx[0].dx; // cartesian only for now
        Real fi,  fj, fk;      // axis radial components
        Real x0   = prob->gXStar[0];
        Real y0   = prob->gYStar[0];
        Real z0   = prob->gZStar[0];
        
        Real x        = 0.0;
        Real y        = 0.0;
        Real z        = 0.0;
        
        for(k = ks; k< ke; k++){
          z      = zx[k].c - z0;
          for(j = js; j< je; j++){
            y      = yx[j].c - y0;
            for(i = is; i< ie; i++){
              x      = xx[i].c - x0;
              
              
              d0   = zoneW.d[k][j][i];
              P0   = zoneW.P[k][j][i];
#if (NTRACERS > 0)    
              tr0  = zoneW.tr[0][k][j][i];
#endif
              
              Pd0  = P0/d0;
              
#ifdef NEQIONISATION
#if ( NIONS == 1)
              nd_HI = zoneW.ion[0][k][j][i]; // HI atom number density
              nd_tot = zlsp_fH_ndHd( fH, nd_HI, d0);
#endif
#if ( NIONS == 2)
              nd_HI = zoneW.ion[0][k][j][i]; // HI atom number density
              nd_HI = zoneW.ion[1][k][j][i]; // HII ion number density
              nd_tot = zlsp_fH_ndHd( fH, nd_HI, nd_HII, d0);
#endif
              gamma  = zlsp_Gamma_fH(fH);
              T = zlsp_T_Pd(Pd0, fH, cp); 
#else
              gamma  = zlsp_Gamma_Pd(Pd0, cp);
              T      = zlsp_T_Pd(Pd0, cp); // works also for variable Mu and gamma, cp may be null if not needed
#ifdef H_IONISATION
              fH[1] = zlsp_XHI_T ( T, cp );  
              fH[2] = zlsp_XHII_T( T, cp );
              fH[0] = 1.0-(fH[1]+fH[2]);
#endif H_IONISATION
#endif !NEQIONISATION
              gamm = gamma -1.0;
              
              e    = P0/gamm;
              kE   = d0*zlsp_KE_L_WHD( zoneW, k, j, i); // 0.5 m v^2
              M    = d0*zlsp_Mom_L_WHD( zoneW, k, j, i); // |m v|
              
              
#ifdef COOLING
              C    = 0.0;
              if ( cp->coolType != kPhysFileCooling) {
#ifdef NEQIONISATION
                C  = d0*d0*zlsp_Lambda(T,fH,cp);
#else
                C  = d0*d0*zlsp_Lambda(T,cp);
#endif !NEQIONISATION
                
              } else {
                
#ifdef NEQIONISATION
                C  = zlsp_LambdaComponents(&Ctot, &CAt, &CH2, &CCO, T, fH, cp);
#else
                C  = zlsp_LambdaComponents(&Ctot, &CAt, &CH2, &CCO, T, cp);
#endif !NEQIONISATION
                Ctot *= d0*d0;
                CAt  *= d0*d0;
                CH2  *= d0*d0;
                CCO  *= d0*d0;
                C    *= d0*d0;
              }
#endif COOLING
              
              if (kE > 1.0e-6) {
                
                bubbleRadius = zls_CvtCoordsToProjRadius3D( &fi, &fj, &fk, x, y, z);
                
                bubbleCells      += 1;
#if (NTRACERS > 0)    
                bubbleWCells     += (tr0>1.0e-6)? 1:0;
                bubbleWint       += tr0;
                bubbleWRadius    += tr0*bubbleRadius;
#endif
                
                bubbleDint       += d0;
                bubbleD2int      += d0*d0;
                bubblePint       += P0;
                
                bubbleDRadius    += d0*bubbleRadius;
                bubbleD2Radius   += d0*d0*bubbleRadius;
                bubblePRadius    += P0*bubbleRadius;
                
                bubbleEint       += e;
                bubbleKEint      += kE;
                bubbleMomint     += M;
                
                bubbleERadius    +=  e*bubbleRadius;
                bubbleKERadius   += kE*bubbleRadius;
                bubbleMomRadius  += M*bubbleRadius;
                
              } 
              
            } // i
          } // j
        } // k
        
        bubbleVolume  = bubbleCells*cellVolume;
#ifdef MPI  
  Real sum;    
  MPI_Allreduce( &bubbleVolume, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleVolume = sum;

  MPI_Allreduce( &bubbleDint, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleDint = sum;
  
  MPI_Allreduce( &bubbleD2int, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleD2int = sum;
  
  MPI_Allreduce( &bubblePint, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubblePint = sum;
  
  MPI_Allreduce( &bubbleEint, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleEint = sum;
  
  MPI_Allreduce( &bubbleKEint, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleKEint = sum;
  
  MPI_Allreduce( &bubbleMomint, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleMomint = sum;
  
  
  MPI_Allreduce( &bubbleDRadius, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleDRadius = sum;
  
  MPI_Allreduce( &bubbleD2Radius, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleD2Radius = sum;
  
  MPI_Allreduce( &bubblePRadius, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubblePRadius = sum;
  
  MPI_Allreduce( &bubbleERadius, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleERadius = sum;
  
  MPI_Allreduce( &bubbleKERadius, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleKERadius = sum;
  
  MPI_Allreduce( &bubbleMomRadius, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleMomRadius = sum;
#endif !MPI

        Real effectiveRadius  = zls_Pow((3/fourpi)*bubbleVolume, 0.3333333333);
#if (NTRACERS > 0)    
        bubbleWVolume = bubbleWCells*cellVolume;
#ifdef MPI  
  MPI_Allreduce( &bubbleWVolume, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleWVolume = sum;

  MPI_Allreduce( &bubbleWint, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleWint = sum;  
  
  MPI_Allreduce( &bubbleWRadius, &sum, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  bubbleWRadius = sum;
#endif !MPI
        Real effectiveWRadius = zls_Pow((3/fourpi)*bubbleWVolume, 0.3333333333);
        bubbleWRadius    /= bubbleWint*effectiveRadius;
#endif
        
        bubbleDRadius    /= bubbleDint*effectiveRadius;
        bubbleD2Radius   /= bubbleD2int*effectiveRadius;
        bubblePRadius    /= bubblePint*effectiveRadius;
        bubbleERadius    /= bubbleEint*effectiveRadius;
        bubbleKERadius   /= bubbleKEint*effectiveRadius;
        bubbleMomRadius  /= bubbleMomint*effectiveRadius;
        
#ifdef MPI
  if ( gMPIrank == 0) {
#endif MPI
#if (NTRACERS > 0)    

        zlsio_ZFPrintf(f,"Level t \t Volume \t R_eff \t RWeff \t ratio \t R_W \t R_D \t R_D2 \t R_P \t R_E \t R_KE \t R_mv \n");
        zlsio_ZFPrintf(f,"L%d %12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \n", 
        level, t, bubbleVolume, effectiveRadius, effectiveWRadius, effectiveWRadius/effectiveRadius,
        bubbleWRadius,bubbleDRadius, bubbleD2Radius, bubblePRadius,
        bubbleERadius, bubbleKERadius, bubbleMomRadius );
        zlsio_LogPrintf("Level t \t Volume \t R_eff \t RWeff \t ratio \t R_W \t R_D \t R_D2 \t R_P \t R_E \t R_KE \t R_mv \n");
        zlsio_LogPrintf("L%d %12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \n", 
        level, t, bubbleVolume, effectiveRadius, effectiveWRadius, effectiveWRadius/effectiveRadius,
        bubbleWRadius,bubbleDRadius, bubbleD2Radius, bubblePRadius,
        bubbleERadius, bubbleKERadius, bubbleMomRadius );
        
        
#else
        zlsio_ZFPrintf(f,"Level t \t Volume \t R_eff \t RWeff \t R_D \t R_D2 \t R_P \t R_E \t R_KE \t R_mv \n");
        zlsio_ZFPrintf(f,"L%d %12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \n", 
        level, t, bubbleVolume, effectiveRadius,
        bubbleDRadius, bubbleD2Radius, bubblePRadius,
        bubbleERadius, bubbleKERadius, bubbleMomRadius );
        zlsio_LogPrintf("Level t \t Volume \t R_eff \t RWeff \t R_D \t R_D2 \t R_P \t R_E \t R_KE \t R_mv \n");
        zlsio_LogPrintf("L%d %12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \t %#12.5g \n", 
        level, t, bubbleVolume, effectiveRadius,
        bubbleDRadius, bubbleD2Radius, bubblePRadius,
        bubbleERadius, bubbleKERadius, bubbleMomRadius );

#endif
#ifdef MPI
  }
#endif MPI
        
#endif SYM3D
        
      } else { err = fileErr; }
      
      zlsp_ReleaseZoneWHD(zoneW);
      
    } // zoneW
    
  }  else { //z, io
    err = memErr;
  }
#endif BUBBLES
  
  return(err);
}

