#include <zls.h>

//====================================================================================
//
// Main sweep entry point, takes care of single or multi-level models.
//
//====================================================================================

// call zlsa_Sweep(l) to process level l and sub levels below if any.

int zlsa_Sweep(Integer level){
  
  int err = noErr;
  
  if ( gNLevels > 1) { 
#ifdef REFINED_SWEEPS
    err = zlsa_RefinedSweep(level, kSubLevelIterations); 
    if (err != noErr){
      zlsio_ErrPrintf( " ERROR: zlsa_Sweep in zlsa_RefinedSweep %d \n", err);
      return(err);
    }
#endif REFINED_SWEEPS
#ifdef UNIFORM_SWEEPS
    err = zlsa_UniformSweep(level, kSubLevelIterations); 
    if (err != noErr){
      zlsio_ErrPrintf( " ERROR: zlsa_Sweep in zlsa_UniformSweep %d \n", err);
      return(err);
    }
#endif UNIFORM_SWEEPS
  } else {
    err = zlsa_DirectSweep(level, 0); 
    if (err != noErr){
      zlsio_ErrPrintf( " ERROR: zlsa_Sweep in zlsa_DirectSweep %d \n", err);
      return(err);
    }
  }
  return(err);
}


#ifdef MEANSPLITSWEEPS
#ifdef SYM2D


int zlsa_Synch2DData    ( ZonePtr z);
int zlsa_Synch2DData( ZonePtr z){
	  
  Integer err = noErr;
  
  if ( z != NULL ){
    
    WHD zoneWL = z->zoneWLeft;    
    WHD zoneWR = z->zoneWRight;    
    
    Integer nti   = z->zoneGrid.nt[kXDim];
    Integer ntj   = z->zoneGrid.nt[kYDim];
    
    Integer i, j;
    
    for(j = 0; j < ntj; j++){
      for(i = 0; i< nti; i++){
        
        zoneWR.d [j][i] = zoneWL.d [j][i];
        zoneWR.P [j][i] = zoneWL.P [j][i];
        zoneWR.vx[j][i] = zoneWL.vx[j][i];
        zoneWR.vy[j][i] = zoneWL.vy[j][i];

#if (NTRACERS > 0)    
        Integer k;
        for (k = 0; k < kNTracers; k++){
          zoneWR.tr[k][j][i]     = zoneWL.tr[k][j][i] ;
        }
#endif
        
      } //i
    } //j
    
    // default
    z->zoneW = z->zoneWLeft;     
    
  } else { err = memErr; } // z
  
  return (err);
}

int zlsa_Average2DSweeps( ZonePtr z);
int zlsa_Average2DSweeps( ZonePtr z){
	
  // symmetrise left and right sweeps
  
  Integer err = noErr;
  
  if ( z != NULL ){
    
    WHD zoneWL = z->zoneWLeft;    
    WHD zoneWR = z->zoneWRight;    
    
    Integer nti   = z->zoneGrid.nt[kXDim];
    Integer ntj   = z->zoneGrid.nt[kYDim];
    
    Integer i, j;
    
    Real d, P; 
    Real vx, vy; 
#if (NTRACERS > 0)    
    Real a; 
#endif    

//    zlsio_Printf("Averaging Sweeps\n");
    
    for(j = 0; j < ntj; j++){
      for(i = 0; i< nti; i++){
        
        d = 0.5*(zoneWL.d [j][i] + zoneWR.d [j][i]);
        zoneWL.d [j][i] = d;
        zoneWR.d [j][i] = d;
        
        P = 0.5*(zoneWL.P [j][i] + zoneWR.P [j][i]);
        zoneWL.P [j][i] = P;
        zoneWR.P [j][i] = P;
        
        vx = 0.5*(zoneWL.vx[j][i] + zoneWR.vx[j][i]);
        zoneWL.vx[j][i] = vx;
        zoneWR.vx[j][i] = vx;
        
        vy = 0.5*(zoneWL.vy[j][i] + zoneWR.vy[j][i]);
        zoneWL.vy[j][i] = vy;
        zoneWR.vy[j][i] = vy;

#if (NTRACERS > 0)    
        Integer k;
        for (k = 0; k < kNTracers; k++){
          a = 0.5*(zoneWR.tr[k][j][i] +  zoneWL.tr[k][j][i]);
          zoneWL.tr[k][j][i]     = a;
          zoneWR.tr[k][j][i]     = a;
        }
#endif
        
      } //i
    } //j
    
    // default
    z->zoneW = z->zoneWLeft;     
    
  } else { err = memErr; } // z
  
  return (err);
}
#endif SYM2D
#endif MEANSPLITSWEEPS

//====================================================================================
//
#ifdef PTHREADS
#include  <pthread.h>
int gSweepLock = 0;
pthread_mutex_t gSweepMutex    = PTHREAD_MUTEX_INITIALIZER; // mutex for gSweepLock and comms
pthread_cond_t  gSweepCond     = PTHREAD_COND_INITIALIZER; // clear to sweep, each independent so no mutexes needed on sweeps themselves
pthread_cond_t  gNextSweepCond = PTHREAD_COND_INITIALIZER; // request next sweep trigger or end


int zlsa_QueuePThreadSweep     ( ZonePtr z );
void *  zlsa_PThreadSweeps        ( void * vz );
int zlsa_WaitPThreadSweeps  ( int nThreads ); // conduct sweeper threads

int zlsa_QueuePThreadSweep  ( ZonePtr z ) {
  int err = noErr;
  pthread_t thread;
  pthread_attr_t attr;
  
  err = pthread_attr_init(&attr);
  err = pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
  err = pthread_create(&thread, &attr, zlsa_PThreadSweeps, (void *)z);
  err = pthread_attr_destroy(&attr);
  
  return(err);
}

void *  zlsa_PThreadSweeps  ( void * vz ) {
  
  int err = noErr;
  ZonePtr z = (ZonePtr)vz;
  
  pthread_mutex_lock            ( &gSweepMutex);
  if (gSweepLock >0) pthread_cond_broadcast  ( &gNextSweepCond );
  pthread_cond_wait             ( &gSweepCond, &gSweepMutex); // first time in, sit and wait for gSweepCond launches
  pthread_mutex_unlock          ( &gSweepMutex);
  
  // sweep zone of thread in parallel, OK as zones are independent here while zlsa_WaitSweepThreads is mastering
  
#ifdef ZLS_LOADSORTING
  zlst_StartCPUTimes ( &z->zoneCPUTimer );
#endif ZLS_LOADSORTING
#ifdef DIVVHCORRECTION
  zlsa_CalcZoneDivV(z);
#endif DIVVHCORRECTION
  
#ifdef MEANSPLITSWEEPS
  
#ifdef SYM1D  
  err = ZLSSWEEPS   ( z ); // macro for sweep method in force. see zls_defs.h
#endif SYM1D  

#ifdef SYM2D  
  err = zlsa_Synch2DData ( z ); // make sure boundaries are on Alt vars too.
  
  z->zoneW = z->zoneWLeft ; // just to be safe, should be set by Synch
  err = ZLSSWEEPS   ( z ) ; // macro for sweep method in force. see zls_defs.h
  z->zoneClock.cycles += 1;
  
  z->zoneW = z->zoneWRight; // sweep Alt vars in opposite sweep order
  err = ZLSSWEEPS   ( z ) ; // macro for sweep method in force. see zls_defs.h
  z->zoneClock.cycles -= 1;
  
  err = zlsa_Average2DSweeps ( z );
#endif SYM2D 
  
#ifdef SYM3D  
  err = ZLSSWEEPS   ( z ); // macro for sweep method in force. see zls_defs.h
#endif SYM3D  
  
#else
  
  err = ZLSSWEEPS   ( z ); // macro for sweep method in force. see zls_defs.h
  
#endif !MEANSPLITSWEEPS
  
  err = zlst_AdvanceClock( &z->zoneClock );
  
  err = zlst_CalcZoneNewTimeScale( z );

#ifdef ZLS_LOADSORTING
  zlst_EndCPUTimes ( &z->zoneCPUTimer );
#endif ZLS_LOADSORTING
  
  pthread_mutex_lock             ( &gSweepMutex);
#if (ZLS_DEBUG > 0)
  zlsio_Printf( " *** Sweep %d done. Cycle: %d \n",(int)z->zoneID, (int)z->zoneClock.cycles );
#endif
  gSweepLock--;     // if gSweepLock hits 0 then it is all over
  pthread_cond_broadcast  ( &gNextSweepCond ); // request next or end of all sweeps (gSweepLock hits 0)
  pthread_mutex_unlock    ( &gSweepMutex);
  pthread_exit(NULL); // thread done
  
}

int zlsa_WaitPThreadSweeps( int nThreads ){ // conduct sweeper threads
  int err = noErr;
  err = pthread_mutex_lock       ( &gSweepMutex );  
  gSweepLock = nThreads;
  while ( gSweepLock > 0 ) {
    err = pthread_cond_broadcast ( &gSweepCond );
    err = pthread_cond_wait      ( &gNextSweepCond, &gSweepMutex );
  }
#if (ZLS_DEBUG > 0)
  zlsio_Printf( " >>>> %d Thread Sweeps Over\n", nThreads );
#endif
  gSweepLock = 0;
  err = pthread_mutex_unlock     ( &gSweepMutex );
  return(err);
}

#endif PTHREADS

//====================================================================================
//
// Sweep and Advance all zones in level, including any refined zones on the level below
// ***and pass results to super level***.  Recursively applied to sub levels.
// Call RefinedSweep(0,0) to process whole structure.  Calls Direct Sweep on each zone, 
// and each zone has a dt set before a call to refined sweep.
//
//====================================================================================

int zlsa_RefinedSweep(Integer level, Integer iteration){ // call RefinedSweep(0, 0) to process whole structure.
	
  int err = noErr;
  Integer iter;
	
  if (level >= 0) {
  
		// 0 = top level, gNLevels-1 is bottom of recursion
		
    if ((level+1) >= gNLevels) {
			
      // bottom of recursion, plus catch any level values that are too large
			
      if (level < gNLevels ) { // only sensible levels
#if ( ZLS_DEBUG > 1 )
        fprintf( stdout,  " Sweeping level %d \n", level); 
#endif
        err = zlsa_DirectSweep ( level, iteration );
				
      } else { 
        err = levErr; 
        zlsio_ErrPrintf( " LEVEL ERROR: level number %d out of range %d \n", gNLevels, level+1); 
        return (err); 
      } // (level < gNLevels )
			
    } else {
			
      if ((level+1) < gNLevels) {
        for (iter = 0; iter<gLevels[level+1].levelRefinement; iter++){
          err = zlsa_RefinedSweep(level+1, iter);
          if (err != noErr) { return(err);}
        }
      }
			
			
      err = zlsa_DirectSweep ( level, iteration );
      
      if ( (level+1) < gNLevels ) {
#ifdef MPI
      // wait for all zones to be the same state before BC exchanges
	  MPI_Barrier(MPI_COMM_WORLD);
#endif MPI
        err = zlsbc_FillSuperLevelFromLevel ( level+1 );
      }// this level needs lower level's active cells and transition cells
      
    }
  } else { 
    err = levErr; 
    zlsio_ErrPrintf( " LEVEL ERROR: zlsa_RefinedSweep level number %d out of range %d \n", gNLevels, level); 
    return (err); 
  } 
	
  return(err);
	
}

//====================================================================================
//
// Sweep and Advance all zones in level, including any refined zones on the level below
// ***and pass results to super level***. 
//
// Difference with zlsa_RefinedSweep is that all levels are swept 
// using the same finest timestep.
//
// requires the timesteps to be set up first in a variation of zlst_SetAllClocksTimeStep
//
//====================================================================================

int zlsa_UniformSweep(Integer level, Integer iteration){ 
  //
  // call zlsa_UniformSweep(0, 0) to process whole structure.
  //
	// step all levels one uniform timestep, needs all level clocks to be 
  // synchronized to the same timestep regardless of refinement.  
  // Usually set timestep to most refined step.
  //
  
  int err = noErr;
  Integer l;
  
  if (level >= 0) {
    
    for (l = (gNLevels-1); l >= level; l--) {
   //   zlsio_Printf( " zlsa_UniformSweep l: %d  it: %d \n", l, iteration);
      
      err = zlsa_DirectSweep ( l, iteration );
      
      if (l < (gNLevels-1)) {
#ifdef MPI
      // wait for all zones in level to be the same state before update
        MPI_Barrier(MPI_COMM_WORLD);
#endif MPI
    //  zlsio_Printf( " zlsbc_FillSuperLevelFromLevel l: %d \n", l+1);
        err = zlsbc_FillSuperLevelFromLevel ( l+1 );
      }
      
    }
    
  } else { 
    err = levErr; 
    zlsio_ErrPrintf( " LEVEL ERROR: zlsa_UniformSweep level number %d out of range %d \n", gNLevels, level); 
    return (err); 
  } 
	
  return(err);
	
}

//====================================================================================
//
// Sweep and Advance all zones in level, without processing any refined zones on the level below
// or updating the level above.  Less overhead, useful for simple tiled levels without
// refinement.  For a single level iteration == 0 == kZLS_DirectSweepIteration
// for refined sweeps, the 0th iteration fills all BCs from super and in subsequent
// sweeps only the outer BCs are filled, as the inner nR BCs form a transition zone that
// is accumulating the fluxes for sending back to the super level.  Used in lagrangian
// code when gNLevels > 1
//====================================================================================


#ifdef ZLS_LOADSORTING

void  QuickSortZoneTimes ( ZonePtr *zoneArr, clock_t * delta,  Integer l, Integer r);
void  QuickSortZoneTimes ( ZonePtr *zoneArr, clock_t * delta,  Integer l, Integer r){

		register Integer		i, j;

		register clock_t		dx;
		register clock_t		dw;
		register ZonePtr		 x;
		register ZonePtr		 w;

			i = l;
			j = r;
      dx = delta  [(l + r) / 2];
      x  = zoneArr[(l + r) / 2];

			do {

				while( delta[i] > dx)
					i++;

				while( dx > delta[j] )
					j--;

				if (i <= j) { // swap
						w          = zoneArr[i];
						zoneArr[i] = zoneArr[j];
						zoneArr[j] = w;
						dw         = delta[i];
						delta[i]   = delta[j];
						delta[j]   = dw;
						i++;
						j--;
					}

			} while (i <= j);

	if (l < j)  QuickSortZoneTimes(zoneArr, delta, l, j);
	if (i < r)  QuickSortZoneTimes(zoneArr, delta, i, r);

}

inline int zlsa_SortZoneListbyLoad( ZonePtr *zoneArr, Integer nZones );
inline int zlsa_SortZoneListbyLoad( ZonePtr *zoneArr, Integer nZones ){
  int err = noErr;
  int i   = 0;
  if ( nZones > 1){
  
    clock_t delta[nZones];
    
    for ( i = 0; i < nZones; i++){
      ZonePtr z = zoneArr[i];
      delta[i]  = z->zoneCPUTimer.cpuusr[1]- z->zoneCPUTimer.cpuusr[0];
    }
    
#if (ZLS_DEBUG >1)    
    for ( i = 0; i < nZones; i++) {
      ZonePtr z = zoneArr[i];
      zlsio_Printf(" z: %2.2d, dCPU: %d \n", z->zoneID, delta[i]);
    }
#endif (ZLS_DEBUG >1)    

    QuickSortZoneTimes ( zoneArr, delta,  0, nZones-1);
    
#if (ZLS_DEBUG >1)    
    for ( i = 0; i < nZones; i++) {
      ZonePtr z = zoneArr[i];
      zlsio_Printf(" QS:  z: %2.2d, dCPU: %d \n", z->zoneID, delta[i]);
    }
#endif (ZLS_DEBUG >1)    
    
   }
  return(err);
}
#endif ZLS_LOADSORTING
//====================================================================================
int zlsa_DirectSweep(Integer level, int iteration){ // call DirectSweep(l, 0) to process just level l.
	
  int err = noErr;
   Integer i;
  Integer zone;
  ZonePtr z;
  
  if ((level >= 0) && (level < gNLevels)){

#ifdef MPI
      // wait for all zones in level to be the same state before BCs are exchanged
	  MPI_Barrier(MPI_COMM_WORLD);
#endif MPI
     err = zlsbc_ImposeAllLevelBCs ( level, iteration );
    if (err != noErr){
      zlsio_ErrPrintf( " ERROR: zlsa_Sweep in zlsbc_ImposeAllLevelBCs: L %d I %d E %d \n",level,iteration,  err);
      return(err);
    }
//
// no need to wait after BCs are set as each may proceed as soon as it has it's own
// data independently
//

// build list of zones for this process to sweep
// number of zones to be dealt with, fixed at startup for now, but could change in runttime.

    Integer nZones   = gLevels[level].nZones; 

#ifdef MPI
// under MPI not all zones of a grid may be swept by this process, build a list
// of actual zones to sweep, and allow for sorting by load if needed.

    Integer nMPIZones = 0;
    for ( i = 0; i < nZones; i++){
        z =  gLevels[level].levelZones[i]; // access zone and increment counter
        if (z->zoneRankID == gMPIrank) nMPIZones++;
      }
    nZones = nMPIZones;
    
#endif MPI
    //
    // A zone list is built even for non-mpi to allow for load 
    // balancing options.
    //
    
    ZonePtr zoneArr[nZones]; // list of zones to sweep
    
    i = 0;
    for ( zone = 0; zone < gLevels[level].nZones; zone++){
#ifdef MPI
      z =  gLevels[level].levelZones[zone]; // access zone and increment counter
      if (z->zoneRankID == gMPIrank) {
      //
      // in MPI list just zones that belong to this rank.
      //
        zoneArr[i++] = z;
      }
#else
      //
      // without MPI list just defaults to whole level list.
      //
      zoneArr[zone] = gLevels[level].levelZones[zone];
#endif !MPI
    }
    
#ifdef ZLS_LOADSORTING
//
// zone load sort, to sort slow zones to the front.
//
//
    zlsa_SortZoneListbyLoad( zoneArr, nZones );
//
#endif ZLS_LOADSORTING

#ifdef PTHREADS

    Integer nThreads     = 0;  // number of zones actually threaded
    Integer queueThreads = 0;  // index in queue for batch = 0 .. ZLS_MAXPTHREADS-1
    Integer batchNumber  = 0;  // the index of the batches going through
	
    gSweepLock           = 0;	// global thread counting mutex variable.

	Integer nZoneBatches     = nZones / ZLS_MAXPTHREADS;
	Integer nZoneRemainder   = nZones % ZLS_MAXPTHREADS;
  
//
// if nZones > 	ZLS_MAXPTHREADS then run in batches of ZLS_MAXPTHREADS and then
// finish off any remainder in a < ZLS_MAXPTHREADS lot.
// as only algorithm zones are swept, if there are a lot of non-algorithm zones
// present then some batches may not be full.  Keep non algorithm zones to a minimum
//
// if nZones == ZLS_MAXPTHREADS then just one batch will run
//
// if nZones < ZLS_MAXPTHREADS then just the remainder will run
//
//
// init actual zone counter, increments even if not queued so the level is 
// fully checked and zones are not missed.  Starts at zone == 0

      zone = 0; 
	  	  
      if ( nZoneBatches > 0 ) {
	  
		    for( batchNumber = 0; batchNumber <nZoneBatches; batchNumber++){

				nThreads = 0;
				
				for ( queueThreads = 0 ; queueThreads < ZLS_MAXPTHREADS; queueThreads++ ) {

					z =  zoneArr[zone++]; // access zone and increment counter
					if (z != NULL){
#ifdef MPI	
						if (z->zoneRankID == gMPIrank) {
#endif MPI
						if ( z->zoneType == kAlgorithmZoneType ){ // only process appropriate zones, ie skip image zones
							zlsa_QueuePThreadSweep( z );
							nThreads++;
						} // z->zoneType == kAlgorithmZoneType
#ifdef MPI	
						}
#endif MPI
					} // z!= NULL
          
          
				}// queue
				
			zlsa_WaitPThreadSweeps( nThreads ); // wait for each batch
			
				
			} // batchNumber
			
	 } // nZoneBatches > 0
	 
      if ( nZoneRemainder > 0 ) {
	  
// finish off any remainder, no batch loop needed
				nThreads = 0;
				
				for ( queueThreads = 0 ; queueThreads < nZoneRemainder; queueThreads++ ) {
					z =  zoneArr[zone++]; // access zone and increment counter
					if (z != NULL){
#ifdef MPI	
						if (z->zoneRankID == gMPIrank) {
#endif MPI
							if ( z->zoneType == kAlgorithmZoneType ){ // only process appropriate zones, ie skip image zones
								zlsa_QueuePThreadSweep( z );
								nThreads++;
							} // z->zoneType == kAlgorithmZoneType
#ifdef MPI	
						}
#endif MPI
					} // z!= NULL
				}  // queue
				
			zlsa_WaitPThreadSweeps( nThreads );
		} // nZoneRemainder > 0
#endif PTHREADS

#ifndef PTHREADS

    for( zone = 0; zone < nZones; zone++){
    
			z =  zoneArr[zone]; // get next zone from list

      if (z != NULL){
#ifdef MPI	
        if (z->zoneRankID == gMPIrank) {
#endif MPI
          if ( z->zoneType == kAlgorithmZoneType ){ // only process appropriate zones, ie skip image zones
            
#ifdef ZLS_LOADSORTING
            zlst_StartCPUTimes ( &z->zoneCPUTimer );
#endif ZLS_LOADSORTING

#ifdef DIVVHCORRECTION
            zlsa_CalcZoneDivV(z);
#endif DIVVHCORRECTION
           
#ifdef MEANSPLITSWEEPS
#ifdef SYM1D  
            err = ZLSSWEEPS   ( z ); // macro for sweep method in force. see zls_defs.h
#endif SYM1D  
            
#ifdef SYM2D  
            err = zlsa_Synch2DData ( z ); // make sure boundaries are on Alt vars too.
            
            z->zoneW = z->zoneWLeft ; // just to be safe, should be set by Synch
            err = ZLSSWEEPS   ( z ) ; // macro for sweep method in force. see zls_defs.h
            z->zoneClock.cycles += 1;
            
            z->zoneW = z->zoneWRight; // sweep Alt vars in opposite sweep order
            err = ZLSSWEEPS   ( z ) ; // macro for sweep method in force. see zls_defs.h
            z->zoneClock.cycles -= 1;
            
            err = zlsa_Average2DSweeps ( z );
#endif SYM2D 
            
#ifdef SYM3D  
            err = ZLSSWEEPS   ( z ); // macro for sweep method in force. see zls_defs.h
#endif SYM3D  

#else
            err = ZLSSWEEPS   ( z ); // macro for sweep method in force. see zls_defs.h

#endif !MEANSPLITSWEEPS
            
            zlst_AdvanceClock( &z->zoneClock );
            
            err = zlst_CalcZoneNewTimeScale( z );
            
#ifdef ZLS_LOADSORTING
            zlst_EndCPUTimes ( &z->zoneCPUTimer );
#endif ZLS_LOADSORTING

          } // z->zoneType == kAlgorithmZoneType 
#ifdef MPI	
        }
#endif MPI
 			} else { err = zonErr; }
     
    } // zone loop
    
#endif !PTHREADS
    
    zlst_AdvanceClock( &gLevels[level].levelClock );
//    zlsio_Printf( " zlsa_DirectSweep level number %d \n",  level); 

  } else { 
    err = levErr; 
    zlsio_ErrPrintf( " LEVEL ERROR: zlsa_DirectSweep level number %d out of range %d \n", level, gNLevels); 
    return (err); 
  } // (level < gNLevels )
  
  return(err);
	
}



