#ifndef ZLS_CONST_H
#define ZLS_CONST_H
#include <zls.h>
//
//========================================
//
// FYRIS - ZLS Constants and Flags
//
//========================================
//
// Flags represent bits and can be added to create combined quantities,
// have 'Flag' in their names, they are designed to be added together
// to produce composite meaningful values.
//
// constants represent unique quantities - ie exclusive values such as
// array indexes, and don't have 'Flag' in their names, mostly.
//
//========================================
//
// This is the overall ZLS Library version constants.  The ZLS_VERSION
// is used in any text output and version checks. 
//
#define ZLS_VERSION     0x00000007
//
//========================================
//
// Basic Errors
//
//========================================
//
enum {  
  noErr     =  0,
  fileErr   = -1,
  memErr    = -2,
  dimErr    = -3,
  zonErr    = -4,    // zone not found in level
  timErr    = -5,    // time error - usually bad dt estimate with a hotspot
  levErr    = -6,    // bad level number - usually to big , ie >= gNLevels
  ioErr     = -7,    // bad io settings reading io.conf etc
  kindErr   = -8,    // zone kind is incompatible with routine, esp image zones
  mpiErr    = -9,    // an error occured in an MPI routine
  nyiErr    = -999  // not yet implemented !
};
//
enum {    
  kZLS_InValid =  0, 
  kZLS_Valid   =  1
};
//
//========================================
//
//  Dimensions and Directions
//
//========================================
//
enum {    
  k0Dim   =  0, 
  kXDim   =  0, 
  kYDim   =  1,
  kZDim   =  2
};
//
enum {    
  k1D   =  1, 
  k2D   =  2,
  k3D   =  3,
  kMaxCoordDims = 3
};
//
//========================================
// constants for counting the prevailing number of dimensions and 
// the corresponding number of directions on the grid..
//========================================
//
#ifdef SYM1D
enum { kSymDims = k1D };
#endif
#ifdef SYM2D
enum { kSymDims = k2D };
#endif
#ifdef SYM3D
enum { kSymDims = k3D };
#endif
//
enum {
#ifdef SYM1D
kNGeomDirs       =  2, // loop limit for the geometric directions 0-1
#endif SYM1D
#ifdef SYM2D
kNGeomDirs       =  4, // loop limit for the geometric directions 0-3
#endif SYM2D
#ifdef SYM3D
kNGeomDirs       =  6 // loop limit for the geometric directions 0-5
#endif SYM3D
};
//
//========================================
//
// Directions as sequential constants - for array indexes etc, 
// Right handed 3D convention:  x cross y = +z
//
//========================================
//
enum {  
  kDirLeft         =  0, // decreasing x
  kDirRight        =  1, // increasing x
  kDirDown         =  2, // decreasing y
  kDirUp           =  3, // increasing y
  kDirYon          =  4, // decreasing z
  kDirHither       =  5, // increasing z (rh'd, out of screen if x is l->r and y is b->t
  kDirSub          =  6, // level down
  kDirSuper        =  7, // level up
  kDirMask         = 0x000f // to mask out just the direction from a composite cell type
};
//
// Direction flags as bit positions for summing multiple directions into a single constant.
//
//  ie transition cell  = kCellTypeFlagTransition + kDirFlagLeft + kDirFlagUp
//  ie boundary   cell  = kBoundaryTypeFlagNextZ  + kDirFlagUp
//
//
enum {  
  kDirFlagLeft      =  0x01, // decreasing x
  kDirFlagRight     =  0x02, // increasing x
  kDirFlagDown      =  0x04, // decreasing y
  kDirFlagUp        =  0x08, // increasing y
  kDirFlagYon       =  0x10, // decreasing z
  kDirFlagHither    =  0x20, // increasing z
  kDirFlagSub       =  0x40, // level down
  kDirFlagSuper     =  0x80 // level up
};
//
//========================================
//  slice geoemtry flags
//========================================
//
enum {    
  kSliceNone  = 0x00, 
  kSliceX     = 0x01, 
  kSliceY     = 0x02,
  kSliceZ     = 0x04,
  kSliceXY    = 0x03, 
  kSliceYZ    = 0x05,
  kSliceZX    = 0x06,
  kSliceXYZ   = 0x07,
  kSliceXSum     = 0x11, 
  kSliceYSum     = 0x12,
  kSliceZSum     = 0x14,
  kSliceXYSum    = 0x13, 
  kSliceYZSum    = 0x15,
  kSliceZXSum    = 0x16,
  kSliceYX    = 0x23, 
  kSliceZY    = 0x25,
  kSliceXZ    = 0x26,
  kSliceYXSum = 0x33, 
  kSliceZYSum = 0x35,
  kSliceXZSum = 0x36
};
//
//====================================================================================
//
// Interpreted zoneData Quantity  Structures - distinct from WArrs
// above, for interfacing to zoneData Qs independently of zoneData
// internal layouts.  No coordinates just the data, multi-dims  and
// convenient mnemonics.
//
//========================================
//
// number of variables, add number of tracers if desired up to 15
// WARNING, large numbers of variables strongly restrict 3D sizes on 32
// bit systems Many small cubes may be needed to avoid 2GB-4GB limits,
// esp in double prec
//

enum { 
  kNVars       = (NVARIABLES),
  kNTracers    = (NTRACERS),
  kNAllVars    = (NVARIABLES+NTRACERS),
#ifdef MEANSPLITSWEEPS
  kAltVarsOffset = kNAllVars,
  kTotalAllVars  = (NVARIABLES+NTRACERS)*2
#else
  kAltVarsOffset = 0,
  kTotalAllVars  = (NVARIABLES+NTRACERS)
#endif
};

//========================================
//
// array locations of quantites in interpreted type elements
//
//========================================
//
//    kUDens      = 0,  // Desity in U
//    kUEneg      = 1,  // Total Energy Density in U
//    kUMx        = 2,  // specific x momentum in U
//    kUMy        = 3,  // specific y momentum in U
//    kUMz        = 4,  // specific z momentum in U
//
//    kWDens      = 0,  // Density    w[kWDens].a = density value,  w[kWDens].c.x.a6 = a6 coefficient of density in x-direction
//    kWPres      = 1,  // Pressure   w[kWPres].a = Pressure value
//    kWVx        = 2,  // x velocity in W
//    kWVy        = 3,  // y velocity in W
//    kWVz        = 4,  // z velocity in W <- max for HD
//
//    kBx         = 5,  // B/Sqrt(4*pi) x - component if mhd used (1D 2D 3D)
//    kBy         = 6,  // B/Sqrt(4*pi) y - component if mhd used (1D 2D 3D)
//    kBz         = 7,  // B/Sqrt(4*pi) z - component if mhd used (1D 2D 3D)
//
//    kBxi        = 8,  // B x interface component if mhd used (1D 2D 3D)
//    kByi        = 9,  // B x interface component if mhd used (2D 3D)
//    kBzi        =10,  // B z interface component if mhd used (3D)
//
enum { 
  kTracer0     = kNVars+0, // tracers add onto the end of the vars as needed, vars are 0 -> nVars-1
  kTracer1     = kNVars+1, //some constants here for cases with just a few tracers
  kTracer2     = kNVars+2, //but the user can always compute tracer array ids programitically
  kTracer3     = kNVars+3, 
  kTracer4     = kNVars+4,
  kTracer5     = kNVars+5,
  kTracer6     = kNVars+6,
  kTracer7     = kNVars+7
};

enum { 
  kIon0     = kNVars+kNTracers+0, // ions add onto the end of the vars and tracers as needed, vars are 0 -> nVars-1
  kIon1     = kNVars+kNTracers+1, //
  kIon2     = kNVars+kNTracers+2, 
  kIon3     = kNVars+kNTracers+3, 
  kIon4     = kNVars+kNTracers+4,
  kIon5     = kNVars+kNTracers+5,
  kIon6     = kNVars+kNTracers+6,
  kIon7     = kNVars+kNTracers+7
};
//
// array locations of quantites in interpreted type elements

enum { 
  kNoVar      = -1,  // no variable selected 
  kAnonVar    = 0,  // an anonymous variable, of no particular interpretation  

  kWDens      = 0,  // in U or W
  kWPres      = 1,  // in W
  kWVx        = 2,  // velocity in W
  kWVy        = 3,  // velocity in W
  kWVz        = 4,  // velocity in W

  kUDens      = 0,  // in U or W
  kUEneg      = 1,  // in U

  kUMx        = 2,  // specific x momentum in U
  kUMy        = 3,  // specific y momentum in U
  kUMz        = 4,  // specific z momentum in U
    
  kBx         = 5,  // B/Sqrt(4*pi) x - component
  kBy         = 6,  // B/Sqrt(4*pi) y - component
  kBz         = 7,  // B/Sqrt(4*pi) z - component

  kBxi        = 8,  // B x interface component if mhd used (1D 2D 3D)
  kByi        = 9,  // B x interface component if mhd used (2D 3D)
  kBzi        =10,  // B z interface component if mhd used (3D)
//
// Derived, virtual variables - ie temperature, computed from P and rho in W
//
  //  
  // Temperature
  //
  kWTemp      = 100, // tep0*P/rho 
  //
  // Energy  (Lagrangian or Eulerian depending on compiler defs)
  //
  kWE         = 101, // Internal Specific Energy
  kWTotE      = 102, // Total Specific Energy 
  kWKE        = 103, // Specific Kinetic Energy 
  //
  kWEE         = 104, // Internal Specific Energy Eulerian always
  kWTotEE      = 105, // Total Specific Energy Eulerian always 
  kWKEE        = 106, // Specific Kinetic Energy Eulerian always 
  //
  kWEL         = 111, // Internal Specific Energy Lagrangian always
  kWTotEL      = 112, // Total Specific Energy Lagrangian always 
  kWKEL        = 113, // Specific Kinetic Energy Lagrangian always 
  //
  // momentum
  //
  kWM         = 107, // d*sqrt(v^2) total momentum
  kWMx        = 108, // d*vx, or mx in U
  kWMy        = 109, // d*vy, or mx in U
  kWMz        = 110, // d*vz, or mx in U
  //
  // EOS
  //
  kWGamma     = 120, // gamma(P/d,cp) or using neq ionisation
  kWMu        = 121, // mu(P/d,cp) or using neq ionisation
  kWXHI       = 122, // xhi (P/d,cp), scalar fraction
  kWXHII      = 123, // xhii(P/d,cp)
  kWXH2       = 124, // xh2 (P/d,cp)
  kWdHI       = 125, // d*xhi (P/d,cp), mass densities here
  kWdHII      = 126, // d*xhii(P/d,cp)
  kWdH2       = 127, // d*xh2 (P/d,cp)
  
  kWEntr      = 128, // A, P/d^g : isoentropic quantity
  
  kWEnth      = 129, // h, P/((g-1) d) + P/d, specific enthalpy  depends in compile options
  kWEnthE     = 130,  // h, P/((g-1)) + P, Eulerian specific enthalpy
  kWEnthL     = 131,  // h, P/((g-1)) + P, Lagrangian specific enthalpy
  
  kWCs        = 140, // Adiabatic Sound speed
  kWMach      = 141, // velm/Sound speed
  kWVelm      = 142, // sqrt(vx^2 +vy^2 ..)
  kWSpeed     = 142, //
  kWAs        = 143, // Isothermal Sound speed

  kWVelmx     = 144, // sqrt(vx^2)
  kWVelmy     = 145, // sqrt(vy^2)
  kWVelmz     = 146, // sqrt(vz^2)
  kWVelmxy    = 147, // sqrt(vx^2+vy^2)
  kWVelmyz    = 148, // sqrt(vy^2+vz^2)
  kWVelmzx    = 149, // sqrt(vz^2+vx^2)

  kWDeltaVx   = 160, // vx velocity slice 
  kWDeltaVy   = 161, // vy velocity slice 
  kWDeltaVz   = 162, // vz velocity slice 
  kWDivV      = 163, // div V
  kWDvDx      = 164, // div Vx
  kWDvDy      = 165, // div Vy
  kWDvDz      = 166, // div Vz
  
// microphysics some reserved for future physics

  kWHeat      = 200, // total heating Rate
  kWHTgr      = 201, // grain-gas heat exchange heating Rate
  kWHCRX      = 202, // Comsic Ray Heating Rate
  kWEmGr      = 203, // Grain Emission = Grain absoprtion rate
  kWHIPE      = 204, // HI photoelectric Heating Rate
  kWnHI       = 205, // nHtot*xhi number densities
  kWnHII      = 206, // nHtot*xhii
  kWnH2       = 207, // nHtot*xh2*0.5
  kRecomb     = 208, // (nHtot*xhii)^2 (T/10e4)^-0.77
  kRayleigh   = 209, // dx*4.73917E-20*(nHtot*xhii)^2 (T/10e4)^-0.77
  
  kWCool      = 210, // total Cooling Rate (emissivity)
  kWCoolAt    = 211, // rho^2 lambda(T,phys), Atomic component
  kWCoolCO    = 212, // rho^2 lambda(T,phys), CO component
  kWCoolH2    = 213, // rho^2 lambda(T,phys), H2 Component
  kWLamb      = 214, // lambda(T,phys)
  kWLambAt    = 215, // lambda(T,phys) Atomic component
  kWLambCO    = 216, // lambda(T,phys) CO cooling
  kWLambH2    = 217, // lambda(T,phys) H2 Cooling

// gravity  
  kGPhi       = 315, // Gravity Potential Field
  kdPhi       = 316, // Gravity spatial derivative  magnitude of all components
  kdPhiX      = 317, // Gravity spatial derivative in x-direction
  kdPhiY      = 318, // Gravity spatial derivative in y-direction
  kdPhiZ      = 319,  // Gravity spatial derivative in z-direction

// tracers  
  kScalar0     = 400, // tracers/density = fraction of tracer by mass
  kScalar1     = 401, //some constants here for cases with just a few tracers
  kScalar2     = 402, //but the user can always compute tracer array ids programitically
  kScalar3     = 403, 
  kScalar4     = 404,
  kScalar5     = 405,
  kScalar6     = 406,
  kScalar7     = 407,  // leave room for more tracers
  
  kRadioJnu0   = 500, // scalar0 times P^1.8 
  kRadioJnu1   = 501, // scalar1 times P^1.8 
  kRadioJnu2   = 502, // scalar2 times P^1.8 
  
// MHD vars - reserved

  kBPres       = 600, // 0.5*(Bx*Bx + By*By + Bz*Bz)
  kBE          = 601, // 0.5*(Bx*Bx + By*By + Bz*Bz)
  kDivB        = 602, // (dBxdx + dBydy [ + dBzdz]) using face centre values 2D + 

// misc variables and displays
  kSymDens     = 999,  // special symmetric difference in density for LW tests
  kVortex2D    = 998,  // 2D vorticity for LW Gresho Vortex test
  kVortex3D    = 997,  // 3D vorticity
  kSclieren    = 996,  // 2D Schlieren image, I prop to partial dd/dx
  kShadow      = 995   // 2D Shadow image, I prop to partial (dd/dx)^2 + (dd/dy)^2
};

// array locations of quantites in fields
enum { 
  kAnonFieldx   = 0,  // an anonymous variable, of no particular interpretation  
  kAnonFieldy   = 1,  // an anonymous variable, of no particular interpretation  
  kAnonFieldz   = 2,  // an anonymous variable, of no particular interpretation  
  kGx           = 0,  //  gravity accelleration in x direction
  kGy           = 1,  //  gravity accelleration in y direction
  kGz           = 2   //  gravity accelleration in z direction
};  
//
//========================================
//
// Zone types - also sets kAlgorithmZoneType, the global type default
// Needed here for kNAllVars and kNTracers used widely
//========================================
//
enum {  
  kZoneTypeS      = 0x0000, // zone has uniterpreted quantities - ie image processing applications
  kZoneTypeU      = 0x0100, // zone has conserved quantities, rho, E, mx, my, mz, ...
  kZoneTypeW      = 0x0200, // zone has physical quantities,  rho, P, vx, vy, vz, ...
  kZoneTypeO      = 0x0300, // zone has some other interpreted  type ...
  kZoneTypeQua    = 0x0000, // zone contains an array of simple Real ...
  kZoneTypeVar    = 0x0010 // zone contains an array of variable structures ...
};
//
//========================================
// SIMPLE_QUAS - zone data are plain blocks/Arrays of Reals, one block per quantity - ie density
// SIMPLE_VARS - zone data are variable blocks/Arrays - quanity + coefficients - not so good for large 3D
// CONSERVED_QUAS zone has Reals for rho, E, mx, my, mz + tracers if any, set NVARIABLES =5 
// CONSERVED_VARS zone has vars for rho, E, mx, my, mz + tracers if any, set NVARIABLES =5 
// PRIMITIVE_QUAS zone has Reals for rho, P, vx, vy, vz + tracers if any, set NVARIABLES =5 
// PRIMITIVE_VARS zone has vars for rho, P, vx, vy, vz + tracers if any, set NVARIABLES =5 
//
#ifdef SIMPLE_QUAS
enum { 
  kAlgorithmZoneType    = (kZoneTypeS+kZoneTypeQua),
  kImageZoneType        = (kZoneTypeS+kZoneTypeQua)
};
#define USEZONEQUAS
#endif

#ifdef SIMPLE_VARS
enum { 
  kAlgorithmZoneType    = (kZoneTypeS+kZoneTypeVar),
  kImageZoneType        = (kZoneTypeS+kZoneTypeQua)
};
#define USEZONEVARS
#endif SIMPLE_VARS

#ifdef CONSERVED_QUAS
enum { 
  kAlgorithmZoneType    = (kZoneTypeU+kZoneTypeQua),
  kImageZoneType        = (kZoneTypeS+kZoneTypeQua)
};
#define USEZONEQUAS
#endif CONSERVED_QUAS

#ifdef CONSERVED_VARS
enum { 
  kAlgorithmZoneType    = ( kZoneTypeU + kZoneTypeVar ),
  kImageZoneType        = (kZoneTypeS+kZoneTypeQua)
};
#define USEZONEVARS
#endif CONSERVED_VARS

#ifdef PRIMITIVE_QUAS
enum { 
  kAlgorithmZoneType    = ( kZoneTypeW + kZoneTypeQua ),
  kImageZoneType        = (kZoneTypeS+kZoneTypeQua)
};
#define USEZONEQUAS
#endif PRIMITIVE_QUAS

#ifdef PRIMITIVE_VARS
enum { 
  kAlgorithmZoneType    = ( kZoneTypeW + kZoneTypeVar ),
  kImageZoneType        = (kZoneTypeS+kZoneTypeQua)
};
#define USEZONEVARS
#endif PRIMITIVE_VARS
//
//========================================
//
//  Zone Cell Map/Boundary constants
//
//========================================
//
enum {  
  kCellTypeFlagNULL          = 0x0000,   // do nothing, skip cell
  kCellTypeFlagNormal        = 0x0100,   // normal cell just compute..
  kCellTypeFlagBoundary      = 0x0200,   // cell is set as a boundary condition + Boundary Type flags + constants
  kCellTypeFlagTransition    = 0x0400,   // cell a transition cell adjacent a refined zone, needs flux corrections
  // type = 0x0400 + direction flag, can be multiple dir flags
  kCellTypeMask              = 0x0f00 // to mask out just the boundary type
};

enum {  
  kBoundaryTypeFlagNULL      = 0x0000, // boundary comes from no other zone
  kBoundaryTypeFlagSuperZ    = 0x1000, // boundary comes from super zone
  kBoundaryTypeFlagNextZ     = 0x2000, // boundary comes from a NextZone, type = kBoundaryTypeNextZ + Dir
  kBoundaryTypeFlagComputed  = 0x4000, // boundary values computed + single type constant:
  kBoundaryTypeFlagUserBC    = 0x8000, // boundary values are also updated by user routine after all other updates
  kBoundaryTypeMask          = 0xf000 // to mask out just the boundary type
};

//
// Computed boundary kinds, CellType = kCellTypeFlagBoundary+kBoundaryTypeFlagComputed+ kind code:
//
enum {  
  kBCCKindFree      =  0, // boundary just copies variable out
  kBCCKindConstant  =  1, // boundary values fixed constant values
  kBCCKindReflected =  2, // boundary mirrors adjacent cells
  kBCCKindPeriodic  =  3, // boundary copies opposite boundary inner active region, non-local: works on extrema only
  kBCCKindExtrapol  =  4, // boundary extrapolates the variables??
  kBCCKindFreer     =  5, // boundary  copies variables out from first BC not edge
  kBCCKindFreeOutOnly  =  6, // boundary  copies variables out enforcing outward velocity
  kBCCKindUser      =  7, // boundary set by user problem parameters, in level[0] levelProblem
  kBCCKindMask      = 0x000f // to separate the calc type from a boundary type, up to 16 types allowed!
};
//
//========================================
// Low level constants follow...
//========================================
//
// Currently the maximum size of the Zone-Level system is predefined, but could be made
// totally dynamic in the future.  For now the max number of levels and zones are set
// here as macros.
//
//========================================
//
#define MAXLEVELS (8)
#define MAXLEVELZONES (64)
//
//========================================
//
#define NREFINEMAX1D    8
#define NREFINEMAX2D    4
#define NREFINEMAX3D    4
//
//========================================
//  var/quantity row limits
//========================================
//
#ifdef SYM1D
enum { 
  kNRefineMax  = NREFINEMAX1D,
};
#define NREFINEMAX NREFINEMAX1D
#endif
#ifdef SYM2D
enum { 
  kNRefineMax  = NREFINEMAX2D,
};
#define NREFINEMAX NREFINEMAX2D
#endif
#ifdef SYM3D
enum { 
  kNRefineMax  = NREFINEMAX3D,
};
#define NREFINEMAX NREFINEMAX3D
#endif
//
//========================================
//
static const Integer kMaxInt    = (0x7fffffff);
static const Integer kMaxIntM1  = (0x7ffffffe);
static const Integer kMaxIntM2  = (0x7ffffffd);
//
//========================================
// limits and small values to control over/underflow conditions
//========================================
//
#ifdef SINGLE_PRECISION
static const Real kZLS_Huge    = 1.0E+38F;  // for setting long times that won't run out
static const Real kZLS_Tiny    = 1.0E-38F;  // smaller even than sensible cooling units.
static const Real kZLS_LogHuge    = +38.0F;  //
static const Real kZLS_LogTiny    = -38.0F;  //
static const Real kZLS_Epsilon = FLT_EPSILON;  // for avoiding div by 0
static const Real kZLS_Small   = (FLT_EPSILON*32);  // for avoiding unphysically small values
static const Real kZLS_PressureVacuum = 1.0e-8F;  // specific small value for pressure minimum, assumes pressure is normalised
static const Real kZLS_MinSpeed = 1.0e-8F;  // specific small value for velocity minimum
static const Real kZLS_MinBfield  = 1.0e-8F;  // specific small value for B threshold
static const Real kZLS_MinDensity  = 1.0e-8F;  // specific small value for d threshold
#endif SINGLE_PRECISION
#ifdef DOUBLE_PRECISION
static const Real kZLS_Huge    =  1.0E+99;  // for setting long times that won't run out
static const Real kZLS_Tiny    =  1.0E-99;  //
static const Real kZLS_LogHuge    = +99.0;  //
static const Real kZLS_LogTiny    = -99.0;  //
static const Real kZLS_Epsilon    = DBL_EPSILON;  // for avoiding div by 0
static const Real kZLS_Small      = (DBL_EPSILON*32.0);  // for avoiding unphysically small values
static const Real kZLS_PressureVacuum = 1.0e-12;  // specific small value for pressure minimum
static const Real kZLS_MinSpeed   = 1.0e-12;  // specific small value for velocity threshold
static const Real kZLS_MinBfield  = 1.0e-12;  // specific small value for B threshold
static const Real kZLS_MinDensity  = 1.0e-12;  // specific small value for d threshold
#endif DOUBLE_PRECISION
//
//========================================
// basic fp constants
//========================================
// common  precison dependent fractions:
static const  Real sixth      = 1.0/6.0;
static const  Real third      = 1.0/3.0;
static const  Real twothd     = 2.0/3.0;
static const  Real fourthd    = 4.0/3.0;
static const  Real fivethd    = 5.0/3.0;
//
static const  Real pi         = 3.1415926535897932;
static const  Real twopi      = 6.2831853071795865;
static const  Real fourpi     = 12.566370614359173;
static const  Real invpi      = 1.0/3.1415926535897932;
static const  Real invtwopi   = 1.0/6.2831853071795865;
static const  Real invfourpi  = 1.0/12.566370614359173;
//
static const  Real kLog10     = 2.3025850929940459; // nat log of 10
//
#endif ZLS_CONST_H
