#ifndef ZLS_DEFS_H
#define ZLS_DEFS_H
#include <zls.h>
//
//================================================================================
//
// Global code and MACRO definitions.
//
// By convention all C MACRO flags are set in ALL UPPERCASE to distinguish
// them from normal C code set in mixed case, mostly lower.  The only exceptions
// to this are some low level inline functions defined elsewhere.  To ensure 
// reliability, define any new MACRO flags in UPPERCASE.
//================================================================================
//  Low level C macros - affect basic structures and are used to set up
//  more complex enums. In general use enum constants when programming - ie
//  kNRefineMax and not NREFINEMAX.
//================================================================================
//
// Fundamental Equations:
// HYDRO: Ideal Hydrodynamics + additional microphysics - Fyris Alpha
//
//================================================================================
//
#define HYDRO
//
//================================================================================
// Geometry MACROS
//
//  define the global dimensions of the problem, ie size of
//  vector/component arrays. Use exactly one of: 
//  SYM1D, SYM2D, SYM3D
//
//#define SYM1D
#define SYM2D
//#define SYM3D
//
#ifdef HYDRO
//
// 1D/2D/3D straight divergence free spatial coordinates
#define CARTESIAN
// 2D/3D only, 1D same as cartesian, Z-R-Phi cylinder coordinates, r
// diverges as 1/r
//#define CYLINDRICAL
// 2D only R-Phi circular coordinates, r diverges as 1/r
//#define CIRCULAR   
// 1D/2D/3D R-Phi-Theta spherical coordinates, r diverges at 1/r^2  Right
// Handed Coords, azimuth phi, equtorial theta +/- pi/2
//#define SPHERICAL
// 3D only, otherwise same as SPHERICAL:  R-Theta-Phi polar spherical
// coordinates, r diverges at 1/r^2  Right Handed Coords,  polar theta
// 0-pi, azimuth phi,
//#define POLARSPHERICAL
//
#endif
//
//================================================================================
//
// Set number of tracer vars in addition to the 
// equation variables
//
#define NTRACERS 0
//
//================================================================================
//
// HYDRO Interpolations/Reconstruction Order - Default PPM, not often changed
// is a 3rd order spatial and 1st order temportal method that is about 2.3 order
// overall.
//
// Method used for cell interpolation and cell left-right hydro reconstruction
//
#define PPM
//
//================================================================================
//
//
//  Riemann Solver used for hydro solutions Default:  RSSTWOSHOCK  
//
// TORO   Exact solver from Toro 1998, slow
//
// PVRS  Approximate primitive variable sover, OK for weak shocks and
// waves
//
// AIRS  Hybrid TORO/PVRS solver - OK
//
// COLLELLATWOSHOCK   original ppmlr solver, and quite fast,
// good at strong rarefactions
//
// GROTHGOTTLIEB  full exact solver from Groth and Gottlieb 1988, iterates
// on Ustar, slower than CW or RSS 
//
// RSSTWOSHOCK :  GROTHGOTTLIEB modified to use two shock approximation by
// RSS, fastest and generally same result as collella except for v strong
// rarefactions, uses PVRS as initial guess by default
//
//#define TORO
//#define PVRS
//#define GROTHGOTTLIEB
#define RSSTWOSHOCK
//#define COLLELLATWOSHOCK
//#define AIRS
//
//
//================================================================================
//
// EOS: cooling, gamma and mu
//
//Whether cooling code and structures (of any kind) are to be included,
//requires double precision
//
//#define COOLING
//
//
//================================================================================
//
//  Equation of State, [e, P, T], (d, mu, gamma)
//
//================================================================================
//
// if COOLING is enabled and the cooling is type == kPhysFileCooling then
// the EOS includes information on the efective molecular weight
// of the gas as it evolves.  The additional calculations for changing
// gamma on a cell by cell basis can be enabled separately.  If
// COOLING is off then neither take effect.
//
// if VARIABLE_EOS is undefined then a single global mu 
// and single gamma are used.
//
// if VARIABLE_EOS is defined then at least one of VARIABLE_MU and VARIABLE_GAMMA
// may be set. H_IONISATION requires VARIABLE_MU and possibly VARIABLE_GAMMA to be set
// as appropriate to the regime being considered, ie if all phases are atomic then 
// VARIABLE_GAMMA may be left off, even as MU and ionisation are changing
//
//#define VARIABLE_EOS
//#define VARIABLE_MU
//#define VARIABLE_GAMMA
//#define H_IONISATION
//
// VARIABLE_EOS alone: allows for potential variable EOS, but the actual variable factor has to be set as well
// VARIABLE_EOS plus:
// A)  VARIABLE_MU -> mu (mean molecular weight) may vary cell by cell, read from cooling function tables (FORMAT 2 or 3)
// B)  VARIABLE_GAMMA -> mu and Gamma (adiabatic index) may vary cell by cell,
//     read from cooling function tables (FORMAT 2 or 3 or 5)
// C) H_IONISATION  -> mu and Gamma vary cell by cell, plus EQUILIBRIUM H2, HI and HII ionisation
//     fractions are read from  cooling function tables (FORMAT 3 or 5) needs VARIABLE_MU and possibly VARIABLE_GAMMA
//    
//
//================================================================================
//
// Gravity Field Calculations
//
//#define GRAVITY
//#define GRAVZFIELD
//
//================================================================================
//
// Artificial Viscosity
//
//If the code is run as split sweeps then in some cases, particularly with
//cooling enabled, a non-linear filter is needed to control the 'carbuncle'
//or 'striping' instability. (see zls_cellupdate.c, zlsa_RemapSmoothW1D.)
//(cartesian or angular, it doesn't matter). Parameters are in
//zlsa_RemapSmoothW1D and should be adjusted on a case by case basis.
//
#define VISCOUS_REMAP
#define DIVVHCORRECTION
//
//================================================================================
//  Algorithm Settings
//================================================================================
//
#ifdef HYDRO
enum {NWAVE = 5, NVAR = 5};
#endif
//
//================================================================================
//
//if LAGRANGIAN is defined then grids keep two axes, an extra one is a
//
//record of the  original axis, so the normal one can be modified by a
//lagrangian update, then restored in a remap. This would not be used in
//an Eulerian code
//
// Currently LAGRANGIAN and SPLITSWEEPS *must* be defined
// Currently MEANSPLITSWEEPS only works for 2D hydro, X-Y and Y-X sweeps are 
// averaged for each update - twice as slow, but will preserve absolute
// symmetry if needed.  Will propagate asymmetry as well of course, but if
// the inital conditions are perfectly symmetrical to floating point precision, then
// the simulation will stay that way indefinitely.  If there is any even minor
// assymetry then it may grow and propogate as normal.  Mostly useful for 
// specific symmetry tests, such as the LW Implosion test.  No normally 
// needed for real problems.
//

#ifdef HYDRO
#define LAGRANGIAN
#define SPLITSWEEPS
//#define MEANSPLITSWEEPS
#define GRAVITY_POTENTIAL
#define SCALAR_LRQS
#define SCALAR_POTENTIAL
#endif HYDRO

//
//================================================================================
//
// Courant-Friedrichs-Levey Condition
//
// define the method for determining the timestep, whether to treat the
// sound speed C and the flow velocities V separately or as a combined
// sum.  Default is combined as it is more conservative and robust, but
// can result in more steps taken unless the coruant number is increased
// to march faster.  CFL for COURANT_VC_SEPARATE should be about 0.6 in
// 2D, and may be 0.8 for COURANT_VC_COMBINED or more in some  cases. 
// reduce CFL by about 20-30% for 3D in both cases.
//
//#define COURANT_VC_SEPARATE
#define COURANT_VC_COMBINED
//
//================================================================================
// Sweep method for multi-level simulations
//
// if multiple levels of resolution are used, define exactly one of the
// following: Refined sweeps does the minimum necessary sweeps per level
// (ie refinement^level) but is not optimum if shocks are crossing grid
// resolution boundaries from low - high resolution.  Works Ok
//  for outflows, from high to loe resolution. Good if lower levels have few cells.
//
// The model progresses in level 0 time steps, same as with only one
// level. UNIFORM_SWEEPS is needed if many shocks cross resolution
// boundaries, and the models use the the finest time step to update all
// levels.  Causes more cells to be calculated per model time (but less
// per seep) and a model can take longer than a refined sweep model to
// reach the the same point.  However the speedup over a full resolution
// model can still be substantial. Use refinements of 4 for 2D and 2-4 in
// 3D.  Best if lower (hires) levels have more cells than upper levels
// 
// Default is UNIFORM_SWEEPS, and is *required* for refined LAGRANGIAN grids
//
//#define REFINED_SWEEPS
#define UNIFORM_SWEEPS
//
//================================================================================
//
//  Here down are automatic/consistency checks, do not edit.
//
//================================================================================
//
#ifdef NDIMS
#undef NDIMS
#endif

#ifdef SYM1D
#define NDIMS 1
#endif SYM1D

#ifdef SYM2D
#define NDIMS 2
#endif SYM2D

#ifdef SYM3D
#define NDIMS 3
#endif SYM3D
//
// Sanity Check incase SYM defs are malformed
// if all else fails, set 1D
//
#ifndef NDIMS

#ifdef SYM1D
#undef SYM1D
#endif

#ifdef SYM2D
#undef SYM2D
#endif

#ifdef SYM3D
#undef SYM3D
#endif

#define SYM1D
#define NDIMS 1

#endif NDIMS
//
// Variable System - default PRIMITIVE_QUAS  for Fyris Hydro/MHD
//
#define PRIMITIVE_QUAS
// SIMPLE_QUAS - zone data are plain blocks/Arrays of Reals, one block per quantity
// CONSERVED_QUAS zone has Reals for rho, E, mx, my, mz + tracers if any, 
// PRIMITIVE_QUAS zone has Reals for rho, P, vx, vy, vz + tracers if any,
//
#ifdef HYDRO
// NVARIABLES must be 2+NDIMS for hydro, W
#define NVARIABLES  (2+NDIMS)
#endif
//
#ifdef MHD
// NVARIABLES must be 8 for mhd, U , M , cell centred B components + B interfaces -reserved
#define NVARIABLES  (8+NDIMS)
#endif
//
#ifdef SYM1D
#ifdef MEANSPLITSWEEPS
#undef MEANSPLITSWEEPS
#endif
#endif SYM1D
#
#ifdef SYM3D
#ifdef MEANSPLITSWEEPS
#undef MEANSPLITSWEEPS
#endif
#endif SYM3D
//
#ifdef VARIABLE_EOS 
// if a variable EOS is defined then at least set mu variable, 
// affecting the caloric EOS
#ifndef VARIABLE_MU 
#define VARIABLE_MU
#endif
#endif

#ifdef H_IONISATION

#ifndef VARIABLE_EOS
#define VARIABLE_EOS
#endif 

#ifndef VARIABLE_MU
#define VARIABLE_MU
#endif 

#endif H_IONISATION

//================================================================================
//
//
// degree of cell value reconstruction for zoneField values
//
// 1 => the cell integral average  value is used, defined by exact left and right values
// by linear interpolation.  
// 1D: left and right are evaluated and averaged, 
//                  a = 1/2(al+ar), 2 points
// 2D: cell corners are evaluated and averaged  , 
//                  a = 1/4(a0+a1+a2+a3), 4 points
// 3D: cell corners are evaluated and averaged  , 
//                  a = 1/8(a0+a1+a2+a3+a4+a5+a6+a7), 8 points
//
// 2 => the cell integral average value is used, defined by exact left centre and right values
// using orthogonal conservative parabolic interpolation.
//
// 1D: centre left and right are evaluated and averaged, 
//                  a  =   1/6*[ 4*a_c + al + ar] 3 points
// 2D: cell corners are evaluated and averaged         , 
//                  a  =  1/36*[16*a_c +  4*[edge middles] + [ corners]] 9 points
// 3D: cell corners are evaluated and averaged         , 
//                  a  = 1/216*[64*a_c + 16*[face centres] + 4*[edge middles] + [corners])  27 points!
// expensive but only done once and is as accurate as the full PPM method, 
// and is conservative
// 
// Code for order 1 or 2 
// See init zone field in zls_gravity.c
#define FIELD_RECON_ORDER 2

//
//================================================================================
//
#ifndef COOLING
#ifdef VARIABLE_EOS 
#undef VARIABLE_EOS
#endif
#ifdef VARIABLE_MU 
#undef VARIABLE_MU
#endif
#ifdef VARIABLE_GAMMA 
#undef VARIABLE_GAMMA
#endif
#ifdef H_IONISATION
#undef H_IONISATION
#endif
#endif !COOLING
//
// Set up sweeps macro
// 
#ifdef HYDRO
#ifdef LAGRANGIAN

#ifdef SYM1D
#ifdef CARTESIAN
#define ZLSSWEEPS(z) zlsa_1DSweep(z)
#endif CARTESIAN
#ifdef SPHERICAL
#define ZLSSWEEPS(z) zlsa_1DSweepSphere(z)
#endif SPHERICAL
#ifdef POLARSPHERICAL
#define ZLSSWEEPS(z) zlsa_1DSweepSphere(z)
#endif POLARSPHERICAL
#ifdef CYLINDRICAL
#define ZLSSWEEPS(z) zlsa_1DSweep(z)
#endif CYLINDRICAL
#ifdef CIRCULAR
#define ZLSSWEEPS(z) zlsa_1DSweepCircle(z)
#endif CIRCULAR
#ifndef ZLSSWEEPS
#define ZLSSWEEPS(z) zlsa_1DSweep(z)
#endif ZLSSWEEPS
#endif SYM1D

#ifdef SYM2D
#ifdef CARTESIAN
#define ZLSSWEEPS(z) zlsa_2DSplitSweeps(z)
#endif CARTESIAN
#ifdef SPHERICAL
#define ZLSSWEEPS(z) zlsa_2DSplitSweepsSphere(z)
#endif SPHERICAL
#ifdef POLARSPHERICAL
#define ZLSSWEEPS(z) zlsa_2DSplitSweepsSphere(z)
#endif POLARSPHERICAL
#ifdef CYLINDRICAL
#define ZLSSWEEPS(z) zlsa_2DSplitSweepsCylinder(z)
#endif CYLINDRICAL
#ifdef CIRCULAR
#define ZLSSWEEPS(z) zlsa_2DSplitSweepsCircle(z)
#endif CIRCULAR
#ifndef ZLSSWEEPS
#define ZLSSWEEPS(z) zlsa_2DSplitSweeps(z)
#endif ZLSSWEEPS
#endif SYM2D

#ifdef SYM3D
#ifdef CARTESIAN
#define ZLSSWEEPS(z) zlsa_3DSplitSweeps(z)
#endif CARTESIAN
#ifdef SPHERICAL
#define ZLSSWEEPS(z) zlsa_3DSplitSweepsSphere(z)
#endif SPHERICAL
#ifdef POLARSPHERICAL
#define ZLSSWEEPS(z) zlsa_3DSplitSweepsPolarSphere(z)
#endif POLARSPHERICAL
#ifdef CYLINDRICAL
#define ZLSSWEEPS(z) zlsa_3DSplitSweepsCylinder(z)
#endif CYLINDRICAL
#ifndef ZLSSWEEPS
#define ZLSSWEEPS(z) zlsa_3DSplitSweeps(z)
#endif ZLSSWEEPS
#endif SYM3D

#endif LAGRANGIAN

#ifdef EULERIAN

#ifdef SYM1D
#ifdef CARTESIAN
#define ZLSSWEEPS(z) zlsa_1DHDSweep(z)
#endif CARTESIAN
#endif SYM1D

#ifdef SYM2D 
#ifdef CARTESIAN
#define ZLSSWEEPS(z) zlsa_2DHDSweep(z)
#endif CARTESIAN
#endif SYM2D

#ifdef SYM3D
#ifdef CARTESIAN
#define ZLSSWEEPS(z) zlsa_3DHDSweep(z)
#endif CARTESIAN
#endif SYM3D

#define INITRIEMANNSOLVER(g)



#endif EULERIAN

#endif HYDRO
//
//
//
#ifdef HYDRO
#ifdef LAGRANGIAN
//
// Set up solver macros
// 
#ifdef TORO
//Exact solver from Toro 1998
#define INITRIEMANNSOLVER(g) zlsr_InitToroRiemannGammas (g)
#define RIEMANNSOLVER(x,y) zlsr_ToroRiemannW1D(x, y)
#endif

#ifdef TOROTWOSHOCK
//Modified solver from Toro 1998, by rss for two shock approximation, 
//simplifies iteration logic
#define INITRIEMANNSOLVER(g) zlsr_InitToroSSRiemannGammas (g)
#define RIEMANNSOLVER(x,y) zlsr_ToroSSRiemannW1D(x, y)
#endif

#ifdef PVRS
//Approximate primitive variable sover, OK for weak shocks and waves
// fixed gamma = global value
#define INITRIEMANNSOLVER(g) 
#define RIEMANNSOLVER(x,y) zlsr_PVRSRiemannW1D(x, y)
#endif

#ifdef AIRS
//Hybrid TORO/PVRS solver - OK, fixed gamma = global value
#define INITRIEMANNSOLVER(g) zlsr_InitAIRSRiemannGammas (g)
#define RIEMANNSOLVER(x,y) zlsr_AIRSRiemannW1D(x, y, 2.0)
#endif

#ifdef COLLELLATWOSHOCK
//original ppmlr solver, approximate and quite fast
#define INITRIEMANNSOLVER(g) zlsr_InitC84RiemannGammas (g)
#define RIEMANNSOLVER(x,y) zlsr_C84RiemannW1D(x, y)
#endif

#ifdef GROTHGOTTLIEB
//full exact solver from Groth and Gottlieb 1988, iterates on Ustar, 
// faster than Toro, van Leer, Godunov etc
#define INITRIEMANNSOLVER(g) zlsr_InitGG88RiemannGammas (g)
#define RIEMANNSOLVER(x,y) zlsr_GG88RiemannW1D(x, y)
#endif

#ifdef RSSTWOSHOCK
//Sutherland twoshock solver based on Groth and Gottlieb 1988, iterates on Ustar, 
//faster than Toro, van Leer, Godunov etc
#define INITRIEMANNSOLVER(g) zlsr_InitRSSRiemannGammas (g)
#define RIEMANNSOLVER(x,y) zlsr_RSSRiemannW1D(x, y)
#endif
#endif LAGRANGIAN

#endif HYDRO

#endif ZLS_DEFS_H
