#include <zls.h>

//====================================================================================
//
// Simulation Variable Update Algorithms
//
//====================================================================================
//
// these are constants, declared as vars only so they can be configured
// (rarely) in ppm.conf.  Regard them as read only.
//
Real    kZLS_SmoothThreshold = 0.001;
Real    kZLS_SmoothScale     = 1.0e6;
Real    kZLS_SmoothLimit     = 1.0;

// below set in zls_conf.c: zlsio_SetUpPPMFromConfigFile
Real    kZLS_RemapAlpha     = 0.025;  
Integer kZLS_SmoothAxes      = kSmoothNone; // Combination of smooth axes flags
#ifdef DIVVHCORRECTION
Real    kZLS_DivVScale       = 2.0;  
Real    kZLS_DivVInvScale    = 0.5;  
Real    kZLS_DivVThreshold   = 1.0;  
Real    kZLS_RemapMaxAlpha  = 0.10;  
#endif DIVVHCORRECTION

//
//====================================================================================
//

#ifdef EULERIAN
int zlsa_EulerianFluxesW1D( Flx1DArr flux,  Wstar1DArr wStar){
  //====================================================================================
  //  After this step the flux contains the godunov fluxes (s = 0) 
  //  for the riemann solutions
  // in xStar
  //====================================================================================
  int err = noErr;
  
  Integer i ;
  
  Real P0;
  Real E0;
  Real d0;
  Real u0;
  
  for ( i = 1; i < wStar.nt; i++){
    
    zlsr_RiemanSample(&P0, &u0, &d0, 0.0, wStar, i );
    flux.f[i].m  = d0*u0; 
    flux.f[i].mx = d0*u0*u0 + P0;
    E0  = (P0 /(kZLS_eos_gamma_1) ) + 0.5*d0*(u0 * u0 );
    flux.f[i].E  = (E0+P0)*u0;
    
  }
  
  return( err );
}

int zlsa_EulerianUpdateW1D( W1DArr w, Flx1DArr flux, Real dt){
  //====================================================================================
  //  After this step the flux contains the godunov fluxes (s = 0) 
  //  for the riemann solutions
  // in xStar
  //====================================================================================
  int err = noErr;
  
  Integer i ;
  
  Real dt_on_dx = dt;
  
  for ( i = 2; i < w.nt-2; i++){
    
    dt_on_dx = dt * w.x[i].idx;
    
    Real u0   = w.vx[i].a;
    Real d0   = w.d [i].a;
    Real P0   = w.P [i].a;
    
    Real mx0  =  u0*d0;
    Real E0   = (P0/kZLS_eos_gamma_1) + 0.5*d0*(u0 * u0);
    
    Real flxd   = flux.f[i].m  - flux.f[i+1].m ;
    Real flxmx  = flux.f[i].mx - flux.f[i+1].mx;
    Real flxE   = flux.f[i].E  - flux.f[i+1].E ;
    
    E0       +=  dt_on_dx*(flxE );
    mx0      +=  dt_on_dx*(flxmx);
    d0       +=  dt_on_dx*(flxd );
    u0        = mx0/d0;
    
    w.d[i].a   = d0;
    w.vx[i].a  = u0;
    w.P[i].a   = (E0 - 0.5*d0*(u0 * u0))*kZLS_eos_gamma_1;
    
  }
  
  return( err );
}

#endif EULERIAN


// conservatve remap smoothing with spatial filter to only smooth
// in regions where up-down noise occurs.

static inline Real dPvEnergy1D ( Real d, Real P, Real vx,  Real  gamm){
  d *= gamm;
  return((P /d ) + 0.5*(vx * vx ));
}

static inline Real dPvEnergy2D ( Real d, Real P, Real vx, Real vy,  Real  gamm){
  d *= gamm;
  return((P /d ) + 0.5*(vx * vx + vy * vy));
}

static inline Real dPvEnergy3D ( Real d, Real P, Real vx, Real vy, Real vz, Real  gamm ){
  d *= gamm;
  return((P /d ) + 0.5*(vx * vx + vy * vy + vz * vz));
}

//#define PSMTH
//#define DSMTH
#define VSMTH
//#define MSMTH

#ifdef LAGRANGIAN

int zlsa_RemapSmoothW1D( W1DArr w, Real dt ){
  
  int err = noErr;
#if (NDIMS > 1)
  
  Integer i;
  Integer     n = w.nt;
  
  Real eps      = kZLS_RemapAlpha;
  Real alpha[3] = { 0.005, 0.99, 0.005 };
  
  Real m;
  Real d[3]; 
  
#ifdef PSMTH
  Real psmooth   [n];
#endif PSMTH
  
#ifdef DSMTH
  Real dsmooth   [n];
#endif DSMTH
  
#ifdef MSMTH
  Real dsmooth    [n];
  Real mxsmooth   [n];
#if (NDIMS>1)
  Real mysmooth   [n];
#endif
#if (NDIMS>2)
  Real mzsmooth   [n];
#endif
  
#endif
  
#ifdef VSMTH
  Real vxsmooth   [n];
#if (NDIMS>1)
  Real vysmooth   [n];
#endif
#if (NDIMS>2)
  Real vzsmooth   [n];
#endif
#endif
  
  
  Real    osca  [n]; 
  Integer oscb  [n]; 
  
  Integer is = kZLS_LimitRemapLimits+1; 
  Integer ie = w.nt-kZLS_LimitRemapLimits-1; 
  
  for(i = is; i < ie; i++){ 
    osca[i]     = zls_min(kZLS_SmoothScale*((w.d[i  ].a-w.d[i-1].a)*(w.d[i+1].a-w.d[i  ].a))+kZLS_SmoothLimit, 0.0);
    oscb[i]     = 0;
  }
  
  Counter smCount = 0;
  
  is += 1;
  ie -= 1;
  oscb[is-1] = 0;
  oscb[ie  ] = 0;
  
  for(i = is; i < ie ; i++){ // set where to smooth cells
    oscb[i]  = ((osca[i-1]*osca[i]*osca[i+1]<0.0))?1:0;
   // oscb[i]  = (osca[i]<0.0)?1:0;
#ifdef DIVVHCORRECTION
#if (NDIMS > 1)
    oscb[i] |= (w.divv[i]>kZLS_DivVThreshold)?1:0;
#endif
#endif DIVVHCORRECTION
    smCount  += oscb[i];
  }
  
  if ( smCount > 0 ){  // only bother if there is something to smooth
    
    smCount = 0; // reset to find actual smoothing number
    
    for ( i = is; i < ie ; i++ ){
      
      eps  = kZLS_RemapAlpha;
#ifdef DIVVHCORRECTION
      eps  = kZLS_RemapAlpha*w.divv[i]*kZLS_DivVInvScale;
      eps  = ( eps > kZLS_RemapMaxAlpha) ? kZLS_RemapMaxAlpha: eps;
#endif DIVVHCORRECTION
      
      alpha[0]  = 0.5*eps*(oscb[i  ]|oscb[i-1]);
      alpha[2]  = 0.5*eps*(oscb[i  ]|oscb[i+1]);
      alpha[1]  = (1.0-(alpha[0]+alpha[2]));
      
      if ( alpha[1] < 0.999999 ){
        
        smCount += 1;
        
#ifdef PSMTH
        d[0] = w.P[i-1].a;
        d[1] = w.P[i  ].a;
        d[2] = w.P[i+1].a;
        m              = (alpha[0]*d[0] + 
                          alpha[1]*d[1] + 
                          alpha[2]*d[2]);
        psmooth  [i]   = m;
#endif PSMTH
        
#ifdef DSMTH
        d[0] = w.d[i-1].a;
        d[1] = w.d[i  ].a;
        d[2] = w.d[i+1].a;
        m              = (alpha[0]*d[0] + 
                          alpha[1]*d[1] + 
                          alpha[2]*d[2]);
        dsmooth  [i]   = m;
#endif DSMTH
        
#ifdef VSMTH
        //       if ( w.sweep == kXDim ){
        d[0] = w.vx[i-1].a;
        d[1] = w.vx[i  ].a;
        d[2] = w.vx[i+1].a;
        m              = (alpha[0]*d[0] + 
                          alpha[1]*d[1] + 
                          alpha[2]*d[2]);
        vxsmooth  [i]   = m;
        //       } else { vxsmooth  [i]   = w.vx[i  ].a; }          
        
#if (NDIMS>1)
        d[0] = w.vy[i-1].a;
        d[1] = w.vy[i  ].a;
        d[2] = w.vy[i+1].a;
        m              = (alpha[0]*d[0] + 
                          alpha[1]*d[1] + 
                          alpha[2]*d[2]);
        vysmooth  [i]   = m;
#endif (NDIMS>1)
#if (NDIMS>2)
        d[0] = w.vz[i-1].a;
        d[1] = w.vz[i  ].a;
        d[2] = w.vz[i+1].a;
        m              = (alpha[0]*d[0] + 
                          alpha[1]*d[1] + 
                          alpha[2]*d[2]);
        vzsmooth  [i]   = m;
#endif (NDIMS>2)
#endif VSMTH
        
#ifdef MSMTH
        d[0] = w.d[i-1].a;
        d[1] = w.d[i  ].a;
        d[2] = w.d[i+1].a;
        m              = (alpha[0]*d[0] + 
                          alpha[1]*d[1] + 
                          alpha[2]*d[2]);
        dsmooth  [i]   = m;
        
        d[0] = w.vx[i-1].a*w.d[i-1].a;
        d[1] = w.vx[i  ].a*w.d[i  ].a;
        d[2] = w.vx[i+1].a*w.d[i+1].a;
        m              = (alpha[0]*d[0] + 
                          alpha[1]*d[1] + 
                          alpha[2]*d[2]);
        mxsmooth  [i]   = m;
        
#if (NDIMS>1)
        d[0] = w.vy[i-1].a*w.d[i-1].a;
        d[1] = w.vy[i  ].a*w.d[i  ].a;
        d[2] = w.vy[i+1].a*w.d[i+1].a;
        m              = (alpha[0]*d[0] + 
                          alpha[1]*d[1] + 
                          alpha[2]*d[2]);
        mysmooth  [i]   = m;
#endif (NDIMS>1)
        
#if (NDIMS>2)
        d[0] = w.vz[i-1].a*w.d[i-1].a;
        d[1] = w.vz[i  ].a*w.d[i  ].a;
        d[2] = w.vz[i+1].a*w.d[i+1].a;
        m              = (alpha[0]*d[0] + 
                          alpha[1]*d[1] + 
                          alpha[2]*d[2]);
        mzsmooth  [i]   = m;
#endif (NDIMS>2)
        
#endif MSMTH
        
      } else {
        
#ifdef PSMTH
        psmooth  [i]  =  w.P[i].a;
#endif PSMTH
        
#ifdef DSMTH
        dsmooth  [i]  =  w.d[i].a;
#endif DSMTH
        
#ifdef VSMTH
        vxsmooth [i]  = w.vx[i].a;
#if (NDIMS>1)
        vysmooth  [i]  = w.vy[i].a;
#endif (NDIMS>1)
#if (NDIMS>2)
        vzsmooth  [i]  = w.vz[i].a;
#endif (NDIMS>2)
#endif VSMTH
        
#ifdef MSMTH
        dsmooth  [i]  =  w.d[i].a;
        mxsmooth [i]  = w.vx[i].a* dsmooth[i];
#if (NDIMS>1)
        mysmooth [i]  = w.vy[i].a* dsmooth[i];
#endif (NDIMS>1)
#if (NDIMS>2)
        mzsmooth [i]  = w.vz[i].a* dsmooth[i];
#endif (NDIMS>2)
#endif MSMTH
      }
      
    }
    
    if (smCount > 0) {
      for(i = is; i < ie ; i++){ // replace
        
#ifdef PSMTH
        w.P [i ].a = psmooth [i];
#endif PSMTH
#ifdef DSMTH
        w.d [i ].a = dsmooth [i];
#endif DSMTH
#ifdef VSMTH
        w.vx[i ].a = vxsmooth[i]; 
#if (NDIMS>1)
        w.vy[i ].a = vysmooth[i]; 
#endif (NDIMS>1)
#if (NDIMS>2)
        w.vz[i ].a = vzsmooth[i]; 
#endif (NDIMS>2)
#endif VSMTH
        
#ifdef MSMTH
        w.d [i ].a = dsmooth [i];
        w.vx[i ].a = mxsmooth[i]/dsmooth[i]; 
#if (NDIMS>1)
        w.vy[i ].a = mysmooth[i]/dsmooth[i]; 
#endif (NDIMS>1)
#if (NDIMS>2)
        w.vz[i ].a = mzsmooth[i]/dsmooth[i]; 
#endif (NDIMS>2)
#endif MSMTH
        
      }
    }
    
  }
  
#endif   
  return(err);
  
}

#ifdef GRAVITY
int zlsa_LagrangianCellAccelleration1D ( Real * a0, Real * a1, Integer nt,  ACoordPtr x0,  ACoordPtr x1, F1DArr fld, Real dt){
  int err = noErr;
  
  Real p1[nt]; // potential at new cell boundary locations in x1
  //
  // compute the mean gravity accelleration in the initial zone cells and evolved cells
  // g0 = dPhi0/dx0, g1 = dPhi1/dx1, a0 = -0.5*dt*g0, a1 = -0.5*dt*g1;
  //
  
  
  register Integer  i = 0;
  register Integer  k = 1;
  
  register Real   f   = 0.0; 
#if (FIELD_RECON_ORDER == 2)
  register Real   cf  = 1.0;
#endif
  
  register Real hdt   = 0.5*dt;  // half time step for averaging normalisation, saving ops in the evolve routine.
  
  register Real deltaX   = (x1[0].l - x0[0].l);
  register Real Phi      =  fld.xi[0].al;
  
  register Real  invDx0  = x0[0].idx; 
  register Real  invDx1  = x1[0].idx;
  
  p1[0] = fld.xi[0].al; // first cell left doesn't move;
  
  // compute potential at left cell boundaries in p1[i], copy to right value of cell - 1, 
  
  for ( i = 1; i < nt; i++ ) {
    
    k = i - 1;
    
    deltaX = (x1[i].l - x0[i].l); // boundary movement
    Phi    =  fld.xi[i].al;  // =  f.xi[k].ar;
    
    if ( deltaX < -kZLS_Small ){ // left edge moved left, into cell i-1 = cell k;
      
      f    = -deltaX*x0[k].idx; // cell fraction, deltaX -ve, f +ve
#if (FIELD_RECON_ORDER == 1)
      Phi -= f*(fld.xi[k].ar - fld.xi[k].al); // PLM interpolation
#endif
#if (FIELD_RECON_ORDER == 2)
      cf   = 1.0 - f;
      Phi -= f*(fld.xi[k].del - cf* fld.xi[k].a6); // PPM interpolation
#endif
      
    } else if ( deltaX  > kZLS_Small ) {
      
      f    = deltaX*x0[i].idx; // cell fraction, deltaX +ve , f +ve
#if (FIELD_RECON_ORDER == 1)
      Phi += f*(fld.xi[i].ar - fld.xi[i].al); // PLM interpolation
#endif
#if (FIELD_RECON_ORDER == 2)
      cf   = 1.0 - f;
      Phi += f*(fld.xi[i].del + cf*fld.xi[i].a6);
#endif
      
    } 
    
    p1[i] = Phi;
    
    invDx0 = x0[k].idx*hdt;
    a0[k]  = -(fld.xi[k].ar - fld.xi[k].al)*invDx0; // mean accelleration in original cell domian in cell i-1
    invDx1 = x1[k].idx*hdt;
    a1[k]  = -(p1[i] - p1[k])*invDx1; // mean accelleration in new cell domian in cell i-1
    
    a0[k] = zls_tolx(a0[k], kZLS_Small);
    a1[k] = zls_tolx(a1[k], kZLS_Small);
    
  }
  
  // last cell
  
  i = nt - 1;
  k =  i - 1;
  
  invDx0 = x0[i].idx*hdt;
  a0[i]  = -( fld.xi[i].ar - fld.xi[i].al)*invDx0; // mean accelleration in original cell domian
  
  invDx1 = x1[i].idx*hdt;
  a1[i]  = -( fld.xi[i].ar - p1[i] )*invDx1; // mean accelleration in new cell domian, right edge fixed
  
  a0[i] = zls_tolx(a0[i], kZLS_Small);
  a1[i] = zls_tolx(a1[i], kZLS_Small);
  
  return(err);
}
#endif GRAVITY

int zlsa_LagrangianSimpleEvolveW1D( W1DArr w,  Wstar1DArr wStar, Real dt){
  //====================================================================================
  //  After this step the axis in w is modified by the cell boundary motions, and
  //  the axis in wStar stays unchanged, so the differences can be used to compute 
  //  remapping fluxes.
  //====================================================================================
  //
  // no correction for striping instability, doesn't require flat array.  Good for 
  // 1D and adiabatic 2D models without extreme values.
  // not compatible with VARIABLE_GAMMA 
  //
  int err = noErr;
	
  Integer i, is, ie ;
  Integer ni, nti   ;
	
  Real cellMass     ;
  Real volumeRatio  ;
  Real dtdM         ;
	
  Real PMid         ;
  Real PMidp1       ;
  Real vxPMid       ;
  Real vxPMidp1     ;
	
  Real P         ;
  Real E         ;
  Real d, V0, V1 ;
	
#ifdef SYM1D
  Real vx;
#endif SYM1D
#ifdef SYM2D
  Real vx, vy;
#endif SYM2D
#ifdef SYM3D
  Real vx, vy, vz;
#endif SYM3D
	
  //
  // move the cell boundaries according to the velocity of the Riemann CD speed
  // in WStar.vx
  //
  // only evolve cells where both boundaries move, and more than remap
  //
  ni  = w.n;
  is  = 2;
  ie  = w.nt-2;
  nti = w.nt  ;
  //
  //========================================
  // Lagangian method!  Moving the cell boundaries with wStar takes place here.
  //========================================
  //
	
  for ( i = 1; i < w.nt; i++){
    w.x[i].l += wStar.v[i].a*dt;
  }
	
  // make the right boundaries consistent with the lefts set above
  for ( i = 0; i < w.nt-1; i++){
    w.x[i].r = w.x[i+1].l;
  }
	
  //
  // w0 axis is modified, recompute dxs and the purely grid parabolic coefficients
  // so ppm a6 and del coeffs will work for the now non-uniformly spaced axis coords.
  //
	
  err = zls_CalcAxisDeltas (w.x, kAxisTypeSpatial, w.nt); 
	
  //
  // now go through and update the densities, energy density and velocities
  //
  // in 1d, the volume is just dx.  wStar has the original dx, and w has the new dx
  //
  for ( i = is; i < ie; i++){
		
    V0          = wStar.x[i].dx   ; // initial volume
    V1          = w.x[i].dx       ; // final volume
		
    PMid        = wStar.P[i].a    ; // left CD pressure
    PMidp1      = wStar.P[i+1].a  ; // right CD pressure
    vxPMid      = wStar.v[i].a    * PMid   ; // left v* pressure - e flux
    vxPMidp1    = wStar.v[i+1].a  * PMidp1 ; // right v* pressure - e flux
		
    d           =  w.d[i].a ; // initial average density
    E           =  w.E[i].a ; // initial average Energy Density
#ifdef SYM1D
    vx          =  w.vx[i].a ; // initial average velx
#endif SYM1D
#ifdef SYM2D
    vx          =  w.vx[i].a ; // initial average velx
    vy          =  w.vy[i].a ; // initial average vely
#endif SYM2D
#ifdef SYM3D
    vx          =  w.vx[i].a ; // initial average velx
    vy          =  w.vy[i].a ; // initial average vely
    vz          =  w.vz[i].a ; // initial average vely
#endif SYM3D
		
    cellMass    = d * V0;
    volumeRatio = V0/V1;
    dtdM        = dt/cellMass ;
		
    d		   *= volumeRatio; // new density
#if (NTRACERS > 0)    
    Integer k;
    for (k = 0; k < kNTracers; k++){
      w.tr[k][i].a *= volumeRatio;  // trs are like density
    }
#endif

    vx        -= dtdM*(  PMidp1 -   PMid );  // new velocity		
    E         -= dtdM*( vxPMidp1 - vxPMid );  // new lagrangian total energy density  (P/(gamm d) + 0.5v^2)
		
#ifdef SYM1D
    P  = zls_max(kZLS_PressureVacuum,d*(E - 0.5*(vx*vx))*kZLS_eos_gamma_1);
#endif SYM1D
#ifdef SYM2D
    P  = zls_max(kZLS_PressureVacuum,d*(E - 0.5*(vx*vx+vy*vy))*kZLS_eos_gamma_1);
#endif SYM2D
#ifdef SYM3D
    P  = zls_max(kZLS_PressureVacuum,d*(E - 0.5*(vx*vx+vy*vy+vz*vz))*kZLS_eos_gamma_1);
#endif SYM3D
		
    w.d[i].a   =  d;  // return values to vector.  mass per volume
    w.E[i].a   =  E;  // internal energy per volume
    w.P[i].a   =  P;  // kZLS_eos_gamma_1 * internal energy per volume
#ifdef ACTIVE_FP_NOISE_CONTROL   
    w.vx[i].a  = zls_tolx(vx, kZLS_Small);  // momentum per mass
#else
    w.vx[i].a  = vx;  // momentum per mass
#endif ACTIVE_FP_NOISE_CONTROL
  }
	
  return(err);
}

enum {
  kNothingMoved   = 0,
  kSomethingMoved = 1
};

int zlsa_LagrangianEvolveW1D( W1DArr w,  Wstar1DArr wStar, Real dt){
  
  int err = noErr;
	
  Integer i, is, ie ;
	
  Real PMid         ;
  Real PMidp1       ;
  Real DeltaPMid    ;
  Real vxPMid       ;
  Real vxPMidp1     ;
  Real DeltavxPMid  ;
	
  Real d, V0, V1;
  Real volumeRatio  ;
  Real cellMass     ;
  Real dtdM         ;
  
  //
  // move the cell boundaries according to the velocity of the Riemann CD speed
  // in WStar.vx
  //
  //========================================
  // Lagangian method!  Moving the cell boundaries with wStar takes place here.
  //========================================
  //
  // move the cell boundary at the CD velocity in the wStar, umid.
  
#ifdef ACTIVE_FP_NOISE_CONTROL   
  Integer recalculate = kNothingMoved;
#endif ACTIVE_FP_NOISE_CONTROL   
  
  // cell 0 boundary is left fixed, cell nt-1 right boundary is left fixed
  
  for ( i = 1; i < w.nt; i++){
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    Real dx = zls_tolx((wStar.v[i].a*dt), kZLS_Small);
    if ( dx != 0.0 ) {
      recalculate = kSomethingMoved;
      w.x[i].l += dx; 
    } else {
      wStar.v[i].a = 0.0;
    }
#else
#ifdef DIVVHCORRECTION
    w.x[i].l +=  (w.divv[i]<kZLS_DivVThreshold)?wStar.v[i].a*dt:0.0; 
#else
    w.x[i].l +=  wStar.v[i].a*dt; 
#endif DIVVHCORRECTION
	  recalculate = kSomethingMoved;
#endif CONTROL_EVOLVE
#else 
    w.x[i].l += wStar.v[i].a*dt; 
#endif ACTIVE_FP_NOISE_CONTROL
  }
  
#ifdef ACTIVE_FP_NOISE_CONTROL   
  if ( recalculate == kSomethingMoved ) {
#endif ACTIVE_FP_NOISE_CONTROL
    // make the right boundaries consistent with the lefts set above
    for ( i = 0; i < w.nt-1; i++){
      w.x[i].r = w.x[i+1].l;
    }
    
    //
    // w0 axis is modified, recompute dxs and the purely grid parabolic coefficients
    // so ppm a6 and del coeffs will work for the now non-uniformly spaced axis coords.
    //
    
    err = zls_CalcAxisDeltas (w.x, kAxisTypeSpatial, w.nt); 
    
#ifdef ACTIVE_FP_NOISE_CONTROL   
  }
#endif ACTIVE_FP_NOISE_CONTROL
  //
  //
  // only evolve cells where both boundaries can move, and more than remap
  //
  is  = kZLS_LimitEvolveLimits; //w.ns-kEvolveLimits;
  ie  = w.nt-kZLS_LimitEvolveLimits; //w.ne+kEvolveLimits;
  // now go through and update the densities, energy density and velocities
  //
  // in 1d, the volume is just dx.  wStar has the original dx, and w has the new dx
  //
  for ( i = is; i < ie; i++){
    
    
    PMid        = wStar.P[  i].a  ; // left CD pressure
    PMidp1      = wStar.P[i+1].a  ; // right CD pressure
    DeltaPMid   = (PMidp1 - PMid);
    
    vxPMid      = wStar.v[  i].a  * PMid   ; // left v* pressure - e flux
    vxPMidp1    = wStar.v[i+1].a  * PMidp1 ; // right v* pressure - e flux
    DeltavxPMid = (vxPMidp1 - vxPMid);
    
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    DeltavxPMid = zls_tolx( DeltavxPMid, kZLS_Small );
    DeltaPMid   = zls_tolx( DeltaPMid,   kZLS_Small );    
    if ((DeltavxPMid != 0.0) || (DeltaPMid != 0.0)) {
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL   
      V0          = wStar.x[i].dx   ; // initial volume
      V1          = w.x[i].dx       ; // final volume
      
      d           =  w.d[i].a ; // initial average density
      cellMass    =  d * V0;
      volumeRatio = ( V0==V1 )?1.0:V0/V1;
      
      d		       *= volumeRatio; // new density
      w.d[i].a    =  d;  // return values to vector.  mass per volume
#if (NTRACERS > 0)    
      Integer k;
      for (k = 0; k < kNTracers; k++){
        w.tr[k][i].a *= volumeRatio;  // trs are like density
      }
#endif
      
      dtdM        = dt/cellMass ;
      DeltavxPMid *= dtdM;
      DeltaPMid   *= dtdM;
      
      w.E[i].a   -=  DeltavxPMid;  
      w.vx[i].a  -=  DeltaPMid;
      w.P[i].a    = zlsp_PFromE_W1D( w, i);
      
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    }
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL   
    
  }; // i, is, ie
  
  
  return(err);
}

int zlsa_LagrangianEvolveW1DF ( W1DArr w,  Wstar1DArr wStar, F1DArr f, Real dt){
  // if f != NULL then it contains the gravity field potential components
#ifndef GRAVITY
#pragma unused(f)
#endif !GRAVITY
  
  int err = noErr;
	
  Integer i, is, ie ;
	
  Real PMid         ;
  Real PMidp1       ;
  Real DeltaPMid    ;
  Real vxPMid       ;
  Real vxPMidp1     ;
  Real DeltavxPMid  ;
#ifdef GRAVITY
  Real dAv          ;
#endif GRAVITY
	
  Real d, V0, V1;
  Real volumeRatio  ;
  Real cellMass     ;
  Real dtdM         ;
  Real vx0;
  Real vx1;
  Real E;
  
  //
  // move the cell boundaries according to the velocity of the Riemann CD speed
  // in WStar.vx
  //
  //========================================
  // Lagangian method!  Moving the cell boundaries with wStar takes place here.
  //========================================
  //
  
#ifdef ACTIVE_FP_NOISE_CONTROL   
  Integer recalculate = kNothingMoved;
#endif ACTIVE_FP_NOISE_CONTROL   
  
  // cell 0 boundary is left fixed, cell nt-1 right boundary is left fixed
  
  for ( i = 1; i < w.nt; i++){
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    Real dx = zls_tolx((wStar.v[i].a*dt), kZLS_Small);
    if ( dx != 0.0 ) {
      recalculate = kSomethingMoved;
      w.x[i].l += dx; 
    } else {
      wStar.v[i].a = 0.0;
    }
#else
#ifdef DIVVHCORRECTION
    w.x[i].l +=  (w.divv[i]<kZLS_DivVThreshold)?wStar.v[i].a*dt:0.0; 
#else
    w.x[i].l +=  wStar.v[i].a*dt; 
#endif DIVVHCORRECTION
	  recalculate = kSomethingMoved;
#endif CONTROL_EVOLVE
#else 
    w.x[i].l += wStar.v[i].a*dt; 
#endif ACTIVE_FP_NOISE_CONTROL
  }
  
#ifdef ACTIVE_FP_NOISE_CONTROL   
  if ( recalculate == kSomethingMoved ) {
#endif ACTIVE_FP_NOISE_CONTROL
    // make the right boundaries consistent with the lefts set above
    for ( i = 0; i < w.nt-1; i++){
      w.x[i].r = w.x[i+1].l;
    }
    
    //
    // w0 axis is modified, recompute dxs and the purely grid parabolic coefficients
    // so ppm a6 and del coeffs will work for the now non-uniformly spaced axis coords.
    //
    
    err = zls_CalcAxisDeltas (w.x, kAxisTypeSpatial, w.nt); 
    
#ifdef ACTIVE_FP_NOISE_CONTROL   
  }
#endif ACTIVE_FP_NOISE_CONTROL
  
  
  
#ifdef GRAVITY
  //
  // compute the mean gravity accelleration in the initial zone cells and evolved cells
  // and average the acceleration throughout dt.  
  // dv = dt * (g0+g1)*0.5,  dE = dt*(v0 * g0 + v1 *g1)*0.5
  //
  
  Real a0[w.nt]; // average g0 = dPhi/dx for initial cell domain, normalised so a0 = .5 x dt *g0
  Real a1[w.nt]; // average g1 = dPhi/dx for final cell domain, normalised so a1 = .5 x dt *g1
  // dv = (a0+a1), dE = (v0 * a0 + v1 *a1) 
  // Conservative accellerations based on cell widths and potential at boundaries using PPM
  // interpolation.
  
  if ( f.xi != NULL) {
    err = zlsa_LagrangianCellAccelleration1D ( a0, a1, w.nt, wStar.x, w.x, f, dt);
  }
  
#endif GRAVITY
  //
  //
  // only evolve cells where both boundaries can move, and more than remap
  //
  is  = kZLS_LimitEvolveLimits; //w.ns-kEvolveLimits;
  ie  = w.nt-kZLS_LimitEvolveLimits; //w.ne+kEvolveLimits;
  //
  // now go through and update the densities, energy density and velocities
  //
  // in 1d, the volume is just dx.  wStar has the original dx, and w has the new dx
  //
  for ( i = is; i < ie; i++){
    
#ifdef GRAVITY
    dAv         = a0[i]+a1[i]; 
#endif GRAVITY
    
    PMid        = wStar.P[  i].a  ; // left CD pressure
    PMidp1      = wStar.P[i+1].a  ; // right CD pressure
    DeltaPMid   = (PMidp1 - PMid);
    
    
    vxPMid      = wStar.v[  i].a  * PMid   ; // left v* pressure - e flux
    vxPMidp1    = wStar.v[i+1].a  * PMidp1 ; // right v* pressure - e flux
    DeltavxPMid = (vxPMidp1 - vxPMid);
    
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    DeltavxPMid = zls_tolx( DeltavxPMid, kZLS_Small );
    DeltaPMid   = zls_tolx( DeltaPMid,   kZLS_Small );  
#ifdef GRAVITY
    int Delta = (DeltavxPMid != 0.0) || (DeltaPMid != 0.0)||( dAv != 0.0);
#else
    int Delta =  (DeltavxPMid != 0.0) || (DeltaPMid != 0.0);
#endif !GRAVITY
    if ( Delta ) {
      
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL   
      V0          = wStar.x[i].dx   ; // initial volume
      V1          = w.x[i].dx       ; // final volume
      
      d           =  w.d[i].a ; // initial average density
      cellMass    =  d * V0;
      volumeRatio = ( V0==V1 )?1.0:V0/V1;
      
      d		       *= volumeRatio; // new density
      w.d[i].a    =  d;  // return values to vector.  mass per volume
#if (NTRACERS > 0)    
      Integer k;
      for (k = 0; k < kNTracers; k++){
        w.tr[k][i].a *= volumeRatio;  // trs are like density
      }
#endif
      
      dtdM        = dt/cellMass ;
      DeltavxPMid *= dtdM;
      DeltaPMid   *= dtdM;
      
      vx0         = w.vx[i].a; // initial average velx, keep for E Equation
#ifdef GRAVITY
      vx1         = vx0 - DeltaPMid + dAv ;  // momentum per mass
#else
      vx1         = vx0 - DeltaPMid ;  // momentum per mass
#endif
      w.vx[i].a   = vx1;
      
      E           =  w.E [i].a ; // initial average Energy Density
#ifdef GRAVITY
      w.E[i].a    =  E - DeltavxPMid + (a0[i]*vx0 + a1[i]*vx1);  // internal energy per volume
#else
      w.E[i].a    =  E - DeltavxPMid;  // internal energy per volume
#endif
      w.P[i].a    = zlsp_PFromE_W1D( w, i);
      
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    }
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL   
    
  }; // i, is, ie
  
  return(err);
}


int zlsa_LagrangianEvolveAxisTypeW1D  ( W1DArr w,  Wstar1DArr wStar, Integer axisType, Real g, Real dt){  
  
  int err = noErr;
	
  Integer i, is, ie ;
  Integer ni, nti   ;
	
  Real cellMass     ;
  Real volumeRatio  ;
  Real dtdM         ;
  Real dA, dAp1     ;
	
  Real PMid         ;
  Real PMidp1       ;
  Real vxPMid       ;
  Real vxPMidp1     ;
	
  Real E;
  Real d, V0, V1;
  
  Real invg;
	
  Real vx;
  
  //
  // move the cell boundaries according to the velocity of the Riemann CD speed
  // in WStar.vx
  //
  // only evolve cells where both boundaries move, and more than remap
  //
  ni  = w.n;
  is = kZLS_LimitEvolveLimits; //w.ns-kEvolveLimits;
  ie = w.nt-kZLS_LimitEvolveLimits; //w.ne+kEvolveLimits;
  nti = w.nt  ;
  
  invg = 1.0/g; // inverse of the angular projection factor  (radius, radius sin theta, etc)
  //
  //========================================
  // Lagangian method  Moving the cell boundaries with wStar takes place here.
  //========================================
  //
  // move the cell boundary at the CD velocity in the wStar, umid.
#ifdef CARTESIAN  
  for ( i = 1; i < w.nt; i++){
    w.x[i].l += wStar.v[i].a*dt;
  }
#endif CARTESIAN  
#ifdef SPHERICAL    
  switch (axisType) {
    default:
    case kAxisTypeSRadius:
      for ( i = 1; i < w.nt; i++){
        w.x[i].l += wStar.v[i].a*dt; //         
      }
      break;
      case kAxisTypePhi:
      case kAxisTypeTheta:
      for ( i = 1; i < w.nt; i++){
        w.x[i].l += wStar.v[i].a*dt*invg; // limit fluxes when in flat zone of orthogonal shock.
      }
  }// switch
#endif SPHERICAL    
  
#ifdef POLARSPHERICAL    
  switch (axisType) {
    default:
    case kAxisTypeSRadius:
      for ( i = 1; i < w.nt; i++){
        w.x[i].l += wStar.v[i].a*dt; // limit fluxes when in flat zone of orthogonal shock.
      }
      break;
      case kAxisTypeTheta:
      case kAxisTypePhi:
      for ( i = 1; i < w.nt; i++){
        w.x[i].l += wStar.v[i].a*dt*invg; // limit fluxes when in flat zone of orthogonal shock.
      }
  }// switch
#endif POLARSPHERICAL    
  
#ifdef CYLINDRICAL    
  switch (axisType) {
    default:
    case kAxisTypeSpatial:
    case kAxisTypeCRadius:
      for ( i = 1; i < w.nt; i++){
        w.x[i].l += wStar.v[i].a*dt; // limit fluxes when in flat zone of orthogonal shock.
      }
      break;
      case kAxisTypePhi:
      for ( i = 1; i < w.nt; i++){
        w.x[i].l += wStar.v[i].a*dt*invg; // limit fluxes when in flat zone of orthogonal shock.
      }
  }// switch
#endif CYLINDRICAL    
  
#ifdef CIRCULAR    
  switch (axisType) {
    default:
    case kAxisTypeCRadius:
      for ( i = 1; i < w.nt; i++){
        w.x[i].l += wStar.v[i].a*dt; // limit fluxes when in flat zone of orthogonal shock.
      }
      break;
      case kAxisTypePhi:
      for ( i = 1; i < w.nt; i++){
        w.x[i].l += wStar.v[i].a*dt*invg; // limit fluxes when in flat zone of orthogonal shock.
      }
  }// switch
#endif CIRCULAR    
  
  // make the right boundaries consistent with the lefts set above
  for ( i = 0; i < w.nt-1; i++){
    w.x[i].r = w.x[i+1].l;
  }
  
  //
  // w0 axis is modified, recompute dxs and the purely grid parabolic coefficients
  // so ppm a6 and del coeffs will work for the now non-uniformly spaced axis coords.
  //
  
  err = zls_CalcAxisDeltas (w.x, axisType, w.nt);  
  
#ifndef CARTESIAN
  err = zls_CalcAxisGeometrySource (w.x, wStar.x, axisType, w.nt);  
#endif !CARTESIAN
  //
  //
  // now go through and update the densities, energy density and velocities
  //
  for ( i = is; i < ie; i++){
    
    dA          = w.x[i].dA;
    dAp1        = w.x[i+1].dA;
    
    //  zlsio_Printf( " Evolve %d %g %g %g %g \n", i, g, V0, V1, dA);
    
    // time averaged area ratio of boundaries, geometry source term, == 1.0 in cartesian coords,
    // this routine is intended for non-cartesian coords, although it will work for them too.
    // Call the plain Evolve routine and this step is not used.
    
    PMid        = wStar.P[i].a    ; // left CD pressure
    PMidp1      = wStar.P[i+1].a  ; // right CD pressure
    vxPMid      = wStar.v[i].a    * PMid   ; // left v* pressure - e flux
    vxPMidp1    = wStar.v[i+1].a  * PMidp1 ; // right v* pressure - e flux
    
    
    V0          = wStar.x[i].dVol*g ; // initial volume
    V1          = w.x[i].dVol*g     ; // final volume
    
    d           =  w.d[i].a ; // initial average density
    
    cellMass    = d * V0;
    volumeRatio = V0/V1;
    dtdM        = dA*dt/cellMass ;
    
    d		   *= volumeRatio; // new density
    w.d[i].a   =  d;  // return values to vector.  mass per volume
#if (NTRACERS > 0)    
    Integer k;
    for (k = 0; k < kNTracers; k++){
      w.tr[k][i].a *= volumeRatio;  // trs are like density
    }
#endif

    E           =  w.E[i].a ; // initial average Energy Density
    vx          =  w.vx[i].a ; // initial average velx
    vx        -= dtdM*(  PMidp1 -   PMid )*( dA + dAp1 )*0.5;  // new velocity		
    E         -= dtdM*( vxPMidp1*dAp1 - vxPMid*dA );  // new lagrangian total energy density  (P/(gamm d) + 0.5v^2)
    w.E[i].a   =  E;  // internal energy per volume
    w.vx[i].a  = vx;  // momentum per mass
    
    
  }
  
  return(err);
}

#ifndef CARTESIAN
// each axis is different and a slightly different evolve type is needed for each axis

int zlsa_LagrangianEvolveSRadialW1D( W1DArr w,  Wstar1DArr wStar, Real dt){
  
  int err = noErr;
	
  Integer i, is, ie ;
	
  Real dA         ;
  Real dAp1       ;
  Real PMid         ;
  Real PMidp1       ;
  Real DeltaPMid    ;
  Real vxPMid       ;
  Real vxPMidp1     ;
  Real DeltavxPMid  ;
	
  Real d, V0, V1;
  Real volumeRatio  ;
  Real cellMass     ;
  Real dtdM         ;
  
  //
  // move the cell boundaries according to the velocity of the Riemann CD speed
  // in WStar.vx
  //
  //========================================
  // Lagangian method!  Moving the cell boundaries with wStar takes place here.
  //========================================
  //
  // move the cell boundary at the CD velocity in the wStar, umid.
  
#ifdef ACTIVE_FP_NOISE_CONTROL   
  Integer recalculate = kNothingMoved;
#endif ACTIVE_FP_NOISE_CONTROL   
  
  // cell 0 boundary is left fixed, cell nt-1 right boundary is left fixed
  
  for ( i = 1; i < w.nt; i++){
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    Real dx = wStar.v[i].a*dt;
	  dx = zls_tolx(dx, kZLS_Small);
    if ( dx != 0.0 ) {
      recalculate = kSomethingMoved;
      w.x[i].l += dx; 
    } else {
      wStar.v[i].a = 0.0;
    }
#else
    w.x[i].l += wStar.v[i].a*dt; 
	  recalculate = kSomethingMoved;
#endif CONTROL_EVOLVE
#else 
    w.x[i].l += wStar.v[i].a*dt; 
#endif ACTIVE_FP_NOISE_CONTROL
  }
  
#ifdef ACTIVE_FP_NOISE_CONTROL   
  if ( recalculate == kSomethingMoved ) {
#endif ACTIVE_FP_NOISE_CONTROL
    // make the right boundaries consistent with the lefts set above
    for ( i = 0; i < w.nt-1; i++){
      w.x[i].r = w.x[i+1].l;
    }
    
    //
    // w0 axis is modified, recompute dxs and the purely grid parabolic coefficients
    // so ppm a6 and del coeffs will work for the now non-uniformly spaced axis coords.
    //
    
    err = zls_CalcAxisDeltas (w.x, kAxisTypeSRadius, w.nt);  
    err = zls_CalcAxisGeometrySource (w.x, wStar.x, kAxisTypeSRadius, w.nt);  
    
#ifdef ACTIVE_FP_NOISE_CONTROL   
  }
#endif ACTIVE_FP_NOISE_CONTROL
  //
  //
  // only evolve cells where both boundaries can move, and more than remap
  //
  
  is  = kZLS_LimitEvolveLimits; //w.ns-kEvolveLimits;
  ie  = w.nt-kZLS_LimitEvolveLimits; //w.ne+kEvolveLimits;
  
  // now go through and update the densities, energy density and velocities
  //
  // in 1d, the volume is just dx.  wStar has the original dx, and w has the new dx
  //
  for ( i = is; i < ie; i++){
    
    dA          = w.x[  i].dA     ; // left CD area channge
    dAp1        = w.x[i+1].dA     ; // right CD area change
    
    PMid        = wStar.P[  i].a  ; // left CD pressure
    PMidp1      = wStar.P[i+1].a  ; // right CD pressure
    DeltaPMid   = (PMidp1 - PMid)*( dA + dAp1 )*0.5;
    
    vxPMid      = wStar.v[  i].a    * PMid    ; // left v* pressure - e flux
    vxPMidp1    = wStar.v[i+1].a  * PMidp1    ; // right v* pressure - e flux
    DeltavxPMid = (vxPMidp1*dAp1 - vxPMid*dA );
    
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    DeltavxPMid = zls_tolx( DeltavxPMid, kZLS_Small );
    DeltaPMid   = zls_tolx( DeltaPMid,   kZLS_Small );  
    
    if ((DeltavxPMid != 0.0) || (DeltaPMid != 0.0)) {
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL   
      
      V0          = wStar.x[i].dVol ; // initial volume
      V1          =     w.x[i].dVol ; // final volume
      d           = w.d[i].a ; // initial average density
      cellMass    =   d * V0 ;
      
      volumeRatio = ( V0==V1 )?1.0:V0/V1;
      d		       *= volumeRatio; // new density
      w.d[i].a    =  d;  // return values to vector.  mass per volume
      
#if (NTRACERS > 0)    
      Integer k;
      for (k = 0; k < kNTracers; k++){
        w.tr[k][i].a *= volumeRatio;  // trs are like density
      }
#endif

      dtdM        = dt/cellMass ;
      DeltavxPMid *= dtdM;
      DeltaPMid   *= dtdM;
      
      w.E[i].a   -= DeltavxPMid;  // internal energy per volume
      w.vx[i].a  -= DeltaPMid;    // momentum per mass
      w.P[i].a    = zlsp_PFromE_W1D( w, i);
      
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    }
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL   
    
  }; // i, is, ie
  
  
  return(err);
}



int zlsa_LagrangianEvolveSRadialW1DF ( W1DArr w,  Wstar1DArr wStar, F1DArr f, Real dt){
#ifndef GRAVITY
#pragma unused(f)
#endif !GRAVITY
  
  int err = noErr;
	
  Integer i, is, ie ;
	
  Real dA         ;
  Real dAp1       ;
  Real PMid         ;
  Real PMidp1       ;
  Real DeltaPMid    ;
  Real vxPMid       ;
  Real vxPMidp1     ;
  Real DeltavxPMid  ;
#ifdef GRAVITY
  Real dAv          ;
#endif GRAVITY
  
  Real d, V0, V1;
  Real volumeRatio  ;
  Real cellMass     ;
  Real dtdM         ;
  Real vx0;
  Real vx1;
  Real E;
  
  //
  // move the cell boundaries according to the velocity of the Riemann CD speed
  // in WStar.vx
  //
  //========================================
  // Lagangian method!  Moving the cell boundaries with wStar takes place here.
  //========================================
  //
  // move the cell boundary at the CD velocity in the wStar, umid.
  
#ifdef ACTIVE_FP_NOISE_CONTROL   
  Integer recalculate = kNothingMoved;
#endif ACTIVE_FP_NOISE_CONTROL   
  
  // cell 0 boundary is left fixed, cell nt-1 right boundary is left fixed
  
  for ( i = 1; i < w.nt; i++){
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    Real dx = wStar.v[i].a*dt;
	  dx = zls_tolx(dx, kZLS_Small);
    if ( dx != 0.0 ) {
      recalculate = kSomethingMoved;
      w.x[i].l += dx; 
    } else {
      wStar.v[i].a = 0.0;
    }
#else
    w.x[i].l += wStar.v[i].a*dt; 
	  recalculate = kSomethingMoved;
#endif CONTROL_EVOLVE
#else 
    w.x[i].l += wStar.v[i].a*dt; 
#endif ACTIVE_FP_NOISE_CONTROL
  }
  
#ifdef ACTIVE_FP_NOISE_CONTROL   
  if ( recalculate == kSomethingMoved ) {
#endif ACTIVE_FP_NOISE_CONTROL
    // make the right boundaries consistent with the lefts set above
    for ( i = 0; i < w.nt-1; i++){
      w.x[i].r = w.x[i+1].l;
    }
    
    //
    // w0 axis is modified, recompute dxs and the purely grid parabolic coefficients
    // so ppm a6 and del coeffs will work for the now non-uniformly spaced axis coords.
    //
    
    err = zls_CalcAxisDeltas (w.x, kAxisTypeSRadius, w.nt);  
    err = zls_CalcAxisGeometrySource (w.x, wStar.x, kAxisTypeSRadius, w.nt);  
    
#ifdef ACTIVE_FP_NOISE_CONTROL   
  }
#endif ACTIVE_FP_NOISE_CONTROL
  //
  
#ifdef GRAVITY
  //
  // compute the mean gravity accelleration in the initial zone cells and evolved cells
  // and average the acceleration throughout dt.  
  // dv = dt * (g0+g1)*0.5,  dE = dt*(v0 * g0 + v1 *g1)*0.5
  //
  
  Real a0[w.nt]; // average g0 = dPhi/dx for initial cell domain, normalised so a0 = .5 x dt *g0
  Real a1[w.nt]; // average g1 = dPhi/dx for final cell domain, normalised so a1 = .5 x dt *g1
  // dv = (a0+a1), dE = (v0 * a0 + v1 *a1) 
  // Conservative accellerations based on cell widths and potential at boundaries using PPM
  // interopolation.
  
  if ( f.xi != NULL) {
    err = zlsa_LagrangianCellAccelleration1D ( a0, a1, w.nt, wStar.x, w.x, f, dt);
  }
  
#endif GRAVITY
  //
  // only evolve cells where both boundaries can move, and more than remap
  //
  
  is  = kZLS_LimitEvolveLimits; //w.ns-kEvolveLimits;
  ie  = w.nt-kZLS_LimitEvolveLimits; //w.ne+kEvolveLimits;
  
  // now go through and update the densities, energy density and velocities
  //
  // in 1d, the volume is just dx.  wStar has the original dx, and w has the new dx
  //
  for ( i = is; i < ie; i++){
    
#ifdef GRAVITY
    dAv         = a0[i]+a1[i];
#endif GRAVITY
    
    dA          = w.x[  i].dA     ; // left CD area channge
    dAp1        = w.x[i+1].dA     ; // right CD area change
    
    PMid        = wStar.P[  i].a  ; // left CD pressure
    PMidp1      = wStar.P[i+1].a  ; // right CD pressure
    DeltaPMid   = (PMidp1 - PMid)*( dA + dAp1 )*0.5;
    
    vxPMid      = wStar.v[  i].a    * PMid    ; // left v* pressure - e flux
    vxPMidp1    = wStar.v[i+1].a  * PMidp1    ; // right v* pressure - e flux
    DeltavxPMid = (vxPMidp1*dAp1 - vxPMid*dA );
    
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    DeltavxPMid = zls_tolx( DeltavxPMid, kZLS_Small );
    DeltaPMid   = zls_tolx( DeltaPMid,   kZLS_Small );  
#ifdef GRAVITY
    int Delta = (DeltavxPMid != 0.0) || (DeltaPMid != 0.0)||( dAv != 0.0);
#else
    int Delta =  (DeltavxPMid != 0.0) || (DeltaPMid != 0.0);
#endif !GRAVITY
    if ( Delta ) {
      
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL   
      
      V0          = wStar.x[i].dVol ; // initial volume
      V1          =     w.x[i].dVol ; // final volume
      d           = w.d[i].a ; // initial average density
      cellMass    =   d * V0 ;
      
      volumeRatio = ( V0==V1 )?1.0:V0/V1;
      d		       *= volumeRatio; // new density
      w.d[i].a    =  d;  // return values to vector.  mass per volume
      
#if (NTRACERS > 0)    
      Integer k;
      for (k = 0; k < kNTracers; k++){
        w.tr[k][i].a *= volumeRatio;  // trs are like density
      }
#endif

      dtdM        = dt/cellMass ;
      DeltavxPMid *= dtdM;
      DeltaPMid   *= dtdM;
      
      vx0         = w.vx[i].a; // initial average velx, keep for E Equation
#ifdef GRAVITY
      vx1         = vx0 - DeltaPMid + dAv ;  // momentum per mass
#else
      vx1         = vx0 - DeltaPMid ;  // momentum per mass
#endif
      w.vx[i].a   = vx1;
      
      E           =  w.E [i].a ; // initial average Energy Density
#ifdef GRAVITY
      w.E[i].a    =  E - DeltavxPMid + (a0[i]*vx0 + a1[i]*vx1);  // internal energy per volume
#else
      w.E[i].a    =  E - DeltavxPMid;  // internal energy per volume
#endif
      
      w.P[i].a    = zlsp_PFromE_W1D( w, i);
      
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    }
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL   
    
  }; // i, is, ie
  
  
  return(err);
}



int zlsa_LagrangianEvolveCRadialW1D( W1DArr w,  Wstar1DArr wStar, Real dt){
  
  int err = noErr;
	
  Integer i, is, ie ;
	
  Real dA         ;
  Real dAp1       ;
  Real PMid         ;
  Real PMidp1       ;
  Real DeltaPMid    ;
  Real vxPMid       ;
  Real vxPMidp1     ;
  Real DeltavxPMid  ;
	
  Real d, V0, V1;
  Real volumeRatio  ;
  Real cellMass     ;
  Real dtdM         ;
  
  //
  // move the cell boundaries according to the velocity of the Riemann CD speed
  // in WStar.vx
  // move the cell boundary at the CD velocity in the wStar, umid.
  
#ifdef ACTIVE_FP_NOISE_CONTROL   
  Integer recalculate = kNothingMoved;
#endif ACTIVE_FP_NOISE_CONTROL   
  
  // cell 0 boundary is left fixed, cell nt-1 right boundary is left fixed
  
  for ( i = 1; i < w.nt; i++){
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    Real dx = wStar.v[i].a*dt;
	  dx = zls_tolx(dx, kZLS_Small);
    if ( dx != 0.0 ) {
      recalculate = kSomethingMoved;
      w.x[i].l += dx; 
    } else {
      wStar.v[i].a = 0.0;
    }
#else
    w.x[i].l += wStar.v[i].a*dt; 
	  recalculate = kSomethingMoved;
#endif CONTROL_EVOLVE
#else 
    w.x[i].l += wStar.v[i].a*dt; 
#endif ACTIVE_FP_NOISE_CONTROL
  }
  
#ifdef ACTIVE_FP_NOISE_CONTROL   
  if ( recalculate == kSomethingMoved ) {
#endif ACTIVE_FP_NOISE_CONTROL
    // make the right boundaries consistent with the lefts set above
    for ( i = 0; i < w.nt-1; i++){
      w.x[i].r = w.x[i+1].l;
    }
    
    //
    // w0 axis is modified, recompute dxs and the purely grid parabolic coefficients
    // so ppm a6 and del coeffs will work for the now non-uniformly spaced axis coords.
    //
    
    err = zls_CalcAxisDeltas (w.x, kAxisTypeCRadius, w.nt);  
    err = zls_CalcAxisGeometrySource (w.x, wStar.x, kAxisTypeCRadius, w.nt);  
    
#ifdef ACTIVE_FP_NOISE_CONTROL   
  }
#endif ACTIVE_FP_NOISE_CONTROL
  //
  //
  // only evolve cells where both boundaries can move, and more than remap
  //
  
  is  = kZLS_LimitEvolveLimits; //w.ns-kEvolveLimits;
  ie  = w.nt-kZLS_LimitEvolveLimits; //w.ne+kEvolveLimits;
  
  // now go through and update the densities, energy density and velocities
  //
  // in 1d, the volume is just dx.  wStar has the original dx, and w has the new dx
  //
  for ( i = is; i < ie; i++){
    
    dA          = w.x[  i].dA     ; // left CD area channge
    dAp1        = w.x[i+1].dA     ; // right CD area change
    
    PMid        = wStar.P[  i].a  ; // left CD pressure
    PMidp1      = wStar.P[i+1].a  ; // right CD pressure
    DeltaPMid   = (PMidp1 - PMid)*( dA + dAp1 )*0.5;
    
    vxPMid      = wStar.v[  i].a    * PMid    ; // left v* pressure - e flux
    vxPMidp1    = wStar.v[i+1].a  * PMidp1    ; // right v* pressure - e flux
    DeltavxPMid = (vxPMidp1*dAp1 - vxPMid*dA );
    
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    DeltavxPMid = zls_tolx( DeltavxPMid, kZLS_Small );
    DeltaPMid   = zls_tolx( DeltaPMid,   kZLS_Small );  
    
    if ((DeltavxPMid != 0.0) || (DeltaPMid != 0.0)) {
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL   
      
      V0          = wStar.x[i].dVol ; // initial volume
      V1          =     w.x[i].dVol ; // final volume
      d           = w.d[i].a ; // initial average density
      cellMass    =   d * V0 ;
      
      volumeRatio = ( V0==V1 )?1.0:V0/V1;
      d		       *= volumeRatio; // new density
      w.d[i].a    =  d;  // return values to vector.  mass per volume
      
#if (NTRACERS > 0)    
      Integer k;
      for (k = 0; k < kNTracers; k++){
        w.tr[k][i].a *= volumeRatio;  // trs are like density
      }
#endif

      dtdM        = dt/cellMass ;
      DeltavxPMid *= dtdM;
      DeltaPMid   *= dtdM;
      
      w.E[i].a   -= DeltavxPMid;  // internal energy per volume
      w.vx[i].a  -= DeltaPMid;    // momentum per mass
      w.P[i].a    = zlsp_PFromE_W1D( w, i);
      
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    }
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL   
    
  }; // i, is, ie
  
  
  return(err);
}

int zlsa_LagrangianEvolveCRadialW1DF( W1DArr w, Wstar1DArr wStar,  F1DArr f, Real dt ){
#ifndef GRAVITY
#pragma unused(f)
#endif !GRAVITY
  
  int err = noErr;
	
  Integer i, is, ie ;
	
  Real dA         ;
  Real dAp1       ;
  Real PMid         ;
  Real PMidp1       ;
  Real DeltaPMid    ;
  Real vxPMid       ;
  Real vxPMidp1     ;
  Real DeltavxPMid  ;
#ifdef GRAVITY
  Real dAv          ;
#endif GRAVITY
	
  Real d, V0, V1;
  Real volumeRatio  ;
  Real cellMass     ;
  Real dtdM         ;
  Real vx0;
  Real vx1;
  Real E;
  
  //
  // move the cell boundaries according to the velocity of the Riemann CD speed
  // in WStar.vx
  // move the cell boundary at the CD velocity in the wStar, umid.
  
#ifdef ACTIVE_FP_NOISE_CONTROL   
  Integer recalculate = kNothingMoved;
#endif ACTIVE_FP_NOISE_CONTROL   
  
  // cell 0 boundary is left fixed, cell nt-1 right boundary is left fixed
  
  for ( i = 1; i < w.nt; i++){
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    Real dx = wStar.v[i].a*dt;
	  dx = zls_tolx(dx, kZLS_Small);
    if ( dx != 0.0 ) {
      recalculate = kSomethingMoved;
      w.x[i].l += dx; 
    } else {
      wStar.v[i].a = 0.0;
    }
#else
    w.x[i].l += wStar.v[i].a*dt; 
	  recalculate = kSomethingMoved;
#endif CONTROL_EVOLVE
#else 
    w.x[i].l += wStar.v[i].a*dt; 
#endif ACTIVE_FP_NOISE_CONTROL
  }
  
#ifdef ACTIVE_FP_NOISE_CONTROL   
  if ( recalculate == kSomethingMoved ) {
#endif ACTIVE_FP_NOISE_CONTROL
    // make the right boundaries consistent with the lefts set above
    for ( i = 0; i < w.nt-1; i++){
      w.x[i].r = w.x[i+1].l;
    }
    
    //
    // w0 axis is modified, recompute dxs and the purely grid parabolic coefficients
    // so ppm a6 and del coeffs will work for the now non-uniformly spaced axis coords.
    //
    
    err = zls_CalcAxisDeltas (w.x, kAxisTypeCRadius, w.nt);  
    err = zls_CalcAxisGeometrySource (w.x, wStar.x, kAxisTypeCRadius, w.nt);  
    
#ifdef ACTIVE_FP_NOISE_CONTROL   
  }
#endif ACTIVE_FP_NOISE_CONTROL
  //
#ifdef GRAVITY
  //
  // compute the mean gravity accelleration in the initial zone cells and evolved cells
  // and average the acceleration throughout dt.  
  // dv = dt * (g0+g1)*0.5,  dE = dt*(v0 * g0 + v1 *g1)*0.5
  //
  
  Real a0[w.nt]; // average g0 = dPhi/dx for initial cell domain, normalised so a0 = .5 x dt *g0
  Real a1[w.nt]; // average g1 = dPhi/dx for final cell domain, normalised so a1 = .5 x dt *g1
  // dv = (a0+a1), dE = (v0 * a0 + v1 *a1) 
  // Conservative accellerations based on cell widths and potential at boundaries using PPM
  // interopolation.
  
  if ( f.xi != NULL) {
    err = zlsa_LagrangianCellAccelleration1D ( a0, a1, w.nt, wStar.x, w.x, f, dt);
  }
  
#endif GRAVITY
  //
  // only evolve cells where both boundaries can move, and more than remap
  //
  
  is  = kZLS_LimitEvolveLimits; //w.ns-kEvolveLimits;
  ie  = w.nt-kZLS_LimitEvolveLimits; //w.ne+kEvolveLimits;
  
  // now go through and update the densities, energy density and velocities
  //
  // in 1d, the volume is just dx.  wStar has the original dx, and w has the new dx
  //
  for ( i = is; i < ie; i++){
    
#ifdef GRAVITY
    dAv         = a0[i]+a1[i];
#endif GRAVITY
    
    
    dA          = w.x[  i].dA     ; // left CD area channge
    dAp1        = w.x[i+1].dA     ; // right CD area change
    
    PMid        = wStar.P[  i].a  ; // left CD pressure
    PMidp1      = wStar.P[i+1].a  ; // right CD pressure
    DeltaPMid   = (PMidp1 - PMid)*( dA + dAp1 )*0.5;
    
    vxPMid      = wStar.v[  i].a    * PMid    ; // left v* pressure - e flux
    vxPMidp1    = wStar.v[i+1].a  * PMidp1    ; // right v* pressure - e flux
    DeltavxPMid = (vxPMidp1*dAp1 - vxPMid*dA );
    
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    DeltavxPMid = zls_tolx( DeltavxPMid, kZLS_Small );
    DeltaPMid   = zls_tolx( DeltaPMid,   kZLS_Small );  
    
    if ((DeltavxPMid != 0.0) || (DeltaPMid != 0.0)) {
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL   
      
      V0          = wStar.x[i].dVol ; // initial volume
      V1          =     w.x[i].dVol ; // final volume
      d           = w.d[i].a ; // initial average density
      cellMass    =   d * V0 ;
      
      volumeRatio = ( V0==V1 )?1.0:V0/V1;
      d		       *= volumeRatio; // new density
      w.d[i].a    =  d;  // return values to vector.  mass per volume
      
#if (NTRACERS > 0)    
      Integer k;
      for (k = 0; k < kNTracers; k++){
        w.tr[k][i].a *= volumeRatio;  // trs are like density
      }
#endif

      dtdM        = dt/cellMass ;
      DeltavxPMid *= dtdM;
      DeltaPMid   *= dtdM;
      
      vx0         = w.vx[i].a; // initial average velx, keep for E Equation
#ifdef GRAVITY
      vx1         = vx0 - DeltaPMid + dAv ;  // momentum per mass
#else
      vx1         = vx0 - DeltaPMid ;  // momentum per mass
#endif
      w.vx[i].a   = vx1;
      
      E           =  w.E [i].a ; // initial average Energy Density
#ifdef GRAVITY
      w.E[i].a    =  E - DeltavxPMid + (a0[i]*vx0 + a1[i]*vx1);  // internal energy per volume
#else
      w.E[i].a    =  E - DeltavxPMid;  // internal energy per volume
#endif
      w.P[i].a    = zlsp_PFromE_W1D( w, i);
      
      
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    }
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL   
    
  }; // i, is, ie
  
  
  return(err);
}



int zlsa_LagrangianEvolvePhiW1D( W1DArr w,  Wstar1DArr wStar, Real g, Real dt){
  // when phi (azimuth) is second axis in 2D, g = r ;
  // when phi (azimuth) is second axis in 3D, g = r cos theta, theta = latitude, equatorial angle;
  // when phi (azimuth) is third axis in 3D, g = r sin theta, theta = co-latitude, or polar angle;
  // g is applied to volumes and boundary speeds
  
  int err = noErr;
	
  Integer i, is, ie ;
	
  Real PMid         ;
  Real PMidp1       ;
  Real DeltaPMid    ;
  Real vxPMid       ;
  Real vxPMidp1     ;
  Real DeltavxPMid  ;
	
  Real d, V0, V1;
  Real volumeRatio  ;
  Real cellMass     ;
  Real dtdM         ;
  
  //
  // move the cell boundaries according to the velocity of the Riemann CD speed
  // in WStar.vx
  //
  //========================================
  // Lagangian method!  Moving the cell boundaries with wStar takes place here.
  //========================================
  //
  // move the cell boundary at the CD velocity in the wStar, umid.
  
#ifdef ACTIVE_FP_NOISE_CONTROL   
  Integer recalculate = kNothingMoved;
#endif ACTIVE_FP_NOISE_CONTROL   
  
  Real invg = 1.0/g;
	
  // cell 0 boundary is left fixed, cell nt-1 right boundary is left fixed
  
  for ( i = 1; i < w.nt; i++){
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    Real dx  = wStar.v[i].a*dt*invg;
    dx = zls_tolx(dx, kZLS_Small);
    if ( dx != 0.0 ) {
      recalculate = kSomethingMoved;
      w.x[i].l += dx; 
    } else {
      wStar.v[i].a = 0.0;
    }
#else
    w.x[i].l += wStar.v[i].a*dt*invg; 
	  recalculate = kSomethingMoved;
#endif CONTROL_EVOLVE
#else 
    w.x[i].l += wStar.v[i].a*dt*invg; 
#endif ACTIVE_FP_NOISE_CONTROL
  }
  
#ifdef ACTIVE_FP_NOISE_CONTROL   
  if ( recalculate == kSomethingMoved ) {
#endif ACTIVE_FP_NOISE_CONTROL
    // make the right boundaries consistent with the lefts set above
    for ( i = 0; i < w.nt-1; i++){
      w.x[i].r = w.x[i+1].l;
    }
    
    //
    // w0 axis is modified, recompute dxs and the purely grid parabolic coefficients
    // so ppm a6 and del coeffs will work for the now non-uniformly spaced axis coords.
    //
    
    err = zls_CalcAxisDeltas (w.x, kAxisTypePhi, w.nt);  // also does cos and sin angles in non cartesian Coords
    
#ifdef ACTIVE_FP_NOISE_CONTROL   
  }
#endif ACTIVE_FP_NOISE_CONTROL
  //
  //
  // only evolve cells where both boundaries can move, and more than remap
  //
  is  = kZLS_LimitEvolveLimits; //w.ns-kEvolveLimits;
  ie  = w.nt-kZLS_LimitEvolveLimits; //w.ne+kEvolveLimits;
  // now go through and update the densities, energy density and velocities
  //
  // in 1d, the volume is just dx.  wStar has the original dx, and w has the new dx
  //
  for ( i = is; i < ie; i++){
    
    PMid        = wStar.P[i].a    ; // left CD pressure
    PMidp1      = wStar.P[i+1].a  ; // right CD pressure
    DeltaPMid   = (PMidp1 - PMid);
    
    vxPMid      = wStar.v[i].a    * PMid   ; // left v* pressure - e flux
    vxPMidp1    = wStar.v[i+1].a  * PMidp1 ; // right v* pressure - e flux
    DeltavxPMid = (vxPMidp1  - vxPMid );
    
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    DeltavxPMid = zls_tolx( DeltavxPMid, kZLS_Small );
    DeltaPMid   = zls_tolx( DeltaPMid,   kZLS_Small );  
    
    if ((DeltavxPMid != 0.0) || (DeltaPMid != 0.0)) {
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL   
      
      V0          = wStar.x[i].dVol *g ; // initial volume
      V1          = w.x[i].dVol*g  ; // final volume
      
      d           = w.d[i].a ; // initial average density
      cellMass    = d * V0;
      volumeRatio = ( V0==V1 )?1.0:V0/V1;
      d		       *= volumeRatio; // new density
      w.d[i].a    =  d;  // return values to vector.  mass per volume
      
#if (NTRACERS > 0)    
      Integer k;
      for (k = 0; k < kNTracers; k++){
        w.tr[k][i].a *= volumeRatio;  // trs are like density
      }
#endif

      dtdM        = dt/cellMass ;
      
      DeltavxPMid *= dtdM;
      DeltaPMid   *= dtdM;
      
      w.E[i].a   -= DeltavxPMid;  // internal energy per volume
      w.vx[i].a  -= DeltaPMid; // momentum per mass
      w.P[i].a    = zlsp_PFromE_W1D( w, i);
      
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    }
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL   
    
  }; // i, is, ie
  
  
  return(err);
}



int zlsa_LagrangianEvolveThetaW1D( W1DArr w,  Wstar1DArr wStar, Real g, Real dt){
	
  int err = noErr;
	
  Integer i, is, ie ;
	
  Real dA           ;
  Real dAp1         ;
  Real PMid         ;
  Real PMidp1       ;
  Real DeltaPMid    ;
  Real vxPMid       ;
  Real vxPMidp1     ;
  Real DeltavxPMid  ;
	
  Real d, V0, V1;
  Real volumeRatio  ;
  Real cellMass     ;
  Real dtdM         ;
  
  Real invg = 1.0/g;
	
  // cell 0 boundary is left fixed, cell nt-1 right boundary is left fixed
#ifdef ACTIVE_FP_NOISE_CONTROL   
  Integer recalculate = kNothingMoved;
#endif ACTIVE_FP_NOISE_CONTROL   
  
  for ( i = 1; i < w.nt; i++){
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    Real dx  = wStar.v[i].a*dt*invg;
    dx = zls_tolx(dx, kZLS_Small);
    if ( dx != 0.0 ) {
      recalculate = kSomethingMoved;
      w.x[i].l += dx; 
    } else {
      wStar.v[i].a = 0.0;
    }
#else
    w.x[i].l += wStar.v[i].a*dt*invg; 
	  recalculate = kSomethingMoved;
#endif CONTROL_EVOLVE
#else 
    w.x[i].l += wStar.v[i].a*dt*invg; 
#endif ACTIVE_FP_NOISE_CONTROL
  }
  
#ifdef ACTIVE_FP_NOISE_CONTROL   
  if ( recalculate == kSomethingMoved ) {
#endif ACTIVE_FP_NOISE_CONTROL
    // make the right boundaries consistent with the lefts set above
    for ( i = 0; i < w.nt-1; i++){
      w.x[i].r = w.x[i+1].l;
    }
    
    //
    // w0 axis is modified, recompute dxs and the purely grid parabolic coefficients
    // so ppm a6 and del coeffs will work for the now non-uniformly spaced axis coords.
    //
    
    err = zls_CalcAxisDeltas (w.x, kAxisTypeTheta, w.nt);  // also does cos and sin angles in Spherical Coords
    err = zls_CalcAxisGeometrySource (w.x, wStar.x, kAxisTypeTheta, w.nt);  
    
#ifdef ACTIVE_FP_NOISE_CONTROL   
  }
#endif ACTIVE_FP_NOISE_CONTROL
  //
  //
  // only evolve cells where both boundaries can move, and more than remap
  //
  is  = kZLS_LimitEvolveLimits; //w.ns-kEvolveLimits;
  ie  = w.nt-kZLS_LimitEvolveLimits; //w.ne+kEvolveLimits;
  // now go through and update the densities, energy density and velocities
  //
  // in 1d, the volume is just dx.  wStar has the original dx, and w has the new dx
  //
  for ( i = is; i < ie; i++){
    
    dA          = w.x[i].dA       ; // left CD area channge
    dAp1        = w.x[i+1].dA     ; // right CD area change
    
    PMid        = wStar.P[i].a    ; // left CD pressure
    PMidp1      = wStar.P[i+1].a  ; // right CD pressure
    DeltaPMid   = (PMidp1 - PMid)*( dA + dAp1 )*0.5;
    
    vxPMid      = wStar.v[i].a    * PMid   ; // left v* pressure - e flux
    vxPMidp1    = wStar.v[i+1].a  * PMidp1 ; // right v* pressure - e flux
    DeltavxPMid = (vxPMidp1*dAp1 - vxPMid*dA );
    
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    DeltavxPMid = zls_tolx( DeltavxPMid, kZLS_Small );
    DeltaPMid   = zls_tolx( DeltaPMid,   kZLS_Small );  
    
    if ((DeltavxPMid != 0.0) || (DeltaPMid != 0.0)) {
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL   
      V0          = wStar.x[i].dVol *g; // initial volume
      V1          = w.x[i].dVol     *g; // final volume
      
      d           = w.d[i].a ; // initial average density
      cellMass    = d * V0;
      volumeRatio = ( V0==V1 )?1.0:V0/V1;
      d		       *= volumeRatio; // new density
      
      w.d[i].a    =  d;  // return values to vector.  mass per volume
      
#if (NTRACERS > 0)    
      Integer k;
      for (k = 0; k < kNTracers; k++){
        w.tr[k][i].a *= volumeRatio;  // trs are like density
      }
#endif

      
      dtdM        = dt/cellMass ;
      
      DeltavxPMid *= dtdM;
      DeltaPMid   *= dtdM;
      
      w.E[i].a   -= DeltavxPMid;  // internal energy per volume
      w.vx[i].a  -= DeltaPMid; // momentum per mass
      w.P[i].a    = zlsp_PFromE_W1D( w, i);
      
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    }
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL   
    
  }
	
  return(err);
} // theta , axis 3 evolve



int zlsa_LagrangianEvolvePolarThetaW1D( W1DArr w,  Wstar1DArr wStar, Real g, Real dt){
	
  int err = noErr;
	
  Integer i, is, ie ;
	
  Real dA           ;
  Real dAp1         ;
  Real PMid         ;
  Real PMidp1       ;
  Real DeltaPMid    ;
  Real vxPMid       ;
  Real vxPMidp1     ;
  Real DeltavxPMid  ;
	
  Real d, V0, V1;
  Real volumeRatio  ;
  Real cellMass     ;
  Real dtdM         ;
  
  Real invg = 1.0/g;
	
  // cell 0 boundary is left fixed, cell nt-1 right boundary is left fixed
#ifdef ACTIVE_FP_NOISE_CONTROL   
  Integer recalculate = kNothingMoved;
#endif ACTIVE_FP_NOISE_CONTROL   
  
  for ( i = 1; i < w.nt; i++){
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    Real dx  = wStar.v[i].a*dt*invg;
    dx = zls_tolx(dx, kZLS_Small);
    if ( dx != 0.0 ) {
      recalculate = kSomethingMoved;
      w.x[i].l += dx; 
    } else {
      wStar.v[i].a = 0.0;
    }
#else
    w.x[i].l += wStar.v[i].a*dt*invg; 
	  recalculate = kSomethingMoved;
#endif CONTROL_EVOLVE
#else 
    w.x[i].l += wStar.v[i].a*dt*invg; 
#endif ACTIVE_FP_NOISE_CONTROL
  }
  
#ifdef ACTIVE_FP_NOISE_CONTROL   
  if ( recalculate == kSomethingMoved ) {
#endif ACTIVE_FP_NOISE_CONTROL
    // make the right boundaries consistent with the lefts set above
    for ( i = 0; i < w.nt-1; i++){
      w.x[i].r = w.x[i+1].l;
    }
    
    //
    // w0 axis is modified, recompute dxs and the purely grid parabolic coefficients
    // so ppm a6 and del coeffs will work for the now non-uniformly spaced axis coords.
    //
    
    err = zls_CalcAxisDeltas (w.x, kAxisTypePolarTheta, w.nt);  // also does cos and sin angles in Spherical Coords
    err = zls_CalcAxisGeometrySource (w.x, wStar.x, kAxisTypePolarTheta, w.nt);  
    
#ifdef ACTIVE_FP_NOISE_CONTROL   
  }
#endif ACTIVE_FP_NOISE_CONTROL
  //
  //
  // only evolve cells where both boundaries can move, and more than remap
  //
  is  = kZLS_LimitEvolveLimits; //w.ns-kEvolveLimits;
  ie  = w.nt-kZLS_LimitEvolveLimits; //w.ne+kEvolveLimits;
  // now go through and update the densities, energy density and velocities
  //
  // in 1d, the volume is just dx.  wStar has the original dx, and w has the new dx
  //
  for ( i = is; i < ie; i++){
    
    dA          = w.x[i].dA       ; // left CD area channge
    dAp1        = w.x[i+1].dA     ; // right CD area change
    
    PMid        = wStar.P[i].a    ; // left CD pressure
    PMidp1      = wStar.P[i+1].a  ; // right CD pressure
    DeltaPMid   = (PMidp1 - PMid)*( dA + dAp1 )*0.5;
    
    vxPMid      = wStar.v[i].a    * PMid   ; // left v* pressure - e flux
    vxPMidp1    = wStar.v[i+1].a  * PMidp1 ; // right v* pressure - e flux
    DeltavxPMid = (vxPMidp1*dAp1 - vxPMid*dA );
    
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    DeltavxPMid = zls_tolx( DeltavxPMid, kZLS_Small );
    DeltaPMid   = zls_tolx( DeltaPMid,   kZLS_Small );  
    
    if ((DeltavxPMid != 0.0) || (DeltaPMid != 0.0)) {
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL   
      V0          = wStar.x[i].dVol *g; // initial volume
      V1          = w.x[i].dVol     *g; // final volume
      
      d           = w.d[i].a ; // initial average density
      cellMass    = d * V0;
      volumeRatio = ( V0==V1 )?1.0:V0/V1;
      d		       *= volumeRatio; // new density
      
      w.d[i].a    =  d;  // return values to vector.  mass per volume
      
#if (NTRACERS > 0)    
      Integer k;
      for (k = 0; k < kNTracers; k++){
        w.tr[k][i].a *= volumeRatio;  // trs are like density
      }
#endif

      
      dtdM        = dt/cellMass ;
      
      DeltavxPMid *= dtdM;
      DeltaPMid   *= dtdM;
      
      w.E[i].a   -= DeltavxPMid;  // internal energy per volume
      w.vx[i].a  -= DeltaPMid; // momentum per mass
      w.P[i].a    = zlsp_PFromE_W1D( w, i);
      
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_EVOLVE
    }
#endif CONTROL_EVOLVE
#endif ACTIVE_FP_NOISE_CONTROL   
    
  }
	
  return(err);
} // polar theta , axis 2 evolve
#endif !CARTESIAN


int zlsa_LagrangianRemapW1D( W1DArr w,  Flx1DArr flux, ACoordPtr x0){
	
  int err = noErr;
  Integer i, k, is, ie;
	
  Real V0; // current volume
  Real V1; // remapped volume
  Real invV1; // 1/V1
  
  Real d,  E2;
  Real dV, deltaX;
#ifdef PLM
  Real f, cf; // cell fractions, cf = co-fraction = 1-f
#endif PLM
#ifdef PPM
  Real f, f2;
#endif PPM
  
#ifdef SYM1D
  Real  vx0;
  Real vx;
#endif SYM1D
#ifdef SYM2D
  Real vx0, vy0;
  Real vx, vy;
#endif SYM2D
#ifdef SYM3D
  Real  vx0, vy0, vz0;
  Real  vx, vy, vz;
#endif SYM3D
  
  Real cellMass0;
  Real cellMass1;
  Real invCellMass1;
	
#ifndef GODUNOV
  err = zlsppm_CalcRemappingCoeffsW1DArr   ( w ); // all vars, no flattening
  err = zlsp_CalcCellLREnergyW1D  ( w );
#endif !GODUNOV
  
  err = zlsh_ClearWFluxes1DArr	 ( flux );
	
  if ( err == noErr ) {
		
    is = kZLS_LimitRemapLimits; //w.ns-kRemapLimits;
    ie = w.nt-kZLS_LimitRemapLimits; //w.ne+kRemapLimits;
    
    for(i = is; i < ie+1; i++){
			
      deltaX = w.x[i].l - x0[i].l;
      dV     = deltaX;
			
#ifdef PLM
      f  = 0.0;
#endif PLM
#ifdef PPM
      f  = 0.0;
      f2 = 1.0;
#endif PPM
      
      if ( deltaX > kZLS_Small ) {
        //
        // cell left boundry moved to the right
        // element of w that needs to be integrated is the right side
        // of cell i - 1;
        //
        k  = i - 1;
        
#ifdef GODUNOV
        flux.f[i].m  =    dV *w.d[k].a;
        flux.f[i].E  =  flux.f[i].m *w.E[k].a;
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *w.vx[k].a;
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *w.vx[k].a;
        flux.f[i].my = flux.f[i].m *w.vy[k].a;
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *w.vx[k].a;
        flux.f[i].my = flux.f[i].m *w.vy[k].a;
        flux.f[i].mz = flux.f[i].m *w.vz[k].a;
#endif SYM3D
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*w.tr[l][k].a;
        }
#endif

#endif GODUNOV
        
        
#ifdef PLM        
        f  = 0.5*deltaX*(w.x[k].idx);
        cf = 1.0-f;
				
        flux.f[i].m  =    dV *( cf*w.d[k].ar + f* w.d[k].al);
        flux.f[i].E  =  flux.f[i].m *(cf*w.E[k].ar + f*w.E[k].al);
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[k].ar + f*w.vx[k].al);
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[k].ar + f*w.vx[k].al);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[k].ar + f*w.vy[k].al);
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[k].ar + f*w.vx[k].al);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[k].ar + f*w.vy[k].al);
        flux.f[i].mz = flux.f[i].m *(cf*w.vz[k].ar + f*w.vz[k].al);
#endif SYM3D
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(cf*w.tr[l][k].ar + f*w.tr[l][k].al);
        }
#endif
        
#endif PLM
        
#ifdef PPM        
        
        f  = 0.5*deltaX*(w.x[k].idx); // deltaX is +ve, CW eqn 1.12 case 1
        f2 = 1.0 - fourthd*f;
				
        flux.f[i].m  =    dV *( w.d[k].ar - f*( w.d[k].del -  f2*w.d[k].a6));
        flux.f[i].E  =  flux.f[i].m *(w.E[k].ar - f*(w.E[k].del - f2*w.E[k].a6));
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(w.vx[k].ar - f*(w.vx[k].del - f2*w.vx[k].a6));
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(w.vx[k].ar - f*(w.vx[k].del - f2*w.vx[k].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[k].ar - f*(w.vy[k].del - f2*w.vy[k].a6));
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(w.vx[k].ar - f*(w.vx[k].del - f2*w.vx[k].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[k].ar - f*(w.vy[k].del - f2*w.vy[k].a6));
        flux.f[i].mz = flux.f[i].m *(w.vz[k].ar - f*(w.vz[k].del - f2*w.vz[k].a6));
#endif SYM3D
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(w.tr[l][k].ar - f*(w.tr[l][k].del - f2*w.tr[l][k].a6));
        }
#endif
        
#endif PPM
				
      } else if ( deltaX < -kZLS_Small ) {
        //
        // cell left boundry moved to the left
        // element of w that needs to be integrated is the left side
        // of cell i;
        //
				
#ifdef GODUNOV
				
        flux.f[i].m  =    dV *w.d[i].a; // mass transferred
        flux.f[i].E  =  flux.f[i].m *w.E[i].a; // total energy
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *w.vx[i].a; // momentum transferred
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *w.vx[i].a;
        flux.f[i].my = flux.f[i].m *w.vy[i].a;
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *w.vx[i].a;
        flux.f[i].my = flux.f[i].m *w.vy[i].a;
        flux.f[i].mz = flux.f[i].m *w.vz[i].a;
#endif SYM2D
				
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*w.tr[l][i].a;
        }
#endif

#endif GODUNOV
				
#ifdef PLM
        f  = -0.5*deltaX*(w.x[i].idx);
				cf = 1.0 - f;
        
        flux.f[i].m  =    dV *( cf*w.d[i].al + f* w.d[i].ar); // mass transferred
        flux.f[i].E  =  flux.f[i].m *(cf*w.E[i].al + f*w.E[i].ar); // total energy
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[i].al + f*w.vx[i].ar); // momentum transferred
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[i].al + f*w.vx[i].ar);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[i].al + f*w.vy[i].ar);
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[i].al + f*w.vx[i].ar);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[i].al + f*w.vy[i].ar);
        flux.f[i].mz = flux.f[i].m *(cf*w.vz[i].al + f*w.vz[i].ar);
#endif SYM2D
				
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(cf*w.tr[l][i].al + f*w.tr[l][i].ar);
        }
#endif

#endif PLM
				
#ifdef PPM
        f  = -0.5*deltaX*(w.x[i].idx); // eqn 1.12 CW, deltaX is -ve here, f is +ve
        f2 = 1.0 - fourthd*f;         // CW eqn 1.12 case 2
				
        flux.f[i].m  =    dV *( w.d[i].al + f*( w.d[i].del +  f2*w.d[i].a6)); // mass transferred
        flux.f[i].E  =  flux.f[i].m *(w.E[i].al + f*(w.E[i].del + f2*w.E[i].a6)); // total energy
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(w.vx[i].al + f*(w.vx[i].del + f2*w.vx[i].a6)); // momentum transferred
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(w.vx[i].al + f*(w.vx[i].del + f2*w.vx[i].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[i].al + f*(w.vy[i].del + f2*w.vy[i].a6));
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(w.vx[i].al + f*(w.vx[i].del + f2*w.vx[i].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[i].al + f*(w.vy[i].del + f2*w.vy[i].a6));
        flux.f[i].mz = flux.f[i].m *(w.vz[i].al + f*(w.vz[i].del + f2*w.vz[i].a6));
#endif SYM2D
				
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(w.tr[l][i].al + f*(w.tr[l][i].del + f2*w.tr[l][i].a6));
        }
#endif

#endif PPM
				
      };
    }; //i
    
		
    is = kZLS_LimitRemapLimits; //w.ns-kRemapLimits;
    ie = w.nt-kZLS_LimitRemapLimits; //w.ne+kRemapLimits;
    
    // update cells with remap 'fluxes'  and restore coordinates
		
    for(i = is; i < ie; i++){ 
      
      if ((flux.f[i].m != 0.0) || (flux.f[i+1].m != 0.0)){
        
        V0        = w.x[i].dVol  ; // current volume
        V1        =  x0[i].dVol  ; // remapped volume
        invV1     = 1.0/V1; // precompute in CalcDeltas???  LATER rss 2006
        
        cellMass0 = ( w.d[i].a*V0 );
        cellMass1 = ( cellMass0 + ( flux.f[i].m - flux.f[i+1].m ) );
        invCellMass1 = 1.0/cellMass1;
        
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          w.tr[l][i].a  = (  w.tr[l][i].a*V0 + ( flux.f[i].mtr[l] - flux.f[i+1].mtr[l] ))*invV1;
        }
#endif
        
        d =   cellMass1*invV1;
        
#ifdef SYM1D
        vx0 = w.vx[i].a;
        vx = ( vx0*cellMass0 + (flux.f[i].mx-flux.f[i+1].mx) )*invCellMass1;
        
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_REMAP
        vx = zls_tolx( (vx-vx0), kZLS_Small)+vx0;
#endif CONTROL_REMAP
#endif ACTIVE_FP_NOISE_CONTROL

        E2 = ( w.E[i].a*cellMass0 +  (flux.f[i].E - flux.f[i+1].E) )*invCellMass1; 

        w.vx[i].a = vx;
        
#endif SYM1D
        
#ifdef SYM2D
        vx0 = w.vx[i].a;
        vy0 = w.vy[i].a;
        
        vx = ( vx0*cellMass0 + (flux.f[i].mx-flux.f[i+1].mx) )*invCellMass1;
        vy = ( vy0*cellMass0 + (flux.f[i].my-flux.f[i+1].my) )*invCellMass1;
        
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_REMAP
        vx = zls_tolx( (vx-vx0), kZLS_Small)+vx0;
        vy = zls_tolx( (vy-vy0), kZLS_Small)+vy0;
#endif CONTROL_REMAP
#endif ACTIVE_FP_NOISE_CONTROL

        E2 = ( w.E[i].a*cellMass0 +  (flux.f[i].E - flux.f[i+1].E) )*invCellMass1; 

        w.vx[i].a = vx;
        w.vy[i].a = vy;
        
#endif SYM2D
        
#ifdef SYM3D
        
        vx0 = w.vx[i].a;
        vy0 = w.vy[i].a;
        vz0 = w.vz[i].a;
        
        vx = ( vx0*cellMass0 + (flux.f[i].mx-flux.f[i+1].mx) )*invCellMass1;
        vy = ( vy0*cellMass0 + (flux.f[i].my-flux.f[i+1].my) )*invCellMass1;
        vz = ( vz0*cellMass0 + (flux.f[i].mz-flux.f[i+1].mz) )*invCellMass1;
        
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_REMAP
        vx = zls_tolx( (vx-vx0), kZLS_Small)+vx0;
        vy = zls_tolx( (vy-vy0), kZLS_Small)+vy0;
        vz = zls_tolx( (vz-vz0), kZLS_Small)+vz0;
#endif CONTROL_REMAP
#endif ACTIVE_FP_NOISE_CONTROL

        E2 = ( w.E[i].a*cellMass0 +  (flux.f[i].E - flux.f[i+1].E) )*invCellMass1;

        w.vx[i].a = vx;
        w.vy[i].a = vy;
        w.vz[i].a = vz;
        
#endif SYM3D
        
        w.E[i].a  = E2;
        w.d[i].a  = d;

        // Derive P after conservative remapping that gives total E
        
        w.P[i].a  = zlsp_PFromE_W1D(w, i);
        
      }
      
    } //i
    
    memcpy(w.x, x0, w.nt*sizeof(AxisCoords));
    
  } else { err = memErr;}
  
  return(err);
}

//
#ifndef CARTESIAN
// each axis is different and a slightly different remap type is needed for each axis
// 3D sphericaly divergent axis, r: r - phi - theta or r - pTheta - phi are RH
int zlsa_LagrangianRemapSRadialW1D( W1DArr w,  Flx1DArr flux,  ACoordPtr x0){
	
  int err = noErr;
  Integer i, k, is, ie;
	
  Real V0; // current volume
  Real V1; // remapped volume
  Real invV1; // 1/V1
  
  Real d,  P2, E2;
  Real dV, dr, r;
#ifdef PLM
  Real f, cf;
#endif PLM
#ifdef PPM
  Real f, f2;
#endif PPM
  
  Real gamm  = kZLS_eos_gamma_1;
  
#ifdef VARIABLE_GAMMA
  PhysicsPtr phys = gLevels[0].levelPhysics; 
  CoolingPtr cp   = &(phys->coolParameters);
#endif VARIABLE_GAMMA
  
#ifdef SYM1D
  Real  vx0;
  Real vx;
#endif SYM1D
#ifdef SYM2D
  Real vx0, vy0;
  Real vx, vy;
#endif SYM2D
#ifdef SYM3D
  Real  vx0, vy0, vz0;
  Real  vx, vy, vz;
#endif SYM3D
  
  Real cellMass0;
  Real cellMass1;
  Real invCellMass1;
	
#ifndef GODUNOV
  err = zlsppm_CalcRemappingCoeffsW1DArr   ( w ); // all vars, no flattening
  err = zlsp_CalcCellLREnergyW1D  ( w );
#endif !GODUNOV
  
  err = zlsh_ClearWFluxes1DArr	 ( flux );
  
  if ( err == noErr ) {
		
    is = kZLS_LimitRemapLimits; //w.ns-kRemapLimits;
    ie = w.nt-kZLS_LimitRemapLimits; //w.ne+kRemapLimits;
		
    for(i = is; i < ie+1; i++){ //
			
      r  = x0[i].l;
      dr = w.x[i].l - x0[i].l; // swept volume, not cell volume, so need to recompute
      dV = dr*(r*(r + dr) + dr*dr*third);  // change in volume over radial difference
			
#ifdef PLM
      f  = 0.0;
      cf = 1.0;
#endif PLM
#ifdef PPM
      f  = 0.0;
      f2 = 1.0;
#endif PPM
      
      
      if ( dr > kZLS_Small ) {
        //
        // cell left boundry moved to the right
        // element of w that needs to be integrated is the right side
        // of cell i - 1;
        //
        k  = i - 1;
        
#ifdef GODUNOV
        flux.f[i].m  =    dV *w.d[k].a;
        flux.f[i].E  =  flux.f[i].m *w.E[k].a;
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *w.vx[k].a;
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *w.vx[k].a;
        flux.f[i].my = flux.f[i].m *w.vy[k].a;
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *w.vx[k].a;
        flux.f[i].my = flux.f[i].m *w.vy[k].a;
        flux.f[i].mz = flux.f[i].m *w.vz[k].a;
#endif SYM3D
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*w.tr[l][k].a;
        }
#endif

#endif GODUNOV
        
#ifdef PLM        
        f  = 0.5*dr*(w.x[k].idx);
        cf = 1.0-f;
				
        flux.f[i].m  =    dV *( cf*w.d[k].ar + f*w.d[k].al);
        flux.f[i].E  =  flux.f[i].m *(cf*w.E[k].ar + f*w.E[k].al);
        
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[k].ar + f*w.vx[k].al);
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[k].ar + f*w.vx[k].al);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[k].ar + f*w.vy[k].al);
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[k].ar + f*w.vx[k].al);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[k].ar + f*w.vy[k].al);
        flux.f[i].mz = flux.f[i].m *(cf*w.vz[k].ar + f*w.vz[k].al);
#endif SYM3D
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(cf*w.tr[l][k].ar + f*w.tr[l][k].al);
        }
#endif
        
#endif PLM
        
#ifdef PPM
        
        f  = 0.5*dr*(w.x[k].idx);
        f2 = 1.0 - fourthd*f;
				
        flux.f[i].m  =    dV *( w.d[k].ar - f*( w.d[k].del -  f2*w.d[k].a6));
        flux.f[i].E  = flux.f[i].m *(w.E[k].ar - f*(w.E[k].del - f2*w.E[k].a6));
        
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(w.vx[k].ar - f*(w.vx[k].del - f2*w.vx[k].a6));
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(w.vx[k].ar - f*(w.vx[k].del - f2*w.vx[k].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[k].ar - f*(w.vy[k].del - f2*w.vy[k].a6));
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(w.vx[k].ar - f*(w.vx[k].del - f2*w.vx[k].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[k].ar - f*(w.vy[k].del - f2*w.vy[k].a6));
        flux.f[i].mz = flux.f[i].m *(w.vz[k].ar - f*(w.vz[k].del - f2*w.vz[k].a6));
#endif SYM3D
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(w.tr[l][k].ar - f*(w.tr[l][k].del - f2*w.tr[l][k].a6));
        }
#endif

#endif PPM
				
      } else if ( dr < -kZLS_Small ) {
        //
        // cell left boundry moved to the left
        // element of w that needs to be interated is the left side
        // of cell i;
        //
        
#ifdef GODUNOV
        flux.f[i].m  =    dV *w.d[i].a; // mass transferred
        flux.f[i].E  =  flux.f[i].m *w.E[i].a; // total energy
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *w.vx[i].a; // momentum transferred
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *w.vx[i].a;
        flux.f[i].my = flux.f[i].m *w.vy[i].a;
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *w.vx[i].a;
        flux.f[i].my = flux.f[i].m *w.vy[i].a;
        flux.f[i].mz = flux.f[i].m *w.vz[i].a;
#endif SYM2D
				
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*w.tr[l][i].a;
        }
#endif

#endif GODUNOV
				
#ifdef PLM
        f  = -0.5*dr*(w.x[i].idx);
				cf = 1.0 - f;
        
        flux.f[i].m  =    dV *( cf*w.d[i].al + f* w.d[i].ar); // mass transferred
        flux.f[i].E  =  flux.f[i].m *(cf*w.E[i].al + f*w.E[i].ar); // total energy
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[i].al + f*w.vx[i].ar); // momentum transferred
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[i].al + f*w.vx[i].ar);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[i].al + f*w.vy[i].ar);
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[i].al + f*w.vx[i].ar);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[i].al + f*w.vy[i].ar);
        flux.f[i].mz = flux.f[i].m *(cf*w.vz[i].al + f*w.vz[i].ar);
#endif SYM2D
				
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(cf*w.tr[l][i].al + f*w.tr[l][i].ar);
        }
#endif
        
#endif PLM
				
        
#ifdef PPM
        
        f  = -0.5*dr*(w.x[i].idx);
        f2 =  1.0 - fourthd*f;
				
        flux.f[i].m  =    dV *( w.d[i].al + f*( w.d[i].del +  f2*w.d[i].a6)); // mass transferred
        flux.f[i].E  =  flux.f[i].m *(w.E[i].al + f*(w.E[i].del + f2*w.E[i].a6)); // total energy
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(w.vx[i].al + f*(w.vx[i].del + f2*w.vx[i].a6)); // momentum transferred
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(w.vx[i].al + f*(w.vx[i].del + f2*w.vx[i].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[i].al + f*(w.vy[i].del + f2*w.vy[i].a6));
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(w.vx[i].al + f*(w.vx[i].del + f2*w.vx[i].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[i].al + f*(w.vy[i].del + f2*w.vy[i].a6));
        flux.f[i].mz = flux.f[i].m *(w.vz[i].al + f*(w.vz[i].del + f2*w.vz[i].a6));
#endif SYM2D
				
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(w.tr[l][i].al + f*(w.tr[l][i].del + f2*w.tr[l][i].a6));
        }
#endif
        
#endif PPM
				
      }
      
    } //i
		
		
    is = kZLS_LimitRemapLimits; //w.ns-kRemapLimits;
    ie = w.nt-kZLS_LimitRemapLimits; //w.ne+kRemapLimits;
		
    
    for(i = is; i < ie; i++){ //
      
      if ((flux.f[i].m != 0.0) || (flux.f[i+1].m != 0.0)){
        
        V0        = w.x[i].dVol; // current volume
        V1        =  x0[i].dVol;  // remapped volume
      	invV1     = 1.0/V1;
        
		    cellMass0 = ( w.d[i].a*V0 );
        cellMass1 = ( cellMass0 + ( flux.f[i].m - flux.f[i+1].m ) );
        invCellMass1 = 1.0/cellMass1;
        
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          w.tr[l][i].a  = (  w.tr[l][i].a*V0 + ( flux.f[i].mtr[l] - flux.f[i+1].mtr[l] ))*invV1;
        }
#endif
        
        d =   cellMass1*invV1;
        
#ifdef SYM1D
        vx0 = w.vx[i].a;
        vx = ( vx0*cellMass0 + (flux.f[i].mx-flux.f[i+1].mx) )*invCellMass1;
        
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_REMAP
        vx = zls_tolx( (vx-vx0), kZLS_Small)+vx0;
#endif CONTROL_REMAP
#endif ACTIVE_FP_NOISE_CONTROL

        E2 = ( w.E[i].a*cellMass0 +  (flux.f[i].E - flux.f[i+1].E) )*invCellMass1; 

        w.vx[i].a = vx;
        
#endif SYM1D
        
#ifdef SYM2D
        vx0 = w.vx[i].a;
        vy0 = w.vy[i].a;
        vx = ( vx0*cellMass0 + (flux.f[i].mx-flux.f[i+1].mx) )*invCellMass1;
        vy = ( vy0*cellMass0 + (flux.f[i].my-flux.f[i+1].my) )*invCellMass1;
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_REMAP
        vx = zls_tolx( (vx-vx0), kZLS_Small)+vx0;
        vy = zls_tolx( (vy-vy0), kZLS_Small)+vy0;
#endif CONTROL_REMAP
#endif ACTIVE_FP_NOISE_CONTROL

        E2 = ( w.E[i].a*cellMass0 +  (flux.f[i].E - flux.f[i+1].E) )*invCellMass1; 

        w.vx[i].a = vx;
        w.vy[i].a = vy;
#endif SYM2D
        
#ifdef SYM3D
        vx0 = w.vx[i].a;
        vy0 = w.vy[i].a;
        vz0 = w.vz[i].a;
        vx = ( vx0*cellMass0 + (flux.f[i].mx-flux.f[i+1].mx) )*invCellMass1;
        vy = ( vy0*cellMass0 + (flux.f[i].my-flux.f[i+1].my) )*invCellMass1;
        vz = ( vz0*cellMass0 + (flux.f[i].mz-flux.f[i+1].mz) )*invCellMass1;
        
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_REMAP
        vx = zls_tolx( (vx-vx0), kZLS_Small)+vx0;
        vy = zls_tolx( (vy-vy0), kZLS_Small)+vy0;
        vz = zls_tolx( (vz-vz0), kZLS_Small)+vz0;
#endif CONTROL_REMAP
#endif ACTIVE_FP_NOISE_CONTROL

        E2 = ( w.E[i].a*cellMass0 +  (flux.f[i].E - flux.f[i+1].E) )*invCellMass1; 

        w.vx[i].a = vx;
        w.vy[i].a = vy;
        w.vz[i].a = vz;
#endif SYM3D
        

        w.E[i].a  = E2;
        w.d[i].a  = d;
        
        w.P[i].a  = zlsp_PFromE_W1D(w, i);

      } // non-zero fluxes
      
			
    } //i
    
    memcpy(w.x, x0, w.nt*sizeof(AxisCoords));
    
  } else { err = memErr;}
	
  return(err);
} // radial Remap

// circular divergent axis, r, cylinder or 2D circular.  z - r - phi is RH.
int zlsa_LagrangianRemapCRadialW1D( W1DArr w,  Flx1DArr flux,  ACoordPtr x0){
	
  int err = noErr;
  Integer i, k, is, ie;
	
  Real V0; // current volume
  Real V1; // remapped volume
  Real invV1; // 1/V1
  
  Real d,  P2, E2;
  Real dV, dr, r;
#ifdef PLM
  Real f, cf;
#endif PLM
#ifdef PPM
  Real f, f2;
#endif PPM
    
#ifdef SYM1D
  Real  vx0;
  Real vx;
#endif SYM1D
#ifdef SYM2D
  Real vx0, vy0;
  Real vx, vy;
#endif SYM2D
#ifdef SYM3D
  Real  vx0, vy0, vz0;
  Real  vx, vy, vz;
#endif SYM3D
  
  Real cellMass0;
  Real cellMass1;
  Real invCellMass1;
	
#ifndef GODUNOV
  err = zlsppm_CalcRemappingCoeffsW1DArr   ( w ); // all vars, no flattening
  err = zlsp_CalcCellLREnergyW1D  ( w ); // derive E_l and E_r from P_l and P_r
#endif !GODUNOV
  
  err = zlsh_ClearWFluxes1DArr	 ( flux );
  
  if ( err == noErr ) {
		
    is = kZLS_LimitRemapLimits; //w.ns-kRemapLimits;
    ie = w.nt-kZLS_LimitRemapLimits; //w.ne+kRemapLimits;
		
    for(i = is; i < ie+1; i++){ //
      
      r  = x0[i].l;
      dr = w.x[i].l - x0[i].l; // swept volume, not cell volume, so need to recompute
      dV = dr*(r+(dr*0.5));  // change in volume over circular radial difference, ie 2D r-theta circular coordinates, or in z-r cylindrical 
			
#ifdef PLM
      f  = 0.0;
      cf = 1.0;
#endif PLM
#ifdef PPM
      f  = 0.0;
      f2 = 1.0;
#endif PPM
      
      
      if ( dr > kZLS_Small ) {
        //
        // cell left boundry moved to the right
        // element of w that needs to be integrated is the right side
        // of cell i - 1;
        //
        k  = i - 1;
        
#ifdef GODUNOV
        flux.f[i].m  =    dV *w.d[k].a;
        flux.f[i].E  =  flux.f[i].m *w.E[k].a;
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *w.vx[k].a;
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *w.vx[k].a;
        flux.f[i].my = flux.f[i].m *w.vy[k].a;
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *w.vx[k].a;
        flux.f[i].my = flux.f[i].m *w.vy[k].a;
        flux.f[i].mz = flux.f[i].m *w.vz[k].a;
#endif SYM3D
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*w.tr[l][k].a;
        }
#endif

#endif GODUNOV
        
#ifdef PLM        
        f  = 0.5*dr*(w.x[k].idx);
        cf = 1.0-f;
				
        flux.f[i].m  =    dV *( cf*w.d[k].ar + f*w.d[k].al);
        flux.f[i].E  =  flux.f[i].m *(cf*w.E[k].ar + f*w.E[k].al);
        
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[k].ar + f*w.vx[k].al);
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[k].ar + f*w.vx[k].al);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[k].ar + f*w.vy[k].al);
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[k].ar + f*w.vx[k].al);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[k].ar + f*w.vy[k].al);
        flux.f[i].mz = flux.f[i].m *(cf*w.vz[k].ar + f*w.vz[k].al);
#endif SYM3D
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(cf*w.tr[l][k].ar + f*w.tr[l][k].al);
        }
#endif
        
#endif PLM
        
#ifdef PPM
        
        f  = 0.5*dr*(w.x[k].idx);
        f2 = 1.0 - fourthd*f;
				
        flux.f[i].m  =    dV *( w.d[k].ar - f*( w.d[k].del -  f2*w.d[k].a6));
        flux.f[i].E  = flux.f[i].m *(w.E[k].ar - f*(w.E[k].del - f2*w.E[k].a6));
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(w.vx[k].ar - f*(w.vx[k].del - f2*w.vx[k].a6));
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(w.vx[k].ar - f*(w.vx[k].del - f2*w.vx[k].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[k].ar - f*(w.vy[k].del - f2*w.vy[k].a6));
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(w.vx[k].ar - f*(w.vx[k].del - f2*w.vx[k].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[k].ar - f*(w.vy[k].del - f2*w.vy[k].a6));
        flux.f[i].mz = flux.f[i].m *(w.vz[k].ar - f*(w.vz[k].del - f2*w.vz[k].a6));
#endif SYM3D
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(w.tr[l][k].ar - f*(w.tr[l][k].del - f2*w.tr[l][k].a6));
        }
#endif

#endif PPM
				
      } else if ( dr < -kZLS_Small ) {
        //
        // cell left boundry moved to the left
        // element of w that needs to be interated is the left side
        // of cell i;
        //
        
#ifdef GODUNOV
        flux.f[i].m  =    dV *w.d[i].a; // mass transferred
        flux.f[i].E  =  flux.f[i].m *w.E[i].a; // total energy
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *w.vx[i].a; // momentum transferred
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *w.vx[i].a;
        flux.f[i].my = flux.f[i].m *w.vy[i].a;
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *w.vx[i].a;
        flux.f[i].my = flux.f[i].m *w.vy[i].a;
        flux.f[i].mz = flux.f[i].m *w.vz[i].a;
#endif SYM2D
				
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*w.tr[l][i].a;
        }
#endif

#endif GODUNOV
				
#ifdef PLM
        f  = -0.5*dr*(w.x[i].idx);
				cf = 1.0 - f;
        
        flux.f[i].m  =    dV *( cf*w.d[i].al + f* w.d[i].ar); // mass transferred
        flux.f[i].E  =  flux.f[i].m *(cf*w.E[i].al + f*w.E[i].ar); // total energy
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[i].al + f*w.vx[i].ar); // momentum transferred
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[i].al + f*w.vx[i].ar);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[i].al + f*w.vy[i].ar);
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[i].al + f*w.vx[i].ar);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[i].al + f*w.vy[i].ar);
        flux.f[i].mz = flux.f[i].m *(cf*w.vz[i].al + f*w.vz[i].ar);
#endif SYM2D
				
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(cf*w.tr[l][i].al + f*w.tr[l][i].ar);
        }
#endif

#endif PLM
				
        
#ifdef PPM
        
        f  = -0.5*dr*(w.x[i].idx);
        f2 =  1.0 - fourthd*f;
				
        flux.f[i].m  =    dV *( w.d[i].al + f*( w.d[i].del +  f2*w.d[i].a6)); // mass transferred
        flux.f[i].E  =  flux.f[i].m *(w.E[i].al + f*(w.E[i].del + f2*w.E[i].a6)); // total energy
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(w.vx[i].al + f*(w.vx[i].del + f2*w.vx[i].a6)); // momentum transferred
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(w.vx[i].al + f*(w.vx[i].del + f2*w.vx[i].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[i].al + f*(w.vy[i].del + f2*w.vy[i].a6));
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(w.vx[i].al + f*(w.vx[i].del + f2*w.vx[i].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[i].al + f*(w.vy[i].del + f2*w.vy[i].a6));
        flux.f[i].mz = flux.f[i].m *(w.vz[i].al + f*(w.vz[i].del + f2*w.vz[i].a6));
#endif SYM2D
				
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(w.tr[l][i].al + f*(w.tr[l][i].del + f2*w.tr[l][i].a6));
        }
#endif
#endif PPM
				
      }
      
    } //i
		
		
    is = kZLS_LimitRemapLimits; //w.ns-kRemapLimits;
    ie = w.nt-kZLS_LimitRemapLimits; //w.ne+kRemapLimits;
		
    
    for(i = is; i < ie; i++){ //
      
      if ((flux.f[i].m != 0.0) || (flux.f[i+1].m != 0.0)){
        
        V0        = w.x[i].dVol; // current volume
        V1        =  x0[i].dVol;  // remapped volume
        invV1     = 1.0/V1;
        
        cellMass0 = ( w.d[i].a*V0 );
        cellMass1 = ( cellMass0 + ( flux.f[i].m - flux.f[i+1].m ) );
        invCellMass1 = 1.0/cellMass1;
        
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          w.tr[l][i].a  = (  w.tr[l][i].a*V0 + ( flux.f[i].mtr[l] - flux.f[i+1].mtr[l] ))*invV1;
        }
#endif
        
        d =   cellMass1*invV1;
        
#ifdef SYM1D
        vx0 = w.vx[i].a;
        vx = ( vx0*cellMass0 + (flux.f[i].mx-flux.f[i+1].mx) )*invCellMass1;
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_REMAP
        vx = zls_tolx( (vx-vx0), kZLS_Small)+vx0;
#endif CONTROL_REMAP
#endif ACTIVE_FP_NOISE_CONTROL

        E2 = ( w.E[i].a*cellMass0 +  (flux.f[i].E - flux.f[i+1].E) )*invCellMass1; // 

        w.vx[i].a = vx;
#endif SYM1D
        
#ifdef SYM2D
        vx0 = w.vx[i].a;
        vy0 = w.vy[i].a;
        vx = ( vx0*cellMass0 + (flux.f[i].mx-flux.f[i+1].mx) )*invCellMass1;
        vy = ( vy0*cellMass0 + (flux.f[i].my-flux.f[i+1].my) )*invCellMass1;
        
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_REMAP
        vx = zls_tolx( (vx-vx0), kZLS_Small)+vx0;
        vy = zls_tolx( (vy-vy0), kZLS_Small)+vy0;
#endif CONTROL_REMAP
#endif ACTIVE_FP_NOISE_CONTROL

        E2 = ( w.E[i].a*cellMass0 +  (flux.f[i].E - flux.f[i+1].E) )*invCellMass1; // 

        w.vx[i].a = vx;
        w.vy[i].a = vy;
#endif SYM2D
        
#ifdef SYM3D
        vx0 = w.vx[i].a;
        vy0 = w.vy[i].a;
        vz0 = w.vz[i].a;
        vx = ( vx0*cellMass0 + (flux.f[i].mx-flux.f[i+1].mx) )*invCellMass1;
        vy = ( vy0*cellMass0 + (flux.f[i].my-flux.f[i+1].my) )*invCellMass1;
        vz = ( vz0*cellMass0 + (flux.f[i].mz-flux.f[i+1].mz) )*invCellMass1;
        
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_REMAP
        vx = zls_tolx( (vx-vx0), kZLS_Small)+vx0;
        vy = zls_tolx( (vy-vy0), kZLS_Small)+vy0;
        vz = zls_tolx( (vz-vz0), kZLS_Small)+vz0;
#endif CONTROL_REMAP
#endif ACTIVE_FP_NOISE_CONTROL

        E2 = ( w.E[i].a*cellMass0 +  (flux.f[i].E - flux.f[i+1].E) )*invCellMass1; // 

        w.vx[i].a = vx;
        w.vy[i].a = vy;
        w.vz[i].a = vz;
#endif SYM3D
        

        w.E[i].a  = E2;
        w.d[i].a  = d;
        
        w.P[i].a  = zlsp_PFromE_W1D(w, i);
       
        
      } // non-zero fluxes
      
			
    } //i
    
    memcpy(w.x, x0, w.nt*sizeof(AxisCoords));
    
  } else { err = memErr;}
	
  return(err);
} // Cylinder radial Remap

// Azimuth (Phi) Remap
int zlsa_LagrangianRemapPhiW1D( W1DArr w,  Flx1DArr   flux, ACoordPtr x0,  Real g ){
  // when phi (azimuth) is second axis in 2D, g = r ;
  // when phi (azimuth) is third axis in 3D Cylinder, g = r ;
  // when phi (azimuth) is second axis in 3D Sphere, g = r cos theta, theta = latitude, equatorial angle;
  // when phi (azimuth) is third axis in 3D, g = r sin theta, theta = co-latitude, or polar angle;
  // g is applied to volumes and boundary speeds
	
  int err = noErr;
  Integer i, k, is, ie;
	
  Real V0; // current volume
  Real V1; // remapped volume
  Real invV1; // 1/V1
  
  Real d,  P2, E2;
  Real dV, dphi;
#ifdef PLM
  Real f, cf; // cf = 1.0-f, fs are cell fractions
#endif PLM
#ifdef PPM
  Real f, f2;
#endif PPM
    
#ifdef SYM1D
  Real  vx0;
  Real vx;
#endif SYM1D
#ifdef SYM2D
  Real vx0, vy0;
  Real vx, vy;
#endif SYM2D
#ifdef SYM3D
  Real  vx0, vy0, vz0;
  Real  vx, vy, vz;
#endif SYM3D
  
  Real cellMass0;
  Real cellMass1;
  Real invCellMass1;
	
#ifndef GODUNOV
  err = zlsppm_CalcRemappingCoeffsW1DArr   ( w ); // all vars, no flattening
  err = zlsp_CalcCellLREnergyW1D  ( w );
#endif !GODUNOV
  
  err = zlsh_ClearWFluxes1DArr	 ( flux );
  
  if ( err == noErr ) {
		
    is = kZLS_LimitRemapLimits; //w.ns-kRemapLimits;
    ie = w.nt-kZLS_LimitRemapLimits; //w.ne+kRemapLimits;
		
    for(i = is; i < ie+1; i++){ //
      
      
      dphi =  w.x[i].l  - x0[i].l; // radians
      dV   =  dphi* g;
			
#ifdef PLM
      f  = 0.0;
      cf = 1.0;
#endif PLM
#ifdef PPM
      f  = 0.0;
      f2 = 1.0;
#endif PPM
      
      if ( dphi > kZLS_Small ) {
        //
        // cell left boundry moved to the right
        // element of w that needs to be interated is the right side
        // of cell i - 1;
        //
        k  = i - 1;
        
#ifdef GODUNOV
        flux.f[i].m  =    dV *w.d[k].a;
        flux.f[i].E  =  flux.f[i].m *w.E[k].a;
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *w.vx[k].a;
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *w.vx[k].a;
        flux.f[i].my = flux.f[i].m *w.vy[k].a;
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *w.vx[k].a;
        flux.f[i].my = flux.f[i].m *w.vy[k].a;
        flux.f[i].mz = flux.f[i].m *w.vz[k].a;
#endif SYM3D
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*w.tr[l][k].a;
        }
#endif

#endif GODUNOV
        
#ifdef PLM        
        f  = 0.5*dphi*(w.x[k].idx);
        cf = 1.0-f;
				
        flux.f[i].m  =    dV *( cf*w.d[k].ar + f*w.d[k].al);
        flux.f[i].E  =  flux.f[i].m *(cf*w.E[k].ar + f*w.E[k].al);
        
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[k].ar + f*w.vx[k].al);
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[k].ar + f*w.vx[k].al);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[k].ar + f*w.vy[k].al);
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[k].ar + f*w.vx[k].al);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[k].ar + f*w.vy[k].al);
        flux.f[i].mz = flux.f[i].m *(cf*w.vz[k].ar + f*w.vz[k].al);
#endif SYM3D
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(cf*w.tr[l][k].ar + f*w.tr[l][k].al);
        }
#endif
        
#endif PLM
        
#ifdef PPM
        f  = 0.5*dphi*(w.x[k].idx);
        f2 = 1.0 - fourthd*f;
				
        flux.f[i].m  =    dV *( w.d[k].ar - f*( w.d[k].del -  f2*w.d[k].a6));
        flux.f[i].E  = flux.f[i].m *(w.E[k].ar - f*(w.E[k].del - f2*w.E[k].a6));
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(w.vx[k].ar - f*(w.vx[k].del - f2*w.vx[k].a6));
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(w.vx[k].ar - f*(w.vx[k].del - f2*w.vx[k].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[k].ar - f*(w.vy[k].del - f2*w.vy[k].a6));
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(w.vx[k].ar - f*(w.vx[k].del - f2*w.vx[k].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[k].ar - f*(w.vy[k].del - f2*w.vy[k].a6));
        flux.f[i].mz = flux.f[i].m *(w.vz[k].ar - f*(w.vz[k].del - f2*w.vz[k].a6));
#endif SYM3D
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(w.tr[l][k].ar - f*(w.tr[l][k].del - f2*w.tr[l][k].a6));
        }
#endif
        
#endif PPM
				
      } else if ( dphi < -kZLS_Small ) {
        //
        // cell left boundry moved to the left
        // element of w that needs to be interated is the left side
        // of cell i;
        //
        
#ifdef GODUNOV
        flux.f[i].m  =    dV *w.d[i].a; // mass transferred
        flux.f[i].E  =  flux.f[i].m *w.E[i].a; // total energy
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *w.vx[i].a; // momentum transferred
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *w.vx[i].a;
        flux.f[i].my = flux.f[i].m *w.vy[i].a;
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *w.vx[i].a;
        flux.f[i].my = flux.f[i].m *w.vy[i].a;
        flux.f[i].mz = flux.f[i].m *w.vz[i].a;
#endif SYM2D
				
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*w.tr[l][i].a;
        }
#endif

#endif GODUNOV
				
#ifdef PLM
        f  = -0.5*dphi*(w.x[i].idx);
				cf = 1.0 - f;
        
        flux.f[i].m  =    dV *( cf*w.d[i].al + f* w.d[i].ar); // mass transferred
        flux.f[i].E  =  flux.f[i].m *(cf*w.E[i].al + f*w.E[i].ar); // total energy
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[i].al + f*w.vx[i].ar); // momentum transferred
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[i].al + f*w.vx[i].ar);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[i].al + f*w.vy[i].ar);
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[i].al + f*w.vx[i].ar);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[i].al + f*w.vy[i].ar);
        flux.f[i].mz = flux.f[i].m *(cf*w.vz[i].al + f*w.vz[i].ar);
#endif SYM2D
				
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(cf*w.tr[l][i].al + f*w.tr[l][i].ar);
        }
#endif

#endif PLM
				
#ifdef PPM
        
        f  = -0.5*dphi*(w.x[i].idx); // OK, just a ratio, pi not needed
        f2 =  1.0 - fourthd*f;
				
        flux.f[i].m  =    dV *( w.d[i].al + f*( w.d[i].del +  f2*w.d[i].a6)); // mass transferred
        
        flux.f[i].E  =  flux.f[i].m *(w.E[i].al + f*(w.E[i].del + f2*w.E[i].a6)); // total energy
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(w.vx[i].al + f*(w.vx[i].del + f2*w.vx[i].a6)); // momentum transferred
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(w.vx[i].al + f*(w.vx[i].del + f2*w.vx[i].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[i].al + f*(w.vy[i].del + f2*w.vy[i].a6));
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(w.vx[i].al + f*(w.vx[i].del + f2*w.vx[i].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[i].al + f*(w.vy[i].del + f2*w.vy[i].a6));
        flux.f[i].mz = flux.f[i].m *(w.vz[i].al + f*(w.vz[i].del + f2*w.vz[i].a6));
#endif SYM2D
				
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(w.tr[l][i].al + f*(w.tr[l][i].del + f2*w.tr[l][i].a6));
        }
#endif
#endif PPM
				
      }       
    } //i
    
		
    is = kZLS_LimitRemapLimits; //w.ns-kRemapLimits;
    ie = w.nt-kZLS_LimitRemapLimits; //w.ne+kRemapLimits;
    
    for(i = is; i < ie; i++){ //
      
      if ((flux.f[i].m != 0.0) || (flux.f[i+1].m != 0.0)){
        
        V0        = w.x[i].dVol * g; // current volume
        V1        =  x0[i].dVol * g;  // remapped volume
      	invV1     = 1.0/V1;
        
        cellMass0 = ( w.d[i].a*V0 );
        cellMass1 = ( cellMass0 + ( flux.f[i].m - flux.f[i+1].m ) );
        invCellMass1 = 1.0/cellMass1;
        
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          w.tr[l][i].a  = (  w.tr[l][i].a*V0 + ( flux.f[i].mtr[l] - flux.f[i+1].mtr[l] ))*invV1;
        }
#endif
        
        d =   cellMass1*invV1;
        
#ifdef SYM1D
        vx0 = w.vx[i].a;
        vx = ( vx0*cellMass0 + (flux.f[i].mx-flux.f[i+1].mx) )*invCellMass1;
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_REMAP
        vx = zls_tolx( (vx-vx0), kZLS_Small)+vx0;
#endif CONTROL_REMAP
#endif ACTIVE_FP_NOISE_CONTROL

        E2 = ( w.E[i].a*cellMass0 +  (flux.f[i].E - flux.f[i+1].E) )*invCellMass1; 

        w.vx[i].a = vx;
#endif SYM1D
        
#ifdef SYM2D
        vx0 = w.vx[i].a;
        vy0 = w.vy[i].a;
        vx = ( vx0*cellMass0 + (flux.f[i].mx-flux.f[i+1].mx) )*invCellMass1;
        vy = ( vy0*cellMass0 + (flux.f[i].my-flux.f[i+1].my) )*invCellMass1;
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_REMAP
        vx = zls_tolx( (vx-vx0), kZLS_Small)+vx0;
        vy = zls_tolx( (vy-vy0), kZLS_Small)+vy0;
#endif CONTROL_REMAP
#endif ACTIVE_FP_NOISE_CONTROL

        E2 = ( w.E[i].a*cellMass0 +  (flux.f[i].E - flux.f[i+1].E) )*invCellMass1; 

        w.vx[i].a = vx;
        w.vy[i].a = vy;
#endif SYM2D
        
#ifdef SYM3D
        vx0 = w.vx[i].a;
        vy0 = w.vy[i].a;
        vz0 = w.vz[i].a;
        vx = ( vx0*cellMass0 + (flux.f[i].mx-flux.f[i+1].mx) )*invCellMass1;
        vy = ( vy0*cellMass0 + (flux.f[i].my-flux.f[i+1].my) )*invCellMass1;
        vz = ( vz0*cellMass0 + (flux.f[i].mz-flux.f[i+1].mz) )*invCellMass1;
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_REMAP
        vx = zls_tolx( (vx-vx0), kZLS_Small)+vx0;
        vy = zls_tolx( (vy-vy0), kZLS_Small)+vy0;
        vz = zls_tolx( (vz-vz0), kZLS_Small)+vz0;
#endif CONTROL_REMAP
#endif ACTIVE_FP_NOISE_CONTROL

        E2 = ( w.E[i].a*cellMass0 +  (flux.f[i].E - flux.f[i+1].E) )*invCellMass1; 

        w.vx[i].a = vx;
        w.vy[i].a = vy;
        w.vz[i].a = vz;
#endif SYM3D
        

        w.E[i].a  = E2;
        w.d[i].a  = d;
        
        w.P[i].a  = zlsp_PFromE_W1D(w, i);
       
        
      } // non-zero fluxes
      
    } //i
		
    
    memcpy(w.x, x0, w.nt*sizeof(AxisCoords));
    
  } else { err = memErr;}
	
  return(err);
  
} // Azimuthal (Phi) Remap

// Altitude (Theta) Remap
//  0 +/- pi/2 altitude/latitude equatorial angle elevation axis.  R-Phi-Theta is right handed (*not* R-Theta-Phi)

int zlsa_LagrangianRemapThetaW1D( W1DArr w,  Flx1DArr   flux, ACoordPtr x0, Real g ){
  // when phi (azimuth) is second axis in 3D, theta = latitude, equatorial angle;
  // when phi (azimuth) is third axis in 3D, theta = co-latitude, or polar angle;
  // g is applied to volumes and boundary speeds
	
  int err = noErr;
  Integer i, k, is, ie;
	
  Real V0; // current volume
  Real V1; // remapped volume
  Real invV1; // 1/V1
  
  Real d,  P2, E2;
  Real dV, dtheta;
#ifdef PLM
  Real f, cf; // cell fractions, cf = co-fraction = 1-f
#endif PLM
#ifdef PPM
  Real f, f2;
#endif PPM
    
  
#ifdef SYM1D
  Real  vx0;
  Real vx;
#endif SYM1D
#ifdef SYM2D
  Real vx0, vy0;
  Real vx, vy;
#endif SYM2D
#ifdef SYM3D
  Real  vx0, vy0, vz0;
  Real  vx, vy, vz;
#endif SYM3D
  
  Real cellMass0;
  Real cellMass1;
  Real invCellMass1;
	
#ifndef GODUNOV
  err = zlsppm_CalcRemappingCoeffsW1DArr   ( w ); // all vars, no flattening
  err = zlsp_CalcCellLREnergyW1D  ( w );
#endif !GODUNOV
  
  err = zlsh_ClearWFluxes1DArr	 ( flux );
  
  if ( err == noErr ) {
		
    is = kZLS_LimitRemapLimits; //w.ns-kRemapLimits;
    ie = w.nt-kZLS_LimitRemapLimits; //w.ne+kRemapLimits;
		
    for(i = is; i < ie+1; i++){ //
			
      dtheta =  w.x[i].l  - x0[i].l; // radians
      dV     = (w.x[i].sl - x0[i].sl) * g;  // sin(theta) - sin(theta0);
      
#ifdef PLM
      f  = 0.0;
      cf = 1.0;
#endif PLM
#ifdef PPM
      f  = 0.0;
      f2 = 1.0;
#endif PPM
      
      if ( dtheta > kZLS_Small ) {
        //
        // cell left boundry moved to the right
        // element of w that needs to be interated is the right side
        // of cell i - 1;
        //
        k  = i - 1;
        
#ifdef GODUNOV
        flux.f[i].m  =    dV *w.d[k].a;
        flux.f[i].E  =  flux.f[i].m *w.E[k].a;
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *w.vx[k].a;
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *w.vx[k].a;
        flux.f[i].my = flux.f[i].m *w.vy[k].a;
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *w.vx[k].a;
        flux.f[i].my = flux.f[i].m *w.vy[k].a;
        flux.f[i].mz = flux.f[i].m *w.vz[k].a;
#endif SYM3D
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*w.tr[l][k].a;
        }
#endif

#endif GODUNOV
        
#ifdef PLM        
        f  = 0.5*dtheta*(w.x[k].idx);
        cf = 1.0-f;
				
        flux.f[i].m  =    dV *( cf*w.d[k].ar + f*w.d[k].al);
        flux.f[i].E  =  flux.f[i].m *(cf*w.E[k].ar + f*w.E[k].al);
        
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[k].ar + f*w.vx[k].al);
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[k].ar + f*w.vx[k].al);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[k].ar + f*w.vy[k].al);
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[k].ar + f*w.vx[k].al);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[k].ar + f*w.vy[k].al);
        flux.f[i].mz = flux.f[i].m *(cf*w.vz[k].ar + f*w.vz[k].al);
#endif SYM3D
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(cf*w.tr[l][k].ar + f*w.tr[l][k].al);
        }
#endif

#endif PLM
        
#ifdef PPM
        f  = 0.5*dtheta*(w.x[k].idx);
        f2 = 1.0 - fourthd*f;
				
        flux.f[i].m  =    dV *( w.d[k].ar - f*( w.d[k].del -  f2*w.d[k].a6));
        flux.f[i].E  = flux.f[i].m *(w.E[k].ar - f*(w.E[k].del - f2*w.E[k].a6));
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(w.vx[k].ar - f*(w.vx[k].del - f2*w.vx[k].a6));
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(w.vx[k].ar - f*(w.vx[k].del - f2*w.vx[k].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[k].ar - f*(w.vy[k].del - f2*w.vy[k].a6));
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(w.vx[k].ar - f*(w.vx[k].del - f2*w.vx[k].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[k].ar - f*(w.vy[k].del - f2*w.vy[k].a6));
        flux.f[i].mz = flux.f[i].m *(w.vz[k].ar - f*(w.vz[k].del - f2*w.vz[k].a6));
#endif SYM3D
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(w.tr[l][k].ar - f*(w.tr[l][k].del - f2*w.tr[l][k].a6));
        }
#endif

#endif PPM
        
      } else if ( dtheta < -kZLS_Small ) {
        //
        // cell left boundry moved to the left
        // element of w that needs to be interated is the left side
        // of cell i;
        //
        
#ifdef GODUNOV
        flux.f[i].m  =    dV *w.d[i].a; // mass transferred
        flux.f[i].E  =  flux.f[i].m *w.E[i].a; // total energy
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *w.vx[i].a; // momentum transferred
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *w.vx[i].a;
        flux.f[i].my = flux.f[i].m *w.vy[i].a;
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *w.vx[i].a;
        flux.f[i].my = flux.f[i].m *w.vy[i].a;
        flux.f[i].mz = flux.f[i].m *w.vz[i].a;
#endif SYM2D
				
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*w.tr[l][i].a;
        }
#endif

#endif GODUNOV
				
#ifdef PLM
        f  = -0.5*dtheta*(w.x[i].idx);
				cf = 1.0 - f;
        
        flux.f[i].m  =    dV *( cf*w.d[i].al + f* w.d[i].ar); // mass transferred
        flux.f[i].E  =  flux.f[i].m *(cf*w.E[i].al + f*w.E[i].ar); // total energy
        
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[i].al + f*w.vx[i].ar); // momentum transferred
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[i].al + f*w.vx[i].ar);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[i].al + f*w.vy[i].ar);
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[i].al + f*w.vx[i].ar);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[i].al + f*w.vy[i].ar);
        flux.f[i].mz = flux.f[i].m *(cf*w.vz[i].al + f*w.vz[i].ar);
#endif SYM2D
				
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(cf*w.tr[l][i].al + f*w.tr[l][i].ar);
        }
#endif

#endif PLM
				
#ifdef PPM
				
        f  = -0.5*dtheta*(w.x[i].idx);
        f2 =  1.0 - fourthd*f;
				
        flux.f[i].m  =    dV *( w.d[i].al + f*( w.d[i].del +  f2*w.d[i].a6)); // mass transferred
        flux.f[i].E  =  flux.f[i].m *(w.E[i].al + f*(w.E[i].del + f2*w.E[i].a6)); // total energy
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(w.vx[i].al + f*(w.vx[i].del + f2*w.vx[i].a6)); // momentum transferred
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(w.vx[i].al + f*(w.vx[i].del + f2*w.vx[i].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[i].al + f*(w.vy[i].del + f2*w.vy[i].a6));
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(w.vx[i].al + f*(w.vx[i].del + f2*w.vx[i].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[i].al + f*(w.vy[i].del + f2*w.vy[i].a6));
        flux.f[i].mz = flux.f[i].m *(w.vz[i].al + f*(w.vz[i].del + f2*w.vz[i].a6));
#endif SYM2D
				
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(w.tr[l][i].al + f*(w.tr[l][i].del + f2*w.tr[l][i].a6));
        }
#endif

#endif PPM
				
      }
    } //i
		
    // update cells with remap 'fluxes'  and restore coordinates
		
    is = kZLS_LimitRemapLimits; //w.ns-kRemapLimits;
    ie = w.nt-kZLS_LimitRemapLimits; //w.ne+kRemapLimits;
    
    for(i = is; i < ie; i++){ //
      
      if ((flux.f[i].m != 0.0) || (flux.f[i+1].m != 0.0)){
        
        V0        = w.x[i].dVol * g; // current volume
        V1        =  x0[i].dVol * g;  // remapped volume
      	invV1     = 1.0/V1;
        
        cellMass0 = ( w.d[i].a*V0 );
        cellMass1 = ( cellMass0 + ( flux.f[i].m - flux.f[i+1].m ) );
        invCellMass1 = 1.0/cellMass1;
        
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          w.tr[l][i].a  = (  w.tr[l][i].a*V0 + ( flux.f[i].mtr[l] - flux.f[i+1].mtr[l] ))*invV1;
        }
#endif
        
        d =   cellMass1*invV1;
        
#ifdef SYM1D
        vx0 = w.vx[i].a;
        vx = ( vx0*cellMass0 + (flux.f[i].mx-flux.f[i+1].mx) )*invCellMass1;
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_REMAP
        vx = zls_tolx( (vx-vx0), kZLS_Small)+vx0;
#endif CONTROL_REMAP
#endif ACTIVE_FP_NOISE_CONTROL

        E2 = ( w.E[i].a*cellMass0 +  (flux.f[i].E - flux.f[i+1].E) )*invCellMass1; 

        w.vx[i].a = vx;
#endif SYM1D
        
#ifdef SYM2D
        vx0 = w.vx[i].a;
        vy0 = w.vy[i].a;
        vx = ( vx0*cellMass0 + (flux.f[i].mx-flux.f[i+1].mx) )*invCellMass1;
        vy = ( vy0*cellMass0 + (flux.f[i].my-flux.f[i+1].my) )*invCellMass1;
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_REMAP
        vx = zls_tolx( (vx-vx0), kZLS_Small)+vx0;
        vy = zls_tolx( (vy-vy0), kZLS_Small)+vy0;
#endif CONTROL_REMAP
#endif ACTIVE_FP_NOISE_CONTROL

        E2 = ( w.E[i].a*cellMass0 +  (flux.f[i].E - flux.f[i+1].E) )*invCellMass1; 

        w.vx[i].a = vx;
        w.vy[i].a = vy;
#endif SYM2D
        
#ifdef SYM3D
        vx0 = w.vx[i].a;
        vy0 = w.vy[i].a;
        vz0 = w.vz[i].a;
        vx = ( vx0*cellMass0 + (flux.f[i].mx-flux.f[i+1].mx) )*invCellMass1;
        vy = ( vy0*cellMass0 + (flux.f[i].my-flux.f[i+1].my) )*invCellMass1;
        vz = ( vz0*cellMass0 + (flux.f[i].mz-flux.f[i+1].mz) )*invCellMass1;
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_REMAP
        vx = zls_tolx( (vx-vx0), kZLS_Small)+vx0;
        vy = zls_tolx( (vy-vy0), kZLS_Small)+vy0;
        vz = zls_tolx( (vz-vz0), kZLS_Small)+vz0;
#endif CONTROL_REMAP
#endif ACTIVE_FP_NOISE_CONTROL

        E2 = ( w.E[i].a*cellMass0 +  (flux.f[i].E - flux.f[i+1].E) )*invCellMass1; 
        
        w.vx[i].a = vx;
        w.vy[i].a = vy;
        w.vz[i].a = vz;
#endif SYM3D
        
        w.E[i].a  = E2;
        w.d[i].a  = d;
        
        w.P[i].a  = zlsp_PFromE_W1D(w, i);
        
      } // non-zero fluxes
      
    } //i
		
		
    memcpy(w.x, x0, w.nt*sizeof(AxisCoords));
    
  } else { err = memErr;}
	
  return(err);
} // Altitude (Theta) Remap

// Co-Latitude Polar Theta Remap
// 0 - pi polar angle, co-latitide, down from z axis, used commonly in physics, R-pTheta-Phi is RH.
int zlsa_LagrangianRemapPolarThetaW1D( W1DArr w,  Flx1DArr   flux,  ACoordPtr x0, Real g ){
  // when phi (azimuth) is second axis in 3D, theta = latitude, equatorial angle;
  // when phi (azimuth) is third axis in 3D, theta = co-latitude, or polar angle;
  // g is applied to volumes and boundary speeds
	
  int err = noErr;
  Integer i, k, is, ie;
	
  Real V0; // current volume
  Real V1; // remapped volume
  Real invV1; // 1/V1
  
  Real d,  P2, E2;
  Real dV, dtheta;
#ifdef PLM
  Real f, cf; // cell fractions, cf = co-fraction = 1-f
#endif PLM
#ifdef PPM
  Real f, f2;
#endif PPM
  
#ifdef SYM1D
  Real  vx0;
  Real vx;
#endif SYM1D
#ifdef SYM2D
  Real vx0, vy0;
  Real vx, vy;
#endif SYM2D
#ifdef SYM3D
  Real  vx0, vy0, vz0;
  Real  vx, vy, vz;
#endif SYM3D
  
  Real cellMass0;
  Real cellMass1;
  Real invCellMass1;
	
#ifndef GODUNOV
  err = zlsppm_CalcRemappingCoeffsW1DArr   ( w ); // all vars, no flattening
  err = zlsp_CalcCellLREnergyW1D  ( w );
#endif !GODUNOV
  
  err = zlsh_ClearWFluxes1DArr	 ( flux );
  
  if ( err == noErr ) {
		
    is = kZLS_LimitRemapLimits; //w.ns-kRemapLimits;
    ie = w.nt-kZLS_LimitRemapLimits; //w.ne+kRemapLimits;
		
    for(i = is; i < ie+1; i++){ //
			
      dtheta =  w.x[i].l  - x0[i].l; // radians
      dV   =  (w.x[i].cl - x0[i].cl)*g; // cos(radians) in deltaX, g = radius
      
#ifdef PLM
      f  = 0.0;
      cf = 1.0;
#endif PLM
#ifdef PPM
      f  = 0.0;
      f2 = 1.0;
#endif PPM
      
      if ( dtheta > kZLS_Small ) {
        //
        // cell left boundry moved to the right
        // element of w that needs to be interated is the right side
        // of cell i - 1;
        //
        k  = i - 1;
        
#ifdef GODUNOV
        flux.f[i].m  =    dV *w.d[k].a;
        flux.f[i].E  =  flux.f[i].m *w.E[k].a;
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *w.vx[k].a;
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *w.vx[k].a;
        flux.f[i].my = flux.f[i].m *w.vy[k].a;
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *w.vx[k].a;
        flux.f[i].my = flux.f[i].m *w.vy[k].a;
        flux.f[i].mz = flux.f[i].m *w.vz[k].a;
#endif SYM3D
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*w.tr[l][k].a;
        }
#endif

#endif GODUNOV
        
#ifdef PLM        
        f  = 0.5*dtheta*(w.x[k].idx);
        cf = 1.0-f;
				
        flux.f[i].m  =    dV *( cf*w.d[k].ar + f*w.d[k].al);
        flux.f[i].E  =  flux.f[i].m *(cf*w.E[k].ar + f*w.E[k].al);
        
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[k].ar + f*w.vx[k].al);
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[k].ar + f*w.vx[k].al);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[k].ar + f*w.vy[k].al);
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[k].ar + f*w.vx[k].al);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[k].ar + f*w.vy[k].al);
        flux.f[i].mz = flux.f[i].m *(cf*w.vz[k].ar + f*w.vz[k].al);
#endif SYM3D
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(cf*w.tr[l][k].ar + f*w.tr[l][k].al);
        }
#endif

#endif PLM
        
#ifdef PPM
        f  = 0.5*dtheta*(w.x[k].idx);
        f2 = 1.0 - fourthd*f;
				
        flux.f[i].m  =    dV *( w.d[k].ar - f*( w.d[k].del -  f2*w.d[k].a6));
        flux.f[i].E  = flux.f[i].m *(w.E[k].ar - f*(w.E[k].del - f2*w.E[k].a6));
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(w.vx[k].ar - f*(w.vx[k].del - f2*w.vx[k].a6));
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(w.vx[k].ar - f*(w.vx[k].del - f2*w.vx[k].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[k].ar - f*(w.vy[k].del - f2*w.vy[k].a6));
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(w.vx[k].ar - f*(w.vx[k].del - f2*w.vx[k].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[k].ar - f*(w.vy[k].del - f2*w.vy[k].a6));
        flux.f[i].mz = flux.f[i].m *(w.vz[k].ar - f*(w.vz[k].del - f2*w.vz[k].a6));
#endif SYM3D
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(w.tr[l][k].ar - f*(w.tr[l][k].del - f2*w.tr[l][k].a6));
        }
#endif

#endif PPM
        
      } else if ( dtheta < -kZLS_Small ) {
        //
        // cell left boundry moved to the left
        // element of w that needs to be interated is the left side
        // of cell i;
        //
        
#ifdef GODUNOV
        flux.f[i].m  =    dV *w.d[i].a; // mass transferred
        flux.f[i].E  =  flux.f[i].m *w.E[i].a; // total energy
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *w.vx[i].a; // momentum transferred
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *w.vx[i].a;
        flux.f[i].my = flux.f[i].m *w.vy[i].a;
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *w.vx[i].a;
        flux.f[i].my = flux.f[i].m *w.vy[i].a;
        flux.f[i].mz = flux.f[i].m *w.vz[i].a;
#endif SYM2D
				
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*w.tr[l][i].a;
        }
#endif

#endif GODUNOV
				
#ifdef PLM
        f  = -0.5*dtheta*(w.x[i].idx);
				cf = 1.0 - f;
        
        flux.f[i].m  =    dV *( cf*w.d[i].al + f* w.d[i].ar); // mass transferred
        flux.f[i].E  =  flux.f[i].m *(cf*w.E[i].al + f*w.E[i].ar); // total energy
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[i].al + f*w.vx[i].ar); // momentum transferred
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[i].al + f*w.vx[i].ar);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[i].al + f*w.vy[i].ar);
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(cf*w.vx[i].al + f*w.vx[i].ar);
        flux.f[i].my = flux.f[i].m *(cf*w.vy[i].al + f*w.vy[i].ar);
        flux.f[i].mz = flux.f[i].m *(cf*w.vz[i].al + f*w.vz[i].ar);
#endif SYM2D
				
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(cf*w.tr[l][i].al + f*w.tr[l][i].ar);
        }
#endif

#endif PLM
				
#ifdef PPM
				
        f  = -0.5*dtheta*(w.x[i].idx);
        f2 =  1.0 - fourthd*f;
				
        flux.f[i].m  =    dV *( w.d[i].al + f*( w.d[i].del +  f2*w.d[i].a6)); // mass transferred
        flux.f[i].E  =  flux.f[i].m *(w.E[i].al + f*(w.E[i].del + f2*w.E[i].a6)); // total energy
				
#ifdef SYM1D
        flux.f[i].mx = flux.f[i].m *(w.vx[i].al + f*(w.vx[i].del + f2*w.vx[i].a6)); // momentum transferred
#endif SYM1D
#ifdef SYM2D
        flux.f[i].mx = flux.f[i].m *(w.vx[i].al + f*(w.vx[i].del + f2*w.vx[i].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[i].al + f*(w.vy[i].del + f2*w.vy[i].a6));
#endif SYM2D
#ifdef SYM3D
        flux.f[i].mx = flux.f[i].m *(w.vx[i].al + f*(w.vx[i].del + f2*w.vx[i].a6));
        flux.f[i].my = flux.f[i].m *(w.vy[i].al + f*(w.vy[i].del + f2*w.vy[i].a6));
        flux.f[i].mz = flux.f[i].m *(w.vz[i].al + f*(w.vz[i].del + f2*w.vz[i].a6));
#endif SYM2D
				
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          flux.f[i].mtr[l]  = dV*(w.tr[l][i].al + f*(w.tr[l][i].del + f2*w.tr[l][i].a6));
        }
#endif

#endif PPM
				
      }
    } //i
		
    // update cells with remap 'fluxes'  and restore coordinates
		
    is = kZLS_LimitRemapLimits; //w.ns-kRemapLimits;
    ie = w.nt-kZLS_LimitRemapLimits; //w.ne+kRemapLimits;
    
    for(i = is; i < ie; i++){ //
      
      if ((flux.f[i].m != 0.0) || (flux.f[i+1].m != 0.0)){
        
        V0        = w.x[i].dVol * g; // current volume
        V1        =  x0[i].dVol * g;  // remapped volume
      	invV1     = 1.0/V1;
        
        cellMass0 = ( w.d[i].a*V0 );
        cellMass1 = ( cellMass0 + ( flux.f[i].m - flux.f[i+1].m ) );
        invCellMass1 = 1.0/cellMass1;
        
#if (NTRACERS > 0)    
        Integer l;
        for (l = 0; l < kNTracers; l++){
          w.tr[l][i].a  = (  w.tr[l][i].a*V0 + ( flux.f[i].mtr[l] - flux.f[i+1].mtr[l] ))*invV1;
        }
#endif
        
        d =   cellMass1*invV1;
        
#ifdef SYM1D
        vx0 = w.vx[i].a;
        vx = ( vx0*cellMass0 + (flux.f[i].mx-flux.f[i+1].mx) )*invCellMass1;
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_REMAP
        vx = zls_tolx( (vx-vx0), kZLS_Small)+vx0;
#endif CONTROL_REMAP
#endif ACTIVE_FP_NOISE_CONTROL

        E2 = ( w.E[i].a*cellMass0 +  (flux.f[i].E - flux.f[i+1].E) )*invCellMass1; 

        w.vx[i].a = vx;
#endif SYM1D
        
#ifdef SYM2D
        vx0 = w.vx[i].a;
        vy0 = w.vy[i].a;
        vx = ( vx0*cellMass0 + (flux.f[i].mx-flux.f[i+1].mx) )*invCellMass1;
        vy = ( vy0*cellMass0 + (flux.f[i].my-flux.f[i+1].my) )*invCellMass1;
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_REMAP
        vx = zls_tolx( (vx-vx0), kZLS_Small)+vx0;
        vy = zls_tolx( (vy-vy0), kZLS_Small)+vy0;
#endif CONTROL_REMAP
#endif ACTIVE_FP_NOISE_CONTROL

        E2 = ( w.E[i].a*cellMass0 +  (flux.f[i].E - flux.f[i+1].E) )*invCellMass1; // 

        w.vx[i].a = vx;
        w.vy[i].a = vy;
#endif SYM2D
        
#ifdef SYM3D
        vx0 = w.vx[i].a;
        vy0 = w.vy[i].a;
        vz0 = w.vz[i].a;
        vx = ( vx0*cellMass0 + (flux.f[i].mx-flux.f[i+1].mx) )*invCellMass1;
        vy = ( vy0*cellMass0 + (flux.f[i].my-flux.f[i+1].my) )*invCellMass1;
        vz = ( vz0*cellMass0 + (flux.f[i].mz-flux.f[i+1].mz) )*invCellMass1;
#ifdef ACTIVE_FP_NOISE_CONTROL   
#ifdef CONTROL_REMAP
        vx = zls_tolx( (vx-vx0), kZLS_Small)+vx0;
        vy = zls_tolx( (vy-vy0), kZLS_Small)+vy0;
        vz = zls_tolx( (vz-vz0), kZLS_Small)+vz0;
#endif CONTROL_REMAP
#endif ACTIVE_FP_NOISE_CONTROL

        E2 = ( w.E[i].a*cellMass0 +  (flux.f[i].E - flux.f[i+1].E) )*invCellMass1;

        w.vx[i].a = vx;
        w.vy[i].a = vy;
        w.vz[i].a = vz;
#endif SYM3D
        
        w.E[i].a  = E2;
        w.d[i].a  = d;
        
        w.P[i].a  = zlsp_PFromE_W1D(w, i);
       
      } // non-zero fluxes
      
    } //i
		
		
    memcpy(w.x, x0, w.nt*sizeof(AxisCoords));
    
  } else { err = memErr;}
	
  return(err);
} // Co-Latitude Polar Theta Remap

#endif !CARTESIAN

#endif LAGRANGIAN

