#ifndef ZLS_CELLUPDATE_H
#define ZLS_CELLUPDATE_H
#include <zls.h>
//
//====================================================================================
//
// Simulation Variable Update Algorithms
//
//====================================================================================
// constants for smoothing/artificial viscosity if used.
extern Integer kZLS_SmoothAxes; // Combination of smooth axes flags
extern Real    kZLS_SmoothThreshold;   // min inflection for vertex algorithm
extern Real    kZLS_RemapAlpha;   // exchange flux fraction, <= 0.005 for stability 
#ifdef DIVVHCORRECTION
extern  Real    kZLS_DivVScale;   // divV for unit smoothing
extern  Real    kZLS_DivVInvScale;   // 1.0/divV for unit smoothing
extern  Real    kZLS_DivVThreshold;   // min divV for smoothing
extern  Real    kZLS_RemapMaxAlpha;   // max smoothing amount
#endif DIVVHCORRECTION

enum{ 
  kSmoothNone  = 0x00,
  kSmoothXFlag = 0x01,
  kSmoothYFlag = 0x02,
  kSmoothZFlag = 0x04,
  kSmoothAll   = 0x07
};

#ifdef EULERIAN
int zlsa_EulerianFluxesW1D( Flx1DArr flux,  Wstar1DArr wStar);
int zlsa_EulerianUpdateW1D( W1DArr w, Flx1DArr flux, Real dt);
#endif EULERIAN

//
#ifdef LAGRANGIAN
//  Primitive Variable Lagrangian Remap Algorithms. 
//  NOTE: to use these routines, the LAGRANGIAN macro must be defined in zls_const.h

//====================================================================================
//
// use the coords in x0 and w to remap the cells in w to their original
// locations the cell location differences result in fluxes between
// cells. zlsa_LagrangianRemapW1D computes the fluxes and updates the
// cells in place, so can't be used for passing fluxes up to higher
// levels.  But it can be used for single level sims and *greatly*
// reduces memory requiements then as the zoneFluxes arrays are not
// used.
//
// Uses the x0 axis, and the axis in w, which is the modified x axis
// afer the evolve step.  At the end the axis in w is restored to a
// copy of the x0 axis, as it is a reference to the zone's axis and is
// needed for subsequent sweeps.  Using a ref to x (the modified axis)
// could be a problem in a multi-tasking environment, breaking
// re-entrancy.  In that case the FillW1D routine will make copies, not
// just take references to the zone axes, which will be slower but safe
// against two threads sweeping the same zone.
//
//====================================================================================
// 
// If a potential field is present, the potential is interpolated
// conservatively using PPM or PLM, and the integral average
// acceleration is <a> = -(F_r - F_l)/(Xr-Xl) This call returns the
// accelerations for the original grid cells and the accelleration
// averaged conservatively over each cell in the lagrangian moved
// cells. a0 and a1 are arrays over [i]  refering to cell i before and
// after the boundaries have moved.  The as are multiplied by 0.5 dt to
// save the muliplies later in the evolve1DF routine.
int zlsa_LagrangianCellAccelleration1D ( Real * a0, Real * a1, Integer nt,  ACoordPtr x0,  ACoordPtr x1, F1DArr fld, Real dt);
//
//  After this step the axis in w is modified by the cell boundary motions, and
//  the axis in wStar stays unchanged, so the differences can be used to compute 
//  remaping fluxes.
//
int zlsa_LagrangianSimpleEvolveW1D   ( W1DArr w, Wstar1DArr wStar,   Real dt);

int zlsa_LagrangianEvolveW1D         ( W1DArr w,  Wstar1DArr wStar,  Real dt); 
int zlsa_LagrangianEvolveW1DF        ( W1DArr w,  Wstar1DArr wStar, F1DArr f, Real dt); 
// inludes a potential field in f.

// each axis is different and a slightly different remap type is needed
// for each axis to track the change in volume elements long the axes,
// ie divergence, or the take into account the geometric projection, g,
// to convert angular steps to spatial ones.
int zlsa_LagrangianEvolveAxisTypeW1D     ( W1DArr w,  Wstar1DArr wStar, Integer axisType, Real g, Real dt); 
 
#ifndef CARTESIAN
int zlsa_LagrangianEvolveSRadialW1D      ( W1DArr w, Wstar1DArr wStar,  Real dt ); // radial coordinate - diverges r^2
int zlsa_LagrangianEvolveSRadialW1DF     ( W1DArr w, Wstar1DArr wStar,  F1DArr f, Real dt ); // radial coordinate - diverges r^2

int zlsa_LagrangianEvolveCRadialW1D      ( W1DArr w, Wstar1DArr wStar,  Real dt ); // radial coordinate - diverges r
int zlsa_LagrangianEvolveCRadialW1DF     ( W1DArr w, Wstar1DArr wStar,  F1DArr f, Real dt ); // radial coordinate - diverges r

// assume no non-radial gravity for now, will change when self-gravity in non-cartesian coords are supported

int zlsa_LagrangianEvolvePhiW1D          ( W1DArr w, Wstar1DArr wStar,  Real g,  Real dt );
int zlsa_LagrangianEvolveThetaW1D        ( W1DArr w, Wstar1DArr wStar,  Real g,  Real dt );
int zlsa_LagrangianEvolvePolarThetaW1D   ( W1DArr w, Wstar1DArr wStar,  Real g,  Real dt );
#endif !CARTESIAN

//====================================================================================
int zlsa_LagrangianRemapW1D      ( W1DArr w,  Flx1DArr flux, ACoordPtr x0); // cartesian remap, no geometric source terms.
int zlsa_RemapSmoothW1D          ( W1DArr w, Real dt );

// int zlsa_LagrangianRemapAxisTypeW1D( W1DArr w, Flx1DArr flux,
// ACoordPtr x0, Integer axisType , Real g); 
// a general, slightly  slower, call for all the below
//
// each axis is different and a slightly different remap type is needed
// for each axis
#ifndef CARTESIAN
int zlsa_LagrangianRemapSRadialW1D    ( W1DArr w, Flx1DArr flux, ACoordPtr x0 ); // radial coordinate - diverges r^2
int zlsa_LagrangianRemapCRadialW1D    ( W1DArr w, Flx1DArr flux, ACoordPtr x0 ); // radial coordinate - diverges r
int zlsa_LagrangianRemapPhiW1D        ( W1DArr w, Flx1DArr flux, ACoordPtr x0, Real g ); // azimuth, all round, Phi,  g = radius sin(Polar Theta) or radius Cos(Theta) in 3D, just radius in 2D
int zlsa_LagrangianRemapThetaW1D      ( W1DArr w, Flx1DArr flux, ACoordPtr x0, Real g ); // altitude, pole-pole, Theta, g = radius
int zlsa_LagrangianRemapPolarThetaW1D ( W1DArr w, Flx1DArr flux, ACoordPtr x0, Real g ); // co-latitude, pole down, Theta, g = radius
#endif !CARTESIAN
//====================================================================================
//
#endif LAGRANGIAN

#endif ZLS_CELLUPDATE_H
