#include <zls.h>

//====================================================================================
//
// PPM Cell Boundary State Reconstruction Algorithms
//
//====================================================================================

static const Size kElementSize = sizeof(Real);
//========================================
// Low level calls for computing coefficents: PPM. PLM and flattening
// on lower level Arrays/  Called by higher level routines on sub
// elements of higher level structures.
//========================================

inline int  zlsppm_CalcLRVar1DArr (V1DArr v,  Q1DArr flat){
#ifdef GODUNOV			
	return(noErr);
#else 			
	/*
	 //
	 // Colella and Woodward,1984, JCompPhys 54, 174
	 //
	 */      
	int err = noErr;
	
	Integer i,n;
	
	Real  fl, onemfl;
	
	Integer is, ie;
	Var1 *      va;
	
	ACoordPtr    x;
	
	if ((v.v!=NULL)&&(v.x!=NULL)) {
		
		va = v.v;
		x  = v.x;
		
		is = 1;
		ie = v.nt-1;
		n  = v.nt;
    Real da[n];
    Real diffa[n];
    
    bzero((void *) da, n*sizeof(Real));
    
		for(i = 0; i < ie; i++){
			diffa[i] =  va[i+1].a - va[i].a;
		}//i

    //
    //     Equation 1.7 of C&W
    //     da(j) = pc3 * (a(j+1) - a(j)) + pc4 * (a(j) - a(j-1))
    //    leave da[n] 0.0 if v[n].a is a local max/min/flat
    
		for (i = is; i < ie; i++){
			if(diffa[i-1]*diffa[i] > 0.0) {
				da[i] = x[i].pc3 * diffa[i] +  x[i].pc4* diffa[i-1];
        da[i] = zls_sign( zls_min3(zls_abs(da[i]), 2.0*zls_abs(diffa[i-1]), 2.0*zls_abs(diffa[i])), da[i] );
			}
		} //i
    
    
#ifdef PPM
    //
    //     Equation 1.6 of C&W
    //     a(j+.5) = a(j) + pc0 * (a(j+1)-a(j)) + pc1 * dma(j) + pc2 * dma(j+1)
    //     MONOT: Limit ar[n] to the range defined by a[n] and a[n+1]
    //
    //     NOTE: we now swap coeff 1 and 2 wrt ppmlr, so the last coeff
    //     applies to delta_a_j+1, and stepping through x in order,
    //     a(j+.5) = a(j) + pc0 * (a(j+1)-a(j)) + pc1 * dma(j) + pc2 * dma(j+1)

    va[is].al = va[0].ar;
		for(i = is ; i < ie ; i++){
			va[i].ar   = va[i].a +  x[i].pc0*diffa[i] +  x[i].pc1*da[i] +  x[i].pc2*da[i+1];
			va[i+1].al = va[i].ar;
		} //i
    va[ie-1].ar = va[ie].al;
#endif PPM	
    
#ifdef PLM
    
		for(i = is ; i < ie ; i++){
			va[i].ar = va[i].a +  0.5*da[i];
			va[i].al = va[i].a -  0.5*da[i];
		} //i
    va[0].ar  = va[is].al;
    va[0].al  = va[0].ar;
    va[ie].al = va[ie-1].ar;
    va[ie].ar = va[ie].al;
    
#endif PLM		
    
		//
		// eqn. 4.1 - flatten interpolation in zones with a shock ( flat[n]->1. )
		//
		
		if( flat !=NULL) {
			for(i = is ; i <= ie ;i++){
      				fl       = flat[i];
				if ( fl > 0.0) {
          onemfl   = 1.0 - fl   ;
          va[i].ar = fl * va[i].a + onemfl * va[i].ar;
          va[i].al = fl * va[i].a + onemfl * va[i].al;
       }
			} //i
		} // 
		
		
	} else { // !v
		err = memErr;
	}
	//
	return (err);
#endif GODUNOV			
}


inline int  zlsppm_CalcDelA6Var1DArr (V1DArr v){
#ifndef PPM
  //  a6 and del are only used in the PPM interpolation method
  // PLM uses just al and ar, while GODUNOV just a.
  return(dimErr);
#else
  /*
	 //
	 // Colella and Woodward,1984, JCompPhys 54, 174 
	 //
	 */      
	int err = noErr;
	
	Integer i;
	
	Real  aj, al, ar, dela, a6a;
	Real  scrch1;
	Real  scrch2;
	
	Integer is, ie;
	Var1 *      va;
	
	//
	//========================================
	//  Compute Delta and a6 if the L/R values are known
	//
	
	if (v.v!=NULL) {
		
		va = v.v;
		
		is = 1;
		ie = v.nt-1;
    
		va [0].al  = va [0].a;
		va [0].ar  = va [0].a;
		va[ie].ar  = va[ie].a;
		va[ie].al  = va[ie].a;
    
#ifdef PPM
		va [0].del  = 0.0;
		va [0].a6   = 0.0;
		va[ie].del  = 0.0;
		va[ie].a6   = 0.0;
#endif PPM
		//
		// compute delta_a, a_6
		// MONOT: if a is a local max/min, flaten zone structure ar,al -> a.
		// MONOT: compute monotonized values using eq. 1.10 of C&W
		//        if parabola exceeds al/ar, reset ar/al so that slope -> 0.
		// Recalculate delta_a and a_6
		//
		for(i = is;i<ie;i++){
			
			ar =  va[i].ar ;
			al =  va[i].al ;
			aj =  va[i].a   ;
			
			dela = ar - al ;
			a6a  = 6.0* (aj - 0.5 * (ar + al)) ;
			
			scrch1 = (ar - aj) * (aj - al) ;
			if(scrch1 < 0.0) {
				ar = aj;
				al = aj;
        dela = a6a = 0.0;
        goto store;
			}
			
			scrch1 = dela * dela;
			scrch2 = dela * a6a;
			
			al = (scrch1 < +scrch2)? (3.0*aj) - (2.0* ar):al;      
			ar = (scrch1 < -scrch2)? (3.0*aj) - (2.0* al):ar;
      
			dela  = ar  - al;
			a6a   = 6.0* (aj - 0.5 * (ar + al));
      
store:			
      va[i].a    = aj;
			va[i].ar   = ar;
			va[i].al   = al;
			va[i].del  = dela;
			va[i].a6   = a6a;
			
		}
		
	} else { //!v
		err = memErr;
	}
	//
	return (err);
#endif PPM
}

int    zlsppm_CalcCoeffsVar1DArr (V1DArr v,  Q1DArr flat){ 
   
   // combine both the above
	 //
	 // Colella and Woodward,1984, JCompPhys 54, 174 eq 1.5-1.8,1.10
	 //

	int err = noErr;
#ifndef GODUNOV  
  
	if (v.v!=NULL) {
		
#ifdef PRETESTPPMCOEFFS
    Integer i, n;
    Integer is, ie;
    
    Var1 *  va = v.v;
    
		n  = v.nt;
    
		is = 1;
		ie = v.nt-1;
    
		va [0].al  = va [0].a;
		va [0].ar  = va [0].a;

		va[ie].ar  = va[ie].a;
		va[ie].al  = va[ie].a;
    
#ifdef PPM
		va [0].del  = 0.0;
		va [0].a6   = 0.0;
		va[ie].del  = 0.0;
		va[ie].a6   = 0.0;
#endif PPM
    
    Real a     = va [is].a;
    
		Var1  v1;
		v1.a  = a;

		v1.al = a;
		v1.ar = a;
#ifdef PPM
		v1.del = 0.0;
		v1.a6  = 0.0;
#endif
		for(i = is; i < ie; i++){
			if (va [i].a != a) goto processPPM;  // only produce PPM coeffs if there are any differences between cells
			va [i]  = v1;
		}//i
    
    // if we get here we are all done with perfectly flat cells		
		return( noErr );
		
processPPM:
#endif PRETESTPPMCOEFFS
      
    zlsppm_CalcLRVar1DArr    ( v, flat); // PLM and PPM
#ifdef PPM
    zlsppm_CalcDelA6Var1DArr ( v );				
#endif PPM		
	} else { //!g !v
		err = memErr;
	}
#endif !GODUNOV  
	//
	return (err);
}


int    zlsppm_CalcMPWENOCoeffsVar1DArr (V1DArr v);
  int    zlsppm_CalcMPWENOCoeffsVar1DArr (V1DArr v){ 
#ifndef PPM			
	return(noErr);
#else
//C
//C Adaptive MP WENO upwind shock capturing 4th/3rd order conservative 
//C reconstruction with non-uniform mesh.  Sutherland 2004
//C Using three symmetrical quadratic fits to conserved integral to
//C determine left and right boundary derivatives.  
//C
//C method is a generalised form of bryson and Lee methods but for non-uniform
//C grids, applied to Collela and Woodward
//C simple quartic interpolation, but here the quartic is decomposed into
//C three quadratics, left, centre and right, which are combined with optimum 
//C weights in smooth regions to recover 4th order fit, and degrade to left or right
//C 2nd order extrapolation at discontinuities determined by smoothness estimates for
//C each quad.  Smoothness is based on sum of squares of 1st and 2nd derivatives of
//C the original data (ie 2nd and 3rd differentials of integral being fit.
//C
//C All the formulae from Osher, Jui, Balsara, Bryson etc have been rederived for 
//C random meshes, allowing for shock capturing cell boundary movements. 
//C interpolated parabolas should not need  flattening
//C
//C For more background on uniform meshes, as applied to central differnce
//C schemes (ie non-reiman) see:
//C
//C Liu, Osher, Chan 1994, J. Comp. Phys. 115, 200
//C Jiang, Shu 1996, J. Comp. Phys. 126, 202 
//C Suresh, Huynh, 1997, J. Comp. Phys. 136, 83
//C Jiang, Tadmor 1998, SAIM J. Sci. Comput. 19, 1892 
//C Balsara, Shu, 2000, J. Comp. Phys. 160, 405
//C Qui, Shu 2002, J. Comp. Phys. 183, 187
//C Bryson, Levey 2003, SIAM J. Numer. Anal. 41, 1339
//C
//C paraset sets up constants which are re-used each time we want to
//C interpolate a parabola on a quantity. First pull out constants
//C A, B, and C, and then compute all the equations in terms of those
//C quantities. 
//C
//C pc0 = dx
//C pc1 = d1
//C pc2 = c0
//C pc3 = c1
//C pc4 = c2
//C
//C
//
//C
  
  int err = noErr;

//C
//C set up conserved sum, and both dx and value difference tables
//C 
//C F = Sum ai * dxi 
//C

	Real  fm,fm2,fp,fp2;
	Real  dm,d,dp,dp2;
	Real  d1m,d1,d1p;
	Real  s0,s1,s2,sc;
	Real  w0,w1,w2,eps;
//
//C
//C monotone constraint functions and values
//C
//   const Real alpha = 2.0;
//   Real aul, alc;
//  do n = nmin-1, nmax+1
//    C
//    C
//    aul   = a(n) + minmod(alpha*(a(n)-a(n-1)), a(n+1)-a(n))
//    ar(n) = median(ar(n), a(n), aul)
//    
//    aul   = a(n) + minmod(alpha*(a(n)-a(n+1)), a(n-1)-a(n))
//    al(n) = median(al(n), a(n), aul)
//    C
//    enddo
    
    
  register Real  aj, al, ar, dela, a6a;
	Real  scrch1;
	Real  scrch2;
  
	Integer is, ie;
	Integer i, nt;
	
	ACoordPtr    x;
	
	if ((v.v!=NULL)&&(v.x!=NULL)) {
    
    Var1 *  va; 

		x   = v.x;
    va  = v.v;
		nt  = v.nt;
//
    Real  f   [nt];
    Real  dadx[nt], da2dx[nt], da2dx2[nt];

		is  =    1;
		ie  = nt-1;
		
    eps  = 1.0e-6;

    i = 0;
    d    = x[i  ].pc0;
    dp   = x[i+1].pc0;
    d1   = x[i  ].pc1;
    
    f    [i]  =  va[i  ].a*d;
    dadx [i]  = (va[i+1].a-va[i].a)/d;
    da2dx[i]  = (va[i+2].a-va[i].a)/d1;
    
    da2dx2[i] = (((d/d1)*va[i+2].a)+((dp/d1)*va[i].a)-va[i+1].a);  
    
		for(i = is; i<ie-1; i++){
      
      d    = x[i].pc0;
      dp   = x[i+1].pc0;
      d1   = x[i].pc1;
      
      f[i]      = f[i-1] + va[i].a*d;
      dadx[i]   = (va[i+1].a-va[i].a)/d;
      da2dx[i]  = (va[i+2].a-va[i].a)/d1;
      da2dx2[i] = (((d/d1)*va[i+2].a)+((dp/d1)*va[i].a)-va[i+1].a);    
      
    }

    i         = ie - 1;
    d         = x[i  ].pc0;
    dp        = x[i+1].pc0;
    f   [i]   =   f[i-1]   + va[i].a*d;
    dadx[i]   = (va[i+1].a - va[i].a)/d;
    da2dx[i]  = 0.0;
    da2dx2[i] = 0.0;    

		is = 2;
		ie = nt-2;
		
    Real  q0  [nt], q1   [nt], q2    [nt];
    
    q0[0]   = 0.0;
    q1[0]   = 0.0;
    q2[0]   = 0.0;
    q0[1]   = 0.0;
    q1[1]   = 0.0;
    q2[1]   = 0.0;

		for(i = is;i<ie;i++){

      fm2   = f[i-2]-f[i];
      fm    = f[i-1]-f[i];
      fp    = f[i+1]-f[i];
      fp2   = f[i+2]-f[i];
//
      dm    = x[i-1].pc0;
      d     = x[i  ].pc0;
      dp    = x[i+1].pc0;
      dp2   = x[i+2].pc0;
//
      d1m   = x[i-1].pc1;
      d1    = x[i  ].pc1;
      d1p   = x[i+1].pc1;
//
// smoothness from a, cs and weights  sum of squares of 1st and second numerical derivatives
// of original data - ie 2nd and 3rd derivs of integral f
//
      s0 = dp*dp * 0.250*(dadx[i  ]*dadx[i  ]+dadx[i+1]*dadx[i+1]) + (da2dx2[i  ]*da2dx2[i  ]);
      s1 =  d* d * (da2dx[i-1]*da2dx[i-1])                          + (da2dx2[i-1]*da2dx2[i-1]);
      s2 = dm*dm * 0.250*(dadx[i-2]*dadx[i-2]+dadx[i-1]*dadx[i-1]) + (da2dx2[i-2]*da2dx2[i-2]);
//
// non-linear weights - tends to optimum in smooth regions
//
      w0 = x[i  ].pc2/((eps+s0)*(eps+s0));
      w1 = x[i  ].pc3/((eps+s1)*(eps+s1));
      w2 = x[i  ].pc4/((eps+s2)*(eps+s2));
      sc = 1.0/(w0+w1+w2);
//
      w0 = w0*sc;
      w1 = w1*sc;
      w2 = w2*sc;
//
// evaluate three quads at xr
//       
      q0[i] = fp *(d1p/(dp2*dp))  - fp2*(dp /(dp2*d1p));
      q1[i] = fp *(d  /(dp *d1))  - fm *(dp /(d  *d1 ));
      q2[i] = fm2*(d  /(d1m*dm))  - fm *(d1m/(dm *d  ));
//
// construct right and left interpolated values, and parabolic coeffs
//              
      va[i].ar = q0[i]  *w0 + q1[i]  *w1 +q2[i]  *w2;
      va[i].al = q0[i-1]*w0 + q1[i-1]*w1 +q2[i-1]*w2;
//
    } // i
    
    
		for(i = is;i<ie;i++){
			
			aj =  va[i].a  ;			
      ar = zls_median( va[i].ar , va[i].a, va[i+1].a );
      al = zls_median( va[i].al , va[i].a, va[i-1].a );

      scrch1 = (ar - aj) * (aj - al) ;
			if(scrch1 < 0.0) { // local max min
				ar = al = aj;
				a6a = dela = 0.0;
			} else {
				
				dela = ar - al ;
        a6a  = 6.0* (aj - 0.5 * (ar + al)) ;
				
				scrch1 = dela * dela;
				scrch2 = dela * a6a;
				
				if(scrch1 <  scrch2) {
					al = (aj+aj+aj) - (ar+ar);
					dela  = ar  - al;
					a6a   = 6.0* (aj - 0.5 * (ar + al));
					if (zls_abs(a6a)<1.0e-8) {
						ar = al = aj;
						a6a = dela = 0.0;
					} 
					goto store;
				}
				
				if( scrch1 < -scrch2 ) {
					ar    = (aj+aj+aj) - (al+al);
					dela  = ar  - al;
					a6a   = 6.0* (aj - 0.5 * (ar + al));
					if (zls_abs(a6a)<1.0e-8) {
						ar = al = aj;
						a6a = dela = 0.0;
					}           
				} 
				
			} // not local max min
			
store: 
				
      va[i].ar   = ar;
			va[i].al   = al;
			va[i].del  = dela;
			va[i].a6   = a6a;
			
		}
		
	} else { //!g !v
		err = memErr;
	}

return (err);
#endif !PPM			
}

//
//int    zlsppm_CalcCoeffsVar2DArr  (V2DArr v, Q2DArr flat ){
//	int err = nyiErr;
//	return (err);
//}
//
//int    zlsppm_CalcCoeffsVar3DArr  (V3DArr v, Q3DArr flat ){
//	int err = nyiErr;
//	return(err);
//}
//

Real kZLS_Omega1       ; // set in zls_conf.c: zlsio_SetUpPPMFromConfigFile
Real kZLS_Omega2       ; 
Real kZLS_CWEpsilon    ; // 0.30 decrease for greater sensitivity to pressure jumps
Real kZLS_PPMFlatMin   ; // 0.0 increase for general diffusivity
Real kZLS_PPMFlatMax   ; // 1.0 decrease for limiting max diffusivity
Real kZLS_PPMFlatDd    ;// dd threshold for continuing calculations
Real kZLS_PPMFlatDP    ;  // dP threshold for continuing calculations
Real kZLS_PPMFlatDv    ; // dv threshold for continuing calculations

int    zlsppm_CalcFlatteningW1DArr( W1DArr w ){ // check for places to reduce order
	
	int err = noErr;
	
	Integer i, is, ie, ni, nt;
	
	 register Real s, t, u;
   Real delP, delP2, delV, omegaJ;
   
   is  = 1;
   ie  = w.nt-1;
   ni  = w.n ;
   nt  = w.nt;
   
   Real steep[nt];
   
   bzero((void *) steep, nt*kElementSize);
 	
   for ( i = is; i < ie; i++ ) {
     
     delP =  (w.P[i+1].a - w.P[i-1].a);
     delV =  (w.vx[i-1].a - w.vx[i+1].a);
     
     if (zls_abs(delP) > (kZLS_PPMFlatDP*w.P[i].a)){
     
     s   = zls_abs(delP)/(zls_min( w.P[i+1].a, w.P[i-1].a )); // always +ve
     s   = (s > kZLS_CWEpsilon)?1.0:0.0; // needs to be large enough, 1 if s >kZLS_CWEpsilon
     if ( s > 0.0) {
       
       u   = (delV> kZLS_PPMFlatDv)?1.0:0.0; // +1 for a potential shock, 0 if not
       if ( u > 0.0){
         
         delP2    = ((w.P[i+2].a - w.P[i-2].a));
         omegaJ   = ((delP/delP2) - kZLS_Omega1) * kZLS_Omega2;
         omegaJ   = zls_max(0.0, omegaJ);
         
         steep[i] = ( omegaJ > kZLS_PPMFlatMax)?kZLS_PPMFlatMax: omegaJ;
         
       }; // s*u
       
     };
     };
     
   };
   
   for ( i = is; i < ie; i++ ) {
      t         = zls_max3(steep[i-1]*0.5, steep[i], steep[i+1]*0.5);
      w.flat[i] = zls_max( kZLS_PPMFlatMin, t);
#ifdef DIVVHCORRECTION
#if (NDIMS > 1)
        t = w.divv[i];
        t = (t>kZLS_DivVThreshold)?t*kZLS_DivVInvScale:0.0;
        t = (t>1.0)?1.0:t;
        w.flat[i] = zls_max(t,w.flat[i]);
#endif
#endif DIVVHCORRECTION
    if ( zlst_GlobalCycles < (NDIMS) ) w.flat[i] = kZLS_PPMFlatMax;
   }
   
   w.flat[0]   = kZLS_PPMFlatMax;
   w.flat[1]   = kZLS_PPMFlatMax;

   w.flat[nt-1] = kZLS_PPMFlatMax;
   w.flat[nt-2] = kZLS_PPMFlatMax;

   return(err);
}


//========================================
//Low level LRQArr calls for when al and ar require grid interpolation due to low order input
// cell data.
//========================================

int    zlsppm_CalcLRQ1DArr  ( LRQ1DArr va, GCoordPtr grid, Q1DArr flat){
#ifdef GODUNOV			
	return(noErr);
#else 			
//
// for LR data with a but *without pre-given LR values*, interps on grid same as PPM, needs coordinates
// as v doesn't contain that infor like a full VArr does. Assumes v have average values in place, only!
// this is not the usual form of LRQArr data so this call will be less common.  Mostly will occur
// is the input data is a simle table of average values and l/r values are not available.
// ie a numerical gravity field read in.  If the data is based on a function routine then l and r can 
// be evaluated without interpolation here and this routine is not needed.
//  
	int err = noErr;
	
	Integer i,n;
	
	Real  fl, onemfl;
	
	Integer is, ie;
  ACoordPtr x;
		
	if ((va != NULL)&& (grid!=NULL)) {
		
		is = 1;
		ie = grid->nt[kXDim]-1;
		n  = grid->nt[kXDim];
		x  = grid->x;
    
    Real diffa[n];
		
		//
		for(i = 0; i < ie; i++){
			diffa[i] =  va[i+1].a - va[i].a;
		}//i
		
    Real da[n];
		//
		for (i = is; i < ie; i++){
			if(diffa[i-1]*diffa[i] < 0.0) {
				da[i] = 0.0;
			} else {
				da[i] = x[i].pc3 * diffa[i] +  x[i].pc4* diffa[i-1];
				da[i] = zls_sign1(da[i])* zls_min3(zls_abs(da[i]), 2.0*zls_abs(diffa[i-1]), 2.0*zls_abs(diffa[i]));
			}
		} //i
    
		for(i = is ; i < ie ; i++){
			va[i].ar   = va[i].a +  x[i].pc0*diffa[i] +  x[i].pc1*da[i+1] +  x[i].pc2*da[i];
			va[i+1].al = va[i].ar;
		} //i
		
		
		if( flat !=NULL) {
			for(i = is ; i <= ie ;i++){
				fl       = flat[i];
				onemfl   = 1.0 - fl   ;
				va[i].ar = fl * va[i].a + onemfl * va[i].ar;
				va[i].al = fl * va[i].a + onemfl * va[i].al;
			} //i
		} // 
		
		
	} else { // !v
		err = memErr;
	}
	//
	return (err);
#endif GODUNOV			
}

//
//
//int    zlsppm_CalcLRQ2DArr  ( LRQ2DArr va, GCoordPtr grid, Q2DArr flat){
//	int err = nyiErr;
//	return (err);
//}
//
//int    zlsppm_CalcLRQ3DArr  ( LRQ3DArr va, GCoordPtr grid, Q3DArr flat){
//	int err = nyiErr;
//	return(err);
//}
//
//========================================

// for LR data with pre-given LR values, and cental values rather than means.
// If FIELD_RECON_ORDER 2 is used the (a) is computed from a_c and al and ar.  
// FIELD_RECON_ORDER 1 is used (a) is computed form al and ar only.
// Simliar to the F1DArr call, but can be used on the underlying zoneField data
// So when copied to and F1DArr, conversion is not needed every time.  Baiscly an option to 
// do the conversion at a lower level, if it is not done in the original intialisation.
//  nti is needed as va has no size info.  but we are not
// interpolating on a grid so the grid info and flatten are not needed.

int  zlsppm_CalcMeanLRQ1DArr  ( LRQ1DArr va, Integer nti){
  int err = noErr;
      // PPM and PLM etc apply to F1DArrs, FIELD_RECON_ORDER applies to LRQ1DArrs

  if ( va != NULL) {
    Real ar = 0.0;
    Real al = 0.0;
		Integer is = 0;
		Integer ie = nti;
		Integer i;
#if (FIELD_RECON_ORDER == 0)
    Real aj = 0.0; // here aj must be the central value from a single evaluation and the cell average.
    for ( i = is; i < ie; i++){
      aj =  va[i].a  ; 
      va[i].ar = aj;
      va[i].al = aj;
    }
    //
    // Not in conservative form now (ie edges are not integral related).
    // If a F1DArr is made from this field then 
    // It will need to use grid interpolation to get valid l/r values
    // Call zlsppm_CalcLRQ1DArr on this LRQ1DArr with additional grid information
    // first, or let F1DArr do that calc with 
#endif
#if (FIELD_RECON_ORDER == 1)
    for ( i = is; i < ie; i++){
      ar =  va[i].ar ;
      al =  va[i].al ;
      va[i].a = 0.5 * (ar + al) ; // compute <a> as average integral across cell, conservative form. 
    }
#endif
#if (FIELD_RECON_ORDER == 2)
    Real aj = 0.0; // here aj is the central value *not* the cell average.
    for ( i = is; i < ie; i++){
      aj =  va[i].a  ; // here aj is the central value *not* the cell average.
      ar =  va[i].ar ;
      al =  va[i].al ;
      va[i].a = kZLS_sixth * ( 4.0*aj - 0.5 * (ar + al)) ; // compute <a> from a central evaluation of a at x.c
    }
#endif
  } else { err = memErr;}
  return(err);
}

//int  zlsppm_CalcMeanLRQ2DArr  ( LRQ2DArr va, Integer ntj, Integer nti){
//	int err = nyiErr;
//	return(err);
//}
//int  zlsppm_CalcMeanLRQ3DArr  ( LRQ3DArr va, Integer ntk, Integer ntj, Integer nti){
//	int err = nyiErr;
//	return(err);
//}
//========================================
//========================================
// Higher level calls for F1DArrs, ie from Fields, with the twist that the l and r
// values are integral and sometimes the central or mean value need to be computed.
// with PPM, F1DArrs have del and a6  elements, but not for PLM or GODUNOV.  Wheras 
// LRQArrs always have just a and l and r.
//========================================

int  zlsppm_CalcDelA6F1DArr  ( F1DArr f ){
#ifndef PPM
//  a6 and del are only used in the PPM interpolation method
  // PLM uses just al and ar, while GODUNOV just a.
  return(noErr);
#else
	int err = noErr;
	
	Integer i;
	
	register Real  aj, al, ar, dela, a6a;
	
	Integer is, ie;
	
	//
	//========================================
	// Compute Delta and a6 if the L/R values are known, and required to be kept
	// unchanged.  In this case we don't fiddle the parabolic monotone limits.
  // but this is OK for fields, if not hyperbolic system variables. :)
	
	if (f.xi!=NULL) {
		
		
		is = 0;
		ie = f.nt;
		
		for(i = is;i<ie;i++){
			
			al =  f.xi[i].al ;
			aj =  f.xi[i].a  ;
			ar =  f.xi[i].ar ;
			
			dela = ar - al ;
			a6a  = 6.0* (aj - 0.5 * (ar + al)) ; // standard formula given a is (a)
			
			f.xi[i].al    = al;
			f.xi[i].a     = aj;
			f.xi[i].ar    = ar;
			
			f.xi[i].del  = dela;
			f.xi[i].a6   = a6a;
		}
		
	} else { //!va
		err = memErr;
	}
	//
	return (err);
#endif PPM
}

int  zlsppm_CalcMeanF1DArr  (F1DArr f){
#ifdef GODUNOV
// GODUNOV has just a, so no point really
  return(noErr);
#else
	int err = noErr;
	
	Integer i;
	
#ifdef PPM
register Real  aj, al, ar, dela, a6a;
#endif PPM
#ifdef PLM
register Real  aj, al, ar;
#endif PLM

	Integer is, ie;
//
//========================================
// Compute (a) Delta and a6 if the given a is actually the central cell value at
// x_c and not the cell average.  Assuming a PPM fit the equivalent average (a) can 
// be derived from a_c, al, and ar, just as a_c can come from the other three.
//  
//  a6  = 6*[a - 1/2(al + ar)] (standard formula
//  a   = al + del/2 + a6/6  ( standard formula
//
//  a in terms of a_c :
//
//  a  = 1/6*[4a_c + al + ar]
//
//  if we have a_c not a, plus al and ar are pre-determined:
//
//  a6  = 4*[a_c - 1/2(al + ar) ] ( complementary formula
//  a_c = al + del/2 + a6/4 ( complementary formulae
//  
//  a_c in terms of a :
//
//  a_c  = 1/4*[6a - al - ar]
//
// As can be seen, working with a_c or a is similar, however we choose to work with 
// a as it is a conservative quantity based on an integral property of the cell
// so if for some reason we ge a_c values and exact al and ar values then we 
// change a_c to a.   This can happen when a grid is filled with just a central and edge 
// evaluations of a function, espcailly in 1D, ie a gravity potential without a convenient 
// integral or differntial form.
//
// IMPORTANT L/R values are fundamental here, are known, and required to be kept
// unchanged, as they are exact and the basis of the integrals.  This is subtly differnt to 
// the lr values from grid cell-cell interpolated values of al and ar.
// In this case we don't fiddle the parabolic monotone limits.
// Since al and ar are not free parameters.  This better be used for well behaves fields!
//
	if (f.xi!=NULL) {
		
		
		is = 0;
		ie = f.nt;
		
		for(i = is;i<ie;i++){
			
			aj =  f.xi[i].a  ; // here aj is the central value *not* the cell average.
			ar =  f.xi[i].ar ;
			al =  f.xi[i].al ;
      // PPM and PLM etc apply to F1DArrs, FIELD_RECON_ORDER applies to LRQ1DArrs
#ifdef PPM
      aj   = kZLS_sixth * ( 4.0*aj - 0.5 * (ar + al)) ; // compute <a> from a central evaluation of a at x.c
      a6a  = 6.0 * (aj - 0.5 * (ar + al)) ;
			dela = ar - al ;
			
			f.xi[i].a    = aj;
			f.xi[i].del  = dela;
			f.xi[i].a6   = a6a;
#endif PPM
#ifdef PLM
      aj           =  0.5 * (ar + al) ; // compute <a> as average integral across cell, conservative form. 
			f.xi[i].a    = aj;
#endif PLM
		}
		
	} else { //!va
		err = memErr;
	}
	//
#endif !GODUNOV
	return (err);
}

int    zlsppm_CalcCoeffsF1DArr ( F1DArr f ){
	
	int err = noErr;
	
	V1DArr  v;
	
	v.n   = f.n;
	v.ns  = f.ns;
	v.ne  = f.ne;
	v.nt  = f.nt;
	if (f.x!=NULL) {
		v.x   = f.x;

		v.v   = f.xi;
		zlsppm_CalcCoeffsVar1DArr( v, NULL );
    
	} else { //!f !x
		err = memErr;
	}
	return (err);
}

// If computeLR == kZLS_ComputeLR then the LRQArr is based on 0 order data
// and the left and right boundaries need to be computed before the other
// coefficients.  In GODUNOV this isn't needed, but will be for PLM and PPM
// modes with 0 order LRQArr data.
// If the LRQArr data already has LR data values and computeLR == kZLS_DontComputeLR
// then there is nothing to do for PLM, but del and a6 are still needed for PPM.

//int    zlsppm_CalcCoeffsF2DArr ( F2DArr f ){
//	
//	int err = nyiErr;
//	return (err);
//}
//
//int    zlsppm_CalcCoeffsF3DArr ( F3DArr f ){
//	
//	int err = nyiErr;
//	return (err);
//}


//========================================
//  Here are routines for going through higher level structures according to dimension
// and other code options, and these call the lower level coefficient routines for their
// components.  These are the ones most often called on higher level simulation structures
// such as primitive variable arrays.
//========================================

int    zlsppm_CalcAllCoeffsW1DArr (W1DArr w,  int iflat ){
	
	int err = noErr;
	
	V1DArr  v;
	
	v.n   = w.n;
	v.ns  = w.ns;
	v.ne  = w.ne;
	v.nt  = w.nt;
	if (w.x!=NULL) {
  
  if ( iflat ) 		err = zlsppm_CalcFlatteningW1DArr ( w ); // check for places to reduce order

		v.x   = w.x;
		if (w.d!=NULL) {
			v.v   = w.d;
			if (iflat) {
				zlsppm_CalcCoeffsVar1DArr( v, w.flat );
			} else {
				zlsppm_CalcCoeffsVar1DArr( v, NULL );
			}
		} else { //!w !x
			err = memErr;
		}
		if (w.P!=NULL) {
			v.v   = w.P;
			if (iflat) {
				zlsppm_CalcCoeffsVar1DArr( v, w.flat );
			} else {
				zlsppm_CalcCoeffsVar1DArr( v, NULL );
			}
		} else { //!w !x
			err = memErr;
		}
		if (w.E!=NULL) {
			v.v   = w.E;
			if (iflat) {
				zlsppm_CalcCoeffsVar1DArr( v, w.flat );
			} else {
				zlsppm_CalcCoeffsVar1DArr( v, NULL );
			}
		} else { //!w !x
			err = memErr;
		}
		
#ifdef SYM1D
		if (w.vx!=NULL) {
			v.v   = w.vx;
			if (iflat) {
				zlsppm_CalcCoeffsVar1DArr( v, w.flat );
			} else {
				zlsppm_CalcCoeffsVar1DArr( v, NULL );
			}
		} else { //!w !x
			err = memErr;
		}
#endif SYM1D
#ifdef SYM2D
		if (w.vx!=NULL) {
			v.v   = w.vx;
			if (iflat) {
				zlsppm_CalcCoeffsVar1DArr( v, w.flat );
			} else {
				zlsppm_CalcCoeffsVar1DArr( v, NULL );
			}
		} else { //!w !x
			err = memErr;
		}
		if (w.vy!=NULL) {
			v.v   = w.vy;
			if (iflat) {
				zlsppm_CalcCoeffsVar1DArr( v, w.flat );
			} else {
				zlsppm_CalcCoeffsVar1DArr( v, NULL );
			}
		} else { //!w !x
			err = memErr;
		}
		
#endif SYM2D
		
#ifdef SYM3D
		if (w.vx!=NULL) {
			v.v   = w.vx;
			if (iflat) {
				zlsppm_CalcCoeffsVar1DArr( v, w.flat );
			} else {
				zlsppm_CalcCoeffsVar1DArr( v, NULL );
			}
		} else { //!w !x
			err = memErr;
		}
		if (w.vy!=NULL) {
			v.v   = w.vy;
			if (iflat) {
				zlsppm_CalcCoeffsVar1DArr( v, w.flat );
			} else {
				zlsppm_CalcCoeffsVar1DArr( v, NULL );
			}
		} else { //!w !x
			err = memErr;
		}
		if (w.vz!=NULL) {
			v.v   = w.vz;
			if (iflat) {
				zlsppm_CalcCoeffsVar1DArr( v, w.flat );
			} else {
				zlsppm_CalcCoeffsVar1DArr( v, NULL );
			}
		} else { //!w !x
			err = memErr;
		}
#endif SYM3D
		
		
#if (NTRACERS > 0)    
		Integer i;
		for (i = 0; i < kNTracers; i++){
			v.v   = w.tr[i];
			if (iflat) {
				zlsppm_CalcCoeffsVar1DArr( v, w.flat );
			} else {
				zlsppm_CalcCoeffsVar1DArr( v, NULL );
			}
		}
#endif
	} else { //!w !x
		err = memErr;
	}
	return (err);
}


int    zlsppm_CalcCoeffsW1DArr ( W1DArr w ){
	
	int err = noErr;
	
	V1DArr  v;
	
	v.n   = w.n;
	v.ns  = w.ns;
	v.ne  = w.ne;
	v.nt  = w.nt;
	if (w.x!=NULL) {
    
	  if (w.d!=NULL) {
      v.v   = w.d;
      zlsppm_CalcCoeffsVar1DArr( v, NULL );
    } else { err = memErr;  return(err);}
    
	  if (w.E!=NULL) {
      v.v   = w.E;
      zlsppm_CalcCoeffsVar1DArr( v, NULL );
    } else { err = memErr;  return(err);}
    
    if (w.P!=NULL) {
      v.v   = w.P;
      zlsppm_CalcCoeffsVar1DArr( v, NULL );
    } else { err = memErr;  return(err);}
    
#ifdef SYM1D
    if (w.vx!=NULL) {
      v.v   = w.vx;
      zlsppm_CalcCoeffsVar1DArr( v, NULL );
    } else { err = memErr;  return(err);}
#endif SYM1D
#ifdef SYM2D
    if (w.vx!=NULL) {
      v.v   = w.vx;
      zlsppm_CalcCoeffsVar1DArr( v, NULL );
    } else { err = memErr;  return(err);}
    if (w.vy!=NULL) {
      v.v   = w.vy;
      zlsppm_CalcCoeffsVar1DArr( v, NULL );
    } else { err = memErr;  return(err);}
#endif SYM2D
		
#ifdef SYM3D
    if (w.vx!=NULL) {
      v.v   = w.vx;
      zlsppm_CalcCoeffsVar1DArr( v, NULL );
    } else { err = memErr;  return(err);}
    if (w.vy!=NULL) {
      v.v   = w.vy;
      zlsppm_CalcCoeffsVar1DArr( v, NULL );
    } else { err = memErr;  return(err);}
    if (w.vz!=NULL) {
      v.v   = w.vz;
      zlsppm_CalcCoeffsVar1DArr( v, NULL );
    } else { err = memErr;  return(err);}
#endif SYM3D
		
		
#if (NTRACERS > 0)    
		Integer i;
		for (i = 0; i < kNTracers; i++){
      if (w.tr[i]!=NULL) {
        v.v   = w.tr[i];
        zlsppm_CalcCoeffsVar1DArr( v, NULL );
      } else { err = memErr;  return(err);}
		}
#endif
	} else { //!w !x
		err = memErr;
	}
	return (err);
}

int    zlsppm_CalcRemappingCoeffsW1DArr ( W1DArr w ){
	//
  // advection scheme
  //
	int err = noErr;
	
	V1DArr  v;
	
	v.n   = w.n;
	v.ns  = w.ns;
	v.ne  = w.ne;
	v.nt  = w.nt;
	if (w.x!=NULL) {
		v.x   = w.x;
		v.v   = w.d;
		zlsppm_CalcCoeffsVar1DArr( v, NULL );
		v.v   = w.P;
		zlsppm_CalcCoeffsVar1DArr( v, NULL );
#ifdef SYM1D
		v.v   = w.vx;
		zlsppm_CalcCoeffsVar1DArr( v, NULL );
#endif SYM1D
#ifdef SYM2D
		v.v   = w.vx;
		zlsppm_CalcCoeffsVar1DArr( v, NULL );
		v.v   = w.vy;
		zlsppm_CalcCoeffsVar1DArr( v, NULL );
#endif SYM2D
		
#ifdef SYM3D
		v.v   = w.vx;
		zlsppm_CalcCoeffsVar1DArr( v, NULL );
		v.v   = w.vy;
		zlsppm_CalcCoeffsVar1DArr( v, NULL );
		v.v   = w.vz;
		zlsppm_CalcCoeffsVar1DArr( v, NULL );
#endif SYM3D
		
		
#if (NTRACERS > 0)    
		Integer i;
		for (i = 0; i < kNTracers; i++){
			v.v   = w.tr[i];
			zlsppm_CalcCoeffsVar1DArr( v, NULL );
		}
#endif
	} else { //!w !x
		err = memErr;
	}
	return (err);
}

int    zlsppm_CalcFlatHydroCoeffsW1DArr (W1DArr w){
	
	int err = noErr;
	
	V1DArr  v;
	
	v.n   = w.n;
	v.ns  = w.ns;
	v.ne  = w.ne;
	v.nt  = w.nt;
	if (w.x!=NULL) {
		v.x   = w.x;
		err = zlsppm_CalcFlatteningW1DArr       ( w ); // check for places to reduce order
		v.v   = w.d;
		err += zlsppm_CalcCoeffsVar1DArr( v, w.flat );
		v.v   = w.P;
		err += zlsppm_CalcCoeffsVar1DArr( v, w.flat );
		v.v   = w.vx;
		err += zlsppm_CalcCoeffsVar1DArr( v, w.flat );
  } else { //!w !x
		err = memErr;
	}
	return (err);
}



//
//int    zlsppm_CalcCoeffsQ         (CPtr c,
//                                   QPtr q,  // q and c must match dimesions exactly
//                                   GCoordPtr g, 
//                                   int iflat, int iskip ){
//	return(nyiErr); // not yet implemented
//}
//
//int    zlsppm_CalcCoeffsQ1DArr     (CPtr c,
//                                    Q1DArr q,  // q and c must match dimesions exactly
//                                    GCoordPtr g, 
//                                    int iflat, int iskip ){
//	return(nyiErr);
//}
//
//
//int    zlsppm_CalcCoeffsQ2DArr     (C2DArr c,
//                                    Q2DArr q,  // q and c must match dimesions exactly
//                                    GCoordPtr g,
//                                    int iflat, int iskip ){
//	return(nyiErr);
//}
//
//int    zlsppm_CalcCoeffsQ3DArr     (C3DArr c,
//                                    Q3DArr q,  // q and c must match dimesions exactly
//                                    GCoordPtr g,
//                                    int iflat, int iskip ){
//	return(nyiErr);
//}
//

int zlsppm_ReconstructPPMWLRStates1D( W1DArr w, Real dt, Real g) { 
	// g is a gemoetric projection factor = 1.0 for all cartesian axes.  
  // Scales dx by 1/g in angular axes, g is constant for a row, so hdt/g is performed outside loop
  
	int err = noErr;
	
	if ((w.cs!=NULL)&&(w.d!=NULL)&&(w.P!=NULL)&&(w.vx!=NULL)&&(w.x!=NULL)) {
		
		err = zlsp_CalcWSoundSpeed1D      ( w );

#ifdef PPM
    err = zlsppm_CalcFlatHydroCoeffsW1DArr ( w );
#endif PPM
		
#ifdef PLM
		err = zlsppm_CalcFlatHydroCoeffsW1DArr ( w );
#endif PLM
		
		
#ifndef GODUNOV
		//
		// if godunov reconstruction is used, no interpolation is needed as the left and right
		// values are copied on read in.
		//
		Integer       i;
		Integer  ni, nt,  is, ie;
		register Real  hdt, f, cf;
		ACoordPtr       x;
		 
		ni    = w.n ;
		nt    = w.nt;
		is    = 1;
		ie    = nt - 1;
		x     = w.x ;
		
		Real       rateCs = 0.0;
		Real       Cdtdx[nt];
#ifdef PPM
		Real      fCdtdx[nt];
#endif PPM

		
		hdt = (g != 1.0)?0.5*dt/g:0.5*dt;
		
		for (i= is; i < ie; i++){
      
      Real  cs = w.cs[i  ].a;
      
      rateCs  = hdt*cs*x[i  ].idx;
      
#ifdef PPM
			fCdtdx[i  ] = 1.0 - fourthd*rateCs;
#endif PPM
			Cdtdx [i  ] = rateCs; 
      

		}
    
		Real Pl, Pr;
		Real dl, dr;
		Real vl, vr;
		
		for (i= is; i < ie; i++){
			
#ifdef PLM			
			// Here right refers to the right edge of the cell i
			f = Cdtdx[i];
      cf = 1.0- f;
      
			Pr =  zls_max(kZLS_PressureVacuum, cf*w.P[i].ar + f*w.P[i].al);
			dr =  zls_max(kZLS_Small,cf*w.d[i].ar + f*w.d[i].al);
			vr =  cf*w.vx[i].ar + f*w.vx[i].al;

			w.P [i].ar =  Pr;
			w.d [i].ar =  dr;
			w.vx[i].ar =  zls_tolx(vr, kZLS_Small);
			
			// Here left refers to the left edge of the cell i
			Pl   =  zls_max(kZLS_PressureVacuum, cf*w.P[i].al + f*w.P[i].ar);
			dl   =  zls_max(kZLS_Small, cf*w.d[i].al + f*w.d[i].ar);
			vl   =  cf*w.vx[i].al + f*w.vx[i].ar;
			
			w.P [i].al   =  Pl;
			w.d [i].al   =  dl;
			w.vx[i].al   =  zls_tolx(vl, kZLS_Small);
			
#endif PLM
			
#ifdef PPM			
			// Here right refers to the right edge of the cell i
			f   = Cdtdx[i];
      cf  = fCdtdx[i];
      
			Pr  = zls_max(kZLS_PressureVacuum, w.P[i].ar - f*( w.P[i].del-cf* w.P[i].a6));
			dr  = zls_max(kZLS_Small, w.d[i].ar - f*( w.d[i].del-cf* w.d[i].a6));
			vr  = w.vx[i].ar - f*(w.vx[i].del-cf*w.vx[i].a6);
			
			w.P [i].ar =  Pr;
			w.d [i].ar =  dr;
			w.vx[i].ar =  zls_tolx(vr, kZLS_Small);
			
			// Here left refers to the left edge of the cell i
			Pl   =  zls_max(kZLS_PressureVacuum, w.P[i].al + f*( w.P[i].del+cf* w.P[i].a6));
			dl   =  zls_max(kZLS_Small, w.d[i].al +f*( w.d[i].del+cf* w.d[i].a6));
			vl   =  w.vx[i].al +f*(w.vx[i].del+cf*w.vx[i].a6);
			
			w.P [i].al   =  Pl;
			w.d [i].al   =  dl;
			w.vx[i].al   =  zls_tolx(vl, kZLS_Small);
			
#endif PPM
		} // i
		
		err = zlsp_CalcCellWLRSoundSpeed1D ( w );
#endif !GODUNOV
		
	} else { err = memErr;}
	
	return(err);
}

int zlsppm_ReconstructPPMWLRStates1DF( W1DArr w, F1DArr phi, Real dt, Real g) { 
#ifndef GRAVITY
#pragma unused(phi)
#endif  
	int err = noErr;
	
	if ((w.cs!=NULL)&&(w.d!=NULL)&&(w.P!=NULL)&&(w.vx!=NULL)&&(w.x!=NULL)) {
		
#ifdef PPM
    err = zlsppm_CalcFlatHydroCoeffsW1DArr ( w );
#endif PPM
		
#ifdef PLM
		err = zlsppm_CalcFlatHydroCoeffsW1DArr ( w );
#endif PLM
		
		err = zlsp_CalcWSoundSpeed1D      ( w );
		
#ifndef GODUNOV
		//
		// if godunov reconstruction is used, no interpolation is needed as the left and right
		// values are copied on read in.
		//
		register Integer       i;
		register Real       rateC0; 
		register Real  hdt, f, cf;
    
		Integer  ni, nt,  is, ie;
		ACoordPtr       x;
#ifdef GRAVITY
	  Real 	accel = 0.0;
#endif GRAVITY
    
		ni    = w.n ;
		nt    = w.nt;
		is    = 0;
		ie    = nt;
		x     = w.x ;
		
		Real       Cdtdx[nt];
#ifdef PPM
		Real      fCdtdx[nt];
#endif PPM
		
		hdt      = (g != 1.0)?0.5*dt/g:0.5*dt;
				
		for (i= is; i < ie; i++){
      rateC0  = hdt*w.cs[i  ].a;
      rateC0  *= x[i  ].idx;
			Cdtdx [i  ] = rateC0; 
#ifdef PPM
			fCdtdx[i  ] = 1.0 - fourthd*rateC0;
#endif PPM
		}
    
    
		Real Pl, Pr;
		Real dl, dr;
		Real vl, vr;

#ifdef GRAVITY
		Real dv, pf;
#endif GRAVITY
		
		for (i= is; i < ie; i++){
			
#ifdef PLM			
			// Here right refers to the right edge of the cell i
      f = Cdtdx[i];
      cf = 1.0-f;
			
			Pr =  zls_max(kZLS_PressureVacuum, cf*w.P[i].ar +f*w.P[i].al);
			dr =  zls_max(kZLS_Small, cf*w.d[i].ar + f*w.d[i].al);
			vr =  cf*w.vx[i].ar + f*w.vx[i].al;

#ifdef GRAVITY
      accel  = 0.0;
      if (phi.xi != NULL) accel = -(phi.xi[i].ar-phi.xi[i].al)*x[i].idx;
      dv  =  hdt*accel;
      vr += dv; 
#endif GRAVITY
      
			w.P [i].ar =  Pr;
			w.d [i].ar =  dr;
			w.vx[i].ar =  zls_tolx(vr, kZLS_Small);
			
			// Here left refers to the left edge of the cell i
			Pl   =  zls_max(kZLS_PressureVacuum, cf*w.P[i].al + f*w.P[i].ar);
			dl   =  zls_max(kZLS_Small, cf*w.d[i].al +f*w.d[i].ar);
			vl   =  cf*w.vx[i].al + f*w.vx[i].ar;

#ifdef GRAVITY
      vl +=  dv; 
#endif GRAVITY
			
			w.P [i].al   =  Pl;
			w.d [i].al   =  dl;
			w.vx[i].al   =  zls_tolx(vl, kZLS_Small);
			
#endif PLM
			
#ifdef PPM			
			// Here right refers to the right edge of the cell i
      f  = Cdtdx[i];
      cf = fCdtdx[i];
      
			Pr =  zls_max(kZLS_PressureVacuum, w.P[i].ar - f*( w.P[i].del-cf*w.P[i].a6));
			dr =  zls_max(kZLS_Small, w.d[i].ar - f*( w.d[i].del-cf* w.d[i].a6));
			vr =  w.vx[i].ar - f*(w.vx[i].del-cf*w.vx[i].a6);

#ifdef GRAVITY
      accel  = 0.0;
      if (phi.xi != NULL) {
        pf = 1.0-f;
        accel = -(phi.xi[i].del-pf*phi.xi[i].a6)*x[i].idx;
      }
      dv  =  hdt*accel; 
//      if ( ( i >= 129) && ( i <= 134)) {
//        zlsio_Printf("dVright: %d %g %g \n", i, dv, vr);
//      }
      vr +=  dv; 
#endif GRAVITY
      
			w.P [i].ar =  Pr;
			w.d [i].ar =  dr;
			w.vx[i].ar =  zls_tolx(vr, kZLS_Small);;
			
			// Here left refers to the left edge of the cell i
			Pl   =  zls_max(kZLS_PressureVacuum, w.P[i].al + f*( w.P[i].del+cf*w.P[i].a6));
			dl   =  zls_max(kZLS_Small, w.d[i].al + f*( w.d[i].del+cf* w.d[i].a6));
			vl   =  w.vx[i].al + f*(w.vx[i].del+cf*w.vx[i].a6);

#ifdef GRAVITY
      accel  = 0.0;
      if (phi.xi != NULL) { 
        pf = 1.0-f;          
        accel = -(phi.xi[i].del+pf*phi.xi[i].a6)*x[i].idx;
      }
      dv  =  hdt*accel; 
//      if ( ( i >= 129) && ( i <= 134)) {
//        zlsio_Printf("dVleft: %d %g %g \n", i, dv, vl);
//      }
      vl +=  dv; 
#endif GRAVITY
			
			w.P [i].al   =  Pl;
			w.d [i].al   =  dl;
			w.vx[i].al   =  zls_tolx(vl, kZLS_Small);
			
#endif PPM

		} // i
		
		err = zlsp_CalcCellWLRSoundSpeed1D ( w );
#endif !GODUNOV
		
	} else { err = memErr;}
	
	return(err);
}


//
// compute v1 at left coordinates in x1, interpolating v0, with values at x0 using PPM coeffs in v0
// works without a g factor because interp fractions are ratios of x1/x0 not absolute coordinates.
// left boundary of i = 0 and right boundary of i = ni-1 are assumed to not move and are
// just copied. Assumes courant condition and xl is always in cell i or i-1 only.
//
// can't do cell centre interp with this routine as the centre may move to i-1, i or i+1,
// use the slower LRC routine for that case. (nyiErr)
//

int zlsppm_InterpolateLRPPM1D( V1DArr v1, V1DArr v0){
  int err = noErr;
  
  Real   deltaX, f;  // x1-x0 and deltaX/dx (fraction)
  Integer        i; // loop over 1 - (ni-1)
  Integer       im; // i -1 
  
  //  interp value = v0.al + f*(v0.del + (1-f)v0.a6), 
  //  set all als in v1, then compute ars by shuffles and then a, del and a6 for v1 by normal formulae
  
  Integer  nt, is, ie;
  
  ACoordPtr  x0 = v0.x;
  ACoordPtr  x1 = v1.x;
  
  Real  al,  del,  a6; // PPM coeffs for v0 cell i
  Real alm, delm, a6m; // PPM coeffs for v0 cell i-1
  
  Real al1; // interpolated left value for v1 cell i
  Real al1m; // interpolated left value v1 cell i-1
  Real invDX; // 1/dx0 in cell i
  Real invDXm; //  1/dx0 in cell i -1
  
  nt    = v0.nt;
  is    = 1;
  ie    = nt-1;
  
  v1.v[ 0].al = v0.v[ 0].al; // sent end values
  v1.v[ie].ar = v0.v[ie].ar;

  im = 0;
  
  alm  = v0.v[0].al;
#ifdef PLM
  delm = v0.v[0].ar-v0.v[0].al;
  a6m  = 0.0;
#endif PLM
#ifdef PPM
  delm = v0.v[0].del;
  a6m  = v0.v[0].a6;
#endif PPM
  invDXm = x0[0].idx;
  al1m = alm;

  // first interpolate all left values, and copy to right of i-1 on the fly
  
  for (i= is; i < ie; i++){
    
    al  = v0.v[i].al;
#ifdef PLM
    del = v0.v[i].ar-v0.v[i].al;
    a6  = 0.0;
#endif PLM
#ifdef PPM
    del = v0.v[i].del;
    a6  = v0.v[i].a6;
#endif PPM
    
    al1 = al; // default if the coords are the same
    
    invDX= x0[i].idx;
   
    deltaX = x0[i].l-x1[i].l;

#ifdef PLM
    if ( deltaX > kZLS_Small ){ // moved to the right, stayed in i
      f = deltaX*invDX;
      al1 = al + f*del;
    } else if ( deltaX < -kZLS_Small ){ // moved to the left into i - 1
      f = 1.0 + (deltaX*invDXm); // deltaX is -ve
      al1 = alm + f*delm;
    } 
#endif PLM
#ifdef PPM
    if ( deltaX > kZLS_Small ){ // moved to the right, stayed in i
      f = deltaX*invDX;
      al1 = al + f*(del+(1.0-f)*a6);
    } else if ( deltaX < -kZLS_Small ){ // moved to the left into i - 1
      f = 1.0 + (deltaX*invDXm); // deltaX is -ve
      al1 = alm + f*(delm+(1.0-f)*a6m);
    } 
#endif PPM
    
    v1.v[i ].al = al1;
    v1.v[im].ar = al1; // equals right of i-1 cell
#ifdef PPM
    v1.v[im].del = al1-al1m; // del of prev cell is now defined
#endif PPM
    al1m = al1; // keep current al for next cycle
    
    alm  = al;
    delm = del;
    a6m  = a6;
    im   = i;
    invDXm = invDX;
    
  }
  
 // finally fix del for last cell
#ifdef PPM
  v1.v[ie].del = v1.v[ie].ar - v1.v[ie].al;  // del of last cell is now defined
#endif PPM

  // without additional central or average values for v1, only the boundaries 
  // can be evaluated here.  This is a fundamental limitation.  After this call
  // only al, ar and del values in v1 are valid.
  
  return(err);
}


//========================================
//                     
// calculating variable PPM interp coeffs - need zlsppm_CalcCoeffs to be called first
// Kept here to make more generally available for non-hydro applications as well.
//
//========================================

const Real kZLS_PPM_Q1 = (1.0/12.0);
const Real kZLS_PPM_Q2 = (1.0/15.0);

int    zlsppm_CalcInterpVar1D (V1DArr  v, InPtr f, int n){
#ifdef GODUNOV
	return(noErr);
#else
	
	int err = noErr;
	Integer i, k;
	Integer is, ie;
	
#ifdef PLM			
	Real  dela;
#endif PLM			
#ifdef PPM			
	Real  dela, a6a;
#endif PPM			
	Var1 * va;
	
	if (v.v!=NULL) {
		
		va = v.v;
		
		if ((n <2)||(n >8)) {
			zlsio_Printf( " Error: Only n = 2-8 interpolation allowed. %d\n\n", n);
			err = dimErr;
			return(err);
		}
		
		is = 0;
		ie = v.nt;
		
		
		for(i = is;i<ie;i++){
			
#ifdef PLM			
//			aj   = va[i].a;
//			dela = va[i].del;
//			for( k = 0; k<n; k++){
//				f[i].c[k]  = aj + gZLS_PPM_fxDelta[k][n]*dela;
//				f[i].c[k]  = (aj == 0.0)? 1.0: f[i].c[k]/ aj;
//			}//k
			dela = va[i].ar - va[i].al;
			for( k = 0; k<n; k++){
				f[i].c[k]  = gZLS_PPM_fxDelta[k][n]*dela;
			}//k
#endif PLM			

#ifdef PPM			
//			aj   = va[i].a;
//			for( k = 0; k<n; k++){
//				f[i].c[k]  = aj + gZLS_PPM_fxDelta[k][n]*dela + gZLS_PPM_fxA6[k][n]*a6a;
//				f[i].c[k]  = (aj == 0.0)? 1.0: f[i].c[k]/ aj;
//			}//k

			dela = va[i].del;
			a6a  = va[i].a6;
			for( k = 0; k<n; k++){
				f[i].c[k]  = gZLS_PPM_fxDelta[k][n]*dela + gZLS_PPM_fxA6[k][n]*a6a;
			}//k
#endif PPM			
			
		}//i
		
		
	}  else { err = memErr; }
	
	return(err);
#endif GODUNOV
	
}


int    zlsppm_CalcInterpVar2D (V2DArr  v, InPtr f, int n){
#pragma unused( v, f, n)
	int err = nyiErr;
	return(err);
}

int    zlsppm_CalcInterpVar3D (V3DArr  v, InPtr f, int n){
#pragma unused( v, f, n)
	int err = nyiErr;
	return(err);
}

int    zlsppm_CalcLinInterpVar1D (V1DArr  v, InPtr f, int n){
	
	int err = noErr;
	Integer i, k;
	Integer is, ie;
	
	Real  dela;
	Var1 * va;
	
	if (v.v!=NULL) {
		
		va = v.v;
		
		if ((n <2)||(n >8)) {
			zlsio_Printf( " Error: Only n = 2-8 interpolation allowed. %d\n\n", n);
			err = dimErr;
			return(err);
		}
		
		is = 0;
		ie = v.nt;
		
		for(i = is;i<ie;i++){
			
#ifdef GODUNOV
			dela = 0.0;
#else
			dela = va[i].ar-va[i].al;
#endif !GODUNOV
			
			for( k = 0; k<n; k++){
				f[i].c[k]  = gZLS_PPM_fxDelta[k][n]*dela;
		    }//k
			
		}//i
		
		
	}  else { err = memErr; }
	
	return(err);
	
}

//
//====================================================================================
//
//  PPM Interpolation coefficients and resampling
//
//====================================================================================
//


Real  gZLS_PPM_fxDelta[kNRefineMax][(kNRefineMax+1)];
Real     gZLS_PPM_fxA6[kNRefineMax][(kNRefineMax+1)];


void   zlsppm_InitVarInterpCoeffs(void){
	
	int i, j;
	
	for(i = 0; i<kNRefineMax; i++){
		for(j = 0; j<(kNRefineMax+1); j++){
			gZLS_PPM_fxDelta[i][j] = 0.0;
			gZLS_PPM_fxA6[i][j] = 0.0;
		}
	}
	
#if (NREFINEMAX<2)
	zlsio_ErrPrintf( " ERROR: nrefinemax: NREFINEMAX (allowed 2-8)\n\n" );
#endif
#if (NREFINEMAX==2)
	gZLS_PPM_fxDelta[0][2] = (Real)k0DeltaN2/kAjDeltaN2;
	gZLS_PPM_fxDelta[1][2] = (Real)k1DeltaN2/kAjDeltaN2;
	gZLS_PPM_fxA6[0][2] = 0.0;
	gZLS_PPM_fxA6[1][2] = 0.0;
#endif
#if (NREFINEMAX==3)
	gZLS_PPM_fxDelta[0][2] = (Real)k0DeltaN2/kAjDeltaN2;
	gZLS_PPM_fxDelta[1][2] = (Real)k1DeltaN2/kAjDeltaN2;
	gZLS_PPM_fxDelta[0][3] = (Real)k0DeltaN3/kAjDeltaN3;
	gZLS_PPM_fxDelta[1][3] = (Real)k1DeltaN3/kAjDeltaN3;
	gZLS_PPM_fxDelta[2][3] = (Real)k2DeltaN3/kAjDeltaN3;
	gZLS_PPM_fxA6[0][2] = 0.0;
	gZLS_PPM_fxA6[1][2] = 0.0;
	gZLS_PPM_fxA6[0][3] = (Real)k0A6N3/kAjA6N3;
	gZLS_PPM_fxA6[1][3] = (Real)k1A6N3/kAjA6N3;
	gZLS_PPM_fxA6[2][3] = (Real)k2A6N3/kAjA6N3;
#endif
#if (NREFINEMAX==4)
	gZLS_PPM_fxDelta[0][2] = (Real)k0DeltaN2/kAjDeltaN2;
	gZLS_PPM_fxDelta[1][2] = (Real)k1DeltaN2/kAjDeltaN2;
	gZLS_PPM_fxDelta[0][3] = (Real)k0DeltaN3/kAjDeltaN3;
	gZLS_PPM_fxDelta[1][3] = (Real)k1DeltaN3/kAjDeltaN3;
	gZLS_PPM_fxDelta[2][3] = (Real)k2DeltaN3/kAjDeltaN3;
	gZLS_PPM_fxDelta[0][4] = (Real)k0DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[1][4] = (Real)k1DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[2][4] = (Real)k2DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[3][4] = (Real)k3DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxA6[0][2] = 0.0;
	gZLS_PPM_fxA6[1][2] = 0.0;
	gZLS_PPM_fxA6[0][3] = (Real)k0A6N3/kAjA6N3;
	gZLS_PPM_fxA6[1][3] = (Real)k1A6N3/kAjA6N3;
	gZLS_PPM_fxA6[2][3] = (Real)k2A6N3/kAjA6N3;
	gZLS_PPM_fxA6[0][4] = (Real)k0A6N4/kAjA6N4;
	gZLS_PPM_fxA6[1][4] = (Real)k1A6N4/kAjA6N4;
	gZLS_PPM_fxA6[2][4] = (Real)k2A6N4/kAjA6N4;
	gZLS_PPM_fxA6[3][4] = (Real)k3A6N4/kAjA6N4;
#endif
#if (NREFINEMAX==5)
	gZLS_PPM_fxDelta[0][2] = (Real)k0DeltaN2/kAjDeltaN2;
	gZLS_PPM_fxDelta[1][2] = (Real)k1DeltaN2/kAjDeltaN2;
	gZLS_PPM_fxDelta[0][3] = (Real)k0DeltaN3/kAjDeltaN3;
	gZLS_PPM_fxDelta[1][3] = (Real)k1DeltaN3/kAjDeltaN3;
	gZLS_PPM_fxDelta[2][3] = (Real)k2DeltaN3/kAjDeltaN3;
	gZLS_PPM_fxDelta[0][4] = (Real)k0DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[1][4] = (Real)k1DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[2][4] = (Real)k2DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[3][4] = (Real)k3DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[0][5] = (Real)k0DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[1][5] = (Real)k1DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[2][5] = (Real)k2DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[3][5] = (Real)k3DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[4][5] = (Real)k4DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxA6[0][2] = 0.0;
	gZLS_PPM_fxA6[1][2] = 0.0;
	gZLS_PPM_fxA6[0][3] = (Real)k0A6N3/kAjA6N3;
	gZLS_PPM_fxA6[1][3] = (Real)k1A6N3/kAjA6N3;
	gZLS_PPM_fxA6[2][3] = (Real)k2A6N3/kAjA6N3;
	gZLS_PPM_fxA6[0][4] = (Real)k0A6N4/kAjA6N4;
	gZLS_PPM_fxA6[1][4] = (Real)k1A6N4/kAjA6N4;
	gZLS_PPM_fxA6[2][4] = (Real)k2A6N4/kAjA6N4;
	gZLS_PPM_fxA6[3][4] = (Real)k3A6N4/kAjA6N4;
	gZLS_PPM_fxA6[0][5] = (Real)k0A6N5/kAjA6N5;
	gZLS_PPM_fxA6[1][5] = (Real)k1A6N5/kAjA6N5;
	gZLS_PPM_fxA6[2][5] = (Real)k2A6N5/kAjA6N5;
	gZLS_PPM_fxA6[3][5] = (Real)k3A6N5/kAjA6N5;
	gZLS_PPM_fxA6[4][5] = (Real)k4A6N5/kAjA6N5;
#endif
#if (NREFINEMAX==6)
	gZLS_PPM_fxDelta[0][2] = (Real)k0DeltaN2/kAjDeltaN2;
	gZLS_PPM_fxDelta[1][2] = (Real)k1DeltaN2/kAjDeltaN2;
	gZLS_PPM_fxDelta[0][3] = (Real)k0DeltaN3/kAjDeltaN3;
	gZLS_PPM_fxDelta[1][3] = (Real)k1DeltaN3/kAjDeltaN3;
	gZLS_PPM_fxDelta[2][3] = (Real)k2DeltaN3/kAjDeltaN3;
	gZLS_PPM_fxDelta[0][4] = (Real)k0DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[1][4] = (Real)k1DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[2][4] = (Real)k2DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[3][4] = (Real)k3DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[0][5] = (Real)k0DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[1][5] = (Real)k1DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[2][5] = (Real)k2DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[3][5] = (Real)k3DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[4][5] = (Real)k4DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[0][6] = (Real)k0DeltaN6/kAjDeltaN6;
	gZLS_PPM_fxDelta[1][6] = (Real)k1DeltaN6/kAjDeltaN6;
	gZLS_PPM_fxDelta[2][6] = (Real)k2DeltaN6/kAjDeltaN6;
	gZLS_PPM_fxDelta[3][6] = (Real)k3DeltaN6/kAjDeltaN6;
	gZLS_PPM_fxDelta[4][6] = (Real)k4DeltaN6/kAjDeltaN6;
	gZLS_PPM_fxDelta[5][6] = (Real)k5DeltaN6/kAjDeltaN6;
	gZLS_PPM_fxA6[0][2] = 0.0;
	gZLS_PPM_fxA6[1][2] = 0.0;
	gZLS_PPM_fxA6[0][3] = (Real)k0A6N3/kAjA6N3;
	gZLS_PPM_fxA6[1][3] = (Real)k1A6N3/kAjA6N3;
	gZLS_PPM_fxA6[2][3] = (Real)k2A6N3/kAjA6N3;
	gZLS_PPM_fxA6[0][4] = (Real)k0A6N4/kAjA6N4;
	gZLS_PPM_fxA6[1][4] = (Real)k1A6N4/kAjA6N4;
	gZLS_PPM_fxA6[2][4] = (Real)k2A6N4/kAjA6N4;
	gZLS_PPM_fxA6[3][4] = (Real)k3A6N4/kAjA6N4;
	gZLS_PPM_fxA6[0][5] = (Real)k0A6N5/kAjA6N5;
	gZLS_PPM_fxA6[1][5] = (Real)k1A6N5/kAjA6N5;
	gZLS_PPM_fxA6[2][5] = (Real)k2A6N5/kAjA6N5;
	gZLS_PPM_fxA6[3][5] = (Real)k3A6N5/kAjA6N5;
	gZLS_PPM_fxA6[4][5] = (Real)k4A6N5/kAjA6N5;
	gZLS_PPM_fxA6[0][6] = (Real)k0A6N6/kAjA6N6;
	gZLS_PPM_fxA6[1][6] = (Real)k1A6N6/kAjA6N6;
	gZLS_PPM_fxA6[2][6] = (Real)k2A6N6/kAjA6N6;
	gZLS_PPM_fxA6[3][6] = (Real)k3A6N6/kAjA6N6;
	gZLS_PPM_fxA6[4][6] = (Real)k4A6N6/kAjA6N6;
	gZLS_PPM_fxA6[5][6] = (Real)k5A6N6/kAjA6N6;
#endif
#if (NREFINEMAX==7)
	gZLS_PPM_fxDelta[0][2] = (Real)k0DeltaN2/kAjDeltaN2;
	gZLS_PPM_fxDelta[1][2] = (Real)k1DeltaN2/kAjDeltaN2;
	gZLS_PPM_fxDelta[0][3] = (Real)k0DeltaN3/kAjDeltaN3;
	gZLS_PPM_fxDelta[1][3] = (Real)k1DeltaN3/kAjDeltaN3;
	gZLS_PPM_fxDelta[2][3] = (Real)k2DeltaN3/kAjDeltaN3;
	gZLS_PPM_fxDelta[0][4] = (Real)k0DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[1][4] = (Real)k1DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[2][4] = (Real)k2DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[3][4] = (Real)k3DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[0][5] = (Real)k0DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[1][5] = (Real)k1DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[2][5] = (Real)k2DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[3][5] = (Real)k3DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[4][5] = (Real)k4DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[0][6] = (Real)k0DeltaN6/kAjDeltaN6;
	gZLS_PPM_fxDelta[1][6] = (Real)k1DeltaN6/kAjDeltaN6;
	gZLS_PPM_fxDelta[2][6] = (Real)k2DeltaN6/kAjDeltaN6;
	gZLS_PPM_fxDelta[3][6] = (Real)k3DeltaN6/kAjDeltaN6;
	gZLS_PPM_fxDelta[4][6] = (Real)k4DeltaN6/kAjDeltaN6;
	gZLS_PPM_fxDelta[5][6] = (Real)k5DeltaN6/kAjDeltaN6;
	gZLS_PPM_fxDelta[0][7] = (Real)k0DeltaN7/kAjDeltaN7;
	gZLS_PPM_fxDelta[1][7] = (Real)k1DeltaN7/kAjDeltaN7;
	gZLS_PPM_fxDelta[2][7] = (Real)k2DeltaN7/kAjDeltaN7;
	gZLS_PPM_fxDelta[3][7] = (Real)k3DeltaN7/kAjDeltaN7;
	gZLS_PPM_fxDelta[4][7] = (Real)k4DeltaN7/kAjDeltaN7;
	gZLS_PPM_fxDelta[5][7] = (Real)k5DeltaN7/kAjDeltaN7;
	gZLS_PPM_fxDelta[6][7] = (Real)k6DeltaN7/kAjDeltaN7;
	gZLS_PPM_fxA6[0][2] = 0.0;
	gZLS_PPM_fxA6[1][2] = 0.0;
	gZLS_PPM_fxA6[0][3] = (Real)k0A6N3/kAjA6N3;
	gZLS_PPM_fxA6[1][3] = (Real)k1A6N3/kAjA6N3;
	gZLS_PPM_fxA6[2][3] = (Real)k2A6N3/kAjA6N3;
	gZLS_PPM_fxA6[0][4] = (Real)k0A6N4/kAjA6N4;
	gZLS_PPM_fxA6[1][4] = (Real)k1A6N4/kAjA6N4;
	gZLS_PPM_fxA6[2][4] = (Real)k2A6N4/kAjA6N4;
	gZLS_PPM_fxA6[3][4] = (Real)k3A6N4/kAjA6N4;
	gZLS_PPM_fxA6[0][5] = (Real)k0A6N5/kAjA6N5;
	gZLS_PPM_fxA6[1][5] = (Real)k1A6N5/kAjA6N5;
	gZLS_PPM_fxA6[2][5] = (Real)k2A6N5/kAjA6N5;
	gZLS_PPM_fxA6[3][5] = (Real)k3A6N5/kAjA6N5;
	gZLS_PPM_fxA6[4][5] = (Real)k4A6N5/kAjA6N5;
	gZLS_PPM_fxA6[0][6] = (Real)k0A6N6/kAjA6N6;
	gZLS_PPM_fxA6[1][6] = (Real)k1A6N6/kAjA6N6;
	gZLS_PPM_fxA6[2][6] = (Real)k2A6N6/kAjA6N6;
	gZLS_PPM_fxA6[3][6] = (Real)k3A6N6/kAjA6N6;
	gZLS_PPM_fxA6[4][6] = (Real)k4A6N6/kAjA6N6;
	gZLS_PPM_fxA6[5][6] = (Real)k5A6N6/kAjA6N6;
	gZLS_PPM_fxA6[0][7] = (Real)k0A6N7/kAjA6N7;
	gZLS_PPM_fxA6[1][7] = (Real)k1A6N7/kAjA6N7;
	gZLS_PPM_fxA6[2][7] = (Real)k2A6N7/kAjA6N7;
	gZLS_PPM_fxA6[3][7] = (Real)k3A6N7/kAjA6N7;
	gZLS_PPM_fxA6[4][7] = (Real)k4A6N7/kAjA6N7;
	gZLS_PPM_fxA6[5][7] = (Real)k5A6N7/kAjA6N7;
	gZLS_PPM_fxA6[6][7] = (Real)k6A6N7/kAjA6N7;
#endif
#if (NREFINEMAX==8)
	gZLS_PPM_fxDelta[0][2] = (Real)k0DeltaN2/kAjDeltaN2;
	gZLS_PPM_fxDelta[1][2] = (Real)k1DeltaN2/kAjDeltaN2;
	gZLS_PPM_fxDelta[0][3] = (Real)k0DeltaN3/kAjDeltaN3;
	gZLS_PPM_fxDelta[1][3] = (Real)k1DeltaN3/kAjDeltaN3;
	gZLS_PPM_fxDelta[2][3] = (Real)k2DeltaN3/kAjDeltaN3;
	gZLS_PPM_fxDelta[0][4] = (Real)k0DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[1][4] = (Real)k1DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[2][4] = (Real)k2DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[3][4] = (Real)k3DeltaN4/kAjDeltaN4;
	gZLS_PPM_fxDelta[0][5] = (Real)k0DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[1][5] = (Real)k1DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[2][5] = (Real)k2DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[3][5] = (Real)k3DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[4][5] = (Real)k4DeltaN5/kAjDeltaN5;
	gZLS_PPM_fxDelta[0][6] = (Real)k0DeltaN6/kAjDeltaN6;
	gZLS_PPM_fxDelta[1][6] = (Real)k1DeltaN6/kAjDeltaN6;
	gZLS_PPM_fxDelta[2][6] = (Real)k2DeltaN6/kAjDeltaN6;
	gZLS_PPM_fxDelta[3][6] = (Real)k3DeltaN6/kAjDeltaN6;
	gZLS_PPM_fxDelta[4][6] = (Real)k4DeltaN6/kAjDeltaN6;
	gZLS_PPM_fxDelta[5][6] = (Real)k5DeltaN6/kAjDeltaN6;
	gZLS_PPM_fxDelta[0][7] = (Real)k0DeltaN7/kAjDeltaN7;
	gZLS_PPM_fxDelta[1][7] = (Real)k1DeltaN7/kAjDeltaN7;
	gZLS_PPM_fxDelta[2][7] = (Real)k2DeltaN7/kAjDeltaN7;
	gZLS_PPM_fxDelta[3][7] = (Real)k3DeltaN7/kAjDeltaN7;
	gZLS_PPM_fxDelta[4][7] = (Real)k4DeltaN7/kAjDeltaN7;
	gZLS_PPM_fxDelta[5][7] = (Real)k5DeltaN7/kAjDeltaN7;
	gZLS_PPM_fxDelta[6][7] = (Real)k6DeltaN7/kAjDeltaN7;
	gZLS_PPM_fxDelta[0][8] = (Real)k0DeltaN8/kAjDeltaN8;
	gZLS_PPM_fxDelta[1][8] = (Real)k1DeltaN8/kAjDeltaN8;
	gZLS_PPM_fxDelta[2][8] = (Real)k2DeltaN8/kAjDeltaN8;
	gZLS_PPM_fxDelta[3][8] = (Real)k3DeltaN8/kAjDeltaN8;
	gZLS_PPM_fxDelta[4][8] = (Real)k4DeltaN8/kAjDeltaN8;
	gZLS_PPM_fxDelta[5][8] = (Real)k5DeltaN8/kAjDeltaN8;
	gZLS_PPM_fxDelta[6][8] = (Real)k6DeltaN8/kAjDeltaN8;
	gZLS_PPM_fxDelta[7][8] = (Real)k7DeltaN8/kAjDeltaN8;
	gZLS_PPM_fxA6[0][2] = 0.0;
	gZLS_PPM_fxA6[1][2] = 0.0;
	gZLS_PPM_fxA6[0][3] = (Real)k0A6N3/kAjA6N3;
	gZLS_PPM_fxA6[1][3] = (Real)k1A6N3/kAjA6N3;
	gZLS_PPM_fxA6[2][3] = (Real)k2A6N3/kAjA6N3;
	gZLS_PPM_fxA6[0][4] = (Real)k0A6N4/kAjA6N4;
	gZLS_PPM_fxA6[1][4] = (Real)k1A6N4/kAjA6N4;
	gZLS_PPM_fxA6[2][4] = (Real)k2A6N4/kAjA6N4;
	gZLS_PPM_fxA6[3][4] = (Real)k3A6N4/kAjA6N4;
	gZLS_PPM_fxA6[0][5] = (Real)k0A6N5/kAjA6N5;
	gZLS_PPM_fxA6[1][5] = (Real)k1A6N5/kAjA6N5;
	gZLS_PPM_fxA6[2][5] = (Real)k2A6N5/kAjA6N5;
	gZLS_PPM_fxA6[3][5] = (Real)k3A6N5/kAjA6N5;
	gZLS_PPM_fxA6[4][5] = (Real)k4A6N5/kAjA6N5;
	gZLS_PPM_fxA6[0][6] = (Real)k0A6N6/kAjA6N6;
	gZLS_PPM_fxA6[1][6] = (Real)k1A6N6/kAjA6N6;
	gZLS_PPM_fxA6[2][6] = (Real)k2A6N6/kAjA6N6;
	gZLS_PPM_fxA6[3][6] = (Real)k3A6N6/kAjA6N6;
	gZLS_PPM_fxA6[4][6] = (Real)k4A6N6/kAjA6N6;
	gZLS_PPM_fxA6[5][6] = (Real)k5A6N6/kAjA6N6;
	gZLS_PPM_fxA6[0][7] = (Real)k0A6N7/kAjA6N7;
	gZLS_PPM_fxA6[1][7] = (Real)k1A6N7/kAjA6N7;
	gZLS_PPM_fxA6[2][7] = (Real)k2A6N7/kAjA6N7;
	gZLS_PPM_fxA6[3][7] = (Real)k3A6N7/kAjA6N7;
	gZLS_PPM_fxA6[4][7] = (Real)k4A6N7/kAjA6N7;
	gZLS_PPM_fxA6[5][7] = (Real)k5A6N7/kAjA6N7;
	gZLS_PPM_fxA6[6][7] = (Real)k6A6N7/kAjA6N7;
	gZLS_PPM_fxA6[0][8] = (Real)k0A6N8/kAjA6N8;
	gZLS_PPM_fxA6[1][8] = (Real)k1A6N8/kAjA6N8;
	gZLS_PPM_fxA6[2][8] = (Real)k2A6N8/kAjA6N8;
	gZLS_PPM_fxA6[3][8] = (Real)k3A6N8/kAjA6N8;
	gZLS_PPM_fxA6[4][8] = (Real)k4A6N8/kAjA6N8;
	gZLS_PPM_fxA6[5][8] = (Real)k5A6N8/kAjA6N8;
	gZLS_PPM_fxA6[6][8] = (Real)k6A6N8/kAjA6N8;
	gZLS_PPM_fxA6[7][8] = (Real)k7A6N8/kAjA6N8;
#endif
	
}

