#include <zls.h>

#ifdef HYDRO

#ifdef SYM2D

// set up permutations and sweep limits for 2D sweeps and orthogonal starting points

static const Integer nSweepOrder[4] = { kXDim, kYDim, kYDim, kXDim }; // std cycle xy-yx sweeps
static const Integer kScanStartX[4] = { 0, kZLS_LimitSecondSweepScan, kZLS_LimitFirstSweepScan,0 };
static const Integer kScanStartY[4] = { kZLS_LimitFirstSweepScan, 0, 0, kZLS_LimitSecondSweepScan };

//static const Integer nSweepOrder[4] = { kXDim, kYDim, kXDim, kYDim }; // cycle xy-xy sweeps
//static const Integer kScanStartX[4] = { 0, kZLS_LimitSecondSweepScan, 0, kZLS_LimitSecondSweepScan };
//static const Integer kScanStartY[4] = { kZLS_LimitFirstSweepScan, 0, kZLS_LimitFirstSweepScan, 0 };

#endif SYM2D

#ifdef SYM3D

// set up permutations and sweep limits for 3D sweeps and orthogonal starting points

static const Integer nSweepOrder[9] = { kXDim, kYDim, kZDim, 
  kYDim, kZDim, kXDim, 
kZDim, kXDim, kYDim }; // cycle xyz-yzx-zxy- sweeps

// each sweep stage scan starts:

static const Integer kScanStartX[9] = { 0, kZLS_LimitFirstSweepScan, kZLS_LimitSecondSweepScan,
  kZLS_LimitFirstSweepScan, kZLS_LimitSecondSweepScan, 0,
kZLS_LimitFirstSweepScan, 0, kZLS_LimitSecondSweepScan }; // cycle xyz-yzx-zxy- sweeps

static const Integer kScanStartY[9] = { kZLS_LimitFirstSweepScan, 0, kZLS_LimitSecondSweepScan,
  0, kZLS_LimitFirstSweepScan, kZLS_LimitSecondSweepScan,
kZLS_LimitFirstSweepScan, kZLS_LimitSecondSweepScan, 0 }; // cycle xyz-yzx-zxy- sweeps

static const Integer kScanStartZ[9] = { kZLS_LimitFirstSweepScan, kZLS_LimitSecondSweepScan, 0,
  kZLS_LimitFirstSweepScan, 0,  kZLS_LimitSecondSweepScan,
0, kZLS_LimitFirstSweepScan, kZLS_LimitSecondSweepScan }; // cycle xyz-yzx-zxy- sweeps

#endif SYM3D

#ifdef LAGRANGIAN
//====================================================================================
//
// Lagrangian 1D Hydrodynamic Sweep of a zone, uses W, 
// primitive variables zone must have timestep ready for hydro updates.  
//
//====================================================================================
#ifdef SYM1D
int zlsa_1DSweep ( ZonePtr z){
	
  int   err = noErr;
  Real           dt;
  ACoordPtr      x0;
	
  if (z != NULL){
    
    WHD        zoneW  =  z->zoneW;
    GridCoords gc = z->zoneGrid;
    
    
#ifdef GRAVITY
    // get the zone field and allocate a 1D array for passing it
    // to the sweeps with interpolation coeffs set up
    Field   zoneF;
    zls_GetZoneField(&zoneF, z); // nothing is allocated here, so no dispose is needed later
    F1DArr  f0 = z->zoneF0X;
    f0.x       = gc.x;
#endif GRAVITY
    
    W1DArr        w0 = z->zoneW0X;
    Wstar1DArr wStar = z->zoneWStarX;
    Flx1DArr    flux = z->zone1DfluxX;
    
    dt  = z->zoneClock.dt;
    
    //=== X - Sweep ============================================================================
    
    w0.x    = gc.x;
#ifdef LAGRANGIAN
    x0 = gc.x0;
#else 
    x0 = gc.x;
#endif				
    wStar.x = x0;
    
    err = zlsh_FillW1DFrom1DZoneWHD         ( w0,  zoneW ); // get a row, w0 will have BCs whether or not zone does
    
#ifdef GRAVITY
    err = zlsh_FillF1DFrom1DField           ( f0,  zoneF ); // get radial potential row with BCs
    err = zlsppm_ReconstructPPMWLRStates1DF ( w0, f0, dt, 1.0); 
#else
    err = zlsppm_ReconstructPPMWLRStates1D  ( w0 , dt, 1.0 ); // averages LR states over domain of dependence from PPM coeffs
#endif !GRAVITY
    
    err = RIEMANNSOLVER                     ( wStar, w0 ); // macro set up in zls_const.h        
    
#ifdef PRETEST_W_VEL
    Integer is  = kZLS_LimitEvolveLimits;
    Integer ie  = gc.nt[kXDim]-kZLS_LimitEvolveLimits;
    
    Integer i;
    for ( i = is; i < ie; i++){
      if ( wStar.v[i].a != 0.0 ) goto process1d;
    }
    goto cooling1d;
  process1d:
#endif PRETEST_W_VEL
    
#ifdef GRAVITY
    err = zlsa_LagrangianEvolveW1DF  ( w0, wStar, f0, dt );
#else
    err = zlsa_LagrangianEvolveW1D  ( w0, wStar, dt );
#endif !GRAVITY
    
    err = zlsa_LagrangianRemapW1D   ( w0, flux, x0 ); // remap w0,
    
#ifdef VISCOUS_REMAP
    if ( kZLS_SmoothAxes & kSmoothXFlag )
      err = zlsa_RemapSmoothW1D( w0, dt );
#endif VISCOUS_REMAP
    
    err = zlsh_Fill1DZoneWHDFromW1D          ( zoneW,  w0 ); // put it back
    
    //====================================================================================
#ifdef PRETEST_W_VEL
  cooling1d:
    ;
#endif PRETEST_W_VEL
    
#ifdef COOLING	
    err = zlsp_CoolZoneWHD( z, dt);
#endif COOLING
    
  } //z
	
  return(err);
}

#ifdef SPHERICAL

int zlsa_1DSweepSphere( ZonePtr z){
	
  int err = noErr;
  Real g;  
  ACoordPtr     x0;
	
  if (z != NULL){
    
    WHD         zoneW =  z->zoneW;
    
    GridCoords gc =  z->zoneGrid;
    
#ifdef GRAVITY
    // get the zone field and allocate a 1D array for passing it
    // to the sweeps with interpolation coeffs set up
    Field   zoneF;
    zls_GetZoneField(&zoneF, z); // nothing is allocated here, so no dispose is needed later
    F1DArr  f0X = z->zoneF0X;
    f0X.x       = gc.x;
#endif GRAVITY
    
    W1DArr        w0X = z->zoneW0X;
    Wstar1DArr wStarX = z->zoneWStarX;
    Flx1DArr    fluxX = z->zone1DfluxX;
    
    Real  dt = z->zoneClock.dt;
    
    //=== R - Radial Sweep ============================================================================
    
    w0X.x    = gc.x;
#ifdef LAGRANGIAN
    x0 = gc.x0;
#else 
    x0 = gc.x;
#endif				
    wStarX.x = x0;
    g   = 1.0; // all first axis sweeps are spatial, so g = 1.0
    
    err = zlsh_FillW1DFrom1DZoneWHD            ( w0X, zoneW ); // get a row with BCs
#ifdef DIVVHCORRECTION
    err = zlsh_FillW1DDivVFrom1DZone           ( w0X, z);
#endif DIVVHCORRECTION
#ifdef GRAVITY
    err = zlsh_FillF1DFrom1DField              ( f0X, zoneF ); // get radial potential row with BCs
    err = zlsppm_ReconstructPPMWLRStates1DF    ( w0X, f0X, dt, g ); 
#else
    err = zlsppm_ReconstructPPMWLRStates1D     ( w0X , dt, 1.0 ); // averages LR states over domain of dependence from PPM coeffs
#endif !GRAVITY
    
    err = RIEMANNSOLVER                        ( wStarX, w0X );      // Note: macro set up in zls_defs.h
    
#ifdef PRETEST_W_VEL
    Integer is  = kZLS_LimitEvolveLimits;
    Integer ie  = gc.nt[kXDim]-kZLS_LimitEvolveLimits;
    Integer i;
    for ( i = is; i < ie; i++){
      if ( wStarX.v[i].a != 0.0 ) goto processR1;
    }
    goto coolingR1;
    
  processR1:
    
#endif PRETEST_W_VEL
    
#ifdef GRAVITY
    zlsa_LagrangianEvolveSRadialW1DF     ( w0X, wStarX, f0X, dt );
#else
    zlsa_LagrangianEvolveSRadialW1D      ( w0X, wStarX, dt );
#endif !GRAVITY
    zlsa_LagrangianRemapSRadialW1D       ( w0X, fluxX, x0);
    
#ifdef VISCOUS_REMAP
    if ( kZLS_SmoothAxes & kSmoothXFlag )
      err = zlsa_RemapSmoothW1D( w0X, dt );
#endif VISCOUS_REMAP
    
    zlsh_Fill1DZoneWHDFromW1D            ( zoneW,  w0X ); // put it back without BCs
    
#ifdef PRETEST_W_VEL
  coolingR1:
    ;
#endif PRETEST_W_VEL
    
#ifdef COOLING	
    err = zlsp_CoolZoneWHD( z, dt );
#endif COOLING	
    
  } //z
  
  return(err);
}

#endif SPHERICAL

#ifdef CIRCULAR
int zlsa_1DSweepCircle( ZonePtr z){
	// use ordinary sweeps for 1D cylinder Z, this is for 1D cylinder/circular r
  int err = noErr;
  Real g;  
  ACoordPtr     x0;
	
  if (z != NULL){
    
    WHD         zoneW =  z->zoneW;
    
    GridCoords gc =  z->zoneGrid;
    
#ifdef GRAVITY
    // get the zone field and allocate a 1D array for passing it
    // to the sweeps with interpolation coeffs set up
    Field   zoneF;
    zls_GetZoneField(&zoneF, z); // nothing is allocated here, so no dispose is needed later
    F1DArr  f0X = z->zoneF0X;
    f0X.x    = gc.x;
#endif GRAVITY
    
    W1DArr        w0X = z->zoneW0X;
    Wstar1DArr wStarX = z->zoneWStarX;
    Flx1DArr    fluxX = z->zone1DfluxX;
    
    Real  dt     = z->zoneClock.dt;
    
    //=== R - Radial Sweep ============================================================================
    
    w0X.x    = gc.x;
#ifdef LAGRANGIAN
    x0 = gc.x0;
#else 
    x0 = gc.x;
#endif				
    wStarX.x = x0;
    g   = 1.0; // all first axis sweeps are spatial, so g = 1.0
    
    err = zlsh_FillW1DFrom1DZoneWHD            ( w0X,  zoneW ); // get a row with BCs
#ifdef DIVVHCORRECTION
    err = zlsh_FillW1DDivVFrom1DZone           ( w0X, z);
#endif DIVVHCORRECTION
#ifdef GRAVITY
    err = zlsh_FillF1DFrom1DField              ( f0X,  zoneF ); // get radial potential row with BCs
    err = zlsppm_ReconstructPPMWLRStates1DF    ( w0X, f0X, dt, g); 
#else
    err = zlsppm_ReconstructPPMWLRStates1D      ( w0X , dt, 1.0); // averages LR states over domain of dependence from PPM coeffs
#endif !GRAVITY
    
    err = RIEMANNSOLVER                         ( wStarX, w0X );      // Note: macro set up in zls_defs.h
    
#ifdef PRETEST_W_VEL
    Integer is  = kZLS_LimitEvolveLimits;
    Integer ie  = gc.nt[kXDim]-kZLS_LimitEvolveLimits;
    Integer i;
    for ( i = is; i < ie; i++){
      if ( wStarX.v[i].a != 0.0 ) goto processC1;
    }
    goto coolingC1;
  processC1:
#endif PRETEST_W_VEL
    
#ifdef GRAVITY
    zlsa_LagrangianEvolveCRadialW1DF     ( w0X, wStarX, f0X, dt );
#else
    zlsa_LagrangianEvolveCRadialW1D      ( w0X, wStarX, dt );
#endif !GRAVITY
    zlsa_LagrangianRemapCRadialW1D       ( w0X, fluxX, x0);
    
#ifdef VISCOUS_REMAP
    if ( kZLS_SmoothAxes & kSmoothXFlag )
      err = zlsa_RemapSmoothW1D( w0X, dt );
#endif VISCOUS_REMAP
    
    zlsh_Fill1DZoneWHDFromW1D            ( zoneW,  w0X ); // put it back without BCs
    
#ifdef PRETEST_W_VEL
  coolingC1:
    ;
#endif PRETEST_W_VEL
    
#ifdef COOLING	
    err = zlsp_CoolZoneWHD( z, dt );
#endif COOLING	
    
  } //z
  
  return(err);
}
#endif CIRCULAR

#endif SYM1D
#endif LAGRANGIAN

#ifdef SPLITSWEEPS

//====================================================================================
//
// Lagrangian 2D Hydrodynamic Sweep of a zone, uses W, 
// primitive variables zone must have timestep ready for hydro updates.  
//
// Split Lagrangian  + remap algorithm
//
//====================================================================================
// conversion from angular to spatial coordinates: 
// Cartesian = x, y, g = 1, 1, 2D Spherical = r, phi, g = 1, r
// 2D Circular = r, phi, g = 1, r, Cylinder = z, r, g = 1, 1

#ifdef SYM2D
int zlsa_2DSplitSweeps( ZonePtr z){
	
  int err = noErr;
	
  if (z != NULL){
    
    WHD         zoneW =  z->zoneW;
    
    GridCoords gc =  z->zoneGrid;
    ACoordPtr     x0;
    
#ifdef GRAVITY
    // get the zone field and allocate a 1D array for passing it
    // to the sweeps with interpolation coeffs set up
    Field   zoneF;
    zls_GetZoneField(&zoneF, z); // nothing is allocated here, so no dispose is needed later
    F1DArr    f0X = z->zoneF0X;
    F1DArr    f0Y = z->zoneF0Y;    
    Real       g  = 1.0;
#endif GRAVITY
    
    W1DArr        w0X = z->zoneW0X;
    Wstar1DArr wStarX = z->zoneWStarX;
    Flx1DArr    fluxX = z->zone1DfluxX;
    
    W1DArr        w0Y = z->zoneW0Y;
    Wstar1DArr wStarY = z->zoneWStarY;
    Flx1DArr    fluxY = z->zone1DfluxY;
    
    Real       dt = z->zoneClock.dt;
    
    Integer   i, j;
    
    Integer is ;
    Integer ie ;
    Integer js ;
    Integer je ;
    
    Integer nc  = (z->zoneClock.cycles) % k2D;
    Integer iSweep, iScan;
    //    zlsio_Printf(" NC %d \n", nc);
    
    for( iSweep = 0; iSweep < k2D; iSweep++){
      iScan = iSweep + (nc * k2D); // 0-4
      
      if ( nSweepOrder[ iScan ] == kXDim ) {	
        
        //        zlsio_Printf(" X - Sweep: %d %d %d YS%d\n", nc, iSweep, iScan, kScanStartY[iScan]);
        
        //=== X - Sweep ============================================================================
        
        w0X.x    = gc.x;
#ifdef LAGRANGIAN
        x0 = gc.x0;
#else 
        x0 = gc.x;
#endif				
        wStarX.x = x0;
        
        js  = kScanStartY[iScan];
        je  = gc.nt[kYDim] - kScanStartY[iScan];
        
        // x-sweeps
        for(j = js; j<je; j++){ // step in y
          
          w0X.y    = gc.y[j];
          
          err = zlsh_FillW1DFrom2DZoneWHDX  ( w0X,  zoneW, j); // get a row
#ifdef DIVVHCORRECTION
          err = zlsh_FillW1DDivVFrom2DZoneX ( w0X, z, j); // get orthogonal divV components
#endif DIVVHCORRECTION
          
#ifdef GRAVITY
          err = zlsh_FillF1DFrom2DFieldX          ( f0X,  zoneF , j); // get a row, f0 will have BCs whether or not zone does            
          err = zlsppm_ReconstructPPMWLRStates1DF ( w0X , f0X, dt, g); // averages LR states over domain of dependence from PPM coeffs
#else
          err = zlsppm_ReconstructPPMWLRStates1D  ( w0X , dt, 1.0); // averages LR states over domain of dependence from PPM coeffs
#endif GRAVITY
          
          err = RIEMANNSOLVER              ( wStarX, w0X ); // macro set up in zls_defs.h
          
#ifdef PRETEST_W_VEL
          
          is  = kZLS_LimitEvolveLimits;
          ie  = gc.nt[kXDim]-kZLS_LimitEvolveLimits;
          
          for ( i = is; i < ie; i++){
            if ( wStarX.v[i].a != 0.0 ) goto processx1;
          }
          
          continue;
        processx1:
#endif PRETEST_W_VEL
          
#ifdef GRAVITY
          err = zlsa_LagrangianEvolveW1DF  ( w0X, wStarX, f0X, dt );
#else
          err = zlsa_LagrangianEvolveW1D   ( w0X, wStarX, dt );
#endif GRAVITY
          
          err = zlsa_LagrangianRemapW1D    ( w0X, fluxX, x0 ); // remap w0,
          
#ifdef VISCOUS_REMAP
          if ( kZLS_SmoothAxes & kSmoothXFlag )
            err = zlsa_RemapSmoothW1D( w0X, dt );
#endif VISCOUS_REMAP
          
          err = zlsh_Fill2DZoneWHDFromW1DX ( zoneW,  w0X,  j ); // put it back without BCs
          z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
          
        } //j
        
        //=== X - Sweep ============================================================================
      }
      
      
      if ( nSweepOrder[ iScan ] == kYDim ) {		
        //=== Y - Sweep ======================================================================
        
        //        zlsio_Printf(" Y - Sweep: %d %d %d XS%d\n", nc, iSweep, iScan, kScanStartX[iScan]);
        
        w0Y.x    = gc.y;
#ifdef LAGRANGIAN
        x0 = gc.y0;
#else 
        x0 = gc.y;
#endif				
        wStarY.x = x0;
        
        is  =  kScanStartX[iScan];
        ie  = gc.nt[kXDim]- kScanStartX[iScan];
        
        for(i = is; i<ie; i++){
          w0X.y    = gc.x[i];
          
          err = zlsh_FillW1DFrom2DZoneWHDY  ( w0Y,  zoneW, i); // get a column
#ifdef DIVVHCORRECTION
          err = zlsh_FillW1DDivVFrom2DZoneY ( w0Y, z, i);
#endif DIVVHCORRECTION
#ifdef GRAVITY
          err = zlsh_FillF1DFrom2DFieldY          ( f0Y,  zoneF , i); // get a row, f0 will have BCs whether or not zone does            
          err = zlsppm_ReconstructPPMWLRStates1DF ( w0Y , f0Y, dt, g); // averages LR states over domain of dependence from PPM coeffs
#else
          err = zlsppm_ReconstructPPMWLRStates1D  ( w0Y , dt, 1.0); // averages LR states over domain of dependence from PPM coeffs
#endif GRAVITY
          
          err = RIEMANNSOLVER               ( wStarY, w0Y ); // macro set up in zls_const.h
          
#ifdef PRETEST_W_VEL
          js  = kZLS_LimitEvolveLimits;
          je  = gc.nt[kYDim]-kZLS_LimitEvolveLimits;
          for ( j = js; j < je; j++){
            if ( wStarY.v[j].a != 0.0 ) goto processy1;
          };
          continue;
        processy1:
#endif PRETEST_W_VEL
          
#ifdef GRAVITY
          err = zlsa_LagrangianEvolveW1DF   ( w0Y, wStarY, f0Y, dt );
#else
          err = zlsa_LagrangianEvolveW1D    ( w0Y, wStarY, dt );
#endif GRAVITY
          
          err = zlsa_LagrangianRemapW1D     ( w0Y, fluxY, x0 ); // remap w0,
          
#ifdef VISCOUS_REMAP
          if ( kZLS_SmoothAxes & kSmoothYFlag )
            err = zlsa_RemapSmoothW1D( w0Y, dt );
#endif VISCOUS_REMAP
          
          err = zlsh_Fill2DZoneWHDFromW1DY  ( zoneW, w0Y, i ); // put it back
          z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
          
        } //i
        
        //=== Y - Sweep ======================================================================
      }
      
    } // iScan
    
#ifdef COOLING	
    err = zlsp_CoolZoneWHD( z, dt );
#endif COOLING	
    
    //   zlsio_Printf(" Sweep: %d %d \n", z->zoneID, z->zoneTimeScaleStatus);
    
  } //z
	
  return(err);
}


// sweeps with sin/cos coordinate requirements at some level

#ifdef SPHERICAL
int zlsa_2DSplitSweepsSphere( ZonePtr z){
	
  int err = noErr;
	
	
  if (z != NULL){
    
    WHD         zoneW =  z->zoneW;
    
    GridCoords gc =  z->zoneGrid;
    ACoordPtr     x0;
    
#ifdef GRAVITY
    // get the zone field and allocate a 1D array for passing it
    // to the sweeps with interpolation coeffs set up
    Field   zoneF;
    zls_GetZoneField(&zoneF, z); // nothing is allocated here, so no dispose is needed later
    F1DArr          f0X = z->zoneF0X;
    //   F1DArr          f0Y = z->zoneF0Y;    
#endif GRAVITY
    
    W1DArr        w0X = z->zoneW0X;
    Wstar1DArr wStarX = z->zoneWStarX;
    Flx1DArr    fluxX = z->zone1DfluxX;
    
    W1DArr        w0Y = z->zoneW0Y;
    Wstar1DArr wStarY = z->zoneWStarY;
    Flx1DArr    fluxY = z->zone1DfluxY;
    
    Real       dt = z->zoneClock.dt;
    Real       g = 1.0;  
    // converion from angular to spatial coordinates: 
    // Cartesian = x, y, g = 1, 1, 2D Spherical = r, phi, g = 1, r
    // 2D Circular = r, phi, g = 1, r, Cylinder = z, r, g = 1, 1
    
    Integer   i, j;
    
    Integer is ;
    Integer ie ;
    Integer js ;
    Integer je ;
    
    //    Integer nbc = z->zoneBCells;
    Integer nc  = z->zoneClock.cycles % k2D;
    Integer iSweep, iScan;
    
    for( iSweep = 0; iSweep < k2D; iSweep++){
      iScan = iSweep + (nc * k2D); //
      
      if ( nSweepOrder[ iScan ] == kXDim ) {		
        
        //=== X - Sweep ===Spherical Radius R ============================================
        
        w0X.x    = gc.x;
#ifdef LAGRANGIAN
        x0 = gc.x0;
#else 
        x0 = gc.x;
#endif				
        wStarX.x = x0;
        
        js  = kScanStartY[iScan];
        je  = gc.nt[kYDim] - kScanStartY[iScan];
        
        g   = 1.0; // all first axis sweeps are spatial, so g = 1.0
        // x-sweeps
        for(j = js; j<je; j++){ // step in y
          
          w0X.y    = gc.y[j];
          
          err = zlsh_FillW1DFrom2DZoneWHDX ( w0X,  zoneW, j); // get a row
#ifdef DIVVHCORRECTION
          err = zlsh_FillW1DDivVFrom2DZoneX ( w0X, z, j); // get orthogonal divV components
#endif DIVVHCORRECTION
          
#ifdef GRAVITY
          err = zlsh_FillF1DFrom2DFieldX          ( f0X,  zoneF , j); // get a row, f0 will have BCs whether or not zone does            
          err = zlsppm_ReconstructPPMWLRStates1DF ( w0X , f0X, dt, g); // averages LR states over domain of dependence from PPM coeffs
#else
          err = zlsppm_ReconstructPPMWLRStates1D  ( w0X , dt, g); // averages LR states over domain of dependence from PPM coeffs
#endif GRAVITY
          err = RIEMANNSOLVER              ( wStarX, w0X ); // macro set up in zls_defs.h
          
#ifdef PRETEST_W_VEL
          
          is  = kZLS_LimitEvolveLimits;
          ie  = gc.nt[kXDim]-kZLS_LimitEvolveLimits;
          
          for ( i = is; i < ie; i++){
            if ( wStarX.v[i].a != 0.0 ) goto processr1;
          }
          
          continue;
        processr1:
#endif PRETEST_W_VEL
          
#ifdef GRAVITY
          err = zlsa_LagrangianEvolveSRadialW1DF ( w0X, wStarX,  f0X, dt );
#else
          err = zlsa_LagrangianEvolveSRadialW1D  ( w0X, wStarX, dt );
#endif GRAVITY
          err = zlsa_LagrangianRemapSRadialW1D   ( w0X, fluxX, x0);
          
#ifdef VISCOUS_REMAP
          if ( kZLS_SmoothAxes & kSmoothXFlag )
            err = zlsa_RemapSmoothW1D( w0X, dt);
#endif VISCOUS_REMAP
          
          err = zlsh_Fill2DZoneWHDFromW1DX ( zoneW,  w0X,  j ); // put it back without BCs
          z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
          
          
        } //j
        
        //=== X - Sweep ===Spherical Radius R ============================================
      }
      
      
      if ( nSweepOrder[ iScan ] == kYDim ) {		
        //=== Y - Phi Sweep ======================================================================
        
        w0Y.x    = gc.y;
#ifdef LAGRANGIAN
        x0 = gc.y0;
#else 
        x0 = gc.y;
#endif				
        wStarY.x = x0;
        
        is  =  kScanStartX[iScan];
        ie  = gc.nt[kXDim]- kScanStartX[iScan];
        
        for(i = is; i<ie; i++){
          
          g =  gc.x[i].c; // radius to scale dphi to spatial units,
          
          w0X.y    = gc.x[i];
          
          // no gravity in angular dimensions yet
          
          err = zlsh_FillW1DFrom2DZoneWHDY  ( w0Y,  zoneW, i); // get a column
#ifdef DIVVHCORRECTION
          err = zlsh_FillW1DDivVFrom2DZoneY ( w0Y, z, i);
#endif DIVVHCORRECTION
          
          err = zlsppm_ReconstructPPMWLRStates1D   ( w0Y , dt, g); // averages LR states over domain of dependence from PPM coeffs
          
          err = RIEMANNSOLVER               ( wStarY, w0Y ); // macro set up in zls_const.h
          
#ifdef PRETEST_W_VEL
          js  = kZLS_LimitEvolveLimits;
          je  = gc.nt[kYDim]-kZLS_LimitEvolveLimits;
          for ( j = js; j < je; j++){
            if ( wStarY.v[j].a != 0.0 ) goto processp1;
          };
          continue;
        processp1:
#endif PRETEST_W_VEL
          
          err = zlsa_LagrangianEvolvePhiW1D    ( w0Y, wStarY, g, dt );
          err = zlsa_LagrangianRemapPhiW1D     ( w0Y, fluxY, x0, g ); // remap w0,
          
#ifdef VISCOUS_REMAP
          if ( kZLS_SmoothAxes & kSmoothYFlag )
            err = zlsa_RemapSmoothW1D( w0Y, dt );
#endif VISCOUS_REMAP
          
          err = zlsh_Fill2DZoneWHDFromW1DY  ( zoneW, w0Y, i ); // put it back
          z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
          
        } //i
        
        //=== Y - Phi Sweep ======================================================================
      }
      
    } // iScan
    
#ifdef COOLING	
    err = zlsp_CoolZoneWHD  ( z, dt );
#endif COOLING	
    
  } //z
	
  return(err);
}
#endif SPHERICAL

// Z-R coordinates
#ifdef CYLINDRICAL

int zlsa_2DSplitSweepsCylinder( ZonePtr z){
	
  int err = noErr;
	
	
  if (z != NULL){
    
    WHD         zoneW =  z->zoneW;
    
    GridCoords gc =  z->zoneGrid;
    ACoordPtr     x0;
    
#ifdef GRAVITY
    // get the zone field and allocate a 1D array for passing it
    // to the sweeps with interpolation coeffs set up
    Field   zoneF;
    zls_GetZoneField(&zoneF, z); // nothing is allocated here, so no dispose is needed later
    F1DArr          f0X = z->zoneF0X;
    F1DArr          f0Y = z->zoneF0Y;    
#endif GRAVITY
    
    W1DArr        w0X = z->zoneW0X;
    Wstar1DArr wStarX = z->zoneWStarX;
    Flx1DArr    fluxX = z->zone1DfluxX;
    
    W1DArr        w0Y = z->zoneW0Y;
    Wstar1DArr wStarY = z->zoneWStarY;
    Flx1DArr    fluxY = z->zone1DfluxY;
    
    Real       dt = z->zoneClock.dt;
    Real       g = 1.0;  
    // converion from angular to spatial coordinates: 
    // Cartesian = x, y, g = 1, 1, 2D Spherical = r, phi, g = 1, r
    // 2D Circular = r, phi, g = 1, r, Cylinder = z, r, g = 1, 1
    
    Integer   i, j;
    
    Integer is ;
    Integer ie ;
    Integer js ;
    Integer je ;
    
    //   Integer nbc = z->zoneBCells;
    Integer nc  = z->zoneClock.cycles % kSymDims;
    Integer iSweep, iScan;
    
    for( iSweep = 0; iSweep < kSymDims; iSweep++){
      iScan = iSweep + (nc * kSymDims); //
      
      if ( nSweepOrder[ iScan ] == kXDim ) {		
        
        //=== X - Cylinder Z-Sweep ============================================================================
        
        w0X.x    = gc.x;
#ifdef LAGRANGIAN
        x0 = gc.x0;
#else 
        x0 = gc.x;
#endif				
        wStarX.x = x0;
        
        js  = kScanStartY[iScan];
        je  = gc.nt[kYDim] - kScanStartY[iScan];
        
        g   = 1.0; // all first axis sweeps are spatial, so g = 1.0
        // x-sweeps
        for(j = js; j<je; j++){ // step in y
          
          w0X.y    = gc.y[j];
          err = zlsh_FillW1DFrom2DZoneWHDX ( w0X,  zoneW, j); // get a row
#ifdef DIVVHCORRECTION
          err = zlsh_FillW1DDivVFrom2DZoneX ( w0X, z, j); // get orthogonal divV components
#endif DIVVHCORRECTION
          
#ifdef GRAVITY
          err = zlsh_FillF1DFrom2DFieldX          ( f0X, zoneF, j); // get a row, f0 will have BCs whether or not zone does            
          err = zlsppm_ReconstructPPMWLRStates1DF ( w0X,   f0X, dt, g); // averages LR states over domain of dependence from PPM coeffs
#else
          err = zlsppm_ReconstructPPMWLRStates1D  ( w0X, dt, g); // averages LR states over domain of dependence from PPM coeffs
#endif GRAVITY
          err = RIEMANNSOLVER              ( wStarX, w0X ); // macro set up in zls_defs.h
          
#ifdef PRETEST_W_VEL
          
          is  = kZLS_LimitEvolveLimits;
          ie  = gc.nt[kXDim]-kZLS_LimitEvolveLimits;
          
          for ( i = is; i < ie; i++){
            if ( wStarX.v[i].a != 0.0 ) goto process2dcz1;
          }
          
          continue;
        process2dcz1:
#endif PRETEST_W_VEL
          
#ifdef GRAVITY
          err = zlsa_LagrangianEvolveW1DF ( w0X, wStarX,  f0X, dt );
#else
          err = zlsa_LagrangianEvolveW1D  ( w0X, wStarX, dt );
#endif GRAVITY
          
          err = zlsa_LagrangianRemapW1D   ( w0X, fluxX, x0);
          
#ifdef VISCOUS_REMAP
          if ( kZLS_SmoothAxes & kSmoothXFlag )
            err = zlsa_RemapSmoothW1D( w0X, dt );
#endif VISCOUS_REMAP
          
          err = zlsh_Fill2DZoneWHDFromW1DX ( zoneW,  w0X,  j ); // put it back without BCs
          z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
          
        } //j
        
        //=== X - Sweep ============================================================================
      }
      
      
      if ( nSweepOrder[ iScan ] == kYDim ) {		
        //=== Y - Cylinder R Sweep ======================================================================
        
        w0Y.x    = gc.y;
#ifdef LAGRANGIAN
        x0 = gc.y0;
#else 
        x0 = gc.y;
#endif				
        wStarY.x = x0;
        
        is  =  kScanStartX[iScan];
        ie  = gc.nt[kXDim]- kScanStartX[iScan];
        
        g   = 1.0; // second axis sweeps are spatial too, so g = 1.0, but evolve and remap include divergence
        
        for(i = is; i<ie; i++){
          
          w0X.y    = gc.x[i];
          
				  err = zlsh_FillW1DFrom2DZoneWHDY       ( w0Y,  zoneW, i); // get a column            
#ifdef DIVVHCORRECTION
          err = zlsh_FillW1DDivVFrom2DZoneY      ( w0Y, z, i); // get orthogonal divV components
#endif DIVVHCORRECTION
				  
#ifdef GRAVITY
          err = zlsh_FillF1DFrom2DFieldY          ( f0Y,  zoneF , i); // get a row, f0 will have BCs whether or not zone does            
          err = zlsppm_ReconstructPPMWLRStates1DF ( w0Y , f0Y, dt, g); // averages LR states over domain of dependence from PPM coeffs
#else
				  err = zlsppm_ReconstructPPMWLRStates1D  ( w0Y , dt, g ); // averages LR states over domain of dependence from PPM coeffs
#endif GRAVITY
          
          err = RIEMANNSOLVER               ( wStarY, w0Y ); // macro set up in zls_const.h
          
#ifdef PRETEST_W_VEL
          js  = kZLS_LimitEvolveLimits;
          je  = gc.nt[kYDim]-kZLS_LimitEvolveLimits;
          for ( j = js; j < je; j++){
            if ( wStarY.v[j].a != 0.0 ) goto process2dcr1;
          };
          continue;
        process2dcr1:
#endif PRETEST_W_VEL
          
#ifdef GRAVITY
          err = zlsa_LagrangianEvolveCRadialW1DF ( w0Y, wStarY, f0Y, dt );
#else
				  err = zlsa_LagrangianEvolveCRadialW1D ( w0Y, wStarY, dt );
#endif GRAVITY
          
				  err = zlsa_LagrangianRemapCRadialW1D  ( w0Y, fluxY, x0  ); // remap w0,
          
#ifdef VISCOUS_REMAP
          if ( kZLS_SmoothAxes & kSmoothYFlag )
            err = zlsa_RemapSmoothW1D( w0Y, dt );
#endif VISCOUS_REMAP
          
          err = zlsh_Fill2DZoneWHDFromW1DY  ( zoneW, w0Y, i ); // put it back
          z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
          
        } //i
        
        //=== Y - Sweep ======================================================================
      }
      
    } // iScan
    
#ifdef COOLING	
    err = zlsp_CoolZoneWHD( z, dt );
#endif COOLING	
    
  } //z
	
  return(err);
}

#endif CYLINDRICAL

// Cylindrical R-Phi coordinates
#ifdef CIRCULAR

int zlsa_2DSplitSweepsCircle( ZonePtr z){
	
  int err = noErr;
	
	
  if (z != NULL){
    
    WHD         zoneW =  z->zoneW;
    
    GridCoords gc =  z->zoneGrid;
    ACoordPtr     x0;
    
#ifdef GRAVITY
    // get the zone field and allocate a 1D array for passing it
    // to the sweeps with interpolation coeffs set up
    Field   zoneF;
    zls_GetZoneField(&zoneF, z); // nothing is allocated here, so no dispose is needed later
    F1DArr          f0X = z->zoneF0X;
    F1DArr          f0Y = z->zoneF0Y;    
#endif GRAVITY
    
    W1DArr        w0X = z->zoneW0X;
    Wstar1DArr wStarX = z->zoneWStarX;
    Flx1DArr    fluxX = z->zone1DfluxX;
    
    W1DArr        w0Y = z->zoneW0Y;
    Wstar1DArr wStarY = z->zoneWStarY;
    Flx1DArr    fluxY = z->zone1DfluxY;
    
    Real       dt = z->zoneClock.dt;
    Real       g = 1.0;  
    // converion from angular to spatial coordinates: 
    // Cartesian = x, y, g = 1, 1, 2D Spherical = r, phi, g = 1, r
    // 2D Circular = r, phi, g = 1, r, Cylinder = z, r, g = 1, 1
    
    Integer   i, j;
    
    Integer is ;
    Integer ie ;
    Integer js ;
    Integer je ;
    
    //   Integer nbc = z->zoneBCells;
    Integer nc  = z->zoneClock.cycles % kSymDims;
    Integer iSweep, iScan;
    
    for( iSweep = 0; iSweep < kSymDims; iSweep++){
      iScan = iSweep + (nc * k2D); // 0-4
      
      if ( nSweepOrder[ iScan ] == kXDim ) {		
        
        //=== X - Cylinder R-Sweep ============================================================================
        
        w0X.x    = gc.x;
#ifdef LAGRANGIAN
        x0 = gc.x0;
#else 
        x0 = gc.x;
#endif				
        wStarX.x = x0;
        
        js  = kScanStartY[iScan];
        je  = gc.nt[kYDim] - kScanStartY[iScan];
        
        g   = 1.0; // Cylinder R is spatial, so g = 1.0
        // x-sweeps
        for(j = js; j<je; j++){ // step in y
          
          w0X.y    = gc.y[j];
          
          err = zlsh_FillW1DFrom2DZoneWHDX  ( w0X,  zoneW, j ); // get a row with BCs
#ifdef DIVVHCORRECTION
          err = zlsh_FillW1DDivVFrom2DZoneX ( w0X, z, j ); // get orthogonal divV components
#endif DIVVHCORRECTION
          
#ifdef GRAVITY
          err = zlsh_FillF1DFrom2DFieldX          ( f0X,  zoneF, j ); // get a row, f0 will have BCs whether or not zone does            
          err = zlsppm_ReconstructPPMWLRStates1DF ( w0X, f0X, dt, g); // averages LR states over domain of dependence from PPM coeffs
#else
          err = zlsppm_ReconstructPPMWLRStates1D  ( w0X, dt, g); // averages LR states over domain of dependence from PPM coeffs
#endif GRAVITY
          
          err = RIEMANNSOLVER              ( wStarX, w0X ); // macro set up in zls_defs.h
          
#ifdef PRETEST_W_VEL
          
          is  = kZLS_LimitEvolveLimits;
          ie  = gc.nt[kXDim]-kZLS_LimitEvolveLimits;
          
          for ( i = is; i < ie; i++){
            if ( wStarX.v[i].a != 0.0 ) goto process2dcr2;
          }
          
          continue;
        process2dcr2:
#endif PRETEST_W_VEL
          
          
#ifdef GRAVITY
          err = zlsa_LagrangianEvolveCRadialW1DF  ( w0X, wStarX, f0X, dt );
#else
          err = zlsa_LagrangianEvolveCRadialW1D   ( w0X, wStarX, dt );
#endif GRAVITY
          
          err = zlsa_LagrangianRemapCRadialW1D    ( w0X, fluxX, x0);
          
#ifdef VISCOUS_REMAP
          if ( kZLS_SmoothAxes & kSmoothXFlag )
            err = zlsa_RemapSmoothW1D( w0X, dt );
#endif VISCOUS_REMAP
          
          err = zlsh_Fill2DZoneWHDFromW1DX ( zoneW,  w0X,  j ); // put it back without BCs
          z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
          
        } //j
        
        //=== X - Sweep ============================================================================
      }
      
      
      if ( nSweepOrder[ iScan ] == kYDim ) {		
        //=== Y - Cylinder Phi Sweep ======================================================================
        
        w0Y.x    = gc.y;
#ifdef LAGRANGIAN
        x0 = gc.y0;
#else 
        x0 = gc.y;
#endif				
        wStarY.x = x0;
        
        is  =  kScanStartX[iScan];
        ie  = gc.nt[kXDim]- kScanStartX[iScan];
        
        for(i = is; i<ie; i++){
          
          w0X.y    = gc.x[i];
          g        = gc.x[i].c; // radius to scale dphi to spatial units
          
          err = zlsh_FillW1DFrom2DZoneWHDY        ( w0Y,  zoneW, i); // get a column            
#ifdef DIVVHCORRECTION
          err = zlsh_FillW1DDivVFrom2DZoneY       ( w0Y, z, i ); // get orthogonal divV components
#endif DIVVHCORRECTION
          err = zlsppm_ReconstructPPMWLRStates1D  ( w0Y , dt, g ); // averages LR states over domain of dependence from PPM coeffs
          
          err = RIEMANNSOLVER                     ( wStarY, w0Y ); // macro set up in zls_const.h
          
#ifdef PRETEST_W_VEL
          js  = kZLS_LimitEvolveLimits;
          je  = gc.nt[kYDim]-kZLS_LimitEvolveLimits;
          for ( j = js; j < je; j++){
            if ( wStarY.v[j].a != 0.0 ) goto process2dcp2;
          };
          continue;
        process2dcp2:
#endif PRETEST_W_VEL
          
          err = zlsa_LagrangianEvolvePhiW1D ( w0Y, wStarY, g,  dt);
          err = zlsa_LagrangianRemapPhiW1D  ( w0Y, fluxY, x0, g ); // remap w0,
          
#ifdef VISCOUS_REMAP
          if ( kZLS_SmoothAxes & kSmoothYFlag )
            err = zlsa_RemapSmoothW1D( w0Y, dt );
#endif VISCOUS_REMAP
          
          err = zlsh_Fill2DZoneWHDFromW1DY  ( zoneW, w0Y, i ); // put it back
          z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
          
        } //i
        
        //=== Y - Sweep ======================================================================
      }
      
    } // iScan
    
#ifdef COOLING	
    err = zlsp_CoolZoneWHD( z, dt );
#endif COOLING	
    
  } //z
	
  return(err);
}

#endif CIRCULAR


#endif SYM2D


#ifdef SYM3D

int zlsa_3DSplitSweeps( ZonePtr z){
	
  int err = noErr;
	
  Real      dt;
  
  Integer    i, j , k;
  
  Real g;  
  
  // conversion from angular to spatial coordinates: 
  // Cartesian = x, y, z g = 1, 1, 1, 3D Spherical = r, phi, theta g = 1, r Cos(theta), r
  // Cylinder = z, r, phi g = 1, 1, r  3D Polar Spherical = r, pTheta, Phi, g = 1, r, r Sin(theta)
	
  ACoordPtr     x0;
	
  if (z != NULL){
    
		// setup preamble, local references to zone vars
    
    WHD         zoneW =  z->zoneW;
    
    GridCoords gc =  z->zoneGrid;
    
#ifdef GRAVITY
    // get the zone field and allocate a 1D array for passing it
    // to the sweeps with interpolation coeffs set up
    Field   zoneF;
    zls_GetZoneField(&zoneF, z); // nothing is allocated here, so no dispose is needed later
    F1DArr          f0X = z->zoneF0X;
    F1DArr          f0Y = z->zoneF0Y;    
    F1DArr          f0Z = z->zoneF0Z;    
#endif GRAVITY
	  
    W1DArr        w0X = z->zoneW0X;
    Wstar1DArr wStarX = z->zoneWStarX;
    Flx1DArr    fluxX = z->zone1DfluxX;
    
    W1DArr        w0Y = z->zoneW0Y;
    Wstar1DArr wStarY = z->zoneWStarY;
    Flx1DArr    fluxY = z->zone1DfluxY;
    
    W1DArr        w0Z = z->zoneW0Z;
    Wstar1DArr wStarZ = z->zoneWStarZ;
    Flx1DArr    fluxZ = z->zone1DfluxZ;
    
    Integer is ;
    Integer ie ;
    Integer js ;
    Integer je ;
    Integer ks ;
    Integer ke ;
    
    dt  = z->zoneClock.dt;
    
    //   Integer nbc = z->zoneBCells;
    Integer nc  = z->zoneClock.cycles % k3D;
    Integer iSweep, iScan;
    
    for( iSweep = 0; iSweep < k3D; iSweep++){
      iScan = iSweep + (nc * k3D); // 0-9
      
      if ( nSweepOrder[ iScan ] == kXDim ) {		
        
        w0X.x    = gc.x;
#ifdef LAGRANGIAN
        x0 = gc.x0;
#else 
        x0 = gc.x;
#endif				
        wStarX.x = x0;
        
        g = 1.0;
        // x-sweeps are all spatial, so g = 1.0.  dA is may not be 1.0!
        
        ks  = kScanStartZ[iScan];
        ke  = gc.nt[kZDim]- kScanStartZ[iScan];        
        js  = kScanStartY[iScan];
        je  = gc.nt[kYDim]-kScanStartY[iScan];        
        
        for( j = js; j<je; j++ ){
          for( k = ks; k<ke; k++ ){
            
            err = zlsh_FillW1DXFrom3DZoneWHDYZ      ( w0X, zoneW, j, k); // get a row
#ifdef DIVVHCORRECTION
            err = zlsh_FillW1DDivVFrom3DZoneYZ       ( w0X, z, j, k); // get orthogonal divV components
#endif DIVVHCORRECTION
#ifdef GRAVITY
            err = zlsh_FillF1DXFrom3DFieldYZ        ( f0X,  zoneF, j, k); // get a row, f0 will have BCs whether or not zone does            
            err = zlsppm_ReconstructPPMWLRStates1DF ( w0X , f0X, dt, g); // averages LR states over domain of dependence from PPM coeffs
#else
            err = zlsppm_ReconstructPPMWLRStates1D  ( w0X , dt, g); // averages LR states over domain of dependence from PPM coeffs
#endif GRAVITY
            
            err = RIEMANNSOLVER                     ( wStarX, w0X ); // macro set up in zls_const.h
            
#ifdef PRETEST_W_VEL
            is  = kZLS_LimitEvolveLimits;
            ie  = gc.nt[kXDim]-kZLS_LimitEvolveLimits;
            for ( i= is; i < ie; i++){
              if ( ( wStarX.v[i].a) != 0.0 ) goto processx1_3d;
            }
            continue;
          processx1_3d:
#endif PRETEST_W_VEL
            
#ifdef GRAVITY
            err = zlsa_LagrangianEvolveW1DF  ( w0X, wStarX, f0X, dt );
#else
            err = zlsa_LagrangianEvolveW1D   ( w0X, wStarX, dt );
#endif GRAVITY
            err = zlsa_LagrangianRemapW1D    ( w0X, fluxX, x0 ); // remap w0,
            
#ifdef VISCOUS_REMAP
            if ( kZLS_SmoothAxes & kSmoothXFlag )
              err = zlsa_RemapSmoothW1D( w0X, dt );
#endif VISCOUS_REMAP
            
            err = zlsh_Fill3DZoneWHDYZFromW1DX      ( zoneW,  w0X, j, k ); // put it back
            z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
            
          } //k
        } //j
      }
      //==== End X ===========================================================================
      
      
      //=== Y - Sweep ==========================================================================
      if ( nSweepOrder[ iScan ] == kYDim ) {		
        
        w0Y.x    = gc.y;
#ifdef LAGRANGIAN
        x0 = gc.y0;
#else 
        x0 = gc.y;
#endif				
        wStarY.x = x0;
        g = 1.0;
        
        is  = kScanStartX[iScan];
        ie  = gc.nt[kXDim]-kScanStartX[iScan];
        ks  = kScanStartZ[iScan];
        ke  = gc.nt[kZDim]-kScanStartZ[iScan];        
        
        for(k = ks; k < ke; k++){
          for(i = is; i < ie; i++){
            
            err = zlsh_FillW1DYFrom3DZoneWHDZX       ( w0Y,  zoneW, k, i ); // get a column
#ifdef DIVVHCORRECTION
            err = zlsh_FillW1DDivVFrom3DZoneZX       ( w0Y, z, k, i); // get orthogonal divV components
#endif DIVVHCORRECTION
#ifdef GRAVITY
            err = zlsh_FillF1DYFrom3DFieldZX         ( f0Y,  zoneF , k, i); // get a row, f0 will have BCs whether or not zone does            
            err = zlsppm_ReconstructPPMWLRStates1DF  ( w0Y , f0Y, dt, g); // averages LR states over domain of dependence from PPM coeffs
#else
            err = zlsppm_ReconstructPPMWLRStates1D   ( w0Y , dt, g); // averages LR states over domain of dependence from PPM coeffs
#endif GRAVITY
            
            err = RIEMANNSOLVER                    ( wStarY, w0Y ); // macro set up in zls_const.h
            
#ifdef PRETEST_W_VEL
            js  = kZLS_LimitEvolveLimits;
            je  = gc.nt[kYDim]-kZLS_LimitEvolveLimits;
            for ( j = js; j < je; j++){
              if ( wStarY.v[j].a != 0.0) goto processy1_3d;
            }
            continue;
          processy1_3d:
#endif PRETEST_W_VEL
            
#ifdef GRAVITY
            err = zlsa_LagrangianEvolveW1DF  ( w0Y, wStarY, f0Y, dt );
#else
            err = zlsa_LagrangianEvolveW1D   ( w0Y, wStarY, dt );
#endif GRAVITY
            
            err = zlsa_LagrangianRemapW1D    ( w0Y, fluxY, x0 ); // remap w0,
            
#ifdef VISCOUS_REMAP
            if ( kZLS_SmoothAxes & kSmoothYFlag )
              err = zlsa_RemapSmoothW1D( w0Y, dt );
#endif VISCOUS_REMAP
            
            err = zlsh_Fill3DZoneWHDZXFromW1DY     ( zoneW, w0Y, k, i ); // put it back
            z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
            
          } //i
        } //k
      }
      //=== End Y =========================================================================
      
      //=== Z - Sweep : Z  etc =====================================================================
      
      if ( nSweepOrder[ iScan ] == kZDim ) {		
        
        is  = kScanStartX[iScan];
        ie  = gc.nt[kXDim]-kScanStartX[iScan];
        js  = kScanStartY[iScan];
        je  = gc.nt[kYDim]-kScanStartY[iScan];        
        
        w0Z.x    = gc.z;
#ifdef LAGRANGIAN
        x0 = gc.z0;
#else 
        x0 = gc.z;
#endif				
        wStarZ.x = x0;
        
        g       = 1.0;  // default g = 1.0, ie any spatial axis
        
        for(i = is; i<ie; i++){
          for(j = js; j<je; j++){
            
            err = zlsh_FillW1DZFrom3DZoneWHDXY     ( w0Z ,  zoneW, i ,j ); // get a z row
#ifdef DIVVHCORRECTION
            err = zlsh_FillW1DDivVFrom3DZoneXY     ( w0Z, z, i, j); // get orthogonal divV components
#endif DIVVHCORRECTION
            
#ifdef GRAVITY
            err = zlsh_FillF1DZFrom3DFieldXY        ( f0Z,  zoneF , i, j); // get a row, f0 will have BCs whether or not zone does            
            err = zlsppm_ReconstructPPMWLRStates1DF ( w0Z , f0Z, dt, g); // averages LR states over domain of dependence from PPM coeffs
#else
            err = zlsppm_ReconstructPPMWLRStates1D ( w0Z , dt, g); // averages LR states over domain of dependence from PPM coeffs
#endif GRAVITY
            
            err = RIEMANNSOLVER                    ( wStarZ, w0Z ); // macro set up in zls_const.h
            
#ifdef PRETEST_W_VEL
            ks  = kZLS_LimitEvolveLimits; //gc.ns[kZDim] - kEvolveLimits;
            ke  = gc.nt[kZDim]-kZLS_LimitEvolveLimits; //gc.ne[kZDim] + kEvolveLimits;
            for ( k = ks; k < ke; k++){
              if ( wStarZ.v[k].a != 0.0 ) goto processz1_3d;
            }
            continue;
          processz1_3d:
#endif PRETEST_W_VEL
            
#ifdef GRAVITY
            err = zlsa_LagrangianEvolveW1DF  ( w0Z, wStarZ, f0Z, dt );
#else
            err = zlsa_LagrangianEvolveW1D   ( w0Z, wStarZ, dt );
#endif GRAVITY
            err = zlsa_LagrangianRemapW1D     ( w0Z, fluxZ, x0 ); // remap w0,
            
#ifdef VISCOUS_REMAP
		        if ( kZLS_SmoothAxes & kSmoothZFlag )
              err = zlsa_RemapSmoothW1D( w0Z, dt );
#endif VISCOUS_REMAP
            
            err = zlsh_Fill3DZoneWHDXYFromW1DZ     ( zoneW, w0Z,  i , j ); // put it back
            z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
            
          } // j				
        } //i
        
      }
      //==== End Z =============================================================================
      
    }  // scans
    
    
#ifdef COOLING	
    err = zlsp_CoolZoneWHD( z, dt );
#endif COOLING	
		
  } //z
	
  return(err);
  
}

#ifdef SPHERICAL
int zlsa_3DSplitSweepsSphere( ZonePtr z){
	
  int err = noErr;
	
  if (z != NULL){
    
		// setup preamble, local references to zone vars
    
    WHD         zoneW =  z->zoneW;
    
    ACoordPtr     x0;
    GridCoords gc =  z->zoneGrid;
    
#ifdef GRAVITY
    // get the zone field and allocate a 1D array for passing it
    // to the sweeps with interpolation coeffs set up
    Field   zoneF;
    zls_GetZoneField(&zoneF, z); // nothing is allocated here, so no dispose is needed later
    F1DArr          f0X = z->zoneF0X;
    F1DArr          f0Y = z->zoneF0Y;    
    F1DArr          f0Z = z->zoneF0Z;    
#endif GRAVITY
	  
    W1DArr        w0X = z->zoneW0X;
    Wstar1DArr wStarX = z->zoneWStarX;
    Flx1DArr    fluxX = z->zone1DfluxX;
    
    W1DArr        w0Y = z->zoneW0Y;
    Wstar1DArr wStarY = z->zoneWStarY;
    Flx1DArr    fluxY = z->zone1DfluxY;
    
    W1DArr        w0Z = z->zoneW0Z;
    Wstar1DArr wStarZ = z->zoneWStarZ;
    Flx1DArr    fluxZ = z->zone1DfluxZ;
    
    Real g;
    //  
    // conversion from angular to spatial coordinates: g
    // Cartesian = x, y, z g = 1, 1, 1, 3D Spherical = r, phi, theta g = 1, r Cos(theta), r
    // Cylinder = z, r, phi g = 1, 1, r  3D Polar Spherical = r, pTheta, Phi, g = 1, r, r Sin(theta)
    //
    
    Integer    i, j , k;
    
    Integer is ;
    Integer ie ;
    Integer js ;
    Integer je ;
    Integer ks ;
    Integer ke ;
    
    Real dt  = z->zoneClock.dt;
    
    //Integer nbc = z->zoneBCells;
    Integer nc  = z->zoneClock.cycles % k3D;
    Integer iSweep, iScan;
    
    for( iSweep = 0; iSweep < k3D; iSweep++){
      iScan = iSweep + (nc * k3D); // 0-9
      
      //=== x - Sweep == Spherical Radius R==============================================
      if ( nSweepOrder[ iScan ] == kXDim ) {		
        
        w0X.x    = gc.x;
#ifdef LAGRANGIAN
        x0 = gc.x0;
#else 
        x0 = gc.x;
#endif				
        wStarX.x = x0;
        
        g = 1.0;
        
        ks  = kScanStartZ[iScan];
        ke  = gc.nt[kZDim]- kScanStartZ[iScan];        
        js  = kScanStartY[iScan];
        je  = gc.nt[kYDim]-kScanStartY[iScan];        
        
        for( k = ks; k<ke; k++ ){
          for( j = js; j<je; j++ ){
            
            err = zlsh_FillW1DXFrom3DZoneWHDYZ      ( w0X, zoneW, j, k); // get a row
#ifdef DIVVHCORRECTION
            err = zlsh_FillW1DDivVFrom3DZoneYZ       ( w0X, z, j, k); // get orthogonal divV components
#endif DIVVHCORRECTION
#ifdef GRAVITY
            err = zlsh_FillF1DXFrom3DFieldYZ        ( f0X,  zoneF , j, k); // get a row, f0 will have BCs whether or not zone does            
            err = zlsppm_ReconstructPPMWLRStates1DF ( w0X , f0X, dt, g); // averages LR states over domain of dependence from PPM coeffs
#else
            err = zlsppm_ReconstructPPMWLRStates1D ( w0X , dt, g); // averages LR states over domain of dependence from PPM coeffs
#endif GRAVITY
            
            err = RIEMANNSOLVER                     ( wStarX, w0X ); // macro set up in zls_const.h
            
            
#ifdef PRETEST_W_VEL
            is  = kZLS_LimitEvolveLimits;
            ie  = gc.nt[kXDim]-kZLS_LimitEvolveLimits;
            for ( i= is; i < ie; i++){
              if ( ( wStarX.v[i].a) != 0.0 ) goto process3dr1;
            }
            continue;
          process3dr1:
#endif PRETEST_W_VEL
            
#ifdef GRAVITY
            err = zlsa_LagrangianEvolveSRadialW1DF  ( w0X, wStarX, f0X, dt );
#else
            err = zlsa_LagrangianEvolveSRadialW1D   ( w0X, wStarX, dt );
#endif GRAVITY
            zlsa_LagrangianRemapSRadialW1D              ( w0X,  fluxX, x0);
            
#ifdef VISCOUS_REMAP
            if ( kZLS_SmoothAxes & kSmoothXFlag )
              err = zlsa_RemapSmoothW1D( w0X, dt  );
#endif VISCOUS_REMAP
            
            err = zlsh_Fill3DZoneWHDYZFromW1DX      ( zoneW,  w0X, j, k ); // put it back
            z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
            
          } //j
        } //k
      }
      //=== x - Sweep == Spherical Radius R==============================================
      
      
      //=== Y - Sweep == Azimuth Phi ======================================================
      if ( nSweepOrder[ iScan ] == kYDim ) {		
        
        w0Y.x    = gc.y;
#ifdef LAGRANGIAN
        x0 = gc.y0;
#else 
        x0 = gc.y;
#endif				
        wStarY.x = x0;
        
        is  = kScanStartX[iScan];
        ie  = gc.nt[kXDim]-kScanStartX[iScan];
        ks  = kScanStartZ[iScan];
        ke  = gc.nt[kZDim]-kScanStartZ[iScan];        
        
        for(k = ks; k < ke; k++){
          for(i = is; i < ie; i++){
            
            g =  gc.x[i].c * gc.z[k].cc; // radius cos theta to scale dPhi to spatial units
            
            err = zlsh_FillW1DYFrom3DZoneWHDZX       ( w0Y,  zoneW, k, i ); // get a column
#ifdef DIVVHCORRECTION
            err = zlsh_FillW1DDivVFrom3DZoneZX       ( w0Y, z, k, i); // get orthogonal divV components
#endif DIVVHCORRECTION
            // no non-radial gravity in 3d spherical coords yet
            err = zlsppm_ReconstructPPMWLRStates1D   ( w0Y , dt, g); // averages LR states over domain of dependence from PPM coeffs
            
            err = RIEMANNSOLVER                    ( wStarY, w0Y ); // macro set up in zls_const.h
            
#ifdef PRETEST_W_VEL
            js  = kZLS_LimitEvolveLimits;
            je  = gc.nt[kYDim]-kZLS_LimitEvolveLimits;
            for ( j = js; j < je; j++){
              if ( wStarY.v[j].a != 0.0) goto proces3dsp1;
            }
            continue;
          proces3dsp1:
#endif PRETEST_W_VEL
            
            err = zlsa_LagrangianEvolvePhiW1D  ( w0Y, wStarY, g, dt );
            err = zlsa_LagrangianRemapPhiW1D   ( w0Y, fluxY, x0, g ); // remap w0,
            
#ifdef VISCOUS_REMAP
            if ( kZLS_SmoothAxes & kSmoothYFlag )
              err = zlsa_RemapSmoothW1D( w0Y, dt );
#endif VISCOUS_REMAP
            
            err = zlsh_Fill3DZoneWHDZXFromW1DY     ( zoneW, w0Y, k, i ); // put it back
            z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
            
          } //i
        } //k
      }
      //=== Y - Sweep == Azimuth Phi ======================================================
      
      //=== Z - Sweep :  Equatorial Lattitude Theta  ========================================
      
      if ( nSweepOrder[ iScan ] == kZDim ) {		
        
        is  = kScanStartX[iScan];
        ie  = gc.nt[kXDim]-kScanStartX[iScan];
        js  = kScanStartY[iScan];
        je  = gc.nt[kYDim]-kScanStartY[iScan];        
        
        w0Z.x    = gc.z;
#ifdef LAGRANGIAN
        x0 = gc.z0;
#else 
        x0 = gc.z;
#endif				
        wStarZ.x = x0;
        
        
        for(j = js; j<je; j++){
          for(i = is; i<ie; i++){
            
            g =  gc.x[i].c ; // radius to scale dTheta to spatial units
            
            err = zlsh_FillW1DZFrom3DZoneWHDXY     ( w0Z ,  zoneW, i ,j ); // get a z row
#ifdef DIVVHCORRECTION
            err = zlsh_FillW1DDivVFrom3DZoneXY     ( w0Z, z, i, j); // get orthogonal divV components
#endif DIVVHCORRECTION
            // no non-radial gravity in 3d spherical coords
            err = zlsppm_ReconstructPPMWLRStates1D ( w0Z , dt, g); // averages LR states over domain of dependence from PPM coeffs
            
            err = RIEMANNSOLVER                    ( wStarZ, w0Z ); // macro set up in zls_const.h
            
#ifdef PRETEST_W_VEL
            ks  = kZLS_LimitEvolveLimits; //gc.ns[kZDim] - kEvolveLimits;
            ke  = gc.nt[kZDim]-kZLS_LimitEvolveLimits; //gc.ne[kZDim] + kEvolveLimits;
            for ( k = ks; k < ke; k++){
              if ( wStarZ.v[k].a != 0.0 ) goto process3dst1;
            }
            continue;
          process3dst1:
#endif PRETEST_W_VEL
            
            // no non-radial gravity in 3d spherical coords
            
            zlsa_LagrangianEvolveThetaW1D          ( w0Z, wStarZ, g, dt );
            zlsa_LagrangianRemapThetaW1D           ( w0Z, fluxZ, x0, g);
            
#ifdef VISCOUS_REMAP
		        if ( kZLS_SmoothAxes & kSmoothZFlag )
              err = zlsa_RemapSmoothW1D( w0Z, dt );
#endif VISCOUS_REMAP
            
            err = zlsh_Fill3DZoneWHDXYFromW1DZ     ( zoneW, w0Z,  i , j ); // put it back
            z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
            
          } //i
        } // j				
        
      }
      //=== Z - Sweep :  Equatorial Lattitude Theta  ========================================
      
    }  // scans
    
    
#ifdef COOLING	
    err = zlsp_CoolZoneWHD( z, dt );
#endif COOLING	
		
  } //z
	
  return(err);
  
} // zlsa_3DSplitSweepsSphere

#endif SPHERICAL

#ifdef POLARSPHERICAL
int zlsa_3DSplitSweepsPolarSphere( ZonePtr z){
	
  int err = noErr;
	
  if (z != NULL){
    
		// setup preamble, local references to zone vars
    
    WHD         zoneW =  z->zoneW;
    
    ACoordPtr     x0;
    GridCoords gc =  z->zoneGrid;
    
#ifdef GRAVITY
    // get the zone field and allocate a 1D array for passing it
    // to the sweeps with interpolation coeffs set up
    Field   zoneF;
    zls_GetZoneField(&zoneF, z); // nothing is allocated here, so no dispose is needed later
    F1DArr          f0X = z->zoneF0X;
    F1DArr          f0Y = z->zoneF0Y;    
    F1DArr          f0Z = z->zoneF0Z;    
#endif GRAVITY
	  
    W1DArr        w0X = z->zoneW0X;
    Wstar1DArr wStarX = z->zoneWStarX;
    Flx1DArr    fluxX = z->zone1DfluxX;
    
    W1DArr        w0Y = z->zoneW0Y;
    Wstar1DArr wStarY = z->zoneWStarY;
    Flx1DArr    fluxY = z->zone1DfluxY;
    
    W1DArr        w0Z = z->zoneW0Z;
    Wstar1DArr wStarZ = z->zoneWStarZ;
    Flx1DArr    fluxZ = z->zone1DfluxZ;
    
    Real g;
    //  
    // conversion from angular to spatial coordinates: g
    // Cartesian = x, y, z g = 1, 1, 1, 3D Spherical = r, phi, theta g = 1, r Cos(theta), r
    // Cylinder = z, r, phi g = 1, 1, r  3D Polar Spherical = r, pTheta, Phi, g = 1, r, r Sin(theta)
    //
    
    Integer    i, j , k;
    
    Integer is ;
    Integer ie ;
    Integer js ;
    Integer je ;
    Integer ks ;
    Integer ke ;
    
    Real dt  = z->zoneClock.dt;
    
    Integer nbc = z->zoneBCells;
    Integer nc  = z->zoneClock.cycles % k3D;
    Integer iSweep, iScan;
    
    for( iSweep = 0; iSweep < k3D; iSweep++){
      iScan = iSweep + (nc * k3D); // 0-9
      
      //=== x - Sweep == Spherical Radius R==============================================
      if ( nSweepOrder[ iScan ] == kXDim ) {		
        
        w0X.x    = gc.x;
#ifdef LAGRANGIAN
        x0 = gc.x0;
#else 
        x0 = gc.x;
#endif				
        wStarX.x = x0;
        
        g = 1.0;
        
        ks  = kScanStartZ[iScan];
        ke  = gc.nt[kZDim]- kScanStartZ[iScan];        
        js  = kScanStartY[iScan];
        je  = gc.nt[kYDim]-kScanStartY[iScan];        
        
        for( k = ks; k<ke; k++ ){
          for( j = js; j<je; j++ ){
            
            err = zlsh_FillW1DXFrom3DZoneWHDYZ      ( w0X, zoneW, j, k); // get a row
#ifdef DIVVHCORRECTION
            err = zlsh_FillW1DDivVFrom3DZoneYZ       ( w0X, z, j, k); // get orthogonal divV components
#endif DIVVHCORRECTION
#ifdef GRAVITY
            err = zlsh_FillF1DXFrom3DFieldYZ        ( f0X,  zoneF , j, k); // get a row, f0 will have BCs whether or not zone does            
            err = zlsppm_ReconstructPPMWLRStates1DF ( w0X , f0X, dt, g); // averages LR states over domain of dependence from PPM coeffs
#else
            err = zlsppm_ReconstructPPMWLRStates1D ( w0X , dt, g); // averages LR states over domain of dependence from PPM coeffs
#endif GRAVITY
            
            err = RIEMANNSOLVER                     ( wStarX, w0X ); // macro set up in zls_const.h
            
            
#ifdef PRETEST_W_VEL
            is  = kZLS_LimitEvolveLimits;
            ie  = gc.nt[kXDim]-kZLS_LimitEvolveLimits;
            for ( i= is; i < ie; i++){
              if ( ( wStarX.v[i].a) != 0.0 ) goto processx1_3d;
            }
            continue;
          processx1_3d:
#endif PRETEST_W_VEL
            
#ifdef GRAVITY
            err = zlsa_LagrangianEvolveSRadialW1DF  ( w0X, wStarX, f0X, dt );
#else
            err = zlsa_LagrangianEvolveSRadialW1D   ( w0X, wStarX, dt );
#endif GRAVITY
            zlsa_LagrangianRemapSRadialW1D              ( w0X,  fluxX, x0);
            
#ifdef VISCOUS_REMAP
            if ( kZLS_SmoothAxes & kSmoothXFlag )
              err = zlsa_RemapSmoothW1D( w0X, dt  );
#endif VISCOUS_REMAP
            
            err = zlsh_Fill3DZoneWHDYZFromW1DX      ( zoneW,  w0X, j, k ); // put it back
            z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
            
          } //j
        } //k
      }
      //=== x - Sweep == Spherical Radius R==============================================
      
      
      //=== Y - Sweep ==Polar Theta ======================================================
      if ( nSweepOrder[ iScan ] == kYDim ) {		
        
        w0Y.x    = gc.y;
#ifdef LAGRANGIAN
        x0 = gc.y0;
#else 
        x0 = gc.y;
#endif				
        wStarY.x = x0;
        
        is  = kScanStartX[iScan];
        ie  = gc.nt[kXDim]-kScanStartX[iScan];
        ks  = kScanStartZ[iScan];
        ke  = gc.nt[kZDim]-kScanStartZ[iScan];        
        
        for(k = ks; k < ke; k++){
          for(i = is; i < ie; i++){
            
            g =  gc.x[i].c; // radius  to scale dpTheta to spatial units
            
            err = zlsh_FillW1DYFrom3DZoneWHDZX       ( w0Y,  zoneW, k, i ); // get a column
#ifdef DIVVHCORRECTION
            err = zlsh_FillW1DDivVFrom3DZoneZX       ( w0Y, z, k, i); // get orthogonal divV components
#endif DIVVHCORRECTION
            // no non-radial gravity in 3d spherical coords yet
            err = zlsppm_ReconstructPPMWLRStates1D   ( w0Y , dt, g); // averages LR states over domain of dependence from PPM coeffs
            
            err = RIEMANNSOLVER                      ( wStarY, w0Y ); // macro set up in zls_const.h
            
#ifdef PRETEST_W_VEL
            js  = kZLS_LimitEvolveLimits;
            je  = gc.nt[kYDim]-kZLS_LimitEvolveLimits;
            for ( j = js; j < je; j++){
              if ( wStarY.v[j].a != 0.0) goto proces3dst2;
            }
            continue;
          proces3dst2:
#endif PRETEST_W_VEL
            
            err = zlsa_LagrangianEvolvePolarThetaW1D  ( w0Y, wStarY, g, dt );
            err = zlsa_LagrangianRemapPolarThetaW1D   ( w0Y, fluxY, x0, g ); // remap w0,
            
#ifdef VISCOUS_REMAP
            if ( kZLS_SmoothAxes & kSmoothYFlag )
              err = zlsa_RemapSmoothW1D( w0Y, dt );
#endif VISCOUS_REMAP
            
            err = zlsh_Fill3DZoneWHDZXFromW1DY     ( zoneW, w0Y, k, i ); // put it back
            z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
            
          } //i
        } //k
      }
      //=== Y - Sweep ==Polar Theta ======================================================
      
      //=== Z - Sweep ==Azimuth Phi ======================================================
      if ( nSweepOrder[ iScan ] == kZDim ) {		
        
        is  = kScanStartX[iScan];
        ie  = gc.nt[kXDim]-kScanStartX[iScan];
        js  = kScanStartY[iScan];
        je  = gc.nt[kYDim]-kScanStartY[iScan];        
        
        w0Z.x    = gc.z;
#ifdef LAGRANGIAN
        x0 = gc.z0;
#else 
        x0 = gc.z;
#endif				
        wStarZ.x = x0;
        
        
        for(j = js; j<je; j++){
          for(i = is; i<ie; i++){
            
            g =  gc.x[i].c * gc.y[j].sc; // radius sin(ptheta) to scale dphi to spatial units
            
            err = zlsh_FillW1DZFrom3DZoneWHDXY     ( w0Z ,  zoneW, i ,j ); // get a z row
#ifdef DIVVHCORRECTION
            err = zlsh_FillW1DDivVFrom3DZoneXY      ( w0Z, z, i, j); // get orthogonal divV components
#endif DIVVHCORRECTION
            
            // no non-radial gravity in 3d spherical coords
            
            err = zlsppm_ReconstructPPMWLRStates1DG ( w0Z , dt, g); // averages LR states over domain of dependence from PPM coeffs
            
            err = RIEMANNSOLVER                    ( wStarZ, w0Z ); // macro set up in zls_const.h
            
#ifdef PRETEST_W_VEL
            ks  = kZLS_LimitEvolveLimits; //gc.ns[kZDim] - kEvolveLimits;
            ke  = gc.nt[kZDim]-kZLS_LimitEvolveLimits; //gc.ne[kZDim] + kEvolveLimits;
            for ( k = ks; k < ke; k++){
              if ( wStarZ.v[k].a != 0.0 ) goto process3dsp2;
            }
            continue;
          process3dsp2:
#endif PRETEST_W_VEL
            
            // no non-radial gravity in 3d spherical coords
            
            zlsa_LagrangianEvolvePhiW1D          ( w0Z, wStarZ, g, dt );
            zlsa_LagrangianRemapPhiW1D           ( w0Z, fluxZ, x0, g);
            
#ifdef VISCOUS_REMAP
		        if ( kZLS_SmoothAxes & kSmoothZFlag )
              err = zlsa_RemapSmoothW1D( w0Z, dt );
#endif VISCOUS_REMAP
            
            err = zlsh_Fill3DZoneWHDXYFromW1DZ     ( zoneW, w0Z,  i , j ); // put it back
            z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
            
          } //i
        } // j				
        
      }
      //=== Z - Sweep == Azimuth Phi ======================================================
      
    }  // scans
    
    
#ifdef COOLING	
    err = zlsp_CoolZoneWHD( z, dt );
#endif COOLING	
		
  } //z
	
  return(err);
  
} // zlsa_3DSplitSweepsPolarSphere

#endif POLARSPHERICAL

#ifdef CYLINDRICAL
int zlsa_3DSplitSweepsCylinder( ZonePtr z){
	
  int err = noErr;
	
  Real      dt;
	
  Integer   nbc, nc;
  Integer    i, j , k;
  
  Real g;  
  // converion from angular to spatial coordinates: 
  // Cylinder = z, r, phi g = 1, 1, r
	
  ACoordPtr     x0;	
	
  if (z != NULL){
    
    WHD         zoneW =  z->zoneW;
	  
    W1DArr        w0X = z->zoneW0X;
    Wstar1DArr wStarX = z->zoneWStarX;
    Flx1DArr    fluxX = z->zone1DfluxX;
    
    W1DArr        w0Y = z->zoneW0Y;
    Wstar1DArr wStarY = z->zoneWStarY;
    Flx1DArr    fluxY = z->zone1DfluxY;
    
    W1DArr        w0Z = z->zoneW0Z;
    Wstar1DArr wStarZ = z->zoneWStarZ;
    Flx1DArr    fluxZ = z->zone1DfluxZ;
    
    Integer is ;
    Integer ie ;
    Integer js ;
    Integer je ;
    Integer ks ;
    Integer ke ;
    
    nbc = z->zoneBCells;
    dt  = z->zoneClock.dt;
    nc  = z->zoneClock.cycles % k3D;
    
    for ( iSweep = 0; iSweep < k3D; iSweep++){
      Integer iScan = iSweep+(nc*k3D);
      if ( nSweepOrder[ iScan ] == kXDim ) {		
        
        //=== X - Sweep : Cylinder Z ===================================================================
        
        w0X.x    = gc.x;
#ifdef LAGRANGIAN
        x0 = gc.x0;
#else 
        x0 = gc.x;
#endif				
        wStarX.x = x0;
        
        g = 1.0;
        // x-sweeps are all spatial, so g = 1.0.  dA is  not 1.0!
        
        ks  = kZLS_LimitSecondSweepScan;
        ke  = gc.nt[kZDim]-kZLS_LimitSecondSweepScan;        
        js  = kZLS_LimitSecondSweepScan;
        je  = gc.nt[kYDim]-kZLS_LimitSecondSweepScan;        
        
        for( k = ks; k<ke; k++ ){
          for( j = js; j<je; j++ ){
            
            err = zlsh_FillW1DXFrom3DZoneWHDYZ      ( w0X, zoneW, j, k); // get a row
#ifdef DIVVHCORRECTION
            err = zlsh_FillW1DDivVFrom3DZoneYZ        ( w0X, z, j, k); // get orthogonal divV components
#endif DIVVHCORRECTION
            
            err = zlsppm_ReconstructPPMWLRStates1D  ( w0X , dt, g); // averages LR states over domain of dependence from PPM coeffs
            
            err = RIEMANNSOLVER ( wStarX, w0X ); // Note: macro set up in zls_defs.h
            
#ifdef PRETEST_W_VEL
            is  = kZLS_LimitEvolveLimits;
            ie  = gc.nt[kXDim]-kZLS_LimitEvolveLimits;
            for ( i = is; i < ie; i++ ){
              if ( wStarX.v[i].a != 0.0 ) goto process3dr1;
            };
            continue;
          process3dr1:
#endif PRETEST_W_VEL
            err = zlsa_LagrangianEvolveW1D  ( w0X, wStarX, dt );
            err = zlsa_LagrangianRemapW1D   ( w0X,  fluxX, x0);
#ifdef VISCOUS_REMAP
            if ( kZLS_SmoothAxes & kSmoothXFlag )
              err = zlsa_RemapSmoothW1D( w0X, dt );
#endif VISCOUS_REMAP
            err = zlsh_Fill3DZoneWHDYZFromW1DX          ( zoneW,  w0X, j, k ); // put it back
            z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
            
          } //j
        } //k
        
      }
      
      //=== End Cylinder Z ===================================================================
      
      
      //=== Y - Sweep : Cylinder Radius ===================================================================
      
      if ( nSweepOrder[ iScan] == kYDim ) {		
        
        w0Y.x    = gc.y;
#ifdef LAGRANGIAN
        x0 = gc.y0;
#else 
        x0 = gc.y;
#endif				
        wStarY.x = x0;
        g = 1.0;
        
        is  = kZLS_LimitFirstSweepScan;
        ie  = gc.nt[kXDim] - kZLS_LimitFirstSweepScan;
        ks  = kZLS_LimitSecondSweepScan;
        ke  = gc.nt[kZDim]-kZLS_LimitSecondSweepScan;        
        
        for(k = ks; k < ke; k++){
          for(i = is; i < ie; i++){
            
            err = zlsh_FillW1DYFrom3DZoneWHDZX     ( w0Y,  zoneW, k, i ); // get a column
#ifdef DIVVHCORRECTION
            err = zlsh_FillW1DDivVFrom3DZoneZX     ( w0Y, z, K, i); // get orthogonal divV components
#endif DIVVHCORRECTION
            err = zlsppm_ReconstructPPMWLRStates1D ( w0Y, dt, g ); // averages LR states over domain of dependence from PPM coeffs
            
            err = RIEMANNSOLVER                    ( wStarY, w0Y ); // macro set up in zls_const.h
            
#ifdef PRETEST_W_VEL
            js  = kZLS_LimitEvolveLimits;
            je  = gc.nt[kYDim]-kZLS_LimitEvolveLimits;
            for ( j = js; j < je; j++ ){
              if ( wStarY.v[j].a != 0.0 ) goto process3dphi1;
            };
            continue;
          process3dphi1:
#endif PRETEST_W_VEL
            err = zlsa_LagrangianEvolveCRadialW1D  ( w0Y, wStarY, g, dt );
            err = zlsa_LagrangianRemapCRadialW1D   ( w0Y, fluxY, x0, g ); // remap w0,
            
#ifdef VISCOUS_REMAP
            if ( kZLS_SmoothAxes & kSmoothYFlag )
              err = zlsa_RemapSmoothW1D( w0Y, dt );
#endif VISCOUS_REMAP
            
            err = zlsh_Fill3DZoneWHDZXFromW1DY ( zoneW, w0Y, k, i ); // put it back
            z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
            
          } //i
        } //k
        
      }
      //=== End Cylinder Radius ===================================================================
      
      
      //=== Z - Sweep : Phi ===================================================================
      
      if ( nSweepOrder[ iScan ] == kZDim ) {		
        
        w0Z.x    = gc.z;
#ifdef LAGRANGIAN
        x0 = gc.z0;
#else 
        x0 = gc.z;
#endif				
        wStarZ.x = x0;
        g       = 1.0;
        
        is  = kZLS_LimitFirstSweepScan;
        ie  = gc.nt[kXDim] - kZLS_LimitFirstSweepScan;
        js  = kZLS_LimitFirstSweepScan;
        je  = gc.nt[kYDim] - kZLS_LimitFirstSweepScan;
        
        for(j = js; j<je; j++){
          g =  gc.y[j].c; // radius to scale dphi to spatial units
          for(i = is; i<ie; i++){
            
            err = zlsh_FillW1DZFrom3DZoneWHDXY     ( w0Z,  zoneW, i ,j ); // get a z row
#ifdef DIVVHCORRECTION
            err = zlsh_FillW1DDivVFrom3DZoneYZ       ( w0Z, z, i, j); // get orthogonal divV components
#endif DIVVHCORRECTION
            err = zlsppm_ReconstructPPMWLRStates1D ( w0Z , dt, g ); // averages LR states over domain of dependence from PPM coeffs
            
            err = RIEMANNSOLVER                    ( wStarZ, w0Z ); // macro set up in zls_const.h
            
#ifdef PRETEST_W_VEL
            ks  = kZLS_LimitEvolveLimits; //gc.ns[kZDim] - kEvolveLimits;
            ke  = gc.nt[kZDim]-kZLS_LimitEvolveLimits; //gc.ne[kZDim] + kEvolveLimits;
            for ( k = ks; k < ke; k++){
              if ( wStarZ.v[k].a != 0.0 ) goto process3dt1;
            }
            continue;
          process3dt1:
#endif PRETEST_W_VEL
            err = zlsa_LagrangianEvolvePhiW1D          ( w0Z, wStarZ, g, dt );
            err = zlsa_LagrangianRemapPhiW1D           ( w0Z, fluxZ, x0, g);
            
#ifdef VISCOUS_REMAP
            if ( kZLS_SmoothAxes & kSmoothZFlag )
              err = zlsa_RemapSmoothW1D( w0Z, dt );
#endif VISCOUS_REMAP
            
            err = zlsh_Fill3DZoneWHDXYFromW1DZ        ( zoneW,  w0Z, i ,j ); // put it back
            z->zoneTimeScaleStatus = kZLS_InValid; // things changed, will need new timestep
            
          } //i
        } // j				
        
      }
      //==== End Phi ==========================================================================
      
    }
    
#ifdef COOLING	
    err = zlsp_CoolZoneWHD( z, dt );
#endif COOLING	
    
    
  } //z
	
  return(err);
}
#endif CYLINDRICAL

#endif SYM3D

#endif SPLITSWEEPS

#endif HYDRO


int zlsa_CalcZoneDivV( ZonePtr z ){
  //
  // compute centered approx for magntude of the
  // multi-D divergence^2 in velocity
  //  = (-4/3 (dv))^2 for dv < 0
  // plus only keeping values for compression regions
  // for use in artificial viscosity and smoothing code.
  //
  int err = noErr;
#ifdef LAGRANGIAN
#ifdef SPLITSWEEPS
#ifdef DIVVHCORRECTION
  
  if (z != NULL){
    
    WHD zoneW = z->zoneW;    
    
#ifdef SYM1D
    register Real    a ;
    register Integer nti   = z->zoneGrid.nt[kXDim];
    register Integer i, il, ir;
    
    for(i = 0; i< nti; i++){
      il = (i>0)?(i-1):i;
      ir = (i<(nti-1))?(i+1):i;
      a  = -fourthd*zls_min( zoneW.vx[ir] - zoneW.vx[il], 0.0); // non zero for compression ( ie shocks)
      z->zoneDivV[i].a = a*a;
    } //i
#endif SYM1D
    
#ifdef SYM2D
    
    register Real    a ;
    register Real    b ;
    register Integer nti   = z->zoneGrid.nt[kXDim];
    register Integer ntj   = z->zoneGrid.nt[kYDim];
    
    register Integer i, il, ir;
    register Integer j, jl, jr;
    
    for(j = 0; j < ntj; j++){
      jl = (j >0)?j-1:j;
      jr = (j < (ntj-1))?j+1:j;
      for(i = 0; i< nti; i++){
        il = (i >0)?i-1:i;
        ir = (i < (nti-1))?i+1:i;
        a  = -fourthd*zls_min( zoneW.vx[j ][ir] - zoneW.vx[j ][il], 0.0);
        b  = -fourthd*zls_min( zoneW.vy[jr][i ] - zoneW.vy[jl][i ], 0.0);
        z->zoneDivV[j][i].a[kXDim] = a*a;
        z->zoneDivV[j][i].a[kYDim] = b*b;
      } //i
    } //j
    
#endif SYM2D
    
#ifdef SYM3D
    
    register Real    a ;
    register Real    b ;
    register Real    c ;
    register Integer nti   = z->zoneGrid.nt[kXDim];
    register Integer ntj   = z->zoneGrid.nt[kYDim];
    register Integer ntk   = z->zoneGrid.nt[kZDim];
    
    register Integer i, il, ir;
    register Integer j, jl, jr;
    register Integer k, kl, kr;
    
    for(k = 0; k < ntk; k++){
      kl = (k >0)?k-1:k;
      kr = (k < (ntk-1))?k+1:k;
      for(j = 0; j < ntj; j++){
        jl = (j >0)?j-1:j;
        jr = (j < (ntj-1))?j+1:j;
        for(i = 0; i< nti; i++){
          il = (i >0)?i-1:i;
          ir = (i < (nti-1))?i+1:i;
          
          a  = -fourthd*zls_min( zoneW.vx[k ][j ][ir] - zoneW.vx[k ][j ][il], 0.0);
          b  = -fourthd*zls_min( zoneW.vy[k ][jr][i ] - zoneW.vy[k ][jl][i ], 0.0);
          c  = -fourthd*zls_min( zoneW.vz[kr][j ][i ] - zoneW.vz[kl][j ][i ], 0.0);
          
          z->zoneDivV[k][j][i].a[kXDim] = a*a;
          z->zoneDivV[k][j][i].a[kYDim] = b*b;
          z->zoneDivV[k][j][i].a[kZDim] = c*c;
          
        } //i
      } //j
    } //k
#endif SYM3D
    
  } else { err = zonErr; } //z
  
#endif LAGRANGIAN
#endif SPLITSWEEPS
#endif DIVVHCORRECTION
  
  return(err);
  
}



int zlsa_CalcZonePressureRatio( Real *dP, ZonePtr z ){
  // find the Pmax and Pmin and return Pmax/Pmin  
  int err = noErr;
  
  *dP = 1.0;
  
  if (z != NULL){
    WHD zoneW = z->zoneW;    
    
    Real P, d, T;
    Real Pmax = kZLS_PressureVacuum;
    Real Pmin = kZLS_Huge;
    
#ifdef SYM1D
    Integer nti   = z->zoneGrid.nt[kXDim];
    Integer i;
    for(i = 0; i< nti; i++){
      P    = zoneW.P[i];
      d    = zoneW.d[i];
      T    = P/d;
      Pmax = (Pmax <  T)?T:Pmax;
      Pmin = (Pmin >= T)?T:Pmin;
    } //i
#endif SYM1D
    
#ifdef SYM2D
    Integer nti   = z->zoneGrid.nt[kXDim];
    Integer ntj   = z->zoneGrid.nt[kYDim];
    Integer i, j;
    for(j = 0; j < ntj; j++){
      for(i = 0; i< nti; i++){
        P    = zoneW.P[j][i];
        d    = zoneW.d[j][i];
        T    = P/d;
        Pmax = (Pmax <  T)?T:Pmax;
        Pmin = (Pmin >= T)?T:Pmin;
      } //i
    } //j
#endif SYM2D
    
#ifdef SYM3D
    Integer nti   = z->zoneGrid.nt[kXDim];
    Integer ntj   = z->zoneGrid.nt[kYDim];
    Integer ntk   = z->zoneGrid.nt[kZDim];
    Integer i, j, k;
    for(k = 0; k < ntk; k++){
      for(j = 0; j < ntj; j++){
        for(i = 0; i< nti; i++){
          P    = zoneW.P[k][j][i];
          d    = zoneW.d[k][j][i];
          T    = P/d;
          Pmax = (Pmax <  T)?T:Pmax;
          Pmin = (Pmin >= T)?T:Pmin;
        } //i
      } //j
    } //k
#endif SYM3D
    
    *dP =zls_tolx(Pmax/Pmin, kZLS_Small);
    
  } else { err = zonErr; }//z
  
  return(err);
  
}


int zlsa_CalcLevelPressureRatio  (Real * dP, Integer level){
  
  //
  // collect a level wide ratio of greatest Max P/Min P for all level zones
  //
  
  int err = noErr;
  
  Integer         zone;
  ZonePtr            z;
  
  Real levelDP = kZLS_Huge; // set large to capture overall min
  Real zoneDP  = *dP;      // set to prev value to allow accellerator limiter to work
  
  
  if (gLevels[level].nZones>0){
    
    for( zone = 0; zone < gLevels[level].nZones; zone++){
      z =  gLevels[level].levelZones[zone];
      if (z != NULL){
        
#ifdef MPI  
        if (z->zoneRankID == gMPIrank) {
#endif MPI
          err = zlsa_CalcZonePressureRatio(&zoneDP, z);
          levelDP = levelDP > zoneDP ? zoneDP : levelDP;
#ifdef MPI      
        }
#endif MPI
        
      } else { err = zonErr; }
    } // zone loop
    
  } else { err = zonErr; }
  
  
#ifdef MPI  
  Real globalDeltaP;    
  MPI_Allreduce( &levelDP, &globalDeltaP, 1, MPI_DOUBLE, MPI_MIN, MPI_COMM_WORLD);
  //  if ( levelDP != globalDeltaP) zlsio_Printf(" Rank: %d  globaldP: %g\n", gMPIrank, levelDP, globalDeltaP);
  *dP = globalDeltaP;
#else 
  *dP = levelDP;
#endif !MPI
  
  return(err);
  
}




