#ifndef ZLS_SWEEPS_H
#define ZLS_SWEEPS_H
#include <zls.h>

enum { kSweepAllLevels       = 0, // used by refined sweep
       kSweepMainLevel       = 0, // used by direct sweep
       kSubLevelIterations   = 0, // used by refined sweep
       kNoSubLevelIterations = 0 // used by direct sweep
	}; 

//
// Active grid values begin at ns[dim] and end at ne[dim]-1 The sweeps
// are performed in loops to load, evolve, remap and finally replace
// the cell values into the grid.
//
// To allow for different numbers of boundary cells, and give the
// option of performing more calculations to move active cell values
// further from incomplete boundary cells, the offset constants are be
// defined as indexes in from the absolute edges, 0 and nt, ie we loop
// from 0 + k to < nti-k.  This means that k <= nBC. and the minimum
// nBC is therefore 4.
//
// if nBC > 4 then the actual grid cells get more cell padding of
// accurate boundary values.  nBC (i.e. ns[dim] should be 4 or more)
//
enum { 
  kZLS_LimitFirstSweepScan  = 0, // perpendicular axes limits for first sweep
  kZLS_LimitSecondSweepScan = 4, // perpendicular axes limits for subsequent sweep/s = Replace limit from sweep0
  kZLS_LimitEvolveLimits    = 2, // only evolve cells wilth full order PPM coeffs
  kZLS_LimitRemapLimits     = 3, // remap cells bounded by fully evolved cells
  kZLS_LimitReplaceLimits   = 4  // active grid cells are replaced.
};

#ifdef UNSPLITSWEEP

//====================================================================================
//
// Dimensionally UnSplit Eulerian Hydrodynamic Sweeps of a zone, 
// uses W, primitive variables
// 
// zone must have timestep ready for hydro updates.  
//
//====================================================================================

#ifdef SYM1D
int zlsa_1DSweep          ( ZonePtr z);
int zlsa_1DSweepSphere    ( ZonePtr z);
int zlsa_1DSweepCircle    ( ZonePtr z);
#endif SYM1D
//
#ifdef SYM2D
int zlsa_2DSweep            ( ZonePtr z); // cartesian
int zlsa_2DSweepSphere      ( ZonePtr z);
int zlsa_2DSweepPolarSphere ( ZonePtr z);
int zlsa_2DSweepCylinder    ( ZonePtr z);
int zlsa_2DSweepCircular    ( ZonePtr z);
#endif SYM2D
//
#ifdef SYM3D
int zlsa_3DSweep            ( ZonePtr z); // cartesian
int zlsa_3DSweepSphere      ( ZonePtr z);
int zlsa_3DSweepPolarSphere ( ZonePtr z);
int zlsa_3DSweepCylinder    ( ZonePtr z);
#endif SYM3D

#endif UNSPLITSWEEP

#ifdef SPLITSWEEPS
//====================================================================================
//
// Dimensionally Split Lagrangian  Hydrodynamic Sweeps of a zone, 
// uses W, primitive variables
// 
// zone must have timestep ready for hydro updates.  
//
//====================================================================================

#ifdef SYM1D
int zlsa_1DSweep          ( ZonePtr z);
int zlsa_1DSweepSphere    ( ZonePtr z);
int zlsa_1DSweepCircle    ( ZonePtr z);
#endif SYM1D
//
#ifdef SYM2D
int zlsa_2DSplitSweeps            ( ZonePtr z); // cartesian
int zlsa_2DSplitSweepsSphere      ( ZonePtr z);
int zlsa_2DSplitSweepsPolarSphere ( ZonePtr z);
int zlsa_2DSplitSweepsCylinder    ( ZonePtr z);
int zlsa_2DSplitSweepsCircular     ( ZonePtr z);
#endif SYM2D
//
#ifdef SYM3D
int zlsa_3DSplitSweeps            ( ZonePtr z); // cartesian
int zlsa_3DSplitSweepsSphere      ( ZonePtr z);
int zlsa_3DSplitSweepsPolarSphere ( ZonePtr z);
int zlsa_3DSplitSweepsCylinder    ( ZonePtr z);
#endif SYM3D
#endif SPLITSWEEPS

//====================================================================================
// -4/3 (dv/dx) for artificial viscosity calculations

int zlsa_CalcZoneDivV( ZonePtr z );

//====================================================================================
//  First versions of pressure range checking routines for adaptive level deactivation 
// use.  If the pressure ratio across each zone in a whole level is small enough,
// then turn off the level
//
//
int zlsa_CalcZonePressureRatio  ( Real * dP, ZonePtr z );
int zlsa_CalcLevelPressureRatio ( Real * dP, Integer level);

#endif ZLS_SWEEPS_H
