#include "zls.h"
//
//========================================
//
// ZLS bottleneck print routines to allow logging of TEXT io.
// 

static int gUserLogFileIsOpen = 0; // 0 = not open yet, 1 = open for writing
static int gLogFileIsOpen = 0; // 0 = not open yet, 1 = open for writing
static int gInfoFileIsOpen = 0; // 0 = not open yet, 1 = open for writing
static int gFilesFileIsOpen = 0; // 0 = not open yet, 1 = open for writing

// sends print to stdout and the log file, only to screen on rank 0 for MPI
void zlsio_Printf (const char *template, ...){
  va_list ap;
  va_start (ap, template);
#if (ZLS_VERBOSE > 0)
#ifdef MPI
  if ( gMPIrank == 0) {
#endif MPI
    vfprintf (stdout, template, ap);
#ifdef MPI
  }
#endif MPI
#endif 
  // if (gLogFileIsOpen) vfprintf (zls_IOLOGFILE, template, ap);
  va_end (ap);
  va_start (ap, template);
  if (gLogFileIsOpen) vfprintf (zls_IOLOGFILE, template, ap);
  va_end (ap);
}

// sends print to stderr and the log file, 1 file per rank
void zlsio_ErrPrintf (const char *template, ...){
  va_list ap;
  va_start (ap, template);
  vfprintf (stderr, template, ap);
  va_end (ap);
  va_start (ap, template);
  if (gLogFileIsOpen) vfprintf (zls_IOLOGFILE, template, ap);
  va_end (ap);
}

// sends print to  the log file, 1 file per rank
void zlsio_LogPrintf (const char *template, ...){
  va_list ap;
  va_start (ap, template);
  if (gLogFileIsOpen) vfprintf (zls_IOLOGFILE, template, ap);
  va_end (ap);
}

// sends print to  the user log file, only 1 file, so treat MPI properly
void zlsio_UserLogPrintf (const char *template, ...){
  va_list ap;
  va_start (ap, template);
#ifdef MPI
  if ( gMPIrank == 0) {
#endif MPI
   if (gUserLogFileIsOpen) vfprintf (zls_IOUSERLOGFILE, template, ap);
#ifdef MPI
  }
#endif MPI
  va_end (ap);
}

// sends print to  the info file, 1 file per rank
void zlsio_InfoPrintf (const char *template, ...){
  va_list ap;
  va_start (ap, template);
  if (gInfoFileIsOpen) vfprintf (zls_IOINFOFILE, template, ap);
  va_end (ap);
}

// sends print to  the files file, 1 file per rank
void zlsio_FilesPrintf (const char *template, ...){
  va_list ap;
  va_start (ap, template);
  if (gFilesFileIsOpen) vfprintf (zls_IOFILESFILE, template, ap);
  va_end (ap);
}

//
// zlsio_FPrintf is no different to fprintf, but have it here for consistency
//
// Default to not log fprintf calls, but could be turned on by uncommenting line
// for v low level debugging.
//
void zlsio_FPrintf ( FILE *f, const char *template, ...){
  va_list ap;
  va_start (ap, template);
#ifdef MPI
  if (( gMPIrank == 0)||( f != stdout )) {
#endif MPI
    vfprintf ( f, template, ap);
#ifdef MPI
  }
#endif MPI
 // if (gLogFileIsOpen) vfprintf (zls_IOLOGFILE, template, ap);
  va_end (ap);
}

//
// zlsio_ZFPrintf uses gzip to compress the text output file, not intended for 
// stdout, just .gz files.  Used on user dump files, but not log files or .conf
// files input or output.
// if IO compression is off then this will operate normally, writing uncompressed
// text files.
//
void zlsio_ZFPrintf ( zFile f, const char *template, ...){
#ifdef MPI
  if (( gMPIrank == 0)||( f != stdout )) {
#endif MPI
    char buf[4096];
    Size len;
    va_list ap;
    va_start (ap, template);
    len = vsprintf(buf, template, ap);
    va_end(ap);
    zlsio_write ( f, (void *)buf, len );
#ifdef MPI
  }
#endif MPI
}

//========================================

int zlsio_CreateLogFile ( int newFile ){
  int err = noErr;
  IOSettingsPtr io = NULL;
  gLogFileIsOpen = 0;
  
  err =  zls_GetLevelIOPrefs ( &io, 0 ); // top level has main ref to IO
  if ( io != NULL ) {
    
    char logFile[PATHLEN];
#ifdef MPI
    // in MPI each rank has a separate log file.
    sprintf(logFile, "%s%s%s%4.4d%s", io->outDir, PATHSEP, io->outLogName,gMPIrank, LOGEXT);
#else
    sprintf(logFile, "%s%s%s%s", io->outDir, PATHSEP, io->outLogName, LOGEXT);
#endif !MPI
    
    if ( newFile == kZLS_CreateNewLogFile ){
      io->outLogFILE = fopen(logFile, "w");
    } else if ( newFile == kZLS_ContinueLogFile ){
      io->outLogFILE = fopen(logFile, "a");
    } else {
      io->outLogFILE = fopen(logFile, "w");
    }
    if ( io->outLogFILE == NULL)   { err = fileErr; return(err); }
    
    err = setvbuf ( io->outLogFILE, NULL, _IOLBF, BUFSIZ ); // set for line buffering
    gLogFileIsOpen = 1;
    
    // keep log open until sim ends
    
  } else { err = fileErr; }
  
  return ( err );
}


int zlsio_CreateUserLogFile ( int newFile ){
  int err = noErr;
  IOSettingsPtr io = NULL;
  
  gUserLogFileIsOpen = 0;
#ifdef MPI  
  if (gMPIrank == 0) {
#endif MPI
    
    err =  zls_GetLevelIOPrefs ( &io, 0 ); // top level has main ref to IO
    if ( io != NULL ) {
      
      char UserlogFile[PATHLEN];
      sprintf(UserlogFile, "%s%s%s%s", io->outDir, PATHSEP, io->outUserLogName, LOGEXT);
      
      if ( newFile == kZLS_CreateNewLogFile ){
        io->outUserLogFILE = fopen(UserlogFile, "w");
      } else if ( newFile == kZLS_ContinueLogFile ){
        io->outUserLogFILE = fopen(UserlogFile, "a");
      } else {
        io->outUserLogFILE = fopen(UserlogFile, "w");
      }
      if ( io->outUserLogFILE == NULL)    { err = fileErr; return(err); }
      
      err = setvbuf ( io->outUserLogFILE, NULL, _IOLBF, BUFSIZ ); // set for line buffering
      gUserLogFileIsOpen = 1;
      
      // keep log open until sim ends
      
    } else { err = fileErr; }
#ifdef MPI  
  }
#endif MPI
  
  return ( err );
}



int zlsio_CreateInfoFile ( int newFile ){
  int err = noErr;
  IOSettingsPtr io = NULL;
  gInfoFileIsOpen = 0;
  
#ifdef MPI  
  if (gMPIrank == 0) {
#endif MPI
    
    err =  zls_GetLevelIOPrefs ( &io, 0 ); // top level has main ref to IO
    if ( io != NULL ) {
      
      char infoFile[PATHLEN];
      sprintf(infoFile, "%s%s%s%s", io->outDir, PATHSEP, io->outInfoName, INFOEXT);
      
      if ( newFile == kZLS_CreateNewInfoFile ){
        io->outInfoFILE = fopen(infoFile, "w");
      } else if ( newFile == kZLS_ContinueInfoFile ){
        io->outInfoFILE = fopen(infoFile, "a");
      } else {
        io->outInfoFILE = fopen(infoFile, "w");
      }
      if ( io->outInfoFILE == NULL)    { err = fileErr; return(err); }
      
      err = setvbuf ( io->outInfoFILE, NULL, _IOLBF, BUFSIZ ); // set for line buffering
      gInfoFileIsOpen = 1;
      
      // keep log open until sim ends
      
    } else { err = fileErr; }
    
#ifdef MPI  
  }
#endif MPI
  
  return ( err );
}


int zlsio_CreateFilesFile ( int newFile ){
  int err = noErr;
  IOSettingsPtr io = NULL;
  
  err =  zls_GetLevelIOPrefs ( &io, 0 ); // top level has main ref to IO
  if ( io != NULL ) {
    
    char fFile[PATHLEN];
#ifdef MPI
    // in MPI each rank has a separate log file.
    sprintf(fFile, "%s%s%s%4.4d%s", io->outDir, PATHSEP, io->outFilesName, gMPIrank, FILESEXT);
#else
    sprintf(fFile, "%s%s%s%s", io->outDir, PATHSEP, io->outFilesName, FILESEXT);
#endif !MPI
    
    if ( newFile == kZLS_CreateNewFilesFile ){
      io->outFilesFILE = fopen(fFile, "w");
    } else if ( newFile == kZLS_ContinueFilesFile ){
      io->outFilesFILE = fopen(fFile, "a");
    } else {
      io->outFilesFILE = fopen(fFile, "w");
    }
    if ( io->outFilesFILE == NULL)    { err = fileErr; return(err); }
    
    err = setvbuf ( io->outFilesFILE, NULL, _IOLBF, BUFSIZ ); // set for line buffering
    gFilesFileIsOpen = 1;
    
    // keep log open until sim ends
    
  } else { err = fileErr; }
  
  return ( err );
}


int zlsio_CloseLogFile ( void ){
  int err = noErr;
  // use macro because we are closing down
  if ( zls_IOLOGFILE ) {
    // keep log open until sim ends, so call this at the end of main
    fclose( zls_IOLOGFILE );
    gLogFileIsOpen = 0;
  }
  
  return ( err );
}

int zlsio_CloseUserLogFile ( void ){
  int err = noErr;
  // use macro because we are closing down
  if ( zls_IOUSERLOGFILE ) {
    // keep log open until sim ends, so call this at the end of main
    fclose( zls_IOUSERLOGFILE );
    gUserLogFileIsOpen = 0;
  }
  
  return ( err );
}


int zlsio_CloseInfoFile ( void ){
  int err = noErr;
  // use macro because we are closing down
  if ( zls_IOINFOFILE ) {
    // keep log open until sim ends, so call this at the end of main
    fclose( zls_IOINFOFILE );
    gInfoFileIsOpen = 0;
  }
  
  return ( err );
}


int zlsio_CloseFilesFile ( void ){
  int err = noErr;
  // use macro because we are closing down
  if ( zls_IOFILESFILE ) {
    // keep log open until sim ends, so call this at the end of main
    fclose( zls_IOFILESFILE );
    gFilesFileIsOpen = 0;
  }
  
  return ( err );
}

//
//========================================
//
//
//

int   zlsio_DisplaySimLimits( FILE * f , char * fname, Integer banner ){
  
  if (banner) {
#if (ZLS_VERBOSE > 0) 
    zlsio_FPrintf( f, "\n%s\n", DIVIDERLINE );
#endif
    zlsio_FPrintf( f, " ZLS Simulation Limits from %s \n", fname);
#if (ZLS_VERBOSE > 0) 
    zlsio_FPrintf( f, "%s\n", DIVIDERLINE );
#endif
  } else {
    
    
    zlsio_FPrintf( f, " Real Time Duration         : %s\n", zls_TimeIntervalString((Real) zlst_GetRealTimeDuration()));
    zlsio_FPrintf( f, "%s\n", DIVIDERLINE );
    zlsio_FPrintf( f, "   End Time: %#12.6g     End Cycle: 0x%8.8x\n",
                  gLevels[0].levelClock.endTime, (unsigned int)gLevels[0].levelClock.endCycles );    
    zlsio_FPrintf( f, "%s\n",  DIVIDERLINE );
    TimeCoords tc = gLevels[0].levelClock;
    zlsio_FPrintf( f, " End Time                   : %12.6g \n",    tc.endTime);
#if (ZLS_VERBOSE>0)
    zlsio_FPrintf( f, " End Cycles                 : 0x%8.8x \n",    tc.endCycles);
#endif (ZLS_VERBOSE>0)
    zlsio_FPrintf( f, " User File Time  Interval   : %12.6g \n",   tc.userTime);
#if (ZLS_VERBOSE>0)
    zlsio_FPrintf( f, " User File Cycle Interval   : 0x%8.8x \n",   tc.userCycles);
#endif (ZLS_VERBOSE>0)
    zlsio_FPrintf( f, " Image Time  Interval       : %12.6g \n",  tc.frameTime);
#if (ZLS_VERBOSE>0)
    zlsio_FPrintf( f, " Image Cycle Interval       : 0x%8.8x \n",  tc.frameCycles);
#endif (ZLS_VERBOSE>0)
    
    zlsio_FPrintf( f, " Data Time  Interval        : %12.6g \n",   tc.dataTime);
#if (ZLS_VERBOSE>0)
    zlsio_FPrintf( f, " Data Cycle Interval        : 0x%8.8x \n",   tc.dataCycles);
#endif (ZLS_VERBOSE>0)
    
    zlsio_FPrintf( f, " Restart Save Time  Interval: %12.6g \n",   tc.saveTime);
#if (ZLS_VERBOSE>0)
    zlsio_FPrintf( f, " Restart Save Cycle Interval: 0x%8.8x \n",   tc.saveCycles);
#endif (ZLS_VERBOSE>0)
    zlsio_FPrintf( f, "%s\n",  DIVIDERLINE );
    
  }
  
  return (noErr);
  
}

int   zlsio_DisplayIOSetup  ( FILE * f , char * fname, Integer banner ){
  
  if (banner) {
#if (ZLS_VERBOSE > 0) 
    zlsio_FPrintf( f, "\n%s\n", DIVIDERLINE );
#endif
    zlsio_FPrintf( f, " ZLS File IO Settings Initialised from %s \n", fname);
#if (ZLS_VERBOSE > 0) 
    zlsio_FPrintf( f, "%s\n", DIVIDERLINE );
#endif
  } else {
    
  }
  
  return (noErr);
  
}



int   zlsio_DisplayPhysicsSetup  ( FILE * f , char * fname, char * cfile, Integer banner ){
  
  if (banner) {
#if (ZLS_VERBOSE > 0) 
    zlsio_FPrintf( f, "\n%s\n", DIVIDERLINE );
#endif
    zlsio_FPrintf( f, " ZLS Physics Initialised from %s \n", fname);
#if (ZLS_VERBOSE > 0) 
    zlsio_FPrintf( f, "%s\n", DIVIDERLINE );
#endif
  } else {
    
#ifdef COOLING
    zlsp_PrintPhysicsConfiguration( f, cfile);
#else
    zlsp_PrintPhysicsConfiguration( f, NULL);
#endif
    
  }
  
  return (noErr);
  
}


int   zlsio_DisplayGridSetup( FILE * f , char * fname, Integer banner ){
  
  if (banner) {
    zlsio_FPrintf( f, "\n%s\n", DIVIDERLINE );
    zlsio_FPrintf( f, " ZLS Zone Level System Initialised from %s \n", fname );
#if (ZLS_VERBOSE > 0) 
    zlsio_FPrintf( f, "%s\n", DIVIDERLINE );
#endif
  } else {
    
    Real ncells = 0.0;
    
    zls_CalcAllLevelsEffectiveCellNumber ( f, &ncells, kZLS_ShowCellCount );
    
    zlsio_FPrintf( f, "%s\n",  DIVIDERLINE );
    
  }
  
  return (noErr);
  
}

//
//========================================
//

int zlsio_LogSetup ( char * zlsConfName, char * ppmConfName, 
                    char * gridConfName, char * physConfName, 
                    char * initConfName, char * ioConfName, 
                    char * restartConfName ){
#pragma unused( zlsConfName)
#pragma unused( restartConfName)
#pragma unused( ppmConfName)
  
  zlsio_DisplayGridSetup     ( zls_IOLOGFILE , gridConfName, 1 );
  zlsio_PrintConfiguration   ( zls_IOLOGFILE , "" );
  zlsio_DisplayGridSetup     ( zls_IOLOGFILE , gridConfName, 0 );
  
  zlsio_DisplayPhysicsSetup  ( zls_IOLOGFILE , physConfName, NULL, 1 );
  zlsio_DisplayPhysicsSetup  ( zls_IOLOGFILE , physConfName, NULL, 0 );
  
  zlsio_DisplaySimLimits     ( zls_IOLOGFILE , initConfName, 1 );
  zlsio_DisplaySimLimits     ( zls_IOLOGFILE , initConfName, 0 );
  
  zlsio_DisplayIOSetup       ( zls_IOLOGFILE , ioConfName, 1 );
  fprintf( zls_IOLOGFILE, "%s\n",  DIVIDERLINE );
  
  
  return (noErr);  
}


int zlsio_SpoolTextFile(zFileNC destfile, char * srcName);
int zlsio_SpoolTextFile(zFileNC destfile, char * srcName){
  char textline[PATHLEN];
  
  zFileNC srcfile =  zlsio_openNC ( srcName, "r");
  if (( destfile != NULL ) && ( srcfile != NULL )){
    
    while (!feof(srcfile)){
      fgets( textline, PATHLEN, srcfile);
      if ( textline[0] != '#') {
        if (strlen(textline)>1) fputs( textline, destfile);
      }
    }
    
    zlsio_closeNC( srcfile );
  };
  
  return(noErr);
  
}

int zlsio_RecordConfiguration ( char * zlsConfName, char * ppmConfName, 
                               char * gridConfName, char * physConfName, 
                               char * initConfName, char * ioConfName, 
                               char * restartConfName ){  
#ifdef MPI  
  if (gMPIrank == 0) {
#endif MPI
    
    if ( zls_IOINFOFILE != NULL) {
      
      fprintf( zls_IOINFOFILE, "#%s\n",  DIVIDERLINE );
      fprintf( zls_IOINFOFILE, "# ZLS CONFIGURATION RECORD. \n" );
      fprintf( zls_IOINFOFILE, "# Fyris Version: %s \n", FYRIS_VERSION );
      time_t  startTime = time(NULL);
      fprintf( zls_IOINFOFILE, "# %s", ctime(&startTime));
      fprintf( zls_IOINFOFILE, "# %s\n", FYRIS_PROBLEMNAME);
      fprintf( zls_IOINFOFILE, "#%s\n", DIVIDERLINE );
      
      fprintf( zls_IOINFOFILE, "#%s\n",  DIVIDERLINE );
      fprintf( zls_IOINFOFILE, "# ZLS CONFFILE. \n" );
      fprintf( zls_IOINFOFILE, "#%s\n", DIVIDERLINE );
      zlsio_SpoolTextFile(zls_IOINFOFILE, zlsConfName);
      fprintf( zls_IOINFOFILE, "#%s\n",  DIVIDERLINE );
      fprintf( zls_IOINFOFILE, "# PPM CONFFILE. \n" );
      fprintf( zls_IOINFOFILE, "#%s\n", DIVIDERLINE );
      zlsio_SpoolTextFile(zls_IOINFOFILE, ppmConfName);
      fprintf( zls_IOINFOFILE, "#%s\n",  DIVIDERLINE );
      fprintf( zls_IOINFOFILE, "# PHYSICS CONFFILE. \n" );
      fprintf( zls_IOINFOFILE, "#%s\n", DIVIDERLINE );
      zlsio_SpoolTextFile(zls_IOINFOFILE, physConfName);
      fprintf( zls_IOINFOFILE, "#%s\n",  DIVIDERLINE );
      fprintf( zls_IOINFOFILE, "# INIT CONFFILE. \n" );
      fprintf( zls_IOINFOFILE, "#%s\n", DIVIDERLINE );
      zlsio_SpoolTextFile(zls_IOINFOFILE, initConfName);
      fprintf( zls_IOINFOFILE, "#%s\n",  DIVIDERLINE );
      fprintf( zls_IOINFOFILE, "# IO CONFFILE. \n" );
      fprintf( zls_IOINFOFILE, "#%s\n", DIVIDERLINE );
      zlsio_SpoolTextFile(zls_IOINFOFILE, ioConfName);
      fprintf( zls_IOINFOFILE, "#%s\n",  DIVIDERLINE );
      fprintf( zls_IOINFOFILE, "# GRID CONFFILE. \n" );
      fprintf( zls_IOINFOFILE, "#%s\n", DIVIDERLINE );
      zlsio_SpoolTextFile(zls_IOINFOFILE, gridConfName);
      fprintf( zls_IOINFOFILE, "#%s\n", DIVIDERLINE );
      //    zlsio_SpoolTextFile(zls_IOINFOFILE, restartConfName);
      
    }
#ifdef MPI  
  }
#endif MPI
  
  return (noErr);  
}


int zlsio_StartLog ( void ){
  int err = noErr;
  time_t  startTime = time(NULL);
  if ( zls_IOLOGFILE ) {
  fprintf( zls_IOLOGFILE, "%s\n",  DIVIDERLINE );
  fprintf( zls_IOLOGFILE, " ZLS LOG START. \n" );
  fprintf( zls_IOLOGFILE, " Fyris Version: %s \n", FYRIS_VERSION );
  fprintf( zls_IOLOGFILE, " %s", ctime(&startTime));
  fprintf( zls_IOLOGFILE, "%s\n",  DIVIDERLINE );
  }
#ifdef MPI
  if ( gMPIrank == 0) {
#endif MPI
  if ( zls_IOUSERLOGFILE ) {
    fprintf( zls_IOUSERLOGFILE, "%s\n",  DIVIDERLINE );
    fprintf( zls_IOUSERLOGFILE, " ZLS USER LOG START. \n" );
    fprintf( zls_IOUSERLOGFILE, " Fyris Version: %s \n", FYRIS_VERSION );
    fprintf( zls_IOUSERLOGFILE, " %s", ctime(&startTime));
    fprintf( zls_IOUSERLOGFILE, "%s\n",  DIVIDERLINE );
    }
#ifdef MPI
  }
#endif MPI
  
  if ( zls_IOFILESFILE ) {
  fprintf( zls_IOFILESFILE, "%s\n",  DIVIDERLINE );
  fprintf( zls_IOFILESFILE, " ZLS DATA FILES LOG START. \n" );
  fprintf( zls_IOFILESFILE, " Fyris Version: %s \n", FYRIS_VERSION );
  fprintf( zls_IOFILESFILE, " %s", ctime(&startTime));
  fprintf( zls_IOFILESFILE, "%s\n",  DIVIDERLINE );
  }
  
  return (err);
}

int zlsio_EndLog ( void ){
  int err = noErr;
  time_t  startTime = time(NULL);
  
  if ( zls_IOLOGFILE ) {
  fprintf( zls_IOLOGFILE, "%s\n",  DIVIDERLINE );
  fprintf( zls_IOLOGFILE, " ZLS LOG CLOSED. \n" );
  fprintf( zls_IOLOGFILE, " %s", ctime(&startTime));
  fprintf( zls_IOLOGFILE, "%s\n",  DIVIDERLINE );
  }
#ifdef MPI
  if ( gMPIrank == 0) {
#endif MPI
  if ( zls_IOUSERLOGFILE ) {
    fprintf( zls_IOUSERLOGFILE, "%s\n",  DIVIDERLINE );
    fprintf( zls_IOUSERLOGFILE, " ZLS LOG CLOSED. \n" );
    fprintf( zls_IOUSERLOGFILE, " %s", ctime(&startTime));
    fprintf( zls_IOUSERLOGFILE, "%s\n",  DIVIDERLINE );
  }
#ifdef MPI
  }
#endif MPI
  
  if ( zls_IOUSERLOGFILE ) {
  fprintf( zls_IOFILESFILE, "%s\n",  DIVIDERLINE );
  fprintf( zls_IOLOGFILE, " ZLS DATA FILES LOG CLOSED. \n" );
  fprintf( zls_IOFILESFILE, " %s", ctime(&startTime));
  fprintf( zls_IOFILESFILE, "%s\n",  DIVIDERLINE );
  }
  
  return (err);
}

int zlsio_StartBanner ( void ){
  int err = noErr;
#if (ZLS_VERBOSE > 0)
  time_t  startTime = time(NULL);

  zlsio_Printf( "%s\n",  DIVIDERLINE );
  zlsio_Printf( " %s/ZLS Simulation Start.\n", FYRIS );
  zlsio_Printf( " %s", ctime(&startTime));
  zlsio_Printf( "%s\n\n",  DIVIDERLINE );
#endif
  return (err);
}

int zlsio_EndBanner ( void ){
  int err = noErr;
#if (ZLS_VERBOSE > 0)
  time_t  endTime = time(NULL);

  zlsio_Printf( "\n%s\n", DIVIDERLINE );
  zlsio_Printf( " %s/ZLS PPM-LR Hydro Simulation End.\n", FYRIS );
  zlsio_Printf( " %s", ctime(&endTime));
  zlsio_Printf( "%s\n\n",  DIVIDERLINE );
#endif 
  return (err);
}


