#ifndef ZLS_LOG_H
#define ZLS_LOG_H
#include "zls.h"

//
//========================================
//
// Routines for recording the setup and running progress of a
// simulation to a log file in the Output directory.  Most useful for
// documenting the run parameters and exactly when any images were
// saved and what time step a given output file corresponds to.  Keeps
// a record of the startup displays and the ongoing performance of the
// run, and records any error  messages.
// 
// It does this by using a replacement for standard C print statements:
// printf and fprintf( stdout, fprintf( stderr and fprintf( f, etc. 
// These zls routines mimic the standard calls, passing their variable
// parameters onto the normal system equivalents and also sending a
// copy to the log file if possible.  One call to zlsio_Printf( "Hello
// World %d\n", n) performs the normal fprintf( stdout, "Hello World
// %d\n", n ); and then does fprintf( zls_IOLOGFILE, "Hello World
// %d\n", n ); to copy to the log.  This makes the main code clean and
// nearly as simple as a normal code that prints formatted text.  If
// the log file doesn't exist yet the second call is not made.
//
//  To ensure any new code is logged, in place of the standard print calls Use:
//========================================
//Instead of printf to print to the screen (stdout), use zlsio_Printf
//with the same arguments: ie 
//          printf(" Time: %g s %d c\n", time, cycles); 
//becomes
//    zlsio_Printf( " Time: %g s %d c\n", time, cycles);
//
//========================================
//Instead of 
//          fprintf( stdout, " Time: %g s %d c\n", time, cycles);
//
//  do the same, but leave off stdout as it isn't needed:
//
//     zlsio_Printf( " Time: %g s %d c\n", time, cycles);
//
//========================================
//For error messages, instead of using the standard stderr stream literally:
//
//          fprintf( stderr, " Time: %g s %d c\n", time, cycles);
//
//use the zls error stream that sends to stderr and also the log file:
//
//     zlsio_ErrPrintf(" Time: %g s %d c\n", time, cycles);
//
//========================================
//To print to a file, f,  and *not* log it you *can* still use:
//
//    fprintf( f, " Time: %g s %d c\n", time, cycles);
//
//or the equivqlent: (this one doesn't log, but could be made to do so
//for debugging purposes):
//
//    zlsio_FPrintf( f, " Time: %g s %d c\n", time, cycles);
//
///========================================
//To print to the log and not to screen or to anywhere else use: 
//
//    zlsio_LogPrintf( " Time: %g s %d c\n", time, cycles);
//
//========================================
//
// macro for rapid access to log file stream.  Only safe once sim is
// running after setup has succeeded and we assume io and the file are
// already created . However has at least a crude attempt is made to check the
// io ptr for NULL first
//
#define zls_IOLOGFILE (gLevels[0].levelIO?gLevels[0].levelIO->outLogFILE:NULL)
#define zls_IOUSERLOGFILE (gLevels[0].levelIO?gLevels[0].levelIO->outUserLogFILE:NULL)
#define zls_IOINFOFILE (gLevels[0].levelIO?gLevels[0].levelIO->outInfoFILE:NULL)
#define zls_IOFILESFILE (gLevels[0].levelIO?gLevels[0].levelIO->outFilesFILE:NULL)

enum {
  kZLS_CreateNewLogFile  = 0, // create a new log file
  kZLS_ContinueLogFile   = 1,  // append to the exisiting file or create a new one
  kZLS_CreateNewInfoFile = 0, // create a new simulation information file
  kZLS_ContinueInfoFile  = 1,  // append to the exisiting file or create a new one
  kZLS_CreateNewFilesFile = 0, // create a new output files record file
  kZLS_ContinueFilesFile = 1  // append to the exisiting file or create a new one
};

//========================================
//
// Logging equivalents of standard variable argument printf statements 
//

void zlsio_Printf    (const char *template, ...);  // print to screen and log, uses stdout
void zlsio_ErrPrintf (const char *template, ...);  // print error and log, uses stderr
void zlsio_LogPrintf (const char *template, ...);  // print only to log.
void zlsio_UserLogPrintf (const char *template, ...);  // print only to user log.
void zlsio_InfoPrintf (const char *template, ...);  // print only to info.
void zlsio_FilesPrintf (const char *template, ...);  // print only to files.
//
// zlsio_FPrintf is no different to fprintf,but have it here for consistency
//
// Defaults to *not* log fprintf calls, but could be turned on by uncommenting 
// one line for v low level debugging.
//

void zlsio_FPrintf ( FILE *f, const char *template, ...);  // print to file like fprintf( f,
void zlsio_ZFPrintf ( zFile f, const char *template, ...);  // print to gz file like fprintf( f,

//
//
//========================================
//
int zlsio_CreateLogFile ( int newFile ); // uses level 0's IOSettings Ptr, error if not ready
int zlsio_CloseLogFile  ( void ); // log is kept open, so call this at end of sim
//
int zlsio_CreateUserLogFile ( int newFile ); // uses level 0's IOSettings Ptr, error if not ready
int zlsio_CloseUserLogFile  ( void ); // user log is kept open, so call this at end of sim
//
int zlsio_CreateInfoFile ( int newFile ); // records the simulations config files
int zlsio_CloseInfoFile  ( void ); // log is kept not open, so call this once config is recorded
//
int zlsio_CreateFilesFile ( int newFile ); // records the time and location of the simulation's output
int zlsio_CloseFilesFile  ( void ); // data files log is kept open, so call this at end of sim
//========================================
//
int zlsio_StartLog ( void );
int zlsio_EndLog   ( void );

int zlsio_LogSetup ( char * zlsConfName, char * ppmConfName, 
                    char * gridConfName, char * physConfName, 
                    char * initConfName, char * ioConfName, 
                    char * restartConfName );

int zlsio_RecordConfiguration ( char * zlsConfName, char * ppmConfName, 
                               char * gridConfName, char * physConfName, 
                               char * initConfName, char * ioConfName, 
                               char * restartConfName );

//========================================

int zlsio_StartBanner ( void );
int zlsio_EndBanner   ( void );

//
//========================================
//

int   zlsio_DisplaySimLimits     ( FILE * f , char * fname, Integer banner );
int   zlsio_DisplayIOSetup       ( FILE * f , char * fname, Integer banner );
int   zlsio_DisplayPhysicsSetup  ( FILE * f , char * fname, char * cfile, Integer banner );
int   zlsio_DisplayGridSetup     ( FILE * f , char * fname, Integer banner );

#endif
