#ifndef ZLS_COOLING_H
#define ZLS_COOLING_H
#include <zls.h>

enum {
	kPhysNoCooling    = 0,
	kPhysFileCooling  = 1,
	kPhysPowerCooling = 2
};

// checks for component by component activation
enum {
	kZLS_NoCoolingComponent = 0,
	kZLS_AllowCoolingComponent = 1,
	kZLS_SingleCoolingFunction = 0,
	kZLS_MultipleCoolingFunction = 1,
	kZLS_ComponentCoolingFunction = 2
};


// cooling aux info flags
enum {
	kZLS_HasGammaMu     = 0x00000001,
	kZLS_HasHFractions  = 0x00000002,
	kZLS_HasXFractions  = 0x00000004
};

// structure for holding the cooling information
typedef struct cool {
  
	Integer coolType;
	Integer collAuxType;  // info on presence of additional info such a HI fractions and Xrays
  
  Real coolScale;       // multiplier to scale the overall cooling of all kinds
	
	Real coolMinT;        // minimum cooling threshold, used for all forms of cooling
	Real coolMaxT;        // maximum cooling limit, used for all forms of cooling
  
	Real coolMinPd;        // P/d corresponding to coolMinT
	Real coolMaxPd;        // P/d corresponding to coolMaxT
	
	// allow some control over components of cooling/heating in a multi-component function
	Integer coolFlagTot;       // = kZLS_SingleCoolingFunction, kZLS_MultipleCoolingFunction, or kZLS_ComponentCoolingFunction
	Integer coolFlagAt;       // = kZLS_AllowCoolingComponent or kZLS_NoCoolingComponent, governs all atomic
	Integer coolFlagH2;       // = kZLS_AllowCoolingComponent or kZLS_NoCoolingComponent, sets or disables H2 cooling
	Integer coolFlagCO;       // = kZLS_AllowCoolingComponent or kZLS_NoCoolingComponent, sets or disables CO cooling
	Integer heatFlagCX;       // = kZLS_AllowCoolingComponent or kZLS_NoCoolingComponent, allow cosmic ay heating
	Integer heatFlagHI;       // = kZLS_AllowCoolingComponent or kZLS_NoCoolingComponent, allow HI photo-electic heating
	Integer heatFlagGR;       // = kZLS_AllowCoolingComponent or kZLS_NoCoolingComponent, all grain heating/cooling effects
	
	// parameters for lookup table
	Integer coolNTable;    // number of entries in table
	Real coolTTableMin;    // minimum T in cooling table
	Real coolTTableMax;    // maximum T in cooling table
	Real   coolDeltaT;    // table temperature step *ratio* (ie generated with log T steps originally)
  
  // used to generate the index interval for an arbitrary temperature.
	
	Q1DArr coolTValues;     // table of temperatures, to map an index to a temp
  
	Q1DArr coolFunction;    // table of total cooling function
	Q1DArr coolFunctionAt;  // table of Atomic cooling function, allocated even if not used in a given run ( 1D and small)
	Q1DArr coolFunctionH2;  // table of H2 cooling function, allocated even if not used in a given run ( 1D and small)
	Q1DArr coolFunctionCO;  // table of CO cooling function
	
	Q1DArr coolDFunction;     // table of cooling derivative
	Q1DArr coolDAtFunction;     // table of At cooling derivative
	Q1DArr coolDH2Function;     // table of H2 cooling derivative
	Q1DArr coolDCOFunction;     // table of CO cooling derivative
  
	Real coolPdTableMax;    // maximum P/d in cooling table
	Real coolPdTableMin;    // maximum P/d in cooling table
	
	Q1DArr coolPdValues;    // table of P/d, to map to an index and look up a Temperature or gammma or mu
	Q1DArr coolmuValues;      // table of mol Weights(Pd)
  
	Q1DArr coolUdValues;       // table of Internal Energies(Pd), to derive P from gamma and U/d
	Q1DArr coolGammaValues;   // table of effective gammas(Pd)
  
	Q1DArr coolfh2Values;      // table of H2 fractions
	Q1DArr coolfhiValues;      // table of HI fractions
	Q1DArr coolfhiiValues;      // table of HII fractions
	
	Q1DArr coolfXtotal;     // table of Xray cooling fractions
	Q1DArr coolfXsoft;      // table of soft 0.1- 1.5 keV Cooling fractions
	Q1DArr coolfXhard;      // table of hard 1.5-10.0 keV Cooling fractions
  
  // parameters for powerlaw cooling
  
	Real coolLambda0;
	Real coolT0;
	Real coolPower;
	
	// parameters for phot equilubrium heating if used (nyiErr)
	Real coolPHeat0;
	Real coolPHSlope;
	Real coolPHFactor;
	  
} Cooling, * CoolingPtr;


//
//====================================================================================
//
int zlsp_InitCooling                ( CoolingPtr cp );
int zlsp_NewCooling                 ( CoolingPtr cp );
int zlsp_DisposeCooling             ( CoolingPtr cp );
int zlsp_SetPhysicsCoolingFile      ( CoolingPtr cp, char * cFile );
//
int zlsp_CoolZoneWHD                ( ZonePtr z, Real dt );
int zlsp_InitialCoolingTimeWHD      ( Real *minDt,  ZonePtr z,  Real dt );
//
#ifdef COOLING
//
//====================================================================================
//
// global inline cooling functions
//
//====================================================================================
//
static inline Real zlsp_LambdaComponents( Real *lam, Real *lamAt, Real *lamH2, Real* lamCO, Real T , CoolingPtr cp);
static inline Real zlsp_LambdaComponents( Real *lam, Real *lamAt, Real *lamH2, Real* lamCO, Real T , CoolingPtr cp){
  
	*lam      = 0.0;
	*lamAt    = 0.0;
	*lamH2    = 0.0;
	*lamCO    = 0.0; 
	Real lam0 = 0.0;
  
	if ( ( T > cp->coolMinT ) &&  ( T <= cp->coolMaxT ) ){
    
		switch( cp->coolType)  {
			default:
			case kPhysNoCooling:
				break;
			case kPhysFileCooling:
     {
       
       Real TMax = cp->coolTTableMax;
       Integer n = cp->coolNTable;
       
       if ( T > TMax ){
         *lamAt = cp->coolFunctionAt[n-1]; 
         *lamH2 = cp->coolFunctionH2[n-1]; 
         *lamCO = cp->coolFunctionCO[n-1];
         lam0 = cp->coolFunction[n-1]*zls_sqrt( T/TMax );
       } else {
         
         register Integer k;
         register Integer klo=0;
         register Integer khi=n-1;
         
         do{
           k=(khi+klo)>>1;
           if(cp->coolTValues[k]>T){
             khi=k;
           } else {
             klo=k;
           }
         } while ( khi-klo > 1) ;
         
         Real xtemp1 = T - cp->coolTValues[klo];
         
         *lamAt = (cp->coolFunctionAt[klo] + xtemp1*cp->coolDAtFunction[klo]);
         *lamH2 = (cp->coolFunctionH2[klo] + xtemp1*cp->coolDH2Function[klo]);
         *lamCO = (cp->coolFunctionCO[klo] + xtemp1*cp->coolDCOFunction[klo]);
         lam0   = (cp->coolFunction[klo] + xtemp1*cp->coolDFunction[klo]);
         
       }
       
       *lam   = lam0;
       
     }
				break;
			case kPhysPowerCooling:
     {
       Real T1 = T*cp->coolT0;
       lam0 = (cp->coolLambda0 * zls_Pow(T1, cp->coolPower));
     }
    };  // switch
  }
	
	return(lam0);
  
}


static inline Real zlsp_Lambda( Real T , CoolingPtr cp);
static inline Real zlsp_Lambda( Real T , CoolingPtr cp){
  
	Real lam = 0.0;
  
  // assumes fH[1] = 1.0 if a kZLS_ComponentCoolingFunction is used without fH option
  // and the total function is summed on read in.
  
	if ( ( T > cp->coolMinT ) &&  ( T <= cp->coolMaxT ) ){
		switch( cp->coolType)  {
			default:
			case kPhysNoCooling:
				break;
			case kPhysFileCooling:
     {
       
       Real TMax = cp->coolTTableMax;
       Integer n = cp->coolNTable;
       
       if ( T > TMax ){
         lam = cp->coolFunction[n-1]*zls_sqrt( T/TMax );
       } else {
         
         register Integer k;
         register Integer klo=0;
         register Integer khi=n-1;
         
         do{
           k=(khi+klo)>>1;
           if(cp->coolTValues[k]>T){
             khi=k;
           } else {
             klo=k;
           }
         } while ( khi-klo > 1) ;
         
         Real xtemp1 = T - cp->coolTValues[klo];
         
         lam  = (cp->coolFunction[klo] + xtemp1*cp->coolDFunction[klo]);
         
       }
       
       
     }
        
				break;
			case kPhysPowerCooling:
     {
       Real T1  = T*cp->coolT0;
       lam = (cp->coolLambda0 * zls_Pow(T1, cp->coolPower));
     }
    };  // switch
  }
	
	return(lam);
  
}


#endif COOLING

#endif ZLS_COOLING_H
