#ifndef ZLS_GRAVITY_H
#define ZLS_GRAVITY_H
#include <zls.h>

enum {
	kPhysNoGravity           = 0,
	kPhysIsothermalGravity   = 1,
	kPhysKingSpheroidGravity = 2,
	kPhysNewtonianGravity    = 3,
	kPhysUniformGravity      = 4,
	kPhysDiskZGravity        = 5,
	kPhysTanhZGravity        = 6,
	kPhysHaloGravity         = 7,
	kPhysCosmicGravity       = 8
};

enum {
	kPhysNoGravityFlag           = 0x0000,
	kPhysIsothermalGravityFlag   = 0x0001, // Specific halo potential, based on velocity dispersion and core radius where rho = rho_0/e
	kPhysKingSpheroidGravityFlag = 0x0002,
	kPhysNewtonianGravityFlag    = 0x0004,
  kPhysUniformGravityFlag      = 0x0008,
  kPhysDiskZGravityFlag        = 0x0010,
  kPhysTanhZGravityFlag        = 0x0020,
  kPhysHaloGravityFlag         = 0x0040, // Generic halo potential, like iso, but based on circular velocity, and dimensionless rho(1) = 1/e
  kPhysCosmicHaloGravityFlag   = 0x0080  // same as Halo, but derived from M_vir,x_vir and density contrast Delta*rho_u
};

enum {
  //
  // composite flags constants
  // IsoGravityTypes, share the r_iso and s_iso isothermal parameters
  //
  kPhysIsoGravityTypes        = kPhysIsothermalGravityFlag |
                                     kPhysTanhZGravityFlag |
                                      kPhysHaloGravityFlag |
                                kPhysCosmicHaloGravityFlag
  
};


typedef struct pot {
  
  Integer gravTypeFlags; 
  // a sum of different gravity component types
  // simple newtonian central mass, M in solar masses
  
  Real M ; 
  Real rN0 ;  // zero point for potential, ie star surface radius, gives a +ve potential
  
  // king sphere approximation parameters, plus convenience reciprocals and squares
  
  Real    r_ss;  
  Real invr_ss;
  Real    m_ss; // m_ss in solar masses
  
  // isothermal  parameters, plus convenience reciprocals and squares
  // used to scale several dimensionless potential fnctions, including the
  // sphere, a uniform field and a Tanh vertical disk field.
  
  Real  r_iso;
  Real  s_iso; // Sqrt2 times the circular, ie s^2 = 2GM/r
  
  // Real rho_iso; is not needed as it is determined by r and s uniquely
  
  Real  invr_iso;
  Real  s2_iso;  // 2GM/r
  Real  invs_iso;
  Real  invs2_iso;
  //
  // Generic Cosmological Halo Potential parameters, Pick M_vir, get x_vir and then r_s rho_s and v_s can be 
  // derived, or r_s and v_s can be set directly
  
  Real  r_s;
  Real  invr_s;

  Real  v_s;  // circular rather than dispersion, smaller by sqrt(2), ie v^2 = GM/r
  Real  rho_s;  // derived if needed
  
  Real M_vir;
  Real M_s;
  Real M_1;
  
  Real x_vir;  // generated by fitting formula from cosmological simulations
               // f_m generated by the integrator used to make the input file, see development/code/potentials
  Real f_mvir; // number of scale masses inside x_vir*r_s, 
  Real f_m1;   // number of scale masses inside r_s, x =1.0
  //
  // test parameters:
  // M_vir = 1.2e9Msol, 2.386920E+42 g
  // x_vir = 26.60904 from comsological sims
  // r_s = 1031.345 pc , 3.182422E+21 cm, independent of potential form
  // 
  // Einasto:
  // f_m(x_vir) = 42.46255 for data file: einlogwlogdw.txt
  // f_m(1.0)   = 3.758872 for data file: einlogwlogdw.txt
  //
  // v_s = 1.085686E+06 cm/s 10.85686 km/s
  // M_s = 2.826020E+07 Msol
  //
  // v_1 = 21.04907 km/s
  // M_1 = 1.0622647E+08 Msol
  // 
  // NFW:
  // f_m(x_vir) = 43.53603  for data file: nfwlogwlogdw.txt
  // f_m(1.0)   = 3.571608  for data file: nfwlogwlogdw.txt
  //
  // v_s = 1.07221698E+06 cm/s 10.7221698 km/s
  // M_s = 2.7563377E+07 Msol
  // 
  // v_1 = 20.263505 km/s
  // M_1 = 9.8445577E+07 Msol
  //
  //
  // HIsothermal:
  // f_m(x_vir) = 68.70434 for data file: hisologwlogdw.txt
  // f_m(1.0)   = 2.166088 for data file: hisologwlogdw.txt
  //
  // v_s = 8.535226E+05 cm/s 8.535226 km/s
  // M_s = 1.746614E+07 Msol
  // 
  // v_1 = 12.561835 km/s
  // M_1 = 3.783321E+07 Msol
  //
  //
  
  Real delta_rho;     // usually 200, or 352 for small halos at z=0
  Real rho_universe;  // 2.76e-30 gm/cm^3 at z= 0
  
  Real  v2_s;  // scale GM_s/r_s
  Real  invv_s;
  Real  invv2_s;
  
  Real  v2_1;  // Core GM_s/r_s, actual not scaled
  Real  v_1;  // Core GM_s/r_s, actual not scaled
  
  // dimensionless spherical potential spline data
  
  Integer gravNWpline;
  
  Q1DArr  gravRWspl;
  Q1DArr  gravWspl;
  Q1DArr  gravWspl2;
  
  Q1DArr gravdWspl;
  Q1DArr gravdWspl2;
  
  // Disk Z  field parameters
  
  Real  d_diskZ;  // pc
  Real s0_diskZ;  // M_0/pc^2
  Real re_diskZ;  // M_0/pc^3
  
} Potential, * PotentialPtr;

int  zls_InitField        ( FieldPtr  f); // clears the field fields (heh!)

int  zls_NewFieldPtr      ( FieldPtr *f); // also allocate the small field structure, not needed for plain Field vars
int  zls_DisposeField     ( Field     f);
int  zls_DisposeFieldPtr  ( FieldPtr  f);

//
int   zlsio_SetUpGravityFromConfigP(PotentialPtr gp, struct ccl_t * configP);

// prototypes and structures for zls hydro calculations that involve gravity 
// potentials and accellerations.

Real  zlsg_phi         ( Real r, PotentialPtr gp );
Real  zlsg_dphi        ( Real r, PotentialPtr gp );

Real  zlsg_phi_newton  ( Real r, PotentialPtr gp ); // M in solar masses, r normalised
Real  zlsg_dphi_newton ( Real r, PotentialPtr gp ); // M in solar masses, r normalised

Real  zlsg_phi_sphere  ( Real r, PotentialPtr gp );
Real  zlsg_dphi_sphere ( Real r, PotentialPtr gp );

Real  zlsg_phi_iso     ( Real r, PotentialPtr gp ); // dimensionless self gravitating sphere, scaled by s^2 and r0
Real  zlsg_dphi_iso    ( Real r, PotentialPtr gp );
//
Real  zlsg_phi_halo    ( Real r, PotentialPtr gp ); // dimensionless self gravitating sphere, scaled bu v_s^2 and r0
Real  zlsg_dphi_halo   ( Real r, PotentialPtr gp );
//
Real  zlsg_phi_Uniform ( Real z, PotentialPtr gp ); // dimensionless uniform accelleration (abs(r/r0)), scaled by s^2 and r0
Real  zlsg_dphi_Uniform( Real z, PotentialPtr gp ); // accell =  s^2/r0, shares iso sphere r and s^2 parameters
//
Real  zlsg_phi_DiskZ   ( Real z, PotentialPtr gp ); // Galactic Disk potential, includes halo conribution term, fit to data
Real  zlsg_dphi_DiskZ  ( Real z, PotentialPtr gp ); // and scaled by central surface density, scaleheight and central halo density
//
Real  zlsg_phi_TanhZ   ( Real z, PotentialPtr gp ); // Phi is actually s^2 ln(cosh(x/x0))
Real  zlsg_dphi_TanhZ  ( Real z, PotentialPtr gp ); // -dphi/dr = s^2 tanh(x/x0), shares iso sphere r and s^2 parameters
//
int zlsg_SetIsothermalGravityFile ( PotentialPtr phys, char * gFile);
//

int zlsg_InitGravityPotential(PotentialPtr gp);
int zlsg_DisposeGravityPotential(PotentialPtr gp);

//
int  zlsg_InitZoneGravityPotentialField ( ZonePtr z ); // fills in a  phi vector field for reuse
int  zlsg_InitZoneGravityAccellField    ( ZonePtr z ); // fills in a dphi vector field for reuse
//
int zlsg_InitLevelGravityField( Integer level );
//
//========================================

#ifdef GRAVITY
// only used for gravity right now, probably for other fields later.
// Fields are set up with average and left-right values pre-calculated
// (dynamically or at startup). So if PPM is in force, just del
// (trivial) and a6 need be computed for f, so that can be done on the
// fly when memory access into zoneField is slow anyway. Therefore f
// won't need to be passed to the normal ppm coeff routines later,
// saving a lot of time.
//

int zlsh_FillF1DFrom1DField ( F1DArr f, Field zoneField ); 

int zlsh_FillF1DFrom2DFieldX ( F1DArr f, Field zoneField, Integer j ); 
int zlsh_FillF1DFrom2DFieldY ( F1DArr f, Field zoneField, Integer i );
int zlsh_FillF2DFrom2DField  ( F2DArr f, Field zoneField ); 

int zlsh_FillF1DXFrom3DFieldYZ ( F1DArr f,  Field zoneField, Integer j, Integer k );
int zlsh_FillF1DYFrom3DFieldZX ( F1DArr f,  Field zoneField, Integer k, Integer i );
int zlsh_FillF1DZFrom3DFieldXY ( F1DArr f,  Field zoneField, Integer i, Integer j );

//int zlsh_FillF2DFrom3DZoneX ( F2DArr f,  Field zoneField, Integer i );
//int zlsh_FillF2DFrom3DZoneY ( F2DArr f,  Field zoneField, Integer j );
//int zlsh_FillF2DFrom3DZoneZ ( F2DArr f,  Field zoneField, Integer k );

int zlsh_FillF3DFrom3DField ( F3DArr f,  Field zoneField );


#endif GRAVITY
#endif ZLS_GRAVITY_H
