#include <zls.h>

//========================================
// physical constants in cgs units
//========================================

const Real  k_clight = 2.99792458e10;
const Real  k_me     = 9.109534e-28;
const Real  k_mp     = 1.6726485e-24;
const Real  k_mh     = 1.6737657e-24;
const Real  k_amu    = 1.66053128e-24;
const Real  k_invamu = 1.0/1.66053128e-24;
const Real  k_boltz  = 1.380622e-16;
const Real  k_stefan = 5.6696079e-5;
const Real  k_plank  = 6.626176e-27;
const Real  k_guniv  = 6.6732e-8;
const Real  k_eesu   = 4.80325e-10;
//   astronomical constants (cgs);
const Real  k_msol   = 1.9891e33;
const Real  k_lsol   = 3.826e33;
//    ionized plasma constants
const Real  k_muhi   = 1.432e0; 
const Real  k_muions = 1.300e0;
Real  k_mu     = ZLS_EOS_MU; 
Real  k_invmu  = 1.0/ZLS_EOS_MU; // density -> n total, depends on ionisation, includes electrons
const Real  k_AbnH   = 1.0/(1.66053128e-24*1.432e0); // density -> nH total, fixed by abundances

//   conversion factors
const Real  f_cmau   =  1.49597e13;
const Real  f_cmpc   =  3.085677577e18;
const Real  f_cmkpc  =  3.085677577e21;
const Real  f_cmkm   =  1.0e5   ;
const Real  f_everg  =  1.6021892e-12;
//
// constants for stellar winds, yrs are useful generally.
//
const Real f_yrs      = 31557600.0; //Julian year
const Real f_invyrs   = 3.168808781402895e-08; //inverse Julian year
const Real f_invMyrs  = 3.168808781402895e-14; //inverse Julian Myear
const Real f_gsunyrs  = 6.303077547088498e+25; // mSun /yr to g/s

//========================================
//
//  kZLS_eos_gamma   - equation of state gamma
//  kZLS_eos_gamma_1  - kZLS_eos_gamma - 1
//      
//
Real kZLS_eos_gamma   = ZLS_EOS_GAMMA;
Real kZLS_eos_gamma_1 = (ZLS_EOS_GAMMA - 1.0);
Real kZLS_eos_gamma_2 = (ZLS_EOS_GAMMA - 2.0);
Real kZLS_eos_invgamm = (1.0/(ZLS_EOS_GAMMA-1.0));

Real kZLS_eos_a_iso   = 1.0;  // isothermal sound speed for isothermal EOS
Real kZLS_eos_a_iso2  = 1.0;
//========================================
//
// fiducial dimensional normalising factors for scaling input so that
// the solvers get an easier time, and cooling/gravity are scaled correctly
//
// not const as they are calculated at startup, all set to 1.0 for scale free adiabatic
// default.
//
Real f_t0      = 1.0;  // time
Real f_invt0   = 1.0;  // 1/time
Real f_x0      = 1.0;  // x
Real f_invx0   = 1.0;  // 1/x
Real f_area0   = 1.0;  // x^2
Real f_vol0    = 1.0;   // x^3
Real f_v0      = 1.0;  // velocity
Real f_invv0   = 1.0;  // 1/velocity
Real f_a0      = 1.0;  // accelleration

Real f_z0      = 1.0;
Real f_r0      = 1.0;
Real f_phi0    = 180.0/3.1415926535897932;
Real f_theta0  = 180.0/3.1415926535897932;

Real f_d0      = 1.0; // density
Real f_invd0   = 1.0; // 1.0/density
Real f_nHtot   = 1.0; // total H atom (all forms) density
Real f_P0      = 1.0; // pressure
Real f_m0      = 1.0; // mass
Real f_E0      = 1.0; // energy
Real f_EL0     = 1.0; // energy per unit mass

Real f_tep0    = 1.0; // temperature
Real f_invtep0    = 1.0; // 1/temperature

Real f_cool0   = 1.0; // total emissivity cooling
Real f_lam0    = 1.0; // cooling function
Real f_invlam0 = 1.0; // inverse cooling function

#ifdef MHD
Real f_muB     =  3.544907701811032; // magnetic field mu = sqrt(4 pi)
Real f_B0      =  3.544907701811032; // magnetic field strength  v mu sqrt(rho)
#endif MHD

//
// some normalised constants	 
//
#ifdef GRAVITY
Real k_g      = 1.0; // gravity
Real k_gmsol  = 1.0; // gravity constant for M in solar masses
#endif GRAVITY
//
//========================================
//
char s_Time     [16];
char s_Length   [16];
char s_Area     [16];
char s_Volume   [16];
char s_Velocity [16];
char s_Accel    [16];
char s_Mass     [16];
char s_Density  [16];
char s_Pressure [16];
char s_Energy   [16];
char s_Temperature[16];
#ifdef MHD
char s_Magnetic [16];
#endif MHD
#ifdef COOLING
char s_Cool     [16];
char s_Lambda   [16];
#endif COOLING
#ifdef GRAVITY
char s_Gravity  [16];
#endif GRAVITY
//
//========================================
//

int zlsp_InitGlobalEOS_Gamma( PhysicsPtr phys, Real gamma ){
  int err = noErr;
  
  if ( phys == NULL) {
    phys = gLevels[0].levelPhysics;
  }
  
  if ( phys != NULL){
    
    phys->gamma       = gamma;
    
    //
    // set global values of gamma.  Most general areas will use these globals
    // when gamma is allowed to vary, some specialised routines will refer to
    // phys->gamma and even compute gamma as a function of T and P or rho
    //
    phys->gamm        = (phys->gamma - 1.0);
    
    kZLS_eos_gamma    = phys->gamma;
    kZLS_eos_gamma_1  = (kZLS_eos_gamma - 1.0);
    kZLS_eos_gamma_2  = (kZLS_eos_gamma_1 - 1.0) ;
    kZLS_eos_invgamm  = (1.0/(kZLS_eos_gamma_1));
    
    
#ifdef HYDRO
    
    //
    // set  EOS constants for the Riemann solver, generally a bunch of constants the 
    // require a lot of divisions to evaluate, but if gamma is fixed then they can 
    // be precomputed once and reused.  Actual constants depend of the solver being used
    // so this is a macro.  Some simple solvers do nothing here.
    //

    INITRIEMANNSOLVER (kZLS_eos_gamma);
    
    zlsr_InitToroRiemannGammas(kZLS_eos_gamma); //( always needed for sample routine)
    
#endif HYDRO
    
  }
  
  return(err);
}

int zlsp_InitPhysics(PhysicsPtr phys){
  int err = noErr;
  
  if (phys != NULL) {
		
		phys->mu = k_mu;
    
    zlsp_InitGlobalEOS_Gamma ( phys, kZLS_eos_gamma);
    
    //    zlsp_SetFiducialScales ( 1.0, 1.0, 1.0 ); // now called in init.c before levels zones or phys are inited.
    
    
#ifdef COOLING
    zlsp_InitCooling(&(phys->coolParameters));
#endif COOLING
		
#ifdef GRAVITY
    zlsg_InitGravityPotential(&(phys->gravPotential));
#endif GRAVITY
		
  }
	
	return (err);
}

//
//========================================
//


int zlsp_NewPhysics(PhysicsPtr * phys){
  int err = noErr;
	
  *phys = (PhysicsPtr)malloc(sizeof(Physics));
	
  if (*phys == NULL) {
    err = memErr;
#if (ZLS_DEBUG >0)
		zlsio_ErrPrintf( " ERROR: zlsp_NewPhysics allocation failed %d \n",err);
#endif
  } else {
    err = zlsp_InitPhysics( *phys );
  }
	
	return (err);
}


int zlsp_DisposePhysics(PhysicsPtr phys){
  int err = noErr;
	
	if ( phys != NULL ) {
#ifdef COOLING
	  zlsp_DisposeCooling  (&phys->coolParameters);
#endif COOLING
    
#ifdef GRAVITY
    zlsg_DisposeGravityPotential(&(phys->gravPotential));
#endif GRAVITY
    
		free(phys);
	}
	
	return (err);
}

//
//========================================
//

int zlsp_SetFiducialScales ( Real x0, Real v0, Real d0, Real muB ){ 
  //
	// scaling length, velocity, density [mag field (B always uses muB=sqrt(4 pi)]
	//
	//
	// When inputting a quantity to the code, divide by the appropriate fiducial quantity to
	// normalise it.  Ususally done at IO so there is no real performance loss./
	//
	// When writing out a quantity, restore the quantity by muliplying by the factor
	//
	// ie density:  
	//  Input d, set w.d[i] = d/f_d0;  thereafter the internal density is scaled
	//  Output d, d = w.d[i] * f_d0; to get back to original units.  this allows input in cgs
	// for example without overflowing the solvers.  
	//
	// Adiabatic, scale free: 
	//      x0 = v0 = d0 = 1.0  => time scale = 1.0
	// problem parameters in init.conf can be set in simple dimentional units.
	//
	// cgs ionised ISM on kpc scale galaxy:  
	//      x0 = 1e20cm, v0 = 1.0e7 cm/s, d0 = 1.0e-24 g/cm^3  t0 = 10^13s ~300 000 yrs
	// problem parameters in init.conf can then be input in cgs units and scaled internally
	//
	// Note grid.conf always specifies grid coordinates in normalised quanities, 
	// ie in units of x0, a grid of 0.0 - 1.0 in the first exmple will be 0.0 - 1.0, 
	// in the second example it will still have the values of 0.0- 1.0, but represent 
	// a range of 0.0 - 1.0e20 cm.  1kpc would be 30.8 in these units.
	//
	//
	Integer err = noErr;
	
	f_x0     = x0;
	f_invx0  = 1.0/x0;
  f_area0  = f_x0*x0;  // x^2
  f_vol0   = f_area0*x0;  // x^3
  
	f_r0     = x0; // radii alias for x0
	f_z0     = x0; // cylinder z alias for x0
	
	f_v0     = v0;
	f_invv0  = 1.0/v0;
	f_d0     = d0;
	f_invd0  = 1.0/d0;
  f_nHtot  = k_AbnH*d0;  // ie f_nHtot*d  = number density of H atoms
  
	f_t0     = f_x0/f_v0;
	f_invt0  = f_v0/f_x0;
	f_a0     = f_v0/f_t0;
	
	f_P0     = f_d0*f_v0*f_v0;
	
	f_E0     = f_P0*(f_x0*f_x0*f_x0);
	f_EL0    = f_P0*(f_x0*f_x0*f_x0)/f_d0;
	f_m0     = f_d0*(f_x0*f_x0*f_x0);
	
  f_tep0    = f_v0*f_v0*k_mu*k_amu/k_boltz; // temperature scaling factor, depends on k_mu:  T = f_tep0 * P/d
  f_invtep0    = 1.0/f_tep0; // inverse temperature scaling factor P = d* T* f_invtep0
  
#ifdef MHD
	f_muB     = muB;          // magnetic field scaling, mu = sqrt(4 pi) Gaussian by default
	f_B0      = f_v0*f_muB*zls_sqrt(f_d0); //
#endif MHD
  
	f_cool0   = f_P0/f_t0;
	f_lam0    = f_cool0/(f_d0*f_d0);
	f_invlam0 = 1.0/f_lam0;
    
	// these are more like scaled constants, than renormalised dimensional quantities, so prefix with k_
#ifdef GRAVITY
	k_g       = f_d0*f_t0*f_t0*k_guniv; // a =  k_g M / r^2 , a, M and r normalised units
	k_gmsol   = k_g*k_msol/f_m0; // special: M in solar masses, ie  a = k_gmsol*M/r^2 (M in solar masses, a, r normal)
#endif GRAVITY
  
	return(err);
}

int zlsp_ApplyFiducialScalesWZone2D ( ZonePtr z ){ 
#ifndef SYM2D
  return(dimErr);
#else
  // a convenience routine to allow you to just fill in a primitve variable 2D zone
  // with initial values then scale it all before commencing calculations
	
  Integer err = noErr;
	
  WHD       zoneW;
  Integer    i, j;
	
  if (z != NULL){
		
    err     = zlsp_ReferenceZoneWHD(&zoneW, z);
		
    if (err == noErr) {
			
      Integer is =  0;
      Integer ie =  z->zoneGrid.nt[kXDim];
      Integer js =  0;
      Integer je =  z->zoneGrid.nt[kYDim];
			
      for(j = js; j < je; j++){
        for(i = is; i< ie; i++){
					
          zoneW.d[j][i]   /=  f_d0;
          zoneW.P[j][i]   /=  f_P0;
          zoneW.vx[j][i]  /=  f_x0;
          zoneW.vy[j][i]  /=  f_x0;
					
        } //i
				
      }//j
			
      zlsp_ReleaseZoneWHD(zoneW);
			
    } // zoneW
		
  } else { err = zonErr;} //z
	
  return(err);
#endif SYM2D
}


int zlsp_UndoFiducialScalesWZone2D ( ZonePtr z ){ 
#ifndef SYM2D
  return(dimErr);
#else
  // a convenience routine to allow you restore a 2D zone of primitve vars 
  // prior to output at the end of a simulation.  More often individual
  // outputs are computed as needed.
	
  Integer err = noErr;
	
  WHD       zoneW;
  Integer    i, j;
	
  if (z != NULL){
		
    err     = zlsp_ReferenceZoneWHD(&zoneW, z);
		
    if (err == noErr) {
      
      Integer is =  0;
      Integer ie =  z->zoneGrid.nt[kXDim];
      Integer js =  0;
      Integer je =  z->zoneGrid.nt[kYDim];
      
			
      for(j = js; j < je; j++){
        for(i = is; i< ie; i++){
					
          zoneW.d[j][i]   /=  f_d0;
          zoneW.P[j][i]   /=  f_P0;
          zoneW.vx[j][i]  /=  f_x0;
          zoneW.vy[j][i]  /=  f_x0;
					
        } //i
				
      }//j
      
      zlsp_ReleaseZoneWHD(zoneW);
			
    } // zoneW
		
  } else { err = zonErr;} //z
	
  return(err);
#endif SYM2D
}


//========================================
// zlsp_PrintPhysicsConfiguration always prints to a std C file, producing screen output to stdout or uncompressed text to f
// The next zlsp_ZPrintPhysicsConfiguration routine , takes a zFile and writes to compressed files if io compression is 
// used.  However if io comression is used then zlsp_ZPrintPhysicsConfiguration will not print to the screen, it is intended
// primarily for user dumps where compressed text is being written.  Normal log files and screen will not be compressed.
//
int zlsp_PrintPhysicsConfiguration( FILE * f, char * cfile ){
	int err = noErr;
	
	PhysicsPtr        phys =  gLevels[0].levelPhysics; // get the global physics settings
	
	if (phys != NULL) {
		
#ifdef COOLING
		zlsio_FPrintf(  f,   "%s\n", DIVIDERLINE );
		zlsio_FPrintf(  f,   " ZLS Physics EOS \n" );
		zlsio_FPrintf(  f,   "%s\n", DIVIDERLINE );
		zlsio_FPrintf(  f,   "   Mean Molecular Weight (mu) : %g \n", phys->mu );
		zlsio_FPrintf(  f,   "   Adiabatic Index    (gamma) : %g \n", phys->gamma );
		zlsio_FPrintf(  f,   "%s\n", DIVIDERLINE );
		zlsio_FPrintf(  f,   " ZLS Physics Cooling \n" );
		zlsio_FPrintf(  f,   "%s\n", DIVIDERLINE );
		if (phys->coolParameters.coolType > kPhysNoCooling ){	
			zlsio_FPrintf(  f,   " Cooling is enabled, and ON\n" );
			zlsio_FPrintf(  f,   " Cooling Timescale Fraction: %g\n", kZLS_CoolTimeFraction);
			
			if ( phys->coolParameters.coolType == kPhysFileCooling ) {
        
				if (cfile) zlsio_FPrintf( f, "\n      Cooling File in Use: %s\n",cfile );
				zlsio_FPrintf( f, "        N values: %d\n", phys->coolParameters.coolNTable );
				zlsio_FPrintf( f, "        Table Min.  T: %g\n", phys->coolParameters.coolTTableMin );
				zlsio_FPrintf( f, "        Table Max.  T: %g\n", phys->coolParameters.coolTTableMax );
				zlsio_FPrintf( f, "        Table delta T: %g\n", phys->coolParameters.coolDeltaT );
        
				if (phys->coolParameters.coolFunction) {
					zlsio_FPrintf( f, "        Main Table     : OK\n" );
				} else {
					zlsio_FPrintf( f, "        ERROR: Main Table: BAD - Not ALLOCATED!\n" );
					err = memErr;
					return(err);
				}
        
        if (phys->coolParameters.coolFunctionAt) {
          zlsio_FPrintf(  f,   "         Atomic Table  : OK\n" );
        }
        if (phys->coolParameters.coolFunctionCO) {
          zlsio_FPrintf(  f,   "        Mol. CO Table  : OK\n" );
        }
        if (phys->coolParameters.coolFunctionH2) {
          zlsio_FPrintf(  f,   "        Mol. H2 Table  : OK\n" );
        }
        
				if (phys->coolParameters.coolTValues) {
					zlsio_FPrintf( f, "        Temperatures   : OK\n" );
				} else {
					zlsio_FPrintf( f, "        ERROR: Temperature Table: BAD - Not ALLOCATED!\n" );
					err = memErr;
					return(err);
				}
        
        if (phys->coolParameters.coolFunction) {
          if ( phys->coolParameters.coolFlagTot == kZLS_SingleCoolingFunction) {
            zlsio_FPrintf(  f,"  Single Total Cooling : ENABLED\n" );
          } else {
            
            zlsio_FPrintf(  f, "  Multi/Component Cooling : ENABLED\n" );
            
            if (phys->coolParameters.coolFunctionAt) {
              if ( phys->coolParameters.coolFlagAt == kZLS_AllowCoolingComponent) {
                zlsio_FPrintf(  f,   "                 Atomic Cooling: ENABLED\n" );
              } else {
                zlsio_FPrintf(  f,   "                 Atomic Cooling: DISABLED\n" );
              }
            }
            if (phys->coolParameters.coolFunctionCO) {
              if ( phys->coolParameters.coolFlagCO == kZLS_AllowCoolingComponent) {
                zlsio_FPrintf(  f,   "                Mol. CO Cooling: ENABLED\n" );
              } else {
                zlsio_FPrintf(  f,   "                Mol. CO Cooling: DISABLED\n" );
              }
            }
            if (phys->coolParameters.coolFunctionH2) {
              if ( phys->coolParameters.coolFlagH2 == kZLS_AllowCoolingComponent) {
                zlsio_FPrintf(  f,   "                Mol. H2 Cooling: ENABLED\n" );
              } else {
                zlsio_FPrintf(  f,   "                Mol. H2 Cooling: DISABLED\n" );
              }
            }
            
          }
				}
				
#ifdef VARIABLE_EOS
				if (phys->coolParameters.coolmuValues) {
					zlsio_FPrintf(  f,   "        EOS mu Table   : OK\n" );
          zlsio_FPrintf(  f,   "           Variable mu : ENABLED\n" );
        } else {
          zlsio_FPrintf(  f,   "           Variable mu : DISABLED\n" );
				}
				
#ifdef VARIABLE_GAMMA
				if (phys->coolParameters.coolGammaValues) {
					zlsio_FPrintf(  f,   "        EOS gamma Table: OK\n" );
          zlsio_FPrintf(  f,   "        Variable gamma : ENABLED\n" );
        } else {
          zlsio_FPrintf(  f,   "        Variable gamma : DISABLED\n" );
				}
				if (phys->coolParameters.coolUdValues) {
					zlsio_FPrintf(  f,   "        EOS U Table: OK\n" );
				}
#endif !VARIABLE_GAMMA
        
#ifdef H_IONISATION
				if (phys->coolParameters.coolfh2Values) {
					zlsio_FPrintf(  f,   "EOS Ionisation H2 Table: OK\n" );
				}
				if (phys->coolParameters.coolfhiValues) {
					zlsio_FPrintf(  f,   "EOS Ionisation HI Table: OK\n" );
				}
				if (phys->coolParameters.coolfhiiValues) {
					zlsio_FPrintf(  f,   "EOS Ionisation HIITable: OK\n" );
				}
#endif !H_IONISATION
        
#endif !VARIABLE_EOS
				
			} else if ( phys->coolParameters.coolType == kPhysPowerCooling ) {

        Real q1    = (k_mu*k_amu); q1   *= q1;

				zlsio_FPrintf(  f,   "      Powerlaw Cooling L_0 (T*T_0)^alpha:\n" );
				zlsio_FPrintf(  f,   "      L_0 (%s): %g\n" ,s_Lambda  , phys->coolParameters.coolLambda0*f_lam0*q1);
				zlsio_FPrintf(  f,   "      T_0   (1/Kelvin): %g\n" , phys->coolParameters.coolT0 );
				zlsio_FPrintf(  f,   "      alpha           : %g\n" , phys->coolParameters.coolPower );
				
			} else {
				zlsio_FPrintf(  f,   "      Unknown Cooling, Error??\n" );
			}
			zlsio_FPrintf(  f,   "\n      Maximum Operational T: %g\n",phys->coolParameters.coolMaxT );
			zlsio_FPrintf(  f,   "      Minumum Operational T: %g\n",phys->coolParameters.coolMinT );
			zlsio_FPrintf(  f,   "\n    Maximum Operational P/d: %g\n",phys->coolParameters.coolMaxPd );
			zlsio_FPrintf(  f,   "    Minumum Operational P/d: %g\n",phys->coolParameters.coolMinPd );
                  
			zlsio_FPrintf(  f,   "%s\n", DIVIDERLINE );
      
		} else {
			zlsio_FPrintf(  f,   " Cooling is enabled, but OFF\n" );
			zlsio_FPrintf(  f,   "%s\n", DIVIDERLINE );
		}
		
		
#else
		zlsio_FPrintf(  f,   "%s\n", DIVIDERLINE );
		zlsio_FPrintf(  f,   " ZLS Physics EOS \n" );
		zlsio_FPrintf(  f,   "%s\n", DIVIDERLINE );
		zlsio_FPrintf(  f,   "   Mean Molecular Weight (mu) : %g \n", phys->mu );
		zlsio_FPrintf(  f,   "   Adiabatic Index    (gamma) : %g \n", phys->gamma );
		zlsio_FPrintf(  f,   "%s\n", DIVIDERLINE );
		zlsio_FPrintf(  f,   " ZLS Physics Cooling is disabled, OFF\n" );
		zlsio_FPrintf(  f,   "%s\n", DIVIDERLINE );
#endif !COOLING
		
	}
	
	zlsio_FPrintf(  f,   "%s\n", DIVIDERLINE );
	zlsio_FPrintf(  f,   " ZLS Scaling Factors Initialised \n" );
	zlsio_FPrintf(  f,   "%s\n", DIVIDERLINE );
  zlsio_FPrintf(  f,   "   Time   Scale     t : %g %s\n", f_t0 , s_Time);
	zlsio_FPrintf(  f,   "   Length Scale     x : %g %s\n", f_x0 , s_Length);
	zlsio_FPrintf(  f,   "   Area Scale       A : %g %s\n", f_area0 , s_Area);
	zlsio_FPrintf(  f,   "   Volume Scale   Vol : %g %s\n", f_vol0 , s_Volume);
	zlsio_FPrintf(  f,   "   Velocity Scale   v : %g %s\n", f_v0 , s_Velocity);
	zlsio_FPrintf(  f,   "   Accel. Scale     a : %g %s\n", f_a0 , s_Accel);
	zlsio_FPrintf(  f,   "   Density Scale    d : %g %s\n", f_d0 , s_Density);
	zlsio_FPrintf(  f,   "   Pressure Scale   P : %g %s\n", f_P0 , s_Pressure);
	zlsio_FPrintf(  f,   "   Energy Scale     E : %g %s\n", f_E0 , s_Energy);
	zlsio_FPrintf(  f,   "   Mass Scale       m : %g %s\n", f_m0 , s_Mass);
	zlsio_FPrintf(  f,   "   Temperature Scale T : %g %s\n", f_tep0 , s_Temperature);
#ifdef COOLING
	zlsio_FPrintf(  f,   "\n" );
	zlsio_FPrintf(  f,   "   Cooling Scale     L : %g %s\n", f_cool0, s_Cool);
#endif
  
#ifdef MHD
	zlsio_FPrintf(  f,   "   Magnetic Scale    B : %g %s\n", f_B0   , s_Magnetic);
#endif MHD

#ifdef GRAVITY
	zlsio_FPrintf(  f,   "%s\n", DIVIDERLINE );
	zlsio_FPrintf(  f,   "   Gravity Constant G: %g %s\n", k_guniv , s_Gravity);
	zlsio_FPrintf(  f,   "   Normalised G      : %g x^3/m/t^2  \n", k_g     );
	zlsio_FPrintf(  f,   "   Normalised GM_sol : %g \n", k_gmsol );
#endif
	zlsio_FPrintf(  f,   "%s\n", DIVIDERLINE );
	
	return(err);
}


int zlsp_ZPrintPhysicsConfiguration( zFile f, char * cfile ){
	int err = noErr;
	
	PhysicsPtr        phys =  gLevels[0].levelPhysics; // get the global physics settings
	
	if (phys != NULL) {
		
#ifdef COOLING
		zlsio_ZFPrintf( f, "%s\n", DIVIDERLINE );
		zlsio_ZFPrintf( f, " ZLS Physics EOS \n" );
		zlsio_ZFPrintf( f, "%s\n", DIVIDERLINE );
		zlsio_ZFPrintf( f, "   Mean Molecular Weight (mu) : %g \n", phys->mu );
		zlsio_ZFPrintf( f, "   Adiabatic Indices  (gamma) : %g global: %g \n", phys->gamma,  kZLS_eos_gamma );
		zlsio_ZFPrintf( f, "%s\n", DIVIDERLINE );
		zlsio_ZFPrintf( f, " ZLS Physics Cooling \n" );
		zlsio_ZFPrintf( f, "%s\n", DIVIDERLINE );
		if (phys->coolParameters.coolType > kPhysNoCooling ){	
      
			zlsio_ZFPrintf( f, " Cooling is enabled, and ON\n" );
			zlsio_ZFPrintf( f, " Cooling Timescale Fraction: %g\n", kZLS_CoolTimeFraction);
      
			if ( phys->coolParameters.coolType == kPhysFileCooling ) {
				if (cfile) zlsio_ZFPrintf( f, "\n      Cooling File in Use: %s\n",cfile );
				zlsio_ZFPrintf( f, "        N values: %d\n", phys->coolParameters.coolNTable );
				zlsio_ZFPrintf( f, "        Table Min.  T: %g\n", phys->coolParameters.coolTTableMin );
				zlsio_ZFPrintf( f, "        Table Max.  T: %g\n", phys->coolParameters.coolTTableMax );
				zlsio_ZFPrintf( f, "        Table delta T: %g\n", phys->coolParameters.coolDeltaT );
				if (phys->coolParameters.coolFunction) {
					zlsio_ZFPrintf( f, "        Main Table     : OK\n" );
				} else {
					zlsio_ZFPrintf( f, "        ERROR: Main Table: BAD - Not ALLOCATED!\n" );
					err = memErr;
					return(err);
				}
        
        if (phys->coolParameters.coolFunctionAt) {
          zlsio_ZFPrintf(  f,   "         Atomic Table  : OK\n" );
        }
        if (phys->coolParameters.coolFunctionCO) {
          zlsio_ZFPrintf(  f,   "        Mol. CO Table  : OK\n" );
        }
        if (phys->coolParameters.coolFunctionH2) {
          zlsio_ZFPrintf(  f,   "        Mol. H2 Table  : OK\n" );
        }
        
				if (phys->coolParameters.coolTValues) {
					zlsio_ZFPrintf( f, "        Temperatures   : OK\n" );
				} else {
					zlsio_ZFPrintf( f, "        ERROR: Temperature Table: BAD - Not ALLOCATED!\n" );
					err = memErr;
					return(err);
				}
        
        if (phys->coolParameters.coolFunction) {
          if ( phys->coolParameters.coolFlagTot == kZLS_SingleCoolingFunction) {
            zlsio_ZFPrintf(  f,"  Single Total Cooling : ENABLED\n" );
          } else {
            
            zlsio_ZFPrintf(  f, "     Component Cooling : ENABLED\n" );
            
            if (phys->coolParameters.coolFunctionAt) {
              if ( phys->coolParameters.coolFlagAt == kZLS_AllowCoolingComponent) {
                zlsio_ZFPrintf(  f,   "                 Atomic Cooling: ENABLED\n" );
              } else {
                zlsio_ZFPrintf(  f,   "                 Atomic Cooling: DISABLED\n" );
              }
            }
            if (phys->coolParameters.coolFunctionCO) {
              if ( phys->coolParameters.coolFlagCO == kZLS_AllowCoolingComponent) {
                zlsio_ZFPrintf(  f,   "                Mol. CO Cooling: ENABLED\n" );
              } else {
                zlsio_ZFPrintf(  f,   "                Mol. CO Cooling: DISABLED\n" );
              }
            }
            if (phys->coolParameters.coolFunctionH2) {
              if ( phys->coolParameters.coolFlagH2 == kZLS_AllowCoolingComponent) {
                zlsio_ZFPrintf(  f,   "                Mol. H2 Cooling: ENABLED\n" );
              } else {
                zlsio_ZFPrintf(  f,   "                Mol. H2 Cooling: DISABLED\n" );
              }
            }
            
          }
				}
				
        
				
#ifdef VARIABLE_EOS
				if (phys->coolParameters.coolmuValues) {
					zlsio_ZFPrintf(  f,   "        EOS mu Table   : OK\n" );
          zlsio_ZFPrintf(  f,   "           Variable mu : ENABLED\n" );
        } else {
          zlsio_ZFPrintf(  f,   "           Variable mu : DISABLED\n" );
				}
				
#ifdef VARIABLE_GAMMA
				if (phys->coolParameters.coolGammaValues) {
					zlsio_ZFPrintf(  f,   "        EOS gamma Table: OK\n" );
          zlsio_ZFPrintf(  f,   "        Variable gamma : ENABLED\n" );
        } else {
          zlsio_ZFPrintf(  f,   "        Variable gamma : DISABLED\n" );
				}
				if (phys->coolParameters.coolUdValues) {
					zlsio_ZFPrintf(  f,   "        EOS U Table: OK\n" );
				}
#endif !VARIABLE_GAMMA
        
#ifdef H_IONISATION
				if (phys->coolParameters.coolfh2Values) {
					zlsio_ZFPrintf(  f,   "EOS Ionisation H2 Table: OK\n" );
				}
				if (phys->coolParameters.coolfhiValues) {
					zlsio_ZFPrintf(  f,   "EOS Ionisation HI Table: OK\n" );
				}
				if (phys->coolParameters.coolfhiiValues) {
					zlsio_ZFPrintf(  f,   "EOS Ionisation HIITable: OK\n" );
				}
#endif !H_IONISATION
        
#endif !VARIABLE_EOS
				
			} else if ( phys->coolParameters.coolType == kPhysPowerCooling ) {
        Real q1    = (k_mu*k_amu); q1   *= q1;
				zlsio_ZFPrintf(  f,   "      Powerlaw Cooling L_0 (T*T_0)^alpha:\n" );
				zlsio_ZFPrintf(  f,   "      L_0 (%s): %g\n" ,s_Lambda  , phys->coolParameters.coolLambda0*f_lam0*q1);
				zlsio_ZFPrintf(  f,   "      T_0   (1/Kelvin): %g\n" , phys->coolParameters.coolT0 );
				zlsio_ZFPrintf(  f,   "                alpha : %g\n" , phys->coolParameters.coolPower );
				
			} else {
				zlsio_ZFPrintf( f, "      Unkonwn Cooling, Error??\n" );
			}
			zlsio_ZFPrintf( f, "\n      Maximum Operational T: %g\n",phys->coolParameters.coolMaxT );
			zlsio_ZFPrintf( f, "      Minumum Operational T: %g\n",phys->coolParameters.coolMinT );
			zlsio_ZFPrintf( f, "\n    Maximum Operational P/d: %g\n",phys->coolParameters.coolMaxPd );
			zlsio_ZFPrintf( f, "    Minumum Operational P/d: %g\n",phys->coolParameters.coolMinPd );
			zlsio_ZFPrintf( f, "%s\n", DIVIDERLINE );
			
		} else {
			zlsio_ZFPrintf( f, " Cooling is enabled, but OFF\n" );
			zlsio_ZFPrintf( f, "%s\n", DIVIDERLINE );
		}
		
		
		
#else
		zlsio_ZFPrintf( f, "%s\n", DIVIDERLINE );
		zlsio_ZFPrintf( f, " ZLS Physics EOS \n" );
		zlsio_ZFPrintf( f, "%s\n", DIVIDERLINE );
		zlsio_ZFPrintf( f, "   Mean Molecular Weight (mu) : %g \n", phys->mu );
		zlsio_ZFPrintf( f, "   Adiabatic Indices  (gamma) : %g \n", kZLS_eos_gamma );
		zlsio_ZFPrintf( f, "%s\n", DIVIDERLINE );
		zlsio_ZFPrintf( f, " ZLS Physics Cooling is disabled, OFF\n" );
		zlsio_ZFPrintf( f, "%s\n", DIVIDERLINE );
#endif !COOLING
		
	}
	
	zlsio_ZFPrintf( f, "%s\n", DIVIDERLINE );
	zlsio_ZFPrintf( f, " ZLS Scaling Factors Initialised \n" );
	zlsio_ZFPrintf( f, "%s\n", DIVIDERLINE );
  zlsio_ZFPrintf(  f,   "   Time   Scale     t : %g %s\n", f_t0 , s_Time);
	zlsio_ZFPrintf(  f,   "   Length Scale     x : %g %s\n", f_x0 , s_Length);
	zlsio_ZFPrintf(  f,   "   Area Scale       A : %g %s\n", f_area0 , s_Area);
	zlsio_ZFPrintf(  f,   "   Volume Scale   Vol : %g %s\n", f_vol0 , s_Volume);
	zlsio_ZFPrintf(  f,   "   Velocity Scale   v : %g %s\n", f_v0 , s_Velocity);
	zlsio_ZFPrintf(  f,   "   Accel. Scale     a : %g %s\n", f_a0 , s_Accel);
	zlsio_ZFPrintf(  f,   "   Density Scale    d : %g %s\n", f_d0 , s_Density);
	zlsio_ZFPrintf(  f,   "   Pressure Scale   P : %g %s\n", f_P0 , s_Pressure);
	zlsio_ZFPrintf(  f,   "   Energy Scale     E : %g %s\n", f_E0 , s_Energy);
	zlsio_ZFPrintf(  f,   "   Mass Scale       m : %g %s\n", f_m0 , s_Mass);
	zlsio_ZFPrintf(  f,   "   Temperature Scale T : %g %s\n", f_tep0 , s_Temperature);
#ifdef COOLING
	zlsio_ZFPrintf(  f,   "\n" );
	zlsio_ZFPrintf(  f,   "   Cooling Scale     L : %g %s\n", f_cool0, s_Cool);
#endif
#ifdef MHD
	zlsio_ZFPrintf(  f,   "\n" );
	zlsio_ZFPrintf( f, "   Magnetic Scale    : %g %s\n", f_B0   , s_Magnetic);
#endif MHD
#ifdef GRAVITY
	zlsio_ZFPrintf(  f,   "%s\n", DIVIDERLINE );
	zlsio_ZFPrintf(  f,   "   Gravity Constant G: %g %s\n", k_guniv , s_Gravity);
	zlsio_ZFPrintf(  f,   "   Normalised G      : %g x^3/m/t^2  \n", k_g     );
	zlsio_ZFPrintf(  f,   "   Normalised GM_sol : %g   \n", k_gmsol );
#endif
	zlsio_ZFPrintf( f, "%s\n", DIVIDERLINE );
	
	return(err);
}


//====================================================================================
//
// Getting and Setting W and U structures from zonedata
//
//====================================================================================

int zlsp_VelocitySlice( QPtr q, QPtr velocity, Real vMin, Real vMax, GridCoords gc){
  // take nDimensional quantity q, and velocity data, can clip q where v is between vMin and vMax
  // using gc to get array sizes
  // q is a plain quantity from a call to zlsp_GetZoneWQuantityCopy
  // velocity is a plain quantity from a call to zlsp_GetZoneWQuantityCopy as well.
  int err = noErr;
#ifdef SYM1D
  Integer    i;
  QPtr   q1d  = q;
  QPtr   v1d  = velocity;
#endif SYM1D
#ifdef SYM2D
  Integer    i, j;
  Q2DArr q2d;
  Q2DArr v2d;
#endif SYM2D
#ifdef SYM3D
  Integer    i, j, k;
  Q3DArr q3d;
  Q3DArr v3d;
#endif SYM3D
  Real val, vel; // srcatch vars
  
#ifdef FILES_NO_BCS    
  
  // since we have used zlsp_GetZoneWQuantityCopy and are just 
  // slicing now, then if the BCs are not being saved then 
  // they will already be removed by zlsp_GetZoneWQuantityCopy
  // in that case the loop will go from 0 -> n-1 , is = 0 etc
  
#ifdef SYM1D
  Integer ni  = gc.n[kXDim];
#endif SYM1D
  
#ifdef SYM2D
  Integer ni  = gc.n[kXDim];
  Integer nj  = gc.n[kYDim];
#endif SYM2D
  
#ifdef SYM3D
  Integer ni  = gc.n[kXDim];
  Integer nj  = gc.n[kYDim];
  Integer nk  = gc.n[kZDim];
#endif SYM3D
  
  
#endif FILES_NO_BCS
  
#ifdef FILES_HAVE_BCS          
  // in this case the loop will go from 0 -> nt-1 , is = 0 etc
  
#ifdef SYM1D
  Integer ni  = gc.nt[kXDim];
#endif SYM1D
  
#ifdef SYM2D
  Integer ni  = gc.nt[kXDim];
  Integer nj  = gc.nt[kYDim];
#endif SYM2D
  
#ifdef SYM3D
  Integer ni  = gc.nt[kXDim];
  Integer nj  = gc.nt[kYDim];
  Integer nk  = gc.nt[kZDim];
#endif SYM3D
  
#endif FILES_HAVE_BCS
  
  if (( q != NULL) && ( velocity != NULL)) {
    
#ifdef SYM1D
    for (i = 0; i<ni; i++){
      val = q1d[i];
      vel = v1d[i];
      val = (vel > vMin)? val: vMin;
      val = (vel < vMax)? val: vMin;
      q1d[i] = val;
    }
#endif SYM1D
#ifdef SYM2D
    err = zls_SetQuantity2DArr( &q2d, q, nj,  ni );
    err = zls_SetQuantity2DArr( &v2d, velocity, nj,  ni );
    
    for (j = 0; j<nj; j++){
      for (i = 0; i<ni; i++){
        val = q2d[j][i];
        vel = v2d[j][i];
        val = (vel > vMin)? val: vMin;
        val = (vel < vMax)? val: vMin;
        
        q2d[j][i] = val;
      }
    }
    
    zls_ReleaseQuantity2DArr(q2d);
    zls_ReleaseQuantity2DArr(v2d);
#endif SYM2D
#ifdef SYM3D
    err = zls_SetQuantity3DArr( &q3d, q, nk, nj,  ni );
    err = zls_SetQuantity3DArr( &v3d, velocity, nk, nj,  ni );
    
    for (k = 0; k<nk; k++){
      for (j = 0; j<nj; j++){
        for (i = 0; i<ni; i++){
          val = q3d[k][j][i];
          vel = v3d[k][j][i];
          val = (vel > vMin)? val: vMin;
          val = (vel < vMax)? val: vMin;
          q3d[k][j][i] = val;
        }
      }
    }
    zls_ReleaseQuantity3DArr(q3d);
    zls_ReleaseQuantity3DArr(v3d);
#endif SYM3D
    
  } else { err = memErr;} // q, velocity
  
  return(err);
}


int zlsp_SymmetricDifference( QPtr q, GridCoords gc){
  // q is a plain quantity from a call to zlsp_GetZoneWQuantityCopy
  int err = noErr;
#ifdef SYM1D
  return(noErr);
#else
#ifdef SYM2D
  Integer    i, j;
  Q2DArr q2d;
#endif SYM2D
#ifdef SYM3D
  Integer    i, j, k;
  Q3DArr q3d;
#endif SYM3D
  Real val; // srcatch var
  
#ifdef FILES_NO_BCS    
  
  // since we have used zlsp_GetZoneWQuantityCopy and are just 
  // slicing now, then if the BCs are not being saved then 
  // they will already be removed by zlsp_GetZoneWQuantityCopy
  // in that case the loop will go from 0 -> n-1 , is = 0 etc
  
#ifdef SYM2D
  Integer nj  = gc.n[kYDim];
  Integer ni  = zls_min(nj, gc.n[kXDim]);
#endif SYM2D
  
#ifdef SYM3D
  Integer nk  = gc.n[kZDim];
  Integer nj  = zls_min(nk, gc.n[kYDim]);
  Integer ni  = zls_min(nj, gc.n[kXDim]);
#endif SYM3D
  
  
#endif FILES_NO_BCS
  
#ifdef FILES_HAVE_BCS          
  // in this case the loop will go from 0 -> nt-1 , is = 0 etc
  
#ifdef SYM2D
  Integer nj  = gc.nt[kYDim];
  Integer ni  = zls_min(nj, gc.nt[kXDim]);
#endif SYM2D
  
#ifdef SYM3D
  Integer nk  = gc.nt[kZDim];
  Integer nj  = zls_min(nk, gc.nt[kYDim]);
  Integer ni  = zls_min(nj, gc.nt[kXDim]);
#endif SYM3D
  
#endif FILES_HAVE_BCS
  
  if ( q != NULL) {
    
#ifdef SYM2D
    err = zls_SetQuantity2DArr( &q2d, q, nj,  ni );
    for (j = 0; j<nj; j++){
      for (i = j; i<ni; i++){
        val = 0.5*(q2d[j][i] + q2d[i][j]);
        if ( i == j)  {
          q2d[j][i] = 0.0;
        } else {
          q2d[j][i] -= val;
          q2d[i][j] -= val;
        }
      }
    }
    zls_ReleaseQuantity2DArr(q2d);
#endif SYM2D
    
#ifdef SYM3D
    err = zls_SetQuantity3DArr( &q3d, q, nk, nj,  ni );
    
    for (k = 0; k<nk; k++){
      for (j = k; j<nj; j++){
        for (i = j; i<ni; i++){
          val = (q3d[k][j][i]+q3d[i][k][j]+q3d[j][i][k])/3.0;
          if ( (i == j)&&(i == k) ) {
            q3d[k][j][i]  = 0.0;
          } else {
            q3d[k][j][i] -= val;
            q3d[i][k][j] -= val;
            q3d[j][i][k] -= val;
          }
        }
      }
    }
    zls_ReleaseQuantity3DArr(q3d);
#endif SYM3D
    
  } else { err = memErr;} // q, velocity
  
  return(err);
#endif !SYM1D
}

#ifdef PRIMITIVE_QUAS
#ifdef HYDRO
int zlsp_ReferenceZoneWHD( WHD * w, ZonePtr z ){ 
  //
  // usually w is a reference to a static stack local var in caller
  // ie err = zlsp_ReferenceZoneWHD(&w, z);  There are no New/Dispose for WHDs.
  //
  Integer err = noErr;
#if (NTRACERS > 0)    
  Integer i;
#endif	
  if ((z != NULL)&&(w != NULL)) {
    
		
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    GridCoords gc = z->zoneGrid;
    //
    // wrap multi-d array wrappers on plain blocks, in 2D and 3D. must release w when finished
    // at least in 2D and 3D.
    //
		
#ifdef SYM3D
    Integer ni = gc.nt[kXDim];
    Integer nj = gc.nt[kYDim];
    Integer nk = gc.nt[kZDim];
#endif SYM3D
		
#ifdef SYM2D
    Integer ni = gc.nt[kXDim];
    Integer nj = gc.nt[kYDim];
#endif SYM2D
		
#ifdef SYM1D
		
    w->d  = z->zoneData[kWDens];
    w->P  = z->zoneData[kWPres];
    w->vx = z->zoneData[kWVx];
#if (NTRACERS > 0)    
    for (i = 0; i < kNTracers; i++){
      w->tr[i] = z->zoneData[i+kTracer0];
    }
#endif

#endif SYM1D
		
#ifdef SYM2D
    err = zls_SetQuantity2DArr(&(w->d) ,  z->zoneData[kWDens], nj, ni);
    err = zls_SetQuantity2DArr(&(w->P) ,  z->zoneData[kWPres], nj, ni);
    err = zls_SetQuantity2DArr(&(w->vx),  z->zoneData[kWVx], nj, ni);
    err = zls_SetQuantity2DArr(&(w->vy),  z->zoneData[kWVy], nj, ni);
		
#if (NTRACERS > 0)    
    Integer iv;
    for (i = 0; i < kNTracers; i++){
      iv = i+kTracer0;
      err = zls_SetQuantity2DArr(&(w->tr[i]),  z->zoneData[iv], nj, ni);
    }
#endif

#endif SYM2D
		
#ifdef SYM3D
		
    err = zls_SetQuantity3DArr(&(w->d) ,  z->zoneData[kWDens], nk, nj, ni);
    err = zls_SetQuantity3DArr(&(w->P) ,  z->zoneData[kWPres], nk, nj, ni);
    err = zls_SetQuantity3DArr(&(w->vx),  z->zoneData[kWVx], nk, nj, ni);
    err = zls_SetQuantity3DArr(&(w->vy),  z->zoneData[kWVy], nk, nj, ni);
    err = zls_SetQuantity3DArr(&(w->vz),  z->zoneData[kWVz], nk, nj, ni);
#if (NTRACERS > 0)    
    for (i = 0; i < kNTracers; i++){
      err = zls_SetQuantity3DArr(&(w->tr[i]),  z->zoneData[i+kTracer0], nk, nj, ni);
    }
#endif
		
#endif SYM3D
		
#endif ZONEDATA_ARE_PLAINBLOCKS
		
#ifdef ZONEDATA_ARE_MULTIDARRAYS
		
#ifdef SYM1D
    w->d = z->zoneData[kWDens];
    w->P = z->zoneData[kWPres];
    w->vx = z->zoneData[kWVx];
#if (NTRACERS > 0)    
    for (i = 0; i < kNTracers; i++){
      w->tr[i] = z->zoneData[i+kTracer0];
    }
#endif
#endif SYM1D
		
#ifdef SYM2D
    w->d = z->zoneData[kWDens];
    w->P = z->zoneData[kWPres];
    w->vx = z->zoneData[kWVx];
    w->vy = z->zoneData[kWVy];
#if (NTRACERS > 0)    
    for (i = 0; i < kNTracers; i++){
      w->tr[i] = z->zoneData[i+kTracer0];
    }
#endif
#endif SYM2D
		
#ifdef SYM3D
    w->d = z->zoneData[kWDens];
    w->P = z->zoneData[kWPres];
    w->vx = z->zoneData[kWVx];
    w->vy = z->zoneData[kWVy];
    w->vz = z->zoneData[kWVz];
#if (NTRACERS > 0)    
    for (i = 0; i < kNTracers; i++){
      w->tr[i] = z->zoneData[i+kTracer0];
    }
#endif
#endif SYM3D
		
#endif ZONEDATA_ARE_MULTIDARRAYS
  } else { err = memErr; }
	
  return(err);
}
#endif HYDRO

#endif PRIMITIVE_QUAS


#ifdef CONSERVED_QUAS
#ifdef HYDRO
int zlsp_ReferenceZoneUHD( UHD * u, ZonePtr z ){ 
  //
  // usually u is a reference to a static stack local var in caller
  // ie err = GetZoneUHD(&u, z);  There are no New/Dispose for UHDs.
  //
  Integer err = noErr;
#if (NTRACERS > 0)    
  Integer i;
#endif	
  if ((z != NULL)&&(w != NULL)) {
    
		
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    GridCoords gc = z->zoneGrid;
    //
    // wrap multi-d array wrappers on plain blocks, in 2D and 3D. must release w when finished
    // at least in 2D and 3D.
    //
		
#ifdef SYM3D
    Integer ni = gc.nt[kXDim];
    Integer nj = gc.nt[kYDim];
    Integer nk = gc.nt[kZDim];
#endif SYM3D
		
#ifdef SYM2D
    Integer ni = gc.nt[kXDim];
    Integer nj = gc.nt[kYDim];
#endif SYM2D
		
#ifdef SYM1D
		
    u->d  = z->zoneData[kUDens];
    u->E  = z->zoneData[kUEneg];
    u->mx = z->zoneData[kUMx];
    
#if (NTRACERS > 0)    
    for (i = 0; i < kNTracers; i++){
      u->tr[i] = z->zoneData[i+kTracer0];
    }
#endif
#endif SYM1D
		
#ifdef SYM2D
    err = zls_SetQuantity2DArr(&(u->d) ,  z->zoneData[kUDens], nj, ni);
    err = zls_SetQuantity2DArr(&(u->E) ,  z->zoneData[kUEneg], nj, ni);
    
    err = zls_SetQuantity2DArr(&(u->mx),  z->zoneData[kUMx], nj, ni);
    err = zls_SetQuantity2DArr(&(u->my),  z->zoneData[kUMy], nj, ni);
		
#if (NTRACERS > 0)    
    Integer iv;
    for (i = 0; i < kNTracers; i++){
      iv = i+kTracer0;
      err = zls_SetQuantity2DArr(&(u->tr[i]),  z->zoneData[iv], nj, ni);
    }
#endif
#endif SYM2D
		
#ifdef SYM3D
		
    err = zls_SetQuantity3DArr(&(u->d) ,  z->zoneData[kUDens], nk, nj, ni);
    err = zls_SetQuantity3DArr(&(u->E) ,  z->zoneData[kUEneg], nk, nj, ni);
    
    err = zls_SetQuantity3DArr(&(u->mx),  z->zoneData[kUMx], nk, nj, ni);
    err = zls_SetQuantity3DArr(&(u->my),  z->zoneData[kUMy], nk, nj, ni);
    err = zls_SetQuantity3DArr(&(u->mz),  z->zoneData[kUMz], nk, nj, ni);
    
#if (NTRACERS > 0)    
    for (i = 0; i < kNTracers; i++){
      err = zls_SetQuantity3DArr(&(u->tr[i]),  z->zoneData[i+kTracer0], nk, nj, ni);
    }
#endif
		
#endif SYM3D
		
#endif ZONEDATA_ARE_PLAINBLOCKS
		
#ifdef ZONEDATA_ARE_MULTIDARRAYS
		
#ifdef SYM1D
    u->d = z->zoneData[kUDens];
    u->E = z->zoneData[kUEneg];
    u->mx = z->zoneData[kUMx];
    
#if (NTRACERS > 0)    
    for (i = 0; i < kNTracers; i++){
      u->tr[i] = z->zoneData[i+kTracer0];
    }
#endif
#endif SYM1D
		
#ifdef SYM2D
    u->d = z->zoneData[kUDens];
    u->E = z->zoneData[kUEneg];
    u->mx = z->zoneData[kUMx];
    u->my = z->zoneData[kUMy];
    
#if (NTRACERS > 0)    
    for (i = 0; i < kNTracers; i++){
      u->tr[i] = z->zoneData[i+kTracer0];
    }
#endif
#endif SYM2D
		
#ifdef SYM3D
    u->d = z->zoneData[kUDens];
    u->E = z->zoneData[kUEneg];
    u->mx = z->zoneData[kUMx];
    u->my = z->zoneData[kUMy];
    u->mz = z->zoneData[kUMz];
    
#if (NTRACERS > 0)    
    for (i = 0; i < kNTracers; i++){
      u->tr[i] = z->zoneData[i+kTracer0];
    }
#endif
#endif SYM3D
		
#endif ZONEDATA_ARE_MULTIDARRAYS
  } else { err = memErr; }
	
  return(err);
}
#endif HYDRO

#endif CONSERVED_QUAS

#ifdef MEANSPLITSWEEPS

int zlsp_ReferenceZoneWHDAlt( WHD * w, ZonePtr z ){ 
  //
  // usually w is a reference to a static stack local var in caller
  // ie err = GetZoneWHD(&w, z);  There are no New/Dispose for WHDs.
  //
  Integer err = noErr;
#if (NTRACERS > 0)    
  Integer i;
#endif	
  if ((z != NULL)&&(w != NULL)) {
    
		
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    GridCoords gc = z->zoneGrid;
    //
    // wrap multi-d array wrappers on plain blocks, in 2D and 3D. must release w when finished
    // at least in 2D and 3D.
    //
		
#ifdef SYM3D
    Integer ni = gc.nt[kXDim];
    Integer nj = gc.nt[kYDim];
    Integer nk = gc.nt[kZDim];
#endif SYM3D
		
#ifdef SYM2D
    Integer ni = gc.nt[kXDim];
    Integer nj = gc.nt[kYDim];
#endif SYM2D
		
#ifdef SYM1D
		
    w->d  = z->zoneData[kWDens];
    w->P  = z->zoneData[kWPres];
    w->vx = z->zoneData[kWVx];
#if (NTRACERS > 0)    
    for (i = 0; i < kNTracers; i++){
      w->tr[i] = z->zoneData[i+kTracer0];
    }
#endif
#endif SYM1D
		
#ifdef SYM2D
    err = zls_SetQuantity2DArr(&(w->d) ,  z->zoneData[kWDens+kAltVarsOffset], nj, ni);
    err = zls_SetQuantity2DArr(&(w->P) ,  z->zoneData[kWPres+kAltVarsOffset], nj, ni);
    err = zls_SetQuantity2DArr(&(w->vx),  z->zoneData[kWVx+kAltVarsOffset], nj, ni);
    err = zls_SetQuantity2DArr(&(w->vy),  z->zoneData[kWVy+kAltVarsOffset], nj, ni);
		
#if (NTRACERS > 0)    
    Integer iv;
    for (i = 0; i < kNTracers; i++){
      iv = i+kTracer0+kAltVarsOffset;
      err = zls_SetQuantity2DArr(&(w->tr[i]),  z->zoneData[iv], nj, ni);
    }
#endif
#endif SYM2D
		
#ifdef SYM3D
		
    err = zls_SetQuantity3DArr(&(w->d) ,  z->zoneData[kWDens], nk, nj, ni);
    err = zls_SetQuantity3DArr(&(w->P) ,  z->zoneData[kWPres], nk, nj, ni);
    err = zls_SetQuantity3DArr(&(w->vx),  z->zoneData[kWVx], nk, nj, ni);
    err = zls_SetQuantity3DArr(&(w->vy),  z->zoneData[kWVy], nk, nj, ni);
    err = zls_SetQuantity3DArr(&(w->vz),  z->zoneData[kWVz], nk, nj, ni);
#if (NTRACERS > 0)    
    for (i = 0; i < kNTracers; i++){
      err = zls_SetQuantity3DArr(&(w->tr[i]),  z->zoneData[i+kTracer0], nk, nj, ni);
    }
#endif
#endif SYM3D
		
#endif ZONEDATA_ARE_PLAINBLOCKS
		
#ifdef ZONEDATA_ARE_MULTIDARRAYS
		
#ifdef SYM1D
    w->d = z->zoneData[kWDens+kAltVarsOffset];
    w->P = z->zoneData[kWPres+kAltVarsOffset];
    w->vx = z->zoneData[kWVx+kAltVarsOffset];
#if (NTRACERS > 0)    
    for (i = 0; i < kNTracers; i++){
      w->tr[i] = z->zoneData[i+kTracer0+kAltVarsOffset];
    }
#endif
#endif SYM1D
		
#ifdef SYM2D
    w->d = z->zoneData[kWDens+kAltVarsOffset];
    w->P = z->zoneData[kWPres+kAltVarsOffset];
    w->vx = z->zoneData[kWVx+kAltVarsOffset];
    w->vy = z->zoneData[kWVy+kAltVarsOffset];
#if (NTRACERS > 0)    
    for (i = 0; i < kNTracers; i++){
      w->tr[i] = z->zoneData[i+kTracer0+kAltVarsOffset];
    }
#endif
#endif SYM2D
		
#ifdef SYM3D
    w->d = z->zoneData[kWDens+kAltVarsOffset];
    w->P = z->zoneData[kWPres+kAltVarsOffset];
    w->vx = z->zoneData[kWVx+kAltVarsOffset];
    w->vy = z->zoneData[kWVy+kAltVarsOffset];
    w->vz = z->zoneData[kWVz+kAltVarsOffset];
#if (NTRACERS > 0)    
    for (i = 0; i < kNTracers; i++){
      w->tr[i] = z->zoneData[i+kTracer0+kAltVarsOffset];
    }
#endif
#endif SYM3D
		
#endif ZONEDATA_ARE_MULTIDARRAYS
  } else { err = memErr; }
	
  return(err);
}
#endif MEANSPLITSWEEPS

#ifdef HYDRO
int zlsp_ReleaseZoneWHD(WHD w){ 
  //
  // usually w is a reference to a static stack local var in caller
  // ie err = GetZoneWHD(&w, z);  There are no New/Dispose for WHDs.
  // do nothing for ZONEDATA_ARE_MULTIDARRAYS
  Integer err = noErr;
#ifdef ZONEDATA_ARE_PLAINBLOCKS
  if (w.d != NULL) {
		
#ifdef SYM1D
    //     nothing needed to do, 1D arrs are just simple refs.
#endif SYM1D
		
#ifdef SYM2D
    zls_ReleaseQuantity2DArr(w.d);
    zls_ReleaseQuantity2DArr(w.P);
    zls_ReleaseQuantity2DArr(w.vx);
    zls_ReleaseQuantity2DArr(w.vy);
#if (NTRACERS>0)
    Integer i;
    for (i = 0; i < kNTracers; i++){
      zls_ReleaseQuantity2DArr(w.tr[i]);
    }
#endif
#endif SYM2D
		
#ifdef SYM3D
    zls_ReleaseQuantity3DArr(w.d);
    zls_ReleaseQuantity3DArr(w.P);
    zls_ReleaseQuantity3DArr(w.vx);
    zls_ReleaseQuantity3DArr(w.vy);
    zls_ReleaseQuantity3DArr(w.vz);
#if (NTRACERS>0)
    Integer i;
    for (i = 0; i < kNTracers; i++){
      zls_ReleaseQuantity3DArr(w.tr[i]);
    }
#endif
#endif SYM3D
		
  } else { err = memErr; }
#else
#pragma unused(w)
#endif !ZONEDATA_ARE_PLAINBLOCKS
  return(err);
}


int zlsp_ReleaseZoneUHD(UHD u){ 
  //
  // usually w is a reference to a static stack local var in caller
  // ie err = GetZoneWHD(&w, z);  There are no New/Dispose for WHDs.
  // do nothing for ZONEDATA_ARE_MULTIDARRAYS
  Integer err = noErr;
#ifdef ZONEDATA_ARE_PLAINBLOCKS
  if (u.d != NULL) {
		
#ifdef SYM1D
    //     nothing needed to do, 1D arrs are just simple refs.
#endif SYM1D
		
#ifdef SYM2D
    zls_ReleaseQuantity2DArr(u.d);
    zls_ReleaseQuantity2DArr(u.E);
    zls_ReleaseQuantity2DArr(u.mx);
    zls_ReleaseQuantity2DArr(u.my);
#if (NTRACERS>0)
    Integer i;
    for (i = 0; i < kNTracers; i++){
      zls_ReleaseQuantity2DArr(w.tr[i]);
    }
#endif
#endif SYM2D
		
#ifdef SYM3D
    zls_ReleaseQuantity3DArr(w.d);
    zls_ReleaseQuantity3DArr(w.P);
    zls_ReleaseQuantity3DArr(w.mx);
    zls_ReleaseQuantity3DArr(w.my);
    zls_ReleaseQuantity3DArr(w.mz);
#if (NTRACERS>0)
    Integer i;
    for (i = 0; i < kNTracers; i++){
      zls_ReleaseQuantity3DArr(w.tr[i]);
    }
#endif
#endif SYM3D
		
  } else { err = memErr; }
#else
#pragma unused(w)
#endif !ZONEDATA_ARE_PLAINBLOCKS
  return(err);
}
#endif HYDRO


//====================================================================================
// Array Access Macros for GetW varaibles.
//====================================================================================
//
#ifdef SYM3D

#define LOOP  \
for (k = 0; k<nk; k++){ \
kc = k+ks; \
for (j = 0; j<nj; j++){ \
jc = j+js; \
for (i = 0; i<ni; i++){ \
ic = i+is; \

#define ENDLOOP  }; }; };        
#define CELL      [kc][jc][ic]

#ifdef SCALAR_POTENTIAL
#define CELLA0    [kc][jc][ic].a
#endif SCALAR_POTENTIAL

#ifdef VECTOR_FIELD
#define CELLA0    [kc][jc][ic].a[k0Dim]
#endif VECTOR_FIELD

#define CELLX0    [kc][jc][ic-1]
#define CELLX1    [kc][jc][ic+1]
#define CELLY0    [kc][jc-1][ic]
#define CELLY1    [kc][jc+1][ic]
#define CELLZ0    [kc-1][jc][ic]
#define CELLZ1    [kc+1][jc][ic]
#define Q        q[k][j][i]

#endif SYM3D

#ifdef SYM2D

#define LOOP  \
for (j = 0; j<nj; j++){ \
jc = j+js; \
for (i = 0; i<ni; i++){ \
ic = i+is; \

#define ENDLOOP    }; };        
#define CELL    [jc][ic]

#ifdef SCALAR_POTENTIAL
#define CELLA0  [jc][ic].a
#endif SCALAR_POTENTIAL

#ifdef VECTOR_FIELD
#define CELLA0  [jc][ic].a[k0Dim]
#endif VECTOR_FIELD

#define CELLX0  [jc][ic-1]
#define CELLX1  [jc][ic+1]
#define CELLY0  [jc-1][ic]
#define CELLY1  [jc+1][ic]
#define Q      q[j][i]

#endif SYM2D

#ifdef SYM1D

#define LOOP  \
for (i = 0; i<ni; i++){ \
ic = i+is; \

#define ENDLOOP  };        
#define CELL    [ic]

#define CELLA0  [ic].a

#define CELLX0  [ic-1]
#define CELLX1  [ic+1]
#define Q      q[i]

#endif SYM1D
//====================================================================================

int   zlsp_GetZoneWQArr( QArr q, ZonePtr z, Integer varIndex ){
  
  int err = noErr;
  
  
  if (q !=NULL){
    
    if (z != NULL) {
      
      
      Integer level     = z->zoneLevel;
      PhysicsPtr   phys = gLevels[level].levelPhysics;
      
      WHD       zoneW   = z->zoneW;
      
      Field     zoneF;
      zls_GetZoneField(&zoneF, z);
      GridCoords gc     = z->zoneGrid;
#ifdef SYM1D
      
      Integer    i;
      Integer    ic;
#ifdef FILES_NO_BCS    
      Integer is =  gc.ns[kXDim];
      Integer ni =  gc.n[kXDim];
#endif FILES_NO_BCS    
#ifdef FILES_HAVE_BCS    
      Integer is =  0;
      Integer ni =  gc.nt[kXDim];
#endif FILES_HAVE_BCS    
#endif SYM1D
      
#ifdef SYM2D
      
      Integer    i , j ;
      Integer    ic, jc;
#ifdef FILES_NO_BCS    
      Integer is =  gc.ns[kXDim];
      Integer js =  gc.ns[kYDim];
      Integer ni =  gc.n[kXDim];
      Integer nj =  gc.n[kYDim];
#endif FILES_NO_BCS    
#ifdef FILES_HAVE_BCS    
      Integer is =  0;
      Integer js =  0;
      Integer ni =  gc.nt[kXDim];
      Integer nj =  gc.nt[kYDim];
#endif FILES_HAVE_BCS    
#endif SYM2D
      
#ifdef SYM3D
      
      Integer    i , j , k ;
      Integer    ic, jc, kc;
#ifdef FILES_NO_BCS    
      Integer is =  gc.ns[kXDim];
      Integer js =  gc.ns[kYDim];
      Integer ks =  gc.ns[kZDim];
      Integer ni =  gc.n[kXDim];
      Integer nj =  gc.n[kYDim];
      Integer nk =  gc.n[kZDim];
#endif FILES_NO_BCS    
#ifdef FILES_HAVE_BCS    
      Integer is =  0;
      Integer js =  0;
      Integer ks =  0;
      Integer ni =  gc.nt[kXDim];
      Integer nj =  gc.nt[kYDim];
      Integer nk =  gc.nt[kZDim];
#endif FILES_HAVE_BCS    
      
#endif SYM3D
      
#if (NTRACERS>0)  
      Real tr;
#endif (NTRACERS>0)              
      Real dens, P, E, KE;
      Real        gamma = kZLS_eos_gamma;
      Real         mu   = k_mu;
                  
      CoolingPtr   cp   = NULL;
      
      Real T; // temperature
#ifdef COOLING  
      Real lam, lamAt, lamH2, lamCO; // cooling rates, total, H2 and CO

      Real fH[3]  = {0.0,0.0,1.0};
      
#endif COOLING  
      
      if (phys != NULL) {
        gamma  = phys->gamma;  // global plasma adiabtic index
        mu     = phys->mu;     // global plasma mu
#ifdef COOLING
        cp    = &(phys->coolParameters);
        
        if ( k_mu < 1.0) {
          fH[0] = 0.0;
          fH[1] = 0.0;
          fH[2] = 1.0; // initially fully ionised guess
        } else  if ( k_mu < 2.0) {
          fH[0] = 0.0;
          fH[1] = 1.0; // initially fully neutral guess
          fH[2] = 0.0; 
        } else {
          fH[0] = 1.0; // initially fully molecular guess
          fH[1] = 0.0;
          fH[2] = 0.0; 
        }
        
#endif COOLING  
      }
      
      Real idx = gc.x[0].idx;
#ifdef SYM2D
      Real idy =   gc.y[0].idx;
#endif SYM2D
#ifdef SYM3D
      Real idy =   gc.y[0].idx;
      // unused Real idz =   gc.z[0].idx;
#endif SYM3D
      
      
      switch (varIndex){
        case kSymDens:
          LOOP  
          Q = zoneW.d CELL;
          ENDLOOP
          break;          
        case kSclieren:
          LOOP  
          Q = 0.5*( zoneW.d CELLX1 - zoneW.d CELLX0 )*idx;
          ENDLOOP
          break;          
#if (NDIMS>1)
        case kVortex2D:
          LOOP  
          Real dvydx = 0.5*( zoneW.vy CELLX1 - zoneW.vy CELLX0 )*idx;
          Real dvxdy = 0.5*( zoneW.vx CELLY1 - zoneW.vx CELLY0 )*idy;
          Q = (dvydx - dvxdy);
          ENDLOOP
          break;          
        case kShadow:
          LOOP  
          Real dddx = 0.5*( zoneW.d CELLX1 - zoneW.d CELLX0 )*idx;
          Real dddy = 0.5*( zoneW.d CELLY1 - zoneW.d CELLY0 )*idy;
          Q = (dddx*dddx + dddy*dddy);
          ENDLOOP
          break;          
#endif (NDIMS>1)
          
#ifdef HYDRO          
#ifdef SYM1D
        case kWMx:
          LOOP  
          dens  = zoneW.d CELL;
          Q = dens*zoneW.vx CELL;
          ENDLOOP
          break;          
        case kWM:
          LOOP  
          dens  = zoneW.d CELL;
          Real vx  = zoneW.vx CELL;
          Q = dens*zls_abs(vx);
          ENDLOOP
          break;          
#endif SYM1D
#ifdef SYM2D
        case kWMx:
          LOOP  
          dens  = zoneW.d CELL;
          Q = dens*zoneW.vx CELL;
          ENDLOOP
          break;          
        case kWMy:
          LOOP  
          dens  = zoneW.d CELL;
          Q = dens*zoneW.vy CELL;
          ENDLOOP
          break;          
        case kWM:
          LOOP  
          dens  = zoneW.d CELL;
          Real vx  = zoneW.vx CELL;
          Real vy  = zoneW.vy CELL;
          Q = dens*zls_sqrt(vx*vx + vy*vy);
          ENDLOOP
          break;          
#endif SYM2D
#ifdef SYM3D
        case kWMx:
          LOOP  
          dens  = zoneW.d CELL;
          Q = dens*zoneW.vx CELL;
          ENDLOOP
          break;          
        case kWMy:
          LOOP  
          dens  = zoneW.d CELL;
          Q = dens*zoneW.vy CELL;
          ENDLOOP
          break;          
        case kWMz:
          LOOP  
          dens  = zoneW.d CELL;
          Q = dens*zoneW.vz CELL;
          ENDLOOP
          break;          
        case kWM:
          LOOP  
          dens  = zoneW.d CELL;
          Real vx  = zoneW.vx CELL;
          Real vy  = zoneW.vy CELL;
          Real vz  = zoneW.vz CELL;
          Q = dens*zls_sqrt(vx*vx + vy*vy + vz*vz);
          ENDLOOP
          break;          
#endif SYM3D
#endif HYDRO             
        case kWTotE: // total specific energy 
          // Lagrangian: P/((g-1) d) + 0.5 v^2 [+ 0.5B^2/d]
          // Eulerian: P/((g-1)) + 0.5 d v^2 [+ 0.5B^2]
          LOOP 
          
          P     = zoneW.P CELL;
          dens  = zoneW.d CELL;
          E = zlsp_Ud_Pd (P/dens, cp);
          
          KE = 0.0;
          
#ifdef HYDRO
#ifdef SYM1D
          Real vx = zoneW.vx CELL;
          KE = 0.5*(vx*vx);
#endif SYM1D
#ifdef SYM2D
          Real vx = zoneW.vx CELL;
          Real vy = zoneW.vy CELL;
          KE = 0.5*(vx*vx+vy*vy);
#endif SYM2D
#ifdef SYM3D
          Real vx = zoneW.vx CELL;
          Real vy = zoneW.vy CELL;
          Real vz = zoneW.vz CELL;
          KE = 0.5*(vx*vx+vy*vy+vz*vz);
#endif SYM3D
#ifdef LAGRANGIAN
          Q = E + KE;
#endif LAGRANGIAN
#ifdef EULERIAN
          Q = dens*(E + KE);
#endif EULERIAN
#endif HYDRO
          
          ENDLOOP
          break;
          
        case kWTotEE: // total specific energy 
          // Eulerian: P/((g-1)) + 0.5 d v^2 [+ 0.5B^2]
          LOOP 
          
          P     = zoneW.P CELL;
          dens  = zoneW.d CELL;
          E = zlsp_Ud_Pd (P/dens, cp);
          
          KE = 0.0;
          
#ifdef HYDRO
#ifdef SYM1D
          Real vx = zoneW.vx CELL;
          KE = 0.5*(vx*vx);
#endif SYM1D
#ifdef SYM2D
          Real vx = zoneW.vx CELL;
          Real vy = zoneW.vy CELL;
          KE = 0.5*(vx*vx+vy*vy);
#endif SYM2D
#ifdef SYM3D
          Real vx = zoneW.vx CELL;
          Real vy = zoneW.vy CELL;
          Real vz = zoneW.vz CELL;
          KE = 0.5*(vx*vx+vy*vy+vz*vz);
#endif SYM3D
          Q = dens*(E + KE);
#endif HYDRO
          
          ENDLOOP
          break;          
          
        case kWTotEL: // total specific energy 
          // Lagrangian: P/((g-1) d) + 0.5 v^2 [+ 0.5B^2/d]
          LOOP 
          
          P     = zoneW.P CELL;
          dens  = zoneW.d CELL;
          E = zlsp_Ud_Pd (P/dens, cp);
          
          KE = 0.0;
          
#ifdef HYDRO
#ifdef SYM1D
          Real vx = zoneW.vx CELL;
          KE = 0.5*(vx*vx);
#endif SYM1D
#ifdef SYM2D
          Real vx = zoneW.vx CELL;
          Real vy = zoneW.vy CELL;
          KE = 0.5*(vx*vx+vy*vy);
#endif SYM2D
#ifdef SYM3D
          Real vx = zoneW.vx CELL;
          Real vy = zoneW.vy CELL;
          Real vz = zoneW.vz CELL;
          KE = 0.5*(vx*vx+vy*vy+vz*vz);
#endif SYM3D
          Q = E + KE;
#endif HYDRO
          
          ENDLOOP
          break;
          
        case kWE: // internal specific energy
          // Lagrangian  P/((g-1) d)
          // Eulerian    P/((g-1))
          LOOP   
          P    = zoneW.P CELL;
          dens  = zoneW.d CELL;
          Real Pd = P/dens;
          E = zlsp_Ud_Pd (Pd, cp);
#ifdef LAGRANGIAN
          Q = E;
#endif LAGRANGIAN
#ifdef EULERIAN
          Q = dens*E;
#endif EULERIAN
          ENDLOOP            
          break;
          
        case kWEnth: // Specific enthalpy 
          // Lagrangian: P/((g-1) d) + P/d
          // Eulerian  : P/((g-1)) + P
          LOOP   
          P    = zoneW.P CELL;
          dens  = zoneW.d CELL;
          Real Pd = P/dens;
          E = zlsp_Ud_Pd (Pd, cp);
#ifdef LAGRANGIAN
          Q = E + Pd;
#endif LAGRANGIAN
#ifdef EULERIAN
          Q = dens*E + P;
#endif EULERIAN
          ENDLOOP            
          break;
          
        case kWEnthE: // Specific enthalpy 
          // Eulerian  : P/((g-1)) + P
          LOOP   
          P    = zoneW.P CELL;
          dens  = zoneW.d CELL;
          Real Pd = P/dens;
          E = zlsp_Ud_Pd (Pd, cp);
          Q = dens*E + P;
          
          ENDLOOP            
          break;
          
        case kWEnthL: // Specific enthalpy 
          // Lagrangian: P/((g-1) d) + P/d
          LOOP   
          P    = zoneW.P CELL;
          dens  = zoneW.d CELL;
          Real Pd = P/dens;
          E = zlsp_Ud_Pd (Pd, cp);
          Q = E + Pd;
          
          ENDLOOP            
          break;
          
        case kWEntr: // Isoentropic variable P/(d^g)
          LOOP   
          P    = zoneW.P CELL;
          dens  = zoneW.d CELL;
          E = zlsp_En_Pd (P, dens, cp);
          Q = E;
          ENDLOOP            
          break;
          
        case kWTemp:
          LOOP   
          P    = zoneW.P CELL;
          dens  = zoneW.d CELL;
          T     = zlsp_T_Pd(P/dens, cp);
          Q = T;
          ENDLOOP            
          break;
        case kWCoolAt:
#ifdef COOLING
          LOOP  
          P  = zoneW.P CELL;
          dens  = zoneW.d CELL;
          
          T     = zlsp_T_Pd(P/dens, cp);
          
          zlsp_LambdaComponents(&lam, &lamAt, &lamH2, &lamCO, T, cp);
          
          Q = dens*dens*lamAt;
          ENDLOOP            
#endif COOLING
          break;
        case kWCoolCO:
#ifdef COOLING
          LOOP
          P     = zoneW.P CELL;
          dens  = zoneW.d CELL;
          T     = zlsp_T_Pd(P/dens, cp);
          zlsp_LambdaComponents(&lam, &lamAt, &lamH2, &lamCO, T, cp);
          
          Q = dens*dens*lamCO;
          ENDLOOP
#endif COOLING
          break;
        case kWCoolH2:
#ifdef COOLING
          LOOP 
          P  = zoneW.P CELL;
          dens  = zoneW.d CELL;
          T     = zlsp_T_Pd(P/dens, cp);
          zlsp_LambdaComponents(&lam, &lamAt, &lamH2, &lamCO, T, cp);
          
          Q = dens*dens*lamH2;
          ENDLOOP
#endif COOLING
          break;
        case kWCool:
#ifdef COOLING
          LOOP 
          P  = zoneW.P CELL;
          dens  = zoneW.d CELL;
          T     = zlsp_T_Pd(P/dens, cp);
          zlsp_LambdaComponents(&lam, &lamAt, &lamH2, &lamCO, T, cp);
          Q = dens*dens*lam;
          ENDLOOP            
#endif COOLING
          break;
        case kWHeat:
#ifdef COOLING
          LOOP 
          Real heat = 0.0;
          P     = zoneW.P CELL;
          dens  = zoneW.d CELL;
          T     = zlsp_T_Pd(P/dens, cp);
#ifdef H_IONISATION
          fH[1] = zlsp_XHI_T ( T, cp );  
          fH[2] = zlsp_XHII_T( T, cp );
          fH[0] = 1.0-(fH[1]+fH[2]);
#endif H_IONISATION
          Q = heat;
          ENDLOOP            
#endif COOLING
          break;
        case kWHTgr:
#ifdef COOLING
          LOOP
          Real heat = 0.0;
          P     = zoneW.P CELL;
          dens  = zoneW.d CELL;
          T     = zlsp_T_Pd(P/dens, cp);
          Q = heat;
          ENDLOOP            
#endif COOLING
          break;
        case kWEmGr:
#ifdef COOLING
          LOOP 
          Real emission = 0.0;
          P     = zoneW.P CELL;
          dens  = zoneW.d CELL;
          T     = zlsp_T_Pd(P/dens, cp);
          Q = dens*emission*1e-12; // normalised to 100K grains
          ENDLOOP            
#endif COOLING
          break;
        case kWHCRX:
          break;
        case kWHIPE:
          break;
        case kWLambAt:
#ifdef COOLING
          LOOP 
          P     = zoneW.P CELL;
          dens  = zoneW.d CELL;
          T     = zlsp_T_Pd(P/dens, cp);
          zlsp_LambdaComponents(&lam, &lamAt, &lamH2, &lamCO, T, cp);
          Q = lamAt;
          ENDLOOP            
#endif COOLING
          break;
        case kWLambCO:
#ifdef COOLING
          LOOP
          P     = zoneW.P CELL;
          dens  = zoneW.d CELL;
          T     = zlsp_T_Pd(P/dens, cp);
          zlsp_LambdaComponents(&lam, &lamAt, &lamH2, &lamCO, T, cp);
          Q = lamCO;
          ENDLOOP            
#endif COOLING
          break;
        case kWLambH2:
#ifdef COOLING
          LOOP
          P     = zoneW.P CELL;
          dens  = zoneW.d CELL;
          T     = zlsp_T_Pd(P/dens, cp);
          zlsp_LambdaComponents(&lam, &lamAt, &lamH2, &lamCO, T, cp);
          Q = lamH2;
          ENDLOOP            
#endif COOLING
          break;
        case kWLamb:
#ifdef COOLING
          LOOP
          P     = zoneW.P CELL;
          dens  = zoneW.d CELL;
          T     = zlsp_T_Pd(P/dens, cp);
          zlsp_LambdaComponents(&lam, &lamAt, &lamH2, &lamCO, T, cp);
          Q = lam;
          ENDLOOP                                    
#endif COOLING
          break;
        case kWGamma:
#ifdef VARIABLE_EOS
          LOOP
          P     = zoneW.P CELL;
          dens  = zoneW.d CELL;
          Q = zlsp_Gamma_Pd(P/dens, cp);                  
          ENDLOOP            
#endif VARIABLE_EOS
          break;
        case kWMu:
#ifdef VARIABLE_EOS
          LOOP
          P       = zoneW.P CELL;
          dens    = zoneW.d  CELL;
          Q = zlsp_Mu_Pd(P/dens, cp);
          ENDLOOP            
#endif VARIABLE_EOS
          break;
        case kWXHI:
#ifdef VARIABLE_EOS
          LOOP
          P       = zoneW.P CELL;
          dens    = zoneW.d  CELL;
          Q = zlsp_XHI_Pd(P/dens, cp);
          ENDLOOP            
#endif VARIABLE_EOS
          break;
        case kWXHII:
#ifdef VARIABLE_EOS
          LOOP
          P       = zoneW.P CELL;
          dens    = zoneW.d  CELL;
          Q = zlsp_XHII_Pd(P/dens, cp);
          ENDLOOP            
#endif VARIABLE_EOS
          break;
        case kWXH2:
#ifdef VARIABLE_EOS
          LOOP
          P       = zoneW.P CELL;
          dens    = zoneW.d  CELL;
          Q = zlsp_XH2_Pd(P/dens, cp);
          ENDLOOP            
#endif VARIABLE_EOS
          break;
        case kWdHI:
#ifdef VARIABLE_EOS
          LOOP
          P       = zoneW.P CELL;
          dens    = zoneW.d  CELL;
          Q = dens*zlsp_XHI_Pd(P/dens, cp);
          ENDLOOP            
#endif VARIABLE_EOS
          break;
        case kWdHII:
#ifdef VARIABLE_EOS
          LOOP
          P       = zoneW.P CELL;
          dens    = zoneW.d  CELL;
          Q = dens*zlsp_XHII_Pd(P/dens, cp);
          ENDLOOP            
#endif VARIABLE_EOS
          break;
        case kWdH2:
#ifdef VARIABLE_EOS
          LOOP
          P      = zoneW.P CELL;
          dens   = zoneW.d  CELL;
          Q = dens*zlsp_XH2_Pd(P/dens, cp);
          ENDLOOP            
#endif VARIABLE_EOS
          break;
        case kWnHI:
#ifdef VARIABLE_EOS
          LOOP
          P       = zoneW.P CELL;
          dens    = zoneW.d  CELL;
          Q       = dens*f_nHtot*zlsp_XHI_Pd(P/dens, cp);
          ENDLOOP            
#endif VARIABLE_EOS
          break;
        case kWnHII:
#ifdef VARIABLE_EOS
          LOOP
          P       = zoneW.P CELL;
          dens    = zoneW.d  CELL;
          Q = dens*f_nHtot*zlsp_XHII_Pd(P/dens, cp);
          ENDLOOP            
#endif VARIABLE_EOS
          break;
        case kWnH2:
#ifdef VARIABLE_EOS
          LOOP
          P       = zoneW.P CELL;
          dens      = zoneW.d  CELL;
          Q = 0.5*dens*f_nHtot*zlsp_XH2_Pd(P/dens, cp);
          ENDLOOP            
#endif VARIABLE_EOS
          break;
        case kWVelm:
          LOOP
#ifdef SYM1D
          Real vx = zoneW.vx CELL;
          Q = zls_sqrt(vx*vx);
#endif SYM1D
#ifdef SYM2D
          Real vx = zoneW.vx CELL;
          Real vy = zoneW.vy CELL;
          Q = zls_sqrt(vx*vx+vy*vy);
#endif SYM2D
#ifdef SYM3D
          Real vx = zoneW.vx CELL;
          Real vy = zoneW.vy CELL;
          Real vz = zoneW.vz CELL;
          Q = zls_sqrt(vx*vx+vy*vy+vz*vz);
#endif SYM3D
          ENDLOOP            
          break;
        case kRecomb:
#ifdef VARIABLE_EOS
          LOOP
          P       = zoneW.P CELL;
          dens    = zoneW.d  CELL;
          T       = zlsp_T_Pd( P/dens, cp)*1.0e-4;
          Real f  =  dens*f_nHtot*zlsp_XHII_Pd(P/dens, cp);
          Q       = f*f*zls_Pow(T,(-0.77)); 
          ENDLOOP            
#endif VARIABLE_EOS
          break;
        case kRayleigh:
#ifdef VARIABLE_EOS
          LOOP
          P       = zoneW.P CELL;
          dens    = zoneW.d  CELL;
          Real dx     =  gc.x[0].dx*f_x0; // square cells only!
          T       = zlsp_T_Pd( P/dens, cp)*1.0e-4;
          Real f  = dens*f_nHtot*zlsp_XHII_Pd(P/dens, cp);
          Q       = dx*4.73917E-20*f*f*zls_Pow(T,(-0.77)); 
          ENDLOOP            
#endif VARIABLE_EOS
          break;
        case kWVelmx:
          LOOP            
          Real vx = zoneW.vx CELL;
          Q = zls_sqrt(vx*vx);
          ENDLOOP            
          break;
#if (NDIMS > 1)
        case kWVelmxy:
          LOOP
          Real vx = zoneW.vx CELL;
          Real vy = zoneW.vy CELL;
          Q = zls_sqrt(vx*vx+vy*vy);
          ENDLOOP            
          break;
        case kWVelmy:
          LOOP            
          Real vy = zoneW.vy CELL;
          Q = zls_sqrt(vy*vy);
          ENDLOOP            
          break;
#endif (NDIMS > 1)
#if (NDIMS > 2)
        case kWVelmyz:
          LOOP
          Real vz = zoneW.vz CELL;
          Real vy = zoneW.vy CELL;
          Q = zls_sqrt(vz*vz+vy*vy);
          ENDLOOP            
          break;
        case kWVelmzx:
          LOOP
          Real vz = zoneW.vz CELL;
          Real vx = zoneW.vx CELL;
          Q = zls_sqrt(vx*vx+vz*vz);
          ENDLOOP            
          break;
        case kWVelmz:
          LOOP            
          Real vz = zoneW.vz CELL;
          Q = zls_sqrt(vz*vz);
          ENDLOOP            
          break;
#endif (NDIMS > 2)
        case kWCs:
          LOOP
          P     = zoneW.P CELL;
          dens  = zoneW.d CELL;
          Q  =zlsp_Cs_Pd(P/dens, cp);
          ENDLOOP            
          break;
        case kWAs:
          LOOP
          P     = zoneW.P CELL;
          dens  = zoneW.d CELL;
          Q  =zls_sqrt(P/dens);
          ENDLOOP            
          break;
        case kWMach:
          LOOP
          P       = zoneW.P CELL;
          dens    = zoneW.d CELL;
          Real cs = zlsp_Cs_Pd(P/dens, cp);
#ifdef SYM1D
          Real vx = zoneW.vx CELL;
          Q = zls_sqrt(vx*vx)/cs;
#endif SYM1D
#ifdef SYM2D
          Real vx = zoneW.vx CELL;
          Real vy = zoneW.vy CELL;
          Q = zls_sqrt(vx*vx+vy*vy)/cs;
#endif SYM2D
#ifdef SYM3D
          Real vx = zoneW.vx CELL;
          Real vy = zoneW.vy CELL;
          Real vz = zoneW.vz CELL;
          Q = zls_sqrt(vx*vx+vy*vy+vz*vz)/cs;
#endif SYM3D
          ENDLOOP            
          break;
          
#ifdef LAGRANGIAN
#ifdef SPLITSWEEPS
#ifdef DIVVHCORRECTION   
          
          
        case kWDivV:
          LOOP
#ifdef SYM1D
          Real px = z->zoneDivV CELL.a;
          Q = (px>kZLS_DivVThreshold)?px:0.0;
#endif SYM1D
#ifdef SYM2D
          Real px = z->zoneDivV CELL.a[kXDim];
          Real py = z->zoneDivV CELL.a[kYDim];
          Real dv = zls_sqrt(px*px+py*py);
          Q = (dv>kZLS_DivVThreshold)?dv:0.0;
#endif SYM2D
#ifdef SYM3D
          Real px = z->zoneDivV CELL.a[kXDim];
          Real py = z->zoneDivV CELL.a[kYDim];
          Real pz = z->zoneDivV CELL.a[kZDim];
          Real dv = zls_sqrt(px*px+py*py+pz*pz);
          Q = (dv>kZLS_DivVThreshold)?dv:0.0;
#endif SYM3D
          ENDLOOP            
          break;
          
        case kWDvDx:
          LOOP
#ifdef SYM1D
          Real px = z->zoneDivV CELL.a;
          Q = (px>kZLS_DivVThreshold)?px:0.0;
#endif SYM1D
#ifdef SYM2D
          Real px = z->zoneDivV CELL.a[kXDim];
          Real dv = zls_abs(px);
          Q = (dv>kZLS_DivVThreshold)?dv:0.0;
#endif SYM2D
#ifdef SYM3D
          Real px = z->zoneDivV CELL.a[kXDim];
          Real dv = zls_abs(px);
          Q = (dv>kZLS_DivVThreshold)?dv:0.0;
#endif SYM3D
          ENDLOOP            
          break;
          
        case kWDvDy:
          LOOP
#ifdef SYM2D
          Real py = z->zoneDivV CELL.a[kYDim];
          Real dv = zls_abs(py);
          Q = (dv>kZLS_DivVThreshold)?dv:0.0;
#endif SYM2D
#ifdef SYM3D
          Real py = z->zoneDivV CELL.a[kYDim];
          Real dv = zls_abs(py*py);
          Q = (dv>kZLS_DivVThreshold)?dv:0.0;
#endif SYM3D
          ENDLOOP            
          break;
          
        case kWDvDz:
#ifdef SYM3D
          LOOP
          Real pz = z->zoneDivV CELL.a[kZDim];
          Real dv = zls_abs(pz*pz);
          Q = (dv>kZLS_DivVThreshold)?dv:0.0;
          ENDLOOP            
#endif SYM3D
          break;
          
#endif DIVVHCORRECTION
#endif LAGRANGIAN
#endif SPLITSWEEPS
#ifdef GRAVITY              
        case kGPhi:
          LOOP
          Q = zoneF.p CELLA0;
          ENDLOOP            
          break;
        case kdPhi:
          LOOP
#ifdef SYM1D
          Real px = (zoneF.p CELL.ar-zoneF.p CELL.al)*gc.x[ic].idx;
          Q = zls_sqrt(px*px);
#endif SYM1D
#ifdef SYM2D
          Real px = (zoneF.p CELL.ar[kXDim]-zoneF.p CELL.al[kXDim])*gc.x[ic].idx;
          Real py = (zoneF.p CELL.ar[kYDim]-zoneF.p CELL.al[kYDim])*gc.y[jc].idx;
          Q = zls_sqrt(px*px+py*py);
#endif SYM2D
#ifdef SYM3D
          Real px = (zoneF.p CELL.ar[kXDim]-zoneF.p CELL.al[kXDim])*gc.x[ic].idx;
          Real py = (zoneF.p CELL.ar[kYDim]-zoneF.p CELL.al[kYDim])*gc.y[jc].idx;
          Real pz = (zoneF.p CELL.ar[kZDim]-zoneF.p CELL.al[kZDim])*gc.z[kc].idx;
          Q = zls_sqrt(px*px+py*py+pz*pz);
#endif SYM3D
          ENDLOOP            
          break;
        case kdPhiX:
          LOOP
#ifdef SYM1D
          Q  = (zoneF.p CELL.ar-zoneF.p CELL.al)*gc.x[ic].idx;
#else
          Q  = (zoneF.p CELL.ar[kXDim]-zoneF.p CELL.al[kXDim])*gc.x[ic].idx;
#endif SYM1D
          ENDLOOP            
          break;
#if (NDIMS > 1)
        case kdPhiY:
          LOOP
          Q  = (zoneF.p CELL.ar[kYDim] - zoneF.p CELL.al[kYDim])*gc.y[jc].idx;
          ENDLOOP            
          break;
#endif (NDIMS > 1)
#if (NDIMS > 2)
        case kdPhiZ:
          LOOP
          Q  = (zoneF.p CELL.ar[kZDim] - zoneF.p CELL.al[kZDim])*gc.z[kc].idx;
          ENDLOOP            
          break;
#endif (NDIMS > 2)
#endif GRAVITY              
#if (NTRACERS>0)              
        case kScalar0:
        case kScalar1:
        case kScalar2:
        case kScalar3:
        case kScalar4:
        case kScalar5:
        case kScalar6:
        case kScalar7:
          LOOP
          tr     = zoneW.tr[varIndex-kScalar0] CELL;
          dens  = zoneW.d CELL;
          Q = dens>0.0?(tr/dens):0.0;
          ENDLOOP            
          break;
        case kRadioJnu0:
        case kRadioJnu1:
        case kRadioJnu2:
          LOOP
          P     = zoneW.P CELL;
          tr    = zoneW.tr[varIndex-kRadioJnu0] CELL;
          dens  = zoneW.d CELL;
          Q = dens>0.0?(tr/dens)*zls_Pow(P, 1.8):0.0;
          ENDLOOP            
          break;
#endif (NTRACERS>0)              
        default:
          LOOP
          Q  = 0.0;
          ENDLOOP            
      } // switch derived vars
      
      
    } else { err = zonErr;} //z
    
  } else {err = memErr;} // q3d
  
  return(err);
}

#undef LOOP      
#undef ENDLOOP   
#undef CELL
#undef CELLA0
#undef CELLX0
#undef CELLX1
#undef CELLY0
#undef CELLY1
#undef CELLZ0
#undef CELLZ1
#undef Q

int   zlsp_GetZoneWQuantityCopy( QPtr * d, ZonePtr z, Integer varIndex ){
  
  Integer err = noErr;
  
  if (z != NULL) {
    
    
    if ((varIndex >= 0)&&(varIndex < kNAllVars)){
      err = zls_GetZoneDataCopy( d, z, varIndex );
    } else  if ( varIndex == kWDeltaVx ){
      err = zls_GetZoneDataCopy( d, z, kWVx );
    } else  if ( varIndex == kWDeltaVy ){
      err = zls_GetZoneDataCopy( d, z, kWVy );
    } else  if ( varIndex == kWDeltaVz ){
      err = zls_GetZoneDataCopy( d, z, kWVz );
    } else {
      
      
    	GridCoords gc = z->zoneGrid;
      QPtr     data = NULL;
      
#ifdef SYM1D
#ifdef FILES_NO_BCS    
      Integer ni =  gc.n[kXDim];
#endif FILES_NO_BCS    
#ifdef FILES_HAVE_BCS    
      Integer ni =  gc.nt[kXDim];
#endif FILES_HAVE_BCS    
      err = zls_NewQuantity1D(&data, ni);
      if (data !=NULL){
        QPtr  q1d = data;
        err = zlsp_GetZoneWQArr( q1d, z, varIndex);
      } else { err = memErr; }
#endif SYM1D
      
#ifdef SYM2D
#ifdef FILES_NO_BCS    
      Integer ni =  gc.n[kXDim];
      Integer nj =  gc.n[kYDim];
#endif FILES_NO_BCS    
#ifdef FILES_HAVE_BCS    
      Integer ni =  gc.nt[kXDim];
      Integer nj =  gc.nt[kYDim];
#endif FILES_HAVE_BCS    
      Q2DArr q2d = NULL;
      err = zls_NewQuantity2D(&data, ni, nj);
      if ( data != NULL ) {
        zls_SetQuantity2DArr(&q2d, data,  nj, ni);
        err = zlsp_GetZoneWQArr( q2d, z, varIndex);
        zls_ReleaseQuantity2DArr(q2d);
      } else { err = memErr ;}
#endif SYM2D
      
#ifdef SYM3D
#ifdef FILES_NO_BCS    
      Integer ni =  gc.n[kXDim];
      Integer nj =  gc.n[kYDim];
      Integer nk =  gc.n[kZDim];
#endif FILES_NO_BCS    
#ifdef FILES_HAVE_BCS    
      Integer ni =  gc.nt[kXDim];
      Integer nj =  gc.nt[kYDim];
      Integer nk =  gc.nt[kZDim];
#endif FILES_HAVE_BCS    
      
      Q3DArr q3d = NULL;
      err = zls_NewQuantity3D(&data, ni, nj, nk);
      
      if ( data != NULL ) {
        
        zls_SetQuantity3DArr(&q3d, data, nk,  nj, ni);
        err = zlsp_GetZoneWQArr( q3d, z, varIndex);
        zls_ReleaseQuantity3DArr(q3d);
        
      } else { err = memErr ;}
      
#endif SYM3D
      *d = data;
      
    } // ! (varIndex >= 0)&&(varIndex < kNAllVars)
    
  } else { err = zonErr;} //z
  
  
  return(err);
  
}

int   zlsp_GetZoneWQuantityCopy2DSlice ( QPtr * d, ZonePtr z, Integer varIndex , Integer slice, Integer sliceIndex){ 
  
  Integer err = noErr;
#if (NDIMS < 2)
	zlsio_ErrPrintf( " WARNING: zls_GetZoneData2DSlice 2D slice not possible in < 2D zone: ndims =  %d \n", NDIMS);
	err = dimErr;
	return(err);
#endif 
#ifdef SYM2D
#pragma unused(slice, sliceIndex)
  //  err = zlsp_GetZoneWQuantityCopy ( d,  z,  varIndex );
  
  
  
  if (z != NULL) {
    
    
    QPtr  data2dd = NULL; // dest
    QPtr  data2ds = NULL; // source
    
    Q2DArr   q2dd = NULL; // 2D dest
    Q2DArr   q2ds = NULL; // 2D source
    
    Integer i, j;
    
	  GridCoords gc = z->zoneGrid;
    
    Counter nx, ny;
    
    err = zls_GetGridImageDims( &nx, &ny, &z->zoneGrid, slice );
    err = zls_NewQuantity2D( &data2dd, nx, ny );
    
    if (( err == noErr) && ( data2dd != NULL )) {
      
      //
      // in 3D get the data3d through the normal zlsp_GetZoneWQuantityCopy, 
      // which will be 3D, and then simply slice or sum
      //
      
      err = zlsp_GetZoneWQuantityCopy( &data2ds, z,  varIndex );
      
      //
      // set up loop limits
      //
      
#ifdef FILES_NO_BCS    
      
      // since we have used zlsp_GetZoneWQuantityCopy and are just 
      // summing now, then if the BCs are not being saved then 
      // they will already be removed by zlsp_GetZoneWQuantityCopy
      // in that case the loop will go from 0 -> n-1 , is = 0 etc
      
      Integer ni  = gc.n[kXDim];
      Integer nj  = gc.n[kYDim];
      
#endif FILES_NO_BCS
      
#ifdef FILES_HAVE_BCS          
      // in this case the loop will go from 0 -> nt-1 , is = 0 etc
      
      Integer ni  = gc.nt[kXDim];
      Integer nj  = gc.nt[kYDim];
      
#endif FILES_HAVE_BCS
      
      
      if (( err == noErr ) && ( data2dd != NULL )) {
        
        err = zls_SetQuantity2DArr( &q2ds, data2ds, nj,  ni );
        err = zls_SetQuantity2DArr( &q2dd, data2dd, ny,  nx );
        
				switch (slice) {
          default:
					case kSliceXYSum:
					case kSliceXY:
						for (j = 0; j<nj; j++){
							for (i = 0; i<ni; i++){
								q2dd[j][i] = q2ds[j][i];
							} //i
						}; //j
						break;
            case kSliceYXSum:
            case kSliceYX:
						for (j = 0; j<nj; j++){
							for (i = 0; i<ni; i++){
								q2dd[i][j] = q2ds[j][i];
							} //i
						}; //j
						break;
				} // switch
        
        zls_ReleaseQuantity2DArr(q2ds);
        zls_ReleaseQuantity2DArr(q2dd);
        zls_DisposeQuantity ( data2ds ); // it can go now.
        
      } else { err = memErr; }
      
    } else { err = memErr; }
    
    *d = data2dd;
    
  } else { err = zonErr;} //z
  
  
  return ( err );
#endif SYM2D
  
#ifdef SYM3D
  
  if (z != NULL) {
    
    
    QPtr  data2d = NULL;
    QPtr  data3d = NULL;
    Q2DArr   q2d = NULL;
    Q3DArr   q3d = NULL;
    
    Integer i, j, k;
    
    GridCoords gc = z->zoneGrid;
    
    Counter nx, ny;
    
    err = zls_GetGridImageDims( &nx, &ny, &z->zoneGrid, slice );
    err = zls_NewQuantity2D( &data2d, nx, ny );
    
    if (( err == noErr) && ( data2d != NULL )) {
      
      //
      // in 3D get the data3d through the normal zlsp_GetZoneWQuantityCopy, 
      // which will be 3D, and then simply slice or sum
      //
      
      err = zlsp_GetZoneWQuantityCopy( &data3d, z,  varIndex );
      
      
      //
      // set up loop limits
      //
      
#ifdef FILES_NO_BCS    
      
      // since we have used zlsp_GetZoneWQuantityCopy and are just 
      // summing now, then if the BCs are not being saved then 
      // they will already be removed by zlsp_GetZoneWQuantityCopy
      // in that case the loop will go from 0 -> n-1 , is = 0 etc
      
      Integer ni  = gc.n[kXDim];
      Integer nj  = gc.n[kYDim];
      Integer nk  = gc.n[kZDim];
      
#endif FILES_NO_BCS
      
#ifdef FILES_HAVE_BCS          
      // in this case the loop will go from 0 -> nt-1 , is = 0 etc
      
      Integer ni  = gc.nt[kXDim];
      Integer nj  = gc.nt[kYDim];
      Integer nk  = gc.nt[kZDim];
      
#endif FILES_HAVE_BCS
      
      
      if (( err == noErr ) && ( data3d != NULL )) {
        
        err = zls_SetQuantity3DArr( &q3d, data3d, nk,  nj,  ni );
        err = zls_SetQuantity2DArr( &q2d, data2d, ny,  nx );
        
				switch (slice) {
            default:
            case kSliceXY:
            if ( sliceIndex < 0 ) sliceIndex = nk/2;
            k = sliceIndex;
						for (j = 0; j<nj; j++){
							for (i = 0; i<ni; i++){
								q2d[j][i] = q3d[k][j][i];
							} //i
						}; //j
						break;
            case kSliceYX:
            if ( sliceIndex < 0 ) sliceIndex = nk/2;
            k = sliceIndex;
						for (j = 0; j<nj; j++){
							for (i = 0; i<ni; i++){
								q2d[i][j] = q3d[k][j][i];
							} //i
						}; //j
						break;
					  case kSliceYZ:
            if ( sliceIndex < 0 ) sliceIndex = ni/2;
						i = sliceIndex;
						for (k = 0; k<nk; k++){
							for (j = 0; j<nj; j++){
								q2d[k][j] = q3d[k][j][i];
							} //j
						}; //k
						break;
            case kSliceZY:
            if ( sliceIndex < 0 ) sliceIndex = ni/2;
						i = sliceIndex;
						for (k = 0; k<nk; k++){
							for (j = 0; j<nj; j++){
								q2d[j][k] = q3d[k][j][i];
							} //j
						}; //k
						break;
            case kSliceZX:
            if ( sliceIndex < 0 ) sliceIndex = nj/2;
						j = sliceIndex;
						for (i = 0; i<ni; i++){
							for (k = 0; k<nk; k++){
								q2d[i][k] = q3d[k][j][i];
							} //k
						}; //i
						break;
            case kSliceXZ:
            if ( sliceIndex < 0 ) sliceIndex = nj/2;
						j = sliceIndex;
						for (i = 0; i<ni; i++){
							for (k = 0; k<nk; k++){
								q2d[k][i] = q3d[k][j][i];
							} //k
						}; //i
						break;
            case kSliceXYSum:
						k = 0;
						for (j = 0; j<nj; j++){
							for (i = 0; i<ni; i++){
								q2d[j][i] = q3d[k][j][i];
							} //i
						}; //j
						for (j = 0; j<nj; j++){
							for (i = 0; i<ni; i++){
                for (k = 1; k<nk; k++){
                  q2d[j][i] += q3d[k][j][i];
                } //k
							} //i
						}; //j
						break;
            case kSliceYXSum:
						k = 0;
						for (j = 0; j<nj; j++){
							for (i = 0; i<ni; i++){
								q2d[i][j] = q3d[k][j][i];
							} //i
						}; //j
						for (j = 0; j<nj; j++){
							for (i = 0; i<ni; i++){
                for (k = 1; k<nk; k++){
                  q2d[i][j] += q3d[k][j][i];
                } //k
							} //i
						}; //j
						break;
            case kSliceXZSum:
						j = 0;
						for (i = 0; i<ni; i++){
							for (k = 0; k<nk; k++){
								q2d[k][i] = q3d[k][j][i];
							} //k
						}; //i
						for (i = 0; i<ni; i++){
							for (k = 0; k<nk; k++){
                for (j = 1; j<nj; j++){
                  q2d[k][i] += q3d[k][j][i];
                } //j
							} //k
						}; //i
						break;
            case kSliceYZSum:
						i = 0;
						for (k = 0; k<nk; k++){
							for (j = 0; j<nj; j++){
								q2d[k][j] = q3d[k][j][i];
							} //j
						}; //k
						for (k = 0; k<nk; k++){
							for (j = 0; j<nj; j++){
                for (i = 1; i<ni; i++){
                  q2d[k][j] += q3d[k][j][i];
                } //j
							} //j
						}; //k
 						break;
            case kSliceZYSum:
						i = 0;
						for (k = 0; k<nk; k++){
							for (j = 0; j<nj; j++){
								q2d[j][k] = q3d[k][j][i];
							} //j
						}; //k
						for (k = 0; k<nk; k++){
							for (j = 0; j<nj; j++){
                for (i = 1; i<ni; i++){
                  q2d[j][k] += q3d[k][j][i];
                } //j
							} //j
						}; //k
						break;
            case kSliceZXSum:
						j = 0;
						for (i = 0; i<ni; i++){
							for (k = 0; k<nk; k++){
								q2d[i][k] = q3d[k][j][i];
							} //k
						}; //i
						for (i = 0; i<ni; i++){
							for (k = 0; k<nk; k++){
                for (j = 1; j<nj; j++){
                  q2d[i][k] += q3d[k][j][i];
                } //j
							} //k
						}; //i
            
				} // switch
        
        zls_ReleaseQuantity2DArr(q2d);
        zls_ReleaseQuantity3DArr(q3d);
        
        zls_DisposeQuantity ( data3d ); // it can go now.
        
      } else { err = memErr; }
      
    } else { err = memErr; }
    
    *d = data2d;
    
  } else { err = zonErr;} //z
  
  return(err);  
  
#endif SYM3D
}


//========================================

int zlsp_CalcEnergyW1D ( W1DArr w ){
  int err = noErr;
  Integer i;
  for ( i = 0; i<w.nt; i++){
    w.E[i].a = zlsp_EFromP_W1D(w, i);
  }
  return(err);
}

int zlsp_CalcPressureW1D ( W1DArr w ){
  int err = noErr;
  Integer i;
  for ( i = 0; i<w.nt; i++){
    w.P[i].a = zlsp_PFromE_W1D(w, i);
  }
  return(err);
}


int zlsp_CalcEnergyU1D ( U1DArr u ){
  int err = noErr;
  Integer i;
  for ( i = 0; i<u.nt; i++){
    u.E[i].a = zlsp_EFromP_U1D(u, i);
  }
  return(err);
}

int zlsp_CalcPressureU1D ( U1DArr u ){
  int err = noErr;
  Integer i;
  for ( i = 0; i<u.nt; i++){
    u.P[i].a = zlsp_PFromE_U1D(u, i);
  }
  return(err);
}

int zlsp_CalcCellLREnergyW1D(W1DArr w){
#ifdef GODUNOV
	return(noErr);
#else
  int err = noErr;
  
  Integer i;
  for ( i = 0; i<w.nt; i++){
    
#ifdef SYM1D
    Real vx = w.vx[i].al;
    Real kE = 0.5*(vx * vx);
#endif
#ifdef SYM2D
    Real vx = w.vx[i].al;
    Real vy = w.vy[i].al;
    Real kE = 0.5*((vx * vx ) + (vy *vy ));
#endif
#ifdef SYM3D
    Real vx = w.vx[i].al;
    Real vy = w.vy[i].al;
    Real vz = w.vz[i].al;
    Real kE = 0.5*((vx * vx ) + (vy *vy ) + (vz *vz ));
#endif
    Real P = w.P[i].al;
    Real d = w.d[i].al;
    w.E[i].al = zlsp_EFromP( P, kE, d);
    
    
#ifdef SYM1D
    vx = w.vx[i].ar;
    kE = 0.5*(vx * vx);
#endif
#ifdef SYM2D
    vx = w.vx[i].ar;
    vy = w.vy[i].ar;
    kE = 0.5*((vx * vx ) + (vy *vy ));
#endif
#ifdef SYM3D
    vx = w.vx[i].ar;
    vy = w.vy[i].ar;
    vz = w.vz[i].ar;
    kE = 0.5*((vx * vx ) + (vy *vy ) + (vz *vz ));
#endif
    P = w.P[i].ar;
    d = w.d[i].ar;
    w.E[i].ar = zlsp_EFromP( P, kE, d);
    
  }
	
  V1DArr v;
  v.n   = w.n;
  v.ns  = w.ns;
  v.ne  = w.ne;
  v.nt  = w.nt;
  v.v = w.E;
  err = zlsppm_CalcDelA6Var1DArr  ( v );
	
  return(err);
#endif !GODUNOV
}


int zlsp_CalcCellLREnergyU1D(U1DArr u){
#ifdef GODUNOV
	return(noErr);
#else
  int err = noErr;
  
  Integer i;
  for ( i = 0; i<u.nt; i++){
    
    Real d = u.d[i].al;
#ifdef SYM1D
    Real mx = u.mx[i].al;
    Real kE = 0.5*(mx * mx)/d;
#endif
#ifdef SYM2D
    Real mx = u.mx[i].al;
    Real my = u.my[i].al;
    Real kE = 0.5*((mx * mx ) + (my *my ))/d;
#endif
#ifdef SYM3D
    Real mx = u.mx[i].al;
    Real my = u.my[i].al;
    Real mz = u.mz[i].al;
    Real kE = 0.5*((mx * mx ) + (my *my ) + (mz *mz ))/d;
#endif
    Real P = u.P[i].al;
    u.E[i].al = zlsp_EFromP( P, kE, d);
    
    
    d = u.d[i].ar;
#ifdef SYM1D
    mx = u.mx[i].al;
    kE = 0.5*(mx * mx)/d;
#endif
#ifdef SYM2D
    mx = u.mx[i].ar;
    my = u.my[i].ar;
    kE = 0.5*((mx * mx ) + (my *my ))/d;
#endif
#ifdef SYM3D
    mx = u.mx[i].ar;
    my = u.my[i].ar;
    mz = u.mz[i].ar;
    kE = 0.5*((mx * mx ) + (my *my ) + (mz *mz ))/d;
#endif
    P = u.P[i].ar;
    u.E[i].ar = zlsp_EFromP( P, kE, d);
    
  }
	
  V1DArr v;
  v.n   = u.n;
  v.ns  = u.ns;
  v.ne  = u.ne;
  v.nt  = u.nt;
  
  zls_SetVar1DArrtLRQ1DArr(v, u.E);  
  
  // v.v   = u.E;
  
  err = zlsppm_CalcDelA6Var1DArr  ( v );
	
  return(err);
#endif !GODUNOV
}

//========================================

int zlsp_CalcWSoundSpeed1D(W1DArr w){
	
  int err = noErr;
  Real g  = kZLS_eos_gamma;
#ifdef VARIABLE_GAMMA
  PhysicsPtr phys = gLevels[0].levelPhysics; 
  CoolingPtr cp   = NULL;
  if ( phys != NULL) cp   = &(phys->coolParameters);
#endif VARIABLE_GAMMA
  
  Integer i;
  for ( i = 0; i<w.nt; i++){
    Real P = w.P[i].a;
    Real d = w.d[i].a;
    
#ifdef VARIABLE_EOS
    
    
#ifdef VARIABLE_GAMMA  
    g = zlsp_Gamma_Pd(P/d, cp);
    w.g [i].a = g;
#endif VARIABLE_GAMMA
    
#endif VARIABLE_EOS
    
    w.cs[i].a = zls_sqrt(g*P/(d+kZLS_Epsilon));
    
  }
  
  return(err);
  
}

int zlsp_CalcWSoundSpeed2D(W2DArr w){
	
  int err = noErr;
  Real g  = kZLS_eos_gamma;
  
#ifdef VARIABLE_GAMMA
  PhysicsPtr phys = gLevels[0].levelPhysics; 
  CoolingPtr cp   = NULL;
  if ( phys != NULL) cp   = &(phys->coolParameters);
#endif VARIABLE_GAMMA
  
  Integer i,j;
  for ( j = 0; j<w.nt[kYDim]; j++){
    for ( i = 0; i<w.nt[kZDim]; i++){
      Real P = w.P[j][i].a[kXDim];
      Real d = w.d[j][i].a[kXDim];
      
#ifdef VARIABLE_GAMMA  
      g = zlsp_Gamma_Pd(P/d, cp);
      w.g [j][i].a = g;
      w.g [j][i].a = g;
#endif VARIABLE_GAMMA
      
      w.cs[j][i].a = zls_sqrt(g*P/(d+kZLS_Epsilon));
      
    }
  }
  return(err);
}

int zlsp_CalcWSoundSpeed3D(W3DArr w){
	
  int err = noErr;
  Real g  = kZLS_eos_gamma;

#ifdef VARIABLE_GAMMA
  PhysicsPtr phys = gLevels[0].levelPhysics; 
  CoolingPtr cp   = NULL;
  if ( phys != NULL) cp   = &(phys->coolParameters);
#endif VARIABLE_GAMMA
  
  Integer i,j,k;
  for ( k = 0; k<w.nt[kXDim]; k++){
    for ( j = 0; j<w.nt[kYDim]; j++){
      for ( i = 0; i<w.nt[kZDim]; i++){
        Real P = w.P[k][j][i].a[k0Dim];
        Real d = w.d[k][j][i].a[k0Dim];
        
#ifdef VARIABLE_GAMMA  
        g = zlsp_Gamma_Pd(P/d, cp);
        w.g [k][j][i].a = g;
#endif VARIABLE_GAMMA
        
        w.cs[k][j][i].a = zls_sqrt(g*P/(d+kZLS_Epsilon));
        
      }
    }
  }
  return(err);
}

//========================================

#ifndef GODUNOV
// there are no left/right value in Godunov 1st order code 

int zlsp_CalcCellWLRSoundSpeed1D(W1DArr w){
  // this is derived from P and d, not via PPM coeffs
  int err = noErr;
  Real g  = kZLS_eos_gamma;
  
#ifdef VARIABLE_EOS
#ifdef VARIABLE_GAMMA
  PhysicsPtr phys = gLevels[0].levelPhysics; 
  CoolingPtr cp   = NULL;
  if ( phys != NULL) cp   = &(phys->coolParameters);
#endif VARIABLE_GAMMA
#endif VARIABLE_EOS
  
  Integer i;
  
  for ( i = 0; i<w.nt; i++){
    
    Real P = w.P[i].al;
    Real d = w.d[i].al;
    
#ifdef VARIABLE_EOS
        
#ifdef VARIABLE_GAMMA  
    g = zlsp_Gamma_Pd(P/d, cp);
    w.g [i].al = g;
#endif VARIABLE_GAMMA
    
#endif VARIABLE_EOS
    
    w.cs[i].al = zls_sqrt(g*P/(d+kZLS_Epsilon));
    
    
    P = w.P[i].ar;
    d = w.d[i].ar;
#ifdef VARIABLE_EOS
        
#ifdef VARIABLE_GAMMA  
    g = zlsp_Gamma_Pd(P/d, cp);
    w.g [i].ar = g;
#endif VARIABLE_GAMMA
    
#endif VARIABLE_EOS
    
    w.cs[i].ar = zls_sqrt(g*P/(d+kZLS_Epsilon));
    
  }
  return(err);
}

int zlsp_CalcCellWLRSoundSpeed2D(W2DArr w){
	
  int err = noErr;
  Real g  = kZLS_eos_gamma;

#ifdef VARIABLE_GAMMA
  PhysicsPtr phys = gLevels[0].levelPhysics; 
  CoolingPtr cp   = NULL;
  if ( phys != NULL) cp   = &(phys->coolParameters);
#endif VARIABLE_GAMMA
  Integer i,j;
  for ( j = 0; j<w.nt[kYDim]; j++){
    for ( i = 0; i<w.nt[kXDim]; i++){
      
      Real P = w.P[j][i].al[kXDim];
      Real d = w.d[j][i].al[kXDim];
      
#ifdef VARIABLE_GAMMA  
      g = zlsp_Gamma_Pd(P/d, cp);
      w.g [j][i].al[kXDim] = g;
#endif VARIABLE_GAMMA
      
      w.cs[j][i].al[kXDim] = zls_sqrt(g*P/(d+kZLS_Epsilon));
      
      P = w.P[j][i].al[kYDim];
      d = w.d[j][i].al[kYDim];
      
#ifdef VARIABLE_GAMMA  
      g = zlsp_Gamma_Pd(P/d, cp);
      w.g [j][i].al[kYDim] = g;
#endif VARIABLE_GAMMA
      
      w.cs[j][i].al[kYDim] = zls_sqrt(g*P/(d+kZLS_Epsilon));
      
      P = w.P[j][i].ar[kXDim];
      d = w.d[j][i].ar[kXDim];
      
#ifdef VARIABLE_GAMMA  
      g = zlsp_Gamma_Pd(P/d, cp);
      w.g [j][i].ar[kXDim] = g;
#endif VARIABLE_GAMMA
      
      P = w.P[j][i].ar[kYDim];
      d = w.d[j][i].ar[kYDim];
      
#ifdef VARIABLE_GAMMA  
      g = zlsp_Gamma_Pd(P/d, cp);
      w.g [j][i].ar[kYDim] = g;
#endif VARIABLE_GAMMA
      
      w.cs[j][i].ar[kYDim] = zls_sqrt(g*P/(d+kZLS_Epsilon));
      
    }
  }
  return(err);
}

int zlsp_CalcCellWLRSoundSpeed3D(W3DArr w){
	
  int err = noErr;
  Real g  = kZLS_eos_gamma;

#ifdef VARIABLE_GAMMA
  PhysicsPtr phys = gLevels[0].levelPhysics; 
  CoolingPtr cp   = NULL;
  if ( phys != NULL) cp   = &(phys->coolParameters);
#endif VARIABLE_GAMMA
  Integer i,j,k;
  for ( k = 0; k<w.nt[kZDim]; k++){
    for ( j = 0; j<w.nt[kYDim]; j++){
      for ( i = 0; i<w.nt[kXDim]; i++){
        
        Real P = w.P[k][j][i].al[kXDim];
        Real d = w.d[k][j][i].al[kXDim];
        
#ifdef VARIABLE_GAMMA  
        g = zlsp_Gamma_Pd(P/d, cp);
        w.g [k][j][i].al[kXDim] = g;
#endif VARIABLE_GAMMA
        
        w.cs[k][j][i].al[kXDim] = zls_sqrt(g*P/(d+kZLS_Epsilon));
        
        P = w.P[k][j][i].al[kYDim];
        d = w.d[k][j][i].al[kYDim];
        
#ifdef VARIABLE_GAMMA  
        g = zlsp_Gamma_Pd(P/d, cp);
        w.g [k][j][i].al[kYDim] = g;
#endif VARIABLE_GAMMA
        
        w.cs[k][j][i].al[kYDim] = zls_sqrt(g*P/(d+kZLS_Epsilon));
        
        P = w.P[k][j][i].al[kZDim];
        d = w.d[k][j][i].al[kZDim];
        
#ifdef VARIABLE_GAMMA  
        g = zlsp_Gamma_Pd(P/d, cp);
        w.g [k][j][i].al[kZDim] = g;
#endif VARIABLE_GAMMA

        w.cs[k][j][i].al[kZDim] = zls_sqrt(g*P/(d+kZLS_Epsilon));
        
        P = w.P[k][j][i].ar[kXDim];
        d = w.d[k][j][i].ar[kXDim];

#ifdef VARIABLE_GAMMA  
        g = zlsp_Gamma_Pd(P/d, cp);
        w.g [k][j][i].ar[kXDim] = g;
#endif VARIABLE_GAMMA
        
        w.cs[k][j][i].ar[kXDim] = zls_sqrt(g*P/(d+kZLS_Epsilon));
        
        P = w.P[k][j][i].ar[kYDim];
        d = w.d[k][j][i].ar[kYDim];

#ifdef VARIABLE_GAMMA  
        g = zlsp_Gamma_Pd(P/d, cp);
        w.g [k][j][i].ar[kYDim] = g;
#endif VARIABLE_GAMMA
        
        w.cs[k][j][i].ar[kYDim] = zls_sqrt(g*P/(d+kZLS_Epsilon));
        
        P = w.P[k][j][i].ar[kZDim];
        d = w.d[k][j][i].ar[kZDim];
        
#ifdef VARIABLE_GAMMA  
        g = zlsp_Gamma_Pd(P/d, cp);
        w.g [k][j][i].ar[kZDim] = g;
#endif VARIABLE_GAMMA
        
        w.cs[k][j][i].ar[kZDim] = zls_sqrt(g*P/(d+kZLS_Epsilon));
        
      }
    }
  }
  return(err);
}
#endif !GODUNOV


