#ifndef ZLS_PHYSICS_H
#define ZLS_PHYSICS_H
#include <zls.h>
//
// prototypes and structures for zls hydro calculations that involve physics
// include in each separate zls file with a physics routine
//
//
// Top level physics structure
//

typedef struct phys {
  
	Real mu;  // constant if no cooling
	Real gamma; // constant if no cooling
	Real gamm; // gamma-1 constant if no cooling
	Real gamp; // gamma+1 constant if no cooling
  
#ifdef COOLING
  Cooling  coolParameters;
#endif COOLING
	
#ifdef GRAVITY
  Potential gravPotential;
#endif GRAVITY
		
} Physics, * PhysicsPtr;


//====================================================================================
//
int zlsp_InitGlobalEOS_Gamma( PhysicsPtr phys, Real gamma );
//
int zlsp_InitPhysics     ( PhysicsPtr   phys); // set default values for EOS/COOLING/GRAVITY
int zlsp_NewPhysics      ( PhysicsPtr * phys);
int zlsp_DisposePhysics  ( PhysicsPtr   phys);
//
//========================================
//
int zlsp_PrintPhysicsConfiguration   ( FILE * f, char * cfile );
int zlsp_ZPrintPhysicsConfiguration  ( zFile  f, char * cfile );
//
//========================================
//
// When inputting a quantity to the code, divide by the appropriate fidcual quantity to
// normalise it.  Ususally done at IO so there is no real performance loss.
//
// When writing out a quantity, restore the quantity by muliplying by the factor
//
// ie density:  
//  Input d, set w.d[i] = d/f_d0;  thereafter the internal density is scaled
//  Output d, d = w.d[i] * f_d0; to get back to original units.  this allows input in cgs
// for example without overflowing the solvers.  
//
// Adiabatic, scale free: 
//      x0 = v0 = d0 = 1.0  => time scale = 1.0
// problem parameters in init.conf can be set in simple dimensional units.  You can skip the
// divide by 1 anmd multiply by 1 if you want :)
//
// cgs ionised ISM on 100s pc scale:  
//      x0 = 1e20cm, v0 = 1.0e7 cm/s, d0 = 1.0e-24 g/cm^3  t0 = 10^13s ~300 000 yrs
// problem parameters in init.conf can then be input in cgs units and scaled internally
//
// Note grid.conf always specifies grid coordinates in normalised quanities, 
// ie in units of x0, a grid of 0.0 - 1.0 in the first example will be 0.0 - 1.0, 
// in the second example it will still have the values of 0.0- 1.0, but represent 
// a range of 0.0 - 1.0e20 cm.  1kpc would be ~30.8 in these units.
//
// scaling length, velocity, density, magnetic mu if used, 
// set to 0.0 for hydro (ignored)
// set to sqrt(4pi) for gaussian units input
// set to 1.0 for ideal MHD field strength input
int zlsp_SetFiducialScales ( Real x0, Real v0, Real d0, Real muB ); 
//
int zlsp_ApplyFiducialScalesWZone2D ( ZonePtr z );
int zlsp_UndoFiducialScalesWZone2D  ( ZonePtr z ); 
//
//========================================
//
// Get references to the primitive HD variables in a zone a mult-d Q arrays
// w returns with multi-dimensional arrays, even in the zone has plain QPtrs
// for data.  Presents a uniform user inteface.

int zlsp_ReferenceZoneWHD     ( WHD *  w, ZonePtr z ); 
int zlsp_ReferenceZoneUHD     ( UHD *  u, ZonePtr z ); 

#ifdef MEANSPLITSWEEPS
int zlsp_ReferenceZoneWHDAlt  ( WHD * w, ZonePtr z );
#endif MEANSPLITSWEEPS

// Free references to the primitive HD variables, without affecting zoneData
// if the zone is already multi-d zoneData then do nothing, but if the zone has 
// plain data blocks then this is needed to release temporary multi-d memory

int zlsp_ReleaseZoneWHD ( WHD w ); 
int zlsp_ReleaseZoneUHD ( UHD u );

//====================================================================================
//
// Get physical variables from a zone into a QPtr the same size and dims as z
// if the variable is a primitive variable then it is copied, if it is
// a derived variable it is computed and returned.
//
// The QPtr is sized with or without boundary cells according to the global
// macros: FILES_NO_BCS and FILES_HAVE_BCS, and the copy/computes are adjusted
// appropriately.
//
//====================================================================================
int zlsp_VelocitySlice( QPtr q, QPtr velocity, Real vMin, Real vMax, GridCoords gc);
int zlsp_SymmetricDifference( QPtr q, GridCoords gc);

int zlsp_GetZoneWQArr                ( QArr q  , ZonePtr z, Integer varIndex ); // q must already be allocated and the right size
int zlsp_GetZoneWQuantityCopy        ( QPtr * d, ZonePtr z, Integer varIndex );
int zlsp_GetZoneWQuantityCopy2DSlice ( QPtr * d, ZonePtr z, Integer varIndex , Integer slice, Integer idx);

//====================================================================================
int zlsp_CalcEnergyW1D          ( W1DArr w ); // compute E given P in w
int zlsp_CalcPressureW1D        ( W1DArr w ); // compute P given E in w

int zlsp_CalcEnergyU1D          ( U1DArr u ); // compute E given P in u
int zlsp_CalcPressureU1D        ( U1DArr u ); // compute P given E in u

int zlsp_CalcCellLREnergyW1D    ( W1DArr w );  // compute E given P in w and set up cell boundaries
int zlsp_CalcCellLREnergyU1D    ( U1DArr u );  // compute E given P in u and set up cell boundaries

//====================================================================================
// compute adiabatic sound speed for each cell, 
// used as a coefficient in many algorithms.
//====================================================================================

int zlsp_CalcWSoundSpeed1D      ( W1DArr w );
int zlsp_CalcWSoundSpeed2D      ( W2DArr w );
int zlsp_CalcWSoundSpeed3D      ( W3DArr w );

int zlsp_CalcCellWLRSoundSpeed1D( W1DArr w );
int zlsp_CalcCellWLRSoundSpeed2D( W2DArr w );
int zlsp_CalcCellWLRSoundSpeed3D( W3DArr w );

//====================================================================================

//Real zlsp_PFromE_W1D ( W1DArr w , Integer i);
//Real zlsp_EFromP_W1D ( W1DArr w , Integer i);
//Real zlsp_PFromE_U1D ( U1DArr u , Integer i);
//Real zlsp_EFromP_U1D ( U1DArr u , Integer i);
//

#endif ZLS_PHYSICS_H
