#ifndef ZLS_PHYSCONST_H
#define ZLS_PHYSCONST_H
#include <zls.h>
// prototypes and structures for zls hydro calculations that involve physics
// include in each separate zls file with a physics routine
//
// Default mu value -> monatomic ideal fully ionised solar metallicity gas.  
// Can be overridden in phys.conf
//
//  Ionised  : 0.6224e0
//  Neutral  : 1.300e0
//  Molecular: 2.37647
//
//
#define ZLS_EOS_MU   (0.6224e0)
//#define ZLS_EOS_MU   (1.300e0)
//#define ZLS_EOS_MU   (2.37647)
//
// Default gamma value -> monatomic ideal gas.  Can be overridden in phys.conf
// or by a computed EOS 
//
#define ZLS_EOS_GAMMA (5.0/3.0)
//#define ZLS_EOS_GAMMA (7.0/5.0)
//
//
//  kZLS_eos_gamma     =  equation of state gamma for fixed gamma case.
//  kZLS_eos_gamma_1   =  kZLS_eos_gamma - 1
//  kZLS_eos_gamma_2   =  kZLS_eos_gamma_1-1  = kZLS_eos_gamma - 2
//      
//
extern Real kZLS_eos_gamma  ;
extern Real kZLS_eos_gamma_1 ;
extern Real kZLS_eos_gamma_2 ;
extern Real kZLS_eos_invgamm;

extern Real kZLS_eos_a_iso;  // isothermal sound speed for isothermal EOS
extern Real kZLS_eos_a_iso2;

//
// Physical Constants - mixed gaussian c.g.s
//

extern const Real  k_clight;
extern const Real  k_me    ;
extern const Real  k_mp    ;
extern const Real  k_mh    ;
extern const Real  k_amu   ;
extern const Real  k_invamu; // 1/amu
extern const Real  k_boltz ;
extern const Real  k_stefan;
extern const Real  k_plank ;
extern const Real  k_guniv ;
extern const Real  k_eesu  ;

//   astronomical constants (cgs;
extern const Real  k_msol  ;
extern const Real  k_lsol  ;

//    ionized plasma constants
extern const Real  k_muhi  ; // mean mol weight *per H Atom*
extern const Real  k_muions; // mean mol weight per nucleus = mu in neutral atomic  medium (ie no electrons)
extern       Real  k_mu    ; // mean mol weight per particle, 0.6 ionised, 1.3 neutral, 2.35 molecular
extern       Real  k_invmu    ; // 1/(mean mol weight per particle)
extern const Real  k_AbnH  ;  // Abundance/(k_muhi * k_amu), total H nuclei number density
// ie nH_2 = 0.5 k_AbnH*d in fully  molecular gas, *for d not normalised*

//   conversion factors
extern const Real  f_cmau   ;
extern const Real  f_cmpc   ;
extern const Real  f_cmkpc  ;
extern const Real  f_cmkm   ;
extern const Real  f_everg  ;

// constants for stellar winds, yrs are useful generally.
extern const Real f_yrs     ; //Julian year
extern const Real f_invyrs  ; //inverse Julian year
extern const Real f_invMyrs ; //inverse Julian Myear
extern const Real f_gsunyrs ; // mSun /yr to g/s
//
// fiducial dimensional normalising factors for scaling input so that
// the solvers get and easier time, and cooling is scaled correctly
//
// not const as they are sometimes re-calculated at startup
//
// Useage:
// value with units / f_xx = scaled (model) value.  
// model value * f_xx = value with units
//
extern Real f_t0;      // time
extern Real f_invt0;   // 1/time
extern Real f_x0;      // x
extern Real f_invx0;   // 1/x
extern Real f_area0;   // x^2
extern Real f_vol0;    // x^3
extern Real f_v0;      // velocity
extern Real f_invv0;   // 1/velocity
extern Real f_a0;      // accelleration

extern Real f_r0;      // x spherical and cylindrical
extern Real f_z0;      // x in cylindrical
extern Real f_phi0;    // angluar scale, fixed radians
extern Real f_theta0;  // angluar scale, fixed radians

extern Real f_d0     ; // density
extern Real f_invd0  ; // 1.0/density
extern Real f_nHtot  ; // total H Atom (all forms) number density (f_nHtot* d),  *for d normalised*
extern Real f_P0     ; // pressure
extern Real f_m0     ; // mass 
extern Real f_E0     ; // energy density
extern Real f_EL0    ; // energy density/mass Lagrangian

extern Real f_tep0   ; // temperature = f_tep0*P/d
extern Real f_invtep0   ; // 1/temperature

extern Real f_cool0  ; // cooling/emmisivity
extern Real f_lam0   ; // cooling function
extern Real f_invlam0; // inverse cooling function

#ifdef GRAVITY
extern Real k_g      ; // scaled gravity constant, a = k_g m/ r^2, a, m and r in normalised units
extern Real k_gmsol  ; // scaled gravity constant using solar masses, a = k_gmsol M/ r^2, M in solar masses, a, and r in normalised units
#endif GRAVITY

//
// short strings for printing out quantity units, set with scaling constants at startup
//
extern char s_Time     [16];
extern char s_Length   [16];
extern char s_Area     [16];
extern char s_Volume   [16];
extern char s_Velocity [16];
extern char s_Accel    [16];
extern char s_Mass     [16];
extern char s_Density  [16];
extern char s_Pressure [16];
extern char s_Energy   [16];
extern char s_Temperature[16];
#ifdef MHD
extern char s_Magnetic  [16];
#endif MHD
#ifdef GRAVITY
extern char s_Gravity[16];
#endif GRAVITY
#ifdef COOLING
extern char s_Cool[16];
extern char s_Lambda[16];
#endif COOLING


#endif ZLS_PHYSCONST_H
