#include "zls.h"
//========================================
//
// The main loop for a complete ZLS simulation. It starts the main
// performance counters, initializes the Zone-Level system, then
// repeats the event loop until an ending condition is met.  The loop
// is fairly generic for any simulation, computing timesteps and
// performing sweeps of the grid.  As such it should not need to be
// modified very heavily, even for a completely different simulation
// type.
//
// Currently it drives a Fyris-Alpha hydrodynamic simulation.
//
//========================================

int main( int argc, char *argv[] ) {

  int err = noErr;
  
  double nCells = 0.0;
  Real timestep = 0.0;
  Integer cycle = 0;

  CPUCoords ct, ct0;
  
//========================================
//
// Set up ZLS and any simulation intialization, 
// plus user startups and set all the events 
// and clocks. Creates the log files and any output
// directories and so on. Initializes MPI if in 
// force.
//
//========================================
  
  err =  zls_Initialize(argc, argv);
  
  if ( err  == noErr ){
   
//========================================
// Ready to go, get some info for performance calculations.
//========================================

    err = zls_CalcAllLevelsEffectiveCellNumber ( stdout, &nCells, kZLS_DontShowCellCount );
  
    timestep  = 1.0; 
    err       = zlst_CalcAllTimeSteps ( &timestep ); // 1st estimate
    timestep *= kZLS_FracCFL; // start slower to allow for initial changes.
  
    zlst_StartCPUTimes ( &ct0 );
    zlst_StartCPUTimes ( &ct );

//========================================
// Start
//========================================
    zlsio_StartBanner  ( );  // start session
    
    err      = zlst_SetAllClocksTimeStep( &timestep);
    
    zlst_ProgressCheck( &ct , nCells );        
    
//========================================
// Loop the event manager checking for IO and other events until the sim ends
//========================================
    for ( cycle = 0;  ( zlst_CheckAllLevelEvents() == goCode ) ; cycle++){
      
//========================================
// Compute the next time step 
//========================================
      err = zlst_CalcAllTimeSteps (&timestep);
      if (err != noErr){
        zlsio_ErrPrintf(  " ERROR: CalcAllTimeSteps %d \n", err);
        zls_Finalize();
        return(err);
      }
      
        err = zlst_SetAllClocksTimeStep( &timestep);
        if (err != noErr){
          zlsio_ErrPrintf(  " ERROR: zlst_SetAllClocksTimeStep %d \n", err );
          zls_Finalize();
          return(err);
        }
      
//========================================
// Update the grid
//========================================
        err = zlsa_Sweep( kSweepMainLevel ); 
        if (err != noErr){
          zlsio_ErrPrintf( " ERROR: zlsa_Sweep %d \n", err );
          zls_Finalize();
          return(err);
        }
      
      zlst_ProgressCheck( &ct , nCells );   
      	  
    }
    
//========================================
// Done
//========================================
    zlsio_EndBanner  ( ); // end of  session

    zlst_EndCPUTimes     ( &ct0 );
    zlst_ElapsedCPUTimes ( &ct0, " All Simulation Cycles." );
    zlst_ElapsedCellRates( &ct0, nCells, " Simulation Cycles." );
    
  } else {
    zlsio_ErrPrintf( " ERROR: zls_Initialize failed %d \n", err );
  }
  
  zlst_EndCPUTimes     ( &gZLS_GlobalCPUTimer );
  zlst_ElapsedCPUTimes ( &gZLS_GlobalCPUTimer, " Overall Process Time" );
  
//========================================
// Cleanup and close log, shut down MPI if used
//========================================
  zls_Finalize();

  return (err);
  
}

