#include <zls.h>

//======================================================================
//
// User problem parameter initialisation/default values routine
// Called before the user's problem config file is read to set
// all problem parameters to sensible defaults in case the init.conf
// is faultly or even missing
//
//  The fields of the problem structure are defined by the user in
//  zls_problem.h and can be any variables desired by the user's routines.
//  The problem data is stored by reference in gLevels[0] and shared by all 
//  other levels, and can be accessed by any zone at need (see zlsp_UserZoneUpdate
//  and zlsp_InitAtmosphere for examples.
//
//======================================================================

int zlsp_InitProblem    ( ProblemPtr   prob){
  int err = noErr;
  
  if (prob != NULL) {
   
    prob->gD             = 1.0; 
    prob->gP             = 1.0; 

    prob->gR             =  0.1; 
    prob->gPRatio        = 10.0; 
    prob->gSmooth        = 0;
    
  }
  
  return (err);

}

//======================================================================
// standard allocation and destructor for problem strucutres, called
// by the ZLS system at the right time, and should not need to be
// modified in any way by the user.  Here to keep all problem code together
//
// The only exception is if levels are destroyed during a run and the user
// needs to deallocate any dynamic memory set up in zlsp_InitProblem
// or in zlsp_SetUpProblemFromConfigFile.  If this is the case then 
// the user needs to deallocate the memory in zlsp_DisposeProblem
// before  free(prob) is called.
//
// Currently the level list is fixed for the entire run and this is not
// necessary, but it will be in AMR.  Will be called by zls_DisposeLevel
// which doesn't exist yet. :)
//======================================================================

int zlsp_NewProblem(ProblemPtr * prob){
  int err = noErr;
  
  *prob = (ProblemPtr)malloc(sizeof(Problem));
  
  if (*prob == NULL) {
    err = memErr;
#if (ZLS_DEBUG >0)
    zlsio_ErrPrintf( " ERROR: zlsp_NewProblem allocation failed %d \n",err);
#endif
  }
  
  return (err);
}

int zlsp_DisposeProblem(ProblemPtr prob){
  int err = noErr;
  
  if ( prob != NULL ) {
    free(prob);
  }
  
  return (err);
}

int zlsp_InitZoneBlast( ZonePtr z );
int zlsp_InitZoneBlast( ZonePtr z ){
  
//======================================================================
//  Initialise a zone system variables (ie W), usually self contained, 
//  here the initialisation is a bit messy so subroutines are called from
//  here.  When called all the variables for the zone are initialised and 
//  ready to go, grid coordinates, problem vars etc.
//======================================================================
  
  int err = noErr;
  
  if (z != NULL){
    
    Integer level = z->zoneLevel;
    
    if ( level == (gNLevels-1)) {

      WHD zoneW = z->zoneW;    
      ProblemPtr prob = gLevels[level].levelProblem;  // get the user's problem parameters
      
#ifdef SYM1D
      ACoordPtr xx  = z->zoneGrid.x;
      
      Integer nti   = z->zoneGrid.nt[kXDim];
      Integer i;
      
      Real fi; // axis radial components
#endif SYM1D
      
#ifdef SYM2D
      ACoordPtr xx  = z->zoneGrid.x;
      ACoordPtr yx  = z->zoneGrid.y;
      
      Integer nti   = z->zoneGrid.nt[kXDim];
      Integer ntj   = z->zoneGrid.nt[kYDim];
      Integer i, j;
      
      Real fi,  fj; // axis radial components
#endif SYM2D
      
#ifdef SYM3D
      ACoordPtr xx  = z->zoneGrid.x;
      ACoordPtr yx  = z->zoneGrid.y;
      ACoordPtr zx  = z->zoneGrid.z;
      
      Integer nti   = z->zoneGrid.nt[kXDim];
      Integer ntj   = z->zoneGrid.nt[kYDim];
      Integer ntk   = z->zoneGrid.nt[kZDim];
      Integer i, j, k;
      
      Real fi,  fj,  fk; // axis radial components
#endif SYM3D
  
      
      Real P0 =  prob->gP;
      Real d0 =  prob->gD;
      
      Real r  = prob->gR;
      Real P1 = prob->gPRatio * P0 ;
          
#ifdef SYM2D
          
          Real x      = 0.0;
          Real y      = 0.0;
          Real dx     = 0.0;
          Real dy     = 0.0;
          Real radius = 0.0;
          Real f      = 1.0;
          
          for(j = 0; j < ntj; j++){
            y      = yx[j].c;
            dy     = yx[j].dx;
            for(i = 0; i< nti; i++){
              x      = xx[i].c;
              dx     = xx[i].dx;
              
              zoneW.d[j][i] = d0;
              zoneW.P[j][i] = P0;
              zoneW.vx[j][i] = 0.0;
              zoneW.vy[j][i] = 0.0;
              
              radius = zls_CvtCoordsToProjRadius2D( &fi, &fj, x, y);
              
              if ( prob->gSmooth == 1){
                if ( radius < 0.8*r ){
                  zoneW.P[j][i] = P1;
                }  else if ( radius < 1.1*r ){
                  f = zls_SmoothCircle2D( r, 0.0, 0.0, x, y, dx, dy, 256, 1.0);
                  zoneW.P[j][i] = f*P1 + (1.0-f)*P0;
                }
              } else { // no smoothing
                radius = zls_CvtCoordsToProjRadius2D( &fi, &fj, x, y);
                if ( radius < r ){
                  zoneW.P[j][i] = P1;
                }
              }
              
            } //i
          } //j
#endif SYM2D
          
#ifdef SYM3D
          
          Real x        = 0.0;
          Real y        = 0.0;
          Real z        = 0.0;
          Real dx       = 0.0;
          Real dy       = 0.0;
          Real dz       = 0.0;
          Real radius   = 0.0;
          Real f        = 1.0;

          for(k = 0; k < ntk; k++){
            z      = zx[k].c;
            dz     = zx[k].dx;
            for(j = 0; j < ntj; j++){
              y      = yx[j].c;
              dy     = yx[j].dx;
              for(i = 0; i< nti; i++){
                x      = xx[i].c;
                dx     = xx[i].dx;
                
                zoneW.d [k][j][i] =  d0;
                zoneW.P [k][j][i] =  P0;
                zoneW.vx[k][j][i] = 0.0;
                zoneW.vy[k][j][i] = 0.0;
                
                radius = zls_CvtCoordsToProjRadius3D( &fi, &fj, &fk, x, y, z);
                if ( radius < 2.0*r ){
                  f      = zls_SmoothSphere3D( r, 0.0, 0.0,0.0,
                                                       x,  y,  z, dx, dy, dz, 8, 1.0);
                   
                  zoneW.P[k][j][i] = f*P1 + (1.0-f)*P0;;

                }
              } //i
            } //j
          } //k
#endif SYM3D
          
      
    } // end wind on lowest level
    
  } else { err = memErr; } //z
  
  return(err);
}

//======================================================================
//  Standard zone initialsation routine
//======================================================================

int zlsp_InitZoneData( ZonePtr z ){
  
//======================================================================
//  Initialise a zone system variables (ie W), usually self contained, 
//  here the initialisation is a bit messy so subroutines are called from
//  here.  When called all the variables for the zone are initialised and 
//  ready to go, grid coordinates, problem vars etc.
//======================================================================
  
  int err = noErr;
  
  if (z != NULL){
    
    Integer level = z->zoneLevel;
    
    if ( level > 1) {
      err = zlsbc_FillZoneFromSuperZone( z, kInterpBCsFromSuper );
    } else {
      err = zlsp_InitZoneBlast(z);   
    }
        
  } else { err = memErr; }//z
  
  return(err);
}

//======================================================================
//  Standard level initialsation routine, loops through all zones of
//   a level. Called by ZLS after all parameters, coordinates, physics
//   etc have been set up.  Sets the gravity field arrays before the
//   calling zlsp_InitZoneData.
//======================================================================

int zlsp_InitLevelZones(Integer level){
  
  
  int err = noErr;
  Integer zone;
  ZonePtr  z;
  
#if (ZLS_VERBOSE >0)
#ifdef MPI
    if ( gMPIrank == 0){
#endif MPI
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " Level %d User Zones Initialisation: \n", level);
      zlsio_Printf( "%s\n",  DIVIDERLINE );
#ifdef MPI
    }
#endif MPI
#endif
  
  for ( zone = 0; zone <gLevels[level].nZones; zone++){
    z =  gLevels[level].levelZones[zone];
	  if (z != NULL){
#ifdef MPI	
		  if (z->zoneRankID == gMPIrank) {
#endif MPI
			  if ( z->zoneType == kAlgorithmZoneType ){ // only process appropriate zones, ie skip image zones
				  
#if (ZLS_DEBUG>0)
				  zlsio_ErrPrintf( " Initialising Zone: %2.2d:%2.2d \n", level, zone );
#endif
				  
				  err = zlsp_InitZoneData( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsp_InitZoneData Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
          
				  err = zls_SetZoneConstantBCValues( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zls_SetZoneConstantBCValues Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
#ifdef MEANSPLITSWEEPS
#ifdef SYM2D
          err =  zlsa_Synch2DData(z);
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsa_Synch2DSymmetricSweeps Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
#endif SYM2D
#endif   MEANSPLITSWEEPS   
          
          
			  } // z->zoneType == kAlgorithmZoneType 
#ifdef MPI			
		  }
#endif MPI
	  } else { err = zonErr; }
  } // zone
  
  return(err);
  
}


//======================================================================
// Optional routine to allow overriding the hard wired zlsp_InitProblem
// parameters from a config file, specified in zls.conf - usually init.conf
// Called by ZLS at the right time(tm)
//======================================================================

int   zlsp_SetUpProblemFromConfigFile(char * confFile){
  
//======================================================================
// Open the config file confFile, and parse for any user definded
// key,value pairs to set problem parameters (not grid initialisation)
// Calls zlsp_InitProblem, and before zlsp_InitLevelZones
//======================================================================
  
  int err = noErr;
  struct ccl_t  config;
  const char    *str1;
  
  Integer level;

#ifdef COOLING
  PhysicsPtr phys = gLevels[0].levelPhysics;  // get the global Physics parameters
  CoolingPtr cp   = NULL; 
  cp   = &(phys->coolParameters);  // get the coolng parameters
#endif COOLING
  
  ProblemPtr  prob = NULL;
  err = zlsp_NewProblem(&prob);
  
  if ( prob != NULL ) {
    
    config.comment_char = '#';
    config.sep_char     = '=';
    config.str_char     = '"';
    
    if (ccl_parse(&config, confFile)==noErr) {
      
#if (ZLS_VERBOSE > 0) 
      zlsio_Printf( "\n%s\n", DIVIDERLINE );
#endif
      zlsio_Printf( " FYRIS Hydro Problem Initialised from %s \n", confFile);
#if (ZLS_VERBOSE > 0) 
      zlsio_Printf(  "%s\n", DIVIDERLINE );
#endif
      
// Default Values
      
      err = zlsp_InitProblem(prob);
      
//======================================================================
// The user is free to define any key strings and interpret the 
// returned values in any way.  Some problems are simple (see the SOD setup)
// this one is more complex as it sets up a multi-component model with many
// different types of parameters.
//======================================================================
          
      Real d = prob->gD;
      if (( str1 = ccl_get( &config, "Ambient_Density"))){
        d = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Ambient_Density", str1);
#endif
      }
      prob->gD = d;
      
      Real P = prob->gP;
      if (( str1 = ccl_get( &config, "Ambient_Pressure"))){
        P = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Ambient_Pressure", str1);
#endif
      }
      prob->gP = P;
      
      Real r = prob->gR;
      if (( str1 = ccl_get( &config, "Initial_Radius"))){
        r = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Initial_Radius", str1);
#endif
      }
      prob->gR = r;
      
      Real Pr = prob->gPRatio;
      if (( str1 = ccl_get( &config, "Pressure_Ratio"))){
        Pr = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Pressure_Ratio", str1);
#endif
      }
      prob->gPRatio = Pr;
      
      Integer s = prob->gSmooth;
      if (( str1 = ccl_get( &config, "Smooth"))){
        s = atoi( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Smooth", str1);
#endif
      }
      s = s < 0 ? 0: s;
      s = s > 1 ? 1: s;
      prob->gSmooth = s;
      
#if (ZLS_VERBOSE>0)
      
      zlsio_Printf("%s\n",  DIVIDERLINE );
      zlsio_Printf(" Blast Parameters:\n");
      zlsio_Printf("    Ambient Density  :  %g \n", prob->gD);    
      zlsio_Printf("    Ambient Pressure :  %g \n", prob->gP);    
      zlsio_Printf("    Bubble Radius    :  %g \n", prob->gR);    
      zlsio_Printf("    Over Pressure    :  %g \n", prob->gPRatio); 
      if ( prob->gSmooth != 0){
      zlsio_Printf("    Smooth Pressure  :  %d \n", prob->gSmooth);  
      }
      zlsio_Printf("%s\n",  DIVIDERLINE );
      
#else
// log even if not verbose
      
      zlsio_LogPrintf("%s\n",  DIVIDERLINE );
      zlsio_LogPrintf(" Blast Parameters:\n");
      zlsio_LogPrintf("    Ambient Density  :  %g \n", prob->gD);    
      zlsio_LogPrintf("    Ambient Pressure :  %g \n", prob->gP);    
      zlsio_LogPrintf("    Bubble Radius    :  %g \n", prob->gR);    
      zlsio_LogPrintf("    Over Pressure    :  %g \n", prob->gPRatio);    
      if ( prob->gSmooth != 0){
      zlsio_LogPrintf("    Smooth Pressure  :  %d \n", prob->gSmooth);  
      }
      zlsio_LogPrintf("%s\n",  DIVIDERLINE );

#endif 
      
      
//======================================================================
// This code sets the problem into the level structures, mulit-zone
// multi-level or simple single structures 
//======================================================================
      
      
      gLevels[0].levelProblem = prob; // register the problem with all levels from 0 down.      
      if (gNLevels>1){
        for (level = 1; level < gNLevels; level++){
          gLevels[level].levelProblem = gLevels[0].levelProblem;
        }
      }
      
//======================================================================
// Now the level contents can be set - standard calls here dont' change
//======================================================================
      
      for (level = 0; level < gNLevels; level++){
        
#ifdef GRAVITY
        err =  zlsg_InitLevelGravityField( level ); // precompute the potential of to the gravity field over the level
        if (err != noErr ) {
          zlsio_ErrPrintf( " ERROR: zlsg_InitLevelGravityField: Level Field Failed: %d,  %d \n",level, err );
          return(err);
        }
#endif GRAVITY
        
        err = zlsp_InitLevelZones( level ); // last: user's levelProblem, initialised top down
        
      }
      
      if (gNLevels>1){
        for (level = (gNLevels-1); level > 0; level--){
          err = zlsbc_FillSuperLevelFromLevel(level);
        }
      }
      
#if (ZLS_VERBOSE>0)
      zlsio_Printf("%s\n",  DIVIDERLINE );
#else
      zlsio_LogPrintf("%s\n",  DIVIDERLINE );
#endif
      
      ccl_release(&config);
      
    }// config  
    
  } else {
    err = memErr;
    zlsio_ErrPrintf( " ERROR: Unable to allocate level 0 ProblemPtr \n");
    return (err);
  }
  
  return (err);
  
}

//======================================================================
// Routine for 'User; boundary cells set in grid.conf not used here
// but often used for jet inlet models
//======================================================================
#ifdef ZLS_USER_BOUNDARIES
int  zlsp_CalcUserZoneBCs ( ZonePtr z, int direction, int bndryKind ){
  return (noErr);
}
#endif ZLS_USER_BOUNDARIES


//======================================================================
// Routine for modifying anywhere in the grid in any way as part of
// a custom problem model. Not restricted to boundary cells, here used for 
// wind bubble/mass loss bubble energy injection and mass injection
// called after all other boundary conditions have been computed.
// Use sparingly, if not careful this routine can impact heavily on 
// the performance of the simulation.
//======================================================================
// 
// Normally deactivated (ie not even checked for calling) unless the macro
// ZLS_USER_UPDATE_ROUTINE is defined in zls_defs.  If not defined then
// this routine is not even callable - and this is the default setting.
// Also normally provided as an empty routine.
//

#ifdef ZLS_USER_UPDATE_ROUTINE

int  zlsp_UserZoneUpdate ( ZonePtr z  ){
  
  Integer err = noErr;
  ProblemPtr prob = NULL;
  PhysicsPtr phys = NULL;
  
  // A different type of BC, not the grid edges as all but setting cell values
  // on the grid.  Uses zoneW and is based on the zone data init routine rather than
  // the grid edge code of the basic boundary value routines.
  
  if ( z != NULL ){
    
    Integer level = z->zoneLevel;
    prob = gLevels[level].levelProblem;  // get the user's problem parameters
    phys = gLevels[level].levelPhysics;  // get the global eos parameters
    
  } else { err = memErr; } // z
  
  return (err);
}
#endif ZLS_USER_UPDATE_ROUTINE
