#include "zls.h"

#ifdef MPI

//=============================================================================
// Highlevel calls for general fyris/zls code to call
//=============================================================================

int zls_MPI_ImposeLevelLinks( Integer level ) {
  
  int err = noErr;
  
  ZonePtr z, zR;
  Integer zone;
  
  Integer dimIndex; // loop over dimensions of simulation
  //  Integer baseDirection; // starting direction of current dimension
  Integer dirIndex; // loop over directions in current dimension
  Integer dir; // actual direction in question
  Integer bndryType;
  
  
  //
  // Apply all the links in a given *dimension* at a time, 
  // It is vital that we go in opposite directions in turn for each dimension in 
  // turn, otherwise corner BCs will not be valid.
  // Unlike straight links we don't just loop dirIndex which looks like it does this
  // because of the intermediate buffering.  We need to setup, send/receive and set
  // each dimension in turn explicitly.
  //
  // ie the Y swaps can't be started until all the X send/recives are completely finished.
  // In particular the Y buffers can't be set and queued for receives until the 
  // X transactions are done, as some of the X infomration is needed and the swaps are not
  // 'in place' like normal links.
  //
  
  for (dimIndex = 0; dimIndex < kSymDims; dimIndex++) {
    
    int  nReqs       = 0;        // int for mpi parameters
    int nReqsCounter = 0; // mpi counter for recieve reqests as they are called
    Integer baseDirection = dimIndex*2; // 0, 2, 4 = Left, Down, Yon
    
    // count number of recieve requests needed for whole level in both directions 
    // constant at present but will be dynamic in an AMR calculation
    
    zls_MPI_CountLevelRecieveRequestsDim( &nReqs,  level, dimIndex ); // get all requests for X, Y  or Z dimensions
    
    MPI_Request reqs[nReqs]; // allocate temp array for request obj list
    zls_MPI_ClearRequests( reqs, nReqs);
    
    //========================================================================================
    // post non-blocking recieves for all links from zones on other ranks, 
    // for zones with zonedata that belong to this rank.
    // Set *all* send buffers for zones of this rank, whether linked remotely or locally
    //========================================================================================
    
    
    for ( zone = 0; zone < gLevels[level].nZones; zone++){
      z =  gLevels[level].levelZones[zone];    
      if ( z != NULL) {
        if ( z->zoneRankID == gMPIrank) { // zone + zone data on current process
          
          for (dirIndex = 0; dirIndex < 2; dirIndex++) {
            dir = dirIndex+baseDirection;
            
            bndryType = z->zoneBCType[dir]&kBoundaryTypeFlagNextZ; 
            if  (bndryType == kBoundaryTypeFlagNextZ) { // link in direction 
              
              zls_SetSendBufferFromGridDir(z, dir); // fill send buffers while here
              
              zR = z->zoneNext[dir]; // get link target
              if ( zR != NULL) {
                
                if ( zR->zoneRankID != gMPIrank) { // target is on other process
                  
                  // Zone data is on other process, need an MPI recieve
                  // Start a nonblocking receive, with next available request
                  
                  err = zls_MPI_NonBlockingReceiveBuffers( z, zR, dir, reqs, nReqsCounter);
                  if (err != noErr) { err = mpiErr; return(err);}
                  // count requests pending for this process
                  nReqsCounter++;
                } // zone data is on other process
                
                // can't get data into recieve buffers directly here, 
                // as not all send buffers are ready yet, but will
                // copy same process buffers directly  in send loop below
                
              } else { err = zonErr;}// zR
            } // link
            
          } // dirINdex loop
        } // local z
      } else { err = zonErr;} // z != NULL
    } // level zone count Loop
    
    //========================================================================================
    // nReqsCounter receive requests for dir are pending, and all our rank's send buffers are set
    // Now send out all this rank's sends to match other rank's recieves,
    // and copy directly from send buffers to recieve buffers of zones on same rank
    //========================================================================================
    
    for ( zone = 0; zone < gLevels[level].nZones; zone++){
      z =  gLevels[level].levelZones[zone];
      if ( z != NULL) {
        if ( z->zoneRankID == gMPIrank) {
          
          for (dirIndex = 0; dirIndex < 2; dirIndex++) {
            
            dir = dirIndex+baseDirection;
            
            bndryType = z->zoneBCType[dir]&kBoundaryTypeFlagNextZ; 
            if  (bndryType == kBoundaryTypeFlagNextZ) {
              
              zR = z->zoneNext[dir];
              if ( zR != NULL) {
                
                if ( zR->zoneRankID != gMPIrank) { 
                  
                  //
                  // Zone data in zR is on other process, need an MPI recieve
                  // already pending, then a ready send or standard send
                  // is used to set remote buffers.
                  //  Should be no delay as receives are already in place.
                  
                  err = zls_MPI_SendBuffers( zR, z, dir);
                  if (err != noErr) { err = mpiErr; return(err);}
                  
                } else {
                  //
                  // destination zone, zR,  is on same process as z, and data
                  // for z and zR are accessible,so we do a direct copy
                  //
                  err = zls_MPI_SetBuffersDirect( zR, z, dir);
                  if (err != noErr) { err = mpiErr; return(err);}
                  
                }
                
              } else { err = zonErr;}// zR
            } // link
          } // dirINdex loop
        } // local z
      } else { err = zonErr;}// z != NULL
    } // level Loop
    
    
    // wait for all pending receives to be
    // satisfied for current process.
    
    
    if (nReqsCounter > 0){
      err = zls_MPI_WaitRequests( reqs, nReqsCounter); 
      if ( err != noErr) { return(err); }
    }
    
    
    // now all buffers for current dimension are filled and ready to set
    // the actual zone BCs with
    
    for ( zone = 0; zone < gLevels[level].nZones; zone++){
      z =  gLevels[level].levelZones[zone];
      if ( z != NULL) {
        if ( z->zoneRankID == gMPIrank) {
          
          for (dirIndex = 0; dirIndex < 2; dirIndex++) {
            dir = dirIndex+baseDirection;
            
            bndryType = z->zoneBCType[dir]&kBoundaryTypeFlagNextZ; 
            if  (bndryType == kBoundaryTypeFlagNextZ) {
              
              // have a link, now set BCs from received buffer in that direction
              err = zls_GetBCsfromReceiveBufferDir(z, dir);
              if (err != noErr) { err = mpiErr; return(err);}
              
            } // link
            
          } // dirINdex loop
          
        } // local z
      } else { err = zonErr;} // z != NULL
      
      
    } // level Loop
    
    
    MPI_Barrier(MPI_COMM_WORLD); // wait for everyone before proceeding to rest of code
    
  } // dimINdex loop
  
  return(err);
  
  
}// zls_MPI_ImposeLevelLinks

int zls_MPI_GetLevelMergedZoneData(ZonePtr mergedZone, Integer level, Integer varIndex ) {
  
  int err = noErr;
  
  ZonePtr z;
  Integer zone;
  
  if ( gMPIrank != level ) { 
    
    // ranks other than level, just send their data to level, which has
    // mergedZone != NULL
    
    for ( zone = 0; zone < gLevels[level].nZones; zone++){
      z =  gLevels[level].levelZones[zone];    
      if ( z != NULL) {
        if ( z->zoneRankID == gMPIrank) {
          
          QPtr subImage = NULL;
          err           = zlsp_GetZoneWQuantityCopy  ( &subImage, z, varIndex); 
          
          //  start a send to rank 0
          
          //          MPI_Status status;
          int dest     = level;
          int tag      = z->zoneID;
          //
          GridCoords gc = z->zoneGrid;
          Counter nx   = 1UL; 
          Counter ny   = 1UL; 
          Counter nz   = 1UL; 
          
          zls_GetGridCubeDims(&nx, &ny, &nz, &gc); // allows for FILES_NO_BCS/FILES_HAVE_BCS macros
          int nElements = nx*ny*nz; // number of elements in zone data buffer, int for MPI
          
          if (( subImage !=NULL )&&( err == noErr)) {
            
#ifdef DOUBLE_PRECISION    
            err = MPI_Send((void*) subImage, nElements, 
                           MPI_DOUBLE, 
                           dest, tag, 
                           MPI_COMM_WORLD); 
#endif DOUBLE_PRECISION    
#ifdef SINGLE_PRECISION    
            err = MPI_Send((void*) subImage, nElements, 
                           MPI_FLOAT, 
                           dest, tag, 
                           MPI_COMM_WORLD); 
#endif SINGLE_PRECISION
            
          }
          
          zls_DisposeQuantity ( subImage ); subImage = NULL;  //  clear for next zone to use
          
        } // zone data is on rank
      } else { err = zonErr;} // z != NULL
    } // level zone Loop
    
  } else { // rank == level    
    for ( zone = 0; zone < gLevels[level].nZones; zone++){
      z =  gLevels[level].levelZones[zone];    
      if ( z != NULL) {
        
        QPtr subImage = NULL;
        
        if ( z->zoneRankID == gMPIrank) { // just get the data normally
          err           = zlsp_GetZoneWQuantityCopy  ( &subImage, z, varIndex); 
        } else {
          
          //  start a recieve for zone
          
          MPI_Status status;
          int source   = z->zoneRankID;
          int tag      = z->zoneID;
          
          //
          GridCoords gc = z->zoneGrid;
          Counter nx   = 1UL; 
          Counter ny   = 1UL; 
          Counter nz   = 1UL; 
          
          zls_GetGridCubeDims(&nx, &ny, &nz, &gc); // allows for FILES_NO_BCS/FILES_HAVE_BCS macros
          int nElements = nx*ny*nz; // number of elements in zone data buffer, int for MPI
          subImage = malloc(sizeof(Real)*nElements);
          
          if ( subImage !=NULL){
#ifdef DOUBLE_PRECISION    
            err = MPI_Recv((void*) subImage, nElements, 
                           MPI_DOUBLE, 
                           source, tag, 
                           MPI_COMM_WORLD, &status); 
#endif DOUBLE_PRECISION    
#ifdef SINGLE_PRECISION    
            err = MPI_Recv((void*) subImage, nElements, 
                           MPI_FLOAT, 
                           source, tag, 
                           MPI_COMM_WORLD, &status); 
#endif SINGLE_PRECISION    
          } else {
            err = memErr;
            return(err);
          }
          
        } // recieve for zone
        
        
        
        if (( subImage != NULL ) && ( mergedZone != NULL )) {
#ifdef FILES_HAVE_BCS 
          err = zlsim_FillLevelImageFromDataAndZone ( mergedZone,  subImage, kZLS_DataHasBCs, z );
#endif FILES_HAVE_BCS 
#ifdef FILES_NO_BCS 
          err = zlsim_FillLevelImageFromDataAndZone ( mergedZone,  subImage, kZLS_DataNoBCs, z );
#endif FILES_NO_BCS 
          if ( err != noErr ){
            err = ioErr;
            return (err);
          }
          zls_DisposeQuantity ( subImage ); subImage = NULL;  //  clear for next zone to use
        } else {
          err = memErr;
          return (err);
        }
        
      } else { err = zonErr;} // z != NULL
    } // level zone Loop
    
  } // rank == 0
  
  MPI_Barrier(MPI_COMM_WORLD); // wait for everyone before proceeding to rest of code
  
  return(err);
  
} // zls_MPI_GetLevelMergedZoneData


int zls_MPI_FillSuperLevelFromLevel( Integer level ) {
  
  int err = noErr;
  
  //  ZonePtr z, zR;
  //  Integer zone;
  //  
  //    for ( zone = 0; zone < gLevels[level].nZones; zone++){
  //      z =  gLevels[level].levelZones[zone];    
  //      if ( z != NULL) {
  //        Integer  zRank =  z->zoneRankID;
  //        zR = z->zoneSuper;
  //        if ( zR != NULL) {
  //          Integer zSuperRank =  zR->zoneRankID;
  //          
  //          if (( zRank == gMPIrank)&&( zSuperRank == gMPIrank)) {
  //            
  //            
  //          } else if ((zRank == gMPIrank)&&( zSuperRank != gMPIrank)) {
  //          
  //
  //          } else if ((zRank != gMPIrank)&&( zSuperRank == gMPIrank)) {
  //          
  //
  //          }
  //          
  //        }
  //        
  //      }
  //
  //     err = zls_MPI_GetRemoteZoneData( ZonePtr  z, ZonePtr zR, Integer varIndex );
  //     if ( err == noErr ){
  //     
  //     }
  //      
  //        if (( subImage != NULL ) && ( mergedZone != NULL )) {
  //#ifdef FILES_HAVE_BCS 
  //          err = zlsbc_FillUpperZoneFromZone ( mergedZone,  subImage, 0 );
  //#endif FILES_HAVE_BCS 
  //#ifdef FILES_NO_BCS 
  //          err = zlsbc_FillUpperZoneFromZone ( mergedZone,  subImage ,0 );
  //#endif FILES_NO_BCS 
  //          if ( err != noErr ){
  //            err = ioErr;
  //            return (err);
  //          }
  //          zls_DisposeQuantity ( subImage ); subImage = NULL;  //  clear for next zone to use
  //        } else {
  //          err = memErr;
  //          return (err);
  //        }
  //          
  //      } else { err = zonErr;} // z != NULL
  //    } // level zone Loop
  //
  //  } // rank == 0
  //  
  //  MPI_Barrier(MPI_COMM_WORLD); // wait for everyone before proceeding to rest of code
  
  return(err);
  
} // zls_MPI_FillSuperLevelFromLevel


int zls_MPI_GetRemoteZoneData( ZonePtr  z, ZonePtr zR, Integer varIndex ){
  int err = noErr;
  
  // get the zone data in remote zone zR, with varIndex into the corresponding zonedata in
  // zone z.  Assumes zR and z are the exact same size.  Otherwise should work for any
  // zone type.  A loop over varIndex for zone zR should transfer a copy of all the zone
  // data in zR to z.  
  
  if ( (z != NULL) && ( zR != NULL ) ) {
    if (( gMPIrank == z->zoneRankID)||( gMPIrank == zR->zoneRankID)) {
      
      GridCoords gc = z->zoneGrid;
      Counter nx   = 1UL; 
      Counter ny   = 1UL; 
      Counter nz   = 1UL; 
      
      zls_GetGridDataDims(&nx, &ny, &nz, &gc);
      int nElements = nx*ny*nz; // number of elements in zone data buffer, int for MPI
      
      if ( gMPIrank == zR->zoneRankID) { // do the send
        
        //          MPI_Status status;
        int dest     = z->zoneRankID;
        int tag      = z->zoneID;
        QPtr q = NULL;
        
        err = zls_GetZoneDataDirect(&q,  zR, varIndex);
        if (( q !=NULL )&&( err == noErr)) {
          
#ifdef DOUBLE_PRECISION    
          err = MPI_Send((void*) q, nElements, 
                         MPI_DOUBLE, 
                         dest, tag, 
                         MPI_COMM_WORLD); 
#endif DOUBLE_PRECISION    
#ifdef SINGLE_PRECISION    
          err = MPI_Send((void*) q, nElements, 
                         MPI_FLOAT, 
                         dest, tag, 
                         MPI_COMM_WORLD); 
#endif SINGLE_PRECISION
          
        }
        
      } else if ( gMPIrank == z->zoneRankID) { // do the recieve
        
        MPI_Status status;
        int source   = zR->zoneRankID;
        int tag      = z->zoneID;
        QPtr q = NULL;
        
        err = zls_GetZoneDataDirect(&q,  z, varIndex);
        if ( (q !=NULL)&&( err == noErr) ) {
          
#ifdef DOUBLE_PRECISION    
          err = MPI_Recv((void*) q, nElements, 
                         MPI_DOUBLE, 
                         source, tag, 
                         MPI_COMM_WORLD, &status); 
#endif DOUBLE_PRECISION    
#ifdef SINGLE_PRECISION    
          err = MPI_Recv((void*) q, nElements, 
                         MPI_FLOAT, 
                         source, tag, 
                         MPI_COMM_WORLD, &status); 
#endif SINGLE_PRECISION    
          
        }
        
      } // recieve
      
    }        
  } else {
    err = zonErr;
    
    return(err);
  }
  
  return(err);
}

int zls_MPI_SetRemoteZoneData(  ZonePtr zR,  ZonePtr z, Integer varIndex ){
  int err = noErr;
  
  // Set the zone data in remote zone zR, with varIndex from the corresponding zonedata in
  // zone z.  Assumes zR and z are the exact same size.  Otherwise should work for any
  // zone type.  A loop over varIndex for zone z should transfer a copy of all the zone
  // data in z to zR.  
  
  if ( (z != NULL) && ( zR != NULL ) ) {
    if (( gMPIrank == z->zoneRankID)||( gMPIrank == zR->zoneRankID)) {
      
      GridCoords gc = z->zoneGrid;
      Counter nx   = 1UL; 
      Counter ny   = 1UL; 
      Counter nz   = 1UL; 
      
      zls_GetGridDataDims(&nx, &ny, &nz, &gc);
      int nElements = nx*ny*nz; // number of elements in zone data buffer, int for MPI
      
      if ( gMPIrank == z->zoneRankID) { // do the send
        
        //        MPI_Status status;
        int dest     = zR->zoneRankID;
        int tag      = zR->zoneID;
        QPtr q = NULL;
        
        err = zls_GetZoneDataDirect(&q,  z, varIndex);
        if (( q !=NULL )&&( err == noErr)) {
          
#ifdef DOUBLE_PRECISION    
          err = MPI_Send((void*) q, nElements, 
                         MPI_DOUBLE, 
                         dest, tag, 
                         MPI_COMM_WORLD); 
#endif DOUBLE_PRECISION    
#ifdef SINGLE_PRECISION    
          err = MPI_Send((void*) q, nElements, 
                         MPI_FLOAT, 
                         dest, tag, 
                         MPI_COMM_WORLD); 
#endif SINGLE_PRECISION
          
        }
        
      } else if ( gMPIrank == zR->zoneRankID) { // do the recieve
        
        MPI_Status status;
        int source   = z->zoneRankID;
        int tag      = zR->zoneID;
        QPtr q = NULL;
        
        err = zls_GetZoneDataDirect(&q,  zR, varIndex);
        if ( (q !=NULL)&&( err == noErr) ) {
          
#ifdef DOUBLE_PRECISION    
          err = MPI_Recv((void*) q, nElements, 
                         MPI_DOUBLE, 
                         source, tag, 
                         MPI_COMM_WORLD, &status); 
#endif DOUBLE_PRECISION    
#ifdef SINGLE_PRECISION    
          err = MPI_Recv((void*) q, nElements, 
                         MPI_FLOAT, 
                         source, tag, 
                         MPI_COMM_WORLD, &status); 
#endif SINGLE_PRECISION    
          
        }
        
      } // recieve
      
    }        
  } else {
    err = zonErr;
    
    return(err);
  }
  
  return(err);
}

//=============================================================================
// lowlevel calls for internal MPI implementation calls
//=============================================================================

int zls_MPI_CountLevelRecieveRequestsDir( int * nReqs, Integer level, Integer dir ){
  
  int err = noErr;
  ZonePtr z, zR;
  Integer zone;
  Integer bndryType;
  Integer n = 0;
  
  for ( zone = 0; zone < gLevels[level].nZones; zone++){
    z =  gLevels[level].levelZones[zone];
    if ( z != NULL) {
      if ( z->zoneRankID == gMPIrank) {
        bndryType = z->zoneBCType[dir]&kBoundaryTypeFlagNextZ; 
        if  (bndryType == kBoundaryTypeFlagNextZ) {
          zR = z->zoneNext[dir];
          if ( zR != NULL) {
            if ( zR->zoneRankID != gMPIrank) { 
              n++;
            } // zone data is on other process
          } else { err = zonErr;} // zR
        } // link
      } // local z
    }else { err = zonErr;}  // z != NULL
  } // level Loop
  
  *nReqs = n;
  
  return(err);
  
}

int zls_MPI_CountLevelRecieveRequestsDim( int * nReqs, Integer level, Integer dim  ){
  
  int err = noErr;
  ZonePtr z, zR;
  Integer zone;
  
  Integer bndryType;
  Integer baseDirection = dim*2; // 0, 2, 4 = Left, Down, Yon
  Integer dirIndex; // loop over directions in current dimension
  Integer dir; // actual direction in question
  
  Integer n = 0;
  
  for ( zone = 0; zone < gLevels[level].nZones; zone++){
    z =  gLevels[level].levelZones[zone];
    if ( z != NULL) {
      if ( z->zoneRankID == gMPIrank) {
        
        for (dirIndex = 0; dirIndex < 2; dirIndex++) {
          dir = dirIndex+baseDirection;
          
          bndryType = z->zoneBCType[dir]&kBoundaryTypeFlagNextZ; 
          if  (bndryType == kBoundaryTypeFlagNextZ) {
            zR = z->zoneNext[dir];
            if ( zR != NULL) {
              if ( zR->zoneRankID != gMPIrank) { 
                n++;
              } // zone data is on other process
            } else { err = zonErr;} // zR
          } // link
          
        } // dirINdex loop
      } // local z
    }else { err = zonErr;}  // z != NULL
  } // level Loop
  
  *nReqs = n;
  
  return(err);
  
}


int zls_MPI_CountLevelRecieveRequests( int * nReqs, Integer level ){
  
  int err = noErr;
  ZonePtr z, zR;
  Integer zone;
  Integer dirIndex;
  Integer bndryType;
  Integer n = 0;
  
  for ( zone = 0; zone < gLevels[level].nZones; zone++){
    z =  gLevels[level].levelZones[zone];
    if ( z != NULL) {
      if ( z->zoneRankID == gMPIrank) {
        for (dirIndex = 0; dirIndex < kNGeomDirs; dirIndex++) {
          bndryType = z->zoneBCType[dirIndex]&kBoundaryTypeFlagNextZ; 
          if  (bndryType == kBoundaryTypeFlagNextZ) {
            zR = z->zoneNext[dirIndex];
            if ( zR != NULL) {
              if ( zR->zoneRankID != gMPIrank) { 
                n++;
              } // zone data is on other process
            } else { err = zonErr;} // zR
          } // link
        } // dirINdex loop
      } // local z
    }else { err = zonErr;}  // z != NULL
  } // level Loop
  
  *nReqs = n;
  
  return(err);
  
}

int zls_MPI_CountLevelFillRequests( int * nReqs, Integer level ){
  
  int err = noErr;
  ZonePtr z, zR;
  Integer zone;
  Integer n = 0;
  
  for ( zone = 0; zone < gLevels[level].nZones; zone++){
    z =  gLevels[level].levelZones[zone];
    if ( z != NULL) {
      if ( z->zoneRankID == gMPIrank) {
        if ( z->zoneType == kAlgorithmZoneType ){ 
          
          zR = z->zoneSuper;
          if ( zR != NULL) {
            if ( z->zoneType == kAlgorithmZoneType ){ 
              if ( zR->zoneRankID != gMPIrank) { 
                n++;
              } // zR rank
            }  // zR type
          } // zR
          
        } // type
      } // local z
    }else { err = zonErr;}  // z != NULL
  } // level Loop
  
  *nReqs = n;
  
  return(err);
  
}


int zls_MPI_ClearRequests( MPI_Request * reqs, int n){
  
  int err = noErr;
  
  if (n > 0) {
    int i;
    for ( i = 0; i < n; i++) reqs[i] = MPI_REQUEST_NULL;
  }
  
  return(err);
  
}


int zls_MPI_WaitRequests( MPI_Request * reqs, int n){
  
  int err = noErr;
  
  MPI_Status stats[n]; // auto array
  
  if (n > 0) {
    err = MPI_Waitall(n, reqs, stats); 
  }
  
  return(err);
  
}





int zls_MPI_NonBlockingReceiveBuffers( ZonePtr z, ZonePtr zR, Integer linkDir, MPI_Request * reqs, int n){
  //
  // Initiate a non-blocking receive request and return the request for 
  // use later in a wait all call after all the ready sends have been sent
  //
  
  int err = noErr;
  
  if (( z != NULL)&&(zR != NULL)) {
    
    
    int source   = zR->zoneRankID;
    
    int sendDir = zls_opposite(linkDir);
    int id      = z->zoneID;
    int tag     = (sendDir<<24)|id;
    
    Real * buf     = z->zoneRecieveBuffer[linkDir];
    int  nElements = z->zoneBufferSize[linkDir];
#if (ZLS_DEBUG>0)   
    zlsio_Printf("NBR : rank:%d srcrank: %d dir: %d dstrank: %d tgtid: %d tag %d \n",gMPIrank, source, sendDir, gMPIrank, id, tag);
#endif (ZLS_DEBUG>0)    
    
    // queue up a recieve for a later ready send
#ifdef DOUBLE_PRECISION    
    err = MPI_Irecv((void*) buf, nElements, 
                    MPI_DOUBLE, 
                    source, tag, 
                    MPI_COMM_WORLD, 
                    &reqs[n]); 
#endif DOUBLE_PRECISION    
#ifdef SINGLE_PRECISION    
    err = MPI_Irecv((void*) buf, nElements, 
                    MPI_FLOAT, 
                    source, tag, 
                    MPI_COMM_WORLD, 
                    &reqs[n]); 
#endif SINGLE_PRECISION    
    
    
  }else { err = zonErr;}
  
  return(err);
  
}


int zls_MPI_ReceiveBuffers( ZonePtr z, ZonePtr zR, Integer linkDir){
  //
  // Initiate a blocking receive for 
  // a send already or yet to be sent
  // if not buffered or the send is not yet
  // sent this may block
  
  int err = noErr;
  
  if (( z != NULL)&&(zR != NULL)) {
    
    MPI_Status status;
    int source  = zR->zoneRankID;
    int sendDir = zls_opposite(linkDir);
    int id      = z->zoneID;
    int tag     = (sendDir<<24)|id;
    
    Real * buf     = z->zoneRecieveBuffer[linkDir];
    int  nElements = z->zoneBufferSize[linkDir];
#ifdef DOUBLE_PRECISION    
    err = MPI_Recv((void*) buf, nElements, 
                   MPI_DOUBLE, 
                   source, tag, 
                   MPI_COMM_WORLD, &status); 
#endif DOUBLE_PRECISION    
#ifdef SINGLE_PRECISION    
    err = MPI_Recv((void*) buf, nElements, 
                   MPI_FLOAT, 
                   source, tag, 
                   MPI_COMM_WORLD, &status); 
#endif SINGLE_PRECISION    
  }else { err = zonErr;}
  
  return(err);
  
}


int zls_MPI_SendBuffers( ZonePtr zR, ZonePtr z, Integer linkDir){
  //
  // Initiate a standard MPI send for a set of zone send buffers
  //
  
  int err = noErr;
  
  if (( z != NULL)&&(zR != NULL)) {
    
    
    int dest     = zR->zoneRankID;
    int sendDir  = linkDir;
    int id       = zR->zoneID;
    int tag      = (sendDir<<24)|id;
    
#if (ZLS_DEBUG>0)   
    zlsio_Printf("Send: rank:%d dstrank: %d dir: %d srcrank: %d tgtid: %d tag %d \n",gMPIrank, dest, sendDir, gMPIrank, id, tag);
#endif (ZLS_DEBUG>0)    
    
    Real * buf     = z->zoneSendBuffer[sendDir];
    int  nElements = z->zoneBufferSize  [sendDir];
#ifdef DOUBLE_PRECISION    
    err = MPI_Send((void*) buf, nElements, 
                   MPI_DOUBLE, 
                   dest, tag, 
                   MPI_COMM_WORLD); 
#endif DOUBLE_PRECISION    
#ifdef SINGLE_PRECISION    
    err = MPI_Send((void*) buf, nElements, 
                   MPI_FLOAT, 
                   dest, tag, 
                   MPI_COMM_WORLD); 
#endif SINGLE_PRECISION    
    
  }else { err = zonErr;}
  
  return(err);
  
}

int zls_MPI_ReadySendBuffers( ZonePtr zR, ZonePtr z, Integer linkDir){
  //
  // Initiate an MPI ready send for a set of zone send buffers
  // must correspond to an existing recieve
  //
  
  int err = noErr;
  
  if (( z != NULL)&&(zR != NULL)) {
    
    int dest     = zR->zoneRankID;
    
    int sendDir  = linkDir;
    int id       = zR->zoneID;
    int tag      = (sendDir<<24)|id;
    
    Real * buf     = z->zoneSendBuffer[sendDir];
    int  nElements = z->zoneBufferSize[sendDir];
#ifdef DOUBLE_PRECISION    
    err = MPI_Rsend((void*) buf, nElements, 
                    MPI_DOUBLE, 
                    dest, tag, 
                    MPI_COMM_WORLD); 
#endif DOUBLE_PRECISION
#ifdef SINGLE_PRECISION    
    err = MPI_Rsend((void*) buf, nElements, 
                    MPI_FLOAT, 
                    dest, tag, 
                    MPI_COMM_WORLD); 
#endif SINGLE_PRECISION
    
  }else { err = zonErr;}
  
  return(err);
  
}


int zls_MPI_SetBuffersDirect(  ZonePtr zR, ZonePtr z, Integer linkDir){
  int err = noErr;
  
  if ((z!= NULL) && (zR!=NULL)) {
    int  nElements = z->zoneBufferSize[linkDir];
    Integer recDir = zls_opposite(linkDir);
    QPtr dst = zR->zoneRecieveBuffer[recDir];
    QPtr src  = z->zoneSendBuffer[linkDir];
    
    if (( src != NULL) && ( dst != NULL) ){
      memcpy(dst, src, sizeof(Real)*nElements);
    } else { err = memErr; return(err);}
    
  } else { err = zonErr; }
  
  return(err);
  
}


// filling and getting buffers to and from grid and BCs

int zls_SetSendBufferFromGridAll ( ZonePtr z ){
  //
  //
  int err = noErr;
  Integer dirIndex;
  
  for (dirIndex = 0; dirIndex < kNGeomDirs; dirIndex++) {
    
    err = zls_SetSendBufferFromGridDir ( z , dirIndex );
    if ( err != noErr) {
      return(err);
    }
    
  } // dirIndex loop
  
  return(err);
}


int zls_SetSendBufferFromGridAllLinks ( ZonePtr z ){
  //
  //
  int err = noErr;
  Integer dirIndex;
  Integer bndryType;
  
  for (dirIndex = 0; dirIndex < kNGeomDirs; dirIndex++) {
    bndryType = z->zoneBCType[dirIndex]&kBoundaryTypeFlagNextZ; 
    if  (bndryType == kBoundaryTypeFlagNextZ) {
      
      // have a link, now set BCs from received buffer in that direction
      err = zls_SetSendBufferFromGridDir ( z , dirIndex );
      if ( err != noErr) {
        return(err);
      }
      
    } // link
  } // dirIndex loop
  
  return(err);
}

int zls_SetSendBufferFromGridDir ( ZonePtr z, Integer dir ){
  
  int err = noErr;
  
  Counter  ni,  nj,  nk;
  Counter nBufferCount;
  Size nBufferSize;
  
  if ( z != NULL) {
    
    Integer nbc    = z->zoneBCells;
    
#ifdef SYM1D
    Counter  i;
    Integer is  =  z->zoneGrid.ns[kXDim];
    Integer ie  =  z->zoneGrid.ne[kXDim];
    Integer nti =  z->zoneGrid.nt[kXDim];
#endif SYM1D
    
#ifdef SYM2D
    Counter  i,  j;
    Integer is  =  z->zoneGrid.ns[kXDim];
    Integer ie  =  z->zoneGrid.ne[kXDim];
    Integer nti =  z->zoneGrid.nt[kXDim];
    
    Integer js  =  z->zoneGrid.ns[kYDim];
    Integer je  =  z->zoneGrid.ne[kYDim];
    Integer ntj =  z->zoneGrid.nt[kYDim];
    
#endif SYM2D
    
#ifdef SYM3D
    Counter  i,  j,  k;
    Integer is  =  z->zoneGrid.ns[kXDim];
    Integer ie  =  z->zoneGrid.ne[kXDim];
    Integer nti =  z->zoneGrid.nt[kXDim];
    
    Integer js  =  z->zoneGrid.ns[kYDim];
    Integer je  =  z->zoneGrid.ne[kYDim];
    Integer ntj =  z->zoneGrid.nt[kYDim];
    
    Integer ks  =  z->zoneGrid.ns[kZDim];
    Integer ke  =  z->zoneGrid.ne[kZDim];
    Integer ntk =  z->zoneGrid.nt[kZDim];
#endif SYM3D
    
    Integer varIndex = 0;
    
    zls_GetGridDirBCDims  ( &ni, &nj, &nk, &z->zoneGrid,  dir );
    nBufferSize  = sizeof(Real)* z->zoneBufferSize[dir];
    nBufferCount = z->zoneBufferSize[dir];
    QPtr qsb =  z->zoneSendBuffer  [dir];
    
    if ( qsb != NULL ) {
      
      for (varIndex = 0; varIndex <  z->nQ; varIndex++){ // loop over all variables
        
        
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM1D
        QPtr q   = z->zoneData[varIndex];
#endif SYM1D
#ifdef SYM2D
        Q2DArr q;
        zls_SetQuantity2DArr(&q  , z->zoneData[varIndex], ntj, nti);
#endif SYM2D
#ifdef SYM3D
        Q3DArr q;
        zls_SetQuantity3DArr(&q, z->zoneData[varINdex], ntk, ntj, nti);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
        
#ifdef ZONEDATA_ARE_MULTIDARRAYS
        QArr q = z->zoneData[varIndex];
#endif ZONEDATA_ARE_MULTIDARRAYS
        
        
#ifdef SYM1D
        switch (dir)  {
          case kDirLeft:
            for(i = 0; i < nbc; i++){
              qsb[i+ni*varIndex] = q[is+i];
            }; //i
            break;
            case kDirRight:
            for(i = 0; i < nbc; i++){
              qsb[i+ni*varIndex] = q[ie-nbc+i];
            }; //i
        } // direction switch
#endif SYM1D
#ifdef SYM2D
        switch (dir)  {
          case kDirLeft:
            for(j = 0; j < ntj; j++){ 
              for(i = 0; i < nbc; i++){
                qsb[i+ni*(j+nj*varIndex)] = q[j][is+i];
              }; //i
            } //j
            break;
            case kDirRight:
            for(j = 0; j < ntj; j++){ 
              for(i = 0; i < nbc; i++){
                qsb[i+ni*(j+nj*varIndex)]  = q[j][ie-nbc+i];
              }; //i
            } //j
            break;
            case kDirDown:
            for(i = 0; i < nti; i++){
              for(j = 0; j < nbc; j++){
                qsb[i+ni*(j+nj*varIndex)]  = q[js+j][i];
              } //j
            } //i
            break;
            case kDirUp:
            for(i = 0; i < nti; i++){
              for(j = 0; j < nbc; j++){
                qsb[i+ni*(j+nj*varIndex)]  = q[je-nbc+j][i];
              }; //j
            } // i
            break;
        } // direction switch
#endif SYM2D
#ifdef SYM3D
        switch (dir)  {
          case kDirLeft:
            for(k = 0; k < ntk; k++){ 
              for(j = 0; j < ntj; j++){ 
                for(i = 0; i < nbc; i++){
                  qsb[i+ni*(j+nj*(k+nk*varIndex))]  = q[k][j][is+i];
                } //i
              } //j
            } //k
            break;
            case kDirRight: 
            for( k = 0; k < ntk; k++ ){ 
              for(j = 0; j < ntj; j++){ 
                for(i = 0; i < nbc; i++){
                  qsb[i+ni*(j+nj*(k+nk*varIndex))] = q[k][j][ie-nbc+i];
                } //i
              } //j
            } //k
            break;
            case kDirDown:
            for(k = 0; k < ntk; k++){ //
              for(i = 0; i < nti; i++){
                for(j = 0; j < nbc; j++){
                  qsb[i+ni*(j+nj*(k+nk*varIndex))] = q[k][js+j][i];
                } //j
              } //i
            } //k
            break;
            case kDirUp:
            for(k = 0; k < ntk; k++){
              for(i = 0; i < nti; i++){
                for(j = 0; j < nbc; j++){ 
                  qsb[i+ni*(j+nj*(k+nk*varIndex))] = q[k][je-nbc+j][i];
                } //j
              } //i
            } //k
            break;
            case kDirYon:
            for(j = 0; j < ntj; j++){ 
              for(i = 0; i < nti; i++){
                for(k = 0; k < nbc; k++){ 
                  qsb[i+ni*(j+nj*(k+nk*varIndex))] = q[ks+k][j][i];
                } //k
              } //j 
            } //i
            break;
            case kDirHither:
            for(j = 0; j < ntj; j++){ 
              for(i = 0; i < nti; i++){
                for(k = 0; k < nbc; k++){ 
                  qsb[i+ni*(j+nj*(k+nk*varIndex))] = q[ke-nbc+k][j][i];
                } //k
              } //j
            } //i
        } // direction switch 
#endif SYM3D
        
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM2D
        zls_ReleaseQuantity2DArr(q);
#endif SYM2D
#ifdef SYM3D
        zls_ReleaseQuantity3DArr(q);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
        
        
      } // zone data variables, varIndex
    } // d != NULL
    
  } else { err = zonErr; }
  
  return(err);
  
}

int zls_GetBCsfromReceiveBufferAll( ZonePtr z ){ 
  //
  //
  int err = noErr;
  Integer dirIndex;
  
  for (dirIndex = 0; dirIndex < kNGeomDirs; dirIndex++) {
    
    err = zls_GetBCsfromReceiveBufferDir ( z , dirIndex );
    if ( err != noErr) {
      return(err);
    }
    
  } // dirIndex loop
  
  return(err);
}

int zls_GetBCsfromReceiveBufferAllLinks( ZonePtr z ){ 
  //
  // checks bdrytype for links, and sets all zone BCs that have link receive buffers
  //
  int err = noErr;
  Integer dirIndex;
  Integer bndryType;
  
  for (dirIndex = 0; dirIndex < kNGeomDirs; dirIndex++) {
    bndryType = z->zoneBCType[dirIndex]&kBoundaryTypeFlagNextZ; 
    if  (bndryType == kBoundaryTypeFlagNextZ) {
      
      // have a link, now set BCs from received buffer in that direction
      err = zls_GetBCsfromReceiveBufferDir ( z , dirIndex );
      if ( err != noErr) {
        return(err);
      }
      
    } // link
  } // dirIndex loop
  
  return(err);
}

int zls_GetBCsfromReceiveBufferDir ( ZonePtr z , Integer dir ){
  
  int err = noErr;
  Counter  ni,  nj,  nk;
  Counter nBufferCount;
  Size nBufferSize;
  
  if ( z != NULL) {
    
    Integer nbc    = z->zoneBCells;
    
#ifdef SYM1D
    Counter  i;
    Integer is  =  z->zoneGrid.ns[kXDim];
    Integer ie  =  z->zoneGrid.ne[kXDim];
    Integer nti =  z->zoneGrid.nt[kXDim];
#endif SYM1D
    
#ifdef SYM2D
    Counter  i;
    Counter  j;
    //    Integer is  =  z->zoneGrid.ns[kXDim];
    Integer ie  =  z->zoneGrid.ne[kXDim];
    Integer nti =  z->zoneGrid.nt[kXDim];
    
    //    Integer js  =  z->zoneGrid.ns[kYDim];
    Integer je  =  z->zoneGrid.ne[kYDim];
    Integer ntj =  z->zoneGrid.nt[kYDim];
    
#endif SYM2D
    
#ifdef SYM3D
    Counter  i;
    Counter  j;
    Counter  k;
    //    Integer is  =  z->zoneGrid.ns[kXDim];
    Integer ie  =  z->zoneGrid.ne[kXDim];
    Integer nti =  z->zoneGrid.nt[kXDim];
    
    //    Integer js  =  z->zoneGrid.ns[kYDim];
    Integer je  =  z->zoneGrid.ne[kYDim];
    Integer ntj =  z->zoneGrid.nt[kYDim];
    
    //    Integer ks  =  z->zoneGrid.ns[kZDim];
    Integer ke  =  z->zoneGrid.ne[kZDim];
    Integer ntk =  z->zoneGrid.nt[kZDim];
#endif SYM3D
    
    Integer varIndex = 0;
    
    
    zls_GetGridDirBCDims  ( &ni, &nj, &nk, &z->zoneGrid,  dir );
    nBufferSize  = sizeof(Real)* z->zoneBufferSize[dir];
    nBufferCount = z->zoneBufferSize[dir];
    QPtr qrb =  z->zoneRecieveBuffer[dir];
    
    if ( qrb != NULL ) {
      
      for (varIndex = 0; varIndex <  z->nQ; varIndex++){ // loop over all variables
        
        
        
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM1D
        QPtr q   = z->zoneData[varIndex];
#endif SYM1D
#ifdef SYM2D
        Q2DArr q;
        zls_SetQuantity2DArr(&q  , z->zoneData[varIndex], ntj, nti);
#endif SYM2D
#ifdef SYM3D
        Q3DArr q;
        zls_SetQuantity3DArr(&q, z->zoneData[varINdex], ntk, ntj, nti);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
        
#ifdef ZONEDATA_ARE_MULTIDARRAYS
        QArr q = z->zoneData[varIndex];
#endif ZONEDATA_ARE_MULTIDARRAYS
        
        
#ifdef SYM1D
        switch (dir)  {
          case kDirLeft:
            for(i = 0; i < nbc; i++){
              q[i]   = qrb[i+ni*varIndex];
            }; //i
            break;
            case kDirRight:
            for(i = 0; i < nbc; i++){
              q[i+ie] = qrb[i+ni*varIndex];
            }; //i
        } // direction switch
#endif SYM1D
#ifdef SYM2D
        switch (dir)  {
          case kDirLeft:
            for(j = 0; j < ntj; j++){ 
              for(i = 0; i < nbc; i++){
                q[j][i] = qrb[i+ni*(j+nj*varIndex)];
              }; //i
            } //j
            break;
            case kDirRight:
            for(j = 0; j < ntj; j++){ 
              for(i = 0; i < nbc; i++){
                q[j][i+ie] = qrb[i+ni*(j+nj*varIndex)];
              }; //i
            } //j
            break;
            case kDirDown:
            for(i = 0; i < nti; i++){
              for(j = 0; j < nbc; j++){
                q[j][i] = qrb[i+ni*(j+nj*varIndex)];
              } //j
            } //i
            break;
            case kDirUp:
            for(i = 0; i < nti; i++){
              for(j = 0; j < nbc; j++){
                q[j+je][i] = qrb[i+ni*(j+nj*varIndex)];
              }; //j
            } // i
            break;
        } // direction switch
#endif SYM2D
#ifdef SYM3D
        switch (dir)  {
          case kDirLeft:
            for(k = 0; k < ntk; k++){ 
              for(j = 0; j < ntj; j++){ 
                for(i = 0; i < nbc; i++){
                  q[k][j][i] = qrb[i+ni*(j+nj*(k+nk*varIndex))];
                } //i
              } //j
            } //k
            break;
            case kDirRight: 
            for( k = 0; k < ntk; k++ ){ 
              for(j = 0; j < ntj; j++){ 
                for(i = 0; i < nbc; i++){
                  q[k][j][i+ie] = qrb[i+ni*(j+nj*(k+nk*varIndex))];
                } //i
              } //j
            } //k
            break;
            case kDirDown:
            for(k = 0; k < ntk; k++){ //
              for(i = 0; i < nti; i++){
                for(j = 0; j < nbc; j++){
                  q[k][j][i] = qrb[i+ni*(j+nj*(k+nk*varIndex))];
                } //j
              } //i
            } //k
            break;
            case kDirUp:
            for(k = 0; k < ntk; k++){
              for(i = 0; i < nti; i++){
                for(j = 0; j < nbc; j++){ 
                  q[k][j+je][i] = qrb[i+ni*(j+nj*(k+nk*varIndex))];
                } //j
              } //i
            } //k
            break;
            case kDirYon:
            for(j = 0; j < ntj; j++){ 
              for(i = 0; i < nti; i++){
                for(k = 0; k < nbc; k++){ 
                  q[k][j][i] = qrb[i+ni*(j+nj*(k+nk*varIndex))];
                } //k
              } //j 
            } //i
            break;
            case kDirHither:
            for(j = 0; j < ntj; j++){ 
              for(i = 0; i < nti; i++){
                for(k = 0; k < nbc; k++){ 
                  q[k+ke][j][i] = qrb[i+ni*(j+nj*(k+nk*varIndex))];
                } //k
              } //j
            } //i
        } // direction switch 
#endif SYM3D
        
#ifdef ZONEDATA_ARE_PLAINBLOCKS
#ifdef SYM2D
        zls_ReleaseQuantity2DArr(q);
#endif SYM2D
#ifdef SYM3D
        zls_ReleaseQuantity3DArr(q);
#endif SYM3D
#endif ZONEDATA_ARE_PLAINBLOCKS
        
        
      } // zone data variables, varIndex
      
    } // d != NULL
    
    
  } else { 
    err = zonErr; 
  }
  
  return(err);
  
}

#endif MPI

