#ifndef ZLS_GRIDS_H
#define ZLS_GRIDS_H
#include "zls.h"

//
//========================================
//
// grid axis constants
//
//========================================
//
enum {    
  kAxisTypeSpatial        =  0,  // spatial coordinate
  kAxisTypeCRadius        =  1,  // Circular  radial coordinate >= 0, 1/r - used in cylinders
  kAxisTypeSRadius        =  2,  // spherical radial coordinate >= 0, 1/r^2
  kAxisTypePhi            =  3,  // 0 +/- 2.0 pi radians, same as azimuth
  kAxisTypeTheta          =  4,  // 0 +/- 0.5 pi radians, same as alitude/latitude,   Projects phi as cos(Theta)
  kAxisTypePolarTheta     =  5  // 0->pi. colatitude, down from +z pole.  Projects phi as sin(polarTheta)
};

enum {    
  kAxisKindLinear        =  0, 
  kAxisKindLog           =  1,
  kAxisKindGauss         =  2
};

enum {    
  kAllocateAxes          =  0, 
  kNoAxes                =  1,
  kFixedGrid             =  0,
  kLagrangeGrid          =  1
};

#ifdef LAGRANGIAN
enum { kAxesType = kLagrangeGrid};
#else
enum { kAxesType = kFixedGrid};
#endif


//====================================================================================
//
// CPU and Performance Timing
//
//====================================================================================

typedef struct cputime{
	
  clock_t  cputot[2]; // cpu clock cycles - not necesarily real time
  clock_t  cpuusr[2];
  clock_t  cpusys[2];
  
  clock_t  realtime[2];    // real (clock on the wall) time, but in cycles too
  Real    modeltime[2];    // model time units
  Counter  modelcycles[2]; // model computation cycles
  
  //   clock_t       cpu[6];    // 
  
} CPUCoords, *CPUCoordPtr;

//
//========================================
// grid/coordinate Space and Time Structures
//========================================
//

typedef struct clock {     // temporal coordinates, timers, def here so it is seen throughout zls
  Real              t;   // simulation time in Zone
  Real             dt;   // timestep in Zone    
  Real        endTime;   // time to end sim, in scaled model units
  
  Counter       cycles;   // computation cycles in Zone
  Counter    endCycles;   // cycle number to end sim
  Counter timingCycles;   // computation cycles , start at 0 even after a restart
  
  // utility counters/timers for various output - see zls_timing in hydro
  
  Real         tFrame;   // increment by dt, reset to 0 when triggered
  Real          tUser;   // timer for user output routine
  Real          tData;   // timer for 16 bit signed data  output routine
  Real          tSave;   // timer for complete state save
  
  Real      frameTime;   // when to trigger the next frame, an interval rather
  Real       userTime;   // than an absolute time
  Real       dataTime;   // 
  Real       saveTime;   // 
  
  Counter nCyclesFrame;   // cycles for frame output, increment by 1, zero on event trigger
  Counter  nCyclesUser;   // cycles for user output
  Counter  nCyclesData;   // cycles for 16 bit signed data output
  Counter  nCyclesSave;   // cycles for restart output
  
  Counter  frameCycles;   // cycles between frame output
  Counter   userCycles;   // cycles between user output
  Counter   dataCycles;   // cycles between 16 bit signed data output
  Counter   saveCycles;   // cycles between state save  output
  
} TimeCoords, * TCoordPtr; // usually a single element, Ptr for pass by reference

typedef struct range{  // integer + float range, uniquely not 0 based indexes
  
  Integer nD;    // dimensions of the range - ie how many array elements are used below
  Integer nG;    // Boundary Cells needed if making a variable, 0 if a quantity
  //         //
  // these are for carrying type and kind info from levels and zones down to lower levels.
  //
  Integer type[kMaxCoordDims];  // 0 = spatial, radial, kAxisTypePhi, phi, azimuth  +/- 2.0 , 2 = theta, altutude +/- 0.5, 
  Integer kind[kMaxCoordDims];  //  0 = linear, 1 = log ...   
  //
  Integer n0[kMaxCoordDims];    // intial indices (i = n0...
  Integer rn[kMaxCoordDims];    // index ranges index ne =  n0+rn :  i = n0; i<ne; i++
  //
  Real    x0[kMaxCoordDims];    // float value of left edge, ie ns- = x0
  Real    rx[kMaxCoordDims];    // right = x0 + rx  = right edge of (ne-1) == left edge of (ne)
  
} Range, * RangePtr; // usually a single element, Ptr for pass by reference

typedef struct axis{ // interlaced array of spatial coordinates and coeffs for each cell
  Real l;        //  cell coords - left edge
  Real r;        //  cell coords - right edge
  Real c;        //  cell coords - centre - convenience - not strictly conserving
  Real dx;       //  cell coords - width -
  Real idx;      //  cell coords - inverse of the width - comp once for main grid and save many divides later
  Real dA;       //  area
  Real dVol;     // volume (dx dA in cartesian)
#ifdef SPHERICAL  
  Real sl, sr, sc; // sine(left), sine(right), sine(centre);
  Real cl, cr, cc; // cosine(left), cosine(right), cosine(centre);
#endif SPHERICAL
#ifdef POLARSPHERICAL 
  Real sl, sr, sc; // sine(left), sine(right), sine(centre);
  Real cl, cr, cc; // cosine(left), cosine(right), cosine(centre);
#endif POLARSPHERICAL
#ifdef CYLINDRICAL  
  Real sl, sr, sc; // sine(left), sine(right), sine(centre);
  Real cl, cr, cc; // cosine(left), cosine(right), cosine(centre);
#endif CYLINDRICAL
#ifdef CIRCULAR 
  Real sl, sr, sc; // sine(left), sine(right), sine(centre);
  Real cl, cr, cc; // cosine(left), cosine(right), cosine(centre);
#endif CIRCULAR
  
  
  //
  // CW 84  - regular and irregular grids para coeffs equation 1.6 a_r =
  // a(j) + pc[0] * (a(j+1)-a(j)) + pc[1] * da(j+1) + pc[2] * da(j)
  // 
  // equation 1.7 da(j) = pc[3] * (a(j+1) - a(j)) + pc[4] * (a(j) -
  // a(j-1)) used to reconstruct the al and ar values if they are not
  // already known by lagrangian interpolation between cells. used to get
  // al and ar for bothPLM and PPM, not needed for GODUNOV
#ifdef PPM
  Real pc0, pc1, pc2, pc3, pc4;  // parabolic grid interp coeffs
#endif PPM
#ifdef PLM
  Real pc0, pc1, pc2, pc3, pc4;  // parabolic grid interp coeffs
#endif PLM
#ifdef GODUNOV
  // a_r = a(j) = al so no coeffs needed
#endif GODUNOV
  
} AxisCoords,  * ACoordArr;  // usually an array of axis: AxisCoords x[n];  x[i].l = left coord of i cell

typedef ACoordArr ACoordPtr; // equivalent, use as appropriate to purpose

typedef struct grid{
  
  Integer           nD; //  number of dimensions, 1, 2 or 3;
  Integer           nG; //  number of boundary cells, same all axes 
  Integer      type[kMaxCoordDims]; //  spatial, radial etc 
  Integer      kind[kMaxCoordDims]; //  kAxisKindLinear = linear, kAxisKindLog = log  etc...  
  
  Counter         n[kMaxCoordDims]; //  number of quantity cells, no bCells in Q arrays
  Counter        ns[kMaxCoordDims]; //  start of active cells, each axis, steps over bCells
  Counter        ne[kMaxCoordDims]; //  end of active cells, each axis
  Counter        nt[kMaxCoordDims]; //  total # of cells, each axis
  
  Real           xl[kMaxCoordDims]; //
  Real           xr[kMaxCoordDims]; //  inner (left) x, outer (right) x of grid - active range
  Real           dx[kMaxCoordDims]; //  simple global dx, range/nt, dxs in detail are in axes
  
  ACoordArr          x; // cell coordinate arrays, NULL if not used
  ACoordArr          y; 
  ACoordArr          z;
  
  ACoordArr         x0; //  initial axes if using lagrangian/remapping, NULL otherwise
  ACoordArr         y0;
  ACoordArr         z0;
  
} GridCoords, * GCoordPtr; //  usually a single element, Ptr for pass by reference
// calculating coeffs

// fundamental index and floating point ranges, call set for each dim
int    zls_SetRangeDims (RangePtr r, Integer ndims, Integer nbc);
int    zls_SetRangeType (RangePtr r, Integer dim, Integer type);
int    zls_SetRangeKind (RangePtr r, Integer dim, Integer kind);

int    zls_SetRange (RangePtr r, Integer dim, Integer nl, Integer nr, Real xl, Real xr);
int    zls_SetRangeDelta (RangePtr r, Integer dim,  Integer n0, Integer n, Real x0, Real x);

int    zls_ScaleRangeCells(RangePtr r, Integer dim,  Integer n);

int    zls_PrintRange         ( FILE *f, RangePtr r);
int    zls_ZPrintRange        ( zFile f, RangePtr r);
int    zls_UserLogPrintRange  ( RangePtr r);

// 1D index <-> floating point cell coordinate maps

int    zls_NewAxis        (ACoordPtr *x, Integer nbc, Integer ni);  
// nt = ni + nbc usually dynamically allocated,
// but could be fixed array too.
// grid coords use dynamics axes.

int    zls_DisposeAxis      ( ACoordPtr x );          // only if dynamically allocated

int    zls_CopyAxis         ( ACoordPtr dst,  ACoordPtr src, Integer nbc, Integer ni);

int    zls_InitAxis            ( ACoordPtr x, Integer type, Integer kind, Integer nbc, Integer ni, Real ix, Real ox); 

int    zls_SetAxisRange        ( ACoordPtr x, Integer dim, RangePtr r);   
int    zls_CopyAxisRange       ( ACoordPtr dst, ACoordPtr src, Integer dim, RangePtr r );

int    zls_RescaleAxis          ( ACoordPtr x, Integer type, Integer ni, Real s);  // 
int    zls_CalcAxisDeltas       ( ACoordPtr x, Integer type, Integer ni);  
int    zls_CalcAxisGeometrySource  ( ACoordPtr x, ACoordPtr x0, Integer type, Integer ni); // ratio of area changes on cell LR boundaries, = dVol/dx
int    zls_CalcAxisParaCoeffs   ( ACoordPtr x, Integer nt);  // private routine called by init axis

int    zls_PrintAxisCoords         ( FILE * f, ACoordPtr x, Integer ns, Integer ne, Integer nt, Integer type);
int    zls_ZPrintAxisCoords        ( zFile f, ACoordPtr x, Integer ns, Integer ne, Integer nt, Integer type);
int    zls_UserLogPrintAxisCoords  ( ACoordPtr x, Integer ns, Integer ne, Integer nt, Integer type);

// composite zone/variable coordinate systems

int    zls_NewGridCoords      (GCoordPtr *g); // if dynamically allocating, usually not necessary 
int    zls_DisposeGridCoords  (GCoordPtr g);  // *only* if dynamically allocated, usually not necessary 
int    zls_ClearGridCoords    (GCoordPtr g);  // just zot grid and the axes, no free g

int    zls_InitGridCoords     (GCoordPtr g, Integer ndims, Integer nbc, Integer allocAxes, RangePtr r); // types and kinds in range

int    zls_SetGridRange       (GCoordPtr g, Integer dim, Integer useAxes, RangePtr r);
int    zls_SetGridDimKind     (GCoordPtr g, Integer dim, Integer kind );
int    zls_SetGridDimType     (GCoordPtr g, Integer dim, Integer type );

int    zls_GetRangeDimsFromGrid( RangePtr r, GCoordPtr g );

int    zls_GetRangeFromGrid    ( RangePtr r, Integer dim, GCoordPtr g );

// returned range has active cells and  number of bcs 

int   zls_GetSubRangeFromGrid        (RangePtr r, Integer dim, GCoordPtr g, Integer n0, Integer n1); 
int   zls_GetSubRangeFromGridDelta   (RangePtr r, Integer dim, GCoordPtr g, Integer n0, Integer rn); 
//
int   zls_GetGridDataDims  ( Counter * nx, Counter * ny, Counter * nz, GCoordPtr g );
//
// get two image dimensions from a grid given a slice type 0 dimensions
// return 1 so at least a single row of pixels can be created. ie a 2D
// grid with 10, 10, 0 in a YZ slice gives 10, 1.  A 3D grid of 10, 10,
// 1 is actually equivalent to the 2D one if boundary cells are not
// included
//
// Dimes returned are consistent with FILES_HAVE_BCS settings and are
// intended for image and other scaled IO, not restart dumps
//
int   zls_GetGridImageDims ( Counter * nx, Counter * ny, GCoordPtr g, Counter slice );
int   zls_GetGridCubeDims  ( Counter * nx, Counter * ny, Counter * nz, GCoordPtr g );

// return dimensions of grid BCs in given direction, setting 1s for unused dims
int    zls_GetGridDirBCDims  ( Counter * ni, Counter * nj, Counter * nk, GCoordPtr g, Integer direction );

// n0 and rn are active area, returned range
int    zls_InitSubGridCoords  (GridCoords * g, GridCoords * sg, Integer allocAxes, RangePtr r);  
int    zls_CloneGridCoords    (GCoordPtr g1, GCoordPtr g0); // copy g0 into g1, creating new axes if needed

int    zls_PrintGridCoords          (  FILE *f, GCoordPtr g);
int    zls_ZPrintGridCoords         ( zFile f, GCoordPtr g);
int    zls_UserLogPrintGridCoords   ( GCoordPtr g);

#endif
