#include "zls.h"

//
// global list head for each level, to access zones across levels not via owner/sub tree
//
Integer                 gNLevels; 
LevelZoneList gLevels[MAXLEVELS];
//

int     zls_InitLevelList   (Integer level, TimeCoords cl, Real CFL, GridCoords gc, 
                             Integer ndims, Integer nbc, Integer refine, Integer interp){  
  
  // returns error code, 0 = noErr must init levels in order 0..n
  // otherwise the level total refinement will be wrong.
  //
  
  int err = noErr;
  static int inited = 0;
  
  if (inited==0) {
    gNLevels = 0;
    gNextZoneID       = 0; // call this Init only ONCE
    zlsppm_InitVarInterpCoeffs();
    inited++;
  }
  
  gLevels[level].level           = level; // 0 = top level, only one list per level allowed
  gLevels[level].nZones          = 0; 
  gLevels[level].maxZones        = 1; 
  gLevels[level].levelNCells     = 0.0; 
  
  gLevels[level].levelZoneKind   = kAlgorithmZoneType; // the interpretation of all zones in this level
  
  gLevels[level].levelClock      = cl; 
  gLevels[level].levelCFL        = CFL; // Courant-Friedrichs-Level numer for all of level
  
  gLevels[level].levelGrid       = gc; 
  gLevels[level].levelDims       = ndims; 
  
  if (level >0) {
    gLevels[level].levelRefinement      = refine; // 1, 2, 4 or 8, 1 only for level 0
    gLevels[level].levelTotalRefinement = gLevels[level-1].levelTotalRefinement*refine; 
  } else {
    gLevels[level].levelRefinement      = 1; // 1, 2, 4 or 8, 1 only for level 0
    gLevels[level].levelTotalRefinement = 1; 
  }
  
  gLevels[level].levelBCells      = nbc; // 
  gLevels[level].levelSuperBCells = nbc/refine; // 
  gLevels[level].levelInterpolation = interp; // 
  
  if ((level!=0)&&((nbc<refine)||(nbc % refine != 0))) {
    zlsio_ErrPrintf( " ERROR: zls_InitLevelList, level %d, nbc and refine incompatible, %d %d\n",level, nbc, refine);
    return(dimErr);
  }
  
  gLevels[level].levelImageZone = NULL; // if used must be a zoneType == kImageZoneType
  gLevels[level].levelIO        = NULL;
  gLevels[level].levelPhysics   = NULL;
  gLevels[level].levelProblem   = NULL; // User Problem Parameters stucture
  gLevels[level].levelZones     = NULL; // treated as an array when allocated
  
  return(err);
}

int  zls_AllocateLevelZoneArray(Integer level,  Integer n){   
  //
  // makes room for n zones in list, can be expanded later, but more mem
  // efficient if a group is initied in the beginning.  n can be v large
  // if needed. only called once for each level in a fixed refined grid
  // AddZoneToLevel used for dynmamically allocated zones in AZ mode
  // levelZones holds copies of zonePtrs, so a lot can be handled here
  // the array of ptr is allocated, not the zones themselves!!
  //
  int err = noErr;
  Integer i;
  
  if (gNLevels <= level) gNLevels = level +1; 
  
  gLevels[level].maxZones     = 0;
  gLevels[level].nZones       = 0;
  gLevels[level].levelZones = (ZonePtrArr) malloc(n*sizeof(ZonePtr)) ; // treated as an array when allocated
  
  if (gLevels[level].levelZones == NULL) {
    err = memErr;
    return(err);
  }
  
  for ( i= 0; i< n; i++){
    gLevels[level].levelZones[i]     = NULL;
  }
  
  gLevels[level].maxZones     = n ; // succeeded, set n
  return(err);
}

int   zls_ClearLevelList(Integer level){             // returns error code, 0 = noErr
#pragma unused(level)
  return(nyiErr);
}

int   zls_CalcAllLevelsUniformSweepEquivNumber ( double * nCells  ){
  int err = noErr;
  Integer level = 0;
  
  double nR     = 1.0;
  double  n     = 0.0;
  double  n0     = 0.0;
  
  for (level = 0; level < gNLevels; level ++){
    n0 += gLevels[level].levelNCells;
  } // level
  
  nR = (double) gLevels[gNLevels-1].levelTotalRefinement;
#ifdef SYM1D
  n = gLevels[0].levelNCells*nR;
#endif SYM1D
#ifdef SYM2D
  n = gLevels[0].levelNCells*nR*nR;
#endif SYM2D
#ifdef SYM3D
  n = gLevels[0].levelNCells*nR*nR*nR;
#endif SYM3D
  
  *nCells = n;
  return(err);
  
}

int   zls_CalcAllLevelsRefinedSweepEquivNumber ( double * nEqivCells ){
  int err = noErr;
  Integer level = 0;
  double  l     = 0;
  double  fraction     = 1.0;
  double  mutliple     = 1.0;
  double  n      = 0.0;
  double  n1     = 0.0;
  double  n0     = 0.0;
  double nR      = 1.0;
  
  nR = (double) gLevels[gNLevels - 1].levelRefinement;
  for (level = 0; level < gNLevels; level ++){
    l = gNLevels - level - 1;
    fraction = pow(nR, -l);  
    mutliple = (double) gLevels[level].levelTotalRefinement;
    n0 += gLevels[level].levelNCells*fraction;
    n1 += gLevels[level].levelNCells*mutliple;
  } // level
  
  nR = (double) gLevels[gNLevels-1].levelTotalRefinement;
#ifdef SYM1D
  n = gLevels[0].levelNCells*nR;
#endif SYM1D
#ifdef SYM2D
  n = gLevels[0].levelNCells*nR*nR;
#endif SYM2D
#ifdef SYM3D
  n = gLevels[0].levelNCells*nR*nR*nR;
#endif SYM3D
  
  *nEqivCells =  n*nR;
  return(err);
}

int   zls_CalcAllLevelsEquivNumberCellNumber   ( double * nEqivCells ){
  int err = noErr;
#ifdef REFINED_SWEEPS
  err = zls_CalcAllLevelsRefinedSweepEquivNumber ( nEqivCells );
#endif REFINED_SWEEPS
#ifdef UNIFORM_SWEEPS
  err = zls_CalcAllLevelsUniformSweepEquivNumber ( nEqivCells );
#endif UNIFORM_SWEEPS
  return(err);
}
//
// This gives the total *effective resolution* of the simulation in
// that cells of level l are swept nRefine times per level above so
// this is the number of cell calculations required per overall sweep.
// Note: high l levels get swept many more times per overall step so
// keep their number to a minimim.
//
int   zls_CalcAllLevelsRefinedSweepCellNumber ( FILE * f,  double * nCells, Integer show ){
  int err = noErr;
  Integer level = 0;
  double  l     = 0;
  double  fraction     = 1.0;
  double  mutliple     = 1.0;
  double  n      = 0.0;
  double  n1     = 0.0;
  double  n0     = 0.0;
  double nR      = 1.0;
  
  if ( show ) {
    zlsio_FPrintf( f, "%s\n",  DIVIDERLINE );
  }
  
  nR = (double) gLevels[gNLevels - 1].levelRefinement;
  for (level = 0; level < gNLevels; level ++){
    l = gNLevels - level - 1;
    fraction = pow(nR, -l);  
    mutliple = (double) gLevels[level].levelTotalRefinement;
    if ( show ) {
      zlsio_FPrintf( f, " Level %d : Zones : %d\n", level,  gLevels[level].nZones);
      zlsio_FPrintf( f, " Level %d : Cells : %10.0f \n", level,  gLevels[level].levelNCells);
      zlsio_FPrintf( f, " Level %d : Cells : %10.0f per Level %d Cycle \n", level,  gLevels[level].levelNCells*mutliple, 0 );
      zlsio_FPrintf( f, " Level %d : Cells : %10.0f per Level %d Cycle \n", level,  gLevels[level].levelNCells*fraction, gNLevels - 1 );
      zlsio_FPrintf( f, "%s\n",  DIVIDERLINE );
    }
    n0 += gLevels[level].levelNCells*fraction;
    n1 += gLevels[level].levelNCells*mutliple;
  } // level
  
  nR = (double) gLevels[gNLevels-1].levelTotalRefinement;
#ifdef SYM1D
  n = gLevels[0].levelNCells*nR;
#endif SYM1D
#ifdef SYM2D
  n = gLevels[0].levelNCells*nR*nR;
#endif SYM2D
#ifdef SYM3D
  n = gLevels[0].levelNCells*nR*nR*nR;
#endif SYM3D
  
  if ( show  && (gNLevels>1)) {
    zlsio_FPrintf( f, " Total   : %10.0f Cells per Level %d Cycle.\n", n0, gNLevels - 1 );
    zlsio_FPrintf( f, " Total   : %10.0f Cells per Level %d Cycle.\n", n1, 0 );
    zlsio_FPrintf( f, " Total   : %10.0f Equivalent Level %d Cells per Level %d Cycle.\n", n, gNLevels - 1,  gNLevels - 1 );
    zlsio_FPrintf( f, " Total   : %10.0f Equivalent Level %d Cells per Level %d Cycle.\n", n*nR, gNLevels - 1 , 0);
    zlsio_FPrintf( f, " Total   : %10.0f Model Time Speed Up.\n", n/n0 );
    zlsio_FPrintf( f, "%s\n",  DIVIDERLINE );
  }
  
  *nCells =  n1;
  return(err);
}

// simpler count assuming all levels are swept the same number of time (ie UniformSweep)

int   zls_CalcAllLevelsUniformSweepCellNumber ( FILE * f,  double * nCells , Integer show ){
  int err = noErr;
  Integer level = 0;
  
  double nR     = 1.0;
  double  n     = 0.0;
  double  n0     = 0.0;
  
  if ( show ) {
    zlsio_FPrintf( f, "%s\n",  DIVIDERLINE );
  }
  for (level = 0; level < gNLevels; level ++){
    if ( show ) {
      zlsio_FPrintf( f, " Level %d : Zones : %d\n", level,  gLevels[level].nZones);
      zlsio_FPrintf( f, " Level %d : Cells : %10.0f \n", level,  gLevels[level].levelNCells);
    }
    n0 += gLevels[level].levelNCells;
  } // level
  
  nR = (double) gLevels[gNLevels-1].levelTotalRefinement;
#ifdef SYM1D
  n = gLevels[0].levelNCells*nR;
#endif SYM1D
#ifdef SYM2D
  n = gLevels[0].levelNCells*nR*nR;
#endif SYM2D
#ifdef SYM3D
  n = gLevels[0].levelNCells*nR*nR*nR;
#endif SYM3D
  
  if ( show  && (gNLevels>1)) {
    zlsio_FPrintf( f, " Total   : %10.0f Cells per Level %d Cycle.\n", n0, gNLevels-1 );
    zlsio_FPrintf( f, " Total   : %10.0f Uniform Cells per Level %d Cycle.\n", n, gNLevels-1 );
    zlsio_FPrintf( f, " Total   : %10.0f Model Time Speed Up.\n", n/n0);
    zlsio_FPrintf( f, "%s\n",  DIVIDERLINE );
  }
  
  *nCells = n0;
  return(err);
  
}

int   zls_CalcAllLevelsEffectiveCellNumber ( FILE * f,  double * nCells, Integer show ){
  int err = noErr;
#ifdef REFINED_SWEEPS
  err = zls_CalcAllLevelsRefinedSweepCellNumber ( f, nCells , show );
#endif REFINED_SWEEPS
#ifdef UNIFORM_SWEEPS
  err = zls_CalcAllLevelsUniformSweepCellNumber ( f, nCells , show );
#endif UNIFORM_SWEEPS
  return(err);
}

// simpler count assuming all levels are swept the same number of time (ie UniformSweep)

int   zls_CalcAllLevelsCellNumber ( FILE * f,  double * nCells , Integer show ){
  int err = noErr;
  Integer level = 0;
  double  n     = 0.0;
  if ( show ) {
    zlsio_FPrintf( f, "%s\n",  DIVIDERLINE );
  }
  for (level = 0; level < gNLevels; level ++){
    if ( show ) {
      zlsio_FPrintf( f, " Level %d : Zones : %d\n", level,  gLevels[level].nZones);
      zlsio_FPrintf( f, " Level %d : Cells : %10.0f \n", level,  gLevels[level].levelNCells);
    }
    n += gLevels[level].levelNCells;
  } // level
  if ( show  && (gNLevels>1)) {
    zlsio_FPrintf( f, " Total   : %10.0f Cells per Cycle.\n", n );
    zlsio_FPrintf( f, "%s\n",  DIVIDERLINE );
  }
  *nCells = n;
  return(err);
}


int   zls_CalcAllLevelsZoneNumber ( FILE * f,  Integer * nZones , Integer show ){
  int err = noErr;
  Integer level = 0;
  Integer  n    = 0;
  if ( show ) {
    zlsio_FPrintf( f, "%s\n",  DIVIDERLINE );
  }
  for (level = 0; level < gNLevels; level ++){
    if ( show ) {
      zlsio_FPrintf( f, " Level %d : Zones : %d\n", level,  gLevels[level].nZones);
    }
    n += gLevels[level].nZones;
  } // level
  if ( show  && (gNLevels>1)) {
    zlsio_FPrintf( f, " Total   : %4.4d Zones per Cycle.\n", n );
    zlsio_FPrintf( f, "%s\n",  DIVIDERLINE );
  }
  *nZones = n;
  return(err);
}

// Accessers/Setters

int  zls_AddLevelZone(Integer level,  ZonePtr z){
  
  int err = noErr;
  int i = gLevels[level].nZones++;
  
  gLevels[level].levelZones[i] = z;
  gLevels[level].levelNCells += (double) z->zoneNCells;
  
  return(err);
  
}

int  zls_RemoveLevelZone(Integer level,  ZonePtr z){ 
  // extracts zone from List and close up gap
#pragma unused(level, z)
  return(nyiErr);
}


int  zls_GetLevelZoneByIndex( ZonePtr *z, Integer level, Integer i ){
  
  int err = noErr;
  
  *z =  gLevels[level].levelZones[i];
  
  return(err);
  
}

int   zls_GetLevelZoneByID(ZonePtr *z, Integer level, Integer zID ){
  
  int err = noErr;
  int i, id, n;
  
  *z = NULL;
  
  n = gLevels[level].nZones;
  
  for (i = 0; i< n; i++){
    
    id = gLevels[level].levelZones[i]->zoneID;
    if (id == zID) *z =  gLevels[level].levelZones[i];
    
  }
  
  return(err);
  
}
int  zls_SetLevelGrid( Integer level, GridCoords g){
  
  int err = noErr;
  
  gLevels[level].levelGrid = g;
  
  return(err);
}

int  zls_GetLevelGrid( GridCoords *g,  Integer level){
  
  int err = noErr;
  
  *g = gLevels[level].levelGrid;
  
  return(err);
}

int  zls_SetLevelRange( Integer level, RangePtr rSuper ){
#pragma unused(level, rSuper)
  int err = nyiErr;
  return(err);
}

int  zls_GetLevelRange( RangePtr r,  Integer level){
  
  int err = noErr;
  
  Integer i;
  Range rng;
  GridCoords gc = gLevels[level].levelGrid;
  Integer dims  = gLevels[level].levelDims;
  
  // set up range parameters for each dimension
  for (i = 0; i< dims; i++){
    zls_GetRangeFromGrid( &rng, i, (GCoordPtr)&gc);
  }
  
  *r = rng;
  
  return(err);
}

int  zls_SetLevelImageZone(Integer level,  ZonePtr z){
  
  int err = noErr;
  
  gLevels[level].levelImageZone = z;
  
  return(err);
  
}

int  zls_GetLevelImageZone( ZonePtr *z, Integer level ){
  int err = noErr;
  
  *z = gLevels[level].levelImageZone;
  
  return(err);
  
}


int  zls_SetLevelIOPrefs(Integer level,  IOSettingsPtr io ){
  
  int err = noErr;
  
  gLevels[level].levelIO = io;
  
  return(err);
  
}

int  zls_GetLevelIOPrefs( IOSettingsPtr *io, Integer level ){
  int err = noErr;
  
  *io = gLevels[level].levelIO;
  
  return(err);
  
}

int  zls_SetLevelPhysics(Integer level,  PhysicsPtr p ){
  
  int err = noErr;
  
  gLevels[level].levelPhysics = p;
  
  return(err);
  
}

int  zls_GetLevelPhysics( PhysicsPtr *p, Integer level ){
  int err = noErr;
  
  *p = gLevels[level].levelPhysics;
  
  return(err);
  
}

int  zls_SetLevelProblem(Integer level,  ProblemPtr p ){
  int err = noErr;
  
  gLevels[level].levelProblem = p;
  
  return(err);
}

int  zls_GetLevelProblem( ProblemPtr *p, Integer level ){
  int err = noErr;
  
  *p = gLevels[level].levelProblem;
  
  return(err);
}
