#ifndef ZLS_LEVELS_H
#define ZLS_LEVELS_H

#include "zls.h"

//  Level Lists  contain zones and algorithm related parameters such as physics and 
// IOSettings.  Each level has a degree of refinements, and a common CFL number and timestep
// for all zones on the level.  Generally algorithm calls deal with levels, not zones directly.
// Each level is swept, and the level clocks/timing determine the simulation progress.

typedef struct lzl { // Level Zones List
  Integer                 level; // 0 = top level, only one list per level allowed
  Integer                nZones; // min 1
  Integer              maxZones; // so you know how many ptrs can be copied.
  double             levelNCells; // cumulative total of active cells in all zones on this level, 
  // never counting BCs.  Actually a fp var to allow for gigantic
  // sims, where performance measures are needed , but exact cell
  // counts are not necessary.  zones keep precise cell counts.
  // exact for billions of cells, and more than 32 bit integers even 
  // on 32 bit machines.
  
  Integer         levelZoneKind; // descibes the type of zones in this level = kAlgorithmZoneType usually, but not necessarily
  
  TimeCoords         levelClock; // time coords for whole level
  Real                 levelCFL; // Courant-Friedrichs-Level numer for all of level
  
  GridCoords          levelGrid; // grid range of whole level, axes not always allocated however
  
  Integer             levelDims; // 1, 2, or 3 = k1D, k2D, k3D
  Integer       levelRefinement; // 2, 3, 4, 5, 6, 7, or 8
  Integer  levelTotalRefinement; // cumulative from level 0
  
  Integer           levelBCells; // boundary cells, usually 4, and n*levelRefinement, n = 1, 2,...
  Integer      levelSuperBCells; // boundary cells/levelRefinement, must be an integral divison, usually 1
  Integer    levelInterpolation; // order of interpolation kInterpNN - nearest neighbor, kInterpPLM or kInterpPPM
  
  ZonePtr        levelImageZone; // must be a zoneType ==  kImageZoneType, NULL if not used
  IOSettingsPtr         levelIO; // reference to IOSettings stucture in level 0
  
  PhysicsPtr       levelPhysics; // reference to Physics stucture in level 0, includes gravity parameters if GRAVITY is on
  ProblemPtr       levelProblem; // reference to User Problem Parameters stucture in level 0
  //
  // if the field is gravity the parameters are in the PhysicsPtr, possibly with additions informationin the ProblemPtr
  // the field arrays may be NULL if the field is faster to just compute in each cell from the parameters
  //
  ZonePtrArr         levelZones; // array of algorithm ZonePtrs in this level, allocated dynamically
  
} LevelZoneList, * LevelZoneListPtr;

//
//========================================
//
// Level Routines
//
//========================================
//
//  Creation/Destruction
//
//  Level Lists are global struct array, no new/malloc, just init member elements
//  Init Level - sets clock, refinement and CFL, and sets Zones to NULLs, and inits global IDs
//  ClearLevel - walks Zone lists and deletes them.
//
//
int   zls_InitLevelList          ( Integer level, TimeCoords cl, Real CFL, GridCoords gc,
                                  Integer ndims, Integer nbc, Integer refine, Integer interp );
int   zls_AllocateLevelZoneArray ( Integer level,  Integer n );
int   zls_ClearLevelList         ( Integer level ); 
//
// Compute the number of cell calculations required per sweep of level 0
// each level is swept refine times per super level, so low level cells are 
// computed many times per level 0 sweep
//
enum {
  kZLS_DontShowCellCount = 0,
  kZLS_ShowCellCount     = 1,
};

int   zls_CalcAllLevelsRefinedSweepEquivNumber ( double * nEqivCells );
int   zls_CalcAllLevelsUniformSweepEquivNumber ( double * nEqivCells );
int   zls_CalcAllLevelsEquivNumberCellNumber   ( double * nEqivCells );

// Compute the effective number of cells in a recursive minimum refined sweep
int   zls_CalcAllLevelsRefinedSweepCellNumber ( FILE * f,  double * nCells, Integer show );

// Compute the effective number of cells in a uniform finest step  sweep, bigger than above
int   zls_CalcAllLevelsUniformSweepCellNumber ( FILE * f,  double * nCells , Integer show );


// Generic call, selects from the above as needed

int   zls_CalcAllLevelsEffectiveCellNumber ( FILE * f,  double * nCells, Integer show );

// Compute the absolute number of cells
int   zls_CalcAllLevelsCellNumber           ( FILE * f,  double * nCells , Integer show );

// Compute the absolute number of zones
int   zls_CalcAllLevelsZoneNumber ( FILE * f,  Integer * nZones , Integer show );

// Accessers/Setters  many could be inlined if performance is an issue

int   zls_AddLevelZone        ( Integer level,   ZonePtr z );
int   zls_RemoveLevelZone     ( Integer level,   ZonePtr z );  // extracts zone from List and close up gap, z still needs disposing
int   zls_GetLevelZoneByIndex ( ZonePtr *z,  Integer level, Integer i );
int   zls_GetLevelZoneByID    ( ZonePtr *z,  Integer level, Integer zID );



int   zls_SetLevelImageZone   ( Integer level,   ZonePtr z );
int   zls_GetLevelImageZone   ( ZonePtr *z, Integer level );

int  zls_SetLevelIOPrefs      ( Integer level,  IOSettingsPtr io );
int  zls_GetLevelIOPrefs      ( IOSettingsPtr *io, Integer level );


int  zls_SetLevelGrid         ( Integer level, GridCoords g);
int  zls_GetLevelGrid         ( GridCoords *g,  Integer level);

int  zls_SetLevelRange( Integer level, RangePtr rSuper ); // NYI
int  zls_GetLevelRange        ( RangePtr r,  Integer level); //acessor only, Set is too fiddly right now and the setup in conf is OK for now

//

int  zls_SetLevelIOPrefs( Integer level,  IOSettingsPtr io );
int  zls_GetLevelIOPrefs( IOSettingsPtr *io, Integer level );

int  zls_SetLevelPhysics( Integer level,  PhysicsPtr p );
int  zls_GetLevelPhysics( PhysicsPtr *p, Integer level );

int  zls_SetLevelProblem( Integer level,  ProblemPtr p );
int  zls_GetLevelProblem( ProblemPtr *p, Integer level );

//
// global list head for each level, to access zones across levels not via owner/sub tree
//
extern  Integer                 gNLevels; 
extern  LevelZoneList gLevels[MAXLEVELS];


#endif
