#include "zls.h"
//==============================================================================
//
// Plain, uniterpreted quantity arrays
//
// QPtr, D1DArr, Q2DArr, Q3DArr = Real 1, 2 and 3D
// blocks or multi-D arrays.
//
//==============================================================================
//==============================================================================
//
// Declared as Q * 1DArr, Q ** 2DArr and Q *** 3DArr, may accessed as
// Q[i], Q[j][i] and Q[k][j][i] or at single ptr from Q, Q[0] or
// Q[0][0] in 1D , 2D and 3D.
//
// zls_qs contains utility routines for allocating and deacllocating
// these arrays dynamically
//==============================================================================
//
//  Allocating/Releasing
//
//
// Basic Quantities - stored in zones, just an array
// and each element is the same size in all dimenstions.
// Convenience 2D and 3D allocators, take j, and k ranges
//
// Floating point Real quantities: QPtr == Q1DArr, Q2DArr, Q3DArr
// Integer quantities: IQPtr == IQ1DArr, IQ2DArr, IQ3DArr
// Complex quantities: CQPtr == CQ1DArr, CQ2DArr, CQ3DArr
//
//  Generally: QPtr variations are simply a pointer to a malloc'd
//  block, ie 
//  QPtr   q, *q is a pointer to a block of Reals.
//  IQPtr iq, *iq is a pointer to a block of Integers.
//  CQPtr cq, *cq is a pointer to a block of Complex values.
//  Can be addressed with an array operator as well, ie q[n]
//  1DArr variations are in fact the same as QPtrs, useages is that 
//  QPtrs are for when block IO is being done or some other block 
//  operation, 1DArr variants are for accessing elements of a 1D array.
//
//  2DArr versions are a single block plus a row offset block allocted
//  in such a way that : Q2DArr q2;  q[j][i] accesses the ith element of 
//  row j in a 2D array.  q2[0] is the pointer the main block, the row offset
//  block is 1D.  3DArrs are the same for 3D, but have two auxilliary blocks,
//  for row and column offsets,ie: q3[k][j[i], and q3[0][0] is the pointer to the main 
//  block.  There are accssor routines to get the QPtr block from any array type,
//  and a Set/Release mechanism to take a raw QPtr, wrap offset arrays around it
//  so it can be accessed as a Q2DArr or Q3DArr, and then have the offset
//  arrays released to leave just the processed QPtr bare and ready for IO.
//  see zls_arrs for more detail.
//
//======================================================================
//
//  Real quantities: Real may be single or double precision
//
//======================================================================

int zls_NewQuantity  (QPtr *q, RangePtr r){
  
  int err = noErr;
  Counter n = 0;
  Counter ni =  r->rn[kXDim];
  Counter nj =  r->rn[kYDim];
  Counter nk =  r->rn[kZDim];
  
  if (r!=NULL){
    
    switch (r->nD){
      case 3:
        n = ni*nj*nk;
        break;
      case 2:
        n = ni*nj;
        break;
      case 1:
        n = ni;
        break;
      default:
        err = dimErr;
        return(err);
    }
    
    if (n>0) {
      *q = (QPtr)malloc(n*sizeof(Real));
    }
    
    if (*q == NULL) {
      err = memErr;
    }
    
  } else {err = memErr;}
  
  return(err);
}


int zls_NewQuantity1D(QPtr *q, Counter ni){
  
  int err = noErr;
  
  *q = (QPtr)malloc(sizeof(Real)*ni);
  
  if (*q == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewQuantity2D(QPtr *q, Counter ni, Counter nj){
  
  int err = noErr;
  Size n =  (ni*nj);
  
  *q = (QPtr)malloc(n*sizeof(Real));
  
  if (*q == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewQuantity3D(QPtr *q, Counter ni, Counter nj, Counter nk){
  int err = noErr;
  Size n =  (ni*nj*nk);
  *q = (QPtr)malloc(n*sizeof(Real));
  
  if (*q == NULL) {
    err = memErr;
  }
  return(err);
}


int zls_DisposeQuantity(QPtr q){
  int err = noErr;
  if (q != NULL) {
    free(q);
  }
  return(err);                 
}

//========================================
//
// multi-dimensional array versions...
//
// Note reversed idices, reference arrays by [i], [j][i] or  [k][j][i]
//
//========================================

int zls_NewQuantity1DArr(Q1DArr *q, Counter ni){ // same as NewQuantity but for Q1DArr syntax
  
  int err = noErr;
  
  *q = (Q1DArr)malloc(sizeof(Real) * ni);
  
  if (*q == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewQuantity2DArr(Q2DArr *q, Counter nj, Counter ni){
  
  int err = noErr;
  
  *q = (Q2DArr)zlsAlloc2DArr(sizeof(Real), nj, ni);
  
  if (*q == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewQuantity3DArr(Q3DArr *q, Counter nk, Counter nj, Counter ni){
  
  int err = noErr;
  
  *q = (Q3DArr)zlsAlloc3DArr(sizeof(Real), nk, nj, ni);
  
  if (*q == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_DisposeQuantity1DArr(Q1DArr q){
  
  int err = noErr;
  
  if (q != NULL) {
    free((void *) q);
  } else {err = memErr;}
  
  return(err);
}

int zls_DisposeQuantity2DArr(Q2DArr q){
  
  int err = noErr;
  
  if (q != NULL) {
    zlsFree2DArr((void **) q);
  } else {err = memErr;}
  
  return(err);
}

int zls_DisposeQuantity3DArr(Q3DArr q){
  
  int err = noErr;
  
  if (q != NULL) {
    zlsFree3DArr((void ***) q);
  } else {err = memErr;}
  
  return(err);
}

int zls_DisposeQuantityArr(QArr q){
  
  int err = noErr;
  
  if (q != NULL) {
#ifdef SYM1D
    free((void *) q);
#endif SYM1D
#ifdef SYM2D
    zlsFree2DArr((void **) q);
#endif SYM2D
#ifdef SYM3D
    zlsFree3DArr((void ***) q);
#endif SYM3D
  } else {err = memErr;}
  
  return(err);
}

//========================================
//
// Wrap an array around an unstructured block, alternative to New, 
// if data already exists Release disposes without affecting data, 
// not needed for 1D data
//
//========================================

int zls_SetQuantity2DArr(Q2DArr *q, QPtr data, Counter nj, Counter ni){
  
  int err = noErr;
  Q2DArr  tempQ = NULL;
  
  tempQ = (Q2DArr)zlsSet2DArr((void *) data, sizeof(Real), nj, ni);
  
  if (tempQ == NULL) {
    err = memErr;
  }
  *q = tempQ;
  
  return(err);
}

int zls_SetQuantity3DArr(Q3DArr *q, QPtr data, Counter nk, Counter nj, Counter ni){
  
  int err = noErr;
  
  *q = (Q3DArr)zlsSet3DArr((void *) data, sizeof(Real), nk, nj, ni);
  
  if (*q == NULL) {
    err = memErr;
  }
  
  return(err);
}

//========================================
//
// remove offset pointer structures from a multi-dim array leaving 
// the data untouched must have a copy of the data reference elsewhere 
// or it will be lost and memory will leak badly. 
//
//========================================

int zls_ReleaseQuantity2DArr(Q2DArr q){
  
  int err = noErr;
  
  if (q != NULL) {
    zlsRelease2DArr((void **) q);
  } else { err = memErr;}
  
  return(err);
}

int zls_ReleaseQuantity3DArr(Q3DArr q){
  
  int err = noErr;
  
  if (q != NULL) {
    zlsRelease3DArr((void ***) q);
  } else { err = memErr;}
  
  return(err);
}

//========================================
// Extract an unstructured block, useful for I/O, returns a reference
// to the data, no extra memory is allocated
//========================================

int zls_GetQuantity2DArr(QPtr *data, Q2DArr q){
  *data = q[0];
  return(noErr);
}

int zls_GetQuantity3DArr(QPtr *data, Q3DArr q){
  *data = q[0][0];
  return(noErr);
}

//========================================
// Copy lower dim quantity arrays from higher dim quantities
//========================================

int zls_SetQ1DArrQ2DArrY(Q1DArr q1, Q2DArr q, Integer nj, Integer i){
  int err = noErr;
  Integer     j;
  if ((q1!=NULL) &&(q!=NULL)){
    for (j = 0; j < nj; j++){
      q1[j]     = q[j][i];
    }
  } else {err = memErr; } 
  return(err);
}


int zls_SetQ1DArrQ2DArrX(Q1DArr q1, Q2DArr q, Integer ni, Integer j){
  int err = noErr;
  Integer     i;
  if ((q1!=NULL) &&(q!=NULL)){
    for (i = 0; i < ni; i++){
      q1[i]     = q[j][i];      
    }
  } else {err = memErr; } 
  return(err);
}

int zls_SetQ1DArrQ3DArrX( Q1DArr q1, Q3DArr q, Integer ni, Integer k, Integer j ){
  int err = noErr;
  Integer     i;
  if ((q1!=NULL) &&(q!=NULL)){
    for (i = 0; i<ni; i++){
      q1[i] = q[k][j][i];
    }
  } else {err = memErr; } 
  return(err);
}

int zls_SetQ1DArrQ3DArrY( Q1DArr q1, Q3DArr q, Integer nj, Integer k, Integer i ){
  int err = noErr;
  Integer     j;
  if ((q1!=NULL) &&(q!=NULL)){
    for (j = 0; j<nj; j++){
      q1[j] = q[k][j][i];
    }
  } else {err = memErr; } 
  return(err);
}

int zls_SetQ1DArrQ3DArrZ( Q1DArr q1, Q3DArr q, Integer nk, Integer j, Integer i ){
  int err = noErr;
  Integer     k;
  if ((q1!=NULL) &&(q!=NULL)){
    for (k = 0; k<nk; k++){
      q1[k] = q[k][j][i];
    }
  } else {err = memErr; } 
  return(err);
}

int zls_SetQ2DArrQ3DArrYZ( Q2DArr q2, Q3DArr q, Integer nj, Integer nk, Integer i ){
  int err = noErr;
  Integer    j,k;
  if ((q2!=NULL) &&(q!=NULL)){
    for (k = 0; k<nk; k++){
      for (j = 0; j<nj; j++){
        q2[k][j] = q[k][j][i];
      }
    }
  } else {err = memErr; } 
  return(err);
}

int zls_SetQ2DArrQ3DArrZX( Q2DArr q2, Q3DArr q, Integer nk, Integer ni, Integer j ){
  int err = noErr;
  Integer     i, k;
  if ((q2!=NULL) &&(q!=NULL)){
    for (k = 0; k<nk; k++){
      for (i = 0; i<ni; i++){
        q2[i][k] = q[k][j][i];
      }
    }
  } else {err = memErr; } 
  return(err);
}

int zls_SetQ2DArrQ3DArrXY( Q2DArr q2, Q3DArr q, Integer ni, Integer nj, Integer k ){
  int err = noErr;
  Integer    i,j;
  if ((q2!=NULL) &&(q!=NULL)){
    for (j = 0; j<nj; j++){
      for (i = 0; i<ni; i++){
        q2[j][i] = q[k][j][i];
      }
    }
  } else {err = memErr; } 
  return(err);
}

int zls_NewIQuantity  (IQPtr *i, RangePtr r){
  
  int err = noErr;
  Counter n = 0;
  Counter ni =  r->rn[kXDim];
  Counter nj =  r->rn[kYDim];
  Counter nk =  r->rn[kZDim];
  
  if (r!=NULL){
    
    switch (r->nD){
      case 3:
        n = ni*nj*nk;
        break;
      case 2:
        n = ni*nj;
        break;
      case 1:
        n = ni;
        break;
      default:
        err = dimErr;
        return(err);
    }
    
    if (n>0) {
      *i = (IQPtr)malloc(n*sizeof(Integer));
    }
    
    if (*i == NULL) {
      err = memErr;
    }
    
  } else {err = memErr;}
  
  return(err);
}


int zls_NewIQuantity1D(IQPtr *i, Counter ni){
  
  int err = noErr;
  Counter n =  ni;
  
  *i = (IQPtr)malloc(n*sizeof(Integer));
  
  if (*i == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewIQuantity2D(IQPtr *i, Counter ni, Counter nj){
  
  int err = noErr;
  Counter n =  (ni*nj);
  
  *i = (IQPtr)malloc(n*sizeof(Integer));
  
  if (*i == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewIQuantity3D(IQPtr *i, Counter ni, Counter nj, Counter nk){
  int err = noErr;
  Counter n =  (ni*nj*nk);
  *i = (IQPtr)malloc(n*sizeof(Integer));
  
  if (*i == NULL) {
    err = memErr;
  }
  return(err);
}


int zls_DisposeIQuantity(IQPtr i){
  int err = noErr;
  if (i != NULL) {
    free(i);
  }
  return(err);                 
}

//========================================
//
// multi-dimensional array versions...
//
// Note reversed idices, reference arrays by [i], [j][i] or  [k][j][i]
//
//========================================

int zls_NewIQuantity2DArr(IQ2DArr *i, Counter nj, Counter ni){
  
  int err = noErr;
  
  *i = (IQ2DArr)zlsAlloc2DArr(sizeof(Integer), nj, ni);
  
  if (*i == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewIQuantity3DArr(IQ3DArr *i, Counter nk, Counter nj, Counter ni){
  
  int err = noErr;
  
  *i = (IQ3DArr)zlsAlloc3DArr(sizeof(Integer), nk, nj, ni);
  
  if (*i == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_DisposeIQuantity2DArr(IQ2DArr i){
  
  int err = noErr;
  
  if (i != NULL) {
    zlsFree2DArr((void **) i);
  } else {err = memErr;}
  
  return(err);
}

int zls_DisposeIQuantity3DArr(IQ3DArr i){
  
  int err = noErr;
  
  if (i != NULL) {
    zlsFree3DArr((void ***) i);
  } else {err = memErr;}
  
  return(err);
}

//========================================
//
// Wrap an array around an unstructured block, alternative to New, 
// if data already exists Release disposes without affecting data, 
// not needed for 1D data
//
//========================================

int zls_SetIQuantity2DArr(IQ2DArr *i, IQPtr data, Counter nj, Counter ni){
  
  int err = noErr;
  IQ2DArr  tempIQ = NULL;
  
  tempIQ = (IQ2DArr)zlsSet2DArr((void *) data, sizeof(Integer), nj, ni);
  
  if (tempIQ == NULL) {
    err = memErr;
  }
  *i = tempIQ;
  
  return(err);
}

int zls_SetIQuantity3DArr(IQ3DArr *i, IQPtr data, Counter nk, Counter nj, Counter ni){
  
  int err = noErr;
  
  *i = (IQ3DArr)zlsSet3DArr((void *) data, sizeof(Integer), nk, nj, ni);
  
  if (*i == NULL) {
    err = memErr;
  }
  
  return(err);
}

//========================================
//
// remove offset pointer structures from a multi-dim array leaving 
// the data untouched must have a copy of the data reference elsewhere 
// or it will be lost and memory will leak badly. 
//
//========================================

int zls_ReleaseIQuantity2DArr(IQ2DArr i){
  
  int err = noErr;
  
  if (i != NULL) {
    zlsRelease2DArr((void **) i);
  } else { err = memErr;}
  
  return(err);
}

int zls_ReleaseIQuantity3DArr(IQ3DArr i){
  
  int err = noErr;
  
  if (i != NULL) {
    zlsRelease3DArr((void ***) i);
  } else { err = memErr;}
  
  return(err);
}

//========================================
// Extract an unstructured block, useful for I/O, returns a reference
// to the data, no extra memory is allocated
//========================================

int zls_GetIQuantity2DArr(IQPtr *data, IQ2DArr i){
  *data = i[0];
  return(noErr);
}

int zls_GetIQuantity3DArr(IQPtr *data, IQ3DArr i){
  *data = i[0][0];
  return(noErr);
}


//
// Complex Quantities - stored in zones, just an array
// and each element is the same size in all dimenstions.
// Convenience 2D and 3D allocators, take j, and k ranges
//
//========================================
//
// Single complex quantities:
//
//========================================

int zls_NewCQuantity1D(CQPtr *cq, Counter ni){
  
  int err = noErr;
  
  *cq = (CQPtr)malloc(sizeof(Complex)*ni);
  
  if (*cq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewCQuantity2D(CQPtr *cq, Counter ni, Counter nj){
  
  int err = noErr;
  Size n =  (ni*nj);
  
  *cq = (CQPtr)malloc(n*sizeof(Complex));
  
  if (*cq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewCQuantity3D(CQPtr *cq, Counter ni, Counter nj, Counter nk){
  int err = noErr;
  Size n =  (ni*nj*nk);
  *cq = (CQPtr)malloc(n*sizeof(Complex));
  
  if (*cq == NULL) {
    err = memErr;
  }
  return(err);
}


int zls_DisposeCQuantity(CQPtr cq){
  int err = noErr;
  if (cq != NULL) {
    free(cq);
  }
  return(err);                 
}

//========================================
//
// multi-dimensional array versions...
//
// Note reversed idices, reference arrays by [i], [j][i] or  [k][j][i]
//
//========================================

int zls_NewCQuantity1DArr(CQ1DArr *cq, Counter ni){ // same as NewCQuantity but for CQ1DArr syntax
  
  int err = noErr;
  
  *cq = (CQ1DArr)malloc(sizeof(Complex) * ni);
  
  if (*cq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewCQuantity2DArr(CQ2DArr *cq, Counter nj, Counter ni){
  
  int err = noErr;
  
  *cq = (CQ2DArr)zlsAlloc2DArr(sizeof(Complex), nj, ni);
  
  if (*cq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewCQuantity3DArr(CQ3DArr *cq, Counter nk, Counter nj, Counter ni){
  
  int err = noErr;
  
  *cq = (CQ3DArr)zlsAlloc3DArr(sizeof(Complex), nk, nj, ni);
  
  if (*cq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_DisposeCQuantity1DArr(CQ1DArr cq){
  
  int err = noErr;
  
  if (cq != NULL) {
    free((void *) cq);
  } else {err = memErr;}
  
  return(err);
}

int zls_DisposeCQuantity2DArr(CQ2DArr cq){
  
  int err = noErr;
  
  if (cq != NULL) {
    zlsFree2DArr((void **) cq);
  } else {err = memErr;}
  
  return(err);
}

int zls_DisposeCQuantity3DArr(CQ3DArr cq){
  
  int err = noErr;
  
  if (cq != NULL) {
    zlsFree3DArr((void ***) cq);
  } else {err = memErr;}
  
  return(err);
}

//========================================
//
// Dimension limited Vector Quantities - stored in zones, 
//
//========================================
//
// Single block vector quantities:
//
//========================================

int zls_NewVQuantity1D(VQPtr *vq, Counter ni){
  
  int err = noErr;
  
  *vq = (VQPtr)malloc(sizeof(QVec)*ni);
  
  if (*vq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewVQuantity2D(VQPtr *vq, Counter ni, Counter nj){
  
  int err = noErr;
  Size n =  (ni*nj);
  
  *vq = (VQPtr)malloc(n*sizeof(QVec));
  
  if (*vq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewVQuantity3D(VQPtr *vq, Counter ni, Counter nj, Counter nk){
  int err = noErr;
  Size n =  (ni*nj*nk);
  *vq = (VQPtr)malloc(n*sizeof(QVec));
  
  if (*vq == NULL) {
    err = memErr;
  }
  return(err);
}


int zls_DisposeVQuantity(VQPtr vq){
  int err = noErr;
  if (vq != NULL) {
    free(vq);
  }
  return(err);                 
}

//========================================
//
// multi-dimensional array versions...
//
// Note reversed idices, reference arrays by [i], [j][i] or  [k][j][i]
//
//========================================

int zls_NewVQuantity1DArr(VQ1DArr *vq, Counter ni){ // same as NewCQuantity but for CQ1DArr syntax
  
  int err = noErr;
  
  *vq = (VQ1DArr)malloc(sizeof(QVec) * ni);
  
  if (*vq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewVQuantity2DArr(VQ2DArr *vq, Counter nj, Counter ni){
  
  int err = noErr;
  
  *vq = (VQ2DArr)zlsAlloc2DArr(sizeof(QVec), nj, ni);
  
  if (*vq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewVQuantity3DArr(VQ3DArr *vq, Counter nk, Counter nj, Counter ni){
  
  int err = noErr;
  
  *vq = (VQ3DArr)zlsAlloc3DArr(sizeof(QVec), nk, nj, ni);
  
  if (*vq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_DisposeVQuantity1DArr(VQ1DArr vq){
  
  int err = noErr;
  
  if (vq != NULL) {
    free((void *) vq);
  } else {err = memErr;}
  
  return(err);
}

int zls_DisposeVQuantity2DArr(VQ2DArr vq){
  
  int err = noErr;
  
  if (vq != NULL) {
    zlsFree2DArr((void **) vq);
  } else {err = memErr;}
  
  return(err);
}

int zls_DisposeVQuantity3DArr(VQ3DArr vq){
  
  int err = noErr;
  
  if (vq != NULL) {
    zlsFree3DArr((void ***) vq);
  } else {err = memErr;}
  
  return(err);
}

//========================================
//
// Wrap an array around an unstructured block, alternative to New, 
// if data already exists Release disposes without affecting data, 
// not needed for 1D data
//
//========================================

int zls_SetCQuantity2DArr(CQ2DArr *cq, CQPtr cdata, Counter nj, Counter ni){
  
  int err = noErr;
  CQ2DArr  tempQ = NULL;
  
  tempQ = (CQ2DArr)zlsSet2DArr((void *) cdata, sizeof(Complex), nj, ni);
  
  if (tempQ == NULL) {
    err = memErr;
  }
  *cq = tempQ;
  
  return(err);
}

int zls_SetCQuantity3DArr(CQ3DArr *cq, CQPtr cdata, Counter nk, Counter nj, Counter ni){
  
  int err = noErr;
  
  *cq = (CQ3DArr)zlsSet3DArr((void *) cdata, sizeof(Complex), nk, nj, ni);
  
  if (*cq == NULL) {
    err = memErr;
  }
  
  return(err);
}

//========================================
//
// remove offset pointer structures from a multi-dim array leaving 
// the cdata untouched must have a copy of the cdata reference elsewhere 
// or it will be lost and memory will leak badly. 
//
//========================================

int zls_ReleaseCQuantity2DArr(CQ2DArr cq){
  
  int err = noErr;
  
  if (cq != NULL) {
    zlsRelease2DArr((void **) cq);
  } else { err = memErr;}
  
  return(err);
}

int zls_ReleaseCQuantity3DArr(CQ3DArr cq){
  
  int err = noErr;
  
  if (cq != NULL) {
    zlsRelease3DArr((void ***) cq);
  } else { err = memErr;}
  
  return(err);
}

//========================================
// Extract an unstructured block, useful for I/O, returns a reference
// to the cdata, no extra memory is allocated
//========================================

int zls_GetCQuantity2DArr(CQPtr *cdata, CQ2DArr cq){
  *cdata = cq[0];
  return(noErr);
}

int zls_GetCQuantity3DArr(CQPtr *cdata, CQ3DArr cq){
  *cdata = cq[0][0];
  return(noErr);
}

//========================================
// Copy lower dim quantity arrays from higher dim quantities
//========================================

int zls_SetCQ1DArrCQ2DArrY(CQ1DArr cq1, CQ2DArr cq, Integer nj, Integer i){
  int err = noErr;
  Integer     j;
  if ((cq1!=NULL) &&(cq!=NULL)){
    for (j = 0; j < nj; j++){
      cq1[j]     = cq[j][i];
    }
  } else {err = memErr; } 
  return(err);
}


int zls_SetCQ1DArrCQ2DArrX(CQ1DArr cq1, CQ2DArr cq, Integer ni, Integer j){
  int err = noErr;
  Integer     i;
  if ((cq1!=NULL) &&(cq!=NULL)){
    for (i = 0; i < ni; i++){
      cq1[i]     = cq[j][i];      
    }
  } else {err = memErr; } 
  return(err);
}

int zls_SetCQ1DArrCQ3DArrX( CQ1DArr cq1, CQ3DArr cq, Integer ni, Integer k, Integer j ){
  int err = noErr;
  Integer     i;
  if ((cq1!=NULL) &&(cq!=NULL)){
    for (i = 0; i<ni; i++){
      cq1[i] = cq[k][j][i];
    }
  } else {err = memErr; } 
  return(err);
}

int zls_SetCQ1DArrCQ3DArrY( CQ1DArr cq1, CQ3DArr cq, Integer nj, Integer k, Integer i ){
  int err = noErr;
  Integer     j;
  if ((cq1!=NULL) &&(cq!=NULL)){
    for (j = 0; j<nj; j++){
      cq1[j] = cq[k][j][i];
    }
  } else {err = memErr; } 
  return(err);
}

int zls_SetCQ1DArrCQ3DArrZ( CQ1DArr cq1, CQ3DArr cq, Integer nk, Integer j, Integer i ){
  int err = noErr;
  Integer     k;
  if ((cq1!=NULL) &&(cq!=NULL)){
    for (k = 0; k<nk; k++){
      cq1[k] = cq[k][j][i];
    }
  } else {err = memErr; } 
  return(err);
}

int zls_SetCQ2DArrCQ3DArrYZ( CQ2DArr cq2, CQ3DArr cq, Integer nj, Integer nk, Integer i ){
  int err = noErr;
  Integer    j,k;
  if ((cq2!=NULL) &&(cq!=NULL)){
    for (k = 0; k<nk; k++){
      for (j = 0; j<nj; j++){
        cq2[k][j] = cq[k][j][i];
      }
    }
  } else {err = memErr; } 
  return(err);
}

int zls_SetCQ2DArrCQ3DArrZX( CQ2DArr cq2, CQ3DArr cq, Integer nk, Integer ni, Integer j ){
  int err = noErr;
  Integer     i, k;
  if ((cq2!=NULL) &&(cq!=NULL)){
    for (k = 0; k<nk; k++){
      for (i = 0; i<ni; i++){
        cq2[i][k] = cq[k][j][i];
      }
    }
  } else {err = memErr; } 
  return(err);
}

int zls_SetCQ2DArrCQ3DArrXY( CQ2DArr cq2, CQ3DArr cq, Integer ni, Integer nj, Integer k ){
  int err = noErr;
  Integer    i,j;
  if ((cq2!=NULL) &&(cq!=NULL)){
    for (j = 0; j<nj; j++){
      for (i = 0; i<ni; i++){
        cq2[j][i] = cq[k][j][i];
      }
    }
  } else {err = memErr; } 
  return(err);
}

//==============================================================================
//
// Some specific format quantities, intended for IO on specific file formats
// 
//==============================================================================
// 32 bit float quantity/arrays, always 32 bit float independent of whether
// Real is single or double precision
//==============================================================================

int zls_NewFQuantity  (FQPtr *fq, RangePtr r){
  
  int err = noErr;
  Counter n = 0;
  Counter ni =  r->rn[kXDim];
  Counter nj =  r->rn[kYDim];
  Counter nk =  r->rn[kZDim];
  
  if (r!=NULL){
    
    switch (r->nD){
      case 3:
        n = ni*nj*nk;
        break;
      case 2:
        n = ni*nj;
        break;
      case 1:
        n = ni;
        break;
      default:
        err = dimErr;
        return(err);
    }
    
    if (n>0) {
      *fq = (FQPtr)malloc(n*sizeof(float));
    }
    
    if (*fq == NULL) {
      err = memErr;
    }
    
  } else {err = memErr;}
  
  return(err);
}


int zls_NewFQuantity1D(FQPtr *fq, Counter ni){
  
  int err = noErr;
  
  *fq = (FQPtr)malloc(sizeof(float)*ni);
  
  if (*fq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewFQuantity2D(FQPtr *fq, Counter ni, Counter nj){
  
  int err = noErr;
  Size n =  (ni*nj);
  
  *fq = (FQPtr)malloc(n*sizeof(float));
  
  if (*fq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewFQuantity3D(FQPtr *fq, Counter ni, Counter nj, Counter nk){
  int err = noErr;
  Size n =  (ni*nj*nk);
  *fq = (FQPtr)malloc(n*sizeof(float));
  
  if (*fq == NULL) {
    err = memErr;
  }
  return(err);
}


int zls_DisposeFQuantity(FQPtr fq){
  int err = noErr;
  if (fq != NULL) {
    free(fq);
  }
  return(err);                 
}

//========================================
//
// multi-dimensional array versions...
//
// Note reversed idices, reference arrays by [i], [j][i] or  [k][j][i]
//
//========================================

int zls_NewFQuantity1DArr(FQ1DArr *fq, Counter ni){ // same as NewFQuantity but for FQ1DArr syntax
  
  int err = noErr;
  
  *fq = (FQ1DArr)malloc(sizeof(float) * ni);
  
  if (*fq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewFQuantity2DArr(FQ2DArr *fq, Counter nj, Counter ni){
  
  int err = noErr;
  
  *fq = (FQ2DArr)zlsAlloc2DArr(sizeof(float), nj, ni);
  
  if (*fq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewFQuantity3DArr(FQ3DArr *fq, Counter nk, Counter nj, Counter ni){
  
  int err = noErr;
  
  *fq = (FQ3DArr)zlsAlloc3DArr(sizeof(float), nk, nj, ni);
  
  if (*fq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_DisposeFQuantity1DArr(FQ1DArr fq){
  
  int err = noErr;
  
  if (fq != NULL) {
    free((void *) fq);
  } else {err = memErr;}
  
  return(err);
}

int zls_DisposeFQuantity2DArr(FQ2DArr fq){
  
  int err = noErr;
  
  if (fq != NULL) {
    zlsFree2DArr((void **) fq);
  } else {err = memErr;}
  
  return(err);
}

int zls_DisposeFQuantity3DArr(FQ3DArr fq){
  
  int err = noErr;
  
  if (fq != NULL) {
    zlsFree3DArr((void ***) fq);
  } else {err = memErr;}
  
  return(err);
}

int zls_DisposeFQuantityArr(FQArr fq){
  
  int err = noErr;
  
  if (fq != NULL) {
#ifdef SYM1D
    free((void *) fq);
#endif SYM1D
#ifdef SYM2D
    zlsFree2DArr((void **) fq);
#endif SYM2D
#ifdef SYM3D
    zlsFree3DArr((void ***) fq);
#endif SYM3D
  } else {err = memErr;}
  
  return(err);
}

//========================================
//
// Wrap an array around an unstructured block, alternative to New, 
// if data already exists Release disposes without affecting data, 
// not needed for 1D data
//
//========================================

int zls_SetFQuantity2DArr(FQ2DArr *fq, FQPtr data, Counter nj, Counter ni){
  
  int err = noErr;
  FQ2DArr  tempFQ = NULL;
  
  tempFQ = (FQ2DArr)zlsSet2DArr((void *) data, sizeof(float), nj, ni);
  
  if (tempFQ == NULL) {
    err = memErr;
  }
  *fq = tempFQ;
  
  return(err);
}

int zls_SetFQuantity3DArr(FQ3DArr *fq, FQPtr data, Counter nk, Counter nj, Counter ni){
  
  int err = noErr;
  
  *fq = (FQ3DArr)zlsSet3DArr((void *) data, sizeof(float), nk, nj, ni);
  
  if (*fq == NULL) {
    err = memErr;
  }
  
  return(err);
}

//========================================
//
// remove offset pointer structures from a multi-dim array leaving 
// the data untouched must have a copy of the data reference elsewhere 
// or it will be lost and memory will leak badly. 
//
//========================================

int zls_ReleaseFQuantity2DArr(FQ2DArr fq){
  
  int err = noErr;
  
  if (fq != NULL) {
    zlsRelease2DArr((void **) fq);
  } else { err = memErr;}
  
  return(err);
}

int zls_ReleaseFQuantity3DArr(FQ3DArr fq){
  
  int err = noErr;
  
  if (fq != NULL) {
    zlsRelease3DArr((void ***) fq);
  } else { err = memErr;}
  
  return(err);
}

//========================================
// Extract an unstructured block, useful for I/O, returns a reference
// to the data, no extra memory is allocated
//========================================

int zls_GetFQuantity2DArr(FQPtr *data, FQ2DArr fq){
  *data = fq[0];
  return(noErr);
}

int zls_GetFQuantity3DArr(FQPtr *data, FQ3DArr fq){
  *data = fq[0][0];
  return(noErr);
}


// 
//==============================================================================
// 64 bit double quantity/arrays, always 64 bit double independent of whether
// Real is single or double precision
//==============================================================================
//

int zls_NewDQuantity  (DQPtr *dq, RangePtr r){
  
  int err = noErr;
  Counter n = 0;
  Counter ni =  r->rn[kXDim];
  Counter nj =  r->rn[kYDim];
  Counter nk =  r->rn[kZDim];
  
  if (r!=NULL){
    
    switch (r->nD){
      case 3:
        n = ni*nj*nk;
        break;
      case 2:
        n = ni*nj;
        break;
      case 1:
        n = ni;
        break;
      default:
        err = dimErr;
        return(err);
    }
    
    if (n>0) {
      *dq = (DQPtr)malloc(n*sizeof(double));
    }
    
    if (*dq == NULL) {
      err = memErr;
    }
    
  } else {err = memErr;}
  
  return(err);
}


int zls_NewDQuantity1D(DQPtr *dq, Counter ni){
  
  int err = noErr;
  
  *dq = (DQPtr)malloc(sizeof(double)*ni);
  
  if (*dq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewDQuantity2D(DQPtr *dq, Counter ni, Counter nj){
  
  int err = noErr;
  Size n =  (ni*nj);
  
  *dq = (DQPtr)malloc(n*sizeof(double));
  
  if (*dq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewDQuantity3D(DQPtr *dq, Counter ni, Counter nj, Counter nk){
  int err = noErr;
  Size n =  (ni*nj*nk);
  *dq = (DQPtr)malloc(n*sizeof(double));
  
  if (*dq == NULL) {
    err = memErr;
  }
  return(err);
}


int zls_DisposeDQuantity(DQPtr dq){
  int err = noErr;
  if (dq != NULL) {
    free(dq);
  }
  return(err);                 
}

//========================================
//
// multi-dimensional array versions...
//
// Note reversed idices, reference arrays by [i], [j][i] or  [k][j][i]
//
//========================================

int zls_NewDQuantity1DArr(DQ1DArr *dq, Counter ni){ // same as NewDQuantity but for DQ1DArr syntax
  
  int err = noErr;
  
  *dq = (DQ1DArr)malloc(sizeof(double) * ni);
  
  if (*dq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewDQuantity2DArr(DQ2DArr *dq, Counter nj, Counter ni){
  
  int err = noErr;
  
  *dq = (DQ2DArr)zlsAlloc2DArr(sizeof(double), nj, ni);
  
  if (*dq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_NewDQuantity3DArr(DQ3DArr *dq, Counter nk, Counter nj, Counter ni){
  
  int err = noErr;
  
  *dq = (DQ3DArr)zlsAlloc3DArr(sizeof(double), nk, nj, ni);
  
  if (*dq == NULL) {
    err = memErr;
  }
  
  return(err);
}

int zls_DisposeDQuantity1DArr(DQ1DArr dq){
  
  int err = noErr;
  
  if (dq != NULL) {
    free((void *) dq);
  } else {err = memErr;}
  
  return(err);
}

int zls_DisposeDQuantity2DArr(DQ2DArr dq){
  
  int err = noErr;
  
  if (dq != NULL) {
    zlsFree2DArr((void **) dq);
  } else {err = memErr;}
  
  return(err);
}

int zls_DisposeDQuantity3DArr(DQ3DArr dq){
  
  int err = noErr;
  
  if (dq != NULL) {
    zlsFree3DArr((void ***) dq);
  } else {err = memErr;}
  
  return(err);
}

int zls_DisposeDQuantityArr(DQArr dq){
  
  int err = noErr;
  
  if (dq != NULL) {
#ifdef SYM1D
    free((void *) dq);
#endif SYM1D
#ifdef SYM2D
    zlsFree2DArr((void **) dq);
#endif SYM2D
#ifdef SYM3D
    zlsFree3DArr((void ***) dq);
#endif SYM3D
  } else {err = memErr;}
  
  return(err);
}

//========================================
//
// Wrap an array around an unstructured block, alternative to New, 
// if data already exists Release disposes without affecting data, 
// not needed for 1D data
//
//========================================

int zls_SetDQuantity2DArr(DQ2DArr *dq, DQPtr data, Counter nj, Counter ni){
  
  int err = noErr;
  DQ2DArr  tempDQ = NULL;
  
  tempDQ = (DQ2DArr)zlsSet2DArr((void *) data, sizeof(double), nj, ni);
  
  if (tempDQ == NULL) {
    err = memErr;
  }
  *dq = tempDQ;
  
  return(err);
}

int zls_SetDQuantity3DArr(DQ3DArr *dq, DQPtr data, Counter nk, Counter nj, Counter ni){
  
  int err = noErr;
  
  *dq = (DQ3DArr)zlsSet3DArr((void *) data, sizeof(double), nk, nj, ni);
  
  if (*dq == NULL) {
    err = memErr;
  }
  
  return(err);
}

//========================================
//
// remove offset pointer structures from a multi-dim array leaving 
// the data untouched must have a copy of the data reference elsewhere 
// or it will be lost and memory will leak badly. 
//
//========================================

int zls_ReleaseDQuantity2DArr(DQ2DArr dq){
  
  int err = noErr;
  
  if (dq != NULL) {
    zlsRelease2DArr((void **) dq);
  } else { err = memErr;}
  
  return(err);
}

int zls_ReleaseDQuantity3DArr(DQ3DArr dq){
  
  int err = noErr;
  
  if (dq != NULL) {
    zlsRelease3DArr((void ***) dq);
  } else { err = memErr;}
  
  return(err);
}

//========================================
// Extract an unstructured block, useful for I/O, returns a reference
// to the data, no extra memory is allocated
//========================================

int zls_GetDQuantity2DArr(DQPtr *data, DQ2DArr dq){
  *data = dq[0];
  return(noErr);
}

int zls_GetDQuantity3DArr(DQPtr *data, DQ3DArr dq){
  *data = dq[0][0];
  return(noErr);
}



