#ifndef ZLS_TIMING_H
#define ZLS_TIMING_H
#include <sys/times.h>
#include <time.h>
#include <zls.h>


//====================================================================================
//
extern CPUCoords gZLS_GlobalCPUTimer;
//
// compute cycles between printing cycle progress reports > 16, even number
//
extern Counter kZLS_CycleProgressPrint; 
//
// compute cycles between cpu/cell/cycles performance checks 
// must be an integer multiple of kZLS_CycleProgressPrint
//
extern Counter kZLS_CPUCyclePerformance; 
//
//====================================================================================
//
// main public timing calls.
//
// returns a string representing the system real wall time and date
//
extern  char gZLS_RealTimeString[80];
char * zlst_TheCurrentRealTime(void);
//
// Start up the global timing, ctp is the *global* cpu timer structure:
// gZLS_GlobalCPUTimer, ie zlst_InitGlobalTiming(&gZLS_GlobalCPUTimer);
//
int zlst_InitGlobalTiming( CPUCoordPtr ctp ); 
//
int zlst_ClearCPUTimes   ( CPUCoordPtr ctp );
int zlst_StartCPUTimes   ( CPUCoordPtr ctp ); // to setup local timers
int zlst_EndCPUTimes     ( CPUCoordPtr ctp );
int zlst_SaveCPUTimes    ( CPUCoordPtr ct_sav, CPUCoordPtr ctp );
int zlst_ResumeCPUTimes  ( CPUCoordPtr ctp, CPUCoordPtr ct_sav );
int zlst_ProgressCheck   ( CPUCoordPtr ctp, double nCells );
//
//====================================================================================
// 
// lower level calls
// 
int zlst_ElapsedCPUTimes ( CPUCoordPtr ct , char * title);
int zlst_ElapsedCellRates( CPUCoordPtr ct, double nCells, char * title );
//====================================================================================
//
// Simulation Timers and Counter Algorithms
//
//====================================================================================

// timer events/codes
enum { 
	goCode    = noErr,
	startCode = 1,
	endCode   = 2,
	frameCode = 3,
	userCode  = 4,
	dataCode  = 5,
	saveCode  = 6
};

// flags for constructing a set of events in a single event variable.

enum {
  kGoEventFlag    = 0x0001,
  kStartEventFlag = 0x0002,
  kEndEventFlag   = 0x0004,
  kFrameEventFlag = 0x0008,
  kUserEventFlag  = 0x0010,
  kDataEventFlag  = 0x0020,
  kSaveEventFlag  = 0x0040,
  kRealDataEventFlag  = 0x0080
};

extern const Real kZLS_TimeEventSlop; 

//
// prevent steps landing too close before an event - resulting in a
// tiny step next and slow recovery of step size. Assumes IO is more
// widely spaced than simulation steps.  For 1.15: If s step lands
// within 15% of the event time, then expand the step 15%. See
// zls_timing.c to set the value. WARNING, can generate race conditions
// if too large or violate the Courant condition if CFL *
// kZLS_TimeEventSlop is too large. 1.15 assumes comensurate event
// timing, CFL ~0.6 and minimum 6-10 steps between I/O.
// 
//====================================================================================
// these are the only routines to advance a clock time, *do not bypass and access a 
// zone clock directly or you get what you deserve!
//
int zlst_AdvanceCounters          ( TimeCoords * zoneclock );
int zlst_AdvanceClock             ( TimeCoords * zoneclock );

//
//====================================================================================
// These are the routines to determine simualtion step sizes and 
// set the  zone clocks prior to a simulation step
//

extern Real kZLS_CFL; // range 0.5 - 0.8, default 0.75
extern Real kZLS_FracCFL; // fraction of CFL for initial step size
#
extern Real kZLS_CoolTimeFraction;   // range 0.1 - 10.0
//
// estimate contstrained timestep for a zone
int zlst_CalcZoneTimeStep        ( ZonePtr z );     

// collect a level wide constrained time step from all level zones
int zlst_CalcLevelTimeStep       ( Real * dt, Integer level ); 

// over all levels and zones
int zlst_CalcAllTimeSteps        ( Real * dt ); 

// sets up dt in every level *and* zone clock in a heirachial manner before a step
int zlst_SetAllClocksTimeStep     ( Real * dt ); 

// sets up dt in every level *and* zone clock in a uniform manner before a step
int zlst_SetAllClocksUniformTimeStep  (Real * dt);
//
//====================================================================================
//
int zlst_InitTimeCoords     ( TimeCoords * tc ); // zero out a TimeCoords stucture
int zlst_DefaultEventTiming ( TimeCoords * tc ); // set initial default events, 
// calls init and wipes tc
//
// Basic Setters, actually not always a good idea to use these, as it is likely to set 
// inconsistent values except at special times, such as setting up after a restart.
// Mostly used to initialise clocks with 0s.
//
//====================================================================================
int zlst_SetRealTime            ( void); // sets wall time in seconds
int zlst_SetRealTimeLimit       ( time_t  rt); // pass the ending time in whole seconds

// pass the run duration in whole seconds, sets *both* realtime and limit accordingly
int zlst_SetRealTimeDuration    ( time_t Delta_rt);
//
time_t zlst_GetRealTime            ( void ); 
time_t zlst_GetRealTimeLimit       ( void ); 
time_t zlst_GetRealTimeDuration    ( void );
time_t zlst_GetRealTimeRemaining   ( void );

//====================================================================================
//
int zlst_SetClockTime           ( TimeCoords * zoneclock,  Real t);
int zlst_SetClockCycles         ( TimeCoords * zoneclock,  Integer c);
int zlst_SetClockTimingCycles   ( TimeCoords * zoneclock,  Integer c); // m
//
int zlst_SetLevelClockTime           ( Integer level,  Real t);
int zlst_SetLevelClockCycles         ( Integer level,  Integer c);
int zlst_SetLevelClockTimingCycles   ( Integer level,  Integer c);
//
int zlst_SetAllLevelClocksTime         ( Real t);
int zlst_SetAllLevelClocksCycles       ( Integer c);
int zlst_SetAllLevelClocksTimingCycles ( Integer c);
//
// these call set and check level IO limits and sim limits, after a
// simulation step They are safe to use.
//
// Checks for output and simulation timing are based on level clocks
// actual advancing zones in algorithms use the dts set in zone clocks.
// The level clock is a master clock, using all fields, and zone clocks
// are slaves an only use the time t and step size dt in practise
//

// set single levels or all levels at once...

int zlst_SetLevelClockEndTime      ( Integer level,  Real t);
int zlst_SetLevelClockFrameTime    ( Integer level,  Real t);
int zlst_SetLevelClockUserTime     ( Integer level,  Real t);
int zlst_SetLevelClockDataTime     ( Integer level,  Real t);
int zlst_SetLevelClockSaveTime     ( Integer level,  Real t);

int zlst_SetLevelClockEndCycle     ( Integer level, Integer i);
int zlst_SetLevelClockFrameCycle   ( Integer level, Integer i);
int zlst_SetLevelClockUserCycle    ( Integer level, Integer i);
int zlst_SetLevelClockDataCycle    ( Integer level, Integer i);
int zlst_SetLevelClockSaveCycle    ( Integer level, Integer i);

int zlst_SetAllLevelClocksEndTime         ( Real t);
int zlst_SetAllLevelClocksFrameTime       ( Real t);
int zlst_SetAllLevelClocksUserTime        ( Real t);
int zlst_SetAllLevelClocksDataTime        ( Real t);
int zlst_SetAllLevelClocksSaveTime        ( Real t);

int zlst_SetAllLevelClocksEndCycle     ( Integer i);
int zlst_SetAllLevelClocksFrameCycle   ( Integer i);
int zlst_SetAllLevelClocksUserCycle    ( Integer i);
int zlst_SetAllLevelClocksDataCycle    ( Integer i);
int zlst_SetAllLevelClocksSaveCycle    ( Integer i);

//
// Here is where the Simulation level timers are actually checked and
// reset as needed.
//

int zlst_CheckIOTimeLimits        ( TimeCoords * zoneclock );
int zlst_CheckIOCycleLimits       ( TimeCoords * zoneclock );
int zlst_CheckRealTimeLimit       (void);

int zlst_CheckLevelIOTimeLimits        ( Integer level );
int zlst_CheckLevelIOCycleLimits       ( Integer level );

//
// Processes all level events and returns a code to proceed or
// terminate, Should be called for each cycle of the main loop
//

int zlst_CheckAllLevelEvents      ( void);
//====================================================================================

//
// Accessor functions/MACROS for the global times
//

#define zlst_GlobalTime      (gLevels[0].levelClock.t)
#define zlst_GlobalTimeStep  (gLevels[0].levelClock.dt)
#define zlst_GlobalEndTime   (gLevels[0].levelClock.endTime)
#define zlst_GlobalFrameTime (gLevels[0].levelClock.frameTime)
//
// in common units
//
#define zlst_GlobalSeconds   (gLevels[0].levelClock.t*f_t0)
#define zlst_GlobalYears     (gLevels[0].levelClock.t*f_t0*f_invyrs)
#define zlst_GlobalSecondsStep (gLevels[0].levelClock.dt*f_t0)
#define zlst_GlobalYearsStep (gLevels[0].levelClock.dt*f_t0*f_invyrs)

//
// Accessor function for the global cycle count, timing cycles always starts from 0 each run
// even if a restart file sets cycles to a non zero value, this is then used in performance
// tests, *only* , ie zlst_ProgressCheck
//

#define zlst_GlobalCycles ((int)gLevels[0].levelClock.cycles)
#define zlst_GlobalTimingCycles ((int)gLevels[0].levelClock.timingCycles)


//====================================================================================


#endif ZLS_TIMING_H
