#ifndef ZLS_UTILITY_H
#define ZLS_UTILITY_H
#include <sys/stat.h>
#include <errno.h>
#include <sys/types.h>

#include <zls.h>

int zls_RangeOptions(char* opt, int *i, int *j);

//========================================

char * zls_TimeIntervalString( Real seconds );

//========================================
// utility functions for directions.
int zls_opposite(int direction); // return zls direction code for the direction opposite to the input direction
char * zls_dirname( Integer direction); // convert direction index into a short standard string


//========================================
// Powerlaw integerator for use in initialisation, integrates over cell
// for correct cell values (ie not just central coordinate value)  ,
// normalised to r0 v. slow. handles alpha = -1, 
//
// x0 should not equal x1, r0 is a scaling factor, Delta x = (x1 - x0)
// evaluates <> = r0^alpha * Int(x0 -> x1) x^(alpha) dx /  Delta x
// Special faster cases for alpha = -3, -2, -1, 0, 1, 2, 3   
//========================================

Real zls_IntegratePowerlaw( Real r0,  Real alpha,  Real x0,  Real x1);

Real zls_erf(Real x);

Real zls_RealSpH0402 (Real x, Real y);
Real zls_RealSpH0804 (Real x, Real y);
Real zls_RealSpH1608 (Real x, Real y);
Real zls_RealSpH3216 (Real x, Real y);
Real zls_RealSpH6432 (Real x, Real y);
Real zls_RealSpH12864(Real x, Real y);
Real zls_RealSpH256128( Real x, Real y);

Real zls_OverlapSin2D (Real ampl, Real f, Real phase, Real offset,
                       Real xin, Real yin, Real  dx, Real dy,
                       Integer n);

Real zls_OverlapCos2D (Real ampl, Real f, Real phase, Real offset,
                       Real xin, Real yin, Real  dx, Real dy,
                       Integer n);

Real zls_OverlapLine2D ( Real a, Real b,
                        Real xin, Real yin, Real  dx, Real dy,
                        Integer n );

Real zls_OverlapCircle2D (Real rad,  Real x_0, Real y_0, 
                          Real xin, Real yin, 
                          Real  dx, Real dy,
                          Integer n);

Real zls_OverlapSphere3D (Real  rad, Real  x_0, Real  y_0, Real z_0,
                          Real  xin, Real yin, Real zin, 
                          Real   dx, Real  dy, Real dz,
                          Integer n );

Real zls_MeanProjectedRadius2D ( Real * r2, Real * fi, Real * fj,  
                                Real xin, Real yin,
                                Real   dx, Real dy,
                                Integer n );

Real zls_MeanProjectedRadius3D ( Real * r2, Real * fi, Real * fj, Real * fk,
                                Real  xin, Real yin, Real zin, 
                                Real   dx, Real  dy,  Real dz,
                                Integer n );

Real zls_SmoothSin2D (Real ampl, Real f, Real phase, Real offset,
                      Real xin, Real yin, Real  dx, Real dy,
                      Integer n, Real hs);

Real zls_SmoothCos2D (Real ampl, Real f, Real phase, Real offset,
                      Real xin, Real yin, Real  dx, Real dy,
                      Integer n, Real hs);

Real zls_SmoothLine2D ( Real a, Real b,
                       Real xin, Real yin, Real  dx, Real dy,
                       Integer n, Real hs);

Real zls_SmoothCircle2D (Real rad, Real  x_0, Real  y_0, 
                         Real xin, Real yin, Real  dx, Real  dy, Integer n, Real hs);

Real zls_SmoothSphere3D (Real rad, Real x_0,  Real y_0,  Real z_0, 
                         Real xin, Real yin, Real zin, Real dx,  Real dy,  Real dz, Integer n,  Real hs);

float zls_ran0_(SInt32 *idum);

float zls_ran1_(SInt32 *idum);

float zls_ran2_(SInt32 *idum);

Real zls_invNormal(Real x);

int zls_splint( Real *y, Real x, Integer n, Real *xa, Real *ya, Real * y2a);

int zls_mkdir(char *fname);

#ifndef OS_SOLARISAMD64
#ifdef OS_MACOSXG4
void  zls_sincos(double x, double *s, double *c);
#else 
static inline void zls_sincos(double x, double *s, double *c){
  *s = sin(x);
  *c = cos(x);
}
#endif !OS_MACOSXG4
#else 
static inline void zls_sincos(double x, double *s, double *c){
  sincos(x, s, c);
}
#endif OS_SOLARISAMD64

Real zls_AxisFiducialScale( Integer axis);

static inline Real zls_CvtCoordsToProjRadius1D( Real * fi, Real x);
static inline Real zls_CvtCoordsToProjRadius1D( Real * fi, Real x){
#ifdef CARTESIAN
  Real d = zls_abs(x);
  if (fi != NULL) *fi    = (x>=0.0)?1.0:-1.0;
#endif CARTESIAN
#ifdef CYLINDRICAL
  // x = z, ie same as cartesian
  Real d = zls_abs(x);
  if (fi != NULL) *fi    = (x>=0.0)?1.0:-1.0;
#endif CYLINDRICAL
#ifdef CIRCULAR
  // x = r,  the radial  component is 1.0
  Real d = x;
  if (fi != NULL) *fi    = 1.0;
#endif CIRCULAR
#ifdef SPHERICAL
  // x = r,  the radial  component is 1.0
  Real d = x;
  if (fi != NULL) *fi    = 1.0;
#endif SPHERICAL
#ifdef POLARSPHERICAL
  // x = r,  the radial  component is 1.0
  Real d = x;
  if (fi != NULL) *fi    = 1.0;
#endif POLARSPHERICAL
  return(d);
}

static inline Real zls_CvtCoordsToProjRadius2D( Real * fi, Real * fj,  Real x, Real y);
static inline Real zls_CvtCoordsToProjRadius2D( Real * fi, Real * fj,  Real x, Real y){
#ifdef CARTESIAN
  Real d = zls_hypot(x,y);
  if (fi != NULL) *fi    = d>0.0?x/d:0.0;
  if (fj != NULL) *fj    = d>0.0?y/d:0.0;
#endif CARTESIAN
#ifdef CYLINDRICAL
  // x = z, y = R ie *still* same as cartesian, except y is always +ve
  Real d = zls_hypot(x,y);
  if (fi != NULL) *fi    = d>0.0?x/d:0.0;
  if (fj != NULL) *fj    = d>0.0?y/d:0.0;
#endif CYLINDRICAL
#ifdef CIRCULAR
#pragma unused(y)
  // x = r, y = phi, the radial  component is 1.0, phi component = 0.0
  Real d = x;
  if (fi != NULL) *fi    = 1.0;
  if (fj != NULL) *fj    = 0.0;
#endif CIRCULAR
#ifdef SPHERICAL
#pragma unused(y)
  // x = r, y = phi, the radial  component is 1.0, phi component = 0.0
  Real d = x;
  if (fi != NULL) *fi    = 1.0;
  if (fj != NULL) *fj    = 0.0;
#endif SPHERICAL
#ifdef POLARSPHERICAL
#pragma unused(y)
  // x = r, y = phi, the radial  component is 1.0, phi component = 0.0
  Real d = x;
  if (fi != NULL) *fi    = 1.0;
  if (fj != NULL) *fj    = 0.0;
#endif POLARSPHERICAL
  return(d);
}


static inline Real zls_CvtCoordsToProjRadius3D( Real * fi, Real * fj, Real * fk, Real x, Real y, Real z);
static inline Real zls_CvtCoordsToProjRadius3D( Real * fi, Real * fj, Real * fk, Real x, Real y, Real z){
#ifdef CARTESIAN
  Real d =  zls_hypot3(x,y,z);
  if (fi != NULL) *fi    = d>0.0?x/d:0.0;
  if (fj != NULL) *fj    = d>0.0?y/d:0.0;
  if (fk != NULL) *fk    = d>0.0?z/d:0.0;
#endif CARTESIAN
#ifdef CYLINDRICAL
#pragma unused(z)
  // x = z, y = R, z = phi, ie *still* same as cartesian, except y is always +ve
  Real d = zls_hypot(x,y);
  if (fi != NULL) *fi    = d>0.0?x/d:0.0;
  if (fj != NULL) *fj    = d>0.0?y/d:0.0;
  if (fk != NULL) *fk    = 0.0;
#endif CYLINDRICAL
#ifdef CIRCULAR
#pragma unused(y,z)
  // x = R, y = phi,, ie *still* same as cartesian, except y is always +ve
  Real d = x;
  if (fi != NULL) *fi    = 1.0;
  if (fj != NULL) *fj    = 0.0;
  if (fk != NULL) *fk    = 0.0;
#endif CIRCULAR
#ifdef SPHERICAL
#pragma unused(y,z)
  // x = r, y = phi, the radial  component is 1.0, phi component = 0.0
  Real d = x;
  if (fi != NULL) *fi    = 1.0;
  if (fj != NULL) *fj    = 0.0;
  if (fk != NULL) *fk    = 0.0;
#endif SPHERICAL
#ifdef POLARSPHERICAL
#pragma unused(y,z)
  // x = r, y = phi, the radial  component is 1.0, phi component = 0.0
  Real d = x;
  if (fi != NULL) *fi    = 1.0;
  if (fj != NULL) *fj    = 0.0;
  if (fk != NULL) *fk    = 0.0;
#endif POLARSPHERICAL
  return(d);
}

static inline Real zls_CvtCoordsToRadius1D( Real x);
static inline Real zls_CvtCoordsToRadius1D( Real x){
#ifdef CARTESIAN
  Real d = zls_abs(x);
#endif CARTESIAN
#ifdef CYLINDRICAL
  // x = z, ie same as cartesian
  Real d = zls_abs(x);
#endif CYLINDRICAL
#ifdef CIRCULAR
  // x = r,  the radial  component is 1.0
  Real d = x;
#endif CIRCULAR
#ifdef SPHERICAL
  // x = r,  the radial  component is 1.0
  Real d = x;
#endif SPHERICAL
#ifdef POLARSPHERICAL
  // x = r,  the radial  component is 1.0
  Real d = x;
#endif POLARSPHERICAL
  return(d);
}

static inline Real zls_CvtCoordsToRadius2D( Real x, Real y);
static inline Real zls_CvtCoordsToRadius2D( Real x, Real y){
#ifdef CARTESIAN
  Real d = zls_hypot(x,y);
#endif CARTESIAN
#ifdef CYLINDRICAL
  // x = z, y = R ie *still* same as cartesian, except y is always +ve
  Real d = zls_hypot(x,y);
#endif CYLINDRICAL
#ifdef CIRCULAR
#pragma unused(y)
  // x = r, y = phi, the radial  component is 1.0, phi component = 0.0
  Real d = x;
#endif CIRCULAR
#ifdef SPHERICAL
#pragma unused(y)
  // x = r, y = phi, the radial  component is 1.0, phi component = 0.0
  Real d = x;
#endif SPHERICAL
#ifdef POLARSPHERICAL
#pragma unused(y)
  // x = r, y = phi, the radial  component is 1.0, phi component = 0.0
  Real d = x;
#endif POLARSPHERICAL
  return(d);
}


static inline Real zls_CvtCoordsToRadius3D( Real x, Real y, Real z);
static inline Real zls_CvtCoordsToRadius3D( Real x, Real y, Real z){
#ifdef CARTESIAN
  Real d = zls_hypot3(x,y,z);
#endif CARTESIAN
#ifdef CYLINDRICAL
#pragma unused(z)
  // x = z, y = R, z = phi, ie *still* same as cartesian, except y is always +ve
  Real d = zls_hypot(x,y);
#endif CYLINDRICAL
#ifdef CIRCULAR
#pragma unused(y,z)
  // x = R, y = phi,, ie *still* same as cartesian, except y is always +ve
  Real d = x;
#endif CIRCULAR
#ifdef SPHERICAL
#pragma unused(y,z)
  // x = r, y = phi, the radial  component is 1.0, phi component = 0.0
  Real d = x;
#endif SPHERICAL
#ifdef POLARSPHERICAL
#pragma unused(y,z)
  // x = r, y = phi, the radial  component is 1.0, phi component = 0.0
  Real d = x;
#endif POLARSPHERICAL
  return(d);
}


static inline Real zls_CvtCoordsToCylRadius1D( Real x);
static inline Real zls_CvtCoordsToCylRadius1D( Real x){
  Real d = 0.0;
#ifdef CIRCULAR
  // x = r,  the radial  component is 1.0
  d = x;
#else
#pragma unused(x)
#endif !CIRCULAR
  return(d);
}

static inline Real zls_CvtCoordsToCylRadius2D( Real x, Real y);
static inline Real zls_CvtCoordsToCylRadius2D( Real x, Real y){
#ifdef CARTESIAN
  Real d = zls_abs(y);
#endif CARTESIAN
#ifdef CYLINDRICAL
  // x = z, y = R ie *still* same as cartesian, except y is always +ve
  Real d = y;
#endif CYLINDRICAL
#ifdef CIRCULAR
  // x = r, y = phi, the radial  component is 1.0, phi component = 0.0
  Real d = x;
#endif CIRCULAR
#ifdef SPHERICAL
  // x = r, y = phi CCW.  cylinder z axis = along phi = 0 axis
  Real d = x * sin( y );
#endif SPHERICAL
#ifdef POLARSPHERICAL
  // x = r, y = phi down from y axis  cylinder z axis = along phi = 90 degrees from pole axis
  Real d = x * cos( y );
#endif POLARSPHERICAL
  return(d);
}

static inline Real zls_CvtCoordsToCylRadius3D( Real x, Real y, Real z);
static inline Real zls_CvtCoordsToCylRadius3D( Real x, Real y, Real z){
#ifdef CARTESIAN
  Real d = zls_hypot(y,z);
#endif CARTESIAN
#ifdef CYLINDRICAL
  Real d = y;
#endif CYLINDRICAL
#ifdef CIRCULAR
  Real d = x;
#endif CIRCULAR
#ifdef SPHERICAL
  // x = r, y = phi = CCW around  x-y plane, z = theta, CCW up from x-y plane the  cylinder z axis = x = along theta = 0 degrees from xy plane
  Real d = x*sin(y)*cos(z); // check this
#endif SPHERICAL
#ifdef POLARSPHERICAL
  // x = r, y = theta CW, down from z axis, z = phi = CCW around  x-y plane, the  cylinder z axis = x = along theta = 90 degrees from pole axis
  Real d = x*sin(y)*sin(z);
#endif POLARSPHERICAL
  return(d);
}


#endif ZLS_UTILITY_H
