#ifndef ZLS_VARS_H
#define ZLS_VARS_H
#include "zls.h"
//
//========================================
//
// Variable = quantity + coefficients - elements change with dimension,
// Input ns refer to grid size including  boundary cells (nBC), ready
// for sweeps, variables are always treated as multidimensional arrays,
// quantities may be Ptr or Arrs variable arrays structures include
// index information.
//
//========================================
// Quantities as composite values + coeffs for cache speed, but limited
// zone sizes OK for 1D/2D, consider separate quantites QPtrs and
// Coeffs CPtrs for large 3D sets
//========================================
//
// Coefficients here are initially for PPM
// interpolation/reconstruction, will make the coefficients a compile
// option and maybe move them to algorithms
//

typedef struct ppmccf {// value coeffcients associated with a variable in cell
  Real del  ;  // right = l + del, default = 0
  Real a6   ;  // quadratic term, can be set to 0 without affecting a, if del = 0 a6 = 0
  Real flat ;  // if cell parabola needs to be flattened
} Coeffs, * CPtr;

typedef Coeffs     * C1DArr;   // plainest block , 1 2 or 3D
typedef Coeffs    ** C2DArr;  // 2D array of Vs v[j][i]  v[0] is VPtr
typedef Coeffs   *** C3DArr;  // 3D array of Vs v[k][j][i]  v[0][0] is VPtr

// these are the main PPM strutures for conservative orthogonal
// piecewise quadratic interpolation.  The PPM algorithm takes a, finds
// al and ar by 4th order lagrangian interopolation of the integral of
// a between cells, then finds the intra-cell coefficients del and a6. 
// strictly only al, ar and a6 are needed, however del is kept locally
// for speed in a number of places. OK for 1D and 2D arrays, for full
// 3D unsplit PPM methods ar could be dropped and the algorthms
// adapted, as the extra 25% space needed for ar may not be worth the
// gains in 3D.

typedef struct varppm1{
  Real       a;
  Real      al;  // left (lower coordinate) value, default = a
  Real      ar;  // right (higher coordinate) value, default = a
  Real     del;  // right = l + del, default = 0
  Real      a6;  // quadratic term, can be set to 0 without affecting a, if del = 0 a6 = 0
} PPMVar1;

typedef struct varppm2{
  Real    a [2];  // cell integral/average value/vector
  Real    al[2];  // left (lower coordinate) value, default = a
  Real    ar[2];  // right (higher coordinate) value, default = a
  Real   del[2];  // right = l + del, default = 0
  Real    a6[2];  // quadratic term, can be set to 0 without affecting a, if del = 0 a6 = 0
} PPMVar2;

typedef struct varppm3{
  Real    a [3];  // cell integral/average value
  Real    al[3];  // left (lower coordinate) value, default = a
  Real    ar[3];  // right (higher coordinate) value, default = a
  Real   del[3];  // right = l + del, default = 0
  Real    a6[3];  // quadratic term, can be set to 0 without affecting a, if del = 0 a6 = 0
} PPMVar3;

//
// smaller cell vars with just the PLM coeffs, mean, left and right, delta is derived on the fly if needed from 
// ar - al and usually as a local variable.
//
// For unsplit algorithms also includes field for riemann * values
//


typedef struct varvectLRQ{
  Real       a;  // cell integral/average value
  Real      al;  // left (lower coordinate) value, default = a
  Real      ar;  // right = l + del, default = a
} VLRQ1;

typedef struct varvectLRQ2{
  Real     a[2];  // cell integral/average value
  Real    al[2];  // left (lower coordinate) value, default = a
  Real    ar[2];  // right (higher coordinate) value, default = a
} VLRQ2;

typedef struct varvectLRQ3{
  Real     a[3];  // cell integral/average value
  Real    al[3]; // left (lower coordinate) value, default = a
  Real    ar[3]; // right (higher coordinate) value, default = a
} VLRQ3;


typedef VLRQ1 SLRQ1;

typedef struct varscalarLRQ2{
  Real     a;  // cell integral/average value
  Real    al[2];  // left (lower coordinate) value, default = a
  Real    ar[2];  // right (higher coordinate) value, default = a
} SLRQ2;

typedef struct varscalarLRQ3{
  Real     a   ;  // cell integral/average value
  Real    al[3]; // left (lower coordinate) value, default = a
  Real    ar[3]; // right (higher coordinate) value, default = a
} SLRQ3;

// smallest cell vars with just the mean  value for Godunov 1st order calcs

typedef struct vargod1{
  Real     a;
} GODVar1;

typedef GODVar1 GODVar2;
typedef GODVar1 GODVar3;

#ifdef PPM
typedef PPMVar1 Var1;
typedef PPMVar2 Var2;
typedef PPMVar3 Var3;
#endif

#ifdef PLM
typedef SLRQ1 Var1;
typedef SLRQ2 Var2;
typedef SLRQ3 Var3;
#endif

#ifdef GODUNOV
typedef GODVar1 Var1;
typedef GODVar2 Var2;
typedef GODVar3 Var3;
#endif


typedef struct v1{
  Integer n, ns, ne, nt;
  ACoordArr  x;
  Var1     * v;
} V1DArr; // 1D array of  v[i].a 

typedef struct v2{
  Integer n[2], ns[2], ne[2], nt[2];
  ACoordArr x;
  ACoordArr y;
  Var2     ** v;
} V2DArr; // 2D array of + index ranges v[j][i].a 

typedef struct v3{
  Integer n[3], ns[3], ne[3], nt[3];
  ACoordArr x;
  ACoordArr y;
  ACoordArr z;
  Var3     *** v;
} V3DArr; // 3D array of Vs v[k][j][i].a

#ifdef SYM1D
typedef V1DArr   VArr;
#endif

#ifdef SYM2D
typedef V2DArr   VArr;
#endif

#ifdef SYM3D
typedef V3DArr   VArr;
#endif

//
// NOTE  Qptr == Q1DArr, CPtr == C1DArr,   == I1DArr,  VPtr != V1DArr
//
//========================================

int zls_NewVar1DArr  (V1DArr *v, Integer nBC, Counter ni);
int zls_NewVar2DArr  (V2DArr *v, Integer nBC, Counter nj, Counter ni);
int zls_NewVar3DArr  (V3DArr *v, Integer nBC, Counter nk, Counter nj, Counter ni);

int zls_DisposeVar1DArr (V1DArr v);
int zls_DisposeVar2DArr (V2DArr v);
int zls_DisposeVar3DArr (V3DArr v);
int zls_DisposeVarArr     (VArr v);

//========================================

int zls_SetVar1DArrQ1DArr(V1DArr v, Q1DArr q); // set var = q values, offset bcs , coeffs 0.0, dims must match
int zls_SetVar2DArrQ2DArr(V2DArr v, Q2DArr q); // v provides dimensions
int zls_SetVar3DArrQ3DArr(V3DArr v, Q3DArr q);

int zls_SetVarArrQArr(VArr v, QArr q);  // dim equals the zoneData dim

//========================================
//
// Extracting 1D rows from 2D and 3D QArrs into vars
//
int zls_SetVar1DArrQ2DArrX(V1DArr v, Q2DArr q, Integer j); 
int zls_SetVar1DArrQ2DArrY(V1DArr v, Q2DArr q, Integer i); 

int zls_SetVar1DArrQ3DArrX(V1DArr v, Q3DArr q, Integer k, Integer j); 
int zls_SetVar1DArrQ3DArrY(V1DArr v, Q3DArr q, Integer k, Integer i); 
int zls_SetVar1DArrQ3DArrZ(V1DArr v, Q3DArr q, Integer j, Integer i); 

//========================================

int zls_SetVar1DArrAxis(V1DArr v, ACoordPtr x); // dims must match
int zls_SetVar2DArrAxis(V2DArr v, ACoordPtr x, ACoordPtr y); // v provides dimensions
int zls_SetVar3DArrAxis(V3DArr v, ACoordPtr x, ACoordPtr y, ACoordPtr z);

int zls_SetVar1DArrQ   (V1DArr v,   QPtr q); // set var = q values, offset bcs , coeffs 0.0, dims must match
int zls_SetVar2DArrQ   (V2DArr v,   QPtr q); // v provides dimensions
int zls_SetVar3DArrQ   (V3DArr v,   QPtr q);

//========================================
// LR equivalents to Q1DArr etc  a + al + ar instead of just a,  3*
// size of QArr but still basicly a plain block.  Note there is no
// plain LRQPtr as these do point to a strucure rather than an atomic. 
// Helpful in large blocks where interleave impoves cache performance.
//========================================
// simple  quantity block type for any with 1D Left-Centre-Right values
// per index, ie Riemann problem Star region. Used more like QArr as a
// bare array without the strutured header basicly for higher order
// quantities, defined here after PLMVar1 is defined
//========================================

#ifdef SCALAR_LRQS
typedef SLRQ1 LRQ1;
typedef SLRQ2 LRQ2;
typedef SLRQ3 LRQ3;
#endif SCALAR_LRQS

#ifdef VECTOR_LRQS
typedef VLRQ1 LRQ1;
typedef VLRQ2 LRQ2;
typedef VLRQ3 LRQ3;
#endif VECTOR_LRQS

//
// LRQArrs are multi-d arrays of 1D l-c-r values, for collections of independent 
// scalar and vector components
// such as potentials and  fields.
//
typedef LRQ1 *   LRQ1DArr;
typedef LRQ2 **  LRQ2DArr;
typedef LRQ3 *** LRQ3DArr;

typedef SLRQ1 *   SLRQ1DArr;
typedef SLRQ2 **  SLRQ2DArr;
typedef SLRQ3 *** SLRQ3DArr;

typedef VLRQ1 *   VLRQ1DArr;
typedef VLRQ2 **  VLRQ2DArr;
typedef VLRQ3 *** VLRQ3DArr;

#ifdef SYM1D
typedef LRQ1DArr   LRQArr;
typedef SLRQ1DArr    SLRQArr;
typedef VLRQ1DArr    VLRQArr;
#endif

#ifdef SYM2D
typedef LRQ2DArr   LRQArr;
typedef SLRQ2DArr    SLRQArr;
typedef VLRQ2DArr    VLRQArr;
#endif

#ifdef SYM3D
typedef LRQ3DArr   LRQArr;
typedef SLRQ3DArr   SLRQArr;
typedef VLRQ3DArr   VLRQArr;
#endif

//========================================

int zls_NewLRQ1DArr  (LRQ1DArr *v, Counter nti);
int zls_NewLRQ2DArr  (LRQ2DArr *v, Counter ntj, Counter nti);
int zls_NewLRQ3DArr  (LRQ3DArr *v, Counter ntk, Counter ntj, Counter nti);

int zls_DisposeLRQ1DArr (LRQ1DArr v);
int zls_DisposeLRQ2DArr (LRQ2DArr v);
int zls_DisposeLRQ3DArr (LRQ3DArr v);

int zls_NewSLRQ1DArr  (SLRQ1DArr *v, Counter nti);
int zls_NewSLRQ2DArr  (SLRQ2DArr *v, Counter ntj, Counter nti);
int zls_NewSLRQ3DArr  (SLRQ3DArr *v, Counter ntk, Counter ntj, Counter nti);

int zls_DisposeSLRQ1DArr (SLRQ1DArr v);
int zls_DisposeSLRQ2DArr (SLRQ2DArr v);
int zls_DisposeSLRQ3DArr (SLRQ3DArr v);

int zls_NewVLRQ1DArr  (VLRQ1DArr *v, Counter nti);
int zls_NewVLRQ2DArr  (VLRQ2DArr *v, Counter ntj, Counter nti);
int zls_NewVLRQ3DArr  (VLRQ3DArr *v, Counter ntk, Counter ntj, Counter nti);

int zls_DisposeVLRQ1DArr (VLRQ1DArr v);
int zls_DisposeVLRQ2DArr (VLRQ2DArr v);
int zls_DisposeVLRQ3DArr (VLRQ3DArr v);

// setters copy plain blocks into LRQs cell by cell

// set LRQ = q values, dims must match, BCs are offset if present
// ni is size of q, nBC are the number of boundary cells in v
// these set the l and r value of v to equal q as well

int zls_SetLRQ1DArrQArr   ( LRQ1DArr v, Q1DArr q, Integer nBC, Counter ni );
int zls_SetLRQ2DArrQArr   ( LRQ2DArr v, Q2DArr q, Integer nBC, Counter nj, Counter ni );
int zls_SetLRQ3DArrQArr   ( LRQ3DArr v, Q3DArr q, Integer nBC, Counter nk, Counter nj, Counter ni );

int zls_SetVar1DArrtLRQ1DArr ( V1DArr v, LRQ1DArr lr ); 

//========================================

int zls_RepeatBoundaryVarArr       (VArr v); // dim equals the zoneData dim

int zls_RepeatBoundaryVar1DArr   (V1DArr v);
int zls_RepeatBoundaryVar2DArr   (V2DArr v);
int zls_RepeatBoundaryVar3DArr   (V3DArr v);

//========================================
//
// calculating variable PPM coeffs
//
//========================================


//========================================
//
// Interpolation coeffs - Conservativbe Sub-divisions of PPM parabolae,
// for resampling data while preserving the PPM monotnicity and
// conserving all quantities precisely
//
//========================================

enum {  
  kInterpNN      =  0,   // nearest neighbor -simple rebin- piecewise constant
  kInterpPLM     =  1,   // piecewise linear, with flatteing, lower order than PPM
  kInterpPPM     =  2   // piecewise parabolic method - ala collella & woodward 1984
};

enum {
  kNoInterpBCsFromSuper = 0, // when interpolating a zone from the super, dont fill the bcs
  kInterpBCsFromSuper   = 1 // when interpolating a zone from the super, fill the bcs too
};


//========================================
//  General Interpolation Coefficients - simple sub cell fractions
//========================================
//

typedef struct interp {//coeffcients associated with a variable in cell
  Real  c[kNRefineMax];
} Interp, * InPtr;

typedef Interp       * In1DArr;  
typedef Interp        ** In2DArr;  
typedef Interp       *** In3DArr;

// just use the dimension and integer parts of r, cast for differnt dim Interp

int zls_NewInterp            (InPtr *f, RangePtr r); 
int zls_NewInterp1D          (InPtr *f, Counter ni);
int zls_NewInterp2D          (InPtr *f, Counter ni, Counter nj);
int zls_NewInterp3D          (InPtr *f, Counter ni, Counter nj, Counter nk);

int zls_DisposeInterp        (InPtr f);

int zls_NewInterp2DArr       (In2DArr *f, Counter nj, Counter ni);
int zls_NewInterp3DArr       (In3DArr *f, Counter nk, Counter nj, Counter ni);

int zls_DisposeInterp2DArr   (In2DArr f);
int zls_DisposeInterp3DArr   (In3DArr f);

//
//========================================
//
// Structures for passing around interpreted zonedata as collections of
// multi-dimensional arrays, independent of zone storage.  These will
// contain references to the originals, usually zoneData.  Temporary
// multi-d wrappers will be allocated for plain block zones.  So, even
// though these are just references, they should be released when
// finished with if the zones are plain.  If the zonedata is already
// multi-d then release does nothing. If the zoneData  plain, then just
// the parts of the arrays allocated with the getter are released and
// the main data is untouched.
//
// Most functions are in zls_zones, but the declarations are here to
// make the typedefs more widely seen.
//
//
// Primitive/Physical (W) HydroDynamic quantities, different dim arrays
// in each case There are no New/Dispose for WHDs or UHDs etc
//
// more will be added for MHD etc in the future.
//

#ifdef HYDRO

#ifdef SYM1D
typedef struct whd {//coeffcients associated with a variable in cell
  Q1DArr  d; // density
  Q1DArr  P; // pressure (note capital)
  Q1DArr  vx; // x velocity
#if (NTRACERS > 0)    
  Q1DArr  tr[kNTracers]; // tracer vars
#endif
} WHD, * WHDPtr;
#endif SYM1D

#ifdef SYM2D
typedef struct whd {//coeffcients associated with a variable in cell
  Q2DArr  d; // density
  Q2DArr  P; // pressure (note capital)
  Q2DArr  vx; // x velocity
  Q2DArr  vy; // y velocity
#if (NTRACERS > 0)    
  Q2DArr  tr[kNTracers]; // tracer vars
#endif
} WHD, * WHDPtr;
#endif SYM2D

#ifdef SYM3D
typedef struct whd {//coeffcients associated with a variable in cell
  Q3DArr  d; // density
  Q3DArr  P; // pressure (note capital)
  Q3DArr  vx; // x velocity
  Q3DArr  vy; // y velocity
  Q3DArr  vz; //  z velocity
#if (NTRACERS > 0)    
  Q3DArr  tr[kNTracers]; // tracer vars
#endif
} WHD, * WHDPtr;
#endif SYM3D

// Conserved (U) HydroDynamic vars, different dim arrays in each case

#ifdef SYM1D
typedef struct uhd {//coeffcients associated with a variable in cell
  Q1DArr  d; // density
  Q1DArr  E; // total specifc Energy
  Q1DArr  mx; // x specifc  momentum 
#if (NTRACERS > 0)    
  Q1DArr  tr[kNTracers]; // tracer vars
#endif
} UHD, * UHDPtr;
#endif SYM1D

#ifdef SYM2D
typedef struct uhd {//coeffcients associated with a variable in cell
  Q2DArr  d; // density
  Q2DArr  E; // total specifc Energy
  Q2DArr  mx; // x specifc  momentum 
  Q2DArr  my; // y specifc  momentum 
#if (NTRACERS > 0)    
  Q2DArr  tr[kNTracers]; // tracer vars
#endif
} UHD, * UHDPtr;
#endif SYM2D

#ifdef SYM3D
typedef struct uhd {//coeffcients associated with a variable in cell
  Q3DArr  d; // density
  Q3DArr  E; // total specifc Energy
  Q3DArr  mx; // x specifc  momentum 
  Q3DArr  my; // y specifc  momentum 
  Q3DArr  mz; // z specifc  momentum 
#if (NTRACERS > 0)    
  Q3DArr  tr[kNTracers]; // tracer vars
#endif
} UHD, * UHDPtr;
#endif SYM3D

#endif HYDRO

//========================================
// external vector fields - components for each direction for gravity
// for example fields that are not part of the hydro/mhd quantities. 
// eg B is part of the MHD system and will be in the w/u strutures.
// Coriorlis, radiation or gravity forces are applied to the quantities
// and so belong in a separate struture. presently only gravity is
// used. An init gravity call will take the zone field and coordinates
// and precompute a fixed gravity field.  In the future the gravity
// field may evolve etc. and an update gravity call will be needed.
// Field s are defined here to maximise their visibility to the code. 
// The details are in physics/gravity modules later.
//
// In line with the external nature of these fiedls, fields will belong
// to levels, and will be closely linked to the ProblemPtr and the
// PhysicsPtr.  Zones on the level will refer to the level fields,
// which in turn may refer to the global level0 field.  Zones may 
// construct a field from the level field information initially and
// re-use it or if the field evolves they will need to recompute the
// field from the level information repeatedly.
//
// These store cell average as well as L-Right values, so can be used
// with linear interp straight up, avoiding the quartic lagrangian edge
// interpolation, and as PPM by just adding an a6 coefficient.
//
//========================================
//

typedef struct f {
  LRQArr    p; // field scalar potential, <a>, conservative integral cell average
} Field, * FieldPtr;

//==================================================================
// hydro fluxes - components for each direction, assumed for lower (left) boundary of the cell
//================================================================================


typedef struct wflx1 {// w fluxes associated with a 1D W cell, can be a row from a higher dimension block
  
  Real  deltaVol;
  
  Real   m; // mass = density * Vol
  
  Real E;
  
#ifdef SYM1D
  Real  mx; // x momentum
#endif SYM1D
#ifdef SYM2D
  Real  mx; // x momentum
  Real  my; // y momentum
#endif SYM2D
#ifdef SYM3D
  Real  mx; // x momentum
  Real  my; // y momentum
  Real  mz; // z momentum
#endif SYM3D
#if (NTRACERS > 0)    
  Real  mtr[kNTracers]; // tracer masses
#endif
} WFlux1;

typedef struct gflx1 {// w fluxes associated with a 1D W cell, can be a row from a higher dimension block
  Real   d; // density flux = density * vx
  Real   E; // energy density flux = e * vx
#ifdef SYM1D
  Real  mx; // x momentum flux  = mx*vx
#endif SYM1D
#ifdef SYM2D
  Real  mx; // x momentum flux  = mx*vx
  Real  my; // y momentum flux  = my*vx
#endif SYM2D
#ifdef SYM3D
  Real  mx; // x momentum flux  = mx*vx
  Real  my; // y momentum flux  = my*vx
  Real  mz; // y momentum flux  = mz*vx
#endif SYM3D
#if (NTRACERS > 0)    
  Real  dtr[kNTracers]; // tracer density fluxes
#endif
} GFlux1;

typedef struct wflx2 {// w fluxes associated with a 2D W cell in two directions on the lower boundaries
  Real  deltaVol[2];
  
  Real   m[2]; 
  Real  E[2];
  
#ifdef SYM1D
  Real  mx[2];    // never used, 1D sims won't create 2D fluxes
#endif SYM1D
#ifdef SYM2D
  Real  mx[2];  
  Real  my[2];  
#endif SYM2D
#ifdef SYM3D
  Real  mx[2];  
  Real  my[2];  
  Real  mz[2];  
#endif SYM3D
#if (NTRACERS > 0)    
  Real  mtr[kNTracers][2]; // tracer vars
#endif
} WFlux2;

typedef struct wflx3 {// w fluxes associated with a 3D W cell, in 3 directions on the lower boundary
  Real  deltaVol[3];
  
  Real   m[3]; 
  Real   E[3];
  
#ifdef SYM1D
  Real  mx[3];
#endif SYM1D
#ifdef SYM2D
  Real  mx[3];
  Real  my[3]; 
#endif SYM2D
#ifdef SYM3D
  Real  mx[3]; 
  Real  my[3]; 
  Real  mz[3]; 
#endif SYM3D
#if (NTRACERS > 0)    
  Real  mtr[kNTracers][3]; // tracer vars
#endif
} WFlux3;

typedef struct f1 {
  Integer n, ns, ne, nt;
  WFlux1   * f;
} Flx1DArr;

typedef struct f2 {
  Integer n[2], ns[2], ne[2], nt[2];
  WFlux2   ** f;
} Flx2DArr;

typedef struct f3 {
  Integer n[3], ns[3], ne[3], nt[3];
  WFlux3   *** f;
} Flx3DArr;


#ifdef SYM1D
typedef Flx1DArr  FlxArr;  
#endif SYM1D

#ifdef SYM2D
typedef Flx2DArr  FlxArr;  
#endif SYM2D

#ifdef SYM3D
typedef Flx3DArr  FlxArr;  
#endif SYM3D


//================================================================================
// working variable arrays - obtained from the zone data as variables,
// independent of the zone data format. Then passed to algorithm
// routines for solving.
//================================================================================

typedef struct w1{
  Integer n, ns, ne, nt;
  //  Integer sweep; // dimension axis represented by the 1D row.
  ACoordArr x; // refs to owning zone coordinates
#ifdef SYM2D
  AxisCoords y; // if 2D then this is the orthogonal cell coords of the row
#endif SYM2D
#ifdef SYM3D
  AxisCoords y; // if 3D then this is the orthogonal cell coords of the row, using right hand permutations
  AxisCoords z; // if 3D then this is the orthogonal cell coords of the row
#endif SYM3D
  Q1DArr   flat; // flattening in sweep direction
#ifdef DIVVHCORRECTION
  Q1DArr   divv; // multi-d velocity divergence
#endif DIVVHCORRECTION
  
  SLRQ1  * cs; // This is adiabatic sound speed, not fit with PPM coeffs, so just an LRQ
#ifdef VARIABLE_GAMMA
  SLRQ1  * g ; // not fit with PPM coeffs, so just an LRQ
#endif VARIABLE_GAMMA
  
  Var1     * d;  // here down are data from the zone
  Var1     * P;
  Var1     * E;
  
#ifdef HYDRO
#ifdef SYM1D
  Var1     * vx;
#endif SYM1D
#ifdef SYM2D
  Var1     * vx;
  Var1     * vy;
#endif SYM2D
#ifdef SYM3D
  Var1     * vx;
  Var1     * vy;
  Var1     * vz;
#endif SYM3D
#endif HYDRO
  
#if (NTRACERS > 0)    
  Var1     * tr[kNTracers];
#endif
  
} W1DArr; // 1D array ie  w1.P[i].a 


typedef struct w2{
  Integer n[2], ns[2], ne[2], nt[2];
  //  Integer slice; // dimension slice represented by the 2D plane.
  ACoordArr x; // refs to owning zone coordinates
  ACoordArr y;
#ifdef SYM3D
  AxisCoords z; // if 3D then this is the orthogonal cell coords of the plane
#endif SYM3D
  Q2DArr   flat; // flattening in sweep direction
#ifdef DIVVHCORRECTION
  Q2DArr   divv; // multi-d velocity divergence
#endif DIVVHCORRECTION
  
  SLRQ2  ** cs;   // Adiabatic sound speed
#ifdef VARIABLE_GAMMA
  SLRQ2  **  g;   // adiabtic index
#endif VARIABLE_GAMMA
  
  Var2     ** d;
  Var2     ** P;
  Var2     ** E;
  
#ifdef HYDRO
#ifdef SYM1D
  Var2     ** vx;
#endif SYM1D
#ifdef SYM2D
  Var2     ** vx;
  Var2     ** vy;
#endif SYM2D
#ifdef SYM3D
  Var2     ** vx;
  Var2     ** vy;
  Var2     ** vz; 
#endif SYM3D
#endif HYDRO
    
#if (NTRACERS > 0)    
  Var2     ** tr[kNTracers]; // tracers may not need coefficients
#endif

} W2DArr; // 2D array of w+ index ranges ie w.d[j][i].a

typedef struct w3{
  Integer n[3], ns[3], ne[3], nt[3];
  ACoordArr x; // refs to owning zone coordinates
  ACoordArr y;
  ACoordArr z;
  Q3DArr   flat; // flattening along sweep
#ifdef DIVVHCORRECTION
  Q3DArr   divv; // multi-d velocity divergence
#endif DIVVHCORRECTION
  
  SLRQ3  *** cs;   // Adiabatic sound speed
#ifdef VARIABLE_GAMMA
  SLRQ3  *** g ;   // Adiabatic index
#endif VARIABLE_GAMMA
  
  // each cell has coeffs for the scalar in each direction, var3 needed
  Var3     *** d;
  Var3     *** P;
  Var3     *** E;
  
#ifdef HYDRO
#ifdef SYM1D
  Var3     *** vx;
#endif SYM1D
#ifdef SYM2D
  Var3     *** vx;
  Var3     *** vy;
#endif SYM2D
#ifdef SYM3D
  Var3     *** vx;
  Var3     *** vy;
  Var3     *** vz; 
#endif SYM3D
#endif HYDRO
  
#if (NTRACERS > 0)    
  Var3     *** tr[kNTracers]; // tracers may not need coefficients
#endif
} W3DArr; // 3D array of + index ranges ie w.d[k][j][i].a

//
// conserved variable versions of W1DArr - W3DArr
//
#ifdef HYDRO

typedef struct u1{
  Integer n, ns, ne, nt;
  //  Integer sweep; // dimension axis represented by the 1D row.
  ACoordArr x; // refs to owning zone coordinates
#ifdef SYM2D
  AxisCoords y; // if 2D then this is the orthogonal cell coords of the row
#endif SYM2D
#ifdef SYM3D
  AxisCoords y; // if 3D then this is the orthogonal cell coords of the row, using right hand permutations
  AxisCoords z; // if 3D then this is the orthogonal cell coords of the row
#endif SYM3D
//  Q1DArr   flat; // flattening in sweep direction
//#ifdef DIVVHCORRECTION
//  Q1DArr   divv; // multi-d velocity divergence
//#endif DIVVHCORRECTION
  
  VLRQ1  * cs; // This is adiabatic sound speed, not fit with PPM coeffs, so just an LRQ
#ifdef VARIABLE_GAMMA
  VLRQ1  * g ; // not fit with PPM coeffs, so just an LRQ
#endif VARIABLE_GAMMA
  
  VLRQ1     * d;  // here down are data from the zone
  VLRQ1     * P;
  VLRQ1     * E;
  
#ifdef SYM1D
  VLRQ1     * mx;
#endif SYM1D
#ifdef SYM2D
  VLRQ1     * mx;
  VLRQ1     * my;
#endif SYM2D
#ifdef SYM3D
  VLRQ1     * mx;
  VLRQ1     * my;
  VLRQ1     * mz;
#endif SYM3D
  
  
#if (NTRACERS > 0)    
  VLRQ1     * tr[kNTracers];
#endif
  
} U1DArr; // 1D array ie  u1.P[i].a 


typedef struct u2{
  Integer n[2], ns[2], ne[2], nt[2];
  //  Integer slice; // dimension slice represented by the 2D plane.
  ACoordArr x; // refs to owning zone coordinates
  ACoordArr y;
#ifdef SYM3D
  AxisCoords z; // if 3D then this is the orthogonal cell coords of the plane
#endif SYM3D
//  Q2DArr   flat; // flattening in sweep direction
//#ifdef DIVVHCORRECTION
//  Q2DArr   divv; // multi-d velocity divergence
//#endif DIVVHCORRECTION
  
  VLRQ2  ** cs;   // Adiabatic sound speed
#ifdef VARIABLE_GAMMA
  VLRQ2  **  g;   // adiabtic index
#endif VARIABLE_GAMMA
  
  VLRQ2     ** d;
  VLRQ2     ** P;
  VLRQ2     ** E;
  
#ifdef SYM1D
  VLRQ2     ** mx;
#endif SYM1D
#ifdef SYM2D
  VLRQ2     ** mx;
  VLRQ2     ** my;
#endif SYM2D
#ifdef SYM3D
  VLRQ2     ** mx;
  VLRQ2     ** my;
  VLRQ2     ** mz; 
#endif SYM3D
  
#if (NTRACERS > 0)    
  VLRQ2     ** tr[kNTracers]; // tracers may not need coefficients
#endif
} U2DArr; // 2D array of u index ranges ie u.d[j][i].a

typedef struct u3{
  Integer n[3], ns[3], ne[3], nt[3];
  ACoordArr x; // refs to owning zone coordinates
  ACoordArr y;
  ACoordArr z;
//  Q3DArr   flat; // flattening along sweep
//#ifdef DIVVHCORRECTION
//  Q3DArr   divv; // multi-d velocity divergence
//#endif DIVVHCORRECTION
  
  VLRQ3  *** cs;   // Adiabatic sound speed
#ifdef VARIABLE_GAMMA
  VLRQ3  *** g ;   // Adiabatic index
#endif VARIABLE_GAMMA
  
  // each cell has coeffs for the scalar in each direction, var3 needed
  VLRQ3     *** d;
  VLRQ3     *** P;
  VLRQ3     *** E;
  
#ifdef SYM1D
  VLRQ3     *** mx;
#endif SYM1D
#ifdef SYM2D
  VLRQ3     *** mx;
  VLRQ3     *** my;
#endif SYM2D
#ifdef SYM3D
  VLRQ3     *** mx;
  VLRQ3     *** my;
  VLRQ3     *** mz; 
#endif SYM3D
  
#if (NTRACERS > 0)    
  VLRQ3     *** tr[kNTracers]; // tracers may not need coefficients
#endif
} U3DArr; // 3D array of u index ranges ie u.d[k][j][i].a

#endif HYDRO

// equivalent to W1Darr for fields
typedef struct fld1{
  Integer n, ns, ne, nt;
  Integer row; // dimension axis represented by the 1D row.
  ACoordArr x; // refs to owning zone coordinates
#ifdef SYM2D
  AxisCoords y; // if 2D then this is the orthogonal cell coords of the row
#endif SYM2D
#ifdef SYM3D
  AxisCoords y; // if 3D then this is the orthogonal cell coords of the row, using right hand permutations
  AxisCoords z; // if 3D then this is the orthogonal cell coords of the row
#endif SYM3D
  
  Var1     * xi; // for just using the central and sweep-axis l/r values
  
} F1DArr; // 1D array ie  f.xi[i].a 

// equivalent to W2Darr for fields
typedef struct fld2{
  Integer n[2], ns[2], ne[2], nt[2];
  Integer slice; // dimension slice represented by the 2D plane.
  ACoordArr x; // refs to owning zone coordinates
  ACoordArr y;
#ifdef SYM3D
  AxisCoords z; // if 3D then this is the orthogonal cell coords of the plane
#endif SYM3D
  Var2     ** xi; // 
} F2DArr; // 2D array ie  f.xi[j][i].a 


typedef struct fld3{
  Integer n[3], ns[3], ne[3], nt[3];
  ACoordArr x; // refs to owning zone coordinates
  ACoordArr y;
  ACoordArr z;
  // each array is the components *averaged over the cell in the other
  // directions* so each has just one set of coefficients along
  // component direction so Var1 arrays are sufficient for a curl free
  // vector field.
  Var3     *** xi; 
} F3DArr; // 3D array of + index ranges ie f.xi[k][j][i].a



//================================================================================
// A specialised subset of W structures just to describe the Riemann Wstar variables
//

// 1D row of WStar variable, avail whatever the SYM settings
typedef struct ws1{
  Integer n, ns, ne, nt;
  ACoordArr x; // refs to owning zone coordinates
  SLRQ1DArr   P; // P.a = solved P*, P.al = input left P, P.ar = input right P
  SLRQ1DArr   v; // v.v = solved v*, v.al = input left v, v.ar = input right v
  SLRQ1DArr   d; //   left and right density input
  SLRQ1DArr  cs; // left and right sound speed input
#ifdef VARIABLE_GAMMA
  SLRQ1DArr   g; // left and right gamma input
#endif !VARIABLE_GAMMA
} Wstar1DArr; 

#ifdef UNSPLITSWEEP

typedef struct ws{
#ifdef SYM1D 
  Integer n, ns, ne, nt;
  ACoordArr x; // refs to owning zone coordinates
#endif SYM1D
#ifdef SYM2D 
  Integer n[2], ns[2], ne[2], nt[2];
  ACoordArr x; // refs to owning zone coordinates
  ACoordArr y;
#endif SYM2D
#ifdef SYM3D 
  Integer n[3], ns[3], ne[3], nt[3];
  ACoordArr x; // refs to owning zone coordinates
  ACoordArr y;
  ACoordArr z;
#endif SYM3D
  
  SLRQArr   P; // P.as = solved P*, P.a = cell average P,  P.al = input left P, P.ar = input right P
  SLRQArr   v; // v.as = solved v*, v.a = cell average v, v.al = input left v, v.ar = input right v
  SLRQArr   d; // cell average, left and right density input
  SLRQArr  cs; // cell average, left and right sound speed input
  
#ifdef VARIABLE_GAMMA
  SLRQArr   g; // cell average, left and right gamma input
#endif VARIABLE_GAMMA
} WStar; 

#ifdef SYM2D
typedef WStar Wstar2DArr;
#endif SYM2D
#ifdef SYM3D
typedef WStar Wstar3DArr;
#endif SYM3D

#endif UNSPLITSWEEP

#endif
