#include "zls.h"
//
// global master Zone, head of zone heirachy owner/sub tree
//

Integer   gNextZoneID; 
Zone     gMasterZone;

#ifdef MPI
int  gMPIrank;
int  gMPIsize;
MPI_Status  gMPIstatus;
int  gMPIBlocking;
#endif MPI

int     zls_NewZone(ZonePtr *z){
  int err = noErr;
  
  *z = (ZonePtr)malloc(sizeof(Zone));
  
  if (*z == NULL) {
    err = memErr;
#if (ZLS_DEBUG >0)
    zlsio_ErrPrintf( " ERROR: Zone allocation failed %d \n",err);
#endif
    return(err);
  }
  
  return(err);
  
}    

int     zls_DisposeZone(ZonePtr  z){
  int err = noErr;
  
  if (z != NULL) {
    
    Integer i;
    
    for (i = 0; i< z->nQ; i++) {
#ifdef MPI
      if ( z->zoneRankID == gMPIrank ){
#endif MPI
        zls_DisposeZoneData   ( z, i );
        z->zoneData[i] = NULL;
#ifdef MEANSPLITSWEEPS
        zls_DisposeZoneDataAlt ( z, i );
#endif MEANSPLITSWEEPS
#ifdef MPI
      }
#endif MPI
    } // i
    
    zls_DisposeZoneAuxilliaryData   ( z);
    
#ifdef GRAVITY
    zls_DisposeZoneField( z );
#endif GRAVITY
        
    free((void*)z);
    
  } else { err = memErr;}
  
  return(err);
  
} 


int zls_InitZone      ( ZonePtr z, Integer kind, Integer level, RangePtr rLevel, ZonePtr Super, RangePtr rSuper){
  //
  // ni, nj, nk in super coords
  //
  // NOTE: Doesn't allocate data arrays.
  //
  int err = noErr;
  Integer nBC      = 0;
  Integer nDims    = NDIMS;
  GridCoords       g;
  Integer          i; 
  
  z->zoneType     = kind;  
  
  z->zoneLevel    = level; 
  z->zoneID       = gNextZoneID++;  // Unique ID, allocated on creation,
  z->zoneRankID   = 0; // used as comparitor to gMPIrank in MPI
#ifdef MPI
  if (kind == kImageZoneType) {
    // image zones are always on rank == level
    z->zoneRankID   = level; // allocate image zones to MPI rank == level
  } else {
    
    if ( level == 0){
      // this is a simple MPI Allocation strategies that could be improved with an actual load balancing scheme
#ifdef MPI_MODALLOCATION
      // often columns allocated to same rank, when row zones == gMPIsize
      z->zoneRankID   = z->zoneID % gMPIsize; // allocate zones to MPI ranks in turn , columns allocated
#endif MPI_MODALLOCATION
#ifdef MPI_BLOCKALLOCATION
      // allocate zones to MPI ranks in in lots of gMPIBlocking = nZones/gMPIsize turn
      // often maximises adjacency
      z->zoneRankID   = (z->zoneID / gMPIBlocking) % gMPIsize; 
#endif MPI_BLOCKALLOCATION
#ifdef MPI_DIVALLOCATION
      // allocate zones to MPI ranks in gMPIsize blocks in turn, gMPIsize < nZones,
      // often rows allocated to same rank, and row zones == gMPIsize
      z->zoneRankID   = (z->zoneID / gMPIsize) % gMPIsize;
#endif MPI_DIVALLOCATION
      //      zlsio_Printf(" MPI zls_InitZone, Blocking = %d, Rank = %d, zoneID = %d zoneRankID = %d\n",
      //             gMPIBlocking, gMPIrank, z->zoneID, z->zoneRankID);
    } else {
      
      // for now subzones get same rankid as owning super zone, will relax this in future
      if (Super != NULL) {
        z->zoneRankID = Super->zoneRankID;
      } else {
#ifdef MPI_MODALLOCATION
        // often columns allocated to same rank, when row zones == gMPIsize
        z->zoneRankID   = z->zoneID % gMPIsize; // allocate zones to MPI ranks in turn , columns allocated
#endif MPI_MODALLOCATION
#ifdef MPI_BLOCKALLOCATION
        // allocate zones to MPI ranks in in lots of gMPIBlocking = nZones/gMPIsize turn
        // often maximises adjacency
        z->zoneRankID   = (z->zoneID / gMPIBlocking) % gMPIsize; 
#endif MPI_BLOCKALLOCATION
#ifdef MPI_DIVALLOCATION
        // allocate zones to MPI ranks in gMPIsize blocks in turn, gMPIsize < nZones,
        // often rows allocated to same rank, and row zones == gMPIsize
        z->zoneRankID   = (z->zoneID / gMPIsize) % gMPIsize;
#endif MPI_DIVALLOCATION
        //        zlsio_Printf(" MPI zls_InitZone, Blocking = %d, Rank = %d, zoneID = %d zoneRankID = %d\n",
        //             gMPIBlocking, gMPIrank, z->zoneID, z->zoneRankID);
      }
    }
  }
#endif MPI
  
  if ( Super != NULL){
    z->zoneRefine   = gLevels[level].levelRefinement ; // spatial refinement factor *from* supergrid, 2..8
  } else {
    z->zoneRefine = 1; 
  }
  //
  // Top level structure - Zone contains clock, grid and variables, plus
  // refs to super,owning Zone
  //
  z->zoneSuper         = Super;  // super zone this zone is embedded in
  z->zoneSuperRange    = *rSuper;
  
  //  zls_ScaleRangeCells(rLevel, kXDim, z->zoneRefine);
  //  zls_ScaleRangeCells(rLevel, kYDim, z->zoneRefine);
  //  zls_ScaleRangeCells(rLevel, kZDim, z->zoneRefine);
  
  z->zoneLevelRange    = *rLevel; // this range is on the current level resolution
  
#ifdef ZLS_LOADSORTING
  zlst_ClearCPUTimes(&z->zoneCPUTimer);  
#endif ZLS_LOADSORTING
  
  z->zoneClock            = gLevels[level].levelClock;  
  z->zoneCFL              = gLevels[level].levelCFL;
  
  z->zoneNewDt            = 0.0;
  z->zoneOldDt            = 0.0;
  z->zoneTimeScaleStatus  = kZLS_InValid;
  
  z->zoneBCells           = gLevels[level].levelBCells;
  z->zoneSuperBCells      = gLevels[level].levelSuperBCells;
  z->zoneBCInterp         = gLevels[level].levelInterpolation; 
  
  // when any fill from super calls are made, use this interpolation code
  
  for (i = 0; i<kNGeomDirs; i++) {
    z->zoneNext[i]    = NULL; 
    // array of zones next to this zone,NULL or up to 6 zones completely sharing boundary
    // could point to self for periodic boundaries!
    z->zoneBCType[i]  = kBoundaryTypeFlagComputed | kBCCKindFree; // default BCs are computed
    // z->zoneBCValues are set after the zone is initlaised only if BCType is a constant
  }
  
  rSuper->rn[kXDim]     *=  z->zoneRefine;
  rSuper->rn[kYDim]     *=  z->zoneRefine;
  rSuper->rn[kZDim]     *=  z->zoneRefine;
  
  nDims     = gLevels[level].levelDims;
  nBC       = gLevels[level].levelBCells;
  
  //  zls_InitGridCoords(&g, nDims, nBC, kAllocateAxes, rSuper);
  
  GridCoords levelGC = gLevels[level].levelGrid;
  
  zls_InitSubGridCoords(&g, &levelGC, kAllocateAxes, rLevel);
  
  //
  // compute number of cells in active area, not counting BCs.  
  // each cell can contain nVars and coeffs, this is just the grid size
  //
#ifdef SYM3D
  z->zoneNCells = (double)g.n[kXDim]*g.n[kYDim]*g.n[kZDim];
#endif SYM3D
  
#ifdef SYM2D
  z->zoneNCells = (double)g.n[kXDim]*g.n[kYDim];
#endif SYM2D
  
#ifdef SYM1D
  z->zoneNCells = (double)g.n[kXDim];
#endif SYM1D
  
  z->zoneGrid   = g;
  
#ifdef PPM
  //  z->zoneFlat = NULL;
#endif PPM
  
#ifdef LAGRANGIAN
#ifdef SYM1D
  
  z->zone1DfluxX.f = NULL;
  
  zlsh_ClearW1DArr(&z->zoneW0X);
  zlsr_ClearWstar1DArr(&z->zoneWStarX);
  
#endif SYM1D
#ifdef SYM2D
  
  z->zone1DfluxX.f = NULL;
  z->zone1DfluxY.f = NULL;
  
  zlsh_ClearW1DArr(&z->zoneW0X);
  zlsh_ClearW1DArr(&z->zoneW0Y);
  
  zlsr_ClearWstar1DArr(&z->zoneWStarX);
  zlsr_ClearWstar1DArr(&z->zoneWStarY);
  
#endif SYM2D
#ifdef SYM3D
  
  z->zone1DfluxX.f = NULL;
  z->zone1DfluxY.f = NULL;
  z->zone1DfluxZ.f = NULL;
  
  zlsh_ClearW1DArr(&z->zoneW0X);
  zlsh_ClearW1DArr(&z->zoneW0Y);
  zlsh_ClearW1DArr(&z->zoneW0Z);
  
  zlsr_ClearWstar1DArr(&z->zoneWStarX);
  zlsr_ClearWstar1DArr(&z->zoneWStarY);
  zlsr_ClearWstar1DArr(&z->zoneWStarZ);
  
#endif SYM3D
  
#endif LAGRANGIAN
    
#ifdef DIVVHCORRECTION
  z->zoneDivV = NULL;
#endif DIVVHCORRECTION
  
  z->nQ   = kNAllVars;
  
  for (i = 0; i<kTotalAllVars; i++) {
#ifdef USEZONEQUAS
    z->zoneData[i]=NULL;
#endif
#ifdef USEZONEVARS
    z->zoneData[i].v=NULL;
#endif
  }
  
#ifdef MPI
  Integer dirIndex = 0;
  for (dirIndex = 0; dirIndex< kNGeomDirs; dirIndex++) {
    z->zoneBufferSize[dirIndex]    = 0L; // mpi uses int (ie signed!)
    z->zoneRecieveBuffer[dirIndex] = NULL;
    z->zoneSendBuffer[dirIndex]    = NULL; // allocate later if data is needed
  }
#endif MPI
  
  // init to no field, allocate later depending on algorithm in force
  z->zoneField.p = NULL;
  
#ifdef COOLING
  z->zoneMinCoolingTime = kZLS_Huge; 
#endif COOLING
  
  // only add algorithm zones to list, levels keep image zones elsewhere
  
  if ( kind == kAlgorithmZoneType ) err = zls_AddLevelZone (level, z);
  
  return(err);
  
}

//
// set up a two way link between two zones on the same level.  It is an error to try to link
// two zones on different levels.   The direction is from the point of view of the first zone z, 
// so kDirLeft will point the left link in z to the right side of zNext, and the right link of zNext
// to the left side of z.  The common axis co-ordinates (in this case y) must match.
// to link to a higher resolution area, use subzones contained within z, which is extended to 
// cover the high res region as normal.  This type of linking is to allow large areas to be 
// divided into smaller chunks to aid memory and parallel processing .  Also sub zones that share
// super boundaries can be linked across within their level, reducing the need to go up and over through 
// their two super zones that are linked.  Also the overall grid could grow during a simulation by adding linked
// zones to level 0.  The idea is that a group of linked zones on a level can be 
// calculated together without reference to the super level during the level update.
//
// Note: works with zone with or without BCs in their data.  
//
int     zls_SetZoneBCLink    ( ZonePtr z, ZonePtr zNext, Integer direction){
  
  int err = noErr;
  Integer zBCtype = kBoundaryTypeFlagNextZ;
  
  // zNext may be NULL, to unset a link in the direction, defaults to super BCs
  
  if ((z != NULL)&&(zNext!=NULL)){
    
    if (z->zoneLevel == zNext->zoneLevel) {
      
      // get  zone area parameters
      
      switch (direction){
#ifdef SYM1D
        case  kDirRight: // increasing index, 1st dim
          z->zoneBCType[kDirRight]    = zBCtype+kDirFlagRight;
          z->zoneNext  [kDirRight]    = zNext;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(z,kDirRight);
#endif MPI
          
          zNext->zoneBCType[kDirLeft] = zBCtype+kDirFlagLeft;
          zNext->zoneNext  [kDirLeft] = z;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(zNext,kDirLeft);
#endif MPI
          
          break;
        case  kDirLeft:// decreasing index, 1st dim
          z->zoneBCType[kDirLeft]    = zBCtype+kDirFlagLeft;
          z->zoneNext  [kDirLeft]    = zNext;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(z,kDirLeft);
#endif MPI
          
          zNext->zoneBCType[kDirRight] = zBCtype+kDirFlagRight;
          zNext->zoneNext  [kDirRight] = z;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(zNext,kDirRight);
#endif MPI
          break;
#endif SYM1D
#ifdef SYM2D
        case  kDirRight: // increasing index, 1st dim
          z->zoneBCType[kDirRight]    = zBCtype+kDirFlagRight;
          z->zoneNext  [kDirRight]    = zNext;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(z,kDirRight);
#endif MPI
          
          zNext->zoneBCType[kDirLeft] = zBCtype+kDirFlagLeft;
          zNext->zoneNext  [kDirLeft] = z;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(zNext,kDirLeft);
#endif MPI
          break;
        case  kDirLeft:// decreasing index, 1st dim
          z->zoneBCType[kDirLeft]    = zBCtype+kDirFlagLeft;
          z->zoneNext  [kDirLeft]    = zNext;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(z,kDirLeft);
#endif MPI
          
          zNext->zoneBCType[kDirRight] = zBCtype+kDirFlagRight;
          zNext->zoneNext  [kDirRight] = z;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(zNext,kDirRight);
#endif MPI
          break;
        case  kDirUp: // increasing index, 2nd dim
          z->zoneBCType[kDirUp]    = zBCtype+kDirFlagUp;
          z->zoneNext  [kDirUp]    = zNext;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(z,kDirUp);
#endif MPI
          
          zNext->zoneBCType[kDirDown] = zBCtype+kDirFlagDown;
          zNext->zoneNext  [kDirDown] = z;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(zNext,kDirDown);
#endif MPI
          break;
        case  kDirDown:// decreasing index, 2nd dim
          z->zoneBCType[kDirDown]    = zBCtype+kDirFlagDown;
          z->zoneNext  [kDirDown]    = zNext;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(z,kDirDown);
#endif MPI
          
          zNext->zoneBCType[kDirUp] = zBCtype+kDirFlagUp;
          zNext->zoneNext  [kDirUp] = z;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(zNext,kDirUp);
#endif MPI
          break;
#endif SYM2D
#ifdef SYM3D
        case  kDirRight: // increasing index, 1st dim
          z->zoneBCType[kDirRight]    = zBCtype+kDirFlagRight;
          z->zoneNext  [kDirRight]    = zNext;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(z,kDirRight);
#endif MPI
          
          zNext->zoneBCType[kDirLeft] = zBCtype+kDirFlagLeft;
          zNext->zoneNext  [kDirLeft] = z;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(zNext,kDirLeft);
#endif MPI
          break;
        case  kDirLeft:// decreasing index, 1st dim
          z->zoneBCType[kDirLeft]    = zBCtype+kDirFlagLeft;
          z->zoneNext  [kDirLeft]    = zNext;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(z,kDirLeft);
#endif MPI
          zNext->zoneBCType[kDirRight] = zBCtype+kDirFlagRight;
          zNext->zoneNext  [kDirRight] = z;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(zNext,kDirRight);
#endif MPI
          break;
        case  kDirUp: // increasing index, 2nd dim
          z->zoneBCType[kDirUp]    = zBCtype+kDirFlagUp;
          z->zoneNext  [kDirUp]    = zNext;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(z,kDirUp);
#endif MPI
          
          zNext->zoneBCType[kDirDown] = zBCtype+kDirFlagDown;
          zNext->zoneNext  [kDirDown] = z;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(zNext,kDirDown);
#endif MPI
          break;
        case  kDirDown:// decreasing index, 2nd dim
          z->zoneBCType[kDirDown]    = zBCtype+kDirFlagDown;
          z->zoneNext  [kDirDown]    = zNext;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(z,kDirDown);
#endif MPI
          
          zNext->zoneBCType[kDirUp] = zBCtype+kDirFlagUp;
          zNext->zoneNext  [kDirUp] = z;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(zNext,kDirUp);
#endif MPI
          break;
        case  kDirHither:// increasing index, 3rd dim
          z->zoneBCType[kDirHither]    = zBCtype+kDirFlagHither;
          z->zoneNext  [kDirHither]    = zNext;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(z,kDirHither);
#endif MPI
          
          zNext->zoneBCType[kDirYon] = zBCtype+kDirFlagYon;
          zNext->zoneNext  [kDirYon] = z;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(zNext,kDirYon);
#endif MPI
          break;
        case  kDirYon: // decreasing index, 3rd dim
          z->zoneBCType[kDirYon]    = zBCtype+kDirFlagYon;
          z->zoneNext  [kDirYon]    = zNext;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(z,kDirYon);
#endif MPI
          
          zNext->zoneBCType[kDirHither] = zBCtype+kDirFlagHither;
          zNext->zoneNext  [kDirHither] = z;
#ifdef MPI
          err = zls_AllocateZoneBufferDir(zNext,kDirHither);
#endif MPI
          break;
#endif SYM3D
        default:
          err = dimErr;
          zlsio_ErrPrintf( " zls_SetZoneBCLink: Invalid Link Direction code: %d , no link made.\n", direction);
      }
#if (ZLS_DEBUG>0)     
      zlsio_Printf( " zls_SetZoneBCLink: Zone: %x Zone id: %d, next Zone id: %d, Direction %s \n",z, z->zoneID,  zNext->zoneID, zls_dirname(direction) );
      zlsio_Printf( " zls_SetZoneBCLink: BC: %d, Link: %x %d \n", z->zoneBCType[direction] , z->zoneNext[direction], z->zoneNext[direction]->zoneID );
      zlsio_Printf( " zls_SetZoneBCLink: nextBC: %d, Link: %x %d \n\n", zNext->zoneBCType[zls_opposite(direction)] , zNext->zoneNext[zls_opposite(direction)], zNext->zoneNext[zls_opposite(direction)]->zoneID  );
#endif
      
    } else {  
      err = levErr;
      zlsio_ErrPrintf( " zls_SetZoneBCLink: Invalid Link, zones on different levels, no link made.\n" );
    }
  } else { 
    err = memErr; 
    zlsio_ErrPrintf( " zls_SetZoneBCLink: NULL zone/s:  no link made.\n" );
  } // z
  
  return(err);
}


int     zls_SetZoneBCInterp  ( ZonePtr z, Integer interp ){
  int err = noErr;
  
  if (z != NULL){
    z->zoneBCInterp = interp;
  } else { 
    err = memErr; 
    zlsio_ErrPrintf( " zls_SetZoneBCInterp: z = NULL .\n" );
  } // z
  
  return(err);
}

int     zls_SetZoneBCType    ( ZonePtr z, ZonePtr zNext, Integer direction, Integer zbctype ){ 
  // zNext can == NULL for types other than link
  int err = noErr;
  
  Integer bdryType     = zbctype & kBoundaryTypeMask; // basic type without qualifiers
  
  if (z != NULL){
    
    z->zoneBCType[direction] = zbctype; // set the type
    
    if ((bdryType & kBoundaryTypeFlagNextZ) == kBoundaryTypeFlagNextZ) {
      err = zls_SetZoneBCLink    ( z,  zNext,  direction);
    }
    
    if ( ( bdryType  & kBoundaryTypeFlagComputed ) == kBoundaryTypeFlagComputed ){ 
      // nothing special to do here.  Constant values are set after the zone is inited usually
    }
    
    
  } else { 
    err = memErr; 
    zlsio_ErrPrintf( " zls_SetZoneBCType: z = NULL .\n" );
  } // z
  
  return(err);
}

int  zls_SetZoneConstantBCValues( ZonePtr z ){
  Integer err = noErr;
  
  Integer dirIndex;
  Integer bdryType;
  Integer calcKind;
  
  //
  // works with or without BCs as it copies first values in active are into BC values array
  // which can be applied as BCs any time
  //
  
  Integer    id; //  loop counters
  
  if (z != NULL){
    
    // get  zone area parameters
    
#ifdef SYM1D
    
    Q1DArr           q;
    
    Integer is    = z->zoneGrid.ns[kXDim];
    Integer ie    = z->zoneGrid.ne[kXDim];
    
    for (id = 0; id <  z->nQ; id++){
      if (z->zoneData[id]!=NULL){
        
        q = z->zoneData[id];
        
        if (q!=NULL){
          
          // Set any constant or user BCs, all other types are set when the zone was created
          
          for (dirIndex = 0; dirIndex<kNGeomDirs;dirIndex++){
            
            bdryType = z->zoneBCType[dirIndex]&kBoundaryTypeFlagComputed;
            calcKind = z->zoneBCType[dirIndex]&kBCCKindMask;
            
            if ( bdryType == kBoundaryTypeFlagComputed ){ 
              
              if ( calcKind == kBCCKindConstant ){ 
                
                switch (dirIndex)  {
                  case kDirLeft:
                    z->zoneBCValues[dirIndex][id] = q[is];
                    break;
                  case kDirRight:
                    z->zoneBCValues[dirIndex][id] = q[ie-1];
                } // direction switch
                
              } //calcKind
              
            } // bdryType
          } // dirIndex
          
        } else { err = memErr; } // q
      } // z->zoneData[id]!=NULL
    } // var id loop
#endif SYM1D
    
#ifdef SYM2D
    
    Q2DArr           q;
    
    Integer is    = z->zoneGrid.ns[kXDim];
    Integer ie    = z->zoneGrid.ne[kXDim];
    
    Integer je    = z->zoneGrid.ne[kYDim];
    Integer js    = z->zoneGrid.ns[kYDim];
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    Integer nti   = z->zoneGrid.nt[kXDim];
    Integer ntj   = z->zoneGrid.nt[kYDim];
#endif ZONEDATA_ARE_PLAINBLOCKS
    
    for (id = 0; id <  z->nQ; id++){
      if (z->zoneData[id]!=NULL){
#ifdef ZONEDATA_ARE_PLAINBLOCKS
        zls_SetQuantity2DArr(&q, z->zoneData[id], ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
#ifdef ZONEDATA_ARE_MULTIDARRAYS
        q = z->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
        if (q!=NULL){
          
          // Set any constant or user BCs, all other types are set when the zone was created
          
          for (dirIndex = 0; dirIndex<kNGeomDirs;dirIndex++){
            
            bdryType = z->zoneBCType[dirIndex]&kBoundaryTypeFlagComputed;
            calcKind = z->zoneBCType[dirIndex]&kBCCKindMask;
            
            if ( bdryType == kBoundaryTypeFlagComputed ){ 
              
              if ( calcKind == kBCCKindConstant ){ 
                
                switch (dirIndex)  {
                  case kDirLeft:
                    z->zoneBCValues[dirIndex][id] = q[(js+je)/2][is];
                    break;
                  case kDirRight:
                    z->zoneBCValues[dirIndex][id] = q[(js+je)/2][ie-1];
                    break;
                  case kDirDown:
                    z->zoneBCValues[dirIndex][id] = q[js][(is+ie)/2];
                    break;
                  case kDirUp:
                    z->zoneBCValues[dirIndex][id] = q[je-1][(is+ie)/2];
                    break;
                } // direction switch
                
              } //calcKind
              
            } // bdryType
          } // dirIndex
          
#ifdef ZONEDATA_ARE_PLAINBLOCKS
          zls_ReleaseQuantity2DArr(q);
#endif ZONEDATA_ARE_PLAINBLOCKS
        } else { err = memErr; } // q
      } // z->zoneData[id]!=NULL
    } // var id loop
    
#endif SYM2D
    
#ifdef SYM3D
    Q3DArr           q; // no significant extra memory, wrapped around each quantity in turn and released if needed
    
    Integer is    = z->zoneGrid.ns[kXDim];
    Integer ie    = z->zoneGrid.ne[kXDim];
    
    Integer je    = z->zoneGrid.ne[kYDim];
    Integer js    = z->zoneGrid.ns[kYDim];
    
    Integer ke    = z->zoneGrid.ne[kZDim];
    Integer ks    = z->zoneGrid.ns[kZDim];
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    Integer nti   = z->zoneGrid.nt[kXDim];
    Integer ntj   = z->zoneGrid.nt[kYDim];
    Integer ntk   = z->zoneGrid.nt[kZDim];
#endif ZONEDATA_ARE_PLAINBLOCKS
    
    for (id = 0; id <  z->nQ; id++){
      if (z->zoneData[id]!=NULL){
#ifdef ZONEDATA_ARE_PLAINBLOCKS
        zls_SetQuantity3DArr(&q, z->zoneData[id], ntk, ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
#ifdef ZONEDATA_ARE_MULTIDARRAYS
        q = z->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
        if (q!=NULL){
          
          // Set any constant or user BCs, all other types are set when the zone was created
          
          for (dirIndex = 0; dirIndex<kNGeomDirs;dirIndex++){
            
            bdryType = z->zoneBCType[dirIndex]&kBoundaryTypeFlagComputed;
            calcKind = z->zoneBCType[dirIndex]&kBCCKindMask;
            
            if ( bdryType == kBoundaryTypeFlagComputed ){ 
              
              if ( calcKind == kBCCKindConstant ){ 
                
                switch (dirIndex)  {
                  case kDirLeft:
                    z->zoneBCValues[dirIndex][id] = q[(ks+ke)/2][(js+je)/2][is];
                    break;
                  case kDirRight:
                    z->zoneBCValues[dirIndex][id] = q[(ks+ke)/2][(js+je)/2][ie-1];
                    break;
                  case kDirDown:
                    z->zoneBCValues[dirIndex][id] = q[(ks+ke)/2][js][(is+ie)/2];
                    break;
                  case kDirUp:
                    z->zoneBCValues[dirIndex][id] = q[(ks+ke)/2][je-1][(is+ie)/2];
                    break;
                  case kDirYon:
                    z->zoneBCValues[dirIndex][id] = q[ks][(js+je)/2][(is+ie)/2];
                    break;
                  case kDirHither:
                    z->zoneBCValues[dirIndex][id] = q[ke-1][(js+je)/2][(is+ie)/2];
                    break;
                } // direction switch
                
              } //calcKind
              
            } // bdryType
          } // dirIndex
          
#ifdef ZONEDATA_ARE_PLAINBLOCKS
          zls_ReleaseQuantity3DArr(q);
#endif ZONEDATA_ARE_PLAINBLOCKS
        } else { err = memErr; } // q
      } // z->zoneData[id]!=NULL
    } // var id loop
    
#endif SYM3D
    
  } else { err = memErr; } // z
  
  return (err);
  
}


// always print coords to a standard file

int    zls_PrintZoneCoords    ( FILE * f, ZonePtr z){
  
  int err = noErr;
  
  if (z!=NULL){
    
    zlsio_FPrintf( f, " Zone cells and coordinate range in level grid:\n" );
    err = zls_PrintRange( f, &(z->zoneSuperRange));
    
    zlsio_FPrintf( f, " Zone Grid:\n" );
    err = zls_PrintGridCoords( f, &(z->zoneGrid));
    
  } else {err = memErr;}
  
  return(err);
  
}


// print coords to a gz file, or a standard file if io compression is off.

int    zls_ZPrintZoneCoords    ( zFile f, ZonePtr z){
  
  int err = noErr;
  
  if (z!=NULL){
    
    zlsio_ZFPrintf( f, " Zone cells and coordinate range in level grid:\n" );
    err = zls_ZPrintRange( f, &(z->zoneSuperRange));
    zlsio_ZFPrintf( f, " Zone Grid:\n" );
    err = zls_ZPrintGridCoords( f, &(z->zoneGrid));
    
  } else {err = memErr;}
  
  return(err);
  
}
// print coords to a gz file, or a standard file if io compression is off.

int    zls_UserLogPrintZoneCoords    ( ZonePtr z){
  
  int err = noErr;
  
  if (z!=NULL){
    
    zlsio_UserLogPrintf( " Zone cells and coordinate range in level grid:\n" );
    err = zls_UserLogPrintRange( &(z->zoneSuperRange));
    zlsio_UserLogPrintf( " Zone Grid:\n" );
    err = zls_UserLogPrintGridCoords( &(z->zoneGrid));
    
  } else {err = memErr;}
  
  return(err);
  
}

int   zls_NewZoneData(ZonePtr z, Integer varIndex){
  int err = noErr;
  
#ifdef SYM1D
  Counter  ni;
#endif SYM1D
#ifdef SYM2D
  Counter  ni,  nj;
#endif SYM2D
#ifdef SYM3D
  Counter  ni,  nj,  nk;
#endif SYM3D
  
#ifdef ZONEDATA_ARE_MULTIDARRAYS
#ifdef SYM1D
  Q1DArr q;
#endif SYM1D
#ifdef SYM2D
  Q2DArr q;
#endif SYM2D
#ifdef SYM3D
  Q3DArr q;
#endif SYM3D
#endif ZONEDATA_ARE_MULTIDARRAYS
  
#ifdef ZONEDATA_ARE_PLAINBLOCKS
  QPtr   d;
#endif ZONEDATA_ARE_PLAINBLOCKS
  
  if (z!=NULL){
#ifdef MPI    
    if (z->zoneRankID == gMPIrank) {
#endif MPI    
      
      if ((varIndex >= 0)&&(varIndex < kNAllVars)){
        
        zls_DisposeZoneData(z, varIndex);
        
#ifdef ZONEDATA_ARE_PLAINBLOCKS
        
#ifdef SYM1D
        ni = z->zoneGrid.nt[kXDim];
        err = zls_NewQuantity1D(&d, ni);
        bzero((void *) d, ni*sizeof(Real)); 
#endif SYM1D
        
#ifdef SYM2D
        ni = z->zoneGrid.nt[kXDim];
        nj = z->zoneGrid.nt[kYDim];
        err = zls_NewQuantity2D(&d, ni, nj);
        bzero((void *) d, ni*nj*sizeof(Real)); 
#endif SYM2D
        
#ifdef SYM3D
        ni = z->zoneGrid.nt[kXDim];
        nj = z->zoneGrid.nt[kYDim];
        nk = z->zoneGrid.nt[kZDim];
        err = zls_NewQuantity3D(&d, ni, nj, nk);
        bzero((void *) d, ni*nj*nk*sizeof(Real)); 
#endif SYM3D
        z->zoneData[varIndex] = d;  
        
#endif ZONEDATA_ARE_PLAINBLOCKS
        
#ifdef ZONEDATA_ARE_MULTIDARRAYS
        
#ifdef SYM1D
        ni = z->zoneGrid.nt[kXDim];
        err = zls_NewQuantity1D(&q, ni);
        bzero((void *) q, ni* sizeof(Real)); 
#endif SYM1D
        
#ifdef SYM2D
        ni = z->zoneGrid.nt[kXDim];
        nj = z->zoneGrid.nt[kYDim];
        err = zls_NewQuantity2DArr(&q, nj, ni);
        zlsZero2DArr((void **)q, sizeof(Real), nj, ni);
#endif SYM2D
        
#ifdef SYM3D
        ni = z->zoneGrid.nt[kXDim];
        nj = z->zoneGrid.nt[kYDim];
        nk = z->zoneGrid.nt[kZDim];
        err = zls_NewQuantity3DArr(&q, nk, nj, ni);
        zlsZero3DArr((void ***)q, sizeof(Real), nk, nj, ni);
#endif SYM3D
        
        z->zoneData[varIndex] = q;
        
#endif ZONEDATA_ARE_MULTIDARRAYS
        
        
      } else {err = dimErr;} // varIndex
      
      
#ifdef MPI    
    }
#endif MPI    
    
  } else {err = memErr;} // z
  
  return(err);
}

#ifdef MEANSPLITSWEEPS
int   zls_NewZoneDataAlt(ZonePtr z, Integer varIndex){
  int err = noErr;
  
#ifdef SYM1D
  Counter  ni;
#endif SYM1D
#ifdef SYM2D
  Counter  ni,  nj;
#endif SYM2D
#ifdef SYM3D
  Counter  ni,  nj,  nk;
#endif SYM3D
  
#ifdef ZONEDATA_ARE_MULTIDARRAYS
#ifdef SYM1D
  Q1DArr q;
#endif SYM1D
#ifdef SYM2D
  Q2DArr q;
#endif SYM2D
#ifdef SYM3D
  Q3DArr q;
#endif SYM3D
#endif ZONEDATA_ARE_MULTIDARRAYS
  
#ifdef ZONEDATA_ARE_PLAINBLOCKS
  QPtr   d;
#endif ZONEDATA_ARE_PLAINBLOCKS
  
  if (z!=NULL){
#ifdef MPI    
    if (z->zoneRankID == gMPIrank) {
#endif MPI    
      
      if ((varIndex >= 0)&&(varIndex < kNAllVars)){
        
        zls_DisposeZoneDataAlt(z, varIndex);
        
#ifdef ZONEDATA_ARE_PLAINBLOCKS
        
#ifdef SYM1D
        ni = z->zoneGrid.nt[kXDim];
        err = zls_NewQuantity1D(&d, ni);
        bzero((void *) d, ni*sizeof(Real)); 
#endif SYM1D
        
#ifdef SYM2D
        ni = z->zoneGrid.nt[kXDim];
        nj = z->zoneGrid.nt[kYDim];
        err = zls_NewQuantity2D(&d, ni, nj);
        bzero((void *) d, ni*nj*sizeof(Real)); 
#endif SYM2D
        
#ifdef SYM3D
        ni = z->zoneGrid.nt[kXDim];
        nj = z->zoneGrid.nt[kYDim];
        nk = z->zoneGrid.nt[kZDim];
        err = zls_NewQuantity3D(&d, ni, nj, nk);
        bzero((void *) d, ni*nj*nk*sizeof(Real)); 
#endif SYM3D
        z->zoneData[varIndex+kAltVarsOffset] = d;  
        
#endif ZONEDATA_ARE_PLAINBLOCKS
        
#ifdef ZONEDATA_ARE_MULTIDARRAYS
        
#ifdef SYM1D
        ni = z->zoneGrid.nt[kXDim];
        err = zls_NewQuantity1D(&q, ni);
        bzero((void *) q, ni* sizeof(Real)); 
#endif SYM1D
        
#ifdef SYM2D
        ni = z->zoneGrid.nt[kXDim];
        nj = z->zoneGrid.nt[kYDim];
        err = zls_NewQuantity2DArr(&q, nj, ni);
        zlsZero2DArr((void **)q, sizeof(Real), nj, ni);
#endif SYM2D
        
#ifdef SYM3D
        ni = z->zoneGrid.nt[kXDim];
        nj = z->zoneGrid.nt[kYDim];
        nk = z->zoneGrid.nt[kZDim];
        err = zls_NewQuantity3DArr(&q, nk, nj, ni);
        zlsZero3DArr((void ***)q, sizeof(Real), nk, nj, ni);
#endif SYM3D
        
        z->zoneData[varIndex+kAltVarsOffset] = q;
        
#endif ZONEDATA_ARE_MULTIDARRAYS
        
        
      } else {err = dimErr;} // varIndex
      
      
#ifdef MPI    
    }
#endif MPI    
    
  } else {err = memErr;} // z
  
  return(err);
}
#endif MEANSPLITSWEEPS

int zls_NewZoneAuxilliaryData( ZonePtr z ){
  int err = noErr;
  
  // axes are allocated elsewhere
  
  
  if ( z != NULL ){
    
#ifdef MPI    
    if (z->zoneRankID == gMPIrank) {
#endif MPI    
      
      
      Integer     nBC   = z->zoneBCells;
      GridCoords  gc    = z->zoneGrid;
      
#ifdef SYM1D
      Counter  nti;
      nti = gc.nt[kXDim];
#endif SYM1D
#ifdef SYM2D
      Counter  nti,  ntj;
      nti = gc.nt[kXDim];
      ntj = gc.nt[kYDim];
#endif SYM2D
#ifdef SYM3D
      Counter  nti,  ntj,  ntk;
      nti = gc.nt[kXDim];
      ntj = gc.nt[kYDim];
      ntk = gc.nt[kZDim];
#endif SYM3D
      
#ifdef SYM1D
      Counter  ni;
      ni = gc.n[kXDim];
#endif SYM1D
#ifdef SYM2D
      Counter  ni,  nj;
      ni = gc.n[kXDim];
      nj = gc.n[kYDim];
#endif SYM2D
#ifdef SYM3D
      Counter  ni,  nj,  nk;
      ni = gc.n[kXDim];
      nj = gc.n[kYDim];
      nk = gc.n[kZDim];
#endif SYM3D
      
#ifdef LAGRANGIAN
      
#ifdef SYM1D
      Flx1DArr flx;
      err = zlsh_NewWFluxes1DArr   (&flx, nBC, ni);
      err = zlsh_ClearWFluxes1DArr (flx);
      z->zone1DfluxX = flx;
      
      W1DArr w1;
      err = zlsh_NewW1DArr (&w1, nBC, ni);
      z->zoneW0X = w1;
      
      Wstar1DArr ws;
      err = zlsr_NewWstar1DArr  (&ws, nBC, ni);
      z->zoneWStarX = ws;
#endif SYM1D
      
#ifdef SYM2D
      Flx1DArr flx;
      err = zlsh_NewWFluxes1DArr   (&flx, nBC, ni);
      err = zlsh_ClearWFluxes1DArr (flx);
      z->zone1DfluxX = flx;
      err = zlsh_NewWFluxes1DArr   (&flx, nBC, nj);
      err = zlsh_ClearWFluxes1DArr (flx);
      z->zone1DfluxY = flx;
      
      W1DArr w1;
      err = zlsh_NewW1DArr (&w1, nBC, ni);
      z->zoneW0X = w1;
      err = zlsh_NewW1DArr (&w1, nBC, nj);
      z->zoneW0Y = w1;
      
      Wstar1DArr ws;
      err = zlsr_NewWstar1DArr  (&ws, nBC, ni);
      z->zoneWStarX = ws;
      err = zlsr_NewWstar1DArr  (&ws, nBC, nj);
      z->zoneWStarY = ws;
#endif SYM2D
#ifdef SYM3D
      Flx1DArr flx;
      err = zlsh_NewWFluxes1DArr   (&flx, nBC, ni);
      err = zlsh_ClearWFluxes1DArr (flx);
      z->zone1DfluxX = flx;
      err = zlsh_NewWFluxes1DArr   (&flx, nBC, nj);
      err = zlsh_ClearWFluxes1DArr (flx);
      z->zone1DfluxY = flx;
      err = zlsh_NewWFluxes1DArr   (&flx, nBC, nk);
      err = zlsh_ClearWFluxes1DArr (flx);
      z->zone1DfluxZ = flx;
      
      W1DArr w1;
      err = zlsh_NewW1DArr (&w1, nBC, ni);
      z->zoneW0X = w1;
      err = zlsh_NewW1DArr (&w1, nBC, nj);
      z->zoneW0Y = w1;
      err = zlsh_NewW1DArr (&w1, nBC, nk);
      z->zoneW0Z = w1;
      
      Wstar1DArr ws;
      err = zlsr_NewWstar1DArr  (&ws, nBC, ni);
      z->zoneWStarX = ws;
      err = zlsr_NewWstar1DArr  (&ws, nBC, nj);
      z->zoneWStarY = ws;
      err = zlsr_NewWstar1DArr  (&ws, nBC, nk);
      z->zoneWStarZ = ws;
#endif SYM3D
#endif LAGRANGIAN
      
      
#ifdef GRAVITY
#ifdef SYM1D
      F1DArr f;
      err = zlsh_NewF1DArr  (&f, nBC, ni);
      z->zoneF0X = f;
#endif SYM1D
#ifdef SYM2D
      F1DArr f;
      err = zlsh_NewF1DArr  (&f, nBC, ni);
      z->zoneF0X = f;
      err = zlsh_NewF1DArr  (&f, nBC, nj);
      z->zoneF0Y = f;
#endif SYM2D
#ifdef SYM3D
      F1DArr f;
      err = zlsh_NewF1DArr  (&f, nBC, ni);
      z->zoneF0X = f;
      err = zlsh_NewF1DArr  (&f, nBC, nj);
      z->zoneF0Y = f;
      err = zlsh_NewF1DArr  (&f, nBC, nk);
      z->zoneF0Z = f;
#endif SYM3D
      
#endif GRAVITY
      
#ifdef DIVVHCORRECTION
      VQArr vq = NULL;
#ifdef SYM1D
      zls_NewVQuantity1DArr(&vq, nti);
      bzero((void *) vq, ni* sizeof(QVec)); 
#endif SYM1D
#ifdef SYM2D
      err = zls_NewVQuantity2DArr(&vq, ntj, nti);
      zlsZero2DArr((void **)vq, sizeof(QVec), ntj, nti);
#endif SYM2D
#ifdef SYM3D
      err = zls_NewVQuantity3DArr(&vq, ntk, ntj, nti);
      zlsZero3DArr((void ***)vq, sizeof(QVec), ntk, ntj, nti);
#endif SYM3D
      z->zoneDivV = vq;
#endif DIVVHCORRECTION  
      
#ifdef MPI
      //
      // under MPI zone data needs contiguous BC buffers for exchange with remote 
      // processes, but they are allocated as links are created
      Integer dirIndex = 0;
      for (dirIndex = 0; dirIndex< kNGeomDirs; dirIndex++) {
        z->zoneBufferSize[dirIndex]     = 0;
        z->zoneRecieveBuffer[dirIndex]  = NULL;
        z->zoneSendBuffer   [dirIndex]  = NULL; 
      }
      // otherwise you could call allocatebuffers here to preallocate all buffers
#endif MPI
      
#ifdef MPI    
    }
#endif MPI    
  }
  
  return( err );
}


int   zls_AllocateZoneBufferDir(ZonePtr z, Integer dir){
  int err = noErr;
  
#ifdef MPI    
  
  Counter  ni,  nj,  nk;
  
  if (z!=NULL){
    if (z->zoneRankID == gMPIrank) {
      
      if (z->zoneRecieveBuffer[dir]!= NULL) free(z->zoneRecieveBuffer[dir]); z->zoneRecieveBuffer[dir] = NULL;
      if (z->zoneSendBuffer[dir]   != NULL) free(z->zoneSendBuffer[dir])   ; z->zoneSendBuffer[dir] = NULL;
      
      //
      // under MPI zone data needs contiguous BC buffers for exchange with remote 
      // processes
      
      GridCoords gc = z->zoneGrid;
      zls_GetGridDirBCDims  ( &ni, &nj, &nk, &gc,  dir );
      
      QPtr d = NULL;
      int bufferSize =0; // MPI uses int, make sure it doesn't overflow
#ifdef SYM1D
      bufferSize = kNAllVars*ni;
#endif SYM1D
#ifdef SYM2D
      bufferSize = kNAllVars*ni*nj;
#endif SYM2D
#ifdef SYM3D
      bufferSize = kNAllVars*ni*nj*nk;
#endif SYM3D
      err = zls_NewQuantity1D(&d, bufferSize);
      bzero((void *) d, bufferSize*sizeof(Real)); 
      z->zoneBufferSize[dir]    = bufferSize;
      
      z->zoneRecieveBuffer[dir] = d;
      
      d = NULL;
      err = zls_NewQuantity1D(&d, bufferSize);
      bzero((void *) d, bufferSize*sizeof(Real)); 
      z->zoneSendBuffer   [dir]  = d; 
      
    }
    
  } else {err = memErr;} // z
#endif MPI    
  
  return(err);
}

int   zls_AllocateZoneBuffers(ZonePtr z){
  int err = noErr;
  
#ifdef MPI    
  Integer dir;
  
  for ( dir = 0; dir < kNGeomDirs ; dir++){
    err = err = zls_AllocateZoneBufferDir(z, dir);
    if (err != noErr) { return(err);}
  }
  
#endif MPI    
  
  return(err);
}


int     zls_DisposeZoneData   (ZonePtr z, Integer varIndex){
  
  int err = noErr;
  
  if (z!=NULL){
    
#ifdef MPI    
    if (z->zoneRankID == gMPIrank) {
#endif MPI   
      
      if ((varIndex >= 0)&&(varIndex < kNAllVars)){
        
#ifdef ZONEDATA_ARE_PLAINBLOCKS 
        if (z->zoneData[varIndex] != NULL) free(z->zoneData[varIndex]); z->zoneData[varIndex] = NULL;
#endif ZONEDATA_ARE_PLAINBLOCKS 
        
#ifdef ZONEDATA_ARE_MULTIDARRAYS  
        
#ifdef SYM1D
        if (z->zoneData[varIndex] != NULL) 
          free(z->zoneData[varIndex]); 
        z->zoneData[varIndex] = NULL;
#endif SYM1D
        
#ifdef SYM2D
        zls_DisposeQuantity2DArr(z->zoneData[varIndex]); z->zoneData[varIndex] = NULL;
#endif SYM2D
        
#ifdef SYM3D
        zls_DisposeQuantity3DArr(z->zoneData[varIndex]); z->zoneData[varIndex] = NULL;
#endif SYM3D
        
#endif ZONEDATA_ARE_MULTIDARRAYS
        
      } else {err = dimErr;}
      
#ifdef MPI    
    }
#endif MPI    
    
    
  } //z
  
  return(err);
}

#ifdef MEANSPLITSWEEPS

int     zls_DisposeZoneDataAlt (ZonePtr z, Integer varIndex){
  
  int err = noErr;
  
  if (z!=NULL){
    
#ifdef MPI    
    if (z->zoneRankID == gMPIrank) {
#endif MPI   
      
      if ((varIndex >= 0)&&(varIndex < kNAllVars)){
        
#ifdef ZONEDATA_ARE_PLAINBLOCKS 
        if (z->zoneData[varIndex+kAltVarsOffset] != NULL) free(z->zoneData[varIndex+kAltVarsOffset]); z->zoneData[varIndex+kAltVarsOffset] = NULL;
#endif ZONEDATA_ARE_PLAINBLOCKS 
        
#ifdef ZONEDATA_ARE_MULTIDARRAYS  
        
#ifdef SYM1D
        if (z->zoneData[varIndex+kAltVarsOffset] != NULL) 
          free(z->zoneData[varIndex+kAltVarsOffset]); 
        z->zoneData[varIndex+kAltVarsOffset] = NULL;
#endif SYM1D
        
#ifdef SYM2D
        zls_DisposeQuantity2DArr(z->zoneData[varIndex+kAltVarsOffset]); z->zoneData[varIndex+kAltVarsOffset] = NULL;
#endif SYM2D
        
#ifdef SYM3D
        zls_DisposeQuantity3DArr(z->zoneData[varIndex+kAltVarsOffset]); z->zoneData[varIndex+kAltVarsOffset] = NULL;
#endif SYM3D
        
#endif ZONEDATA_ARE_MULTIDARRAYS
        
      } else {err = dimErr;}
      
#ifdef MPI    
    }
#endif MPI    
    
    
  } //z
  
  return(err);
}
#endif MEANSPLITSWEEPS

int     zls_DisposeZoneAuxilliaryData  (ZonePtr z){
  
  int err = noErr;
  
  if (z!=NULL){
    
#ifdef SYM1D
    zls_DisposeAxis (z->zoneGrid.x);
    zls_DisposeAxis (z->zoneGrid.x0);
#endif SYM1D
#ifdef SYM2D
    zls_DisposeAxis (z->zoneGrid.x);
    zls_DisposeAxis (z->zoneGrid.x0);
    zls_DisposeAxis (z->zoneGrid.y);
    zls_DisposeAxis (z->zoneGrid.y0);
#endif SYM2D
#ifdef SYM3D
    zls_DisposeAxis (z->zoneGrid.x);
    zls_DisposeAxis (z->zoneGrid.x0);
    zls_DisposeAxis (z->zoneGrid.y);
    zls_DisposeAxis (z->zoneGrid.y0);
    zls_DisposeAxis (z->zoneGrid.z);
    zls_DisposeAxis (z->zoneGrid.z0);
#endif SYM3D
    
#ifdef MPI
    if ( z->zoneRankID == gMPIrank ){
#endif MPI      
      
#ifdef LAGRANGIAN
      
#ifdef SYM1D
      zlsh_DisposeWFluxes1DArr(z->zone1DfluxX);
      zlsh_DisposeW1DArr(z->zoneW0X);
      zlsr_DisposeWstar1DArr(z->zoneWStarX);
#endif SYM1D
      
#ifdef SYM2D
      zlsh_DisposeWFluxes1DArr(z->zone1DfluxX);
      zlsh_DisposeWFluxes1DArr(z->zone1DfluxY);
      
      zlsh_DisposeW1DArr(z->zoneW0X);
      zlsh_DisposeW1DArr(z->zoneW0Y);
      
      zlsr_DisposeWstar1DArr(z->zoneWStarX);
      zlsr_DisposeWstar1DArr(z->zoneWStarY);
      
#endif SYM2D
      
#ifdef SYM3D
      zlsh_DisposeWFluxes1DArr(z->zone1DfluxX);
      zlsh_DisposeWFluxes1DArr(z->zone1DfluxY);
      zlsh_DisposeWFluxes1DArr(z->zone1DfluxZ);
      
      zlsh_DisposeW1DArr(z->zoneW0X);
      zlsh_DisposeW1DArr(z->zoneW0Y);
      zlsh_DisposeW1DArr(z->zoneW0Z);
      
      zlsr_DisposeWstar1DArr(z->zoneWStarX);
      zlsr_DisposeWstar1DArr(z->zoneWStarY);
      zlsr_DisposeWstar1DArr(z->zoneWStarZ);
#endif SYM3D
#endif LAGRANGIAN
          
#ifdef GRAVITY
      
#ifdef SYM1D
      zlsh_DisposeF1DArr(z->zoneF0X);
#endif SYM1D
      
#ifdef SYM2D
      zlsh_DisposeF1DArr(z->zoneF0X);
      zlsh_DisposeF1DArr(z->zoneF0Y);
#endif SYM2D
      
#ifdef SYM3D
      zlsh_DisposeF1DArr(z->zoneF0X);
      zlsh_DisposeF1DArr(z->zoneF0Y);
      zlsh_DisposeF1DArr(z->zoneF0Z);
#endif SYM3D
      
#endif GRAVITY
      
#ifdef DIVVHCORRECTION
#ifdef SYM1D
      zls_DisposeVQuantity1DArr ( z->zoneDivV );
#endif SYM1D
#ifdef SYM2D
      zls_DisposeVQuantity2DArr( z->zoneDivV );
#endif SYM2D
#ifdef SYM3D
      zls_DisposeVQuantity3DArr( z->zoneDivV );
#endif SYM3D
      z->zoneDivV=NULL;
#endif DIVVHCORRECTION
      
#ifdef MPI
      if ( z->zoneRankID == gMPIrank ){
        Integer dirIndex = 0;
        for (dirIndex = 0; dirIndex< kNGeomDirs; dirIndex++) {
          zls_DisposeQuantity ( z->zoneRecieveBuffer[dirIndex]);
          zls_DisposeQuantity ( z->zoneSendBuffer[dirIndex]);
          z->zoneBufferSize [dirIndex] = 0;
        }
      }
#endif MPI
      
#ifdef MPI    
    }
#endif MPI    
    
    
  } //z
  
  return(err);
}


int   zls_NewZoneField(ZonePtr z){
  
  int err = noErr;
#ifndef GRAVITY
#pragma unused(z)
#else
  
#ifdef SYM1D
  Integer  nti;
#endif SYM1D
  
#ifdef SYM2D
  Integer  nti,  ntj;
#endif SYM2D
  
#ifdef SYM3D
  Integer  nti,  ntj,  ntk;
#endif SYM3D
  
  Field f;
  
  if (z!=NULL){
#ifdef MPI    
    if (z->zoneRankID == gMPIrank) {
#endif MPI    
      
      zls_DisposeZoneField(z);
      
#ifdef SYM1D
      nti = z->zoneGrid.nt[kXDim];
      err = zls_NewLRQ1DArr  (&f.p, nti);
      if ( err != noErr) {
        return(err);
      }
#endif SYM1D
      
#ifdef SYM2D
      nti = z->zoneGrid.nt[kXDim];
      ntj = z->zoneGrid.nt[kYDim];
      
      err = zls_NewLRQ2DArr  (&f.p, ntj, nti);
      if ( err != noErr) {
        return(err);
      }
#endif SYM2D
      
#ifdef SYM3D
      nti = z->zoneGrid.nt[kXDim];
      ntj = z->zoneGrid.nt[kYDim];
      ntk = z->zoneGrid.nt[kZDim];
      err = zls_NewLRQ3DArr  (&f.p, ntk, ntj, nti);
      if ( err != noErr) {
        return(err);
      }
#endif SYM3D
      
      zls_SetZoneField( z, f);
#ifdef MPI    
    }
#endif MPI    
    
  } else {err = memErr;} // z
  
#endif GRAVITY
  
  return(err);
}

int     zls_DisposeZoneField  (ZonePtr z){
  
  int err = noErr;
  
  if (z!=NULL){
#ifdef MPI    
    if (z->zoneRankID == gMPIrank) {
#endif MPI    
      
      Field f;
      zls_GetZoneField( &f, z );
      
      zls_DisposeField(f);
      z->zoneField.p = NULL;
#ifdef MPI    
    }
#endif MPI    
    
  } //z
  
  return(err);
}


int zls_SetZoneField( ZonePtr   z, Field f){
  int err = noErr;
  
  if (z!=NULL){
    z->zoneField.p = f.p;
  } //z
  
  return(err);
}

int zls_GetZoneField( Field * f,  ZonePtr z){
  int err = noErr;
  
  if (z!=NULL){
    f->p = z->zoneField.p;
  } //z
  
  return(err);
}


#ifdef PRIMITIVE_QUAS

int   zls_GetZoneWHD( WHD *w, ZonePtr z ){
  // not to be confused with zlsp_ReferenceZoneWHD which creates the references.
  int   err = noErr;
  *w =  z->zoneW;
  return(err);
}

// *must* be called after all the datas have been allocted successfully;
int   zls_SetZoneWHD( ZonePtr z ){
  int   err = noErr;
  WHD  w;
  err = zlsp_ReferenceZoneWHD( &w, z ); // this rotine checks data validity so OK to just go on err;
  if (err == noErr) { 
    z->zoneW = w;
#ifdef MEANSPLITSWEEPS
    z->zoneWLeft = w;
#endif MEANSPLITSWEEPS
  }
#ifdef MEANSPLITSWEEPS
  err = zlsp_ReferenceZoneWHDAlt( &w, z ); // this rotine checks data validity so OK to just go on err;
  if (err == noErr) { 
    z->zoneWRight = w;
  }
#endif MEANSPLITSWEEPS
  return(err);
}
#endif PRIMITIVE_QUAS

#ifdef CONSERVED_QUAS
int   zls_GetZoneUHD( UHD *u, ZonePtr z ){
  // not to be confused with zlsp_ReferenceZoneUHD which creates the references.
  int   err = noErr;
  *u =  z->zoneU;
  return(err);
}

// *must* be called after all the datas have been allocted successfully;
int   zls_SetZoneUHD( ZonePtr z ){
  int   err = noErr;
  UHD  u;
  err = zlsp_ReferenceZoneUHD( &u, z ); // this rotine checks data validity so OK to just go on err;
  if (err == noErr) { 
    z->zoneU = u;
#ifdef MEANSPLITSWEEPS
    z->zoneULeft = u;
#endif MEANSPLITSWEEPS
  }
#ifdef MEANSPLITSWEEPS
  err = zlsp_GetZoneUHDAlt(&u, z); // this rotine checks data validity so OK to just go on err;
  if (err == noErr) { 
    z->zoneURight = u;
  }
#endif MEANSPLITSWEEPS
  return(err);
}
#endif CONSERVED_QUAS

int   zls_SetZoneDataCopy(ZonePtr z, QPtr d, Integer varIndex){
  //
  // set the active area of zoneData with copy of d
  //
  // If data is from the outside (or otherwise without BCs), assume it matches 
  // the zone 0-n dimensions, then allocate a proper data 
  // with new zoneData and parse d, offsetting by ns.
  // if the zone doesn't use BCs then it is a simple copy, and
  // d is assumed to not include BCs either.
  // (Note zonedata now always includes BCs, only early development versions of ZLS had
  // zone data without and the conditional compilation became too complex,so the option
  // for zone data to be BC free was dropped.  They can still be plain blocks or multi-d
  // arrays though.
  //
  // Unfortunately there is no simple way to query the size of a malloc's block in plain C.
  // On the mac I could use GetPtrSize(d), but that is part of that specific OS.
  //
  // Big NOTE: here we just modify the contents of the zone data, so there is no need to 
  // fiddle around with setting the primitive quantity (or other) references, they should already be
  // set up by either zls_SetZoneDataDirect or zls_NewZoneData.  *NEVER* init the zonedata by hand,
  // Use NZD or SZDD and future references will also be OK.
  //
  int err = noErr;
  Integer nD;
#ifdef SYM1D
  Integer i;
  Integer ni;
  QPtr   q1z;
#endif SYM1D
#ifdef SYM2D
  Integer i, j;
  Integer ni, nj;
  Q2DArr q2z, q2d;
#endif SYM2D
#ifdef SYM3D
  Integer i, j, k;
  Integer ni, nj, nk;
  Q3DArr q3z, q3d;
#endif SYM3D
  
#ifdef SYM1D
  Integer nti;
  Integer is;
#endif SYM1D
#ifdef SYM2D
  Integer nti, ntj;
  Integer is, js;
#endif SYM2D
#ifdef SYM3D
  Integer nti, ntj, ntk;
  Integer is, js, ks;
#endif SYM3D
  
  if ((z!=NULL)&&(d!=NULL)){
    if ((varIndex >= 0)&&(varIndex < kNAllVars)){
      if (z->zoneData[varIndex] != NULL) free(z->zoneData[varIndex]);
      z->zoneData[varIndex] = NULL;
      err = zls_NewZoneData   (z, varIndex); 
      
      if (z->zoneData[varIndex]!=NULL){
        nD = z->zoneGrid.nD;
        
#ifdef SYM1D
        is = z->zoneGrid.ns[kXDim];
        
        ni = z->zoneGrid.n[kXDim];
        
        nti = z->zoneGrid.nt[kXDim];
#endif SYM1D
        
        
#ifdef SYM2D
        is = z->zoneGrid.ns[kXDim];
        js = z->zoneGrid.ns[kYDim];
        
        ni = z->zoneGrid.n[kXDim];
        nj = z->zoneGrid.n[kYDim];
        
        nti = z->zoneGrid.nt[kXDim];
        ntj = z->zoneGrid.nt[kYDim];
#endif SYM2D
        
        
#ifdef SYM3D
        is = z->zoneGrid.ns[kXDim];
        js = z->zoneGrid.ns[kYDim];
        ks = z->zoneGrid.ns[kZDim];
        
        ni = z->zoneGrid.n[kXDim];
        nj = z->zoneGrid.n[kYDim];
        nk = z->zoneGrid.n[kZDim];
        
        nti = z->zoneGrid.nt[kXDim];
        ntj = z->zoneGrid.nt[kYDim];
        ntk = z->zoneGrid.nt[kZDim];
#endif SYM3D
        
        
#ifdef SYM1D
        q1z =  z->zoneData[varIndex];
        for (i = 0; i<ni; i++){
          q1z[i+is] = d[i];
        };
#endif SYM1D
        
#ifdef SYM2D
        
#ifdef ZONEDATA_ARE_PLAINBLOCKS
        zls_SetQuantity2DArr(&q2z, z->zoneData[varIndex], ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
#ifdef ZONEDATA_ARE_MULTIDARRAYS
        q2z = z->zoneData[varIndex];
#endif ZONEDATA_ARE_MULTIDARRAYS
        zls_SetQuantity2DArr(&q2d, d, nj, ni);
        
        for (j = 0; j<nj; j++){
          for (i = 0; i<ni; i++){
            q2z[j+js][i+is] = q2d[j][i];
          }
        }
        
#ifdef ZONEDATA_ARE_PLAINBLOCKS
        zls_ReleaseQuantity2DArr(q2z);
#endif ZONEDATA_ARE_PLAINBLOCKS
        zls_ReleaseQuantity2DArr(q2d);
        
#endif SYM2D
        
#ifdef SYM3D
        
#ifdef ZONEDATA_ARE_PLAINBLOCKS
        zls_SetQuantity3DArr(&q3z, z->zoneData[varIndex], ntk, ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
#ifdef ZONEDATA_ARE_MULTIDARRAYS
        q3z = z->zoneData[varIndex];
#endif ZONEDATA_ARE_MULTIDARRAYS
        zls_SetQuantity3DArr(&q3d, d, nk, nj, ni);
        
        for (k = 0; k<nk; k++){
          for (j = 0; j<nj; j++){
            for (i = 0; i<ni; i++){
              q3z[k+ks][j+js][i+is] = q3d[k][j][i];
            } //i
          } //j
        } //k
        
#ifdef ZONEDATA_ARE_PLAINBLOCKS
        zls_ReleaseQuantity3DArr(q3z);
#endif ZONEDATA_ARE_PLAINBLOCKS
        zls_ReleaseQuantity3DArr(q3d);
        
#endif SYM3D
        
        
      } else { err = memErr;} // zone data ok 
      
    } else {err = dimErr;}
    
  }else {err = memErr;}
  
  return(err);
  
}

int   zls_SetZoneDataDirect(ZonePtr z, QPtr d, Integer varIndex){
  
  // set zoneData to equal the data ref d
  //
  // if the zone data has BCs included then d must be sized appropriately
  // or bad things will happen.
  //
  
  int err = noErr;
  
  // add code to check d is right size later, careful to fill in order
  
  if ((z!=NULL)&&(d!=NULL)){
    
    if ((varIndex >= 0)&&(varIndex < kNAllVars)){
      
#ifdef ZONEDATA_ARE_PLAINBLOCKS
      
      if (z->zoneData[varIndex] != NULL) free(z->zoneData[varIndex]);
      z->zoneData[varIndex] = NULL;
      
      // caller cannot free d, as it is still reffered to:
      z->zoneData[varIndex] = d; // simple isn't it!!!
      
#endif ZONEDATA_ARE_PLAINBLOCKS
      
#ifdef ZONEDATA_ARE_MULTIDARRAYS
      
#ifdef SYM2D
      Integer ni, nj;
#endif SYM2D
      
#ifdef SYM3D
      Integer ni, nj, nk;
#endif SYM3D
      
#ifdef SYM1D
      Q1DArr q;
      if (z->zoneData[varIndex] != NULL) free(z->zoneData[varIndex]);
      z->zoneData[varIndex] = NULL;
      q = d; // if needed below.
#endif SYM1D
      
#ifdef SYM2D
      Q2DArr q;
      zls_DisposeQuantity2DArr(z->zoneData[varIndex]);
      z->zoneData[varIndex] = NULL;
      
      ni = z->zoneGrid.nt[kXDim];
      nj = z->zoneGrid.nt[kYDim];
      
      zls_SetQuantity2DArr(&q, d, nj, ni);
#endif SYM2D
      
#ifdef SYM3D
      Q3DArr q;
      zls_DisposeQuantity3DArr(z->zoneData[varIndex]);
      z->zoneData[varIndex] = NULL;
      
      ni = z->zoneGrid.nt[kXDim];
      nj = z->zoneGrid.nt[kYDim];
      nk = z->zoneGrid.nt[kZDim];
      
      zls_SetQuantity3DArr(&q, d, nk, nj, ni);
#endif SYM3D
      
      z->zoneData[varIndex] = q;
      
#endif ZONEDATA_ARE_MULTIDARRAYS
      
    } else {err = dimErr;}
    
  } else {err = memErr;}
  
  return(err);
  
}


int   zls_GetZoneDataDirect(QPtr *d, ZonePtr z, Integer varIndex){ // copy zone data reference, BCs n all if present
  
  int err = noErr;
  
  if (z!=NULL){
    if ((varIndex >= 0)&&(varIndex < kNAllVars)){
#ifdef ZONEDATA_ARE_PLAINBLOCKS
      *d = z->zoneData[varIndex];
#endif ZONEDATA_ARE_PLAINBLOCKS
#ifdef ZONEDATA_ARE_MULTIDARRAYS
      
#ifdef SYM3D
      *d = z->zoneData[varIndex][0][0];
#endif SYM3D
#ifdef SYM2D
      *d = z->zoneData[varIndex][0];
#endif SYM2D
#ifdef SYM1D
      *d =  z->zoneData[varIndex];
#endif SYM1D
      
#endif ZONEDATA_ARE_MULTIDARRAYS
    } else {err = dimErr;}
  } else {err = memErr;}
  
  return(err);
  
}

int   zls_GetZoneDataCopy( QPtr * d, ZonePtr z, Integer varIndex ){ 
  // copy only the active area of zoneData into a new QPtr
  //
  // If data is from the outside without BCs, assume it matches 
  // the zone 0-n dimensions, then allocate a proper data 
  // with new zoneData and parse d, offsetting by ns.
  // if the zone doesn't use BCs then it is a simple copy, and
  // d is assumed to not include BCs either.
  //
  // Unfortunately there is no simple way to query the size of a malloc's block in plain C.
  // In the original macOS I could use GetPtrSize(d), but that is part of that specific OS.
  //
  //
  int err = noErr;
  QPtr data = NULL;
  
  Integer nD;
  
#ifdef SYM3D          
  Integer i, j, k;
  Integer ni, nj, nk;
  Q3DArr q3z, q3d;
#endif SYM3D
  
#ifdef SYM2D          
  Integer i, j;
  Integer ni, nj;
  Q2DArr q2z, q2d;
#endif SYM2D
  
#ifdef SYM1D          
  Integer i;
  Integer ni;
  QPtr   q1z;
#endif SYM1D          
  
#ifdef SYM3D                  
  Integer is, js, ks;
  Integer nti, ntj, ntk;
#endif SYM3D
  
#ifdef SYM2D                  
  Integer is,   js;
  Integer nti, ntj;
#endif SYM2D
  
#ifdef SYM1D                  
  Integer is;
  Integer nti;
#endif SYM1D
  
  if (z!=NULL){
    
    nD = z->zoneGrid.nD;
    
#ifdef FILES_NO_BCS    
#ifdef SYM1D                
    is = z->zoneGrid.ns[kXDim];
    ni = z->zoneGrid.n[kXDim];
    nti = z->zoneGrid.nt[kXDim];
#endif SYM1D
    
#ifdef SYM2D                
    is = z->zoneGrid.ns[kXDim];
    js = z->zoneGrid.ns[kYDim];
    
    ni = z->zoneGrid.n[kXDim];
    nj = z->zoneGrid.n[kYDim];
    
    nti = z->zoneGrid.nt[kXDim];
    ntj = z->zoneGrid.nt[kYDim];
#endif SYM2D
    
#ifdef SYM3D                
    is = z->zoneGrid.ns[kXDim];
    js = z->zoneGrid.ns[kYDim];
    ks = z->zoneGrid.ns[kZDim];
    
    ni = z->zoneGrid.n[kXDim];
    nj = z->zoneGrid.n[kYDim];
    nk = z->zoneGrid.n[kZDim];
    
    nti = z->zoneGrid.nt[kXDim];
    ntj = z->zoneGrid.nt[kYDim];
    ntk = z->zoneGrid.nt[kZDim];
#endif SYM3D
#endif FILES_NO_BCS
    
#ifdef FILES_HAVE_BCS    
#ifdef SYM1D                
    is = 0;
    ni = z->zoneGrid.nt[kXDim];
    nti = z->zoneGrid.nt[kXDim];
#endif SYM1D
    
#ifdef SYM2D                
    is = 0;
    js = 0;
    
    ni = z->zoneGrid.nt[kXDim];
    nj = z->zoneGrid.nt[kYDim];
    
    nti = z->zoneGrid.nt[kXDim];
    ntj = z->zoneGrid.nt[kYDim];
#endif SYM2D
    
#ifdef SYM3D                
    is = 0;
    js = 0;
    ks = 0;
    
    ni = z->zoneGrid.nt[kXDim];
    nj = z->zoneGrid.nt[kYDim];
    nk = z->zoneGrid.nt[kZDim];
    
    nti = z->zoneGrid.nt[kXDim];
    ntj = z->zoneGrid.nt[kYDim];
    ntk = z->zoneGrid.nt[kZDim];
#endif SYM3D
#endif FILES_HAVE_BCS
    
#ifdef SYM1D          
    err = zls_NewQuantity1D(&data, ni);
    
    if (data !=NULL){
      
      if ((varIndex >= 0)&&(varIndex < kNAllVars)){
        
        q1z = z->zoneData[varIndex];
        
        for (i = 0; i<ni; i++){
          data[i] = q1z[i+is];
        } //i
        
      } else {
        err = dimErr;
      } //((varIndex >= 0)&&(varIndex < kNAllVars))
      
    } else {err = memErr;}
#endif SYM1D
    
#ifdef SYM2D          
    
    err = zls_NewQuantity2D(&data, ni, nj);
    
    if (data !=NULL){
      
      zls_SetQuantity2DArr(&q2d, data,  nj, ni);
      
      if ((varIndex >= 0)&&(varIndex < kNAllVars)){
        
#ifdef ZONEDATA_ARE_PLAINBLOCKS
        zls_SetQuantity2DArr(&q2z, z->zoneData[varIndex],  ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
#ifdef ZONEDATA_ARE_MULTIDARRAYS
        q2z = z->zoneData[varIndex];
#endif ZONEDATA_ARE_MULTIDARRAYS
        
        for (j = 0; j<nj; j++){
          for (i = 0; i<ni; i++){
            q2d[j][i] = q2z[j+js][i+is];
          } //i
        } //j
#ifdef ZONEDATA_ARE_PLAINBLOCKS
        zls_ReleaseQuantity2DArr(q2z);
#endif ZONEDATA_ARE_PLAINBLOCKS
        
      } else {
        
        err = dimErr;
        
      } //((varIndex >= 0)&&(varIndex < kNAllVars))
      
      zls_ReleaseQuantity2DArr(q2d);
      
    } else {err = memErr;}
    
#endif SYM2D
    
    
#ifdef SYM3D          
    err = zls_NewQuantity3D(&data, ni, nj, nk);
    
    if (data !=NULL){
      
      zls_SetQuantity3DArr(&q3d, data, nk, nj, ni);
      
      if ((varIndex >= 0)&&(varIndex < kNAllVars)){
        
#ifdef ZONEDATA_ARE_PLAINBLOCKS
        zls_SetQuantity3DArr(&q3z, z->zoneData[varIndex], ntk, ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
#ifdef ZONEDATA_ARE_MULTIDARRAYS
        q3z = z->zoneData[varIndex];
#endif ZONEDATA_ARE_MULTIDARRAYS
        
        for (k = 0; k<nk; k++){
          for (j = 0; j<nj; j++){
            for (i = 0; i<ni; i++){
              q3d[k][j][i] = q3z[k+ks][j+js][i+is];
            } //i
          } //j
        } //k
        
#ifdef ZONEDATA_ARE_PLAINBLOCKS
        zls_ReleaseQuantity3DArr(q3z);
#endif ZONEDATA_ARE_PLAINBLOCKS
        
      } else {
        err = dimErr;
      } //((varIndex >= 0)&&(varIndex < kNAllVars))
      
      zls_ReleaseQuantity3DArr(q3d);
      
    } else {err = memErr;}
    
#endif SYM3D
    
    
  }else {err = memErr;} //((z!=NULL)&&(d!=NULL))
  
  *d = data;
  
  return(err);
  
}


int   zls_GetZoneData1DRow( QPtr * d, ZonePtr z, Integer varIndex, Integer slice, Integer sliceIndex, Integer sliceIndex2){
  //
  // copy  a slice perp to axis at index of the active volume of zoneData a into a new Q2DPtr
  // meant for 2D slices of 3D zones.  For 2D zones it just returns the data plane an ignores axis and 
  // index.  Err for 1D zones
  // May also sum along the index axis rather than take one plane.
  //
  int err = noErr;
  
#ifdef SYM1D    
#pragma unused( slice, sliceIndex)
  // in 1D we can just allocate and use the existing zls_GetZoneDataCopy routine
  if (z!=NULL){
    err =  zls_GetZoneDataCopy( d, z, varIndex ); // simply call the existing copy routine
  } else { err = memErr;}
  
  return(err);
#endif SYM1D
  
#ifdef SYM2D    
  Integer nD;
  
  Integer     i, j;
  Integer   ni, nj;
  Integer   is, js;
  Integer nti, ntj;
  
  Q2DArr       q2z;
  
  QPtr data = NULL;
  
  if (z!=NULL){
    
    if ((varIndex >= 0)&&(varIndex < kNAllVars)){
      
      nD = z->zoneGrid.nD;
      
      
#ifdef FILES_NO_BCS    
      
      is = z->zoneGrid.ns[kXDim];
      js = z->zoneGrid.ns[kYDim];
      
      nti = z->zoneGrid.nt[kXDim];
      ntj = z->zoneGrid.nt[kYDim];
      
      ni = z->zoneGrid.n[kXDim];
      nj = z->zoneGrid.n[kYDim];
      
      switch (slice) {
        default:
        case kSliceX:
        case kSliceXSum:
          err = zls_NewQuantity1D(&data, ni);
          if ( sliceIndex < 0 ) sliceIndex = nj/2;
          break;
          case kSliceY:
          case kSliceYSum:
          err = zls_NewQuantity1D(&data, nj);
          if ( sliceIndex < 0 ) sliceIndex = ni/2;
          break;
      }
      
#endif FILES_NO_BCS
      
#ifdef FILES_HAVE_BCS          
      
      is = 0;
      js = 0;
      
      nti = z->zoneGrid.nt[kXDim];
      ntj = z->zoneGrid.nt[kYDim];
      
      ni = z->zoneGrid.nt[kXDim];
      nj = z->zoneGrid.nt[kYDim];
      
      switch (slice) {
        default:
        case kSliceX:
        case kSliceXSum:
          err = zls_NewQuantity1D(&data, ni);
          if ( sliceIndex < 0 ) sliceIndex = nj/2;
          break;
          case kSliceY:
          case kSliceYSum:
          err = zls_NewQuantity1D(&data, nj);
          if ( sliceIndex < 0 ) sliceIndex = ni/2;
          break;
      }
      
#endif FILES_HAVE_BCS
      
      if (data !=NULL){
        
#ifdef ZONEDATA_ARE_PLAINBLOCKS
        zls_SetQuantity2DArr(&q2z, z->zoneData[varIndex], ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
#ifdef ZONEDATA_ARE_MULTIDARRAYS
        q2z = z->zoneData[varIndex];
#endif ZONEDATA_ARE_MULTIDARRAYS
        
        switch (slice) {
          default:
          case kSliceX:
            
            j = sliceIndex;
            for (i = 0; i<ni; i++){
              data[i] = q2z[j+js][i+is];
            }; //i
            
            break;
            
            case kSliceXSum:
            
            j = 0;
            for (i = 0; i<ni; i++){
              data[i] = q2z[j+js][i+is];
            } //i
            for (i = 0; i<ni; i++){
              for (j = 0; j<nj; j++){
                data[i] += q2z[j+js][i+is];
              } //j
            }; //i
            break;
            
            case kSliceY:
            i = sliceIndex;
            for (j = 0; j<nj; j++){
              data[j] = q2z[j+js][i+is];
            } //i
            break;
            
            case kSliceYSum:
            i = 0;
            for (j = 0; j<nj; j++){
              data[j] = q2z[j+js][i+is];
            } //i
            for (j =  0; j<nj; j++){
              for (i = 1; i<ni; i++){
                data[j] += q2z[j+js][i+is];
              } //i
            }; //i
            
            break;
        } // switch
#ifdef ZONEDATA_ARE_PLAINBLOCKS
        zls_ReleaseQuantity2DArr(q2z);
#endif ZONEDATA_ARE_PLAINBLOCKS
        
      } else {err = memErr;} // data != NULL
      
    } else {err = dimErr;} //((varIndex >= 0)&&(varIndex < kNAllVars))
    
  }else {err = memErr;} //(z!=NULL)
  
  *d = data;
  
  return(err);
  
#endif SYM2D    
  
#ifdef SYM3D    
  Integer nD;
  
  Integer i, j, k;
  Integer ni, nj, nk;
  Integer is, js, ks;
  Integer nti, ntj, ntk;
  Q3DArr q3z;
  
  QPtr data = NULL;
  //Q2DArr q2d;
  
  if (z!=NULL){
    
    if ((varIndex >= 0)&&(varIndex < kNAllVars)){
      
      nD = z->zoneGrid.nD;
      
      
      //if sliceIndex is -ve it is set to the midplane according to 
      // the slice plane below.
      
#ifdef FILES_NO_BCS    
      
      is = z->zoneGrid.ns[kXDim];
      js = z->zoneGrid.ns[kYDim];
      ks = z->zoneGrid.ns[kZDim];
      
      nti = z->zoneGrid.nt[kXDim];
      ntj = z->zoneGrid.nt[kYDim];
      ntk = z->zoneGrid.nt[kZDim];
      
      ni = z->zoneGrid.n[kXDim];
      nj = z->zoneGrid.n[kYDim];
      nk = z->zoneGrid.n[kZDim];
      
      
#endif FILES_NO_BCS
      
#ifdef FILES_HAVE_BCS          
      
      is = 0;
      js = 0;
      ks = 0;
      nti = z->zoneGrid.nt[kXDim];
      ntj = z->zoneGrid.nt[kYDim];
      ntk = z->zoneGrid.nt[kZDim];
      ni = z->zoneGrid.nt[kXDim];
      nj = z->zoneGrid.nt[kYDim];
      nk = z->zoneGrid.nt[kZDim];
#endif FILES_HAVE_BCS
      
      switch (slice) {
        default:
        case kSliceX:
        case kSliceXSum:
          err = zls_NewQuantity1D(&data, ni);
          if ( sliceIndex < 0 ) sliceIndex = nj/2;
          if ( sliceIndex2 < 0 ) sliceIndex2 = nk/2;
          break;
          case kSliceY:
          case kSliceYSum:
          err = zls_NewQuantity1D(&data, nj);
          if ( sliceIndex < 0 ) sliceIndex = ni/2;
          if ( sliceIndex2 < 0 ) sliceIndex2 = nk/2;
          break;
          case kSliceZ:
          case kSliceZSum:
          err = zls_NewQuantity1D(&data, nk);
          if ( sliceIndex < 0 ) sliceIndex = ni/2;
          if ( sliceIndex2 < 0 ) sliceIndex2 = nj/2;
          break;
      }
      
      if (data !=NULL){
        
        
#ifdef ZONEDATA_ARE_PLAINBLOCKS
        zls_SetQuantity3DArr(&q3z, z->zoneData[varIndex], ntk, ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
#ifdef ZONEDATA_ARE_MULTIDARRAYS
        q3z = z->zoneData[varIndex];
#endif ZONEDATA_ARE_MULTIDARRAYS
        
        switch (slice) {
          default:
          case kSliceX:
            j = sliceIndex;
            k = sliceIndex2;
            for (i = 0; i<ni; i++){
              data[i] = q3z[k+ks][j+js][i+is];
            }; //i
            break;
            
            case kSliceXSum:
            j = 0;
            k = 0;
            for (i = 0; i<ni; i++){
              data[i] = q3z[k+ks][j+js][i+is];
            }; //i
            for (i = 0; i<ni; i++){
              for (j = 1; j<nj; j++){
                for (k = 1; k<nk; k++){
                  data[i] += q3z[k+ks][j+js][i+is];
                } //k
              } //j
            }; //i
            break;
            
            case kSliceY:
            i = sliceIndex;
            k = sliceIndex2;
            for (j = 0; j<nj; j++){
              data[j] = q3z[k+ks][j+js][i+is];
            }; //j
            break;
            
            case kSliceYSum:
            i = 0;
            k = 0;
            for (j = 0; j<nj; j++){
              data[j] = q3z[k+ks][j+js][i+is];
            }; //j
            
            for (j = 0; j<nj; j++){
              for (i = 1; i<ni; i++){
                for (k = 1; k<nk; k++){
                  data[j] += q3z[k+ks][j+js][i+is];
                } //k
              } //i
            }; //j
            break;
            
            
            case kSliceZ:
            
            i = sliceIndex;
            j = sliceIndex2;
            for (k = 0; k<nk; k++){
              data[k] = q3z[k+ks][j+js][i+is];
            }; //k
            break;
            case kSliceZSum:
            i = 0;
            j = 0;
            for (k = 0; k<nk; k++){
              data[k] = q3z[k+ks][j+js][i+is];
            }; //k
            for (k = 0; k<nk; k++){
              for (j = 1; j<nj; j++){
                for (i = 1; i<ni; i++){
                  data[k] += q3z[k+ks][j+js][i+is];
                } //j
              } //j
            }; //k
            break;
            
        } // switch
#ifdef ZONEDATA_ARE_PLAINBLOCKS
        zls_ReleaseQuantity3DArr(q3z);
#endif ZONEDATA_ARE_PLAINBLOCKS
        
//        zls_ReleaseQuantity2DArr(q2d);
        
        
      } else {err = memErr;} // data != NULL
      
    } else {err = dimErr;} //((varIndex >= 0)&&(varIndex < kNAllVars))
    
  }else {err = memErr;} //(z!=NULL)
  
  *d = data;
  
  return(err);
  
#endif SYM3D    
  
}

int   zls_GetZoneData2DSlice( QPtr * d, ZonePtr z, Integer varIndex, Integer slice, Integer sliceIndex){ 
  //
  // copy  a slice perp to axis at index of the active volume of zoneData a into a new Q2DPtr
  // meant for 2D slices of 3D zones.  For 2D zones it just returns the data plane an ignores axis and 
  // index.  Err for 1D zones
  // May also sum along the index axis rather than take one plane.
  //
  int err = noErr;
  
#if (NDIMS < 2)
#pragma unused( d, z, varIndex, slice, sliceIndex)
  zlsio_ErrPrintf( " WARNING: zls_GetZoneData2DSlice 2D slice not possible in < 2D zone: ndims =  %d \n", NDIMS);
  err = dimErr;
  return(err);
#endif 
  
#ifdef SYM2D    
#pragma unused( slice, sliceIndex)
  // in 2D we can just allocate and use the existing zls_GetZoneDataCopy routine
  if (z!=NULL){
    err =  zls_GetZoneDataCopy( d, z, varIndex ); // simply call the existing copy routine
  } else { err = memErr;}
  
  return(err);
#endif SYM2D
  
#ifdef SYM3D    
  Integer nD;
  
  Integer i, j, k;
  Integer ni, nj, nk;
  Integer is, js, ks;
  Integer nti, ntj, ntk;
  Q3DArr q3z;
  
  QPtr data = NULL;
  Q2DArr q2d;
  
  if (z!=NULL){
    
    if ((varIndex >= 0)&&(varIndex < kNAllVars)){
      
      nD = z->zoneGrid.nD;
      
      
      //if sliceIndex is -ve it is set to the midplane according to 
      // the slice plane below.
      
#ifdef FILES_NO_BCS    
      
      is = z->zoneGrid.ns[kXDim];
      js = z->zoneGrid.ns[kYDim];
      ks = z->zoneGrid.ns[kZDim];
      
      nti = z->zoneGrid.nt[kXDim];
      ntj = z->zoneGrid.nt[kYDim];
      ntk = z->zoneGrid.nt[kZDim];
      
      ni = z->zoneGrid.n[kXDim];
      nj = z->zoneGrid.n[kYDim];
      nk = z->zoneGrid.n[kZDim];
      
      switch (slice) {
        case kSliceYZ:
        case kSliceYZSum:
          err = zls_NewQuantity2D(&data, nj, nk);
          if ( sliceIndex < 0 ) sliceIndex = ni/2;
          break;
          case kSliceYX:
          case kSliceYXSum:
          err = zls_NewQuantity2D(&data, nj, ni);
          if ( sliceIndex < 0 ) sliceIndex = nk/2;
          break;
          case kSliceZX:
          case kSliceZXSum:
          err = zls_NewQuantity2D(&data, nk, ni);
          if ( sliceIndex < 0 ) sliceIndex = nj/2;
          break;
          case kSliceZY:
          case kSliceZYSum:
          err = zls_NewQuantity2D(&data, nk, nj);
          if ( sliceIndex < 0 ) sliceIndex = ni/2;
          break;
          case kSliceXY:
          case kSliceXYSum:
          err = zls_NewQuantity2D(&data, ni, nj);
          if ( sliceIndex < 0 ) sliceIndex = nk/2;
          break;
          case kSliceXZ:
          case kSliceXZSum:
          err = zls_NewQuantity2D(&data, ni, nk);
          if ( sliceIndex < 0 ) sliceIndex = nj/2;
          break;
          default:
          zlsio_ErrPrintf( " WARNING: zls_GetZoneData2DSlice Bad Slice Type in %d \n", slice);
          err = dimErr;
          return(err);
      }
      
#endif FILES_NO_BCS
      
#ifdef FILES_HAVE_BCS          
      
      is = 0;
      js = 0;
      ks = 0;
      nti = z->zoneGrid.nt[kXDim];
      ntj = z->zoneGrid.nt[kYDim];
      ntk = z->zoneGrid.nt[kZDim];
      ni = z->zoneGrid.nt[kXDim];
      nj = z->zoneGrid.nt[kYDim];
      nk = z->zoneGrid.nt[kZDim];
      switch (slice) {
        case kSliceYZ:
        case kSliceYZSum:
          err = zls_NewQuantity2D(&data, nj, nk);
          if ( sliceIndex < 0 ) sliceIndex = ni/2;
          break;
          case kSliceYX:
          case kSliceYXSum:
          err = zls_NewQuantity2D(&data, nj, ni);
          if ( sliceIndex < 0 ) sliceIndex = nk/2;
          break;
          case kSliceZX:
          case kSliceZXSum:
          err = zls_NewQuantity2D(&data, nk, ni);
          if ( sliceIndex < 0 ) sliceIndex = nj/2;
          break;
          case kSliceZY:
          case kSliceZYSum:
          err = zls_NewQuantity2D(&data, nk, nj);
          if ( sliceIndex < 0 ) sliceIndex = ni/2;
          break;
          case kSliceXY:
          case kSliceXYSum:
          err = zls_NewQuantity2D(&data, ni, nj);
          if ( sliceIndex < 0 ) sliceIndex = nk/2;
          break;
          case kSliceXZ:
          case kSliceXZSum:
          err = zls_NewQuantity2D(&data, ni, nk);
          if ( sliceIndex < 0 ) sliceIndex = nj/2;
          break;
          default:
          zlsio_ErrPrintf( " WARNING: zls_GetZoneData2DSlice Bad Slice Type in %d \n", slice);
          err = dimErr;
          return(err);
      }
      
#endif FILES_HAVE_BCS
      
      if (data !=NULL){
        
        
#ifdef ZONEDATA_ARE_PLAINBLOCKS
        zls_SetQuantity3DArr(&q3z, z->zoneData[varIndex], ntk, ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
#ifdef ZONEDATA_ARE_MULTIDARRAYS
        q3z = z->zoneData[varIndex];
#endif ZONEDATA_ARE_MULTIDARRAYS
        
        switch (slice) {
          case kSliceYZ:
            zls_SetQuantity2DArr(&q2d, data, nk, nj);
            i = sliceIndex;
            for (k = 0; k<nk; k++){
              for (j = 0; j<nj; j++){
                q2d[k][j] = q3z[k+ks][j+js][i+is];
              } //j
            }; //k
            break;
            case kSliceYZSum:
            zls_SetQuantity2DArr(&q2d, data, nk, nj);
            i = 0;
            for (k = 0; k<nk; k++){
              for (j = 0; j<nj; j++){
                q2d[k][j] = q3z[k+ks][j+js][i+is];
              } //j
            }; //k
            for (k = 0; k<nk; k++){
              for (j = 0; j<nj; j++){
                for (i = 1; i<ni; i++){
                  q2d[k][j] += q3z[k+ks][j+js][i+is];
                } //j
              } //j
            }; //k
            break;
            case kSliceZY:
            zls_SetQuantity2DArr(&q2d, data, nj, nk);
            i = sliceIndex;
            for (k = 0; k<nk; k++){
              for (j = 0; j<nj; j++){
                q2d[j][k] = q3z[k+ks][j+js][i+is];
              } //j
            }; //k
            break;
            case kSliceZYSum:
            zls_SetQuantity2DArr(&q2d, data, nj, nk);
            i = 0;
            for (k = 0; k<nk; k++){
              for (j = 0; j<nj; j++){
                q2d[j][k] = q3z[k+ks][j+js][i+is];
              } //j
            }; //k
            for (k = 0; k<nk; k++){
              for (j = 0; j<nj; j++){
                for (i = 1; i<ni; i++){
                  q2d[j][k] += q3z[k+ks][j+js][i+is];
                } //j
              } //j
            }; //k
            break;
            case kSliceZX:
            zls_SetQuantity2DArr(&q2d, data, ni, nk);
            j = sliceIndex;
            for (i = 0; i<ni; i++){
              for (k = 0; k<nk; k++){
                q2d[i][k] = q3z[k+ks][j+js][i+is];
              } //k
            }; //i
            break;
            case kSliceZXSum:
            zls_SetQuantity2DArr(&q2d, data, ni, nk);
            j = 0;
            for (i = 0; i<ni; i++){
              for (k = 0; k<nk; k++){
                q2d[i][k] = q3z[k+ks][j+js][i+is];
              } //k
            }; //i
            for (i = 0; i<ni; i++){
              for (k = 0; k<nk; k++){
                for (j = 1; j<nj; j++){
                  q2d[i][k] += q3z[k+ks][j+js][i+is];
                } //j
              } //k
            }; //i
            break;
            case kSliceXZ:
            zls_SetQuantity2DArr(&q2d, data, nk, ni);
            j = sliceIndex;
            for (i = 0; i<ni; i++){
              for (k = 0; k<nk; k++){
                q2d[k][i] = q3z[k+ks][j+js][i+is];
              } //k
            }; //i
            break;
            case kSliceXZSum:
            zls_SetQuantity2DArr(&q2d, data, nk, ni);
            j = 0;
            for (i = 0; i<ni; i++){
              for (k = 0; k<nk; k++){
                q2d[k][i] = q3z[k+ks][j+js][i+is];
              } //k
            }; //i
            for (i = 0; i<ni; i++){
              for (k = 0; k<nk; k++){
                for (j = 1; j<nj; j++){
                  q2d[k][i] += q3z[k+ks][j+js][i+is];
                } //j
              } //k
            }; //i
            break;
            case kSliceXY:
            zls_SetQuantity2DArr(&q2d, data, nj, ni);
            k = sliceIndex;
            for (j = 0; j<nj; j++){
              for (i = 0; i<ni; i++){
                q2d[j][i] = q3z[k+ks][j+js][i+is];
              } //i
            }; //j
            break;
            case kSliceXYSum:
            zls_SetQuantity2DArr(&q2d, data, nj, ni);
            k = 0;
            for (j = 0; j<nj; j++){
              for (i = 0; i<ni; i++){
                q2d[j][i] = q3z[k+ks][j+js][i+is];
              } //i
            }; //j
            for (j = 0; j<nj; j++){
              for (i = 0; i<ni; i++){
                for (k = 1; k<nk; k++){
                  q2d[j][i] += q3z[k+ks][j+js][i+is];
                } //k
              } //i
            }; //j
            break;
            case kSliceYX:
            zls_SetQuantity2DArr(&q2d, data, ni, nj);
            k = sliceIndex;
            for (j = 0; j<nj; j++){
              for (i = 0; i<ni; i++){
                q2d[i][j] = q3z[k+ks][j+js][i+is];
              } //i
            }; //j
            break;
            case kSliceYXSum:
            zls_SetQuantity2DArr(&q2d, data, ni, nj);
            k = 0;
            for (j = 0; j<nj; j++){
              for (i = 0; i<ni; i++){
                q2d[i][j] = q3z[k+ks][j+js][i+is];
              } //i
            }; //j
            for (j = 0; j<nj; j++){
              for (i = 0; i<ni; i++){
                for (k = 1; k<nk; k++){
                  q2d[i][j] += q3z[k+ks][j+js][i+is];
                } //k
              } //i
            }; //j
            break;
            default:
            zlsio_ErrPrintf( " WARNING: zls_GetZoneData2DSlice Bad Slice Type in %d \n", slice);
            err = dimErr;
            return(err);
        } // switch
#ifdef ZONEDATA_ARE_PLAINBLOCKS
        zls_ReleaseQuantity3DArr(q3z);
#endif ZONEDATA_ARE_PLAINBLOCKS
        
        zls_ReleaseQuantity2DArr(q2d);
        
        
      } else {err = memErr;} // data != NULL
      
    } else {err = dimErr;} //((varIndex >= 0)&&(varIndex < kNAllVars))
    
  }else {err = memErr;} //(z!=NULL)
  
  *d = data;
  
  return(err);
  
#endif SYM3D    
  
}

//
//  Getters for extracting a row, slice or volume, ie 1D rows for split sweeps
//  creates and returns g and v, caller's resposiblity to dispose afterwards
//  

int     zls_GetZoneVar1D  (V1DArr *v, ZonePtr z, Integer varIndex){// 1D Zone
#ifndef SYM1D
#pragma unused( v, z, varIndex)
  int err = dimErr;
  return(err);
#else
  int err = noErr;
  Integer ni;
  
  Integer nD, nG;
  
  Q1DArr      q;
  
  V1DArr  tempV;
  
  nD = z->zoneGrid.nD;
  
  if (nD == 1){
    
    nG = z->zoneGrid.nG;
    ni = z->zoneGrid.n[0];
    
    err = zls_NewVar1DArr(&tempV, nG, ni);
    q = z->zoneData[varIndex];
    
    err = zls_SetVar1DArrQ1DArr(tempV, q);
    
    tempV.x = z->zoneGrid.x; // copy axis ref
    
  } else { // !nD1
    err = dimErr;
  }
  
  *v = tempV;
  return(err);
#endif SYM1D
}


int     zls_GetZoneVar2D  (V2DArr *v, ZonePtr z, Integer varIndex){// 2D Zone
  
#ifndef SYM2D
  int err = dimErr;
  return(err);
#else
  
  int err = noErr;
  Integer ni, nj;
  Integer nti, ntj;
  
  Integer nD, nG;
  
  Q2DArr      q;
  
  V2DArr  tempV;
  
  nD = z->zoneGrid.nD;
  
  if (nD == 2){
    
    nG = z->zoneGrid.nG;
    
    ni = z->zoneGrid.n[kXDim];
    nj = z->zoneGrid.n[kYDim];
    nti = z->zoneGrid.nt[kXDim];
    ntj = z->zoneGrid.nt[kYDim];
    
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    zls_SetQuantity2DArr(&q, z->zoneData[varIndex], ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
    
#ifdef ZONEDATA_ARE_MULTIDARRAYS
    q = z->zoneData[varIndex];
#endif ZONEDATA_ARE_MULTIDARRAYS
    
    err = zls_NewVar2DArr(&tempV, nG, nj, ni);
    
    err = zls_SetVar2DArrQ2DArr(tempV, q);
    
    tempV.x = z->zoneGrid.x; // copy refernce, dispose var should not dispose, they remain
    tempV.y = z->zoneGrid.y; // the property of the grid.
    
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    zls_ReleaseQuantity2DArr(q);
#endif ZONEDATA_ARE_PLAINBLOCKS
    
  } else { // !nD2
    err = dimErr;
  }
  
  *v = tempV;
  return(err);
  
#endif SYM2D
}


int     zls_GetZoneVar3D  (V3DArr *v, ZonePtr z, Integer varIndex){// 3D Zone
#ifndef SYM3D
#pragma unused( v, z, varIndex)
  int err = dimErr;
  return(err);
#else
  int err = noErr;
  Integer ni, nj, nk;
  Integer nti, ntj, ntk;
  
  Integer nD, nG;
  
  Q3DArr      q;
  
  V3DArr  tempV;
  
  nD = z->zoneGrid.nD;
  
  if (nD == 3){
    
    nG = z->zoneGrid.nG;
    
    ni = z->zoneGrid.n[kXDim];
    nj = z->zoneGrid.n[kYDim];
    nk = z->zoneGrid.n[kZDim];
    
    nti = z->zoneGrid.nt[kXDim];
    ntj = z->zoneGrid.nt[kYDim];
    ntk = z->zoneGrid.nt[kZDim];
    
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    zls_SetQuantity3DArr(&q, z->zoneData[varIndex], ntk, ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
    
#ifdef ZONEDATA_ARE_MULTIDARRAYS
    q = z->zoneData[varIndex];
#endif ZONEDATA_ARE_MULTIDARRAYS
    
    err = zls_NewVar3DArr(&tempV, nG, nk, nj, ni);
    err = zls_SetVar3DArrQ3DArr(tempV, q);
    
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    zls_ReleaseQuantity3DArr(q);
#endif ZONEDATA_ARE_PLAINBLOCKS
    
    tempV.x = z->zoneGrid.x; // copy refernce, dispose var should not dispose, they remain
    tempV.y = z->zoneGrid.y; // the property of the grid.
    tempV.z = z->zoneGrid.z; //
    
  } else { // !nD3
    err = dimErr;
  }
  
  *v = tempV;
  return(err);
#endif SYM3D
}



int     zls_GetZone2DVar1DX (V1DArr *v, ZonePtr z, Integer varIndex, Integer j){
#ifndef SYM2D
  int err = dimErr;
  return(err);
#else
  //
  //  get 1D x var from 2D zone, at j, j in active area
  //
  
  int err = noErr;
  
  Integer ni, nj;
  Integer nti, ntj;
  
  Integer nD, nG;
  
  Q2DArr  q =NULL;
  V1DArr    tempV;
  
  nD  = z->zoneGrid.nD;
  nG  = z->zoneGrid.nG;
  
  ni  = z->zoneGrid.n[kXDim];
  nj  = z->zoneGrid.n[kYDim];
  
  nti = z->zoneGrid.nt[kXDim];
  ntj = z->zoneGrid.nt[kYDim];
  
  if (nD == 2){
    
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    zls_SetQuantity2DArr(&q, z->zoneData[varIndex],ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
    
#ifdef ZONEDATA_ARE_MULTIDARRAYS
    q = z->zoneData[varIndex];
#endif ZONEDATA_ARE_MULTIDARRAYS
    
    err     = zls_NewVar1DArr(&tempV, nG, ni);
    zls_SetVar1DArrQ2DArrX(tempV, q, j);
    tempV.x = z->zoneGrid.x;
    
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    zls_ReleaseQuantity2DArr(q);
#endif ZONEDATA_ARE_PLAINBLOCKS
    
  } else { // !nD2
  err = dimErr; }
  
  *v = tempV;
  
  return(err);
#endif SYM2D
} //zls_GetZone2DVar1DX

int     zls_GetZone2DVar1DY (V1DArr *v, ZonePtr z, Integer varIndex, Integer i){ // in in active area/range coords
#ifndef SYM2D
  int err = dimErr;
  return(err);
#else
  //
  //  get 1D x var from 2D zone, at j, j in active area
  //
  
  int err = noErr;
  
  Integer ni, nj;
  Integer nti, ntj;
  
  Integer nD, nG;
  
  Q2DArr  q =NULL;
  V1DArr    tempV;
  
  nD  = z->zoneGrid.nD;
  nG  = z->zoneGrid.nG;
  
  ni  = z->zoneGrid.n [kXDim];
  nj  = z->zoneGrid.n[kYDim];
  
  nti = z->zoneGrid.nt[kXDim];
  ntj = z->zoneGrid.nt[kYDim];
  
  if (nD == 2){
    
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    zls_SetQuantity2DArr(&q, z->zoneData[varIndex],ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
    
#ifdef ZONEDATA_ARE_MULTIDARRAYS
    q = z->zoneData[varIndex];
#endif ZONEDATA_ARE_MULTIDARRAYS
    
    err     = zls_NewVar1DArr(&tempV, nG, nj);
    zls_SetVar1DArrQ2DArrY(tempV, q, i);
    tempV.x = z->zoneGrid.y;
    
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    zls_ReleaseQuantity2DArr(q);
#endif ZONEDATA_ARE_PLAINBLOCKS
    
  } else { // !nD2
  err = dimErr; }
  
  *v = tempV;
  
  return(err);
#endif SYM2D
} //zls_GetZone2DVar1DY



int     zls_SetVar1DXZone2DQArr (V1DArr v, Q2DArr q, ZonePtr  z, Integer j){
#ifndef SYM2D
  int err = dimErr;
  return(err);
#else
  //
  //  get 1D x var from 2D zone, at j, j in active area
  //
  
  int err = noErr;
  Integer nD = z->zoneGrid.nD;
  
  if (nD == 2){
    
    zls_SetVar1DArrQ2DArrX(v, q, j);
    v.x = z->zoneGrid.x;
    
  } else { // !nD2
  err = dimErr; }
  
  return(err);
#endif SYM2D
} //zls_SetVar1DXZone2DQArr

int     zls_SetVar1DYZone2DQArr (V1DArr v, Q2DArr q, ZonePtr  z, Integer i){ 
  // like Get, but sets a pre allocated var1D, and Q2DArr rather than creating one each time.
  // useful for then a single var can be determined before a long loop on the same zone
  // avoiding setup and dismantling of the q arrays as well.
#ifndef SYM2D
  int err = dimErr;
  return(err);
#else
  //
  //  get 1D x var from 2D zone, at j, j in active area
  //
  
  int err = noErr;
  
  Integer nD  = z->zoneGrid.nD;
  
  if (nD == 2){
    
    zls_SetVar1DArrQ2DArrY(v, q, i);
    v.x = z->zoneGrid.y;
    
  } else { // !nD2
  err = dimErr; }
  
  return(err);
#endif SYM2D
} //SetVar1DYZone2D

int     zls_GetZone3DVar1DX (V1DArr *v, ZonePtr z, Integer varIndex, Integer k, Integer j){
#ifndef SYM3D
#pragma unused( v, z, varIndex, k, j)
  int err = dimErr;
  return(err);
#else
  //
  //  get 1D x var from 3D zone, at j, k in active area
  //
  
  int err = noErr;
  
  Integer ni, nj, nk;
  Integer nti, ntj, ntk;
  
  Integer nD, nG;
  
  Q3DArr  q =NULL;
  V1DArr    tempV;
  
  nD  = z->zoneGrid.nD;
  nG  = z->zoneGrid.nG;
  
  ni  = z->zoneGrid.n[kXDim];
  nj  = z->zoneGrid.n[kYDim];
  nk  = z->zoneGrid.n[kZDim];
  
  nti = z->zoneGrid.nt[kXDim];
  ntj = z->zoneGrid.nt[kYDim];
  ntk = z->zoneGrid.nt[kZDim];
  
  if (nD == 3){
    
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    zls_SetQuantity3DArr(&q, z->zoneData[varIndex], ntk, ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
    
#ifdef ZONEDATA_ARE_MULTIDARRAYS
    q = z->zoneData[varIndex];
#endif ZONEDATA_ARE_MULTIDARRAYS
    
    err     = zls_NewVar1DArr(&tempV, nG, ni);
    err     = zls_SetVar1DArrQ3DArrX(tempV, q, k, j);
    tempV.x = z->zoneGrid.x;
    
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    zls_ReleaseQuantity3DArr(q);
#endif ZONEDATA_ARE_PLAINBLOCKS
    
  } else { // !nD3
  err = dimErr; }
  
  *v = tempV;
  
  return(err);
#endif SYM3D
}

int     zls_GetZone3DVar1DY (V1DArr *v, ZonePtr z, Integer varIndex, Integer k, Integer i){
#ifndef SYM3D
#pragma unused( v, z, varIndex, k, i)
  int err = dimErr;
  return(err);
#else
  
  //
  //  get 1D y var from 3D zone, at j, k in active area
  //
  
  int err = noErr;
  
  Integer ni, nj, nk;
  Integer nti, ntj, ntk;
  
  Integer nD, nG;
  
  Q3DArr  q =NULL;
  V1DArr    tempV;
  
  nD  = z->zoneGrid.nD;
  nG  = z->zoneGrid.nG;
  
  ni  = z->zoneGrid.n[kXDim];
  nj  = z->zoneGrid.n[kYDim];
  nk  = z->zoneGrid.n[kZDim];
  
  nti = z->zoneGrid.nt[kXDim];
  ntj = z->zoneGrid.nt[kYDim];
  ntk = z->zoneGrid.nt[kZDim];
  
  if (nD == 3){
    
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    zls_SetQuantity3DArr(&q, z->zoneData[varIndex], ntk, ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
    
#ifdef ZONEDATA_ARE_MULTIDARRAYS
    q = z->zoneData[varIndex];
#endif ZONEDATA_ARE_MULTIDARRAYS
    
    err     = zls_NewVar1DArr(&tempV, nG, nj);
    err     = zls_SetVar1DArrQ3DArrY(tempV, q, k, i);
    tempV.x = z->zoneGrid.y;
    
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    zls_ReleaseQuantity3DArr(q);
#endif ZONEDATA_ARE_PLAINBLOCKS
    
  } else { // !nD3
  err = dimErr; }
  
  *v = tempV;
  
  return(err);
#endif SYM3D
}

int     zls_GetZone3DVar1DZ (V1DArr *v, ZonePtr z, Integer varIndex, Integer j, Integer i){
#ifndef SYM3D
#pragma unused( v, z, varIndex, j, i)
  return(dimErr);
#else
  //
  //  get 1D z var from 3D zone, at i, j in active area
  //
  
  int err = noErr;
  
  Integer ni, nj, nk;
  Integer nti, ntj, ntk;
  
  Integer nD, nG;
  
  Q3DArr  q =NULL;
  V1DArr    tempV;
  
  nD  = z->zoneGrid.nD;
  nG  = z->zoneGrid.nG;
  
  ni  = z->zoneGrid.n[kXDim];
  nj  = z->zoneGrid.n[kYDim];
  nk  = z->zoneGrid.n[kZDim];
  
  nti = z->zoneGrid.nt[kXDim];
  ntj = z->zoneGrid.nt[kYDim];
  ntk = z->zoneGrid.nt[kZDim];
  
  if (nD == 3){
    
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    zls_SetQuantity3DArr(&q, z->zoneData[varIndex], ntk, ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
    
#ifdef ZONEDATA_ARE_MULTIDARRAYS
    q = z->zoneData[varIndex];
#endif ZONEDATA_ARE_MULTIDARRAYS
    
    err     = zls_NewVar1DArr(&tempV, nG, nk);
    err     = zls_SetVar1DArrQ3DArrZ(tempV, q, j, i);
    tempV.x = z->zoneGrid.z;
    
#ifdef ZONEDATA_ARE_PLAINBLOCKS
    zls_ReleaseQuantity3DArr(q);
#endif ZONEDATA_ARE_PLAINBLOCKS
    
  } else { // !nD3
  err = dimErr; }
  
  *v = tempV;
  
  return(err);
#endif SYM3D
}

int     zls_GetZone3DVar2DXY(V2DArr *v, ZonePtr z, Integer varIndex, Integer k){
#pragma unused( v, z, varIndex, k)
  // x-y slice at k
  int err = nyiErr;
#if (ZLS_DEBUG>0) 
  zlsio_ErrPrintf( " Not Yet Implemented: zls_GetZone3DVar2DXY\n" );
#endif  
  return(err);
} 

int     zls_GetZone3DVar2DXZ(V2DArr *v, ZonePtr z, Integer varIndex, Integer j){
#pragma unused( v, z, varIndex, j)
  // x-z slice at j
  int err = nyiErr;
#if (ZLS_DEBUG>0) 
  zlsio_ErrPrintf( " Not Yet Implemented: zls_GetZone3DVar2DXZ\n" );
#endif  
  return(err);
}

int     zls_GetZone3DVar2DYZ(V2DArr *v, ZonePtr z, Integer varIndex, Integer i){
#pragma unused( v, z, varIndex, i)
  // y-z slice at i
  int err = nyiErr;
#if (ZLS_DEBUG>0) 
  zlsio_ErrPrintf( " Not Yet Implemented: zls_GetZone3DVar2DYZ\n" );
#endif  
  return(err);
}

