#ifndef ZLS_ZONES_H
#define ZLS_ZONES_H

#include "zls.h"

typedef struct zone { 
  //
  // Top level structure - Zone contains clock, grid and variables, plus
  // refs to super,owning Zone
  //
  //  Zone Rules: 
  //          1) All zones except top superzone are completely containted in exactly one superzone
  //          2) All zones have at least zoneBCells active cells
  //          3) A zone can have up to 6 neighbors perfectly abutting in the same level
  //            a) same extent along common boundary.
  //            b) up to one neighbour for each direction. 
  //            c) neighbours must have a reference to each other.
  //          4) Boundary values come from one source only - super, a single next zone, or computed
  //
  Integer       zoneLevel; // level this zone belongs to
  Integer          zoneID; // Unique ID, allocated on creation
  Integer      zoneRankID; // zones of same rank are swept by same process - ie  mpi rank under MPI, curently 0 out of MPI, set in grid.conf
  Integer        zoneType; // Flag for type of zone, see zls_consts.h
  Integer      zoneRefine; // spatial refinement factor from supergrid, 2..8
  Integer      zoneBCells; // number of cells in boundary zones, often same as zoneRefine so 1 supercell fills boundary
  double       zoneNCells; // total actve cells ( arrays*vars) in zone, using a double to allow for v large data
                           // yet still cell accurate for billions of cells.  *Not* used for exact output, but for
                           // overall performance calculations.
#ifdef ZLS_LOADSORTING
  CPUCoords      zoneCPUTimer; // cpu timer for sweep load balancing
#endif ZLS_LOADSORTING
  
  GridCoords     zoneGrid; // zone internal coordinates
  TimeCoords    zoneClock; // time and current timestep and cycle count
  
  Real            zoneCFL; // Courant-Friedrichs-Levey number for the zone - keep the same for each level. 
  Real            zoneNewDt; // dynamic/cooling/ionising timescale for zone, used to compute time step
  Real            zoneOldDt;
  Integer   zoneTimeScaleStatus; // whether timescale needs recomputing or not
#ifdef COOLING
  Real zoneMinCoolingTime; // keep a local zone minimum cooling timescale,
  // form the cooling during a sweep 
  // to be merged into a global time after any  multi-threading sweeps
  // saves recomputing expensive cooling rates when estimating a new timestep
  // ater sweeps the zone mins are combined into the level physics min cooling
  // time global. Needs to be zone by zone to not violate global access when
  // PTHREADS are being used.
#endif COOLING
  
  struct zone *    zoneSuper; // super zone this zone is embedded in
  Range       zoneSuperRange; // zone location in Super zone
  Range       zoneLevelRange; // zone location in Level Grid
  
  Integer    zoneSuperBCells; // number of cells in super zone equal to zone BCells, precomputed for speed.
  
  Integer       zoneBCInterp; // when any fill from super calls are made, use this interpolation code
  
  struct zone *  zoneNext [kNGeomDirs];  // array of zones next to this zone, up to 6 zones completely sharing boundary
  Integer      zoneBCType [kNGeomDirs];  // whether the BCs in a direction are copied or computed, etc
  Real       zoneBCValues [kNGeomDirs][kNAllVars]; // a value for each var in each direction for simple constant boundaries 
  
#ifdef MPI
  // if on a shared memory system , ie PTHREADS, then BCs are exchanged in place.  Under MPI
  // we need buffers to receive remote BC values to be placed in arrays and buffers to send BCs to other 
  // processes.  Basicly need two contiguous QPtrs per direction.
  // only allocated on zones that belong to current process.
  int        zoneBufferSize [kNGeomDirs]; // MPI uses int.  The number of elements = l x m x n x nVars
  QPtr    zoneRecieveBuffer [kNGeomDirs]; // single contig block for all vars in each direction
  QPtr       zoneSendBuffer [kNGeomDirs]; 
#endif MPI
  
  Integer              nQ; // number of vars/quantities used
  
#ifdef USEZONEQUAS
#ifdef ZONEDATA_ARE_PLAINBLOCKS
  QPtr    zoneData[kTotalAllVars]; // cell quantities ( values).
#endif ZONEDATA_ARE_PLAINBLOCKS
  
#ifdef ZONEDATA_ARE_MULTIDARRAYS
  QArr    zoneData[kTotalAllVars]; // cell quantities ( values). equal to QXDArr set by SYM MACRO
#endif ZONEDATA_ARE_MULTIDARRAYS
#endif USEZONEQUAS
  
#ifdef USEZONEVARS
  VArr    zoneData[kTotalAllVars]; // cell variable ( values + coeffs), always incl ghost cells
#endif USEZONEVARS
  
#ifdef PRIMITIVE_QUAS
  
  WHD         zoneW; // multi dimensional W array accessor for zone data, set up after data allocated
  // giving parallel access as a multi-d array even if data are Qptrs.  No overhead
  // if zonedata are multi-Darrays already, fields are references
#ifdef MEANSPLITSWEEPS
  WHD         zoneWLeft;
  WHD         zoneWRight; 
#endif MEANSPLITSWEEPS
  
#endif PRIMITIVE_QUAS
  
#ifdef CONSERVED_QUAS
  UHD         zoneU; // multi dimensional U array accessor for zone data, set up after data allocated
  // giving parallel access as a multi-d array even if data are Qptrs.  No overhead
  // if zonedata are multi-Darrays already, fields are references
#ifdef MEANSPLITSWEEPS
  UHD         zoneULeft;
  UHD         zoneURight; 
#endif MEANSPLITSWEEPS
  
#endif CONSERVED_QUAS
  
#ifdef DIVVHCORRECTION
  VQArr   zoneDivV;
#endif DIVVHCORRECTION
  
#ifdef LAGRANGIAN
  //
  // Auxilliary Arrays for hydrodynamics calculations
  //
  // 1D row scratch variables, pre-allocated so sweeps are threadsafe, one set for each dim as sizes will change
#ifdef SYM1D
  Flx1DArr  zone1DfluxX; // variable fluxes for 1D sweeps, separate from FlxArr above
  W1DArr        zoneW0X;
  Wstar1DArr zoneWStarX;
#endif SYM1D
  
#ifdef SYM2D
  Flx1DArr   zone1DfluxX; // variable fluxes for 1D sweeps, separate from FlxArr above
  Flx1DArr   zone1DfluxY; // variable fluxes for 1D sweeps, separate from FlxArr above
  W1DArr         zoneW0X;
  Wstar1DArr  zoneWStarX;
  W1DArr         zoneW0Y;
  Wstar1DArr  zoneWStarY;
  
#endif SYM2D
  
#ifdef SYM3D
  Flx1DArr   zone1DfluxX; // variable fluxes for 1D sweeps, separate from FlxArr above
  Flx1DArr   zone1DfluxY; // variable fluxes for 1D sweeps, separate from FlxArr above
  Flx1DArr   zone1DfluxZ; // variable fluxes for 1D sweeps, separate from FlxArr above
  W1DArr        zoneW0X;
  Wstar1DArr zoneWStarX;
  W1DArr        zoneW0Y;
  Wstar1DArr zoneWStarY;
  W1DArr        zoneW0Z;
  Wstar1DArr zoneWStarZ;
#endif SYM3D
  
#endif LAGRANGIAN
  
#ifdef GRAVITY
#ifdef SYM1D
  F1DArr        zoneF0X; // gravity field row
#endif SYM1D
  
#ifdef SYM2D
  F1DArr       zoneF0X; // gravity field row
  F1DArr       zoneF0Y; // gravity field row
#endif SYM2D
  
#ifdef SYM3D
  F1DArr       zoneF0X; // gravity field row
  F1DArr       zoneF0Y; // gravity field row
  F1DArr       zoneF0Z; // gravity field row
#endif SYM3D
#endif GRAVITY
  
  Field       zoneField; // reference to the vector force field arrays, usually gravity potential, params in level physics
    
} Zone,  *ZonePtr;

typedef ZonePtr *ZonePtrArr;

//========================================
//
// Zone Routines
//
//========================================
//
// New and Dispose calll malloc or equivalent to allocate memory, Init calls set values  below.
//
int zls_NewZone     ( ZonePtr * z );         // returns error code, 0 = noErr
int zls_DisposeZone ( ZonePtr   z );         // returns error code, 0 = noErr
//
//    zls_InitZone - Super may be NULL for top zone.  Dimensions in Range should match Super if Super is defined.
//    Level zero needs a refine of 1, and the range is then 1:1 for the zone,  for other zones ( levels 1-max)
//    the level refine factor is used to scale the given range to a higher resolution grid for the zone.
//    the range in the super zone resolution is kept for reference by the zone for passing flux corrections back up
//    the chain to the super.
//
//    Most zones get their kind from levelZoneKind of their owning level.  zones set up for 
//    images can be set with ( kZoneTypeS+kZoneTypeQua).  When a level is swept by hydro etc
//    only zones with the right kind will be processed according to the algorithm.
//    Zone kinds containing 'kZoneTypeS' won't be processed by the algorithm.
//    kAlgorithmZoneType is the type used by the current algorith will sweep
//    a zone where z->zoneType == kAlgorithmZoneType,  kAlgorithmZoneType = ( kZoneTypeW + kZoneTypeQua )
//    for ppm/hydro when PRIMITIVE_QUAS is in force.
//    
int zls_InitZone      ( ZonePtr z, Integer kind, Integer level, RangePtr rLevel, ZonePtr Super, RangePtr rSuper ); 
//
//========================================
//
// set up a two way link between two zones on the same level.  It is an error to try to link
// two zones on different levels.   The direction is from the point of view of the first zone z, 
// so kDirLeft will point the left link in z to the right side of zNext, and the right link of zNext
// to the left side of z.  The common axis co-ordinates ( in this case y) must match.
// to link to a higher resolution area, use subzones contained within z, which is extended to 
// cover the high res region as normal.  This type of linking is to allow large areas to be 
// divided into smaller chunks to aid memory and parallel processing .  Also sub zones that share
// super boundaries can be linked across within their level, reducing the need to go up and over through 
// their two super zones that are linked.  Also the overall grid could grow during a simulation by adding linked
// zones to level 0.  The idea is that a group of linked zones on a level can be 
// calculated together without reference to the super level during the level update.
//
int zls_SetZoneBCLink    ( ZonePtr z, ZonePtr zNext, Integer direction );
// when removing a link, you must provide replacement boundary types for each zone.
int zls_RemoveZoneBCLink ( ZonePtr z, Integer direction, Integer zbctype, Integer znextbctype );

int zls_SetZoneBCInterp  ( ZonePtr z, Integer interp ); // set the interpolation order for all fill from super calls

// zNext can == NULL for types other than link, if any BC is set to fill from super the fill zone from super flag is set
// This call is the only safe way to change a zone BC type value.
// as it maintains consistency as well as setting the value.  independent of dimensions.
int zls_SetZoneBCType     ( ZonePtr z, ZonePtr zNext, Integer direction, Integer zbctype );


// works with or without BCS in zoneData, as it is just copying BC Values from first cell on grid

int   zls_SetZoneConstantBCValues( ZonePtr z );
//
int   zls_PrintZoneCoords   ( FILE * f, ZonePtr z );
int   zls_ZPrintZoneCoords  ( zFile  f, ZonePtr z );
int   zls_UserLogPrintZoneCoords  ( ZonePtr z );
//
int   zls_NewZoneData         ( ZonePtr z, Integer varIndex );
int   zls_DisposeZoneData     ( ZonePtr z, Integer varIndex );

#ifdef MEANSPLITSWEEPS
int   zls_NewZoneDataAlt      ( ZonePtr z, Integer varIndex );
int   zls_DisposeZoneDataAlt  ( ZonePtr z, Integer varIndex );
#endif  MEANSPLITSWEEPS

int   zls_NewZoneAuxilliaryData      ( ZonePtr z );
int   zls_DisposeZoneAuxilliaryData  ( ZonePtr z );

int   zls_AllocateZoneBuffers        ( ZonePtr z ); // for preallocating all MPI link buffers at once
int   zls_AllocateZoneBufferDir      ( ZonePtr z, Integer dir ); // for allocating all MPI link buffers on link creation

#ifdef PRIMITIVE_QUAS
int   zls_GetZoneWHD( WHD * w, ZonePtr z );
// must be called after all the datas have been allocted successfully to set zoneW reference
int   zls_SetZoneWHD( ZonePtr z );
#endif PRIMITIVE_QUAS

#ifdef CONSERVED_QUAS
int   zls_GetZoneUHD( UHD * u, ZonePtr z );
int   zls_SetZoneUHD( ZonePtr z );
#endif  CONSERVED_QUAS
//
int  zls_NewZoneField     ( ZonePtr z );
int  zls_DisposeZoneField ( ZonePtr z );
int  zls_SetZoneField     ( ZonePtr z,  Field f );
int  zls_GetZoneField     ( Field *f, ZonePtr z );
//========================================
//
// Getting and Setting zonedata from outside sources.  Always returns a QPtr no matter what the
// zoneData internal format is.
//
// Get the data from a z->zonedata.  At the  zonedata are QPtrs/QArrs, and include
// BCs -  returned d is a simple pointer to a single block, and equals the dimensions of zonedata
//
// The Direct versions just copy the d ptr to and from the zonedata ptr, in effect a 
// new reference to the data.  d must/will equal zonedata, including any BCs.
// the caller must never dispose of d from direct calls, or the zonedata will be corrupted.
//
// Unless memory is tight or speed is an issue, use the copy routines of I/O where possible.
// If you want to get the ptr to the data to scan and set values then use the direct version,
// but don't dispose of d.  
//

// caller should  dispose d after.
// assumes the data is from outside and
// does not include BCs.
int zls_SetZoneDataCopy   ( ZonePtr z, QPtr d, Integer varIndex ); 
//
// caller mustn't dispose d after.
// d is assumed to include BCs or not
// as appropriate, and it DOESN'T check ( can't)
int zls_SetZoneDataDirect   ( ZonePtr z, QPtr d, Integer varIndex ); 
//
// no serious checks, d will include BCs according to the FILES_HAVE_BCS/FILES_NO_BCS 
// flags  macro.  Caller should dispose d after use.
int zls_GetZoneDataCopy   ( QPtr *d, ZonePtr z, Integer varIndex ); 
//
// int zls_GetZoneDataRemapped    ( QPtr *d, ZonePtr z, Integer varIndex ); 
// copy only the active area of zoneData into a new QPtr, remapping 
// any non-linear/non-cartesian coordinates - experimental
// output is the same number of pixels as original
//
// no serious checks, data will include BCs
// id the zone data does too., just copy the ptr.
// caller shouldn't dispose d, the zone keeps a ref
int zls_GetZoneDataDirect ( QPtr *d, ZonePtr z, Integer varIndex ); 
//
// get a 1D row from a zone, if the zone is 2D then use slice value to specify the type of row (X, Y, XSum YSum).  
// SliceIndex gives the row number to select in  2D, counting from the edge of the active area, -1 gives the halfway row. 
// In 3D two indices can be used to specify the row: sliceIndex and sliceIndex2 .  
// In 1D the row is just copied and the indices are ignored.
//
int   zls_GetZoneData1DRow( QPtr * d, ZonePtr z, Integer varIndex, Integer slice, Integer sliceIndex, Integer sliceIndex2);
//
// get a 2D slice from a zone,  If the zone is 3D then use slice type an index to extract a plane,
// if the zone is 2D then just copy the zone, ignoring slice and index,  If the zone is 1D then 
// it is an error.
int zls_GetZoneData2DSlice( QPtr * d, ZonePtr z, Integer varIndex, Integer slice, Integer sliceIndex );
//
//========================================
//
//========================================
//
//  Getters for extracting zone row, slice or volume Vars, ie 1D rows for split sweeps
//  creates and returns v, caller's resposiblity to dispose afterwards.
//  var arrays always have boundary cells added, quantity arrays may or may not.
//

int zls_GetZoneVar1D  ( V1DArr *v, ZonePtr z, Integer varIndex ); // 1D Zone
int zls_GetZoneVar2D  ( V2DArr *v, ZonePtr z, Integer varIndex ); // 2D Zone
int zls_GetZoneVar3D  ( V3DArr *v, ZonePtr z, Integer varIndex ); // 3D Zone
int zls_GetZoneVar      ( VArr   *v, ZonePtr z, Integer varIndex ); // dim equals zone dim automatically

//========================================

int zls_GetZone2DVar1DX ( V1DArr *v, ZonePtr z, Integer varIndex, Integer j );  
int zls_GetZone2DVar1DY ( V1DArr *v, ZonePtr z, Integer varIndex, Integer i );  

int zls_SetVar1DXZone2DQArr ( V1DArr v, Q2DArr q, ZonePtr  z, Integer j );
int zls_SetVar1DYZone2DQArr ( V1DArr v, Q2DArr q, ZonePtr  z, Integer i );

int zls_GetZone3DVar1DX ( V1DArr *v, ZonePtr z, Integer varIndex, Integer k, Integer j ); 
int zls_GetZone3DVar1DY ( V1DArr *v, ZonePtr z, Integer varIndex, Integer k, Integer i ); 
int zls_GetZone3DVar1DZ ( V1DArr *v, ZonePtr z, Integer varIndex, Integer j, Integer i ); 

int zls_GetZone3DVar2DXY( V2DArr *v, ZonePtr z, Integer varIndex, Integer k ); // x-y slice at k
int zls_GetZone3DVar2DXZ( V2DArr *v, ZonePtr z, Integer varIndex, Integer j ); // x-z slice at j
int zls_GetZone3DVar2DYZ( V2DArr *v, ZonePtr z, Integer varIndex, Integer i ); // y-z slice at i

//========================================


int zls_GetZoneFluxCopy( QPtr * d, ZonePtr z, Integer varIndex, Integer flxIndex ); 
int zls_Clear2DZoneFluxes( ZonePtr z );
int zls_Clear3DZoneFluxes( ZonePtr z );

//========================================

// each zone gets a unique ID
extern  Integer      gNextZoneID; 
extern  Zone         gMasterZone;


#endif
