#include <zls.h>

//======================================================================
//
// User problem parameter initialisation/default values routine
// Called before the user's problem config file is read to set
// all problem parameters to sensible defaults in case the init.conf
// is faultly or even missing
//
//  The fields of the problem structure are defined by the user in
//  zls_problem.h and can be any variables desired by the user's routines.
//  The problem data is stored by reference in gLevels[0] and shared by all 
//  other levels, and can be accessed by any zone at need (see zlsp_UserZoneUpdate
//  and zlsp_InitAtmosphere for examples.
//
//======================================================================
#define WIDTH 0.75
#define SUBSAMPLE 8

int zlsp_InitProblem    ( ProblemPtr   prob){
  int err = noErr;
  
  if (prob != NULL) {

    // Mach 10 shock at 30 degrees to razer edge,
    // post shock velocity projected onto grid.
    // upper boundary has analytical solution updates.
    // lower boundary is  reflecting only from x = 1/6 and above
    //  free elsewhere.  left bundary is fixed, right boundary is free
    //  Run until t = 0.25, result in 0 < x < 4.0 and 0  < y < 1.0
    //  grid is 0< x< 4.0, 0 < y < 2.0
    
    // 
    
    Real g      = kZLS_eos_gamma;

     prob->gM0 = 10.0;
     prob->gT0 = 30.0;
     prob->gEdgeX = sixth;
     
     prob->gD0 = 1.4;
     prob->gP0 = 1.0;
     prob->gv0x = 0.0;
     prob->gv0y = 0.0;
     
     Real theta = prob->gT0 * pi/ 180.0; // radians
     Real Mach  = prob->gM0;

    Real d0    = prob->gD0;
    Real P0    = prob->gP0;
    Real v0    = zls_sqrt(g*P0/d0)*Mach; // wrt to shock front
     prob->gv0 = v0;
    
    
    Real P1      = P0*((2.0*g*Mach*Mach)-(g-1.0))/(g+1.0);
    Real d1      = d0*(((g-1.0)*P0 + (g+1.0)*P1)/((g+1.0)*P0 + (g-1.0)*P1));
    Real v1      = v0 - (v0*d0/d1); // wrt preshock medium
    
    prob->gD0   = d0;
    prob->gP0   = P0;
    prob->gv0x  = 0.0;
    prob->gv0y  = 0.0;

    prob->gD1   = d1;
    prob->gP1   = P1;
    
    Real fc      = cos(theta);
    Real fs      = sin(theta);
    
    prob->gv1x   =  v1*fc;
    prob->gv1y   = -v1*fs;
    
  }
  return (err);
}

//======================================================================
// standard allocation and destructor for problem strucutres, called
// by the ZLS system at the right time, and should not need to be
// modified in any way by the user.  Here to keep all problem code together
//
// The only exception is if levels are destroyed during a run and the user
// needs to deallocate any dynamic memory set up in zlsp_InitProblem
// or in zlsp_SetUpProblemFromConfigFile.  If this is the case then 
// the user needs to deallocate the memory in zlsp_DisposeProblem
// before  free(prob) is called.
//
// Currently the level list is fixed for the entire run and this is not
// necessary, but it will be in AMR.  Will be called by zls_DisposeLevel
// which doesn't exist yet. :)
//======================================================================

int zlsp_NewProblem(ProblemPtr * prob){
  int err = noErr;
  
  *prob = (ProblemPtr)malloc(sizeof(Problem));
  
  if (*prob == NULL) {
    err = memErr;
#if (ZLS_DEBUG >0)
    zlsio_ErrPrintf( " ERROR: zlsp_NewProblem allocation failed %d \n",err);
#endif
  }
  
  return (err);
}

int zlsp_DisposeProblem(ProblemPtr prob){
  int err = noErr;
  
  if ( prob != NULL ) {
    free(prob);
  }
  
  return (err);
}


//======================================================================
// Routine to set up a hydrostatic atmosphere in a gravity field or 
// a uniform/cloudy medium if no gravity field is present
//======================================================================

int zlsp_InitAtmosphere( ZonePtr z);
int zlsp_InitAtmosphere( ZonePtr z){

//======================================================================
// Routine to set up a hydrostatic atmosphere in a gravity field or 
// a uniform/cloudy medium if no gravity field is present
//======================================================================
  
  int err = noErr;
  
  if (z != NULL){
    
    WHD zoneW = z->zoneW;    // get zone physical variables
    
    Integer level = z->zoneLevel;
    
    ProblemPtr prob = gLevels[level].levelProblem;  // get the user's problem parameters

#ifdef SYM2D 
    Integer    i,  j;   
    Integer nti  = z->zoneGrid.nt[kXDim];
    Integer ntj  = z->zoneGrid.nt[kYDim];
    ACoordPtr xx  = z->zoneGrid.x;
    ACoordPtr yx  = z->zoneGrid.y;
    
    Real dx  = xx[0].dx;
    Real dy  = yx[0].dx;
    
//    Integer seed = -1234;


//======================================================================
// Atmosphere parameters
//======================================================================
    
//     Real Mach  = prob->gM0;
     Real theta = prob->gT0 * pi/ 180.0; // radians
     Real fc    = cos(theta);
     Real fs    = sin(theta);
     Real a = (fc/fs); // slope of shockfront
    
    Real d0  = prob->gD0;
    Real P0  = prob->gP0;
    Real v_0x = prob->gv0x;
    Real v_0y = prob->gv0y;
    
    Real d1  = prob->gD1;
    Real P1  = prob->gP1;
    Real v_1x = prob->gv1x;
    Real v_1y = prob->gv1y;

//======================================================================
//
// Put on grid first to fill everywhere, replaced later by
// cool atmosphere in some or all locations
//
//======================================================================
    
      for(j = 0; j < ntj; j++){
      Real  y = yx[j].c;
      for(i = 0; i< nti; i++){
        Real x = xx[i].c;
                
        Real b = -a*sixth; // initial y-intercept of shockfront
        
        Real f  = zls_SmoothLine2D( a, b, x, y,  dx, dy, SUBSAMPLE, WIDTH);
        Real cf = (1.0-f);

        zoneW.d [j][i]      = (d1*cf) + (d0*f);
        zoneW.P [j][i]      = (P1*cf) + (P0*f);
        zoneW.vx[j][i]      = (v_1x*cf) + (v_0x*f);
        zoneW.vy[j][i]      = (v_1y*cf) + (v_0y*f);
        
        
      } //i
    } //j
#endif SYM2D 
        
  } // z != NULL
  
  return(err);
}

//======================================================================
//  Standard zone initialsation routine
//======================================================================

int zlsp_InitZoneData( ZonePtr z ){
  
//======================================================================
//  Initialise a zone system variables (ie W), usually self contained, 
//  here the initialisation is a bit messy so subroutines are called from
//  here.  When called all the variables for the zone are initialised and 
//  ready to go, grid coordinates, problem vars etc.
//======================================================================
  
  int err = noErr;
  
  if (z != NULL){
    
    Integer level = z->zoneLevel;
    
    if ( level > 0) {
      err = zlsbc_FillZoneFromSuperZone( z, kInterpBCsFromSuper );
    } else {
      err = zlsp_InitAtmosphere(z);   
    }
        
  } else { err = memErr; }//z
  
  return(err);
}

//======================================================================
//  Standard level initialsation routine, loops through all zones of
//   a level. Called by ZLS after all parameters, coordinates, physics
//   etc have been set up.  Sets the gravity field arrays before the
//   calling zlsp_InitZoneData.
//======================================================================

int zlsp_InitLevelZones(Integer level){
  
  
  int err = noErr;
  Integer zone;
  ZonePtr  z;
  
#if (ZLS_VERBOSE >0)
#ifdef MPI
    if ( gMPIrank == 0){
#endif MPI
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " Level %d User Zones Initialisation: \n", level);
      zlsio_Printf( "%s\n",  DIVIDERLINE );
#ifdef MPI
    }
#endif MPI
#endif
  
  for ( zone = 0; zone <gLevels[level].nZones; zone++){
    z =  gLevels[level].levelZones[zone];
	  if (z != NULL){
#ifdef MPI	
		  if (z->zoneRankID == gMPIrank) {
#endif MPI
			  if ( z->zoneType == kAlgorithmZoneType ){ // only process appropriate zones, ie skip image zones
				  
#if (ZLS_DEBUG>0)
				  zlsio_ErrPrintf( " Initialising Zone: %2.2d:%2.2d \n", level, zone );
#endif
				  
				  err = zlsp_InitZoneData( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsp_InitZoneData Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
          
				  err = zls_SetZoneConstantBCValues( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zls_SetZoneConstantBCValues Failed: %d,  %d \n",zone, err );
					  return(err);
				  }

#ifdef MEANSPLITSWEEPS
#ifdef SYM2D
          err =  zlsa_Synch2DData(z);
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsa_Synch2DSymmetricSweeps Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
#endif SYM2D
#endif   MEANSPLITSWEEPS   
          
         
			  } // z->zoneType == kAlgorithmZoneType 
#ifdef MPI			
		  }
#endif MPI
	  } else { err = zonErr; }
  } // zone
    
  return(err);
  
}

//======================================================================
// Optional routine to allow overriding the hard wired zlsp_InitProblem
// parameters from a config file, specified in zls.conf - usually init.conf
// Called by ZLS at the right time(tm)
//======================================================================

int   zlsp_SetUpProblemFromConfigFile(char * confFile){
  
//======================================================================
// Open the config file confFile, and parse for any user definded
// key,value pairs to set problem parameters (not grid initialisation)
// Calls zlsp_InitProblem, and before zlsp_InitLevelZones
//======================================================================
  
  int err = noErr;
  struct ccl_t  config;
  const char    *str1;
  
  Integer level;
  
  ProblemPtr  prob = NULL;
  err = zlsp_NewProblem(&prob);
  
  if ( prob != NULL ) {
    
    config.comment_char = '#';
    config.sep_char     = '=';
    config.str_char     = '"';
    
    if (ccl_parse(&config, confFile)==noErr) {
      
#if (ZLS_VERBOSE > 0) 
      zlsio_Printf( "\n%s\n", DIVIDERLINE );
#endif
      zlsio_Printf( " FYRIS Hydro Problem Initialised from %s \n", confFile);
#if (ZLS_VERBOSE > 0) 
      zlsio_Printf(  "%s\n", DIVIDERLINE );
#endif
      
// Default Values
      
      err = zlsp_InitProblem(prob);
      
//======================================================================
// The user is free to define any key strings and interpret the 
// returned values in any way.  Some problems are simple (see the SOD setup)
// this one is more complex as it sets up a multi-component model with many
// different types of parameters.
//======================================================================
            
      Real m_0 = prob->gM0;
      if (( str1 = ccl_get( &config,"Mach"))){
        m_0 = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Mach", str1);
#endif
      }
      prob->gM0 =  m_0; 


      Real t_0 = prob->gT0;
      if (( str1 = ccl_get( &config,"Theta"))){
        t_0 = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Theta", str1);
#endif
      }
      prob->gT0 =  t_0; 

      Real x_0 = prob->gEdgeX;
      if (( str1 = ccl_get( &config,"EdgeX"))){
        x_0 = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "EdgeX", str1);
#endif
      }
      prob->gEdgeX =  x_0; 

      Real d_0 = prob->gD0;
      if (( str1 = ccl_get( &config,"Density"))){
        d_0 = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Density", str1);
#endif
      }
      prob->gD0 =  d_0; 


      Real p_0 = prob->gP0;
      if (( str1 = ccl_get( &config,"Pressure"))){
        p_0 = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Pressure", str1);
#endif
      }
      prob->gP0 =  p_0; 

//======================================================================
// Atmosphere parameters
//======================================================================
      {
//
// WIERD C NOTE:
// brace sets up a local scope for temp variables to help reduce brief 
// named variables colliding with any used later in the routine. ie Th 
// could be used here and later in the cluster parameter setup without
// interfering with eachother
//
        
      Real g      = kZLS_eos_gamma;
      Real theta = prob->gT0 * pi/ 180.0; // radians
      Real Mach  = prob->gM0;

      Real d0    = prob->gD0;
      Real P0    = prob->gP0;
      Real v0    = zls_sqrt(g*P0/d0)*Mach; // wrt to shock front
      prob->gv0 = v0;


      Real P1      = P0*((2.0*g*Mach*Mach)-(g-1.0))/(g+1.0);
      Real d1      = d0*(((g-1.0)*P0 + (g+1.0)*P1)/((g+1.0)*P0 + (g-1.0)*P1));
      Real v1      = v0 - (v0*d0/d1); // wrt preshock medium

      prob->gD1   = d1;
      prob->gP1   = P1;

      Real fc      = cos(theta);
      Real fs      = sin(theta);

      prob->gv1x   =  v1*fc;
      prob->gv1y   = -v1*fs;

      zlsio_Printf( " Flow Parameters:\n");
      zlsio_Printf("   d        %g \n", prob->gD1);    
      zlsio_Printf("   P        %g \n", prob->gP1); 
      zlsio_Printf("   v        %g \n", v1);    
      zlsio_Printf("   v_x      %g \n", prob->gv1x);    
      zlsio_Printf("   v_y      %g \n", prob->gv1y); 
      zlsio_Printf("%s\n",  DIVIDERLINE );
      zlsio_Printf( " Atmosphere Parameters:\n");
      zlsio_Printf("   d        %g \n", prob->gD0);    
      zlsio_Printf("   P        %g \n", prob->gP0); 
      zlsio_Printf("%s\n",  DIVIDERLINE );
      zlsio_Printf( " Other Parameters:\n");
      zlsio_Printf("   Theta       %g \n", prob->gT0);    
      zlsio_Printf("   Mach        %g \n", prob->gM0); 
      zlsio_Printf("   Gamma       %g \n", g);    
      zlsio_Printf("   EdgeX       %g \n", prob->gEdgeX); 
      }      
      
//======================================================================
// This code sets the problem into the level structures, mulit-zone
// multi-level or simple single structures 
//======================================================================
      
      
      gLevels[0].levelProblem = prob; // register the problem with all levels from 0 down.      
      if (gNLevels>1){
        for (level = 1; level < gNLevels; level++){
          gLevels[level].levelProblem = gLevels[0].levelProblem;
        }
      }
      
//======================================================================
// Now the level contents can be set - standard calls here dont' change
//======================================================================
      
      for (level = 0; level < gNLevels; level++){
        
#ifdef GRAVITY
        err =  zlsg_InitLevelGravityField( level ); // precompute the potential of to the gravity field over the level
        if (err != noErr ) {
          zlsio_ErrPrintf( " ERROR: zlsg_InitLevelGravityField: Level Field Failed: %d,  %d \n",level, err );
          return(err);
        }
#endif GRAVITY
#ifdef NEQIONISATION
        err =  zlsi_InitLevelIonisationState( level ); 
        if (err != noErr ) {
          zlsio_ErrPrintf( " ERROR: zlsi_InitLevelIonisationState: Level Failed: %d,  %d \n",level, err );
          return(err);
        }
#endif NEQIONISATION
        
        err = zlsp_InitLevelZones( level ); // last: user's levelProblem, initialised top down
        
      }
      
      if (gNLevels>1){
        for (level = (gNLevels-1); level > 0; level--){
          err = zlsbc_FillSuperLevelFromLevel(level);
        }
      }
      
#if (ZLS_VERBOSE>0)
      zlsio_Printf("%s\n",  DIVIDERLINE );
#else
      zlsio_LogPrintf("%s\n",  DIVIDERLINE );
#endif
      
      ccl_release(&config);
      
    }// config  
    
  } else {
    err = memErr;
    zlsio_ErrPrintf( " ERROR: Unable to allocate level 0 ProblemPtr \n");
    return (err);
  }
  
  return (err);
  
}

//======================================================================
// Routine for 'User; boundary cells set in grid.conf not used here
// but often used for jet inlet models
//======================================================================
#ifdef ZLS_USER_BOUNDARIES
int  zlsp_CalcUserZoneBCs ( ZonePtr z, int direction, int bndryKind ){
  Integer err = noErr;
  ProblemPtr prob = NULL;
  PhysicsPtr phys = NULL;
  if ( z != NULL ){
    
    Integer level = z->zoneLevel;
    prob = gLevels[level].levelProblem;  // get the user's problem parameters
    phys = gLevels[level].levelPhysics;  // get the global eos parameters

//    WHD zoneW = z->zoneW;    
    
    Integer   id;   
    QArr       q;

    ACoordPtr xx  = z->zoneGrid.x;
    ACoordPtr yx  = z->zoneGrid.y;
    
    Integer is    = z->zoneGrid.ns[kXDim];
    Integer ie    = z->zoneGrid.ne[kXDim];
    
    Integer js    = z->zoneGrid.ns[kYDim];
    Integer je    = z->zoneGrid.ne[kYDim];
    
    Integer nti   = z->zoneGrid.nt[kXDim];
    Integer ntj   = z->zoneGrid.nt[kYDim];
    
    Real dx  = xx[0].dx;
    Real dy  = yx[0].dx;
    
    Integer i, j;

    Integer  nbc  = z->zoneBCells;
   
     Real v0    =  prob->gv0;
     Real theta = prob->gT0 * pi/ 180.0; // radians
     Real fc    = cos(theta);
     Real fs    = sin(theta);
     Real a     = (fc/fs);
     Real edge  = prob->gEdgeX;
    
    Real P0     = prob->gP0;
    Real d0     = prob->gD0;
    
    Real v_0x     = prob->gv0x;
    Real v_0y     = prob->gv0y;
    
    Real P1     = prob->gP1;
    Real d1     = prob->gD1;

    Real v_1x     = prob->gv1x;
    Real v_1y     = prob->gv1y;

    Real x      = 0.0;
    Real y      = 0.0;
//    Real fi, fj;

	   Real t      = zlst_GlobalTime;
//     Real dt     = zlst_GlobalTimeStep;
    
 		for (id = 0; id <  z->nQ; id++){
      
			if (z->zoneData[id]!=NULL){
				
#ifdef ZONEDATA_ARE_PLAINBLOCKS
				zls_SetQuantity2DArr(&q, z->zoneData[id], ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
				
#ifdef ZONEDATA_ARE_MULTIDARRAYS
				q = z->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
				
				if (q!=NULL){
                        
          if (direction == kDirLeft)  {
            for(i = 0; i < nbc; i++){
              x = xx[is-i-1].c;
              for(j = 0; j < ntj; j++){ 
                y = yx[j].c;
                
                switch (id) {
                  case kWDens:
                    q[j][is-i-1] = d1;
                    break;
                  case kWPres:
                    q[j][is-i-1] = P1;
                    break;
                  case kWVx:
                    q[j][is-i-1] = v_1x;
                    break;
                  case kWVy:
                    q[j][is-i-1] = v_1y;
                    break;
                  default:
                    ;
                } // id switch
              } //j
            } //i
          } // direction switch
          
          if (direction == kDirRight)  {
            for(i = 0; i < nbc; i++){
              x = xx[ie+i].c;
              for(j = 0; j < ntj; j++){ 
                y = yx[j].c;
                
                
                switch (id) {
                  case kWDens:
                    q[j][ie+i] = d1;
                    break;
                  case kWPres:
                    q[j][ie+i] = P1;
                    break;
                  case kWVx:
                    q[j][ie+i] = v_1x;
                    break;
                  case kWVy:
                    q[j][ie+i] = v_1y;
                    break;
                  default:
                    ;
                } // id switch
              } //j
            } //i
          } // direction switch
          
          
          if (direction == kDirUp)  {
                        
            for(j = 0; j < nbc; j++){ 
              y = yx[je+j].c;
              
              for(i = 0; i < nti; i++){
                x = xx[i].c;
                
         //       Real s = f*(x - (sixth)) - ((v0/fs)*t);
               Real b = (-a*sixth) - ((v0/fs)*t);
        
              // Real f  = zls_OverlapLine2D( a, b, x, y,  dx, dy, SUBSAMPLE );
               Real f  = zls_SmoothLine2D( a, b, x, y,  dx, dy, SUBSAMPLE, WIDTH);
               Real cf = (1.0-f);
                switch (id) {
                  case kWDens:
                    q[je+j][i] = (d1*cf) + (d0*f);
                    break;
                  case kWPres:
                    q[je+j][i] = (P1*cf) + (P0*f);
                    break;
                  case kWVx:
                    q[je+j][i] = (v_1x*cf) + (v_0x*f);
                    break;
                  case kWVy:
                    q[je+j][i] = (v_1y*cf) + (v_0y*f);
                    break;
                  default:
                    ;
                } // id switch
                  
            } //i
            } //j
          } // direction switch
          
          if (direction == kDirDown)  {
            for(i = 0; i < nti; i++){
              x = xx[i].c;
              for(j = 0; j < nbc; j++){ 
                y = yx[je+j].c;
                  
                if ( x < edge ) { // free
											q[j][i] = q[js][i];
                } else { // reflect
                  switch (id) {
                    case kWDens:
                      q[js-j-1][i] =  q[js+j][i];
                      break;
                    case kWPres:
                      q[js-j-1][i] =  q[js+j][i];
                      break;
                    case kWVx:
                      q[js-j-1][i] =  q[js+j][i];
                      break;
                    case kWVy:
                      q[js-j-1][i] = -q[js+j][i];
                      break;
                    default:
                      ;
                  } // id switch
                  
                }
                
              } //j
            } //i
          } // direction switch
          
#ifdef ZONEDATA_ARE_PLAINBLOCKS
					zls_ReleaseQuantity2DArr(q);
#endif ZONEDATA_ARE_PLAINBLOCKS
          
				} else { err = memErr; } // q
			} // z->zoneData[id]!=NULL
		} // var id loop

          
  } else { err = memErr; } // z
  
  return (err);
}
#endif ZLS_USER_BOUNDARIES


//======================================================================
// Routine for modifying anywhere in the grid in any way as part of
// a custom problem model. Not restricted to boundary cells, here used for 
// wind bubble/mass loss bubble energy injection and mass injection
// called after all other boundary conditions have been computed.
// Use sparingly, if not careful this routine can impact heavily on 
// the performance of the simulation.
//======================================================================
// 
// Normally deactivated (ie not even checked for calling) unless the macro
// ZLS_USER_UPDATE_ROUTINE is defined in zls_defs.  If not defined then
// this routine is not even callable - and this is the default setting.
// Also normally provided as an empty routine.
//

#ifdef ZLS_USER_UPDATE_ROUTINE

int  zlsp_UserZoneUpdate ( ZonePtr z  ){
  
  Integer err = noErr;
  ProblemPtr prob = NULL;
  PhysicsPtr phys = NULL;
  
  // A different type of BC, not the grid edges as all but setting cell values
  // on the grid.  Uses zoneW and is based on the zone data init routine rather than
  // the grid edge code of the basic boundary value routines.
  
  if ( z != NULL ){
    
    Integer level = z->zoneLevel;
    prob = gLevels[level].levelProblem;  // get the user's problem parameters
    phys = gLevels[level].levelPhysics;  // get the global eos parameters
    
    if ( level == (gNLevels-1)) {
      
#ifdef COOLING
    PhysicsPtr phys = gLevels[0].levelPhysics;  // get the global Physics parameters
    CoolingPtr cp   = NULL;  // get the coolng parameters
    cp   = &(phys->coolParameters);  // get the coolng parameters
#endif
        
    } // level
    
  } else { err = memErr; } // z
  
  return (err);
}
#endif ZLS_USER_UPDATE_ROUTINE
