#include <zls.h>

//======================================================================
//
// User problem parameter initialisation/default values routine
// Called before the user's problem config file is read to set
// all problem parameters to sensible defaults in case the init.conf
// is faulty or even missing
//
//  The fields of the problem structure are defined by the user in
//  zls_problem.h and can be any variables desired by the user's routines.
//  The problem data is stored by reference in gLevels[0] and shared by all 
//  other levels, and can be accessed by any zone at need (see zlsp_UserZoneUpdate
//  and zlsp_InitAtmosphere for examples.
//
//======================================================================

int zlsp_InitProblem    ( ProblemPtr   prob){
  int err = noErr;
  
  if (prob != NULL) {
    
    prob->gD0      = 1.0;
    prob->gP0      = 1.0;
    prob->gV0      = 15.0;
    
  }
  return (err);
}

//======================================================================
// standard allocation and destructor for problem strucutres, called
// by the ZLS system at the right time, and should not need to be
// modified in any way by the user.  Here to keep all problem code together
//
// The only exception is if levels are destroyed during a run and the user
// needs to deallocate any dynamic memory set up in zlsp_InitProblem
// or in zlsp_SetUpProblemFromConfigFile.  If this is the case then 
// the user needs to deallocate the memory in zlsp_DisposeProblem
// before  free(prob) is called.
//
// Currently the level list is fixed for the entire run and this is not
// necessary, but it will be in AMR.  Will be called by zls_DisposeLevel
// which doesn't exist yet. :)
//======================================================================

int zlsp_NewProblem(ProblemPtr * prob){
  int err = noErr;
  
  *prob = (ProblemPtr)malloc(sizeof(Problem));
  
  if (*prob == NULL) {
    err = memErr;
#if (ZLS_DEBUG >0)
    zlsio_ErrPrintf( " ERROR: zlsp_NewProblem allocation failed %d \n",err);
#endif
  }
  
  return (err);
}

int zlsp_DisposeProblem(ProblemPtr prob){
  int err = noErr;
  
  if ( prob != NULL ) {
    free(prob);
  }
  
  return (err);
}

//======================================================================

//======================================================================
//  Standard zone initialsation routine
//======================================================================
                    
int zlsp_InitZoneData( ZonePtr z ){
  
//======================================================================
//  Initialise a zone system variables (ie W), usually self contained, 
//  here the initialisation is a bit messy so subroutines are called from
//  here.  When called all the variables for the zone are initialised and 
//  ready to go, grid coordinates, problem vars etc.
//======================================================================
  
  int err = noErr;
  
  if (z != NULL){
    
    WHD zoneW = z->zoneW;    // get zone physical variables
    
    ProblemPtr prob = gLevels[0].levelProblem;  // get the user's problem parameters
    
    Real d0  = prob->gD0;
    Real P0  = prob->gP0;
    Real v0  = prob->gV0;
    
#ifdef SYM1D
    Integer    i ;
    Integer nti   = z->zoneGrid.nt[kXDim];
    ACoordPtr xx  = z->zoneGrid.x;
    
    for(i = 0; i< nti; i++){
        zoneW.d [i] = d0;
        zoneW.P [i] = P0;
        zoneW.vx[i] = v0;
    }//i
      // set wind density after outburst to low value for BC  updates
#endif SYM1D
    
#ifdef SYM2D
    
    Integer    i, j;
    Integer nti   = z->zoneGrid.nt[kXDim];
    Integer ntj   = z->zoneGrid.nt[kYDim];
    
    for(j = 0; j < ntj; j++){
      for(i = 0; i < nti; i++){
        
        zoneW.d [j][i] = d0;
        zoneW.P [j][i] = P0;
        zoneW.vx[j][i] = v0;
        zoneW.vy[j][i] = 0.0;
      
    }//i
    } //j

#endif SYM2D
    
#ifdef SYM3D

    Integer    i, j, k;
    
    Integer nti   = z->zoneGrid.nt[kXDim];
    Integer ntj   = z->zoneGrid.nt[kYDim];
    Integer ntk   = z->zoneGrid.nt[kZDim];
    
    for(k = 0; k < ntk; k++){
      for(j = 0; j < ntj; j++){
        for(i = 0; i < nti; i++){
            zoneW.d [k][j][i]      =  d0;
            zoneW.P [k][j][i]      =  P0;
            zoneW.vx[k][j][i]      =  v0;
            zoneW.vy[k][j][i]      =  0.0;
            zoneW.vz[k][j][i]      =  0.0;
        }//i
      } //j
    } //k
    
#endif SYM3D
    
  } else { err = memErr; }//z
  
  return(err);
}


//======================================================================
//  Standard level initialsation routine, loops through all zones of
//   a level. Called by ZLS after all parameters, coordinates, physics
//   etc have been set up.  Sets the gravity field arrays before the
//   calling zlsp_InitZoneData.
//======================================================================

int zlsp_InitLevelZones(Integer level){
  
  
  int err = noErr;
  Integer zone;
  ZonePtr  z;
  
#if (ZLS_VERBOSE >0)
#ifdef MPI
    if ( gMPIrank == 0){
#endif MPI
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " Level %d User Zones Initialisation: \n", level);
      zlsio_Printf( "%s\n",  DIVIDERLINE );
#ifdef MPI
    }
#endif MPI
#endif

  for ( zone = 0; zone <gLevels[level].nZones; zone++){
    z =  gLevels[level].levelZones[zone];
	  if (z != NULL){
#ifdef MPI	
		  if (z->zoneRankID == gMPIrank) {
#endif MPI
			  if ( z->zoneType == kAlgorithmZoneType ){ // only process appropriate zones, ie skip image zones
				  
#if (ZLS_DEBUG>0)
				  zlsio_ErrPrintf( " Initialising Zone: %2.2d:%2.2d \n", level, zone );
#endif
				  
				  err = zlsp_InitZoneData( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsp_InitZoneData Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
          
				  err = zls_SetZoneConstantBCValues( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zls_SetZoneConstantBCValues Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
#ifdef MEANSPLITSWEEPS
#ifdef SYM2D
          err =  zlsa_Synch2DData(z);
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsa_Synch2DSymmetricSweeps Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
#endif SYM2D
#endif   MEANSPLITSWEEPS   
          
          
			  } // z->zoneType == kAlgorithmZoneType 
#ifdef MPI			
		  }
#endif MPI
	  } else { err = zonErr; }
  } // zone
    
  return(err);
  
}


//======================================================================
// Optional routine to allow overriding the hard wired zlsp_InitProblem
// parameters from a config file, specified in zls.conf - usually init.conf
// Called by ZLS at the right time(tm)
//======================================================================

int   zlsp_SetUpProblemFromConfigFile(char * confFile){
  
//======================================================================
// Open the config file confFile, and parse for any user definded
// key,value pairs to set problem parameters (not grid initialisation)
// Calls zlsp_InitProblem, and before zlsp_InitLevelZones
//======================================================================
  
  int err = noErr;
  struct ccl_t  config;
  const char    *str1;
  
  Integer level;

#ifdef COOLING
  PhysicsPtr phys = gLevels[0].levelPhysics;  // get the global Physics parameters
  CoolingPtr cp   = NULL; 
  cp   = &(phys->coolParameters);  // get the coolng parameters
  Real  mu  = k_mu;
#endif COOLING
  
  ProblemPtr  prob = NULL;
  err = zlsp_NewProblem(&prob);
  
  if ( prob != NULL ) {
    
    config.comment_char = '#';
    config.sep_char     = '=';
    config.str_char     = '"';
    
    if (ccl_parse(&config, confFile)==noErr) {
      
#if (ZLS_VERBOSE > 0) 
      zlsio_Printf( "\n%s\n", DIVIDERLINE );
#endif
      zlsio_Printf( " FYRIS Hydro Problem Initialised from %s \n", confFile);
#if (ZLS_VERBOSE > 0) 
      zlsio_Printf(  "%s\n", DIVIDERLINE );
#endif
      
// Default Values
      
      err = zlsp_InitProblem(prob);
      
//======================================================================
// The user is free to define any key strings and interpret the 
// returned values in any way.  Some problems are simple (see the SOD setup)
// this one is more complex as it sets up a multi-component model with many
// different types of parameters.
//======================================================================
          
      Real va = prob->gV0;
      if (( str1 = ccl_get( &config, "V0"))){
        va = atof(str1)/f_v0;
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "V0", str1);
#endif
      }
      prob->gV0 = va;
      
      
      Real P = prob->gP0;
      if (( str1 = ccl_get( &config, "P0"))){
        P = atof(str1)/f_P0;
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "P0", str1);
#endif
      }
      prob->gP0 = P;
      
      Real d = prob->gD0;
      if (( str1 = ccl_get( &config, "D0"))){
        d = atof( str1)/f_d0;
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "D0", str1);
#endif
      }
      prob->gD0   = d;
      
      
      zlsio_Printf("%s\n",  DIVIDERLINE );
      zlsio_Printf( " Parameters:\n");
      zlsio_Printf("%s\n",  DIVIDERLINE );
      zlsio_Printf("   d0   %g \n", prob->gD0);    
      zlsio_Printf("   P0   %g \n", prob->gP0); 
      zlsio_Printf("   V0   %g \n", prob->gV0); 
      zlsio_Printf("%s\n",  DIVIDERLINE );
      
      
//======================================================================
// This code sets the problem into the level structures, mulit-zone
// multi-level or simple single structures 
//======================================================================
      
      
      gLevels[0].levelProblem = prob; // register the problem with all levels from 0 down.      
      if (gNLevels>1){
        for (level = 1; level < gNLevels; level++){
          gLevels[level].levelProblem = gLevels[0].levelProblem;
        }
      }
      
//======================================================================
// Now the level contents can be set - standard calls here dont' change
//======================================================================
      
      for (level = 0; level < gNLevels; level++){
        
#ifdef GRAVITY
        err =  zlsg_InitLevelGravityField( level ); // precompute the potential of to the gravity field over the level
        if (err != noErr ) {
          zlsio_ErrPrintf( " ERROR: zlsg_InitLevelGravityField: Level Field Failed: %d,  %d \n",level, err );
          return(err);
        }
#endif GRAVITY
#ifdef NEQIONISATION
        err =  zlsi_InitLevelIonisationState( level ); 
        if (err != noErr ) {
          zlsio_ErrPrintf( " ERROR: zlsi_InitLevelIonisationState: Level Failed: %d,  %d \n",level, err );
          return(err);
        }
#endif NEQIONISATION
        
        err = zlsp_InitLevelZones( level ); // last: user's levelProblem, initialised top down
        
      }
      
      if (gNLevels>1){
        for (level = (gNLevels-1); level > 0; level--){
          err = zlsbc_FillSuperLevelFromLevel(level);
        }
      }
      
#if (ZLS_VERBOSE>0)
      zlsio_Printf("%s\n",  DIVIDERLINE );
#else
      zlsio_LogPrintf("%s\n",  DIVIDERLINE );
#endif
      
      ccl_release(&config);
      
    }// config  
    
  } else {
    err = memErr;
    zlsio_ErrPrintf( " ERROR: Unable to allocate level 0 ProblemPtr \n");
    return (err);
  }
  
  return (err);
  
}

//======================================================================
// Routine for 'User; boundary cells set in grid.conf not used here
// but often used for jet inlet models
//======================================================================
#ifdef ZLS_USER_BOUNDARIES
int  zlsp_CalcUserZoneBCs ( ZonePtr z, int direction, int bndryKind ){
  return (noErr);
}
#endif ZLS_USER_BOUNDARIES


//======================================================================
// Routine for modifying anywhere in the grid in any way as part of
// a custom problem model. Not restricted to boundary cells, here used for 
// wind bubble/mass loss bubble energy injection and mass injection
// called after all other boundary conditions have been computed.
// Use sparingly, if not careful this routine can impact heavily on 
// the performance of the simulation.
//======================================================================
// 
// Normally deactivated (ie not even checked for calling) unless the macro
// ZLS_USER_UPDATE_ROUTINE is defined in zls_defs.  If not defined then
// this routine is not even callable - and this is the default setting.
// Also normally provided as an empty routine.
//

#ifdef ZLS_USER_UPDATE_ROUTINE

int  zlsp_UserZoneUpdate ( ZonePtr z  ){
  
  Integer err = noErr;
  ProblemPtr prob = NULL;
  PhysicsPtr phys = NULL;
  
  // A different type of BC, not the grid edges as all but setting cell values
  // on the grid.  Uses zoneW and is based on the zone data init routine rather than
  // the grid edge code of the basic boundary value routines.
  
  if ( z != NULL ){
    
    Integer level = z->zoneLevel;
    prob = gLevels[level].levelProblem;  // get the user's problem parameters
    phys = gLevels[level].levelPhysics;  // get the global eos parameters
    
    if ( level == (gNLevels-1)) {
      
#ifdef COOLING
    PhysicsPtr phys = gLevels[0].levelPhysics;  // get the global Physics parameters
    CoolingPtr cp   = NULL;  // get the coolng parameters
    cp   = &(phys->coolParameters);  // get the coolng parameters
#endif

    
    } // level
    
  } else { err = memErr; } // z
  
  return (err);
}
#endif ZLS_USER_UPDATE_ROUTINE
