#include <zls.h>

//======================================================================
//
// User problem parameter initialisation/default values routine
// Called before the user's problem config file is read to set
// all problem parameters to sensible defaults in case the init.conf
// is faultly or even missing
//
//  The fields of the problem structure are defined by the user in
//  zls_problem.h and can be any variables desired by the user's routines.
//  The problem data is stored by reference in gLevels[0] and shared by all 
//  other levels, and can be accessed by any zone at need (see zlsp_UserZoneUpdate
//  and zlsp_InitAtmosphere for examples.
//
//======================================================================

int zlsp_InitProblem    ( ProblemPtr   prob){
  int err = noErr;
  
  if (prob != NULL) {
    
    Real g      = kZLS_eos_gamma;

     prob->gM0 = 3.0;
     
     prob->gD0 = 1.4;
     prob->gP0 = 1.0;
     prob->gv0 = 0.0;
     
     Real Mach  = prob->gM0;

     Real d0    = prob->gD0;
     Real P0    = prob->gP0;
     Real v0    = zls_sqrt(g*P0/d0)*Mach; 
     prob->gv0 = v0;
    
  }
  return (err);
}

//======================================================================
// standard allocation and destructor for problem strucutres, called
// by the ZLS system at the right time, and should not need to be
// modified in any way by the user.  Here to keep all problem code together
//
// The only exception is if levels are destroyed during a run and the user
// needs to deallocate any dynamic memory set up in zlsp_InitProblem
// or in zlsp_SetUpProblemFromConfigFile.  If this is the case then 
// the user needs to deallocate the memory in zlsp_DisposeProblem
// before  free(prob) is called.
//
// Currently the level list is fixed for the entire run and this is not
// necessary, but it will be in AMR.  Will be called by zls_DisposeLevel
// which doesn't exist yet. :)
//======================================================================

int zlsp_NewProblem(ProblemPtr * prob){
  int err = noErr;
  
  *prob = (ProblemPtr)malloc(sizeof(Problem));
  
  if (*prob == NULL) {
    err = memErr;
#if (ZLS_DEBUG >0)
    zlsio_ErrPrintf( " ERROR: zlsp_NewProblem allocation failed %d \n",err);
#endif
  }
  
  return (err);
}

int zlsp_DisposeProblem(ProblemPtr prob){
  int err = noErr;
  
  if ( prob != NULL ) {
    free(prob);
  }
  
  return (err);
}


//======================================================================
// Routine to set up a hydrostatic atmosphere in a gravity field or 
// a uniform/cloudy medium if no gravity field is present
//======================================================================

int zlsp_InitAtmosphere( ZonePtr z);
int zlsp_InitAtmosphere( ZonePtr z){

//======================================================================
// Routine to set up a hydrostatic atmosphere in a gravity field or 
// a uniform/cloudy medium if no gravity field is present
//======================================================================
  
  int err = noErr;
  
  if (z != NULL){
    
    WHD zoneW = z->zoneW;    // get zone physical variables
    
    Integer level = z->zoneLevel;
    
    ProblemPtr prob = gLevels[level].levelProblem;  // get the user's problem parameters

#ifdef SYM2D 
    Integer    i,  j;   
    Integer nti  = z->zoneGrid.nt[kXDim];
    Integer ntj  = z->zoneGrid.nt[kYDim];
    
    
//======================================================================
// Atmosphere parameters
//======================================================================
    
    
    Real d0  = prob->gD0;
    Real P0  = prob->gP0;
    Real v_0 = prob->gv0;

//======================================================================
//
// Put on grid first to fill everywhere, replaced later by
// cool atmosphere in some or all locations
//
//======================================================================
    
      for(j = 0; j < ntj; j++){
      for(i = 0; i< nti; i++){
                
        zoneW.d [j][i]  = d0;
        zoneW.P [j][i]  = P0;
        zoneW.vx[j][i]  = v_0;
        zoneW.vy[j][i]  = 0.0;
        
        
      } //i
    } //j
#endif SYM2D 
        
  } // z != NULL
  
  return(err);
}

//======================================================================
//  Standard zone initialsation routine
//======================================================================

int zlsp_InitZoneData( ZonePtr z ){
  
//======================================================================
//  Initialise a zone system variables (ie W), usually self contained, 
//  here the initialisation is a bit messy so subroutines are called from
//  here.  When called all the variables for the zone are initialised and 
//  ready to go, grid coordinates, problem vars etc.
//======================================================================
  
  int err = noErr;
  
  if (z != NULL){
    
    Integer level = z->zoneLevel;
    
    if ( level > 0) {
      err = zlsbc_FillZoneFromSuperZone( z, kInterpBCsFromSuper );
    } else {
      err = zlsp_InitAtmosphere(z);   
    }
        
  } else { err = memErr; }//z
  
  return(err);
}

//======================================================================
//  Standard level initialsation routine, loops through all zones of
//   a level. Called by ZLS after all parameters, coordinates, physics
//   etc have been set up.  Sets the gravity field arrays before the
//   calling zlsp_InitZoneData.
//======================================================================

int zlsp_InitLevelZones(Integer level){
  
  
  int err = noErr;
  Integer zone;
  ZonePtr  z;
  
#ifdef MPI
    if ( gMPIrank == 0){
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " ERROR: zlsp_InitLevelZones: \n",);
      zlsio_Printf( "       THIS PROBLEM WILL NOT RUN \n",);
      zlsio_Printf( "       UNDER MPI DUE TO zlsp_UserZoneUpdate \n",);
      zlsio_Printf( "       REFERRING ACROSS ZONES FOR ENTROPY\n",);
      zlsio_Printf( "       CONDITION, USE PTHREADS OR SINGLE CPU.\n",);
      zlsio_Printf( "%s\n",  DIVIDERLINE );
    }
    return ( mpiErr );
#endif MPI

  
#if (ZLS_VERBOSE >0)
#ifdef MPI
    if ( gMPIrank == 0){
#endif MPI
      zlsio_Printf( "%s\n",  DIVIDERLINE );
      zlsio_Printf( " Level %d User Zones Initialisation: \n", level);
      zlsio_Printf( "%s\n",  DIVIDERLINE );
#ifdef MPI
    }
#endif MPI
#endif
  
  for ( zone = 0; zone <gLevels[level].nZones; zone++){
    z =  gLevels[level].levelZones[zone];
	  if (z != NULL){
#ifdef MPI	
		  if (z->zoneRankID == gMPIrank) {
#endif MPI
			  if ( z->zoneType == kAlgorithmZoneType ){ // only process appropriate zones, ie skip image zones
				  
#if (ZLS_DEBUG>0)
				  zlsio_ErrPrintf( " Initialising Zone: %2.2d:%2.2d \n", level, zone );
#endif
				  
				  err = zlsp_InitZoneData( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsp_InitZoneData Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
          
				  err = zls_SetZoneConstantBCValues( z );
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zls_SetZoneConstantBCValues Failed: %d,  %d \n",zone, err );
					  return(err);
				  }

#ifdef MEANSPLITSWEEPS
#ifdef SYM2D
          err =  zlsa_Synch2DData(z);
				  if (err != noErr ) {
					  zlsio_ErrPrintf( " ERROR: zlsp_InitLevelZones: zlsa_Synch2DSymmetricSweeps Failed: %d,  %d \n",zone, err );
					  return(err);
				  }
#endif SYM2D
#endif   MEANSPLITSWEEPS   
          
         
			  } // z->zoneType == kAlgorithmZoneType 
#ifdef MPI			
		  }
#endif MPI
	  } else { err = zonErr; }
  } // zone
    
  return(err);
  
}

//======================================================================
// Optional routine to allow overriding the hard wired zlsp_InitProblem
// parameters from a config file, specified in zls.conf - usually init.conf
// Called by ZLS at the right time(tm)
//======================================================================

int   zlsp_SetUpProblemFromConfigFile(char * confFile){
  
//======================================================================
// Open the config file confFile, and parse for any user definded
// key,value pairs to set problem parameters (not grid initialisation)
// Calls zlsp_InitProblem, and before zlsp_InitLevelZones
//======================================================================
  
  int err = noErr;
  struct ccl_t  config;
  const char    *str1;
  
  Integer level;
  
  ProblemPtr  prob = NULL;
  err = zlsp_NewProblem(&prob);
  
  if ( prob != NULL ) {
    
    config.comment_char = '#';
    config.sep_char     = '=';
    config.str_char     = '"';
    
    if (ccl_parse(&config, confFile)==noErr) {
      
#if (ZLS_VERBOSE > 0) 
      zlsio_Printf( "\n%s\n", DIVIDERLINE );
#endif
      zlsio_Printf( " FYRIS Hydro Problem Initialised from %s \n", confFile);
#if (ZLS_VERBOSE > 0) 
      zlsio_Printf(  "%s\n", DIVIDERLINE );
#endif
      
// Default Values
      
      err = zlsp_InitProblem(prob);
      
//======================================================================
// The user is free to define any key strings and interpret the 
// returned values in any way.  Some problems are simple (see the SOD setup)
// this one is more complex as it sets up a multi-component model with many
// different types of parameters.
//======================================================================
            
      Real m_0 = prob->gM0;
      if (( str1 = ccl_get( &config,"Mach"))){
        m_0 = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Mach", str1);
#endif
      }
      prob->gM0 =  m_0; 


      Real d_0 = prob->gD0;
      if (( str1 = ccl_get( &config,"Density"))){
        d_0 = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Density", str1);
#endif
      }
      prob->gD0 =  d_0; 


      Real p_0 = prob->gP0;
      if (( str1 = ccl_get( &config,"Pressure"))){
        p_0 = atof( str1);
#if (ZLS_DEBUG>0)
        zlsio_Printf( "   %s = %s\n", "Pressure", str1);
#endif
      }
      prob->gP0 =  p_0; 

//======================================================================
// Atmosphere parameters
//======================================================================
      {
//
// WIERD C NOTE:
// brace sets up a local scope for temp variables to help reduce brief 
// named variables colliding with any used later in the routine. ie Th 
// could be used here and later in the cluster parameter setup without
// interfering with eachother
//
        
      Real g      = kZLS_eos_gamma;
      Real Mach  = prob->gM0;

      Real d0    = prob->gD0;
      Real P0    = prob->gP0;
      Real v0    = zls_sqrt(g*P0/d0)*Mach; // wrt to shock front
      prob->gv0 = v0;

      zlsio_Printf( " Flow Parameters:\n");
      zlsio_Printf("   d        %g \n", prob->gD0);    
      zlsio_Printf("   P        %g \n", prob->gP0); 
      zlsio_Printf("   v        %g \n", v0);    
      zlsio_Printf("%s\n",  DIVIDERLINE );
      zlsio_Printf( " Other Parameters:\n");
      zlsio_Printf("   Mach        %g \n", prob->gM0); 
      zlsio_Printf("   Gamma       %g \n", g);    
      }      
      
//======================================================================
// This code sets the problem into the level structures, mulit-zone
// multi-level or simple single structures 
//======================================================================
      
      
      gLevels[0].levelProblem = prob; // register the problem with all levels from 0 down.      
      if (gNLevels>1){
        for (level = 1; level < gNLevels; level++){
          gLevels[level].levelProblem = gLevels[0].levelProblem;
        }
      }
      
//======================================================================
// Now the level contents can be set - standard calls here dont' change
//======================================================================
      
      for (level = 0; level < gNLevels; level++){
        
#ifdef GRAVITY
        err =  zlsg_InitLevelGravityField( level ); // precompute the potential of to the gravity field over the level
        if (err != noErr ) {
          zlsio_ErrPrintf( " ERROR: zlsg_InitLevelGravityField: Level Field Failed: %d,  %d \n",level, err );
          return(err);
        }
#endif GRAVITY
#ifdef NEQIONISATION
        err =  zlsi_InitLevelIonisationState( level ); 
        if (err != noErr ) {
          zlsio_ErrPrintf( " ERROR: zlsi_InitLevelIonisationState: Level Failed: %d,  %d \n",level, err );
          return(err);
        }
#endif NEQIONISATION
        
        err = zlsp_InitLevelZones( level ); // last: user's levelProblem, initialised top down
        
      }
      
      if (gNLevels>1){
        for (level = (gNLevels-1); level > 0; level--){
          err = zlsbc_FillSuperLevelFromLevel(level);
        }
      }
      
#if (ZLS_VERBOSE>0)
      zlsio_Printf("%s\n",  DIVIDERLINE );
#else
      zlsio_LogPrintf("%s\n",  DIVIDERLINE );
#endif
      
      ccl_release(&config);
      
    }// config  
    
  } else {
    err = memErr;
    zlsio_ErrPrintf( " ERROR: Unable to allocate level 0 ProblemPtr \n");
    return (err);
  }
  
  return (err);
  
}

//======================================================================
// Routine for 'User; boundary cells set in grid.conf not used here
// but often used for jet inlet models
//======================================================================
#ifdef ZLS_USER_BOUNDARIES
int  zlsp_CalcUserZoneBCs ( ZonePtr z, int direction, int bndryKind ){
  Integer err = noErr;
  ProblemPtr prob = NULL;
  PhysicsPtr phys = NULL;
  if ( z != NULL ){
    
    Integer level = z->zoneLevel;
    prob = gLevels[level].levelProblem;  // get the user's problem parameters
    phys = gLevels[level].levelPhysics;  // get the global eos parameters

    WHD zoneW = z->zoneW;    
    
    Integer   id;   
    QArr       q;

    ACoordPtr xx  = z->zoneGrid.x;
    ACoordPtr yx  = z->zoneGrid.y;
    
    Integer is    = z->zoneGrid.ns[kXDim];
    Integer ie    = z->zoneGrid.ne[kXDim];
    
    Integer js    = z->zoneGrid.ns[kYDim];
    Integer je    = z->zoneGrid.ne[kYDim];
    
    Integer nti   = z->zoneGrid.nt[kXDim];
    Integer ntj   = z->zoneGrid.nt[kYDim];
    
    Real dx  = xx[0].dx;
    Real dy  = yx[0].dx;
    
    Integer i, j;

    Integer  nbc  = z->zoneBCells;
   

	   Real t      = zlst_GlobalTime;
//     Real dt     = zlst_GlobalTimeStep;
    
 		for (id = 0; id <  z->nQ; id++){
      
			if (z->zoneData[id]!=NULL){
				
#ifdef ZONEDATA_ARE_PLAINBLOCKS
				zls_SetQuantity2DArr(&q, z->zoneData[id], ntj, nti);
#endif ZONEDATA_ARE_PLAINBLOCKS
				
#ifdef ZONEDATA_ARE_MULTIDARRAYS
				q = z->zoneData[id];
#endif ZONEDATA_ARE_MULTIDARRAYS
				
				if (q!=NULL) {
        
        } else { err = memErr; } // q
			} // z->zoneData[id]!=NULL
		} // var id loop

          
  } else { err = memErr; } // z
  
  return (err);
}
#endif ZLS_USER_BOUNDARIES


//======================================================================
// Routine for modifying anywhere in the grid in any way as part of
// a custom problem model. Not restricted to boundary cells, here used for 
// wind bubble/mass loss bubble energy injection and mass injection
// called after all other boundary conditions have been computed.
// Use sparingly, if not careful this routine can impact heavily on 
// the performance of the simulation.
//======================================================================
// 
// Normally deactivated (ie not even checked for calling) unless the macro
// ZLS_USER_UPDATE_ROUTINE is defined in zls_defs.  If not defined then
// this routine is not even callable - and this is the default setting.
// Also normally provided as an empty routine.
//

#ifdef ZLS_USER_UPDATE_ROUTINE

static Real z0P0  = 0.0;
static Real z0d0  = 0.0;
static Real zvx0  = 0.0;
static Real zvy0  = 0.0;
static Real z0E0  = 0.0;
static Real z0En0 = 0.0;
static Real z0H0  = 0.0;
static Real z0K0  = 0.0;
static Real z0Tot = 0.0;

int  zlsp_UserZoneUpdate ( ZonePtr z  ){
  
  Integer err = noErr;
  ProblemPtr prob = NULL;
  PhysicsPtr phys = NULL;
  
  // A different type of BC, not the grid edges as all but setting cell values
  // on the grid.  Uses zoneW and is based on the zone data init routine rather than
  // the grid edge code of the basic boundary value routines.
  
  if ( z != NULL ){
    
    Integer level = z->zoneLevel;
    Integer zid   = z->zoneID;
    prob = gLevels[level].levelProblem;  // get the user's problem parameters
    phys = gLevels[level].levelPhysics;  // get the global eos parameters

    Integer is    = z->zoneGrid.ns[kXDim];
    Integer js    = z->zoneGrid.ns[kYDim];
    Integer ie    = z->zoneGrid.ne[kXDim];
    Integer je    = z->zoneGrid.ne[kYDim];
    
    WHD zoneW     = z->zoneW; 
    
    
    if (( level == (gNLevels-1))&&( zid == 0)){
      
        z0d0  = zoneW.d [je-1][ie-1];
        z0P0  = zoneW.P [je-1][ie-1];
        Real g      = kZLS_eos_gamma;
        z0En0 = z0P0*zls_Pow(z0d0, (-g)); // prop to entropy P/d^g

        z0E0  = z0P0/(z0d0*(g-1.0));
        z0H0  = z0E0 + (z0P0/z0d0);
        zvx0  = zoneW.vx[je-1][ie-1];
        zvy0  = zoneW.vy[je-1][ie-1];
        z0K0  = 0.5*(zvx0*zvx0 + zvy0*zvy0);

        z0Tot = z0K0 + z0H0;
     }
    
    if (( level == (gNLevels-1))&&( zid == 2)){
      
      Real vz0 = sqrt(zvx0*zvx0 + zvy0*zvy0);
      vz0      = (vz0 < 1.e-8)?1.e-8:vz0;
      Real g  = kZLS_eos_gamma;
//
//---- 1st cell above corner
//    
      Real vx = zoneW.vx[js][is];
      Real vy = zoneW.vy[js][is];
      Real v0 = sqrt(vx*vx + vy*vy);
      v0 = (v0 < 1.e-8)?1.e-8:v0;
      
      Real P  = zoneW.P [js][is];
      Real d  = zls_Pow((P/z0En0), (1.0/g));
      
      Real E  = P/(d*(g-1.0));
      Real H  = E + (P/d);
      Real KE = z0Tot - H;
      KE      = (KE < 0.0)?0.0:KE;
      Real v  = sqrt(2.0*KE);
      Real beta = v/v0;
//    
//      zlsio_Printf("b0 = %g\n", beta);
     
      zoneW.d [js][is]  = d;
      zoneW.vx[js][is]   *= beta;     
      zoneW.vy[js][is]   *= beta;  
      
          
       vz0 = sqrt(zvx0*zvx0 + zvy0*zvy0);
      vz0      = (vz0 < 1.e-8)?1.e-8:vz0;
//
//---- second cell above corner
//    
       vx = zoneW.vx[js+1][is];
       vy = zoneW.vy[js+1][is];
       v0 = sqrt(vx*vx + vy*vy);
       v0 = (v0 < 1.e-8)?1.e-8:v0;
      
       P  = zoneW.P [js+1][is];
       d  = zls_Pow((P/z0En0), (1.0/g));
      
       E  = P/(d*(g-1.0));
       H  = E + (P/d);
       KE = z0Tot - H;
       KE      = (KE < 0.0)?0.0:KE;
       v  = sqrt(2.0*KE);
       beta = v/v0;
//    
//      zlsio_Printf("b1 = %g\n", beta);
     
      zoneW.d [js+1][is]  = d;
      zoneW.vx[js+1][is]   *= beta;     
      zoneW.vy[js+1][is]   *= beta;    

//
//---- second cell along from corner
//    
       vx = zoneW.vx[js][is+1];
       vy = zoneW.vy[js][is+1];
       v0 = sqrt(vx*vx + vy*vy);
       v0 = (v0 < 1.e-8)?1.e-8:v0;
      
       P  = zoneW.P [js][is+1];
       d  = zls_Pow((P/z0En0), (1.0/g));
      
       E  = P/(d*(g-1.0));
       H  = E + (P/d);
       KE = z0Tot - H;
       KE      = (KE < 0.0)?0.0:KE;
       v  = sqrt(2.0*KE);
       beta = v/v0;
//    
//      zlsio_Printf("b2 = %g\n", beta);
     
      zoneW.d [js][is+1]  = d;
      zoneW.vx[js][is+1]   *= beta;     
      zoneW.vy[js][is+1]   *= beta;    
//
//---- third cell along from corner
//    
       vx = zoneW.vx[js][is+2];
       vy = zoneW.vy[js][is+2];
       v0 = sqrt(vx*vx + vy*vy);
       v0 = (v0 < 1.e-8)?1.e-8:v0;
      
       P  = zoneW.P [js][is+2];
       d  = zls_Pow((P/z0En0), (1.0/g));
      
       E  = P/(d*(g-1.0));
       H  = E + (P/d);
       KE = z0Tot - H;
       KE      = (KE < 0.0)?0.0:KE;
       v  = sqrt(2.0*KE);
       beta = v/v0;
//    
//      zlsio_Printf("b3 = %g\n", beta);
     
      zoneW.d [js][is+2]  = d;
      zoneW.vx[js][is+2]   *= beta;     
      zoneW.vy[js][is+2]   *= beta;    
//
//----
//    

    } // level
    
  } else { err = memErr; } // z
  
  return (err);
}
#endif ZLS_USER_UPDATE_ROUTINE
